/*
* Copyright (C) fastman92 <fastman92@gmail.com>, website: http://fastman92.com
* Licensed under the MIT License, see LICENSE at top level directory.
*
*/

#include "FileIDlimit.h"

#include "DynamicLimits.h"
#include "InlinedCodeRemoving.h"
#include "IMGlimits.h"
#include "PlayerLimits.h"
#include "MapLimits.h"

#include "../Core/CGenericLogStorage.h"
#include "../Core/FormattedOutput.h"
#include "../Core/LimitAdjuster.h"
#include "../Core/ProjectConstants.h"

#include "../GameStructures/Rockstar Games/CBaseModelInfo.h"
#include "../GameStructures/Rockstar Games/CColModel.h"
#include "../GameStructures/Rockstar Games/CGenericGameStorage.h"
#include "../GameStructures/Rockstar Games/CKeyGen.h"

#ifdef IS_PLATFORM_WIN_X86
#include "../GameStructures/Rockstar Games/eModelID.h"
#endif

#include "ModelSpecialFeatures.h"
#include "SaveOfVariableLength.h"

#include <Array\countof.h>
#include <stdint.h>

using namespace Game;

FileIDlimit g_fileIDlimits;

int32_t GetFullModelIDfrom16bitRegisterBefore_32_bit_value(int32_t modelID)
{
	return modelID;
}

int32_t GetFullModelIDfrom16bitRegisterBefore_16_bit_value(int32_t modelID)
{
	uint16_t uID = *(uint16_t*)&modelID;

	if (uID > MAX_FILE_ID_UINT16)	// is it negative ID?
		return *(int16_t*)&modelID;
	else
		return uID;
}

int32_t(__cdecl* GetFullModelIDfrom16bitRegisterBefore)(int32_t modelID) = GetFullModelIDfrom16bitRegisterBefore_16_bit_value;

// Constructor
CIDHashContainerDelayedSorting::CIDHashContainerDelayedSorting()
{
	this->bIsArraySorted = true;
}

// Allocate element
void CIDHashContainerDelayedSorting::AllocateElement(int32_t ID, uint32_t hash)
{
	tIDhashPair pair;
	pair.ID = ID;
	pair.hash = hash;
	this->array.push_back(pair);

	this->bIsArraySorted = false;

	// CGenericLogStorage::SaveFormattedTextLn("Adding ID: %d", pair.ID);
	// CGenericLogStorage::SaveWorkBuffer();

	//if (ID == 344)
		//printf_MessageBox("this");
}

// Updates element
void CIDHashContainerDelayedSorting::UpdateElement(int32_t ID, uint32_t oldHash, uint32_t newHash)
{
	auto it = this->GetElementItByID(ID);

	if (oldHash == it->hash)
		return;

	this->array.erase(it);

	tIDhashPair pair;
	pair.ID = ID;
	pair.hash = newHash;
	this->array.push_back(pair);
	this->bIsArraySorted = false;


	// CGenericLogStorage::SaveFormattedTextLn("Updating ID: %d", pair.ID);
	// CGenericLogStorage::SaveWorkBuffer();
}

bool HashPairsCompareHash(const tIDhashPair& a, const tIDhashPair& b)
{
	return a.hash < b.hash;
}

// Returns ID by hash
int32_t CIDHashContainerDelayedSorting::GetIDbyHashBinarySearch(uint32_t hash)
{
	// Sort the array if neccessary
	this->SortTheArray();

	// Keep halving the search space until we reach the end of the vector
	auto begin = this->array.begin();
	auto end = this->array.end();

	while (begin < end)
	{
		// Find the median value between the iterators
		const auto Middle = begin + (std::distance(begin, end) / 2);

		// Re-adjust the iterators based on the median value
		if (Middle->hash == hash)
			return Middle->ID;
		else if (Middle->hash > hash)
			end = Middle;
		else
			begin = Middle + 1;
	}

	return UNDEFINED_FILE_ID_INT32;
}

// Returns ID by hash
int32_t CIDHashContainerDelayedSorting::GetIDbyHashLinearSearch(uint32_t hash)
{
	for (auto it = this->array.begin(); it != this->array.end(); ++it)
		if (it->hash == hash)
			return it->ID;

	return UNDEFINED_FILE_ID_INT32;
}

// Sorts the array
void CIDHashContainerDelayedSorting::SortTheArray()
{
	if (!this->bIsArraySorted)
	{
		std::sort(this->array.begin(), this->array.end(), &HashPairsCompareHash);	
		// CGenericLogStorage::SaveTextLn("Sorting the array!");
		// CGenericLogStorage::SaveWorkBuffer();
		this->bIsArraySorted = true;
	}
}

// Clears the array
void CIDHashContainerDelayedSorting::Clear()
{
	this->array.clear();
}

// Returns element iterator by ID
CIDHashContainerDelayedSorting::tEntryContainer::iterator CIDHashContainerDelayedSorting::GetElementItByID(int32_t ID)
{
	for (auto it = this->array.begin(); it != this->array.end(); ++it)
		if (it->ID == ID)
			return it;

	return this->array.end();
}

///////////

// Allocate element
void CIDHashContainerAlwaysSorted::AllocateElement(int32_t ID, uint32_t hash)
{
	this->array[hash] = ID;
}

// Updates element
void CIDHashContainerAlwaysSorted::UpdateElement(int32_t ID, uint32_t oldHash, uint32_t newHash)
{
	this->array[oldHash] = newHash;
}

// Updates element
void CIDHashContainerAlwaysSorted::RemoveElement(int32_t ID, uint32_t oldHash)
{
	auto oldHashIt = this->array.find(oldHash);

	if (oldHashIt != this->array.end() && oldHashIt->second == ID)
		this->array.erase(oldHashIt);
	else
	{
		for (auto it = this->array.begin(); it != this->array.end(); ++it)
			if (it->second == ID)
			{
				this->array.erase(it);
				return;
			}
	}
}

// Returns ID by hash
int32_t CIDHashContainerAlwaysSorted::GetIDbyHashBinarySearch(uint32_t hash)
{
	auto it = this->array.find(hash);

	return it != this->array.end() ? it->second : UNDEFINED_FILE_ID_INT32;

}

// Clears the array
void CIDHashContainerAlwaysSorted::Clear()
{
	this->array.clear();
}

#define MAKE_FILE_TYPE_PAIR(member, generalFileType) { member, #member, generalFileType },

const tFileTypeDescriptor CFileTypeInfo::fileDescriptorArray[] =
{
	MAKE_FILE_TYPE_LIST(MAKE_FILE_TYPE_PAIR)
};

/* Finds enum member by name */
eFileType CFileTypeInfo::GetEnumMemberByName(const char* name)
{
	for (unsigned int i = 0; i != _countof(fileDescriptorArray); i++)
		if (!strcmp(name, fileDescriptorArray[i].fileTypeName))
			return fileDescriptorArray[i].fileType;

	return FILE_TYPE_UNDEFINED;
}

/* Finds name by enum member */
const char* CFileTypeInfo::GetEnumNameByMember(eFileType member)
{
	const tFileTypeDescriptor* ptr = GetFileInfoDescriptorByMember(member);

	return ptr ? ptr->fileTypeName : NULL;
}

/* Finds general file type descriptor by enum member */
const tFileTypeDescriptor* CFileTypeInfo::GetFileInfoDescriptorByMember(eFileType member)
{
	for (unsigned int i = 0; i != _countof(fileDescriptorArray); i++)
		if (fileDescriptorArray[i].fileType == member)
			return &fileDescriptorArray[i];

	return NULL;
}

// Returns array index
int FileIDlimit::GetFileTypeArrayIndex(eFileType type)
{
	for (int i = 0; i < countOfTypes; i++)
		if (fileTypeArray[i].fileType == type)
			return i;

	return -1;
}

// Returns pointer to file type info
const tFileTypeInfo* details::FileIDlimit_base1::GetFileTypeInfo(eFileType type) const
{
	for (int i = 0; i < countOfTypes; i++)
		if (fileTypeArray[i].fileType == type)
			return &fileTypeArray[i];

	return nullptr;
}

// Returns pointer to file type info
const tFileTypeInfo* details::FileIDlimit_base1::GetFileTypeInfo(eFileTypeGeneral type) const
{
	for (int i = 0; i < countOfTypes; i++)
		if (fileTypeArray[i].pFileTypeDescriptor->generalFileType == type)
			return &fileTypeArray[i];

	return nullptr;
}

// Registers file type
void FileIDlimit::RegisterFileType(eFileType fileType, int defaultCount, tFileTypeLimit maxLimitValue)
{
	int index = this->countOfTypes++;

	tFileTypeInfo* pTypeInfo = &this->fileTypeArray[index];
	
	pTypeInfo->maxLimitValue = maxLimitValue;
	pTypeInfo->fileType = fileType;
	pTypeInfo->pFileTypeDescriptor = CFileTypeInfo::GetFileInfoDescriptorByMember(fileType);

	pTypeInfo->currentBaseID = 0;
	pTypeInfo->defaultBaseID = 0;
	pTypeInfo->defaultLimitValue = defaultCount;
	pTypeInfo->currentLimitValue = defaultCount;
}

// Returns maximum number of file IDs
uint32_t FileIDlimit::GetMaxNumberOfFileIDs() const
{
	return this->maxCountOfIDs;
}

// Returns true, if there is any ID limit that can be modified
bool FileIDlimit::IsThereAnyIDlimitThatCanBeModified() const
{
	for (int i = 0; i < countOfTypes; i++)
		if (fileTypeArray[i].maxLimitValue > 0)
			return true;

	return false;
}

// Updates base IDs, needs to be executed 
void FileIDlimit::UpdateBaseIDs()
{
	int curBaseID = 0;
	int defBaseID = 0;

	this->fileTypeArray[0].currentBaseID = 0;
	this->fileTypeArray[0].defaultBaseID = 0;

	for (int i = 0; i < countOfTypes; ++i)
	{
		tFileTypeBaseID currentLimitValue = this->fileTypeArray[i].currentLimitValue;
		tFileTypeBaseID defaultLimitValue = this->fileTypeArray[i].defaultLimitValue;

		this->fileTypeArray[i + 1].currentBaseID = curBaseID + currentLimitValue;
		this->fileTypeArray[i + 1].defaultBaseID = defBaseID + defaultLimitValue;

		curBaseID += currentLimitValue;
		defBaseID += defaultLimitValue;
	};
}

// Enables/disables ID limits patch.
void FileIDlimit::EnableIDlimitsPatch(bool bEnabled)
{
	if (this->countOfTypes > 0)
		bIDlimitsPatchEnabled = bEnabled;
}

// Returns true if ID limits are known
bool FileIDlimit::AreIDlimitsKnown() const
{
	return this->GetCountOfTypes() > 0;
}

// Returns true if ID limits patch is enabled
bool FileIDlimit::IsIDlimitsPatchEnabled() const
{
	return this->bIDlimitsPatchEnabled;
}

// Writes to log if limits are changed
bool FileIDlimit::WriteToLogIfLimitsAreChanged(bool bModelSpecialFeaturesImplemented)
{
	// if debug, set ID limits to default values 
	if (!CPatch::IsDebugModeActive())
	{
		if (IsAnyIDlimitChanged())
			CGenericLogStorage::SaveTextLn("At least one ID limit is changed. ID limits will be patched.");
		else if(bModelSpecialFeaturesImplemented && g_modelSpecialFeatures.AreSpecialFeaturesEnabled())
			CGenericLogStorage::SaveTextLn("Model special features are enabled. Because of this, the ID limits will be patched.");
		else
		{
			CGenericLogStorage::SaveTextLn("None of ID limits is changed. Nothing to do.");
			CGenericLogStorage::WriteLineSeparator();
			return false;
		}
	}
	else
	{
		SetIDlimitsToDefaultValues();
		CGenericLogStorage::SaveTextLn("ID limits working in debug mode.");
	}

	return true;
}

// Returns true if any ID limit is increased.
bool FileIDlimit::IsAnyIDlimitIncreased() const
{
	for (int i = 0; i < countOfTypes; ++i)
	{
		const tFileTypeInfo* pFileInfo = &this->fileTypeArray[i];

		if (pFileInfo->currentLimitValue > pFileInfo->defaultLimitValue)
			return true;
	}

	return false;
}

// Returns true if any ID limit is changed.
bool FileIDlimit::IsAnyIDlimitChanged() const
{
	for (int i = 0; i < countOfTypes; ++i)
	{
		const tFileTypeInfo* pFileInfo = &this->fileTypeArray[i];

		if (pFileInfo->currentLimitValue != pFileInfo->defaultLimitValue)
			return true;
	}

	return false;
}

// Sets ID limits to default values.
void FileIDlimit::SetIDlimitsToDefaultValues()
{
	for (int i = 0; i <= countOfTypes; ++i)
	{
		tFileTypeInfo* pFileInfo = &this->fileTypeArray[i];
		
		pFileInfo->currentBaseID = pFileInfo->defaultBaseID;
		pFileInfo->currentLimitValue = pFileInfo->defaultLimitValue;
	}
}

// Returns file type info by index
tFileTypeInfo* FileIDlimit::GetFileTypeInfoByIndex(int index)
{
	return &this->fileTypeArray[index];
}

// GTA III
namespace Game_GTAIII
{
	// Initializes CStreamingInfo
	void CStreamingInfo::Init()
	{
		this->uiLoadFlag = 0;
		this->usNext = 0;
		this->usPrev = 0;
		this->usNextOnCd = -1;
		this->iBlockCount = 0;
		this->iBlockOffset = 0;
	}

	void __cdecl patch_CStreaming__Init_doIt()
	{
		tFileTypeBaseID count = g_fileIDlimits.GetCountOfAllFileIDs();

		for (tFileTypeBaseID ID = 0; ID < count; ID++)
			g_fileIDlimits.CStreaming__ms_aInfoForModel.gta_iii[ID].Init();
	}

#ifdef IS_PLATFORM_WIN_X86
	void NAKED patch_CStreaming__Init_406440()
	{
		__asm
		{
			call patch_CStreaming__Init_doIt;

			// go back
			mov eax, 0x406678;
			jmp eax;
		}
	}
#endif
}

#ifdef IS_PLATFORM_WIN_X86
// Patches ID limits if neccessary.
void FileIDlimit::PatchFileIDlimits_GTA_III_1_0()
{
	using namespace Game_GTAIII;

	if(!WriteToLogIfLimitsAreChanged())
		return;
	
	// ID limit checks
	#if TRUE

	// See if count is unsigned
	if (GetCountOfAllFileIDs() >= 32768)
		throw f92_runtime_error("Count of file IDs should be less than 32768."
			"\nUnsigned ID for GTA III is not supported yet."
			"\nCount of IDs: %d", GetCountOfAllFileIDs()
		);
	#endif

	CGenericLogStorage::SaveFormattedTextLn(
		"\nNew ID limits:\n"
		"%d - %d (%d) - DFF models defined within IDE files\n"
		"%d - %d (%d) - TXD texture archives.\n"
		"%d - count of all file IDs\n"
		,

		GetBaseID(FILE_TYPE_DFF), GetLastIDforType(FILE_TYPE_DFF), GetFileIDcurrentLimit(FILE_TYPE_DFF),
		GetBaseID(FILE_TYPE_TXD), GetLastIDforType(FILE_TYPE_TXD), GetFileIDcurrentLimit(FILE_TYPE_TXD),

		GetCountOfAllFileIDs()
		);

	// Allocate variables
	#if TRUE
	if (!CPatch::IsDebugModeActive())
	{
		// CModelInfo::ms_modelInfoPtrs
		if(GetFileIDcurrentLimit(FILE_TYPE_DFF) > GetFileIDdefaultLimit(FILE_TYPE_DFF))
		{
			CModelInfo__ms_modelInfoPtrs.gta_iii = new CBaseModelInfo*[GetFileIDcurrentLimit(FILE_TYPE_DFF)];

			memset(
				CModelInfo__ms_modelInfoPtrs.gta_iii,
				NULL,
				GetFileIDcurrentLimit(FILE_TYPE_DFF) * sizeof(CBaseModelInfo*)
			);

			CModelInfo__ms_modelInfoPtrs.bIsAllocated.Set(true);
		}

		// CStreaming::ms_aInfoForModel
		if(FileIDlimit::GetCountOfAllFileIDs() > GetDefaultCountOfAllFileIDs())
		{
			CStreaming__ms_aInfoForModel.gta_iii = new CStreamingInfo[GetCountOfAllFileIDs()];
			memset(
				CStreaming__ms_aInfoForModel.gta_iii,
				NULL,
				GetCountOfAllFileIDs() * sizeof(CStreamingInfo)
			);

			CStreaming__ms_aInfoForModel.bIsAllocated.Set(true);
		}
	}
	#endif

	// Patch pointers to CStreaming::ms_aInfoForModel
	#if TRUE
	{
		CPatch::PatchPointer(0x406442 + 2, &CStreaming__ms_aInfoForModel.gta_iii->uiLoadFlag);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[edi], 0
		CPatch::PatchPointer(0x40644A + 2, CStreaming__ms_aInfoForModel.gta_iii);	// mov     dword ptr ds:_ZN10CStreaming16ms_aInfoForModelE.usNext[edi], 0; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x406454 + 2, &CStreaming__ms_aInfoForModel.gta_iii->usPrev);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usPrev[edi], 0
		CPatch::PatchPointer(0x40645E + 3, &CStreaming__ms_aInfoForModel.gta_iii->usNextOnCd);	// or      ds:_ZN10CStreaming16ms_aInfoForModelE.usNextOnCd[edi], 0FFFFh
		CPatch::PatchPointer(0x406466 + 2, &CStreaming__ms_aInfoForModel.gta_iii->iBlockCount);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.iBlockCount[edi], 0
		CPatch::PatchPointer(0x406470 + 2, &CStreaming__ms_aInfoForModel.gta_iii->iBlockOffset);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.iBlockOffset[edi], 0
		CPatch::PatchPointer(0x40648A + 2, &CStreaming__ms_aInfoForModel.gta_iii->uiLoadFlag);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[ebx], 0
		CPatch::PatchPointer(0x406491 + 2, CStreaming__ms_aInfoForModel.gta_iii);	// mov     dword ptr ds:_ZN10CStreaming16ms_aInfoForModelE.usNext[ebx], 0; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40649B + 2, &CStreaming__ms_aInfoForModel.gta_iii->usPrev);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usPrev[ebx], 0
		CPatch::PatchPointer(0x4064A5 + 3, &CStreaming__ms_aInfoForModel.gta_iii->usNextOnCd);	// or      ds:_ZN10CStreaming16ms_aInfoForModelE.usNextOnCd[ebx], 0FFFFh
		CPatch::PatchPointer(0x4064AD + 2, &CStreaming__ms_aInfoForModel.gta_iii->iBlockCount);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.iBlockCount[ebx], 0
		CPatch::PatchPointer(0x4064B7 + 2, &CStreaming__ms_aInfoForModel.gta_iii->iBlockOffset);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.iBlockOffset[ebx], 0
		CPatch::PatchPointer(0x4064CF + 2, &CStreaming__ms_aInfoForModel.gta_iii->uiLoadFlag);	// mov     ss:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[ebp], 0
		CPatch::PatchPointer(0x4064D6 + 2, CStreaming__ms_aInfoForModel.gta_iii);	// mov     dword ptr ss:_ZN10CStreaming16ms_aInfoForModelE.usNext[ebp], 0; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x4064E0 + 2, &CStreaming__ms_aInfoForModel.gta_iii->usPrev);	// mov     ss:_ZN10CStreaming16ms_aInfoForModelE.usPrev[ebp], 0
		CPatch::PatchPointer(0x4064EA + 3, &CStreaming__ms_aInfoForModel.gta_iii->usNextOnCd);	// or      ss:_ZN10CStreaming16ms_aInfoForModelE.usNextOnCd[ebp], 0FFFFh
		CPatch::PatchPointer(0x4064F2 + 2, &CStreaming__ms_aInfoForModel.gta_iii->iBlockCount);	// mov     ss:_ZN10CStreaming16ms_aInfoForModelE.iBlockCount[ebp], 0
		CPatch::PatchPointer(0x4064FC + 2, &CStreaming__ms_aInfoForModel.gta_iii->iBlockOffset);	// mov     ss:_ZN10CStreaming16ms_aInfoForModelE.iBlockOffset[ebp], 0
		CPatch::PatchPointer(0x406513 + 2, &CStreaming__ms_aInfoForModel.gta_iii->uiLoadFlag);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[esi], 0
		CPatch::PatchPointer(0x40651A + 2, CStreaming__ms_aInfoForModel.gta_iii);	// mov     dword ptr ds:_ZN10CStreaming16ms_aInfoForModelE.usNext[esi], 0; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x406524 + 2, &CStreaming__ms_aInfoForModel.gta_iii->usPrev);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usPrev[esi], 0
		CPatch::PatchPointer(0x40652E + 3, &CStreaming__ms_aInfoForModel.gta_iii->usNextOnCd);	// or      ds:_ZN10CStreaming16ms_aInfoForModelE.usNextOnCd[esi], 0FFFFh
		CPatch::PatchPointer(0x406536 + 2, &CStreaming__ms_aInfoForModel.gta_iii->iBlockCount);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.iBlockCount[esi], 0
		CPatch::PatchPointer(0x406540 + 2, &CStreaming__ms_aInfoForModel.gta_iii->iBlockOffset);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.iBlockOffset[esi], 0
		CPatch::PatchPointer(0x406557 + 2, &CStreaming__ms_aInfoForModel.gta_iii->uiLoadFlag);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[edx], 0
		CPatch::PatchPointer(0x40655E + 2, CStreaming__ms_aInfoForModel.gta_iii);	// mov     dword ptr ds:_ZN10CStreaming16ms_aInfoForModelE.usNext[edx], 0; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x406568 + 2, &CStreaming__ms_aInfoForModel.gta_iii->usPrev);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usPrev[edx], 0
		CPatch::PatchPointer(0x406572 + 3, &CStreaming__ms_aInfoForModel.gta_iii->usNextOnCd);	// or      ds:_ZN10CStreaming16ms_aInfoForModelE.usNextOnCd[edx], 0FFFFh
		CPatch::PatchPointer(0x40657A + 2, &CStreaming__ms_aInfoForModel.gta_iii->iBlockCount);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.iBlockCount[edx], 0
		CPatch::PatchPointer(0x406584 + 2, &CStreaming__ms_aInfoForModel.gta_iii->iBlockOffset);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.iBlockOffset[edx], 0
		CPatch::PatchPointer(0x40659B + 2, &CStreaming__ms_aInfoForModel.gta_iii->uiLoadFlag);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[ecx], 0
		CPatch::PatchPointer(0x4065A2 + 2, CStreaming__ms_aInfoForModel.gta_iii);	// mov     dword ptr ds:_ZN10CStreaming16ms_aInfoForModelE.usNext[ecx], 0; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x4065AC + 2, &CStreaming__ms_aInfoForModel.gta_iii->usPrev);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usPrev[ecx], 0
		CPatch::PatchPointer(0x4065B6 + 3, &CStreaming__ms_aInfoForModel.gta_iii->usNextOnCd);	// or      ds:_ZN10CStreaming16ms_aInfoForModelE.usNextOnCd[ecx], 0FFFFh
		CPatch::PatchPointer(0x4065BE + 2, &CStreaming__ms_aInfoForModel.gta_iii->iBlockCount);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.iBlockCount[ecx], 0
		CPatch::PatchPointer(0x4065C8 + 2, &CStreaming__ms_aInfoForModel.gta_iii->iBlockOffset);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.iBlockOffset[ecx], 0
		CPatch::PatchPointer(0x4065DF + 2, &CStreaming__ms_aInfoForModel.gta_iii->uiLoadFlag);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[ebx], 0
		CPatch::PatchPointer(0x4065E6 + 2, CStreaming__ms_aInfoForModel.gta_iii);	// mov     dword ptr ds:_ZN10CStreaming16ms_aInfoForModelE.usNext[ebx], 0; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x4065F0 + 2, &CStreaming__ms_aInfoForModel.gta_iii->usPrev);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usPrev[ebx], 0
		CPatch::PatchPointer(0x4065FA + 3, &CStreaming__ms_aInfoForModel.gta_iii->usNextOnCd);	// or      ds:_ZN10CStreaming16ms_aInfoForModelE.usNextOnCd[ebx], 0FFFFh
		CPatch::PatchPointer(0x406602 + 2, &CStreaming__ms_aInfoForModel.gta_iii->iBlockCount);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.iBlockCount[ebx], 0
		CPatch::PatchPointer(0x40660C + 2, &CStreaming__ms_aInfoForModel.gta_iii->iBlockOffset);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.iBlockOffset[ebx], 0
		CPatch::PatchPointer(0x406627 + 2, &CStreaming__ms_aInfoForModel.gta_iii->uiLoadFlag);	// mov     ss:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[ebp], 0
		CPatch::PatchPointer(0x40662E + 2, CStreaming__ms_aInfoForModel.gta_iii);	// mov     dword ptr ss:_ZN10CStreaming16ms_aInfoForModelE.usNext[ebp], 0; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x406638 + 2, &CStreaming__ms_aInfoForModel.gta_iii->usPrev);	// mov     ss:_ZN10CStreaming16ms_aInfoForModelE.usPrev[ebp], 0
		CPatch::PatchPointer(0x406642 + 3, &CStreaming__ms_aInfoForModel.gta_iii->usNextOnCd);	// or      ss:_ZN10CStreaming16ms_aInfoForModelE.usNextOnCd[ebp], 0FFFFh
		CPatch::PatchPointer(0x40664F + 2, &CStreaming__ms_aInfoForModel.gta_iii->iBlockCount);	// mov     ss:_ZN10CStreaming16ms_aInfoForModelE.iBlockCount[ebp], 0
		CPatch::PatchPointer(0x406659 + 2, &CStreaming__ms_aInfoForModel.gta_iii->iBlockOffset);	// mov     ss:_ZN10CStreaming16ms_aInfoForModelE.iBlockOffset[ebp], 0
		CPatch::PatchPointer(0x40667F + 2, &CStreaming__ms_aInfoForModel.gta_iii->uiLoadFlag);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[esi], 0
		CPatch::PatchPointer(0x406686 + 2, CStreaming__ms_aInfoForModel.gta_iii);	// mov     dword ptr ds:_ZN10CStreaming16ms_aInfoForModelE.usNext[esi], 0; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x406690 + 2, &CStreaming__ms_aInfoForModel.gta_iii->usPrev);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usPrev[esi], 0
		CPatch::PatchPointer(0x40669A + 3, &CStreaming__ms_aInfoForModel.gta_iii->usNextOnCd);	// or      ds:_ZN10CStreaming16ms_aInfoForModelE.usNextOnCd[esi], 0FFFFh
		CPatch::PatchPointer(0x4066A2 + 2, &CStreaming__ms_aInfoForModel.gta_iii->iBlockCount);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.iBlockCount[esi], 0
		CPatch::PatchPointer(0x4066AC + 2, &CStreaming__ms_aInfoForModel.gta_iii->iBlockOffset);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.iBlockOffset[esi], 0
		CPatch::PatchPointer(0x4066D4 + 2, &CStreaming__ms_aInfoForModel.gta_iii->uiLoadFlag);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[edx], 0
		CPatch::PatchPointer(0x4066DB + 2, CStreaming__ms_aInfoForModel.gta_iii);	// mov     dword ptr ds:_ZN10CStreaming16ms_aInfoForModelE.usNext[edx], 0; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x4066E5 + 2, &CStreaming__ms_aInfoForModel.gta_iii->usPrev);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usPrev[edx], 0
		CPatch::PatchPointer(0x4066EF + 3, &CStreaming__ms_aInfoForModel.gta_iii->usNextOnCd);	// or      ds:_ZN10CStreaming16ms_aInfoForModelE.usNextOnCd[edx], 0FFFFh
		CPatch::PatchPointer(0x4066F7 + 2, &CStreaming__ms_aInfoForModel.gta_iii->iBlockCount);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.iBlockCount[edx], 0
		CPatch::PatchPointer(0x406701 + 2, &CStreaming__ms_aInfoForModel.gta_iii->iBlockOffset);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.iBlockOffset[edx], 0
		CPatch::PatchPointer(0x406736 + 2, &CStreaming__ms_aInfoForModel.gta_iii->uiLoadFlag);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[ecx], 0
		CPatch::PatchPointer(0x40673D + 2, CStreaming__ms_aInfoForModel.gta_iii);	// mov     dword ptr ds:_ZN10CStreaming16ms_aInfoForModelE.usNext[ecx], 0; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x406747 + 2, &CStreaming__ms_aInfoForModel.gta_iii->usPrev);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usPrev[ecx], 0
		CPatch::PatchPointer(0x406751 + 3, &CStreaming__ms_aInfoForModel.gta_iii->usNextOnCd);	// or      ds:_ZN10CStreaming16ms_aInfoForModelE.usNextOnCd[ecx], 0FFFFh
		CPatch::PatchPointer(0x406759 + 2, &CStreaming__ms_aInfoForModel.gta_iii->iBlockCount);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.iBlockCount[ecx], 0
		CPatch::PatchPointer(0x406763 + 2, &CStreaming__ms_aInfoForModel.gta_iii->iBlockOffset);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.iBlockOffset[ecx], 0
		CPatch::PatchPointer(0x406798 + 2, &CStreaming__ms_aInfoForModel.gta_iii->uiLoadFlag);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[ebx], 0
		CPatch::PatchPointer(0x40679F + 2, CStreaming__ms_aInfoForModel.gta_iii);	// mov     dword ptr ds:_ZN10CStreaming16ms_aInfoForModelE.usNext[ebx], 0; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x4067A9 + 2, &CStreaming__ms_aInfoForModel.gta_iii->usPrev);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usPrev[ebx], 0
		CPatch::PatchPointer(0x4067B3 + 3, &CStreaming__ms_aInfoForModel.gta_iii->usNextOnCd);	// or      ds:_ZN10CStreaming16ms_aInfoForModelE.usNextOnCd[ebx], 0FFFFh
		CPatch::PatchPointer(0x4067BB + 2, &CStreaming__ms_aInfoForModel.gta_iii->iBlockCount);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.iBlockCount[ebx], 0
		CPatch::PatchPointer(0x4067C5 + 2, &CStreaming__ms_aInfoForModel.gta_iii->iBlockOffset);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.iBlockOffset[ebx], 0
		CPatch::PatchPointer(0x4067F3 + 2, &CStreaming__ms_aInfoForModel.gta_iii->uiLoadFlag);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[edi], 0
		CPatch::PatchPointer(0x4067FA + 2, CStreaming__ms_aInfoForModel.gta_iii);	// mov     dword ptr ds:_ZN10CStreaming16ms_aInfoForModelE.usNext[edi], 0; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x406804 + 2, &CStreaming__ms_aInfoForModel.gta_iii->usPrev);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usPrev[edi], 0
		CPatch::PatchPointer(0x40680E + 3, &CStreaming__ms_aInfoForModel.gta_iii->usNextOnCd);	// or      ds:_ZN10CStreaming16ms_aInfoForModelE.usNextOnCd[edi], 0FFFFh
		CPatch::PatchPointer(0x406816 + 2, &CStreaming__ms_aInfoForModel.gta_iii->iBlockCount);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.iBlockCount[edi], 0
		CPatch::PatchPointer(0x406820 + 2, &CStreaming__ms_aInfoForModel.gta_iii->iBlockOffset);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.iBlockOffset[edi], 0
		CPatch::PatchPointer(0x406845 + 2, &CStreaming__ms_aInfoForModel.gta_iii->uiLoadFlag);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[eax], 0
		CPatch::PatchPointer(0x40684C + 2, CStreaming__ms_aInfoForModel.gta_iii);	// mov     dword ptr ds:_ZN10CStreaming16ms_aInfoForModelE.usNext[eax], 0; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x406856 + 2, &CStreaming__ms_aInfoForModel.gta_iii->usPrev);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usPrev[eax], 0
		CPatch::PatchPointer(0x406860 + 3, &CStreaming__ms_aInfoForModel.gta_iii->usNextOnCd);	// or      ds:_ZN10CStreaming16ms_aInfoForModelE.usNextOnCd[eax], 0FFFFh
		CPatch::PatchPointer(0x406868 + 2, &CStreaming__ms_aInfoForModel.gta_iii->iBlockCount);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.iBlockCount[eax], 0
		CPatch::PatchPointer(0x406872 + 2, &CStreaming__ms_aInfoForModel.gta_iii->iBlockOffset);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.iBlockOffset[eax], 0
		CPatch::PatchPointer(0x4068F8 + 2, &CStreaming__ms_aInfoForModel.gta_iii->uiLoadFlag);	// mov     ss:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[ebp], 1
		CPatch::PatchPointer(0x4068FF + 2, &CStreaming__ms_aInfoForModel.gta_iii->ucFlags);	// mov     ss:_ZN10CStreaming16ms_aInfoForModelE.ucFlags[ebp], 1
		CPatch::PatchPointer(0x406969 + 2, &CStreaming__ms_aInfoForModel.gta_iii[GetBaseID(FILE_TYPE_TXD)].uiLoadFlag);	// mov     ds:(_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+1ADB0h)[edx], 1
		CPatch::PatchPointer(0x406EA9 + 2, CStreaming__ms_aInfoForModel.gta_iii);	// add     ecx, offset _ZN10CStreaming16ms_aInfoForModelE; this
		CPatch::PatchPointer(0x406EF0 + 2, CStreaming__ms_aInfoForModel.gta_iii);	// add     ecx, offset _ZN10CStreaming16ms_aInfoForModelE; this
		CPatch::PatchPointer(0x406F09 + 4, &CStreaming__ms_aInfoForModel.gta_iii->usNextOnCd);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usNextOnCd[ebx*4], ax
		CPatch::PatchPointer(0x406FD1 + 2, CStreaming__ms_aInfoForModel.gta_iii);	// add     edi, offset _ZN10CStreaming16ms_aInfoForModelE; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x407023 + 4, &CStreaming__ms_aInfoForModel.gta_iii->usNextOnCd);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usNextOnCd[ebx*4], ax
		CPatch::PatchPointer(0x4077AD + 2, CStreaming__ms_aInfoForModel.gta_iii);	// sub     edx, offset _ZN10CStreaming16ms_aInfoForModelE; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x407E88 + 2, CStreaming__ms_aInfoForModel.gta_iii);	// add     ecx, offset _ZN10CStreaming16ms_aInfoForModelE; this
		CPatch::PatchPointer(0x407EB7 + 2, &CStreaming__ms_aInfoForModel.gta_iii->uiLoadFlag);	// mov     al, ss:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[ebp]
		CPatch::PatchPointer(0x407EC8 + 3, &CStreaming__ms_aInfoForModel.gta_iii->ucFlags);	// movzx   eax, ss:_ZN10CStreaming16ms_aInfoForModelE.ucFlags[ebp]
		CPatch::PatchPointer(0x407EDA + 2, &CStreaming__ms_aInfoForModel.gta_iii->ucFlags);	// or      ss:_ZN10CStreaming16ms_aInfoForModelE.ucFlags[ebp], 8
		CPatch::PatchPointer(0x407EEA + 2, &CStreaming__ms_aInfoForModel.gta_iii->ucFlags);	// or      ss:_ZN10CStreaming16ms_aInfoForModelE.ucFlags[ebp], bl
		CPatch::PatchPointer(0x407EF0 + 2, &CStreaming__ms_aInfoForModel.gta_iii->uiLoadFlag);	// mov     al, ss:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[ebp]
		CPatch::PatchPointer(0x407EFA + 3, &CStreaming__ms_aInfoForModel.gta_iii->ucFlags);	// movzx   eax, ss:_ZN10CStreaming16ms_aInfoForModelE.ucFlags[ebp]
		CPatch::PatchPointer(0x407F26 + 2, CStreaming__ms_aInfoForModel.gta_iii);	// cmp     dword ptr ss:_ZN10CStreaming16ms_aInfoForModelE.usNext[ebp], 0; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x407F2F + 1, CStreaming__ms_aInfoForModel.gta_iii);	// mov     ebx, offset _ZN10CStreaming16ms_aInfoForModelE; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x407F3D + 3, &CStreaming__ms_aInfoForModel.gta_iii->ucFlags);	// movzx   eax, ss:_ZN10CStreaming16ms_aInfoForModelE.ucFlags[ebp]
		CPatch::PatchPointer(0x407F95 + 1, CStreaming__ms_aInfoForModel.gta_iii);	// mov     ecx, offset _ZN10CStreaming16ms_aInfoForModelE; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x407FB9 + 2, &CStreaming__ms_aInfoForModel.gta_iii->uiLoadFlag);	// mov     ss:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[ebp], 2
		CPatch::PatchPointer(0x407FC0 + 2, &CStreaming__ms_aInfoForModel.gta_iii->ucFlags);	// mov     ss:_ZN10CStreaming16ms_aInfoForModelE.ucFlags[ebp], bl
		CPatch::PatchPointer(0x4082A1 + 3, &CStreaming__ms_aInfoForModel.gta_iii->uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[eax*4], 1
		CPatch::PatchPointer(0x4082B3 + 3, &CStreaming__ms_aInfoForModel.gta_iii->uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[eax*4], 1
		CPatch::PatchPointer(0x4082CE + 3, &CStreaming__ms_aInfoForModel.gta_iii->uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[eax*4], 1
		CPatch::PatchPointer(0x4082E0 + 3, &CStreaming__ms_aInfoForModel.gta_iii->uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[eax*4], 1
		CPatch::PatchPointer(0x4082FD + 3, &CStreaming__ms_aInfoForModel.gta_iii->uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[eax*4], 1
		CPatch::PatchPointer(0x40830F + 3, &CStreaming__ms_aInfoForModel.gta_iii->uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[eax*4], 1
		CPatch::PatchPointer(0x408373 + 3, &CStreaming__ms_aInfoForModel.gta_iii->uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[eax*4], 1
		CPatch::PatchPointer(0x408844 + 2, &CStreaming__ms_aInfoForModel.gta_iii->uiLoadFlag);	// mov     al, ss:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[ebp]
		CPatch::PatchPointer(0x40887D + 1, CStreaming__ms_aInfoForModel.gta_iii);	// mov     ecx, offset _ZN10CStreaming16ms_aInfoForModelE; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x408892 + 2, CStreaming__ms_aInfoForModel.gta_iii);	// cmp     dword ptr ss:_ZN10CStreaming16ms_aInfoForModelE.usNext[ebp], 0; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40889B + 2, &CStreaming__ms_aInfoForModel.gta_iii->uiLoadFlag);	// cmp     ss:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[ebp], 2
		CPatch::PatchPointer(0x4088A4 + 3, &CStreaming__ms_aInfoForModel.gta_iii->ucFlags);	// movzx   eax, ss:_ZN10CStreaming16ms_aInfoForModelE.ucFlags[ebp]
		CPatch::PatchPointer(0x4088B6 + 2, &CStreaming__ms_aInfoForModel.gta_iii->ucFlags);	// and     ss:_ZN10CStreaming16ms_aInfoForModelE.ucFlags[ebp], 0F7h
		CPatch::PatchPointer(0x4088C3 + 1, CStreaming__ms_aInfoForModel.gta_iii);	// mov     ecx, offset _ZN10CStreaming16ms_aInfoForModelE; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x4088D1 + 2, &CStreaming__ms_aInfoForModel.gta_iii->uiLoadFlag);	// cmp     ss:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[ebp], 3
		CPatch::PatchPointer(0x408908 + 2, &CStreaming__ms_aInfoForModel.gta_iii->uiLoadFlag);	// cmp     ss:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[ebp], 4
		CPatch::PatchPointer(0x408932 + 2, &CStreaming__ms_aInfoForModel.gta_iii->uiLoadFlag);	// mov     ss:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[ebp], 0
		CPatch::PatchPointer(0x408960 + 2, &CStreaming__ms_aInfoForModel.gta_iii->uiLoadFlag);	// cmp     ss:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[ebp], 1
		CPatch::PatchPointer(0x408969 + 3, &CStreaming__ms_aInfoForModel.gta_iii->ucFlags);	// movzx   eax, ss:_ZN10CStreaming16ms_aInfoForModelE.ucFlags[ebp]
		CPatch::PatchPointer(0x408989 + 2, &CStreaming__ms_aInfoForModel.gta_iii->ucFlags);	// mov     ss:_ZN10CStreaming16ms_aInfoForModelE.ucFlags[ebp], 0
		CPatch::PatchPointer(0x4089C7 + 2, CStreaming__ms_aInfoForModel.gta_iii);	// sub     edx, offset _ZN10CStreaming16ms_aInfoForModelE; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40937A + 3, CStreaming__ms_aInfoForModel.gta_iii);	// cmp     dword ptr ds:_ZN10CStreaming16ms_aInfoForModelE.usNext[eax*4], 0; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x409403 + 3, CStreaming__ms_aInfoForModel.gta_iii);	// cmp     dword ptr ds:_ZN10CStreaming16ms_aInfoForModelE.usNext[eax*4], 0; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40946C + 2, CStreaming__ms_aInfoForModel.gta_iii);	// sub     edx, offset _ZN10CStreaming16ms_aInfoForModelE; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x4094DB + 2, CStreaming__ms_aInfoForModel.gta_iii);	// sub     edx, offset _ZN10CStreaming16ms_aInfoForModelE; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40958D + 1, CStreaming__ms_aInfoForModel.gta_iii);	// mov     edi, offset _ZN10CStreaming16ms_aInfoForModelE; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x409696 + 2, &CStreaming__ms_aInfoForModel.gta_iii->uiLoadFlag);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[esi], 1
		CPatch::PatchPointer(0x4096B7 + 3, &CStreaming__ms_aInfoForModel.gta_iii->ucFlags);	// movzx   eax, ds:_ZN10CStreaming16ms_aInfoForModelE.ucFlags[esi]
		CPatch::PatchPointer(0x409757 + 4, CStreaming__ms_aInfoForModel.gta_iii);	// mov     [esp+30h+var_18], offset _ZN10CStreaming16ms_aInfoForModelE; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x4098B5 + 3, &CStreaming__ms_aInfoForModel.gta_iii->ucFlags);	// movzx   eax, ds:_ZN10CStreaming16ms_aInfoForModelE.ucFlags[edi]
		CPatch::PatchPointer(0x40991A + 3, &CStreaming__ms_aInfoForModel.gta_iii->ucFlags);	// movzx   eax, ds:_ZN10CStreaming16ms_aInfoForModelE.ucFlags[edi]
		CPatch::PatchPointer(0x409945 + 3, &CStreaming__ms_aInfoForModel.gta_iii->ucFlags);	// movzx   eax, ds:_ZN10CStreaming16ms_aInfoForModelE.ucFlags[edi]
		CPatch::PatchPointer(0x4099F1 + 3, &CStreaming__ms_aInfoForModel.gta_iii->ucFlags);	// movzx   eax, ds:_ZN10CStreaming16ms_aInfoForModelE.ucFlags[edi]
		CPatch::PatchPointer(0x409A34 + 3, &CStreaming__ms_aInfoForModel.gta_iii->ucFlags);	// movzx   eax, ds:_ZN10CStreaming16ms_aInfoForModelE.ucFlags[edi]
		CPatch::PatchPointer(0x409AB0 + 3, &CStreaming__ms_aInfoForModel.gta_iii->ucFlags);	// movzx   eax, ds:_ZN10CStreaming16ms_aInfoForModelE.ucFlags[edi]
		CPatch::PatchPointer(0x409C92 + 1, CStreaming__ms_aInfoForModel.gta_iii);	// mov     ecx, offset _ZN10CStreaming16ms_aInfoForModelE; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x409CDB + 3, &CStreaming__ms_aInfoForModel.gta_iii->ucFlags);	// movzx   eax, ds:_ZN10CStreaming16ms_aInfoForModelE.ucFlags[edi]
		CPatch::PatchPointer(0x409D04 + 3, &CStreaming__ms_aInfoForModel.gta_iii->ucFlags);	// movzx   eax, ds:_ZN10CStreaming16ms_aInfoForModelE.ucFlags[edi]
		CPatch::PatchPointer(0x409D66 + 2, &CStreaming__ms_aInfoForModel.gta_iii->uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[edi], 4
		CPatch::PatchPointer(0x409E8C + 2, CStreaming__ms_aInfoForModel.gta_iii);	// sub     edx, offset _ZN10CStreaming16ms_aInfoForModelE; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x409ED5 + 4, &CStreaming__ms_aInfoForModel.gta_iii[GetBaseID(FILE_TYPE_TXD)].uiLoadFlag);	// movzx   eax, ds:(_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+1ADB0h)[edi*4]
		CPatch::PatchPointer(0x409EF0 + 4, &CStreaming__ms_aInfoForModel.gta_iii->ucFlags);	// movzx   eax, ds:_ZN10CStreaming16ms_aInfoForModelE.ucFlags[ebx*4]
		CPatch::PatchPointer(0x409F15 + 1, CStreaming__ms_aInfoForModel.gta_iii);	// mov     ecx, offset _ZN10CStreaming16ms_aInfoForModelE; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x409F66 + 3, &CStreaming__ms_aInfoForModel.gta_iii->ucFlags);	// movzx   eax, ds:_ZN10CStreaming16ms_aInfoForModelE.ucFlags[edi]
		CPatch::PatchPointer(0x409F78 + 2, &CStreaming__ms_aInfoForModel.gta_iii->ucFlags);	// and     ds:_ZN10CStreaming16ms_aInfoForModelE.ucFlags[edi], 0F7h
		CPatch::PatchPointer(0x40A0AB + 1, CStreaming__ms_aInfoForModel.gta_iii);	// mov     ecx, offset _ZN10CStreaming16ms_aInfoForModelE; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40A0E6 + 3, &CStreaming__ms_aInfoForModel.gta_iii->ucFlags);	// movzx   eax, ss:_ZN10CStreaming16ms_aInfoForModelE.ucFlags[ebp]
		CPatch::PatchPointer(0x40A121 + 2, CStreaming__ms_aInfoForModel.gta_iii);	// add     ecx, offset _ZN10CStreaming16ms_aInfoForModelE; this
		CPatch::PatchPointer(0x40A196 + 2, &CStreaming__ms_aInfoForModel.gta_iii->uiLoadFlag);	// cmp     ss:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[ebp], 2
		CPatch::PatchPointer(0x40A1A3 + 1, CStreaming__ms_aInfoForModel.gta_iii);	// mov     esi, offset _ZN10CStreaming16ms_aInfoForModelE; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40A1C4 + 3, &CStreaming__ms_aInfoForModel.gta_iii->ucFlags);	// movzx   eax, ss:_ZN10CStreaming16ms_aInfoForModelE.ucFlags[ebp]
		CPatch::PatchPointer(0x40A211 + 3, &CStreaming__ms_aInfoForModel.gta_iii->uiLoadFlag);	// mov     dl, ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[eax*4]
		CPatch::PatchPointer(0x40A2B1 + 3, &CStreaming__ms_aInfoForModel.gta_iii->ucFlags);	// movzx   eax, ss:_ZN10CStreaming16ms_aInfoForModelE.ucFlags[ebp]
		CPatch::PatchPointer(0x40A2C3 + 2, &CStreaming__ms_aInfoForModel.gta_iii->ucFlags);	// and     ss:_ZN10CStreaming16ms_aInfoForModelE.ucFlags[ebp], 0F7h
		CPatch::PatchPointer(0x40A2D0 + 3, &CStreaming__ms_aInfoForModel.gta_iii->usNextOnCd);	// movsx   ebx, ss:_ZN10CStreaming16ms_aInfoForModelE.usNextOnCd[ebp]
		CPatch::PatchPointer(0x40A4AC + 1, CStreaming__ms_aInfoForModel.gta_iii);	// mov     ebp, offset _ZN10CStreaming16ms_aInfoForModelE; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40A4C6 + 3, &CStreaming__ms_aInfoForModel.gta_iii->ucFlags);	// movzx   eax, ds:_ZN10CStreaming16ms_aInfoForModelE.ucFlags[esi]
		CPatch::PatchPointer(0x40A4DA + 2, &CStreaming__ms_aInfoForModel.gta_iii->ucFlags);	// and     ds:_ZN10CStreaming16ms_aInfoForModelE.ucFlags[esi], 0F7h
		CPatch::PatchPointer(0x40A697 + 2, CStreaming__ms_aInfoForModel.gta_iii);	// sub     edx, offset _ZN10CStreaming16ms_aInfoForModelE; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40A707 + 2, CStreaming__ms_aInfoForModel.gta_iii);	// sub     edx, offset _ZN10CStreaming16ms_aInfoForModelE; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40A7A1 + 2, &CStreaming__ms_aInfoForModel.gta_iii->ucFlags);	// and     ds:_ZN10CStreaming16ms_aInfoForModelE.ucFlags[ebx], 0FEh
		CPatch::PatchPointer(0x40A7BD + 3, &CStreaming__ms_aInfoForModel.gta_iii->ucFlags);	// movzx   eax, ds:_ZN10CStreaming16ms_aInfoForModelE.ucFlags[ebx]
		CPatch::PatchPointer(0x40A7C9 + 2, &CStreaming__ms_aInfoForModel.gta_iii->uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[ebx], 1
		CPatch::PatchPointer(0x40A7D2 + 2, CStreaming__ms_aInfoForModel.gta_iii);	// cmp     dword ptr ds:_ZN10CStreaming16ms_aInfoForModelE.usNext[ebx], 0; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40A7DB + 1, CStreaming__ms_aInfoForModel.gta_iii);	// mov     ecx, offset _ZN10CStreaming16ms_aInfoForModelE; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40A831 + 2, &CStreaming__ms_aInfoForModel.gta_iii->ucFlags);	// and     ds:_ZN10CStreaming16ms_aInfoForModelE.ucFlags[ebx], 0FDh
		CPatch::PatchPointer(0x40A84D + 3, &CStreaming__ms_aInfoForModel.gta_iii->ucFlags);	// movzx   eax, ds:_ZN10CStreaming16ms_aInfoForModelE.ucFlags[ebx]
		CPatch::PatchPointer(0x40A859 + 2, &CStreaming__ms_aInfoForModel.gta_iii->uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[ebx], 1
		CPatch::PatchPointer(0x40A862 + 2, CStreaming__ms_aInfoForModel.gta_iii);	// cmp     dword ptr ds:_ZN10CStreaming16ms_aInfoForModelE.usNext[ebx], 0; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40A86B + 1, CStreaming__ms_aInfoForModel.gta_iii);	// mov     ecx, offset _ZN10CStreaming16ms_aInfoForModelE; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40A9D5 + 2, CStreaming__ms_aInfoForModel.gta_iii);	// add     ecx, offset _ZN10CStreaming16ms_aInfoForModelE; this
		CPatch::PatchPointer(0x40ADC7 + 3, &CStreaming__ms_aInfoForModel.gta_iii[MODEL_SPECIAL01].uiLoadFlag);	// cmp     ds:(_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+208h)[eax*4], 1
		CPatch::PatchPointer(0x40AEE7 + 2, &CStreaming__ms_aInfoForModel.gta_iii[MODEL_ENFORCER].uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+924h, 1
		CPatch::PatchPointer(0x40AF2B + 2, &CStreaming__ms_aInfoForModel.gta_iii[MODEL_FBICAR].uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+85Ch, 1
		CPatch::PatchPointer(0x40AF82 + 2, &CStreaming__ms_aInfoForModel.gta_iii[MODEL_BARRACKS].uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+99Ch, 1
		CPatch::PatchPointer(0x40AF8C + 2, &CStreaming__ms_aInfoForModel.gta_iii[MODEL_RHINO].uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+988h, 1
		CPatch::PatchPointer(0x40B018 + 3, &CStreaming__ms_aInfoForModel.gta_iii->uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[eax*4], 0
		CPatch::PatchPointer(0x40B0C7 + 4, &CStreaming__ms_aInfoForModel.gta_iii->ucFlags);	// movzx   ecx, ds:_ZN10CStreaming16ms_aInfoForModelE.ucFlags[edx*4]
		CPatch::PatchPointer(0x40B1C3 + 3, &CStreaming__ms_aInfoForModel.gta_iii->ucFlags);	// movzx   eax, ds:_ZN10CStreaming16ms_aInfoForModelE.ucFlags[ebx]
		CPatch::PatchPointer(0x40B1DD + 2, &CStreaming__ms_aInfoForModel.gta_iii->uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[ebx], 1
		CPatch::PatchPointer(0x40B223 + 2, &CStreaming__ms_aInfoForModel.gta_iii->uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[ebx], 1
		CPatch::PatchPointer(0x40B22C + 2, &CStreaming__ms_aInfoForModel.gta_iii->ucFlags);	// mov     al, ds:_ZN10CStreaming16ms_aInfoForModelE.ucFlags[ebx]
		CPatch::PatchPointer(0x40B260 + 2, &CStreaming__ms_aInfoForModel.gta_iii->uiLoadFlag);	// cmp     ss:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[ebp], 1
		CPatch::PatchPointer(0x40B270 + 2, &CStreaming__ms_aInfoForModel.gta_iii->ucFlags);	// mov     ss:_ZN10CStreaming16ms_aInfoForModelE.ucFlags[ebp], al
		CPatch::PatchPointer(0x4180F4 + 3, &CStreaming__ms_aInfoForModel.gta_iii->uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[ecx*4], 1
		CPatch::PatchPointer(0x418204 + 2, &CStreaming__ms_aInfoForModel.gta_iii[MODEL_ENFORCER].uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+924h, 1
		CPatch::PatchPointer(0x41820D + 2, &CStreaming__ms_aInfoForModel.gta_iii[MODEL_SWAT].uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+28h, 1
		CPatch::PatchPointer(0x41824A + 2, &CStreaming__ms_aInfoForModel.gta_iii[MODEL_FBICAR].uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+85Ch, 1
		CPatch::PatchPointer(0x418253 + 2, &CStreaming__ms_aInfoForModel.gta_iii[MODEL_FBI].uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+3Ch, 1
		CPatch::PatchPointer(0x418276 + 2, &CStreaming__ms_aInfoForModel.gta_iii[MODEL_RHINO].uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+988h, 1
		CPatch::PatchPointer(0x41827F + 2, &CStreaming__ms_aInfoForModel.gta_iii[MODEL_BARRACKS].uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+99Ch, 1
		CPatch::PatchPointer(0x418288 + 2, &CStreaming__ms_aInfoForModel.gta_iii[MODEL_ARMY].uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+50h, 1
		CPatch::PatchPointer(0x4182C9 + 2, &CStreaming__ms_aInfoForModel.gta_iii[MODEL_GANG01].uiLoadFlag);	// cmp     ds:(_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+0C8h)[eax], 1
		CPatch::PatchPointer(0x4182D2 + 2, &CStreaming__ms_aInfoForModel.gta_iii[MODEL_GANG02].uiLoadFlag);	// cmp     ds:(_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+0DCh)[eax], 1
		CPatch::PatchPointer(0x41FD28 + 2, &CStreaming__ms_aInfoForModel.gta_iii[MODEL_AMBULAN].uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+848h, 1
		CPatch::PatchPointer(0x41FD33 + 2, &CStreaming__ms_aInfoForModel.gta_iii[MODEL_MEDIC].uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+64h, 1
		CPatch::PatchPointer(0x41FD63 + 2, &CStreaming__ms_aInfoForModel.gta_iii[MODEL_AMBULAN].uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+848h, 1
		CPatch::PatchPointer(0x41FDF5 + 2, &CStreaming__ms_aInfoForModel.gta_iii[MODEL_FIRETRUK].uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+794h, 1
		CPatch::PatchPointer(0x41FE00 + 2, &CStreaming__ms_aInfoForModel.gta_iii[MODEL_FIREMAN].uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+78h, 1
		CPatch::PatchPointer(0x41FE30 + 2, &CStreaming__ms_aInfoForModel.gta_iii[MODEL_FIRETRUK].uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+794h, 1
		CPatch::PatchPointer(0x4276A4 + 3, &CStreaming__ms_aInfoForModel.gta_iii->uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[eax*4], 1
		CPatch::PatchPointer(0x4355A5 + 3, &CStreaming__ms_aInfoForModel.gta_iii->uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[eax*4], 1
		CPatch::PatchPointer(0x4371EF + 3, &CStreaming__ms_aInfoForModel.gta_iii->uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[eax*4], 1
		CPatch::PatchPointer(0x43790A + 3, &CStreaming__ms_aInfoForModel.gta_iii->uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[eax*4], 1
		CPatch::PatchPointer(0x43A778 + 2, &CStreaming__ms_aInfoForModel.gta_iii->uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag, 1
		CPatch::PatchPointer(0x444606 + 3, &CStreaming__ms_aInfoForModel.gta_iii->uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[ebx*4], 1
		CPatch::PatchPointer(0x490F11 + 2, &CStreaming__ms_aInfoForModel.gta_iii[MODEL_RHINO].uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+988h, 1
		CPatch::PatchPointer(0x491171 + 4, &CStreaming__ms_aInfoForModel.gta_iii->ucFlags);	// movzx   ebx, ds:_ZN10CStreaming16ms_aInfoForModelE.ucFlags[eax*4]
		CPatch::PatchPointer(0x4A9ED0 + 3, &CStreaming__ms_aInfoForModel.gta_iii->uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[eax*4], 1
		CPatch::PatchPointer(0x542726 + 3, &CStreaming__ms_aInfoForModel.gta_iii->uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[eax*4], 1
		CPatch::PatchPointer(0x54A5FA + 3, &CStreaming__ms_aInfoForModel.gta_iii->uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[ebx*4], 1
		CPatch::PatchPointer(0x54DD63 + 3, &CStreaming__ms_aInfoForModel.gta_iii->uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[eax*4], 1
		CPatch::PatchPointer(0x54FF17 + 2, &CStreaming__ms_aInfoForModel.gta_iii[MODEL_TRAIN].uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+9B0h, 1
		CPatch::PatchPointer(0x585C27 + 3, &CStreaming__ms_aInfoForModel.gta_iii->uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[eax*4], 1
		CPatch::PatchPointer(0x5860DE + 3, &CStreaming__ms_aInfoForModel.gta_iii->uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[eax*4], 1
		CPatch::PatchPointer(0x5872D0 + 3, &CStreaming__ms_aInfoForModel.gta_iii->uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[eax*4], 1
		CPatch::PatchPointer(0x587401 + 3, &CStreaming__ms_aInfoForModel.gta_iii->uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[eax*4], 1
		CPatch::PatchPointer(0x5953AE + 3, &CStreaming__ms_aInfoForModel.gta_iii->uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[eax*4], 1
		CPatch::PatchPointer(0x5955B5 + 3, &CStreaming__ms_aInfoForModel.gta_iii->uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[edi*4], 1
	}
	#endif
		
	// Patch base IDs and multiplies
	#if TRUE
	{
		// 5500
		{
			CPatch::PatchUINT32(0x401121 + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     ebp, 157Ch	; sub_4010D0
			CPatch::PatchUINT32(0x406919 + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     esi, 157Ch	; _ZN10CStreaming4InitEv
			CPatch::PatchUINT32(0x406FC0 + 2, GetBaseID(FILE_TYPE_TXD));	// lea     edi, [esi+157Ch]	; _ZN10CStreaming15LoadCdDirectoryEPKci
			CPatch::PatchUINT32(0x40701A + 2, GetBaseID(FILE_TYPE_TXD));	// lea     eax, [esi+157Ch]	; _ZN10CStreaming15LoadCdDirectoryEPKci
			CPatch::PatchUINT32(0x40702B + 2, GetBaseID(FILE_TYPE_TXD));	// add     esi, 157Ch	; _ZN10CStreaming15LoadCdDirectoryEPKci
			CPatch::PatchUINT32(0x407F06 + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     ecx, 157Ch	; ?requestModel@CStreaming@@QAEXHH@Z
			CPatch::PatchUINT32(0x407F74 + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     ecx, 157Ch	; ?requestModel@CStreaming@@QAEXHH@Z
			CPatch::PatchUINT32(0x407F88 + 1, GetBaseID(FILE_TYPE_TXD));	// add     eax, 157Ch	; ?requestModel@CStreaming@@QAEXHH@Z
			CPatch::PatchUINT32(0x408856 + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     ebx, 157Ch	; RemoveModel__10CStreamingFi
			CPatch::PatchUINT32(0x408911 + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     ebx, 157Ch	; RemoveModel__10CStreamingFi
			CPatch::PatchUINT32(0x408995 + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     ebx, 157Ch	; sub_408940
			CPatch::PatchUINT32(0x4089D9 + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     ebp, 157Ch	; RemoveLeastUsedModel__10CStreamingFv
			CPatch::PatchUINT32(0x409481 + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     ebp, 157Ch	; sub_409450
			CPatch::PatchUINT32(0x4094ED + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     ebp, 157Ch	; sub_4094C0
			CPatch::PatchUINT32(0x40952C + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     ebp, 157Ch	; sub_4094C0
			CPatch::PatchUINT32(0x409554 + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     ebx, 157Ch	; sub_4094C0
			CPatch::PatchUINT32(0x4095BE + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     ebx, 157Ch	; FinishLoadingLargeFile__10CStreamingFPci
			CPatch::PatchUINT32(0x409607 + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     ebx, 157Ch	; FinishLoadingLargeFile__10CStreamingFPci
			CPatch::PatchUINT32(0x4096F5 + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     ebx, 157Ch	; FinishLoadingLargeFile__10CStreamingFPci
			CPatch::PatchUINT32(0x4097AA + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     ebp, 157Ch	; sub_409740
			CPatch::PatchUINT32(0x40990F + 1, GetBaseID(FILE_TYPE_TXD));	// add     eax, 157Ch	; sub_409740
			CPatch::PatchUINT32(0x409A42 + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     ebp, 157Ch	; sub_409740
			CPatch::PatchUINT32(0x409A84 + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     ebp, 157Ch	; sub_409740
			CPatch::PatchUINT32(0x409B18 + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     ebp, 157Ch	; sub_409740
			CPatch::PatchUINT32(0x409CA8 + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     esi, 157Ch	; ProcessLoadingChannel__10CStreamingFi
			CPatch::PatchUINT32(0x409D2E + 2, GetBaseID(FILE_TYPE_TXD));	// add     ebx, 157Ch	; ProcessLoadingChannel__10CStreamingFi
			CPatch::PatchUINT32(0x409EBD + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     ebx, 157Ch	; sub_409E50
			CPatch::PatchUINT32(0x409EEA + 2, GetBaseID(FILE_TYPE_TXD));	// add     ecx, 157Ch	; sub_409E50
			CPatch::PatchUINT32(0x40A0D1 + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     ebx, 157Ch	; sub_40A060
			CPatch::PatchUINT32(0x40A1D4 + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     ebx, 157Ch	; sub_40A060
			CPatch::PatchUINT32(0x40A209 + 1, GetBaseID(FILE_TYPE_TXD));	// add     eax, 157Ch	; sub_40A060
			CPatch::PatchUINT32(0x40A273 + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     ebx, 157Ch	; sub_40A060
			CPatch::PatchUINT32(0x40A58C + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     edi, 157Ch	; LoadAllRequestedModels__10CStreamingFb
			CPatch::PatchUINT32(0x40A7A8 + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     edx, 157Ch	; SetModelIsDeletable__10CStreamingFi
			CPatch::PatchUINT32(0x40A80F + 1, GetBaseID(FILE_TYPE_TXD));	// add     eax, 157Ch	; sub_40A800
			CPatch::PatchUINT32(0x40A838 + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     edx, 157Ch	; SetMissionDoesntRequireModel__10CStreamingFi
			CPatch::PatchUINT32(0x476D9C + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     edi, 157Ch	; LoadObjectTypes__11CFileLoaderFPCc
			CPatch::PatchUINT32(0x4A6087 + 1, GetBaseID(FILE_TYPE_TXD));	// add     eax, 157Ch	; sub_4A6060
			CPatch::PatchUINT32(0x4A60D0 + 1, GetBaseID(FILE_TYPE_TXD));	// add     eax, 157Ch	; sub_4A60A0
			CPatch::PatchUINT32(0x50B5A1 + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     ebx, 157Ch	; _setup2dfx
			CPatch::PatchUINT32(0x50B901 + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     edi, 157Ch	; GetModelInfo__10CModelInfoFPCcPi
			CPatch::PatchUINT32(0x50BBEB + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     ebp, 157Ch	; RemoveColModelsFromOtherLevels__10CModelInfoF10eLevelName
			CPatch::PatchUINT32(0x517C51 + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     ebp, 157Ch	; sub_517C00
			CPatch::PatchUINT32(0x517D43 + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     ecx, 157Ch	; CTimeModelInfo::FindOtherTimeModel
			CPatch::PatchUINT32(0x517D4B + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     ecx, 157Ch	; CTimeModelInfo::FindOtherTimeModel
			CPatch::PatchUINT32(0x5279A5 + 2, GetBaseID(FILE_TYPE_TXD));	// lea     eax, [ebx+157Ch]	; RemoveRef__9CTxdStoreFi
			CPatch::PatchUINT32(0x585066 + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     ebx, 157Ch	; sub_585050
			CPatch::PatchUINT32(0x592D7E + 2, GetBaseID(FILE_TYPE_TXD));	// lea     eax, [ebx+157Ch]	; sub_592C70
			CPatch::PatchUINT32(0x592D93 + 2, GetBaseID(FILE_TYPE_TXD));	// lea     eax, [ebx+157Ch]	; sub_592C70
			CPatch::PatchUINT32(0x592E71 + 2, GetBaseID(FILE_TYPE_TXD));	// add     ebx, 157Ch	; sub_592C70
			CPatch::PatchUINT32(0x592F0A + 2, GetBaseID(FILE_TYPE_TXD));	// lea     eax, [ebx+157Ch]	; sub_592C70
		}
			
		// -5500
		{
			CPatch::PatchUINT32(0x408870 + 2, -GetBaseID(FILE_TYPE_TXD));	// lea     eax, [ebx-157Ch]	; RemoveModel__10CStreamingFi
			CPatch::PatchUINT32(0x408920 + 2, -GetBaseID(FILE_TYPE_TXD));	// lea     eax, [ebx-157Ch]	; RemoveModel__10CStreamingFi
			CPatch::PatchUINT32(0x408A00 + 2, -GetBaseID(FILE_TYPE_TXD));	// lea     eax, [ebp-157Ch]	; RemoveLeastUsedModel__10CStreamingFv
			CPatch::PatchUINT32(0x408A11 + 2, -GetBaseID(FILE_TYPE_TXD));	// lea     eax, [ebp-157Ch]	; RemoveLeastUsedModel__10CStreamingFv
			CPatch::PatchUINT32(0x409489 + 2, -GetBaseID(FILE_TYPE_TXD));	// lea     eax, [ebp-157Ch]	; sub_409450
			CPatch::PatchUINT32(0x409654 + 2, -GetBaseID(FILE_TYPE_TXD));	// lea     eax, [ebx-157Ch]	; FinishLoadingLargeFile__10CStreamingFPci
			CPatch::PatchUINT32(0x409660 + 2, -GetBaseID(FILE_TYPE_TXD));	// lea     eax, [ebx-157Ch]	; FinishLoadingLargeFile__10CStreamingFPci
			CPatch::PatchUINT32(0x409676 + 2, -GetBaseID(FILE_TYPE_TXD));	// lea     eax, [ebx-157Ch]	; FinishLoadingLargeFile__10CStreamingFPci
			CPatch::PatchUINT32(0x409711 + 2, -GetBaseID(FILE_TYPE_TXD));	// lea     eax, [ebx-157Ch]	; FinishLoadingLargeFile__10CStreamingFPci
			CPatch::PatchUINT32(0x409951 + 2, -GetBaseID(FILE_TYPE_TXD));	// lea     eax, [ebp-157Ch]	; sub_409740
			CPatch::PatchUINT32(0x409996 + 2, -GetBaseID(FILE_TYPE_TXD));	// lea     eax, [ebp-157Ch]	; sub_409740
			CPatch::PatchUINT32(0x4099B4 + 2, -GetBaseID(FILE_TYPE_TXD));	// lea     eax, [ebp-157Ch]	; sub_409740
			CPatch::PatchUINT32(0x4099D1 + 2, -GetBaseID(FILE_TYPE_TXD));	// lea     eax, [ebp-157Ch]	; sub_409740
			CPatch::PatchUINT32(0x409A60 + 2, -GetBaseID(FILE_TYPE_TXD));	// lea     eax, [ebp-157Ch]	; sub_409740
			CPatch::PatchUINT32(0x409B45 + 2, -GetBaseID(FILE_TYPE_TXD));	// lea     eax, [ebp-157Ch]	; sub_409740
			CPatch::PatchUINT32(0x40A0F2 + 2, -GetBaseID(FILE_TYPE_TXD));	// lea     eax, [ebx-157Ch]	; sub_40A060
		}
	}
	#endif

	// Patch limits itself
	#if TRUE
	{
		// Model DFF limit, patch pointers to CModelInfo::ms_modelInfoPtrs
		#if TRUE
		{
			CPatch::PatchPointer(0x4010E0 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     esi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs	; sub_4010D0
			CPatch::PatchPointer(0x40394A + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     ebp, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; LoadAnimFiles__12CAnimManagerFv
			CPatch::PatchPointer(0x404C07 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs	; CreateCutsceneObject__12CCutsceneMgrFi
			CPatch::PatchPointer(0x404C1D + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs	; CreateCutsceneObject__12CCutsceneMgrFi
			CPatch::PatchPointer(0x405862 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; ReplaceWithNewModel__9CBuildingFi
			CPatch::PatchPointer(0x4068E0 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     ebx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; _ZN10CStreaming4InitEv
			CPatch::PatchPointer(0x407CBC + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     esi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_407C50
			CPatch::PatchPointer(0x407E28 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_407DD0
			CPatch::PatchPointer(0x407F0E + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; ?requestModel@CStreaming@@QAEXHH@Z
			CPatch::PatchPointer(0x407F7C + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; ?requestModel@CStreaming@@QAEXHH@Z
			CPatch::PatchPointer(0x40843D + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     ebp, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4083D0
			CPatch::PatchPointer(0x4084CE + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     ebp, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4083D0
			CPatch::PatchPointer(0x408568 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     ebp, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4083D0
			CPatch::PatchPointer(0x4085FE + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     ebp, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4083D0
			CPatch::PatchPointer(0x4086F7 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     ebp, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; RemoveBigBuildings__10CStreamingF10eLevelName
			CPatch::PatchPointer(0x40885E + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs	; RemoveModel__10CStreamingFi
			CPatch::PatchPointer(0x408975 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs	; sub_408940
			CPatch::PatchPointer(0x4089E1 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs	; RemoveLeastUsedModel__10CStreamingFv
			CPatch::PatchPointer(0x409389 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs	; sub_409340
			CPatch::PatchPointer(0x409414 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs	; sub_4093C0
			CPatch::PatchPointer(0x4094F5 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs	; sub_4094C0
			CPatch::PatchPointer(0x409534 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs	; sub_4094C0
			CPatch::PatchPointer(0x40955C + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs	; sub_4094C0
			CPatch::PatchPointer(0x4095C6 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs	; FinishLoadingLargeFile__10CStreamingFPci
			CPatch::PatchPointer(0x409613 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     ebp, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs	; FinishLoadingLargeFile__10CStreamingFPci
			CPatch::PatchPointer(0x4096FE + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs	; FinishLoadingLargeFile__10CStreamingFPci
			CPatch::PatchPointer(0x4097BA + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs	; sub_409740
			CPatch::PatchPointer(0x409844 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs	; sub_409740
			CPatch::PatchPointer(0x409898 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs	; sub_409740
			CPatch::PatchPointer(0x409B21 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs	; sub_409740
			CPatch::PatchPointer(0x409CB6 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; ProcessLoadingChannel__10CStreamingFi
			CPatch::PatchPointer(0x409CF3 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; ProcessLoadingChannel__10CStreamingFi
			CPatch::PatchPointer(0x409EC5 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs	; sub_409E50
			CPatch::PatchPointer(0x40A1E1 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs	; sub_40A060
			CPatch::PatchPointer(0x40A27B + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     ebx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs	; sub_40A060
			CPatch::PatchPointer(0x40A594 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; LoadAllRequestedModels__10CStreamingFb
			CPatch::PatchPointer(0x40A7B0 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; SetModelIsDeletable__10CStreamingFi
			CPatch::PatchPointer(0x40A804 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_40A800
			CPatch::PatchPointer(0x40A840 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; SetMissionDoesntRequireModel__10CStreamingFi
			CPatch::PatchPointer(0x40A89B + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     ebx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs	; RequestSpecialModel__10CStreamingFiPCci
			CPatch::PatchPointer(0x40B02E + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs	; StreamVehiclesAndPeds__10CStreamingFv
			CPatch::PatchPointer(0x40B0D4 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs	; sub_40B060
			CPatch::PatchPointer(0x40B1CF + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; RemoveLoadedVehicle__10CStreamingFv
			CPatch::PatchPointer(0x40B4F6 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; GetCollisionInSectorList__FR8CPtrList
			CPatch::PatchPointer(0x416C6B + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; GenerateOneRandomCar__8CCarCtrlFv
			CPatch::PatchPointer(0x416DE0 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; GenerateOneRandomCar__8CCarCtrlFv
			CPatch::PatchPointer(0x416FA2 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; GenerateOneRandomCar__8CCarCtrlFv
			CPatch::PatchPointer(0x417C00 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; GenerateOneRandomCar__8CCarCtrlFv
			CPatch::PatchPointer(0x417C6C + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; GenerateOneRandomCar__8CCarCtrlFv
			CPatch::PatchPointer(0x417CCA + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; this	; GenerateOneRandomCar__8CCarCtrlFv
			CPatch::PatchPointer(0x417E6A + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_417E60
			CPatch::PatchPointer(0x41931C + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     esi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs	; SlowCarDownForPedsSectorList__8CCarCtrlFR8CPtrListP11CAutomobileffffPff
			CPatch::PatchPointer(0x419529 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; SlowCarDownForPedsSectorList__8CCarCtrlFR8CPtrListP11CAutomobileffffPff
			CPatch::PatchPointer(0x41A05A + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; TestCollisionBetween2MovingRects__8CCarCtrlFP8CVehicleP8CVehicleffP7CVectorP7CVectorUc
			CPatch::PatchPointer(0x41A081 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; TestCollisionBetween2MovingRects__8CCarCtrlFP8CVehicleP8CVehicleffP7CVectorP7CVectorUc
			CPatch::PatchPointer(0x41ADBD + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_41ACF0
			CPatch::PatchPointer(0x41ADCF + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_41ACF0
			CPatch::PatchPointer(0x41AE73 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_41ACF0
			CPatch::PatchPointer(0x41AEC6 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_41ACF0
			CPatch::PatchPointer(0x41AF08 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_41ACF0
			CPatch::PatchPointer(0x41AF43 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_41ACF0
			CPatch::PatchPointer(0x41AF8A + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_41ACF0
			CPatch::PatchPointer(0x41AFC5 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_41ACF0
			CPatch::PatchPointer(0x41B00A + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_41ACF0
			CPatch::PatchPointer(0x41B35C + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_41B2D0
			CPatch::PatchPointer(0x41B88A + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_41B640
			CPatch::PatchPointer(0x41DE22 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     esi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; SteerAICarWithPhysics_OnlyMission__8CCarCtrlFP11CAutomobilePfPfPfPb
			CPatch::PatchPointer(0x41DE36 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; SteerAICarWithPhysics_OnlyMission__8CCarCtrlFP11CAutomobilePfPfPfPb
			CPatch::PatchPointer(0x422878 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4222D0
			CPatch::PatchPointer(0x4228A1 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4222D0
			CPatch::PatchPointer(0x4253FC + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_425370
			CPatch::PatchPointer(0x4255DC + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_4254F0
			CPatch::PatchPointer(0x4257BC + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_425740
			CPatch::PatchPointer(0x425A51 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; IsEntityTouching3D__7CGarageFP7CEntity
			CPatch::PatchPointer(0x425B45 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; EntityHasASphereWayOutsideGarage__7CGarageFP7CEntityf
			CPatch::PatchPointer(0x425D2F + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; IsAnyOtherCarTouchingGarage__7CGarageFP8CVehicle
			CPatch::PatchPointer(0x425EBF + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; IsAnyOtherPedTouchingGarage__7CGarageFP4CPed
			CPatch::PatchPointer(0x426044 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; IsAnyCarBlockingDoor__7CGarageFv
			CPatch::PatchPointer(0x427C14 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_427C10
			CPatch::PatchPointer(0x427E5E + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_427D90
			CPatch::PatchPointer(0x42F1C6 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_42F1A0
			CPatch::PatchPointer(0x435484 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; RestoreInfoForCar__19CRecordDataForChaseFP11CAutomobileP18CCarStateEachFrameb
			CPatch::PatchPointer(0x437173 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; GenerateRoadBlocks__11CRoadBlocksFv
			CPatch::PatchPointer(0x437205 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; GenerateRoadBlocks__11CRoadBlocksFv
			CPatch::PatchPointer(0x437447 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; GenerateRoadBlocks__11CRoadBlocksFv			
			CPatch::PatchPointer(0x437808 + 2, CModelInfo__ms_modelInfoPtrs.gta_iii + MODEL_POLICE);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE+1D0h	; GenerateRoadBlockCopsForCar__11CRoadBlocksFP8CVehicleis
			CPatch::PatchPointer(0x437816 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; GenerateRoadBlockCopsForCar__11CRoadBlocksFP8CVehicleis
			CPatch::PatchPointer(0x43C5E6 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; ProcessCommands100To199__14CRunningScriptFi
			CPatch::PatchPointer(0x449A85 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs	; ProcessCommands800To899__14CRunningScriptFi
			CPatch::PatchPointer(0x449FEA + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; ProcessCommands800To899__14CRunningScriptFi
			CPatch::PatchPointer(0x44A2E8 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs	; ProcessCommands800To899__14CRunningScriptFi
			CPatch::PatchPointer(0x44A84A + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; ProcessCommands800To899__14CRunningScriptFi
			CPatch::PatchPointer(0x44C073 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs	; ProcessCommands800To899__14CRunningScriptFi
			CPatch::PatchPointer(0x44C083 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs	; ProcessCommands800To899__14CRunningScriptFi
			CPatch::PatchPointer(0x44C0A3 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; ProcessCommands800To899__14CRunningScriptFi
			CPatch::PatchPointer(0x44C0B3 + 1, CModelInfo__ms_modelInfoPtrs.gta_iii + MODEL_MALE01);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE+1Ch	; ProcessCommands800To899__14CRunningScriptFi
			CPatch::PatchPointer(0x44ED9E + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; ProcessCommands900To999__14CRunningScriptFi
			CPatch::PatchPointer(0x44EDED + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; ProcessCommands900To999__14CRunningScriptFi
			CPatch::PatchPointer(0x454150 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; ClearSpaceForMissionEntity__11CTheScriptsFRC7CVectorP7CEntity
			CPatch::PatchPointer(0x4541A2 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; ClearSpaceForMissionEntity__11CTheScriptsFRC7CVectorP7CEntity
			CPatch::PatchPointer(0x4541B6 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; ClearSpaceForMissionEntity__11CTheScriptsFRC7CVectorP7CEntity
			CPatch::PatchPointer(0x4541D8 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; ClearSpaceForMissionEntity__11CTheScriptsFRC7CVectorP7CEntity
			CPatch::PatchPointer(0x4541EF + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     esi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; ClearSpaceForMissionEntity__11CTheScriptsFRC7CVectorP7CEntity
			CPatch::PatchPointer(0x454B13 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; UpdateObjectIndices__11CTheScriptsFv
			CPatch::PatchPointer(0x455865 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_455800
			CPatch::PatchPointer(0x457B06 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; FixCamWhenObscuredByVehicle__4CCamFRC7CVector
			CPatch::PatchPointer(0x45A8B7 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_459FA0
			CPatch::PatchPointer(0x45C170 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     esi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; Process_Cam_On_A_String__4CCamFRC7CVectorfff
			CPatch::PatchPointer(0x45C177 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; Process_Cam_On_A_String__4CCamFRC7CVectorfff
			CPatch::PatchPointer(0x467ABD + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     esi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs	; ProcessSpecialHeightRoutines__4CCamFv
			CPatch::PatchPointer(0x467AC4 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; ProcessSpecialHeightRoutines__4CCamFv
			CPatch::PatchPointer(0x467BED + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; ProcessSpecialHeightRoutines__4CCamFv
			CPatch::PatchPointer(0x467BF4 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; ProcessSpecialHeightRoutines__4CCamFv
			CPatch::PatchPointer(0x467D24 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; ProcessSpecialHeightRoutines__4CCamFv
			CPatch::PatchPointer(0x467D2B + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     ebp, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; ProcessSpecialHeightRoutines__4CCamFv
			CPatch::PatchPointer(0x4689EF + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; CamControl__7CCameraFv
			CPatch::PatchPointer(0x4689F6 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs	; CamControl__7CCameraFv
			CPatch::PatchPointer(0x473EA9 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     ebx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; CreateRwObject__7CEntityFv
			CPatch::PatchPointer(0x473F44 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; AttachToRwObject__7CEntityFP8RwObject
			CPatch::PatchPointer(0x473F6D + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; DetachFromRwObject__7CEntityFv
			CPatch::PatchPointer(0x473FD7 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; DeleteRwObject__7CEntityFv
			CPatch::PatchPointer(0x47401A + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; GetBoundRect__7CEntityFv
			CPatch::PatchPointer(0x474297 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; GetBoundCentre__7CEntityFv
			CPatch::PatchPointer(0x4742CF + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; GetBoundCentre__7CEntityFR9CVuVector
			CPatch::PatchPointer(0x474314 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; GetBoundRadius__7CEntityFv
			CPatch::PatchPointer(0x474549 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     ebx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs	; PreRender__7CEntityFv
			CPatch::PatchPointer(0x474BAF + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs	; PreRender__7CEntityFv
			CPatch::PatchPointer(0x474E7F + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_474D20
			CPatch::PatchPointer(0x4755C4 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; GetDistanceFromCentreOfMassToBaseOfModel__7CEntityFv
			CPatch::PatchPointer(0x4755E9 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     ebp, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; SetupBigBuilding__7CEntityFv
			CPatch::PatchPointer(0x47695A + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_476930
			CPatch::PatchPointer(0x4769BA + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     esi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs	; sub_476990
			CPatch::PatchPointer(0x476A3C + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_476A20
			CPatch::PatchPointer(0x476A75 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs	; FinishLoadClumpFile__11CFileLoaderFP8RwStreamUi
			CPatch::PatchPointer(0x476A9B + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs	; FinishLoadClumpFile__11CFileLoaderFP8RwStreamUi
			CPatch::PatchPointer(0x476D80 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; LoadObjectTypes__11CFileLoaderFPCc
			CPatch::PatchPointer(0x4773AF + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     ebx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4772B0
			CPatch::PatchPointer(0x4777D9 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     ebx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs	; readMloInstance
			CPatch::PatchPointer(0x477EB0 + 1, CModelInfo__ms_modelInfoPtrs.gta_iii + MODEL_LOPOLYGUY);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE+31Ch	; sub_477DE0
			CPatch::PatchPointer(0x478144 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     ebp, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4780E0
			CPatch::PatchPointer(0x478724 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     esi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; LoadObjectInstance__11CFileLoaderFPCc
			CPatch::PatchPointer(0x479AEB + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; ProcessFire__5CFireFv
			CPatch::PatchPointer(0x49114F + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// cmp     ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4], 0; CModelInfo::ms_modelInfoPtrs	; sub_4910B0
			CPatch::PatchPointer(0x49B8FA + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; ProcessCollisionSectorList__9CPhysicalFP8CPtrList
			CPatch::PatchPointer(0x49BAB2 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; ProcessCollisionSectorList__9CPhysicalFP8CPtrList
			CPatch::PatchPointer(0x49DCD5 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     esi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; ProcessShiftSectorList__9CPhysicalFP8CPtrList
			CPatch::PatchPointer(0x49DE73 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     ebp, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; ProcessShiftSectorList__9CPhysicalFP8CPtrList
			CPatch::PatchPointer(0x49F7AA + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; ProcessEntityCollision__9CPhysicalFP7CEntityP9CColPoint
			CPatch::PatchPointer(0x49F7B9 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; ProcessEntityCollision__9CPhysicalFP7CEntityP9CColPoint
			CPatch::PatchPointer(0x4A2B18 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     esi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4A29B0
			CPatch::PatchPointer(0x4A4D48 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     esi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_4A4C70
			CPatch::PatchPointer(0x4A4E99 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     esi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_4A4C70
			CPatch::PatchPointer(0x4A9364 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     ebx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4A9350
			CPatch::PatchPointer(0x4A938D + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_4A9350
			CPatch::PatchPointer(0x4A93B6 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_4A9350
			CPatch::PatchPointer(0x4A9853 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4A9840
			CPatch::PatchPointer(0x4A992F + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     ebx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4A9920
			CPatch::PatchPointer(0x4A9945 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4A9920
			CPatch::PatchPointer(0x4A9A5E + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4A9920
			CPatch::PatchPointer(0x4AD63F + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; GetAndSetVehicleName__4CHudFv
			CPatch::PatchPointer(0x4B0D56 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; ProcessLineOfSightSectorList__6CWorldFR8CPtrListRC8CColLineR9CColPointRfRP7CEntitybb
			CPatch::PatchPointer(0x4B0D6D + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; ProcessLineOfSightSectorList__6CWorldFR8CPtrListRC8CColLineR9CColPointRfRP7CEntitybb
			CPatch::PatchPointer(0x4B10E7 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_4B1090
			CPatch::PatchPointer(0x4B21C8 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_4B2160
			CPatch::PatchPointer(0x4B4534 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4B4300
			CPatch::PatchPointer(0x4B4607 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_4B4300
			CPatch::PatchPointer(0x4B4CB1 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_4B4AC0
			CPatch::PatchPointer(0x4B58D9 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_4B57B0
			CPatch::PatchPointer(0x4BAE2F + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; __dt__7CObjectFv
			CPatch::PatchPointer(0x4BB224 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; this	; Render__7CObjectFv
			CPatch::PatchPointer(0x4BBD8B + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; RefModelInfo__7CObjectFi
			CPatch::PatchPointer(0x4BC278 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; SetObjectData__11CObjectDataFiR7CObject
			CPatch::PatchPointer(0x4C52D0 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     ebp, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; SetModelIndex__4CPedFUi
			CPatch::PatchPointer(0x4C8F6C + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     esi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; ProcessControl__4CPedFv
			CPatch::PatchPointer(0x4C8F8E + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; ProcessControl__4CPedFv
			CPatch::PatchPointer(0x4C977B + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; ProcessControl__4CPedFv
			CPatch::PatchPointer(0x4CBEDA + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs	; sub_4CBB30
			CPatch::PatchPointer(0x4CBF50 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs	; sub_4CBB30
			CPatch::PatchPointer(0x4CC2B1 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_4CBB30
			CPatch::PatchPointer(0x4CC2C9 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_4CBB30
			CPatch::PatchPointer(0x4CCEDA + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; SetDirectionToWalkAroundObject__4CPedFP7CEntity
			CPatch::PatchPointer(0x4CF904 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs	; AddWeaponModel__4CPedFi
			CPatch::PatchPointer(0x4D019D + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4CFEF4
			CPatch::PatchPointer(0x4D1067 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     esi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_4D0F20
			CPatch::PatchPointer(0x4D107C + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_4D0F20
			CPatch::PatchPointer(0x4D172F + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; Seek__4CPedFv
			CPatch::PatchPointer(0x4D4343 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_4D3F90
			CPatch::PatchPointer(0x4D4AE2 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     ebx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_4D4970
			CPatch::PatchPointer(0x4D5240 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4D5040
			CPatch::PatchPointer(0x4D5335 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4D5040
			CPatch::PatchPointer(0x4D5435 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4D5040
			CPatch::PatchPointer(0x4D5534 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4D5040
			CPatch::PatchPointer(0x4D6B0B + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_4D6A00
			CPatch::PatchPointer(0x4D7527 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4D7490
			CPatch::PatchPointer(0x4E1A42 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4E1A30
			CPatch::PatchPointer(0x4E1F50 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_4E1F30
			CPatch::PatchPointer(0x4E2D84 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     esi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4E2D70
			CPatch::PatchPointer(0x4E2E70 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     ebp, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4E2E50
			CPatch::PatchPointer(0x4E3402 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4E33D0
			CPatch::PatchPointer(0x4E475B + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_4E4730
			CPatch::PatchPointer(0x4E4C9B + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; SeekBoatPosition__4CPedFv
			CPatch::PatchPointer(0x4E5238 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs	; PositionPedOutOfCollision__4CPedFv
			CPatch::PatchPointer(0x4E58A0 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     esi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; ServiceTalking__4CPedFv
			CPatch::PatchPointer(0x4E9010 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; FightStrike__4CPedFR7CVector
			CPatch::PatchPointer(0x4EC692 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; KillPedWithCar__4CPedFP8CVehiclef
			CPatch::PatchPointer(0x4F46C7 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_4F4690
			CPatch::PatchPointer(0x4F4714 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_4F4700
			CPatch::PatchPointer(0x4F490D + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_4F4700
			CPatch::PatchPointer(0x4F498B + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_4F4700
			CPatch::PatchPointer(0x4F4D41 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     esi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; AddToPopulation__11CPopulationFffff
			CPatch::PatchPointer(0x4F4E9F + 2, CModelInfo__ms_modelInfoPtrs.gta_iii + MODEL_COP);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE+4	; AddToPopulation__11CPopulationFffff
			CPatch::PatchPointer(0x4F4EC5 + 2, CModelInfo__ms_modelInfoPtrs.gta_iii + MODEL_FBI);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE+0Ch	; AddToPopulation__11CPopulationFffff
			CPatch::PatchPointer(0x4F4EE5 + 2, CModelInfo__ms_modelInfoPtrs.gta_iii + MODEL_SWAT);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE+8	; AddToPopulation__11CPopulationFffff
			CPatch::PatchPointer(0x4F4F05 + 2, CModelInfo__ms_modelInfoPtrs.gta_iii + MODEL_ARMY);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE+10h	; AddToPopulation__11CPopulationFffff
			CPatch::PatchPointer(0x4F4F20 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; AddToPopulation__11CPopulationFffff
			CPatch::PatchPointer(0x4F5942 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     esi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; AddPedInCar__11CPopulationFP8CVehicle
			CPatch::PatchPointer(0x4F5972 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs	; AddPedInCar__11CPopulationFP8CVehicle
			CPatch::PatchPointer(0x4F598E + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs	; AddPedInCar__11CPopulationFP8CVehicle
			CPatch::PatchPointer(0x4F59A2 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs	; AddPedInCar__11CPopulationFP8CVehicle
			CPatch::PatchPointer(0x4F59B6 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs	; AddPedInCar__11CPopulationFP8CVehicle
			CPatch::PatchPointer(0x4FA58A + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs	; ProcessLightsForEntity__7CEntityFv
			CPatch::PatchPointer(0x5034F3 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; RenderEntityInGlass__6CGlassFP7CEntity
			CPatch::PatchPointer(0x503F4C + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; WindowRespondsToCollision__6CGlassFP7CEntityf7CVector7CVectorb
			CPatch::PatchPointer(0x50B1F6 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_50B1A0
			CPatch::PatchPointer(0x50B313 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4], 0; CModelInfo::ms_modelInfoPtrs	; sub_50B310
			CPatch::PatchPointer(0x50B31E + 3, CModelInfo__ms_modelInfoPtrs.gta_iii + MODEL_COP);	// mov     ds:(_ZN10CModelInfo16ms_modelInfoPtrsE+4)[eax*4], 0	; sub_50B310
			CPatch::PatchPointer(0x50B329 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii + MODEL_SWAT);	// mov     ds:(_ZN10CModelInfo16ms_modelInfoPtrsE+8)[eax*4], 0	; sub_50B310
			CPatch::PatchPointer(0x50B334 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii + MODEL_FBI);	// mov     ds:(_ZN10CModelInfo16ms_modelInfoPtrsE+0Ch)[eax*4], 0	; sub_50B310
			CPatch::PatchPointer(0x50B33F + 3, CModelInfo__ms_modelInfoPtrs.gta_iii + MODEL_ARMY);	// mov     ds:(_ZN10CModelInfo16ms_modelInfoPtrsE+10h)[eax*4], 0	; sub_50B310
			CPatch::PatchPointer(0x50B34A + 3, CModelInfo__ms_modelInfoPtrs.gta_iii + MODEL_MEDIC);	// mov     ds:(_ZN10CModelInfo16ms_modelInfoPtrsE+14h)[eax*4], 0	; sub_50B310
			CPatch::PatchPointer(0x50B355 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii + MODEL_FIREMAN);	// mov     ds:(_ZN10CModelInfo16ms_modelInfoPtrsE+18h)[eax*4], 0	; sub_50B310
			CPatch::PatchPointer(0x50B360 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii + MODEL_MALE01);	// mov     ds:(_ZN10CModelInfo16ms_modelInfoPtrsE+1Ch)[eax*4], 0	; sub_50B310
			CPatch::PatchPointer(0x50B384 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4], 0; CModelInfo::ms_modelInfoPtrs	; sub_50B310
			CPatch::PatchPointer(0x50B399 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii + MODEL_COP);	// mov     ds:(_ZN10CModelInfo16ms_modelInfoPtrsE+4)[eax*4], 0	; sub_50B310
			CPatch::PatchPointer(0x50B3AE + 3, CModelInfo__ms_modelInfoPtrs.gta_iii + MODEL_SWAT);	// mov     ds:(_ZN10CModelInfo16ms_modelInfoPtrsE+8)[eax*4], 0	; sub_50B310
			CPatch::PatchPointer(0x50B3C3 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii + MODEL_FBI);	// mov     ds:(_ZN10CModelInfo16ms_modelInfoPtrsE+0Ch)[eax*4], 0	; sub_50B310
			CPatch::PatchPointer(0x50B590 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; this	; _setup2dfx
			CPatch::PatchPointer(0x50B870 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; GetModelInfo__10CModelInfoFPCcPi
			CPatch::PatchPointer(0x50B964 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4], ebp; CModelInfo::ms_modelInfoPtrs	; CModelInfo::AddSimpleModel
			CPatch::PatchPointer(0x50B9B7 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4], eax; CModelInfo::ms_modelInfoPtrs	; CModelInfo::AddMloModel
			CPatch::PatchPointer(0x50B9FE + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4], ebp; CModelInfo::ms_modelInfoPtrs	; CModelInfo::AddTimeModel
			CPatch::PatchPointer(0x50BA49 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4], eax; CModelInfo::ms_modelInfoPtrs	; CModelInfo::AddClumpModel
			CPatch::PatchPointer(0x50BAC2 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4], eax; CModelInfo::ms_modelInfoPtrs	; CClumpModel::AddVehicleModel
			CPatch::PatchPointer(0x50BB09 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4], eax; CModelInfo::ms_modelInfoPtrs	; CModelInfo::AddPedModel
			CPatch::PatchPointer(0x50BB94 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; IsBoatModel__10CModelInfoFi
			CPatch::PatchPointer(0x50BBC8 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs	; RemoveColModelsFromOtherLevels__10CModelInfoF10eLevelName
			CPatch::PatchPointer(0x50CE59 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     ebp, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; AddSteamsFromGround__7CEntityFP7CVector
			CPatch::PatchPointer(0x5127EB + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs	; StirUp__8CRubbishFP11CAutomobile
			CPatch::PatchPointer(0x51392C + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs	; sub_513830
			CPatch::PatchPointer(0x514D6F + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; CastShadowSectorList__8CShadowsFR8CPtrListffffP7CVectorffffsUcUcUcffPP10CPolyBunch
			CPatch::PatchPointer(0x514EC4 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_514E30
			CPatch::PatchPointer(0x517412 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; RenderExtraPlayerShadows__8CShadowsFv
			CPatch::PatchPointer(0x517C10 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs	; sub_517C00
			CPatch::PatchPointer(0x517D02 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; CTimeModelInfo::FindOtherTimeModel
			CPatch::PatchPointer(0x518D9C + 1, CModelInfo__ms_modelInfoPtrs.gta_iii + MODEL_BAT);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE+2B0h	; sub_518D40
			CPatch::PatchPointer(0x5203E7 + 2, CModelInfo__ms_modelInfoPtrs.gta_iii + MODEL_RHINO);	// cmp     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE+1E8h	; CVehicleModelInfo::SetVehicleComponentFlags
			CPatch::PatchPointer(0x52078F + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; CVehicleModelInfo::PreProcessHierarchy
			CPatch::PatchPointer(0x52103F + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// cmp     ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4], ebx; CModelInfo::ms_modelInfoPtrs	; CVehicleModelInfo::AvoidSameVehicleColour
			CPatch::PatchPointer(0x52199E + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     ebp, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; CVehicleModelInfo::SetEnvironmentMap
			CPatch::PatchPointer(0x5219F0 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; GetVehicleModelMaxPassengers
			CPatch::PatchPointer(0x529091 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     esi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_529070
			CPatch::PatchPointer(0x52C75C + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; ??0CAutomobile@@QAE@HE@Z
			CPatch::PatchPointer(0x52D21F + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     ebp, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_52D210
			CPatch::PatchPointer(0x52D4FD + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; HydraulicControl__11CAutomobileFv
			CPatch::PatchPointer(0x530147 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; CAutomobile::SetPanelDamage
			CPatch::PatchPointer(0x530227 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs	; SetDoorDamage__11CAutomobileFi6eDoors
			CPatch::PatchPointer(0x53085B + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; SpawnFlyingComponent__11CAutomobileFiUi
			CPatch::PatchPointer(0x530871 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; SpawnFlyingComponent__11CAutomobileFiUi
			CPatch::PatchPointer(0x5314BB + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     esi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; ProcessControl__11CAutomobileFv
			CPatch::PatchPointer(0x534366 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; ProcessControl__11CAutomobileFv
			CPatch::PatchPointer(0x535504 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; AddDamagedVehicleParticles__11CAutomobileFv
			CPatch::PatchPointer(0x536DF4 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; PreRender__11CAutomobileFv
			CPatch::PatchPointer(0x5370E1 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; PreRender__11CAutomobileFv
			CPatch::PatchPointer(0x53833C + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; PreRender__11CAutomobileFv
			CPatch::PatchPointer(0x538CBD + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; PreRender__11CAutomobileFv
			CPatch::PatchPointer(0x53982E + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs	; PreRender__11CAutomobileFv
			CPatch::PatchPointer(0x539EC6 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     ebp, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; Render__11CAutomobileFv
			CPatch::PatchPointer(0x53B2C4 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; ProcessEntityCollision__11CAutomobileFP7CEntityP9CColPoint
			CPatch::PatchPointer(0x53B320 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     esi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; ProcessEntityCollision__11CAutomobileFP7CEntityP9CColPoint
			CPatch::PatchPointer(0x53BF81 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     esi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_53BF70
			CPatch::PatchPointer(0x53C5ED + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_53C5B0
			CPatch::PatchPointer(0x53CBB8 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     esi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_53CBA0
			CPatch::PatchPointer(0x53CCB6 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     ebp, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_53CBA0
			CPatch::PatchPointer(0x53D10B + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; MakeTyresMuddySectorList__4CPedFR8CPtrList
			CPatch::PatchPointer(0x53D12F + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; MakeTyresMuddySectorList__4CPedFR8CPtrList
			CPatch::PatchPointer(0x53D14B + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; MakeTyresMuddySectorList__4CPedFR8CPtrList
			CPatch::PatchPointer(0x53D165 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; MakeTyresMuddySectorList__4CPedFR8CPtrList
			CPatch::PatchPointer(0x53E0A0 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; PlaceOnRoadProperly__11CAutomobileFv
			CPatch::PatchPointer(0x53E420 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     esi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs	; CBoat::Ctor
			CPatch::PatchPointer(0x53E963 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; this	; sub_53E890
			CPatch::PatchPointer(0x540594 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs	; sub_53F4E4
			CPatch::PatchPointer(0x5411F0 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_53F4E4
			CPatch::PatchPointer(0x5429B9 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_5426E0
			CPatch::PatchPointer(0x542D11 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_542C40
			CPatch::PatchPointer(0x542D31 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_542C40
			CPatch::PatchPointer(0x543D75 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     ebp, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs	; Update__6CCraneFv
			CPatch::PatchPointer(0x543DDB + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; Update__6CCraneFv
			CPatch::PatchPointer(0x5443F1 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; Update__6CCraneFv
			CPatch::PatchPointer(0x546391 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_546360
			CPatch::PatchPointer(0x547241 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs	; __ct__5CHeliFiUc
			CPatch::PatchPointer(0x549970 + 1, CModelInfo__ms_modelInfoPtrs.gta_iii + MODEL_ESCAPE);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE+24Ch	; InitHelis__5CHeliFv
			CPatch::PatchPointer(0x5499D2 + 1, CModelInfo__ms_modelInfoPtrs.gta_iii + MODEL_CHOPPER);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE+1F4h	; InitHelis__5CHeliFv
			CPatch::PatchPointer(0x54B191 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs	; __ct__6CPlaneFiUc
			CPatch::PatchPointer(0x54B3CD + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; PreRender__6CPlaneFv
			CPatch::PatchPointer(0x54B76A + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; PreRender__6CPlaneFv
			CPatch::PatchPointer(0x54DD55 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; ProcessControl__6CPlaneFv
			CPatch::PatchPointer(0x54DDEB + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; ProcessControl__6CPlaneFv
			CPatch::PatchPointer(0x54E2DE + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs	; __ct__6CTrainFiUc
			CPatch::PatchPointer(0x54E5AD + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; PreRender__6CTrainFv
			CPatch::PatchPointer(0x54E947 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; PreRender__6CTrainFv
			CPatch::PatchPointer(0x54FFCD + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; ProcessControl__6CTrainFv
			CPatch::PatchPointer(0x552663 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; IsSphereTouchingVehicle__8CVehicleFffff
			CPatch::PatchPointer(0x552BD7 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_552BB0
			CPatch::PatchPointer(0x562508 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; CWeapon::FireInstantHitFromCar
			CPatch::PatchPointer(0x562565 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; CWeapon::FireInstantHitFromCar
			CPatch::PatchPointer(0x562705 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     esi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; CWeapon::FireInstantHitFromCar
			CPatch::PatchPointer(0x5747A1 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     ebp, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs	; sub_574790
			CPatch::PatchPointer(0x585072 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     ebp, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs	; sub_585050
			CPatch::PatchPointer(0x596DEA + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     esi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_5966E0
			CPatch::PatchPointer(0x596E30 + 3, CModelInfo__ms_modelInfoPtrs.gta_iii);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_5966E0
		}
		#endif

		// TXD limit
		#if TRUE
		{
			CPatch::PatchUINT32(0x406977 + 2, GetFileIDcurrentLimit(FILE_TYPE_TXD));	// cmp     esi, 352h	; _ZN10CStreaming4InitEv
			CPatch::PatchUINT32(0x527457 + 1, GetFileIDcurrentLimit(FILE_TYPE_TXD));	// push    850	; sub_527440
			CPatch::PatchUINT32(0x5274CF + 2, GetFileIDcurrentLimit(FILE_TYPE_TXD));	// cmp     ebx, 352h	; sub_527490
			CPatch::PatchUINT32(0x592C98 + 1, GetFileIDcurrentLimit(FILE_TYPE_TXD));	// push    352h	; sub_592C70
			CPatch::PatchUINT32(0x592F22 + 2, GetFileIDcurrentLimit(FILE_TYPE_TXD));	// cmp     ebx, 352h	; sub_592C70
		}
		#endif
	}
	#endif

	// Patch functions, so the last file ID doesn't have to be divisible by 
	#if TRUE
	{
		// Patch CStreaming::Init
		if(true)
		{
			CPatch::RedirectCode(0x406440, &patch_CStreaming__Init_406440);

			CPatch::RedirectCodeOneByte(0x40667F, 0x4066B9);
			CPatch::NOPinstructions(0x4066C3, 7);
			CPatch::RedirectCodeOneByte(0x4066D4, 0x40670E);
			CPatch::NOPinstructions(0x406718, 3);
			CPatch::NOPinstructions(0x406725, 7);
			CPatch::RedirectCodeOneByte(0x406736, 0x406770);
			CPatch::NOPinstructions(0x40677A, 3);
			CPatch::RedirectCodeOneByte(0x406787, 0x4067D8);
			CPatch::NOPinstructions(0x4067E2, 7);
			CPatch::RedirectCodeOneByte(0x4067F3, 0x40682D);
			CPatch::NOPinstructions(0x406837, 7);
			CPatch::RedirectCodeOneByte(0x406845, 0x40687C);
		}

		// Patch LoadScene
		if(true)
		{

		}
	}
	#endif

	// finish
	CGenericLogStorage::WriteLineSeparator();
}
#endif

// GTA Vice City
namespace Game_GTAVC
{
	// Initializes CStreamingInfo
	void CStreamingInfo::Init()
	{
		this->uiLoadFlag = 0;
		this->usNext = 0;
		this->usPrev = 0;
		this->usNextOnCd = -1;
		this->_unk_inCdImage = 0;
		this->FilePositionInBlocks = 0;
	}

	void __cdecl patch_CStreaming__Init_doIt()
	{
		tFileTypeBaseID count = g_fileIDlimits.GetCountOfAllFileIDs();

		for (tFileTypeBaseID ID = 0; ID < count; ID++)
			g_fileIDlimits.CStreaming__ms_aInfoForModel.gta_vc[ID].Init();
	}

#ifdef IS_PLATFORM_WIN_X86
	void NAKED patch_CStreaming__Init_4100A8()
	{
		__asm
		{
			call patch_CStreaming__Init_doIt;

			// go back
			mov eax, 0x4102E0;
			jmp eax;
		}
	}
#endif


#ifdef IS_ARCHITECTURE_32_BIT
	void __cdecl LoadScene_initializeFlags()
	{
		uint16_t count = g_fileIDlimits.GetCountOfAllFileIDs();

		for (uint16_t ID = 0; ID < count; ID++)
			g_fileIDlimits.CStreaming__ms_aInfoForModel.gta_vc[ID].flags &= 0xDF;
	}
#endif

#ifdef IS_PLATFORM_WIN_X86
	void NAKED patch_LoadScene_40B3DA()
	{
		__asm
		{
			call LoadScene_initializeFlags;

			// go back
			mov eax, 0x40B4B2;
			jmp eax;
		}
	}
#endif
}

#ifdef IS_PLATFORM_WIN_X86
// Patches ID limits if neccessary.
void FileIDlimit::PatchFileIDlimits_GTA_VC_1_0()
{
	using namespace Game_GTAVC;

	if(!WriteToLogIfLimitsAreChanged())
		return;
	
	// ID limit checks
	#if TRUE		
	// See if count is unsigned
	if (GetCountOfAllFileIDs() >= 32768)
		throw f92_runtime_error("Count of file IDs should be lower than 32768."
			"\nUnsigned ID for GTA VC is not supported yet."
			"\nCount of IDs: %d", GetCountOfAllFileIDs()
		);
	#endif

	CGenericLogStorage::SaveFormattedTextLn(
		"\nNew ID limits:\n"
		"%d - %d (%d) - DFF models defined within IDE files\n"
		"%d - %d (%d) - TXD texture archives.\n"
		"%d - %d (%d) - COL collision archives.\n"
		"%d - %d (%d) - IFP animation archives.\n"
		"%d - count of all file IDs\n"
		,

		GetBaseID(FILE_TYPE_DFF), GetBaseID(FILE_TYPE_DFF), GetFileIDcurrentLimit(FILE_TYPE_DFF),
		GetBaseID(FILE_TYPE_TXD), GetBaseID(FILE_TYPE_TXD), GetFileIDcurrentLimit(FILE_TYPE_TXD),
		GetBaseID(FILE_TYPE_COL), GetBaseID(FILE_TYPE_COL), GetFileIDcurrentLimit(FILE_TYPE_COL),
		GetBaseID(FILE_TYPE_IFP), GetBaseID(FILE_TYPE_IFP), GetFileIDcurrentLimit(FILE_TYPE_IFP),

		GetCountOfAllFileIDs()
		);

	// Allocate variables
	#if TRUE
	if (!CPatch::IsDebugModeActive())
	{
		// CModelInfo::ms_modelInfoPtrs
		if(GetFileIDcurrentLimit(FILE_TYPE_DFF) > FileIDlimit::GetFileIDdefaultLimit(FILE_TYPE_DFF))
		{
			CModelInfo__ms_modelInfoPtrs.gta_vc = new CBaseModelInfo*[GetFileIDcurrentLimit(FILE_TYPE_DFF)];

			memset(
				CModelInfo__ms_modelInfoPtrs.gta_vc,
				NULL,
				GetFileIDcurrentLimit(FILE_TYPE_DFF) * sizeof(CBaseModelInfo*)
			);

			CModelInfo__ms_modelInfoPtrs.bIsAllocated.Set(true);
		}

		// CStreaming::ms_aInfoForModel
		if(GetCountOfAllFileIDs() > GetDefaultCountOfAllFileIDs())
		{
			CStreaming__ms_aInfoForModel.gta_vc = new CStreamingInfo[GetCountOfAllFileIDs()];

			memset(
				CStreaming__ms_aInfoForModel.gta_vc,
				NULL,
				GetCountOfAllFileIDs() * sizeof(CStreamingInfo)
			);

			CStreaming__ms_aInfoForModel.bIsAllocated.Set(true);
		}

		// CAnimManager::ms_aAnimBlocks
		if(GetFileIDcurrentLimit(FILE_TYPE_IFP) > FileIDlimit::GetFileIDdefaultLimit(FILE_TYPE_IFP))
		{
			CAnimManager__ms_aAnimBlocks.gta_vc = new CAnimBlock[GetFileIDcurrentLimit(FILE_TYPE_IFP)];

			memset(
				CAnimManager__ms_aAnimBlocks.gta_vc,
				NULL,
				GetFileIDcurrentLimit(FILE_TYPE_IFP) * sizeof(CAnimBlock)
			);

			CAnimManager__ms_aAnimBlocks.bIsAllocated.Set(true);
		}
	}
	#endif		

	// Patch pointers to CModelInfo::ms_modelInfoPtrs
	#if TRUE
	{
		CPatch::PatchPointer(0x4014EF + 1, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x40150B + 2, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x401530 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     esi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4054C7 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ebp, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x406557 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x406562 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4065B6 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4065C1 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x406657 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x406662 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x40676E + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4067A3 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4067B9 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x406851 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ebx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x406A74 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     esi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x407DC2 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x408AEF + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x408B1A + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x408BBB + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x408BF6 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x408C61 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x408D16 + 1, CModelInfo__ms_modelInfoPtrs.gta_vc + MODEL_ENFORCER);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE+274h
		CPatch::PatchPointer(0x408D84 + 1, CModelInfo__ms_modelInfoPtrs.gta_vc + MODEL_SWAT);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE+8
		CPatch::PatchPointer(0x408E20 + 1, CModelInfo__ms_modelInfoPtrs.gta_vc + MODEL_FBIRANCH);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE+370h
		CPatch::PatchPointer(0x408E94 + 1, CModelInfo__ms_modelInfoPtrs.gta_vc + MODEL_FBI);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE+0Ch
		CPatch::PatchPointer(0x408F37 + 1, CModelInfo__ms_modelInfoPtrs.gta_vc + MODEL_BARRACKS);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE+28Ch
		CPatch::PatchPointer(0x408F9B + 1, CModelInfo__ms_modelInfoPtrs.gta_vc + MODEL_RHINO);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE+288h
		CPatch::PatchPointer(0x409011 + 1, CModelInfo__ms_modelInfoPtrs.gta_vc + MODEL_ARMY);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE+10h
		CPatch::PatchPointer(0x4090A0 + 1, CModelInfo__ms_modelInfoPtrs.gta_vc + MODEL_CHOPPER);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE+294h
		CPatch::PatchPointer(0x409123 + 1, CModelInfo__ms_modelInfoPtrs.gta_vc + MODEL_VICE1);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE+184h
		CPatch::PatchPointer(0x409188 + 1, CModelInfo__ms_modelInfoPtrs.gta_vc + MODEL_VICE2);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE+188h
		CPatch::PatchPointer(0x4091E9 + 1, CModelInfo__ms_modelInfoPtrs.gta_vc + MODEL_VICE3);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE+18Ch
		CPatch::PatchPointer(0x40924A + 1, CModelInfo__ms_modelInfoPtrs.gta_vc + MODEL_VICE4);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE+190h
		CPatch::PatchPointer(0x4092AB + 1, CModelInfo__ms_modelInfoPtrs.gta_vc + MODEL_VICE5);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE+194h
		CPatch::PatchPointer(0x40930C + 1, CModelInfo__ms_modelInfoPtrs.gta_vc + MODEL_VICE6);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE+198h
		CPatch::PatchPointer(0x40936D + 1, CModelInfo__ms_modelInfoPtrs.gta_vc + MODEL_VICE7);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE+19Ch
		CPatch::PatchPointer(0x4093CE + 1, CModelInfo__ms_modelInfoPtrs.gta_vc + MODEL_VICE8);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE+1A0h
		CPatch::PatchPointer(0x4094D0 + 1, CModelInfo__ms_modelInfoPtrs.gta_vc + MODEL_VICECHEE);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE+3B0h
		CPatch::PatchPointer(0x40953B + 1, CModelInfo__ms_modelInfoPtrs.gta_vc + MODEL_VICE1);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE+184h
		CPatch::PatchPointer(0x40959C + 1, CModelInfo__ms_modelInfoPtrs.gta_vc + MODEL_VICE2);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE+188h
		CPatch::PatchPointer(0x4095FD + 1, CModelInfo__ms_modelInfoPtrs.gta_vc + MODEL_VICE3);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE+18Ch
		CPatch::PatchPointer(0x40965E + 1, CModelInfo__ms_modelInfoPtrs.gta_vc + MODEL_VICE4);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE+190h
		CPatch::PatchPointer(0x4096BF + 1, CModelInfo__ms_modelInfoPtrs.gta_vc + MODEL_VICE5);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE+194h
		CPatch::PatchPointer(0x409720 + 1, CModelInfo__ms_modelInfoPtrs.gta_vc + MODEL_VICE6);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE+198h
		CPatch::PatchPointer(0x409788 + 1, CModelInfo__ms_modelInfoPtrs.gta_vc + MODEL_VICE7);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE+19Ch
		CPatch::PatchPointer(0x4097E9 + 1, CModelInfo__ms_modelInfoPtrs.gta_vc + MODEL_VICE8);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE+1A0h
		CPatch::PatchPointer(0x4098E8 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x40999B + 3, CModelInfo__ms_modelInfoPtrs.gta_vc + MODEL_SPECIAL01);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE+1B4h[ecx*4]
		CPatch::PatchPointer(0x409AA6 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x409B35 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x409B97 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x409BDC + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ebp, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x409C98 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x409CFD + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x409D40 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x409DD7 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x409E3D + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x409E80 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x409F17 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x409F7D + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x409FC0 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x40A0E6 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x40A147 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x40A18A + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x40A2BD + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x40A364 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x40A3AD + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x40A40C + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x40A451 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x40A4BD + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x40A620 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x40A6A8 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x40A70D + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x40A750 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x40A7BB + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x40A7FE + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x40A8D9 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x40A93E + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x40A981 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x40AA76 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ebx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x40AA89 + 2, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     esi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x40AD87 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x40ADF9 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x40AE39 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x40AEC7 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x40B07F + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ebp, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x40B821 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x40BA3A + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x40BAF8 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x40BD19 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x40BF81 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x40BFDA + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x40C24D + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x40C308 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x40C338 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x40C39C + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x40C6BA + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x40C7B1 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x40C7F6 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x40C90C + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x40C9F7 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x40CA33 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x40CA63 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x40CAC5 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x40CB04 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x40CB2C + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x40CBAD + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x40CC35 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x40D5ED + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x40D70E + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x40D935 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     esi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x40D980 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ebp, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x40DA49 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ebp, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x40DD6D + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ebp, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x40DE08 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ebp, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x40DE9E + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ebp, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x40DF38 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ebp, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x40E0C9 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ebp, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x40E384 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x40E431 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x40E825 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x40E901 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     esi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x40FF89 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x40FFC0 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4105C0 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ebx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x41FDB8 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x41FDCD + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x420D3A + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x422FB4 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x42424A + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x424271 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x424C8C + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     esi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x424E96 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4276CB + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x427A35 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x427AF1 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x428C8F + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ebx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x428CF1 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x428D6E + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x428E1B + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x428F26 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; jumptable 00428F1F case 0
		CPatch::PatchPointer(0x428F38 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; jumptable 00428F1F case 1
		CPatch::PatchPointer(0x428F4A + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; jumptable 00428F1F case 2
		CPatch::PatchPointer(0x428F5C + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; jumptable 00428F1F case 3
		CPatch::PatchPointer(0x429945 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4299F1 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x429CCA + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x42D6B1 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x42F738 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x42F963 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x42FB88 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x42FE68 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x430009 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x43019B + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x430270 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4304B0 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x430720 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x431380 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4313A9 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4345D7 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x43D2E5 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x43D3D6 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x43D827 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x43F280 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x43F9F6 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x43FAD9 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x43FD6A + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x43FDB2 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x43FDDE + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4408A2 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x440A28 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x440D14 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4410E4 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x443709 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4438CB + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x443D26 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x443F79 + 2, CModelInfo__ms_modelInfoPtrs.gta_vc + MODEL_POLICE);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE+270h
		CPatch::PatchPointer(0x443F87 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x44549F + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x455442 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4555DD + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     esi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x45576D + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x456FB3 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x45C31C + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x45C88A + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ebp, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x45CB88 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x45D0EA + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ebp, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x45DFE7 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x45DFFB + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x45E01A + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x45E02A + 1, CModelInfo__ms_modelInfoPtrs.gta_vc + MODEL_MALE01);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE+1Ch
		CPatch::PatchPointer(0x45E863 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x45F1C1 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     esi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x45F27E + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x45F2CF + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x45F2E3 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x45F308 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x45F31F + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ebp, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x464186 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x467062 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x470D4D + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x470D6A + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x474C12 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x474C19 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x474D37 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x474D3E + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x474E66 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x474E6D + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x475EAA + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4775E0 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x47FF7B + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x47FF94 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x48111A + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x481133 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x483219 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x484476 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x484571 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x484B91 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x484C8C + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x485E26 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x487C78 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ebp, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x487D14 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x488263 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ebp, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4883CD + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4885DA + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4885FF + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x48874F + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x488774 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x48881D + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     esi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x488CAA + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     esi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x489274 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ebp, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x48938F + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4893D7 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x489439 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x489724 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x48979D + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4897F4 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x489819 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ebx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x48AC94 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x48ADA6 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x48AF5C + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x48BAE7 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ebp, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x48D500 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x48D545 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x48D56B + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x48D5AC + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x48D608 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x48D65A + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x48DD04 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x48E976 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4AE525 + 2, CModelInfo__ms_modelInfoPtrs.gta_vc + MODEL_INFERNUS);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE+234h
		CPatch::PatchPointer(0x4AE541 + 2, CModelInfo__ms_modelInfoPtrs.gta_vc + MODEL_CHEETAH);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE+244h
		CPatch::PatchPointer(0x4AE55D + 2, CModelInfo__ms_modelInfoPtrs.gta_vc + MODEL_PHEONIX);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE+33Ch
		CPatch::PatchPointer(0x4AE579 + 2, CModelInfo__ms_modelInfoPtrs.gta_vc + MODEL_DELUXO);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE+34Ch
		CPatch::PatchPointer(0x4AE850 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// cmp     ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4], 0; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4AF259 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4AF26C + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4AF35A + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4B0315 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ebx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4B05A0 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     esi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4B0761 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4B10D3 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ebp, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4B1387 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     esi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4B155B + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4B3533 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4B36C7 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4B684B + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     esi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4B6914 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ebp, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4B6E99 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4B7045 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ebp, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4B9950 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4B9B05 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4BB5E9 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4BEF2D + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     esi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4C4000 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4C4160 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4C762E + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     esi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4C7808 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     esi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4C7936 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ebp, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4C7983 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4C79B5 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4C7ACA + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4C7D7A + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ebx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4C7E84 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ebx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4C7EAD + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4C7EE9 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4C9F94 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4CA1CC + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4D1472 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4D3079 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4D3E6F + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4D3E92 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4D455B + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4D461C + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4D4664 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4D4775 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     esi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4D47FB + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4D572B + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4D5AF0 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4D5EE6 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4D632A + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4D6F1F + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4D6F9F + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     esi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4D7026 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4D709F + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     esi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4D7126 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4D71A0 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4D7263 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4D7313 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4D73A0 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4D7420 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4D7CF8 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4D7FB8 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4D853D + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4D85D0 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4D866C + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4D86FD + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4D879B + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4D882C + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4D88CB + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4D895C + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4D89FB + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4D8A8C + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4D8E2B + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4D8E4C + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4E097B + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4E0C3C + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4E0C90 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4E1BE5 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4E27C4 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4E2A0D + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4E2C5D + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4E3110 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4E3360 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4E3F97 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4E40AF + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4E4498 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4E8A59 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ebp, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4EA1D5 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     esi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4EA36A + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     esi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4EFBDF + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ebp, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4EFE27 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4F0624 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4F091B + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4F3740 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4F382E + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4F3925 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4F3A24 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4F44CE + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ebx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4F4EF0 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4F57EB + 2, CModelInfo__ms_modelInfoPtrs.gta_vc + MODEL_CELLPHONE);	// mov     ebp, ds:_ZN10CModelInfo16ms_modelInfoPtrsE+408h
		CPatch::PatchPointer(0x4F79B0 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4F79F7 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     esi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4F94D2 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4F94E7 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4FA925 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4FAB8D + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4FBDF2 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4FD1C9 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     esi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4FD1DE + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4FF2FB + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     esi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4FF608 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     esi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4FF64F + 2, CModelInfo__ms_modelInfoPtrs.gta_vc + MODEL_MINIGUN2);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE+498h
		CPatch::PatchPointer(0x4FF9AB + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     esi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4FF9F2 + 2, CModelInfo__ms_modelInfoPtrs.gta_vc + MODEL_MINIGUN2);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE+498h
		CPatch::PatchPointer(0x4FFB9A + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4FFBF0 + 2, CModelInfo__ms_modelInfoPtrs.gta_vc + MODEL_MINIGUN2);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE+498h
		CPatch::PatchPointer(0x4FFD99 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     esi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4FFE5D + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     esi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x4FFEA4 + 2, CModelInfo__ms_modelInfoPtrs.gta_vc + MODEL_MINIGUN2);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE+498h
		CPatch::PatchPointer(0x5019CA + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x504DE8 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x504F43 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x504FBE + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x50549D + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x5054B5 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x50605D + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     esi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x506082 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x506AB2 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x50D930 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ebp, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x511AA8 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x511B57 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x51276B + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x512D34 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x5130B6 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x514670 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ebp, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x514A84 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     esi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x515D03 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x5164EB + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x517700 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x51781A + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x523F2C + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x5283E3 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ebp, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x5283FC + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x528451 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x538B50 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x538BE9 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x538CB4 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x538DF5 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x53948A + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x5394BC + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x539C6B + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x539C9A + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x53A2A5 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x53A3D6 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x53AA99 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x53AB82 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x53ABBD + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x53ABEC + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x53AC01 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x53AC11 + 2, CModelInfo__ms_modelInfoPtrs.gta_vc + MODEL_MALE01);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE+1Ch
		CPatch::PatchPointer(0x53AF0E + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x53AF52 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x53AF63 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ebp, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x53B430 + 2, CModelInfo__ms_modelInfoPtrs.gta_vc + MODEL_MALE01);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE+1Ch
		CPatch::PatchPointer(0x53B56B + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x53B585 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x53C07F + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x53C1B9 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x53C604 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x53CA53 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x53CF99 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x53D0BC + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x53D19B + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x53D21B + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x53D2D1 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x53D396 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x53D4D7 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x53D916 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x53D98C + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x5415F3 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x54B10C + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x5526B6 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ebp, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x552700 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x552E39 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     esi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x553155 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x553AFD + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x553C6A + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x554473 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x5584EE + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x55F4E4 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x55F514 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x55F544 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x55F5BF + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4], eax; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x55F633 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4], eax; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x55F679 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4], eax; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x55F6CE + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4], ebp; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x55F724 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4], ebp; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x55F774 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4], ebp; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x55F796 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     esi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x55F7E0 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     esi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x55FAC0 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4], 0; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x55FACB + 3, CModelInfo__ms_modelInfoPtrs.gta_vc + MODEL_COP);	// mov     ds:_ZN10CModelInfo16ms_modelInfoPtrsE+4[eax*4], 0
		CPatch::PatchPointer(0x55FAD6 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc + MODEL_SWAT);	// mov     ds:_ZN10CModelInfo16ms_modelInfoPtrsE+8[eax*4], 0
		CPatch::PatchPointer(0x55FAE1 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc + MODEL_FBI);	// mov     ds:_ZN10CModelInfo16ms_modelInfoPtrsE+0Ch[eax*4], 0
		CPatch::PatchPointer(0x55FAEC + 3, CModelInfo__ms_modelInfoPtrs.gta_vc + MODEL_ARMY);	// mov     ds:_ZN10CModelInfo16ms_modelInfoPtrsE+10h[eax*4], 0
		CPatch::PatchPointer(0x55FAF7 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc + MODEL_MEDIC);	// mov     ds:_ZN10CModelInfo16ms_modelInfoPtrsE+14h[eax*4], 0
		CPatch::PatchPointer(0x55FB02 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc + MODEL_FIREMAN);	// mov     ds:_ZN10CModelInfo16ms_modelInfoPtrsE+18h[eax*4], 0
		CPatch::PatchPointer(0x55FB0D + 3, CModelInfo__ms_modelInfoPtrs.gta_vc + MODEL_MALE01);	// mov     ds:_ZN10CModelInfo16ms_modelInfoPtrsE+1Ch[eax*4], 0
		CPatch::PatchPointer(0x55FB40 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4], 0; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x55FB55 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc + MODEL_COP);	// mov     ds:_ZN10CModelInfo16ms_modelInfoPtrsE+4[eax*4], 0
		CPatch::PatchPointer(0x55FB6A + 3, CModelInfo__ms_modelInfoPtrs.gta_vc + MODEL_SWAT);	// mov     ds:_ZN10CModelInfo16ms_modelInfoPtrsE+8[eax*4], 0
		CPatch::PatchPointer(0x55FB7F + 3, CModelInfo__ms_modelInfoPtrs.gta_vc + MODEL_FBI);	// mov     ds:_ZN10CModelInfo16ms_modelInfoPtrsE+0Ch[eax*4], 0
		CPatch::PatchPointer(0x55FBD2 + 2, CModelInfo__ms_modelInfoPtrs.gta_vc + 240);	// mov     ds:_ZN10CModelInfo16ms_modelInfoPtrsE+3C0h, ebx
		CPatch::PatchPointer(0x55FC36 + 2, CModelInfo__ms_modelInfoPtrs.gta_vc + 241);	// mov     ds:_ZN10CModelInfo16ms_modelInfoPtrsE+3C4h, ebx
		CPatch::PatchPointer(0x55FC9A + 2, CModelInfo__ms_modelInfoPtrs.gta_vc + 242);	// mov     ds:_ZN10CModelInfo16ms_modelInfoPtrsE+3C8h, ebx
		CPatch::PatchPointer(0x55FCFE + 2, CModelInfo__ms_modelInfoPtrs.gta_vc + 243);	// mov     ds:_ZN10CModelInfo16ms_modelInfoPtrsE+3CCh, ebx
		CPatch::PatchPointer(0x55FD62 + 2, CModelInfo__ms_modelInfoPtrs.gta_vc + 244);	// mov     ds:_ZN10CModelInfo16ms_modelInfoPtrsE+3D0h, ebx
		CPatch::PatchPointer(0x55FDC6 + 2, CModelInfo__ms_modelInfoPtrs.gta_vc + 245);	// mov     ds:_ZN10CModelInfo16ms_modelInfoPtrsE+3D4h, ebx
		CPatch::PatchPointer(0x55FE2A + 2, CModelInfo__ms_modelInfoPtrs.gta_vc + 246);	// mov     ds:_ZN10CModelInfo16ms_modelInfoPtrsE+3D8h, ebx
		CPatch::PatchPointer(0x55FE8E + 2, CModelInfo__ms_modelInfoPtrs.gta_vc + 247);	// mov     ds:_ZN10CModelInfo16ms_modelInfoPtrsE+3DCh, ebx
		CPatch::PatchPointer(0x567C5C + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x569186 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x569D34 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x56BA21 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x56C0B4 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x56C271 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x56E10F + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x56F3B2 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x56F470 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x578A74 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x579221 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// cmp     ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4], ebx; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x579DC4 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x579EA7 + 2, CModelInfo__ms_modelInfoPtrs.gta_vc + MODEL_RHINO);	// cmp     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE+288h
		CPatch::PatchPointer(0x57A916 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x58137F + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ebp, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x585B8F + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x585D8F + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ebx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x586120 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x587229 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x58724D + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x587269 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x587283 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x5874E3 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x5874FC + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x5878F8 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     esi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x587A0F + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ebp, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x587D8D + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x58833B + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x5883EA + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x588863 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     esi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x588B46 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x588BD3 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x589534 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x5895CC + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x5897AB + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x58B293 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x58B5C4 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x58EEBB + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ebx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x591AA4 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x592142 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x592220 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x59223B + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x592256 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x59227E + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x593084 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x59B0DB + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x59B0F1 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x59B181 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x59B397 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x59BB0F + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x59BC88 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x59D27D + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x59E2BF + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ebp, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x59E6CC + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x5A154D + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x5A1B7F + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x5A2AD9 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x5A2C5F + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x5A2DE9 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x5A2F7D + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x5A30F9 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x5A326C + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x5A4DF5 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x5A64C3 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x5A6D54 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x5A6FCB + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x5A703A + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x5A73D2 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x5A8713 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x5A87CF + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ebp, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x5A8EAE + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ebp, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x5AA8C1 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x5AF811 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x5B18B2 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x5B194C + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x5B266D + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x5B2A2D + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x5B2B71 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x5B410B + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x5B44B6 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x5B4846 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x5B485C + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x5B5D23 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x5B7E19 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x5B7F43 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x5BE34C + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x5C88D1 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x5CAB96 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x5CB0D8 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x5CB2D8 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x5CB479 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x5D2EF7 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x5D2FB2 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x5D4CA4 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x5D4CC1 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x5D4D83 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x5D4D94 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x5D56C3 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x6047A3 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x604820 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x604D47 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ebp, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x604DC2 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ebp, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x6079BB + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// cmp     ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4], 0; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x6097DB + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x609925 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x609D67 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x60A254 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x60AB1E + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     esi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x60ABA9 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x60AD57 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x60AE26 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ebp, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x60B12B + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     esi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x60B503 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x60B9EA + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x60BFBE + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x60C5D2 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ebp, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x60D193 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x60DDB4 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x60E40E + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x60E426 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x61227B + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x615097 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x6157B5 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     esi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x61FDDB + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     esi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x61FE22 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x62910D + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x62A780 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x62A9A2 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x632086 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x63318F + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x633310 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x634BE2 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x634C4B + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x634D62 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     esi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs
		CPatch::PatchPointer(0x634D98 + 3, CModelInfo__ms_modelInfoPtrs.gta_vc);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs
	}
	#endif

	// Patch pointers to CStreaming::ms_aInfoForModel
	#if TRUE
	{
		CPatch::PatchPointer(0x406C56 + 4, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// movzx   ebx, ds:_ZN10CStreaming16ms_aInfoForModelE.flags[eax*4]
		CPatch::PatchPointer(0x408A30 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> uiLoadFlag);	// cmp     ss:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[ebp], 1
		CPatch::PatchPointer(0x408A40 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// mov     ss:_ZN10CStreaming16ms_aInfoForModelE.flags[ebp], al
		CPatch::PatchPointer(0x408A73 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[ebx], 1
		CPatch::PatchPointer(0x408A7C + 2, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// mov     al, ds:_ZN10CStreaming16ms_aInfoForModelE.flags[ebx]
		CPatch::PatchPointer(0x408AE3 + 3, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// movzx   eax, ds:_ZN10CStreaming16ms_aInfoForModelE.flags[edx]
		CPatch::PatchPointer(0x408AFD + 2, &CStreaming__ms_aInfoForModel.gta_vc -> uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[edx], 1
		CPatch::PatchPointer(0x408BAE + 4, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// movzx   edx, ds:_ZN10CStreaming16ms_aInfoForModelE.flags[ecx*4]
		CPatch::PatchPointer(0x408D1B + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_ENFORCER].flags);	// and     ds:_ZN10CStreaming16ms_aInfoForModelE.flags+0C44h, 0FEh
		CPatch::PatchPointer(0x408D28 + 3, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_ENFORCER].flags);	// movzx   eax, ds:_ZN10CStreaming16ms_aInfoForModelE.flags+0C44h
		CPatch::PatchPointer(0x408D34 + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_ENFORCER].uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+0C44h, 1
		CPatch::PatchPointer(0x408D3D + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_ENFORCER].usNext);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.usNext+0C44h, 0
		CPatch::PatchPointer(0x408D4B + 1, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_ENFORCER].usNext);	// mov     edx, (offset _ZN10CStreaming16ms_aInfoForModelE.usNext+0C44h)
		CPatch::PatchPointer(0x408D50 + 1, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_ENFORCER].usNext);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usNext+0C44h, eax
		CPatch::PatchPointer(0x408D55 + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_ENFORCER].usPrev);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usPrev+0C44h, offset dword_A0DAE0
		CPatch::PatchPointer(0x408D5F + 1, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_ENFORCER].usNext);	// mov     eax, ds:_ZN10CStreaming16ms_aInfoForModelE.usNext+0C44h
		CPatch::PatchPointer(0x408D7B + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_ENFORCER].uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+0C44h, 1
		CPatch::PatchPointer(0x408D89 + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_SWAT].flags);	// and     ds:_ZN10CStreaming16ms_aInfoForModelE.flags+28h, 0FEh
		CPatch::PatchPointer(0x408D96 + 3, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_SWAT].flags);	// movzx   eax, ds:_ZN10CStreaming16ms_aInfoForModelE.flags+28h
		CPatch::PatchPointer(0x408DA2 + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_SWAT].uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+28h, 1
		CPatch::PatchPointer(0x408DAB + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_SWAT].usNext);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.usNext+28h, 0
		CPatch::PatchPointer(0x408DB9 + 1, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_SWAT].usNext);	// mov     esi, (offset _ZN10CStreaming16ms_aInfoForModelE.usNext+28h)
		CPatch::PatchPointer(0x408DBE + 1, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_SWAT].usNext);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usNext+28h, eax
		CPatch::PatchPointer(0x408DC3 + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_SWAT].usPrev);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usPrev+28h, offset dword_A0DAE0
		CPatch::PatchPointer(0x408DCD + 1, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_SWAT].usNext);	// mov     eax, ds:_ZN10CStreaming16ms_aInfoForModelE.usNext+28h
		CPatch::PatchPointer(0x408E25 + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_FBIRANCH].flags);	// and     ds:_ZN10CStreaming16ms_aInfoForModelE.flags+1130h, 0FEh
		CPatch::PatchPointer(0x408E32 + 3, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_FBIRANCH].flags);	// movzx   eax, ds:_ZN10CStreaming16ms_aInfoForModelE.flags+1130h
		CPatch::PatchPointer(0x408E3E + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_FBIRANCH].uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+1130h, 1
		CPatch::PatchPointer(0x408E47 + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_FBIRANCH].usNext);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.usNext+1130h, 0
		CPatch::PatchPointer(0x408E55 + 1, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_FBIRANCH].usNext);	// mov     edi, (offset _ZN10CStreaming16ms_aInfoForModelE.usNext+1130h)
		CPatch::PatchPointer(0x408E5A + 1, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_FBIRANCH].usNext);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usNext+1130h, eax
		CPatch::PatchPointer(0x408E5F + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_FBIRANCH].usPrev);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usPrev+1130h, offset dword_A0DAE0
		CPatch::PatchPointer(0x408E69 + 1, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_FBIRANCH].usNext);	// mov     eax, ds:_ZN10CStreaming16ms_aInfoForModelE.usNext+1130h
		CPatch::PatchPointer(0x408E8B + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_FBIRANCH].uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+1130h, 1
		CPatch::PatchPointer(0x408E99 + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_FBI].flags);	// and     ds:_ZN10CStreaming16ms_aInfoForModelE.flags+3Ch, 0FEh
		CPatch::PatchPointer(0x408EA6 + 3, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_FBI].flags);	// movzx   eax, ds:_ZN10CStreaming16ms_aInfoForModelE.flags+3Ch
		CPatch::PatchPointer(0x408EB2 + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_FBI].uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+3Ch, 1
		CPatch::PatchPointer(0x408EBB + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_FBI].usNext);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.usNext+3Ch, 0
		CPatch::PatchPointer(0x408EC9 + 1, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_FBI].usNext);	// mov     ebp, (offset _ZN10CStreaming16ms_aInfoForModelE.usNext+3Ch)
		CPatch::PatchPointer(0x408ECE + 1, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_FBI].usNext);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usNext+3Ch, eax
		CPatch::PatchPointer(0x408ED3 + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_FBI].usPrev);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usPrev+3Ch, offset dword_A0DAE0
		CPatch::PatchPointer(0x408EDD + 1, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_FBI].usNext);	// mov     eax, ds:_ZN10CStreaming16ms_aInfoForModelE.usNext+3Ch
		CPatch::PatchPointer(0x408F3C + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_BARRACKS].flags);	// and     ds:_ZN10CStreaming16ms_aInfoForModelE.flags+0CBCh, 0FEh
		CPatch::PatchPointer(0x408F49 + 3, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_BARRACKS].flags);	// movzx   eax, ds:_ZN10CStreaming16ms_aInfoForModelE.flags+0CBCh
		CPatch::PatchPointer(0x408F55 + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_BARRACKS].uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+0CBCh, 1
		CPatch::PatchPointer(0x408F5E + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_BARRACKS].usNext);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.usNext+0CBCh, 0
		CPatch::PatchPointer(0x408F6C + 1, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_BARRACKS].usNext);	// mov     ecx, (offset _ZN10CStreaming16ms_aInfoForModelE.usNext+0CBCh)
		CPatch::PatchPointer(0x408F71 + 1, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_BARRACKS].usNext);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usNext+0CBCh, eax
		CPatch::PatchPointer(0x408F76 + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_BARRACKS].usPrev);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usPrev+0CBCh, offset dword_A0DAE0
		CPatch::PatchPointer(0x408F80 + 1, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_BARRACKS].usNext);	// mov     eax, ds:_ZN10CStreaming16ms_aInfoForModelE.usNext+0CBCh
		CPatch::PatchPointer(0x408FA0 + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_RHINO].flags);	// and     ds:_ZN10CStreaming16ms_aInfoForModelE.flags+0CA8h, 0FEh
		CPatch::PatchPointer(0x408FAD + 3, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_RHINO].flags);	// movzx   eax, ds:_ZN10CStreaming16ms_aInfoForModelE.flags+0CA8h
		CPatch::PatchPointer(0x408FB9 + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_RHINO].uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+0CA8h, 1
		CPatch::PatchPointer(0x408FC2 + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_RHINO].usNext);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.usNext+0CA8h, 0
		CPatch::PatchPointer(0x408FD0 + 1, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_RHINO].usNext);	// mov     edx, (offset _ZN10CStreaming16ms_aInfoForModelE.usNext+0CA8h)
		CPatch::PatchPointer(0x408FD5 + 1, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_RHINO].usNext);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usNext+0CA8h, eax
		CPatch::PatchPointer(0x408FDA + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_RHINO].usPrev);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usPrev+0CA8h, offset dword_A0DAE0
		CPatch::PatchPointer(0x408FE4 + 1, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_RHINO].usNext);	// mov     eax, ds:_ZN10CStreaming16ms_aInfoForModelE.usNext+0CA8h
		CPatch::PatchPointer(0x408FFF + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_BARRACKS].uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+0CBCh, 1
		CPatch::PatchPointer(0x409008 + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_RHINO].uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+0CA8h, 1
		CPatch::PatchPointer(0x409016 + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_ARMY].flags);	// and     ds:_ZN10CStreaming16ms_aInfoForModelE.flags+50h, 0FEh
		CPatch::PatchPointer(0x409023 + 3, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_ARMY].flags);	// movzx   eax, ds:_ZN10CStreaming16ms_aInfoForModelE.flags+50h
		CPatch::PatchPointer(0x40902F + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_ARMY].uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+50h, 1
		CPatch::PatchPointer(0x409038 + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_ARMY].usNext);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.usNext+50h, 0
		CPatch::PatchPointer(0x409046 + 1, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_ARMY].usNext);	// mov     esi, (offset _ZN10CStreaming16ms_aInfoForModelE.usNext+50h)
		CPatch::PatchPointer(0x40904B + 1, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_ARMY].usNext);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usNext+50h, eax
		CPatch::PatchPointer(0x409050 + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_ARMY].usPrev);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usPrev+50h, offset dword_A0DAE0
		CPatch::PatchPointer(0x40905A + 1, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_ARMY].usNext);	// mov     eax, ds:_ZN10CStreaming16ms_aInfoForModelE.usNext+50h
		CPatch::PatchPointer(0x4090A5 + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_CHOPPER].flags);	// and     ds:_ZN10CStreaming16ms_aInfoForModelE.flags+0CE4h, 0FEh
		CPatch::PatchPointer(0x4090B2 + 3, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_CHOPPER].flags);	// movzx   eax, ds:_ZN10CStreaming16ms_aInfoForModelE.flags+0CE4h
		CPatch::PatchPointer(0x4090BE + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_CHOPPER].uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+0CE4h, 1
		CPatch::PatchPointer(0x4090C7 + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_CHOPPER].usNext);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.usNext+0CE4h, 0
		CPatch::PatchPointer(0x4090D5 + 1, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_CHOPPER].usNext);	// mov     edi, (offset _ZN10CStreaming16ms_aInfoForModelE.usNext+0CE4h)
		CPatch::PatchPointer(0x4090DA + 1, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_CHOPPER].usNext);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usNext+0CE4h, eax
		CPatch::PatchPointer(0x4090DF + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_CHOPPER].usPrev);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usPrev+0CE4h, offset dword_A0DAE0
		CPatch::PatchPointer(0x4090E9 + 1, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_CHOPPER].usNext);	// mov     eax, ds:_ZN10CStreaming16ms_aInfoForModelE.usNext+0CE4h
		CPatch::PatchPointer(0x409128 + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE1].flags);	// and     ds:_ZN10CStreaming16ms_aInfoForModelE.flags+794h, 0FEh
		CPatch::PatchPointer(0x409135 + 3, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE1].flags);	// movzx   eax, ds:_ZN10CStreaming16ms_aInfoForModelE.flags+794h
		CPatch::PatchPointer(0x409141 + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE1].uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+794h, 1
		CPatch::PatchPointer(0x40914A + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE1].usNext);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.usNext+794h, 0
		CPatch::PatchPointer(0x409158 + 1, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE1].usNext);	// mov     ebp, (offset _ZN10CStreaming16ms_aInfoForModelE.usNext+794h)
		CPatch::PatchPointer(0x40915D + 1, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE1].usNext);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usNext+794h, eax
		CPatch::PatchPointer(0x409162 + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE1].usPrev);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usPrev+794h, offset dword_A0DAE0
		CPatch::PatchPointer(0x40916C + 1, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE1].usNext);	// mov     eax, ds:_ZN10CStreaming16ms_aInfoForModelE.usNext+794h
		CPatch::PatchPointer(0x40918D + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE2].flags);	// and     ds:_ZN10CStreaming16ms_aInfoForModelE.flags+7A8h, 0FEh
		CPatch::PatchPointer(0x40919A + 3, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE2].flags);	// movzx   eax, ds:_ZN10CStreaming16ms_aInfoForModelE.flags+7A8h
		CPatch::PatchPointer(0x4091A6 + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE2].uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+7A8h, 1
		CPatch::PatchPointer(0x4091AF + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE2].usNext);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.usNext+7A8h, 0
		CPatch::PatchPointer(0x4091BD + 1, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE2].usNext);	// mov     ecx, (offset _ZN10CStreaming16ms_aInfoForModelE.usNext+7A8h)
		CPatch::PatchPointer(0x4091C2 + 1, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE2].usNext);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usNext+7A8h, eax
		CPatch::PatchPointer(0x4091C7 + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE2].usPrev);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usPrev+7A8h, offset dword_A0DAE0
		CPatch::PatchPointer(0x4091D1 + 1, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE2].usNext);	// mov     eax, ds:_ZN10CStreaming16ms_aInfoForModelE.usNext+7A8h
		CPatch::PatchPointer(0x4091EE + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE3].flags);	// and     ds:_ZN10CStreaming16ms_aInfoForModelE.flags+7BCh, 0FEh
		CPatch::PatchPointer(0x4091FB + 3, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE3].flags);	// movzx   eax, ds:_ZN10CStreaming16ms_aInfoForModelE.flags+7BCh
		CPatch::PatchPointer(0x409207 + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE3].uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+7BCh, 1
		CPatch::PatchPointer(0x409210 + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE3].usNext);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.usNext+7BCh, 0
		CPatch::PatchPointer(0x40921E + 1, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE3].usNext);	// mov     edx, (offset _ZN10CStreaming16ms_aInfoForModelE.usNext+7BCh)
		CPatch::PatchPointer(0x409223 + 1, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE3].usNext);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usNext+7BCh, eax
		CPatch::PatchPointer(0x409228 + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE3].usPrev);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usPrev+7BCh, offset dword_A0DAE0
		CPatch::PatchPointer(0x409232 + 1, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE3].usNext);	// mov     eax, ds:_ZN10CStreaming16ms_aInfoForModelE.usNext+7BCh
		CPatch::PatchPointer(0x40924F + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE4].flags);	// and     ds:_ZN10CStreaming16ms_aInfoForModelE.flags+7D0h, 0FEh
		CPatch::PatchPointer(0x40925C + 3, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE4].flags);	// movzx   eax, ds:_ZN10CStreaming16ms_aInfoForModelE.flags+7D0h
		CPatch::PatchPointer(0x409268 + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE4].uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+7D0h, 1
		CPatch::PatchPointer(0x409271 + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE4].usNext);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.usNext+7D0h, 0
		CPatch::PatchPointer(0x40927F + 1, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE4].usNext);	// mov     esi, (offset _ZN10CStreaming16ms_aInfoForModelE.usNext+7D0h)
		CPatch::PatchPointer(0x409284 + 1, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE4].usNext);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usNext+7D0h, eax
		CPatch::PatchPointer(0x409289 + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE4].usPrev);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usPrev+7D0h, offset dword_A0DAE0
		CPatch::PatchPointer(0x409293 + 1, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE4].usNext);	// mov     eax, ds:_ZN10CStreaming16ms_aInfoForModelE.usNext+7D0h
		CPatch::PatchPointer(0x4092B0 + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE5].flags);	// and     ds:_ZN10CStreaming16ms_aInfoForModelE.flags+7E4h, 0FEh
		CPatch::PatchPointer(0x4092BD + 3, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE5].flags);	// movzx   eax, ds:_ZN10CStreaming16ms_aInfoForModelE.flags+7E4h
		CPatch::PatchPointer(0x4092C9 + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE5].uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+7E4h, 1
		CPatch::PatchPointer(0x4092D2 + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE5].usNext);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.usNext+7E4h, 0
		CPatch::PatchPointer(0x4092E0 + 1, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE5].usNext);	// mov     edi, (offset _ZN10CStreaming16ms_aInfoForModelE.usNext+7E4h)
		CPatch::PatchPointer(0x4092E5 + 1, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE5].usNext);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usNext+7E4h, eax
		CPatch::PatchPointer(0x4092EA + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE5].usPrev);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usPrev+7E4h, offset dword_A0DAE0
		CPatch::PatchPointer(0x4092F4 + 1, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE5].usNext);	// mov     eax, ds:_ZN10CStreaming16ms_aInfoForModelE.usNext+7E4h
		CPatch::PatchPointer(0x409311 + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE6].flags);	// and     ds:_ZN10CStreaming16ms_aInfoForModelE.flags+7F8h, 0FEh
		CPatch::PatchPointer(0x40931E + 3, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE6].flags);	// movzx   eax, ds:_ZN10CStreaming16ms_aInfoForModelE.flags+7F8h
		CPatch::PatchPointer(0x40932A + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE6].uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+7F8h, 1
		CPatch::PatchPointer(0x409333 + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE6].usNext);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.usNext+7F8h, 0
		CPatch::PatchPointer(0x409341 + 1, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE6].usNext);	// mov     ebp, (offset _ZN10CStreaming16ms_aInfoForModelE.usNext+7F8h)
		CPatch::PatchPointer(0x409346 + 1, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE6].usNext);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usNext+7F8h, eax
		CPatch::PatchPointer(0x40934B + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE6].usPrev);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usPrev+7F8h, offset dword_A0DAE0
		CPatch::PatchPointer(0x409355 + 1, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE6].usNext);	// mov     eax, ds:_ZN10CStreaming16ms_aInfoForModelE.usNext+7F8h
		CPatch::PatchPointer(0x409372 + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE7].flags);	// and     ds:_ZN10CStreaming16ms_aInfoForModelE.flags+80Ch, 0FEh
		CPatch::PatchPointer(0x40937F + 3, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE7].flags);	// movzx   eax, ds:_ZN10CStreaming16ms_aInfoForModelE.flags+80Ch
		CPatch::PatchPointer(0x40938B + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE7].uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+80Ch, 1
		CPatch::PatchPointer(0x409394 + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE7].usNext);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.usNext+80Ch, 0
		CPatch::PatchPointer(0x4093A2 + 1, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE7].usNext);	// mov     edx, (offset _ZN10CStreaming16ms_aInfoForModelE.usNext+80Ch)
		CPatch::PatchPointer(0x4093A7 + 1, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE7].usNext);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usNext+80Ch, eax
		CPatch::PatchPointer(0x4093AC + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE7].usPrev);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usPrev+80Ch, offset dword_A0DAE0
		CPatch::PatchPointer(0x4093B6 + 1, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE7].usNext);	// mov     eax, ds:_ZN10CStreaming16ms_aInfoForModelE.usNext+80Ch
		CPatch::PatchPointer(0x4093D3 + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE8].flags);	// and     ds:_ZN10CStreaming16ms_aInfoForModelE.flags+820h, 0FEh
		CPatch::PatchPointer(0x4093E0 + 3, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE8].flags);	// movzx   eax, ds:_ZN10CStreaming16ms_aInfoForModelE.flags+820h
		CPatch::PatchPointer(0x4093EC + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE8].uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+820h, 1
		CPatch::PatchPointer(0x4093F5 + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE8].usNext);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.usNext+820h, 0
		CPatch::PatchPointer(0x409403 + 1, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE8].usNext);	// mov     esi, (offset _ZN10CStreaming16ms_aInfoForModelE.usNext+820h)
		CPatch::PatchPointer(0x409408 + 1, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE8].usNext);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usNext+820h, eax
		CPatch::PatchPointer(0x40940D + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE8].usPrev);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usPrev+820h, offset dword_A0DAE0
		CPatch::PatchPointer(0x409417 + 1, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE8].usNext);	// mov     eax, ds:_ZN10CStreaming16ms_aInfoForModelE.usNext+820h
		CPatch::PatchPointer(0x4094D5 + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICECHEE].flags);	// and     ds:_ZN10CStreaming16ms_aInfoForModelE.flags+1270h, 0FEh
		CPatch::PatchPointer(0x4094E2 + 3, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICECHEE].flags);	// movzx   eax, ds:_ZN10CStreaming16ms_aInfoForModelE.flags+1270h
		CPatch::PatchPointer(0x4094EE + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICECHEE].uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+1270h, 1
		CPatch::PatchPointer(0x4094F7 + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICECHEE].usNext);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.usNext+1270h, 0
		CPatch::PatchPointer(0x409505 + 1, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICECHEE].usNext);	// mov     edi, (offset _ZN10CStreaming16ms_aInfoForModelE.usNext+1270h)
		CPatch::PatchPointer(0x40950A + 1, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICECHEE].usNext);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usNext+1270h, eax
		CPatch::PatchPointer(0x40950F + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICECHEE].usPrev);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usPrev+1270h, offset dword_A0DAE0
		CPatch::PatchPointer(0x409519 + 1, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICECHEE].usNext);	// mov     eax, ds:_ZN10CStreaming16ms_aInfoForModelE.usNext+1270h
		CPatch::PatchPointer(0x409540 + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE1].flags);	// and     ds:_ZN10CStreaming16ms_aInfoForModelE.flags+794h, 0FEh
		CPatch::PatchPointer(0x40954D + 3, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE1].flags);	// movzx   eax, ds:_ZN10CStreaming16ms_aInfoForModelE.flags+794h
		CPatch::PatchPointer(0x409559 + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE1].uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+794h, 1
		CPatch::PatchPointer(0x409562 + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE1].usNext);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.usNext+794h, 0
		CPatch::PatchPointer(0x409570 + 1, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE1].usNext);	// mov     ebp, (offset _ZN10CStreaming16ms_aInfoForModelE.usNext+794h)
		CPatch::PatchPointer(0x409575 + 1, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE1].usNext);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usNext+794h, eax
		CPatch::PatchPointer(0x40957A + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE1].usPrev);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usPrev+794h, offset dword_A0DAE0
		CPatch::PatchPointer(0x409584 + 1, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE1].usNext);	// mov     eax, ds:_ZN10CStreaming16ms_aInfoForModelE.usNext+794h
		CPatch::PatchPointer(0x4095A1 + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE2].flags);	// and     ds:_ZN10CStreaming16ms_aInfoForModelE.flags+7A8h, 0FEh
		CPatch::PatchPointer(0x4095AE + 3, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE2].flags);	// movzx   eax, ds:_ZN10CStreaming16ms_aInfoForModelE.flags+7A8h
		CPatch::PatchPointer(0x4095BA + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE2].uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+7A8h, 1
		CPatch::PatchPointer(0x4095C3 + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE2].usNext);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.usNext+7A8h, 0
		CPatch::PatchPointer(0x4095D1 + 1, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE2].usNext);	// mov     edx, (offset _ZN10CStreaming16ms_aInfoForModelE.usNext+7A8h)
		CPatch::PatchPointer(0x4095D6 + 1, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE2].usNext);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usNext+7A8h, eax
		CPatch::PatchPointer(0x4095DB + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE2].usPrev);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usPrev+7A8h, offset dword_A0DAE0
		CPatch::PatchPointer(0x4095E5 + 1, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE2].usNext);	// mov     eax, ds:_ZN10CStreaming16ms_aInfoForModelE.usNext+7A8h
		CPatch::PatchPointer(0x409602 + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE3].flags);	// and     ds:_ZN10CStreaming16ms_aInfoForModelE.flags+7BCh, 0FEh
		CPatch::PatchPointer(0x40960F + 3, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE3].flags);	// movzx   eax, ds:_ZN10CStreaming16ms_aInfoForModelE.flags+7BCh
		CPatch::PatchPointer(0x40961B + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE3].uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+7BCh, 1
		CPatch::PatchPointer(0x409624 + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE3].usNext);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.usNext+7BCh, 0
		CPatch::PatchPointer(0x409632 + 1, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE3].usNext);	// mov     esi, (offset _ZN10CStreaming16ms_aInfoForModelE.usNext+7BCh)
		CPatch::PatchPointer(0x409637 + 1, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE3].usNext);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usNext+7BCh, eax
		CPatch::PatchPointer(0x40963C + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE3].usPrev);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usPrev+7BCh, offset dword_A0DAE0
		CPatch::PatchPointer(0x409646 + 1, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE3].usNext);	// mov     eax, ds:_ZN10CStreaming16ms_aInfoForModelE.usNext+7BCh
		CPatch::PatchPointer(0x409663 + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE4].flags);	// and     ds:_ZN10CStreaming16ms_aInfoForModelE.flags+7D0h, 0FEh
		CPatch::PatchPointer(0x409670 + 3, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE4].flags);	// movzx   eax, ds:_ZN10CStreaming16ms_aInfoForModelE.flags+7D0h
		CPatch::PatchPointer(0x40967C + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE4].uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+7D0h, 1
		CPatch::PatchPointer(0x409685 + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE4].usNext);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.usNext+7D0h, 0
		CPatch::PatchPointer(0x409693 + 1, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE4].usNext);	// mov     ecx, (offset _ZN10CStreaming16ms_aInfoForModelE.usNext+7D0h)
		CPatch::PatchPointer(0x409698 + 1, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE4].usNext);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usNext+7D0h, eax
		CPatch::PatchPointer(0x40969D + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE4].usPrev);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usPrev+7D0h, offset dword_A0DAE0
		CPatch::PatchPointer(0x4096A7 + 1, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE4].usNext);	// mov     eax, ds:_ZN10CStreaming16ms_aInfoForModelE.usNext+7D0h
		CPatch::PatchPointer(0x4096C4 + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE5].flags);	// and     ds:_ZN10CStreaming16ms_aInfoForModelE.flags+7E4h, 0FEh
		CPatch::PatchPointer(0x4096D1 + 3, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE5].flags);	// movzx   eax, ds:_ZN10CStreaming16ms_aInfoForModelE.flags+7E4h
		CPatch::PatchPointer(0x4096DD + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE5].uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+7E4h, 1
		CPatch::PatchPointer(0x4096E6 + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE5].usNext);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.usNext+7E4h, 0
		CPatch::PatchPointer(0x4096F4 + 1, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE5].usNext);	// mov     edi, (offset _ZN10CStreaming16ms_aInfoForModelE.usNext+7E4h)
		CPatch::PatchPointer(0x4096F9 + 1, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE5].usNext);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usNext+7E4h, eax
		CPatch::PatchPointer(0x4096FE + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE5].usPrev);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usPrev+7E4h, offset dword_A0DAE0
		CPatch::PatchPointer(0x409708 + 1, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE5].usNext);	// mov     eax, ds:_ZN10CStreaming16ms_aInfoForModelE.usNext+7E4h
		CPatch::PatchPointer(0x409725 + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE6].flags);	// and     ds:_ZN10CStreaming16ms_aInfoForModelE.flags+7F8h, 0FEh
		CPatch::PatchPointer(0x409732 + 3, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE6].flags);	// movzx   eax, ds:_ZN10CStreaming16ms_aInfoForModelE.flags+7F8h
		CPatch::PatchPointer(0x40973E + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE6].uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+7F8h, 1
		CPatch::PatchPointer(0x409747 + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE6].usNext);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.usNext+7F8h, 0
		CPatch::PatchPointer(0x409755 + 1, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE6].usNext);	// mov     edx, (offset _ZN10CStreaming16ms_aInfoForModelE.usNext+7F8h)
		CPatch::PatchPointer(0x40975A + 1, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE6].usNext);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usNext+7F8h, eax
		CPatch::PatchPointer(0x40975F + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE6].usPrev);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usPrev+7F8h, offset dword_A0DAE0
		CPatch::PatchPointer(0x409769 + 1, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE6].usNext);	// mov     eax, ds:_ZN10CStreaming16ms_aInfoForModelE.usNext+7F8h
		CPatch::PatchPointer(0x40978D + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE7].flags);	// and     ds:_ZN10CStreaming16ms_aInfoForModelE.flags+80Ch, 0FEh
		CPatch::PatchPointer(0x40979A + 3, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE7].flags);	// movzx   eax, ds:_ZN10CStreaming16ms_aInfoForModelE.flags+80Ch
		CPatch::PatchPointer(0x4097A6 + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE7].uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+80Ch, 1
		CPatch::PatchPointer(0x4097AF + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE7].usNext);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.usNext+80Ch, 0
		CPatch::PatchPointer(0x4097BD + 1, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE7].usNext);	// mov     ecx, (offset _ZN10CStreaming16ms_aInfoForModelE.usNext+80Ch)
		CPatch::PatchPointer(0x4097C2 + 1, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE7].usNext);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usNext+80Ch, eax
		CPatch::PatchPointer(0x4097C7 + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE7].usPrev);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usPrev+80Ch, offset dword_A0DAE0
		CPatch::PatchPointer(0x4097D1 + 1, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE7].usNext);	// mov     eax, ds:_ZN10CStreaming16ms_aInfoForModelE.usNext+80Ch
		CPatch::PatchPointer(0x4097EE + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE8].flags);	// and     ds:_ZN10CStreaming16ms_aInfoForModelE.flags+820h, 0FEh
		CPatch::PatchPointer(0x4097FB + 3, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE8].flags);	// movzx   eax, ds:_ZN10CStreaming16ms_aInfoForModelE.flags+820h
		CPatch::PatchPointer(0x409807 + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE8].uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+820h, 1
		CPatch::PatchPointer(0x409810 + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE8].usNext);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.usNext+820h, 0
		CPatch::PatchPointer(0x40981E + 1, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE8].usNext);	// mov     edx, (offset _ZN10CStreaming16ms_aInfoForModelE.usNext+820h)
		CPatch::PatchPointer(0x409823 + 1, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE8].usNext);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usNext+820h, eax
		CPatch::PatchPointer(0x409828 + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE8].usPrev);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usPrev+820h, offset dword_A0DAE0
		CPatch::PatchPointer(0x409832 + 1, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE8].usNext);	// mov     eax, ds:_ZN10CStreaming16ms_aInfoForModelE.usNext+820h
		CPatch::PatchPointer(0x4098DE + 3, &CStreaming__ms_aInfoForModel.gta_vc -> uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[ecx*4], 0
		CPatch::PatchPointer(0x409980 + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_SPECIAL01].flags);	// and     ds:(_ZN10CStreaming16ms_aInfoForModelE.flags+884h)[edx], 0FDh
		CPatch::PatchPointer(0x4099A8 + 3, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_SPECIAL01].flags);	// movzx   eax, ds:(_ZN10CStreaming16ms_aInfoForModelE.flags+884h)[edx]
		CPatch::PatchPointer(0x4099B4 + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_SPECIAL01].uiLoadFlag);	// cmp     ds:(_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+884h)[edx], 1
		CPatch::PatchPointer(0x4099BD + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_SPECIAL01].usNext);	// cmp     ds:(_ZN10CStreaming16ms_aInfoForModelE.usNext+884h)[edx], 0
		CPatch::PatchPointer(0x4099CB + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_SPECIAL01].usNext);	// mov     ds:(_ZN10CStreaming16ms_aInfoForModelE.usNext+884h)[edx], eax
		CPatch::PatchPointer(0x4099D1 + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_SPECIAL01].usPrev);	// mov     ds:(_ZN10CStreaming16ms_aInfoForModelE.usPrev+884h)[edx], offset dword_A0DAE0
		CPatch::PatchPointer(0x4099DE + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_SPECIAL01].usNext);	// mov     eax, ds:(_ZN10CStreaming16ms_aInfoForModelE.usNext+884h)[edx]
		CPatch::PatchPointer(0x4099EE + 2, CStreaming__ms_aInfoForModel.gta_vc);	// add     ecx, offset _ZN10CStreaming16ms_aInfoForModelE; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x409A17 + 3, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_SPECIAL01].uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+884h[eax*4], 1
		CPatch::PatchPointer(0x409A91 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[eax], 1
		CPatch::PatchPointer(0x409A9A + 3, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// movzx   eax, ds:_ZN10CStreaming16ms_aInfoForModelE.flags[eax]
		CPatch::PatchPointer(0x409B1E + 2, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// and     ds:_ZN10CStreaming16ms_aInfoForModelE.flags[ecx], 0FEh
		CPatch::PatchPointer(0x409B42 + 3, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// movzx   eax, ds:_ZN10CStreaming16ms_aInfoForModelE.flags[ecx]
		CPatch::PatchPointer(0x409B4E + 2, &CStreaming__ms_aInfoForModel.gta_vc -> uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[ecx], 1
		CPatch::PatchPointer(0x409B57 + 2, CStreaming__ms_aInfoForModel.gta_vc);	// cmp     dword ptr ds:_ZN10CStreaming16ms_aInfoForModelE.usNext[ecx], 0; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x409B60 + 1, CStreaming__ms_aInfoForModel.gta_vc);	// mov     edx, offset _ZN10CStreaming16ms_aInfoForModelE; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x409B6C + 2, CStreaming__ms_aInfoForModel.gta_vc);	// mov     dword ptr ds:_ZN10CStreaming16ms_aInfoForModelE.usNext[ecx], eax; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x409B72 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> usPrev);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usPrev[ecx], offset dword_A0DAE0
		CPatch::PatchPointer(0x409B7C + 2, CStreaming__ms_aInfoForModel.gta_vc);	// mov     eax, dword ptr ds:_ZN10CStreaming16ms_aInfoForModelE.usNext[ecx]; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x409BB9 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// and     ds:_ZN10CStreaming16ms_aInfoForModelE.flags[eax], 0FEh
		CPatch::PatchPointer(0x409BE9 + 3, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// movzx   edx, ds:_ZN10CStreaming16ms_aInfoForModelE.flags[eax]
		CPatch::PatchPointer(0x409BF5 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[eax], 1
		CPatch::PatchPointer(0x409BFE + 2, CStreaming__ms_aInfoForModel.gta_vc);	// cmp     dword ptr ds:_ZN10CStreaming16ms_aInfoForModelE.usNext[eax], 0; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x409C0D + 2, CStreaming__ms_aInfoForModel.gta_vc);	// mov     dword ptr ds:_ZN10CStreaming16ms_aInfoForModelE.usNext[eax], ebp; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x409C13 + 1, CStreaming__ms_aInfoForModel.gta_vc);	// mov     ebp, offset _ZN10CStreaming16ms_aInfoForModelE; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x409C1A + 2, &CStreaming__ms_aInfoForModel.gta_vc -> usPrev);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usPrev[eax], offset dword_A0DAE0
		CPatch::PatchPointer(0x409C24 + 2, CStreaming__ms_aInfoForModel.gta_vc);	// mov     edx, dword ptr ds:_ZN10CStreaming16ms_aInfoForModelE.usNext[eax]; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x409C81 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// and     ds:_ZN10CStreaming16ms_aInfoForModelE.flags[edx], 0FEh
		CPatch::PatchPointer(0x409CA5 + 3, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// movzx   eax, ds:_ZN10CStreaming16ms_aInfoForModelE.flags[edx]
		CPatch::PatchPointer(0x409CB1 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[edx], 1
		CPatch::PatchPointer(0x409CBA + 2, CStreaming__ms_aInfoForModel.gta_vc);	// cmp     dword ptr ds:_ZN10CStreaming16ms_aInfoForModelE.usNext[edx], 0; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x409CC3 + 1, CStreaming__ms_aInfoForModel.gta_vc);	// mov     ecx, offset _ZN10CStreaming16ms_aInfoForModelE; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x409CCF + 2, CStreaming__ms_aInfoForModel.gta_vc);	// mov     dword ptr ds:_ZN10CStreaming16ms_aInfoForModelE.usNext[edx], eax; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x409CD5 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> usPrev);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usPrev[edx], offset dword_A0DAE0
		CPatch::PatchPointer(0x409CDF + 2, CStreaming__ms_aInfoForModel.gta_vc);	// mov     eax, dword ptr ds:_ZN10CStreaming16ms_aInfoForModelE.usNext[edx]; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x409D1F + 2, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// and     ds:_ZN10CStreaming16ms_aInfoForModelE.flags[ecx], 0FEh
		CPatch::PatchPointer(0x409D4D + 3, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// movzx   eax, ds:_ZN10CStreaming16ms_aInfoForModelE.flags[ecx]
		CPatch::PatchPointer(0x409D59 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[ecx], 1
		CPatch::PatchPointer(0x409D62 + 2, CStreaming__ms_aInfoForModel.gta_vc);	// cmp     dword ptr ds:_ZN10CStreaming16ms_aInfoForModelE.usNext[ecx], 0; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x409D6B + 1, CStreaming__ms_aInfoForModel.gta_vc);	// mov     edx, offset _ZN10CStreaming16ms_aInfoForModelE; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x409D77 + 2, CStreaming__ms_aInfoForModel.gta_vc);	// mov     dword ptr ds:_ZN10CStreaming16ms_aInfoForModelE.usNext[ecx], eax; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x409D7D + 2, &CStreaming__ms_aInfoForModel.gta_vc -> usPrev);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usPrev[ecx], offset dword_A0DAE0
		CPatch::PatchPointer(0x409D87 + 2, CStreaming__ms_aInfoForModel.gta_vc);	// mov     eax, dword ptr ds:_ZN10CStreaming16ms_aInfoForModelE.usNext[ecx]; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x409DC0 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// and     ds:_ZN10CStreaming16ms_aInfoForModelE.flags[ecx], 0FEh
		CPatch::PatchPointer(0x409DE4 + 3, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// movzx   eax, ds:_ZN10CStreaming16ms_aInfoForModelE.flags[ecx]
		CPatch::PatchPointer(0x409DF0 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[ecx], 1
		CPatch::PatchPointer(0x409DF9 + 2, CStreaming__ms_aInfoForModel.gta_vc);	// cmp     dword ptr ds:_ZN10CStreaming16ms_aInfoForModelE.usNext[ecx], 0; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x409E02 + 1, CStreaming__ms_aInfoForModel.gta_vc);	// mov     edx, offset _ZN10CStreaming16ms_aInfoForModelE; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x409E0E + 2, CStreaming__ms_aInfoForModel.gta_vc);	// mov     dword ptr ds:_ZN10CStreaming16ms_aInfoForModelE.usNext[ecx], eax; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x409E14 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> usPrev);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usPrev[ecx], offset dword_A0DAE0
		CPatch::PatchPointer(0x409E1E + 2, CStreaming__ms_aInfoForModel.gta_vc);	// mov     eax, dword ptr ds:_ZN10CStreaming16ms_aInfoForModelE.usNext[ecx]; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x409E5F + 2, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// and     ds:_ZN10CStreaming16ms_aInfoForModelE.flags[edx], 0FEh
		CPatch::PatchPointer(0x409E8D + 3, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// movzx   eax, ds:_ZN10CStreaming16ms_aInfoForModelE.flags[edx]
		CPatch::PatchPointer(0x409E99 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[edx], 1
		CPatch::PatchPointer(0x409EA2 + 2, CStreaming__ms_aInfoForModel.gta_vc);	// cmp     dword ptr ds:_ZN10CStreaming16ms_aInfoForModelE.usNext[edx], 0; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x409EAB + 1, CStreaming__ms_aInfoForModel.gta_vc);	// mov     ecx, offset _ZN10CStreaming16ms_aInfoForModelE; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x409EB7 + 2, CStreaming__ms_aInfoForModel.gta_vc);	// mov     dword ptr ds:_ZN10CStreaming16ms_aInfoForModelE.usNext[edx], eax; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x409EBD + 2, &CStreaming__ms_aInfoForModel.gta_vc -> usPrev);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usPrev[edx], offset dword_A0DAE0
		CPatch::PatchPointer(0x409EC7 + 2, CStreaming__ms_aInfoForModel.gta_vc);	// mov     eax, dword ptr ds:_ZN10CStreaming16ms_aInfoForModelE.usNext[edx]; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x409F00 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// and     ds:_ZN10CStreaming16ms_aInfoForModelE.flags[edx], 0FEh
		CPatch::PatchPointer(0x409F24 + 3, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// movzx   eax, ds:_ZN10CStreaming16ms_aInfoForModelE.flags[edx]
		CPatch::PatchPointer(0x409F30 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[edx], 1
		CPatch::PatchPointer(0x409F39 + 2, CStreaming__ms_aInfoForModel.gta_vc);	// cmp     dword ptr ds:_ZN10CStreaming16ms_aInfoForModelE.usNext[edx], 0; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x409F42 + 1, CStreaming__ms_aInfoForModel.gta_vc);	// mov     ecx, offset _ZN10CStreaming16ms_aInfoForModelE; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x409F4E + 2, CStreaming__ms_aInfoForModel.gta_vc);	// mov     dword ptr ds:_ZN10CStreaming16ms_aInfoForModelE.usNext[edx], eax; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x409F54 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> usPrev);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usPrev[edx], offset dword_A0DAE0
		CPatch::PatchPointer(0x409F5E + 2, CStreaming__ms_aInfoForModel.gta_vc);	// mov     eax, dword ptr ds:_ZN10CStreaming16ms_aInfoForModelE.usNext[edx]; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x409F9F + 2, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// and     ds:_ZN10CStreaming16ms_aInfoForModelE.flags[ecx], 0FEh
		CPatch::PatchPointer(0x409FCD + 3, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// movzx   eax, ds:_ZN10CStreaming16ms_aInfoForModelE.flags[ecx]
		CPatch::PatchPointer(0x409FD9 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[ecx], 1
		CPatch::PatchPointer(0x409FE2 + 2, CStreaming__ms_aInfoForModel.gta_vc);	// cmp     dword ptr ds:_ZN10CStreaming16ms_aInfoForModelE.usNext[ecx], 0; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x409FEB + 1, CStreaming__ms_aInfoForModel.gta_vc);	// mov     edx, offset _ZN10CStreaming16ms_aInfoForModelE; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x409FF7 + 2, CStreaming__ms_aInfoForModel.gta_vc);	// mov     dword ptr ds:_ZN10CStreaming16ms_aInfoForModelE.usNext[ecx], eax; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x409FFD + 2, &CStreaming__ms_aInfoForModel.gta_vc -> usPrev);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usPrev[ecx], offset dword_A0DAE0
		CPatch::PatchPointer(0x40A007 + 2, CStreaming__ms_aInfoForModel.gta_vc);	// mov     eax, dword ptr ds:_ZN10CStreaming16ms_aInfoForModelE.usNext[ecx]; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40A0CF + 2, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// and     ds:_ZN10CStreaming16ms_aInfoForModelE.flags[ecx], 0FEh
		CPatch::PatchPointer(0x40A0F3 + 3, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// movzx   eax, ds:_ZN10CStreaming16ms_aInfoForModelE.flags[ecx]
		CPatch::PatchPointer(0x40A0FF + 2, &CStreaming__ms_aInfoForModel.gta_vc -> uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[ecx], 1
		CPatch::PatchPointer(0x40A108 + 2, CStreaming__ms_aInfoForModel.gta_vc);	// cmp     dword ptr ds:_ZN10CStreaming16ms_aInfoForModelE.usNext[ecx], 0; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40A111 + 1, CStreaming__ms_aInfoForModel.gta_vc);	// mov     ebx, offset _ZN10CStreaming16ms_aInfoForModelE; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40A11D + 2, CStreaming__ms_aInfoForModel.gta_vc);	// mov     dword ptr ds:_ZN10CStreaming16ms_aInfoForModelE.usNext[ecx], eax; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40A123 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> usPrev);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usPrev[ecx], offset dword_A0DAE0
		CPatch::PatchPointer(0x40A12D + 2, CStreaming__ms_aInfoForModel.gta_vc);	// mov     eax, dword ptr ds:_ZN10CStreaming16ms_aInfoForModelE.usNext[ecx]; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40A169 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// and     ds:_ZN10CStreaming16ms_aInfoForModelE.flags[edx], 0FEh
		CPatch::PatchPointer(0x40A197 + 3, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// movzx   eax, ds:_ZN10CStreaming16ms_aInfoForModelE.flags[edx]
		CPatch::PatchPointer(0x40A1A3 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[edx], 1
		CPatch::PatchPointer(0x40A1AC + 2, CStreaming__ms_aInfoForModel.gta_vc);	// cmp     dword ptr ds:_ZN10CStreaming16ms_aInfoForModelE.usNext[edx], 0; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40A1B5 + 1, CStreaming__ms_aInfoForModel.gta_vc);	// mov     ebx, offset _ZN10CStreaming16ms_aInfoForModelE; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40A1C1 + 2, CStreaming__ms_aInfoForModel.gta_vc);	// mov     dword ptr ds:_ZN10CStreaming16ms_aInfoForModelE.usNext[edx], eax; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40A1C7 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> usPrev);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usPrev[edx], offset dword_A0DAE0
		CPatch::PatchPointer(0x40A1D1 + 2, CStreaming__ms_aInfoForModel.gta_vc);	// mov     eax, dword ptr ds:_ZN10CStreaming16ms_aInfoForModelE.usNext[edx]; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40A396 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// and     ds:_ZN10CStreaming16ms_aInfoForModelE.flags[edx], 0FEh
		CPatch::PatchPointer(0x40A3BA + 3, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// movzx   eax, ds:_ZN10CStreaming16ms_aInfoForModelE.flags[edx]
		CPatch::PatchPointer(0x40A3C6 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[edx], 1
		CPatch::PatchPointer(0x40A3CF + 2, CStreaming__ms_aInfoForModel.gta_vc);	// cmp     dword ptr ds:_ZN10CStreaming16ms_aInfoForModelE.usNext[edx], 0; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40A3D8 + 1, CStreaming__ms_aInfoForModel.gta_vc);	// mov     ebp, offset _ZN10CStreaming16ms_aInfoForModelE; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40A3E4 + 2, CStreaming__ms_aInfoForModel.gta_vc);	// mov     dword ptr ds:_ZN10CStreaming16ms_aInfoForModelE.usNext[edx], eax; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40A3EA + 2, &CStreaming__ms_aInfoForModel.gta_vc -> usPrev);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usPrev[edx], offset dword_A0DAE0
		CPatch::PatchPointer(0x40A3F4 + 2, CStreaming__ms_aInfoForModel.gta_vc);	// mov     eax, dword ptr ds:_ZN10CStreaming16ms_aInfoForModelE.usNext[edx]; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40A42E + 2, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// and     ds:_ZN10CStreaming16ms_aInfoForModelE.flags[eax], 0FEh
		CPatch::PatchPointer(0x40A45E + 3, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// movzx   ebx, ds:_ZN10CStreaming16ms_aInfoForModelE.flags[eax]
		CPatch::PatchPointer(0x40A46A + 2, &CStreaming__ms_aInfoForModel.gta_vc -> uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[eax], 1
		CPatch::PatchPointer(0x40A473 + 2, CStreaming__ms_aInfoForModel.gta_vc);	// cmp     dword ptr ds:_ZN10CStreaming16ms_aInfoForModelE.usNext[eax], 0; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40A47C + 1, CStreaming__ms_aInfoForModel.gta_vc);	// mov     ebx, offset _ZN10CStreaming16ms_aInfoForModelE; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40A489 + 2, CStreaming__ms_aInfoForModel.gta_vc);	// mov     dword ptr ds:_ZN10CStreaming16ms_aInfoForModelE.usNext[eax], ebp; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40A48F + 2, &CStreaming__ms_aInfoForModel.gta_vc -> usPrev);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usPrev[eax], offset dword_A0DAE0
		CPatch::PatchPointer(0x40A499 + 2, CStreaming__ms_aInfoForModel.gta_vc);	// mov     ecx, dword ptr ds:_ZN10CStreaming16ms_aInfoForModelE.usNext[eax]; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40A609 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// and     ds:_ZN10CStreaming16ms_aInfoForModelE.flags[ecx], 0FEh
		CPatch::PatchPointer(0x40A62D + 3, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// movzx   eax, ds:_ZN10CStreaming16ms_aInfoForModelE.flags[ecx]
		CPatch::PatchPointer(0x40A639 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[ecx], 1
		CPatch::PatchPointer(0x40A642 + 2, CStreaming__ms_aInfoForModel.gta_vc);	// cmp     dword ptr ds:_ZN10CStreaming16ms_aInfoForModelE.usNext[ecx], 0; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40A64B + 1, CStreaming__ms_aInfoForModel.gta_vc);	// mov     edx, offset _ZN10CStreaming16ms_aInfoForModelE; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40A657 + 2, CStreaming__ms_aInfoForModel.gta_vc);	// mov     dword ptr ds:_ZN10CStreaming16ms_aInfoForModelE.usNext[ecx], eax; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40A65D + 2, &CStreaming__ms_aInfoForModel.gta_vc -> usPrev);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usPrev[ecx], offset dword_A0DAE0
		CPatch::PatchPointer(0x40A667 + 2, CStreaming__ms_aInfoForModel.gta_vc);	// mov     eax, dword ptr ds:_ZN10CStreaming16ms_aInfoForModelE.usNext[ecx]; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40A691 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// and     ds:_ZN10CStreaming16ms_aInfoForModelE.flags[edx], 0FEh
		CPatch::PatchPointer(0x40A6B5 + 3, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// movzx   eax, ds:_ZN10CStreaming16ms_aInfoForModelE.flags[edx]
		CPatch::PatchPointer(0x40A6C1 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[edx], 1
		CPatch::PatchPointer(0x40A6CA + 2, CStreaming__ms_aInfoForModel.gta_vc);	// cmp     dword ptr ds:_ZN10CStreaming16ms_aInfoForModelE.usNext[edx], 0; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40A6D3 + 1, CStreaming__ms_aInfoForModel.gta_vc);	// mov     ecx, offset _ZN10CStreaming16ms_aInfoForModelE; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40A6DF + 2, CStreaming__ms_aInfoForModel.gta_vc);	// mov     dword ptr ds:_ZN10CStreaming16ms_aInfoForModelE.usNext[edx], eax; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40A6E5 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> usPrev);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usPrev[edx], offset dword_A0DAE0
		CPatch::PatchPointer(0x40A6EF + 2, CStreaming__ms_aInfoForModel.gta_vc);	// mov     eax, dword ptr ds:_ZN10CStreaming16ms_aInfoForModelE.usNext[edx]; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40A72F + 2, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// and     ds:_ZN10CStreaming16ms_aInfoForModelE.flags[ecx], 0FEh
		CPatch::PatchPointer(0x40A75D + 3, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// movzx   eax, ds:_ZN10CStreaming16ms_aInfoForModelE.flags[ecx]
		CPatch::PatchPointer(0x40A769 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[ecx], 1
		CPatch::PatchPointer(0x40A772 + 2, CStreaming__ms_aInfoForModel.gta_vc);	// cmp     dword ptr ds:_ZN10CStreaming16ms_aInfoForModelE.usNext[ecx], 0; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40A77B + 1, CStreaming__ms_aInfoForModel.gta_vc);	// mov     edx, offset _ZN10CStreaming16ms_aInfoForModelE; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40A787 + 2, CStreaming__ms_aInfoForModel.gta_vc);	// mov     dword ptr ds:_ZN10CStreaming16ms_aInfoForModelE.usNext[ecx], eax; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40A78D + 2, &CStreaming__ms_aInfoForModel.gta_vc -> usPrev);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usPrev[ecx], offset dword_A0DAE0
		CPatch::PatchPointer(0x40A797 + 2, CStreaming__ms_aInfoForModel.gta_vc);	// mov     eax, dword ptr ds:_ZN10CStreaming16ms_aInfoForModelE.usNext[ecx]; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40A7DD + 2, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// and     ds:_ZN10CStreaming16ms_aInfoForModelE.flags[edx], 0FEh
		CPatch::PatchPointer(0x40A80B + 3, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// movzx   eax, ds:_ZN10CStreaming16ms_aInfoForModelE.flags[edx]
		CPatch::PatchPointer(0x40A817 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[edx], 1
		CPatch::PatchPointer(0x40A820 + 2, CStreaming__ms_aInfoForModel.gta_vc);	// cmp     dword ptr ds:_ZN10CStreaming16ms_aInfoForModelE.usNext[edx], 0; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40A829 + 1, CStreaming__ms_aInfoForModel.gta_vc);	// mov     ecx, offset _ZN10CStreaming16ms_aInfoForModelE; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40A835 + 2, CStreaming__ms_aInfoForModel.gta_vc);	// mov     dword ptr ds:_ZN10CStreaming16ms_aInfoForModelE.usNext[edx], eax; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40A83B + 2, &CStreaming__ms_aInfoForModel.gta_vc -> usPrev);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usPrev[edx], offset dword_A0DAE0
		CPatch::PatchPointer(0x40A845 + 2, CStreaming__ms_aInfoForModel.gta_vc);	// mov     eax, dword ptr ds:_ZN10CStreaming16ms_aInfoForModelE.usNext[edx]; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40A8C4 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// and     ss:_ZN10CStreaming16ms_aInfoForModelE.flags[ebp], 0FEh
		CPatch::PatchPointer(0x40A8E6 + 3, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// movzx   ecx, ss:_ZN10CStreaming16ms_aInfoForModelE.flags[ebp]
		CPatch::PatchPointer(0x40A8F2 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> uiLoadFlag);	// cmp     ss:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[ebp], 1
		CPatch::PatchPointer(0x40A8FB + 2, CStreaming__ms_aInfoForModel.gta_vc);	// cmp     dword ptr ss:_ZN10CStreaming16ms_aInfoForModelE.usNext[ebp], 0; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40A904 + 1, CStreaming__ms_aInfoForModel.gta_vc);	// mov     ecx, offset _ZN10CStreaming16ms_aInfoForModelE; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40A910 + 2, CStreaming__ms_aInfoForModel.gta_vc);	// mov     dword ptr ss:_ZN10CStreaming16ms_aInfoForModelE.usNext[ebp], eax; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40A916 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> usPrev);	// mov     ss:_ZN10CStreaming16ms_aInfoForModelE.usPrev[ebp], offset dword_A0DAE0
		CPatch::PatchPointer(0x40A920 + 2, CStreaming__ms_aInfoForModel.gta_vc);	// mov     eax, dword ptr ss:_ZN10CStreaming16ms_aInfoForModelE.usNext[ebp]; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40A960 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// and     ds:_ZN10CStreaming16ms_aInfoForModelE.flags[ecx], 0FEh
		CPatch::PatchPointer(0x40A98E + 3, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// movzx   eax, ds:_ZN10CStreaming16ms_aInfoForModelE.flags[ecx]
		CPatch::PatchPointer(0x40A99A + 2, &CStreaming__ms_aInfoForModel.gta_vc -> uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[ecx], 1
		CPatch::PatchPointer(0x40A9A3 + 2, CStreaming__ms_aInfoForModel.gta_vc);	// cmp     dword ptr ds:_ZN10CStreaming16ms_aInfoForModelE.usNext[ecx], 0; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40A9AC + 1, CStreaming__ms_aInfoForModel.gta_vc);	// mov     ebp, offset _ZN10CStreaming16ms_aInfoForModelE; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40A9B8 + 2, CStreaming__ms_aInfoForModel.gta_vc);	// mov     dword ptr ds:_ZN10CStreaming16ms_aInfoForModelE.usNext[ecx], eax; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40A9BE + 2, &CStreaming__ms_aInfoForModel.gta_vc -> usPrev);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usPrev[ecx], offset dword_A0DAE0
		CPatch::PatchPointer(0x40A9C8 + 2, CStreaming__ms_aInfoForModel.gta_vc);	// mov     eax, dword ptr ds:_ZN10CStreaming16ms_aInfoForModelE.usNext[ecx]; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40AD3B + 2, &CStreaming__ms_aInfoForModel.gta_vc -> FilePositionInBlocks);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.FilePositionInBlocks[esi], edx
		CPatch::PatchPointer(0x40AD41 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> _unk_inCdImage);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE._unk_inCdImage[esi], eax
		CPatch::PatchPointer(0x40AD70 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// and     ds:_ZN10CStreaming16ms_aInfoForModelE.flags[edx], 0FDh
		CPatch::PatchPointer(0x40AD94 + 3, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// movzx   eax, ds:_ZN10CStreaming16ms_aInfoForModelE.flags[edx]
		CPatch::PatchPointer(0x40ADA0 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[edx], 1
		CPatch::PatchPointer(0x40ADA9 + 2, CStreaming__ms_aInfoForModel.gta_vc);	// cmp     dword ptr ds:_ZN10CStreaming16ms_aInfoForModelE.usNext[edx], 0; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40ADB2 + 1, CStreaming__ms_aInfoForModel.gta_vc);	// mov     ecx, offset _ZN10CStreaming16ms_aInfoForModelE; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40ADBE + 2, CStreaming__ms_aInfoForModel.gta_vc);	// mov     dword ptr ds:_ZN10CStreaming16ms_aInfoForModelE.usNext[edx], eax; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40ADC4 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> usPrev);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usPrev[edx], offset dword_A0DAE0
		CPatch::PatchPointer(0x40ADCE + 2, CStreaming__ms_aInfoForModel.gta_vc);	// mov     eax, dword ptr ds:_ZN10CStreaming16ms_aInfoForModelE.usNext[edx]; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40AE16 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// and     ds:_ZN10CStreaming16ms_aInfoForModelE.flags[eax], 0FEh
		CPatch::PatchPointer(0x40AE46 + 3, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// movzx   edx, ds:_ZN10CStreaming16ms_aInfoForModelE.flags[eax]
		CPatch::PatchPointer(0x40AE52 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[eax], 1
		CPatch::PatchPointer(0x40AE5B + 2, CStreaming__ms_aInfoForModel.gta_vc);	// cmp     dword ptr ds:_ZN10CStreaming16ms_aInfoForModelE.usNext[eax], 0; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40AE6A + 2, CStreaming__ms_aInfoForModel.gta_vc);	// mov     dword ptr ds:_ZN10CStreaming16ms_aInfoForModelE.usNext[eax], ecx; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40AE70 + 1, CStreaming__ms_aInfoForModel.gta_vc);	// mov     ecx, offset _ZN10CStreaming16ms_aInfoForModelE; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40AE77 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> usPrev);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usPrev[eax], offset dword_A0DAE0
		CPatch::PatchPointer(0x40AE81 + 2, CStreaming__ms_aInfoForModel.gta_vc);	// mov     edx, dword ptr ds:_ZN10CStreaming16ms_aInfoForModelE.usNext[eax]; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40AEB0 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// and     ds:_ZN10CStreaming16ms_aInfoForModelE.flags[edx], 0FEh
		CPatch::PatchPointer(0x40AED4 + 3, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// movzx   eax, ds:_ZN10CStreaming16ms_aInfoForModelE.flags[edx]
		CPatch::PatchPointer(0x40AEE0 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[edx], 1
		CPatch::PatchPointer(0x40AEE9 + 2, CStreaming__ms_aInfoForModel.gta_vc);	// cmp     dword ptr ds:_ZN10CStreaming16ms_aInfoForModelE.usNext[edx], 0; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40AEF2 + 1, CStreaming__ms_aInfoForModel.gta_vc);	// mov     ecx, offset _ZN10CStreaming16ms_aInfoForModelE; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40AEFE + 2, CStreaming__ms_aInfoForModel.gta_vc);	// mov     dword ptr ds:_ZN10CStreaming16ms_aInfoForModelE.usNext[edx], eax; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40AF04 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> usPrev);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usPrev[edx], offset dword_A0DAE0
		CPatch::PatchPointer(0x40AF0E + 2, CStreaming__ms_aInfoForModel.gta_vc);	// mov     eax, dword ptr ds:_ZN10CStreaming16ms_aInfoForModelE.usNext[edx]; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40AF98 + 2, CStreaming__ms_aInfoForModel.gta_vc);	// sub     edx, offset _ZN10CStreaming16ms_aInfoForModelE; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40B3E2 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// and     ds:_ZN10CStreaming16ms_aInfoForModelE.flags[eax], 0DFh
		CPatch::PatchPointer(0x40B3F2 + 3, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// and     ds:_ZN10CStreaming16ms_aInfoForModelE.flags[edx*4], 0DFh
		CPatch::PatchPointer(0x40B400 + 3, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// and     ds:_ZN10CStreaming16ms_aInfoForModelE.flags[ebx*4], 0DFh
		CPatch::PatchPointer(0x40B40E + 3, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// and     ds:_ZN10CStreaming16ms_aInfoForModelE.flags[esi*4], 0DFh
		CPatch::PatchPointer(0x40B41C + 3, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// and     ds:_ZN10CStreaming16ms_aInfoForModelE.flags[edx*4], 0DFh
		CPatch::PatchPointer(0x40B42A + 3, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// and     ds:_ZN10CStreaming16ms_aInfoForModelE.flags[ebx*4], 0DFh
		CPatch::PatchPointer(0x40B438 + 3, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// and     ds:_ZN10CStreaming16ms_aInfoForModelE.flags[esi*4], 0DFh
		CPatch::PatchPointer(0x40B449 + 3, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// and     ds:_ZN10CStreaming16ms_aInfoForModelE.flags[edx*4], 0DFh
		CPatch::PatchPointer(0x40B461 + 3, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// and     ds:_ZN10CStreaming16ms_aInfoForModelE.flags[ebx*4], 0DFh
		CPatch::PatchPointer(0x40B46C + 3, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// and     ds:_ZN10CStreaming16ms_aInfoForModelE.flags[eax*4], 0DFh
		CPatch::PatchPointer(0x40B47A + 3, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// and     ds:_ZN10CStreaming16ms_aInfoForModelE.flags[esi*4], 0DFh
		CPatch::PatchPointer(0x40B488 + 3, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// and     ds:_ZN10CStreaming16ms_aInfoForModelE.flags[eax*4], 0DFh
		CPatch::PatchPointer(0x40B496 + 3, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// and     ds:_ZN10CStreaming16ms_aInfoForModelE.flags[edx*4], 0DFh
		CPatch::PatchPointer(0x40B4A7 + 3, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// and     ds:_ZN10CStreaming16ms_aInfoForModelE.flags[eax*4], 0DFh
		CPatch::PatchPointer(0x40B4B7 + 3, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// and     ds:_ZN10CStreaming16ms_aInfoForModelE.flags[ecx*4], 0DFh
		CPatch::PatchPointer(0x40B4E7 + 2, CStreaming__ms_aInfoForModel.gta_vc);	// sub     edx, offset _ZN10CStreaming16ms_aInfoForModelE; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40B6E2 + 1, CStreaming__ms_aInfoForModel.gta_vc);	// mov     ebx, offset _ZN10CStreaming16ms_aInfoForModelE; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40B71D + 3, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// movzx   eax, ds:_ZN10CStreaming16ms_aInfoForModelE.flags[ecx]
		CPatch::PatchPointer(0x40B72B + 2, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// and     ds:_ZN10CStreaming16ms_aInfoForModelE.flags[ecx], 0F7h
		CPatch::PatchPointer(0x40B738 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> _unk_inCdImage);	// mov     edx, ds:_ZN10CStreaming16ms_aInfoForModelE._unk_inCdImage[ecx]
		CPatch::PatchPointer(0x40B746 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> FilePositionInBlocks);	// mov     eax, ds:_ZN10CStreaming16ms_aInfoForModelE.FilePositionInBlocks[ecx]
		CPatch::PatchPointer(0x40B8D8 + 1, CStreaming__ms_aInfoForModel.gta_vc);	// mov     ebx, offset _ZN10CStreaming16ms_aInfoForModelE; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40B918 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// cmp     ss:_ZN10CStreaming16ms_aInfoForModelE.flags[ebp], 0
		CPatch::PatchPointer(0x40B97D + 2, &CStreaming__ms_aInfoForModel.gta_vc -> _unk_inCdImage);	// mov     edx, ds:_ZN10CStreaming16ms_aInfoForModelE._unk_inCdImage[ecx]
		CPatch::PatchPointer(0x40B987 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> FilePositionInBlocks);	// mov     eax, ds:_ZN10CStreaming16ms_aInfoForModelE.FilePositionInBlocks[ecx]
		CPatch::PatchPointer(0x40B9FA + 2, &CStreaming__ms_aInfoForModel.gta_vc -> uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[esi], 2
		CPatch::PatchPointer(0x40BA07 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> _unk_inCdImage);	// mov     eax, ds:_ZN10CStreaming16ms_aInfoForModelE._unk_inCdImage[esi]
		CPatch::PatchPointer(0x40BA1D + 3, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// movzx   eax, ds:_ZN10CStreaming16ms_aInfoForModelE.flags[esi]
		CPatch::PatchPointer(0x40BA6B + 3, &CStreaming__ms_aInfoForModel.gta_vc -> uiLoadFlag);	// mov     al, ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[edx*4]
		CPatch::PatchPointer(0x40BA90 + 3, &CStreaming__ms_aInfoForModel.gta_vc -> uiLoadFlag);	// mov     al, ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[eax*4]
		CPatch::PatchPointer(0x40BB22 + 1, CStreaming__ms_aInfoForModel.gta_vc);	// mov     edi, offset _ZN10CStreaming16ms_aInfoForModelE; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40BB4F + 3, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// movzx   eax, ds:_ZN10CStreaming16ms_aInfoForModelE.flags[esi]
		CPatch::PatchPointer(0x40BB5B + 2, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// and     ds:_ZN10CStreaming16ms_aInfoForModelE.flags[esi], 0F7h
		CPatch::PatchPointer(0x40BB68 + 3, &CStreaming__ms_aInfoForModel.gta_vc -> usNextOnCd);	// movsx   ebx, ds:_ZN10CStreaming16ms_aInfoForModelE.usNextOnCd[esi]
		CPatch::PatchPointer(0x40BCDC + 2, CStreaming__ms_aInfoForModel.gta_vc);	// sub     edx, offset _ZN10CStreaming16ms_aInfoForModelE; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40BD29 + 4, &CStreaming__ms_aInfoForModel.gta_vc[GetBaseID(FILE_TYPE_TXD)].uiLoadFlag);	// movzx   eax, ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+1FBD0h[edi*4]
		CPatch::PatchPointer(0x40BD44 + 4, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// movzx   eax, ds:_ZN10CStreaming16ms_aInfoForModelE.flags[ebx*4]
		CPatch::PatchPointer(0x40BD73 + 4, &CStreaming__ms_aInfoForModel.gta_vc -> uiLoadFlag);	// movzx   ecx, ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[edx*4]
		CPatch::PatchPointer(0x40BD8D + 4, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// movzx   edx, ds:_ZN10CStreaming16ms_aInfoForModelE.flags[ebx*4]
		CPatch::PatchPointer(0x40BDC6 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> _unk_inCdImage);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE._unk_inCdImage[edi], 0
		CPatch::PatchPointer(0x40BDD3 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> FilePositionInBlocks);	// mov     eax, ds:_ZN10CStreaming16ms_aInfoForModelE.FilePositionInBlocks[edi]
		CPatch::PatchPointer(0x40BE10 + 3, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// movzx   eax, ds:_ZN10CStreaming16ms_aInfoForModelE.flags[edi]
		CPatch::PatchPointer(0x40BE22 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// and     ds:_ZN10CStreaming16ms_aInfoForModelE.flags[edi], 0F7h
		CPatch::PatchPointer(0x40BF72 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> _unk_inCdImage);	// mov     eax, ds:_ZN10CStreaming16ms_aInfoForModelE._unk_inCdImage[esi]
		CPatch::PatchPointer(0x40BFB2 + 3, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// movzx   eax, ds:_ZN10CStreaming16ms_aInfoForModelE.flags[esi]
		CPatch::PatchPointer(0x40BFEB + 3, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// movzx   eax, ds:_ZN10CStreaming16ms_aInfoForModelE.flags[esi]
		CPatch::PatchPointer(0x40C08B + 2, &CStreaming__ms_aInfoForModel.gta_vc -> uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[esi], 4
		CPatch::PatchPointer(0x40C1ED + 1, CStreaming__ms_aInfoForModel.gta_vc);	// mov     esi, offset _ZN10CStreaming16ms_aInfoForModelE; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40C3B9 + 3, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// movzx   eax, ds:_ZN10CStreaming16ms_aInfoForModelE.flags[edi]
		CPatch::PatchPointer(0x40C3EA + 3, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// movzx   eax, ds:_ZN10CStreaming16ms_aInfoForModelE.flags[edi]
		CPatch::PatchPointer(0x40C42D + 3, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// movzx   eax, ds:_ZN10CStreaming16ms_aInfoForModelE.flags[edi]
		CPatch::PatchPointer(0x40C4DA + 3, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// movzx   eax, ds:_ZN10CStreaming16ms_aInfoForModelE.flags[edi]
		CPatch::PatchPointer(0x40C551 + 3, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// movzx   eax, ds:_ZN10CStreaming16ms_aInfoForModelE.flags[edi]
		CPatch::PatchPointer(0x40C588 + 3, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// movzx   eax, ds:_ZN10CStreaming16ms_aInfoForModelE.flags[edi]
		CPatch::PatchPointer(0x40C61B + 3, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// movzx   eax, ds:_ZN10CStreaming16ms_aInfoForModelE.flags[edi]
		CPatch::PatchPointer(0x40C684 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> _unk_inCdImage);	// mov     eax, ds:_ZN10CStreaming16ms_aInfoForModelE._unk_inCdImage[edi]
		CPatch::PatchPointer(0x40C77D + 1, CStreaming__ms_aInfoForModel.gta_vc);	// mov     esi, offset _ZN10CStreaming16ms_aInfoForModelE; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40C8A5 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> uiLoadFlag);	// mov     ss:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[ebp], 1
		CPatch::PatchPointer(0x40C8AC + 2, &CStreaming__ms_aInfoForModel.gta_vc -> _unk_inCdImage);	// mov     eax, ss:_ZN10CStreaming16ms_aInfoForModelE._unk_inCdImage[ebp]
		CPatch::PatchPointer(0x40C8CA + 3, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// movzx   eax, ss:_ZN10CStreaming16ms_aInfoForModelE.flags[ebp]
		CPatch::PatchPointer(0x40C9DD + 2, CStreaming__ms_aInfoForModel.gta_vc);	// sub     edx, offset _ZN10CStreaming16ms_aInfoForModelE; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40CAAB + 2, CStreaming__ms_aInfoForModel.gta_vc);	// sub     edx, offset _ZN10CStreaming16ms_aInfoForModelE; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40CB9C + 3, CStreaming__ms_aInfoForModel.gta_vc);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.usNext[eax*4], 0; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40CC17 + 3, CStreaming__ms_aInfoForModel.gta_vc);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.usNext[eax*4], 0; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40D447 + 2, CStreaming__ms_aInfoForModel.gta_vc);	// sub     edx, offset _ZN10CStreaming16ms_aInfoForModelE; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40D5C7 + 2, CStreaming__ms_aInfoForModel.gta_vc);	// sub     edx, offset _ZN10CStreaming16ms_aInfoForModelE; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40D6F4 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> uiLoadFlag);	// mov     al, ss:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[ebp]
		CPatch::PatchPointer(0x40D775 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> _unk_inCdImage);	// mov     eax, ss:_ZN10CStreaming16ms_aInfoForModelE._unk_inCdImage[ebp]
		CPatch::PatchPointer(0x40D784 + 2, CStreaming__ms_aInfoForModel.gta_vc);	// cmp     dword ptr ss:_ZN10CStreaming16ms_aInfoForModelE.usNext[ebp], 0; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40D78D + 2, &CStreaming__ms_aInfoForModel.gta_vc -> uiLoadFlag);	// cmp     ss:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[ebp], 2
		CPatch::PatchPointer(0x40D796 + 3, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// movzx   eax, ss:_ZN10CStreaming16ms_aInfoForModelE.flags[ebp]
		CPatch::PatchPointer(0x40D7A8 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// and     ss:_ZN10CStreaming16ms_aInfoForModelE.flags[ebp], 0F7h
		CPatch::PatchPointer(0x40D7B5 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> usPrev);	// mov     eax, ss:_ZN10CStreaming16ms_aInfoForModelE.usPrev[ebp]
		CPatch::PatchPointer(0x40D7BB + 2, CStreaming__ms_aInfoForModel.gta_vc);	// mov     ecx, dword ptr ss:_ZN10CStreaming16ms_aInfoForModelE.usNext[ebp]; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40D7C4 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> usPrev);	// mov     edx, ss:_ZN10CStreaming16ms_aInfoForModelE.usPrev[ebp]
		CPatch::PatchPointer(0x40D7CA + 2, CStreaming__ms_aInfoForModel.gta_vc);	// mov     eax, dword ptr ss:_ZN10CStreaming16ms_aInfoForModelE.usNext[ebp]; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40D7D2 + 2, CStreaming__ms_aInfoForModel.gta_vc);	// mov     dword ptr ss:_ZN10CStreaming16ms_aInfoForModelE.usNext[ebp], 0; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40D7DC + 2, &CStreaming__ms_aInfoForModel.gta_vc -> usPrev);	// mov     ss:_ZN10CStreaming16ms_aInfoForModelE.usPrev[ebp], 0
		CPatch::PatchPointer(0x40D7E8 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> uiLoadFlag);	// cmp     ss:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[ebp], 3
		CPatch::PatchPointer(0x40D81B + 2, &CStreaming__ms_aInfoForModel.gta_vc -> uiLoadFlag);	// cmp     ss:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[ebp], 4
		CPatch::PatchPointer(0x40D885 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> uiLoadFlag);	// mov     ss:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[ebp], 0
		CPatch::PatchPointer(0x40E329 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> uiLoadFlag);	// mov     al, ss:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[ebp]
		CPatch::PatchPointer(0x40E33A + 3, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// movzx   eax, ss:_ZN10CStreaming16ms_aInfoForModelE.flags[ebp]
		CPatch::PatchPointer(0x40E34C + 2, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// or      ss:_ZN10CStreaming16ms_aInfoForModelE.flags[ebp], 8
		CPatch::PatchPointer(0x40E35C + 2, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// or      ss:_ZN10CStreaming16ms_aInfoForModelE.flags[ebp], bl
		CPatch::PatchPointer(0x40E362 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> uiLoadFlag);	// mov     al, ss:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[ebp]
		CPatch::PatchPointer(0x40E370 + 3, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// movzx   eax, ss:_ZN10CStreaming16ms_aInfoForModelE.flags[ebp]
		CPatch::PatchPointer(0x40E3A1 + 2, CStreaming__ms_aInfoForModel.gta_vc);	// mov     esi, dword ptr ss:_ZN10CStreaming16ms_aInfoForModelE.usNext[ebp]; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40E3AB + 2, &CStreaming__ms_aInfoForModel.gta_vc -> usPrev);	// mov     eax, ss:_ZN10CStreaming16ms_aInfoForModelE.usPrev[ebp]
		CPatch::PatchPointer(0x40E3B4 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> usPrev);	// mov     ebx, ss:_ZN10CStreaming16ms_aInfoForModelE.usPrev[ebp]
		CPatch::PatchPointer(0x40E3BA + 2, CStreaming__ms_aInfoForModel.gta_vc);	// mov     eax, dword ptr ss:_ZN10CStreaming16ms_aInfoForModelE.usNext[ebp]; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40E3C2 + 2, CStreaming__ms_aInfoForModel.gta_vc);	// mov     dword ptr ss:_ZN10CStreaming16ms_aInfoForModelE.usNext[ebp], 0; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40E3CC + 2, &CStreaming__ms_aInfoForModel.gta_vc -> usPrev);	// mov     ss:_ZN10CStreaming16ms_aInfoForModelE.usPrev[ebp], 0
		CPatch::PatchPointer(0x40E3D6 + 3, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// movzx   eax, ss:_ZN10CStreaming16ms_aInfoForModelE.flags[ebp]
		CPatch::PatchPointer(0x40E3E2 + 1, CStreaming__ms_aInfoForModel.gta_vc);	// mov     edi, offset _ZN10CStreaming16ms_aInfoForModelE; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40E3EE + 2, CStreaming__ms_aInfoForModel.gta_vc);	// mov     dword ptr ss:_ZN10CStreaming16ms_aInfoForModelE.usNext[ebp], eax; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40E3F4 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> usPrev);	// mov     ss:_ZN10CStreaming16ms_aInfoForModelE.usPrev[ebp], offset dword_A0DAE0
		CPatch::PatchPointer(0x40E3FE + 2, CStreaming__ms_aInfoForModel.gta_vc);	// mov     eax, dword ptr ss:_ZN10CStreaming16ms_aInfoForModelE.usNext[ebp]; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40E465 + 1, CStreaming__ms_aInfoForModel.gta_vc);	// mov     edx, offset _ZN10CStreaming16ms_aInfoForModelE; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40E471 + 2, CStreaming__ms_aInfoForModel.gta_vc);	// mov     dword ptr ss:_ZN10CStreaming16ms_aInfoForModelE.usNext[ebp], eax; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40E477 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> usPrev);	// mov     ss:_ZN10CStreaming16ms_aInfoForModelE.usPrev[ebp], offset dword_97871C
		CPatch::PatchPointer(0x40E481 + 2, CStreaming__ms_aInfoForModel.gta_vc);	// mov     eax, dword ptr ss:_ZN10CStreaming16ms_aInfoForModelE.usNext[ebp]; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40E4A3 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> uiLoadFlag);	// mov     ss:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[ebp], 2
		CPatch::PatchPointer(0x40E4AA + 2, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// mov     ss:_ZN10CStreaming16ms_aInfoForModelE.flags[ebp], bl
		CPatch::PatchPointer(0x40E4C7 + 3, &CStreaming__ms_aInfoForModel.gta_vc -> _unk_inCdImage);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE._unk_inCdImage[eax*4], 0
		CPatch::PatchPointer(0x40F0B8 + 2, CStreaming__ms_aInfoForModel.gta_vc);	// sub     edx, offset _ZN10CStreaming16ms_aInfoForModelE; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40FD69 + 2, CStreaming__ms_aInfoForModel.gta_vc);	// add     edx, offset _ZN10CStreaming16ms_aInfoForModelE; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40FDC2 + 4, &CStreaming__ms_aInfoForModel.gta_vc ->usNextOnCd);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usNextOnCd[esi*4], ax
		CPatch::PatchPointer(0x40FED8 + 2, CStreaming__ms_aInfoForModel.gta_vc);	// sub     edx, offset _ZN10CStreaming16ms_aInfoForModelE; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40FF80 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[ebx], 1
		CPatch::PatchPointer(0x40FF9D + 2, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.flags[ebx], 0
		CPatch::PatchPointer(0x40FFB5 + 1, CStreaming__ms_aInfoForModel.gta_vc);	// mov     esi, offset _ZN10CStreaming16ms_aInfoForModelE; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x40FFCB + 3, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// movzx   eax, ss:_ZN10CStreaming16ms_aInfoForModelE.flags[ebp]
		CPatch::PatchPointer(0x40FFD7 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// and     ss:_ZN10CStreaming16ms_aInfoForModelE.flags[ebp], 0FDh
		CPatch::PatchPointer(0x40FFF4 + 3, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// movzx   eax, ss:_ZN10CStreaming16ms_aInfoForModelE.flags[ebp]
		CPatch::PatchPointer(0x410000 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> uiLoadFlag);	// cmp     ss:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[ebp], 1
		CPatch::PatchPointer(0x410009 + 2, CStreaming__ms_aInfoForModel.gta_vc);	// cmp     dword ptr ss:_ZN10CStreaming16ms_aInfoForModelE.usNext[ebp], 0; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x41001D + 2, CStreaming__ms_aInfoForModel.gta_vc);	// mov     dword ptr ss:_ZN10CStreaming16ms_aInfoForModelE.usNext[ebp], eax; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x410023 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> usPrev);	// mov     ss:_ZN10CStreaming16ms_aInfoForModelE.usPrev[ebp], offset dword_A0DAE0
		CPatch::PatchPointer(0x41002D + 2, CStreaming__ms_aInfoForModel.gta_vc);	// mov     eax, dword ptr ss:_ZN10CStreaming16ms_aInfoForModelE.usNext[ebp]; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x4100AA + 2, &CStreaming__ms_aInfoForModel.gta_vc -> uiLoadFlag);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[edx], 0
		CPatch::PatchPointer(0x4100B2 + 2, CStreaming__ms_aInfoForModel.gta_vc);	// mov     dword ptr ds:_ZN10CStreaming16ms_aInfoForModelE.usNext[edx], 0; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x4100BC + 2, &CStreaming__ms_aInfoForModel.gta_vc -> usPrev);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usPrev[edx], 0
		CPatch::PatchPointer(0x4100C6 + 3, &CStreaming__ms_aInfoForModel.gta_vc ->usNextOnCd);	// or      ds:_ZN10CStreaming16ms_aInfoForModelE.usNextOnCd[edx], 0FFFFh
		CPatch::PatchPointer(0x4100CE + 2, &CStreaming__ms_aInfoForModel.gta_vc -> _unk_inCdImage);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE._unk_inCdImage[edx], 0
		CPatch::PatchPointer(0x4100D8 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> FilePositionInBlocks);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.FilePositionInBlocks[edx], 0
		CPatch::PatchPointer(0x4100F2 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> uiLoadFlag);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[edi], 0
		CPatch::PatchPointer(0x4100F9 + 2, CStreaming__ms_aInfoForModel.gta_vc);	// mov     dword ptr ds:_ZN10CStreaming16ms_aInfoForModelE.usNext[edi], 0; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x410103 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> usPrev);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usPrev[edi], 0
		CPatch::PatchPointer(0x41010D + 3, &CStreaming__ms_aInfoForModel.gta_vc ->usNextOnCd);	// or      ds:_ZN10CStreaming16ms_aInfoForModelE.usNextOnCd[edi], 0FFFFh
		CPatch::PatchPointer(0x410115 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> _unk_inCdImage);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE._unk_inCdImage[edi], 0
		CPatch::PatchPointer(0x41011F + 2, &CStreaming__ms_aInfoForModel.gta_vc -> FilePositionInBlocks);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.FilePositionInBlocks[edi], 0
		CPatch::PatchPointer(0x410136 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> uiLoadFlag);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[ebx], 0
		CPatch::PatchPointer(0x41013D + 2, CStreaming__ms_aInfoForModel.gta_vc);	// mov     dword ptr ds:_ZN10CStreaming16ms_aInfoForModelE.usNext[ebx], 0; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x410147 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> usPrev);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usPrev[ebx], 0
		CPatch::PatchPointer(0x410151 + 3, &CStreaming__ms_aInfoForModel.gta_vc ->usNextOnCd);	// or      ds:_ZN10CStreaming16ms_aInfoForModelE.usNextOnCd[ebx], 0FFFFh
		CPatch::PatchPointer(0x410159 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> _unk_inCdImage);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE._unk_inCdImage[ebx], 0
		CPatch::PatchPointer(0x410163 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> FilePositionInBlocks);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.FilePositionInBlocks[ebx], 0
		CPatch::PatchPointer(0x41017B + 2, &CStreaming__ms_aInfoForModel.gta_vc -> uiLoadFlag);	// mov     ss:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[ebp], 0
		CPatch::PatchPointer(0x410182 + 2, CStreaming__ms_aInfoForModel.gta_vc);	// mov     dword ptr ss:_ZN10CStreaming16ms_aInfoForModelE.usNext[ebp], 0; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x41018C + 2, &CStreaming__ms_aInfoForModel.gta_vc -> usPrev);	// mov     ss:_ZN10CStreaming16ms_aInfoForModelE.usPrev[ebp], 0
		CPatch::PatchPointer(0x410196 + 3, &CStreaming__ms_aInfoForModel.gta_vc ->usNextOnCd);	// or      ss:_ZN10CStreaming16ms_aInfoForModelE.usNextOnCd[ebp], 0FFFFh
		CPatch::PatchPointer(0x41019E + 2, &CStreaming__ms_aInfoForModel.gta_vc -> _unk_inCdImage);	// mov     ss:_ZN10CStreaming16ms_aInfoForModelE._unk_inCdImage[ebp], 0
		CPatch::PatchPointer(0x4101A8 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> FilePositionInBlocks);	// mov     ss:_ZN10CStreaming16ms_aInfoForModelE.FilePositionInBlocks[ebp], 0
		CPatch::PatchPointer(0x4101BF + 2, &CStreaming__ms_aInfoForModel.gta_vc -> uiLoadFlag);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[esi], 0
		CPatch::PatchPointer(0x4101C6 + 2, CStreaming__ms_aInfoForModel.gta_vc);	// mov     dword ptr ds:_ZN10CStreaming16ms_aInfoForModelE.usNext[esi], 0; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x4101D0 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> usPrev);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usPrev[esi], 0
		CPatch::PatchPointer(0x4101DA + 3, &CStreaming__ms_aInfoForModel.gta_vc ->usNextOnCd);	// or      ds:_ZN10CStreaming16ms_aInfoForModelE.usNextOnCd[esi], 0FFFFh
		CPatch::PatchPointer(0x4101E2 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> _unk_inCdImage);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE._unk_inCdImage[esi], 0
		CPatch::PatchPointer(0x4101EC + 2, &CStreaming__ms_aInfoForModel.gta_vc -> FilePositionInBlocks);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.FilePositionInBlocks[esi], 0
		CPatch::PatchPointer(0x410203 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> uiLoadFlag);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[ecx], 0
		CPatch::PatchPointer(0x41020A + 2, CStreaming__ms_aInfoForModel.gta_vc);	// mov     dword ptr ds:_ZN10CStreaming16ms_aInfoForModelE.usNext[ecx], 0; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x410214 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> usPrev);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usPrev[ecx], 0
		CPatch::PatchPointer(0x41021E + 3, &CStreaming__ms_aInfoForModel.gta_vc ->usNextOnCd);	// or      ds:_ZN10CStreaming16ms_aInfoForModelE.usNextOnCd[ecx], 0FFFFh
		CPatch::PatchPointer(0x410226 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> _unk_inCdImage);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE._unk_inCdImage[ecx], 0
		CPatch::PatchPointer(0x410230 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> FilePositionInBlocks);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.FilePositionInBlocks[ecx], 0
		CPatch::PatchPointer(0x410247 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> uiLoadFlag);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[edi], 0
		CPatch::PatchPointer(0x41024E + 2, CStreaming__ms_aInfoForModel.gta_vc);	// mov     dword ptr ds:_ZN10CStreaming16ms_aInfoForModelE.usNext[edi], 0; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x410258 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> usPrev);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usPrev[edi], 0
		CPatch::PatchPointer(0x410262 + 3, &CStreaming__ms_aInfoForModel.gta_vc ->usNextOnCd);	// or      ds:_ZN10CStreaming16ms_aInfoForModelE.usNextOnCd[edi], 0FFFFh
		CPatch::PatchPointer(0x41026A + 2, &CStreaming__ms_aInfoForModel.gta_vc -> _unk_inCdImage);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE._unk_inCdImage[edi], 0
		CPatch::PatchPointer(0x410274 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> FilePositionInBlocks);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.FilePositionInBlocks[edi], 0
		CPatch::PatchPointer(0x41028E + 2, &CStreaming__ms_aInfoForModel.gta_vc -> uiLoadFlag);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[ebx], 0
		CPatch::PatchPointer(0x410295 + 2, CStreaming__ms_aInfoForModel.gta_vc);	// mov     dword ptr ds:_ZN10CStreaming16ms_aInfoForModelE.usNext[ebx], 0; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x41029F + 2, &CStreaming__ms_aInfoForModel.gta_vc -> usPrev);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usPrev[ebx], 0
		CPatch::PatchPointer(0x4102A9 + 3, &CStreaming__ms_aInfoForModel.gta_vc ->usNextOnCd);	// or      ds:_ZN10CStreaming16ms_aInfoForModelE.usNextOnCd[ebx], 0FFFFh
		CPatch::PatchPointer(0x4102B6 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> _unk_inCdImage);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE._unk_inCdImage[ebx], 0
		CPatch::PatchPointer(0x4102C0 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> FilePositionInBlocks);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.FilePositionInBlocks[ebx], 0
		CPatch::PatchPointer(0x4102E7 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> uiLoadFlag);	// mov     ss:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[ebp], 0
		CPatch::PatchPointer(0x4102EE + 2, CStreaming__ms_aInfoForModel.gta_vc);	// mov     dword ptr ss:_ZN10CStreaming16ms_aInfoForModelE.usNext[ebp], 0; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x4102F8 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> usPrev);	// mov     ss:_ZN10CStreaming16ms_aInfoForModelE.usPrev[ebp], 0
		CPatch::PatchPointer(0x410302 + 3, &CStreaming__ms_aInfoForModel.gta_vc ->usNextOnCd);	// or      ss:_ZN10CStreaming16ms_aInfoForModelE.usNextOnCd[ebp], 0FFFFh
		CPatch::PatchPointer(0x41030A + 2, &CStreaming__ms_aInfoForModel.gta_vc -> _unk_inCdImage);	// mov     ss:_ZN10CStreaming16ms_aInfoForModelE._unk_inCdImage[ebp], 0
		CPatch::PatchPointer(0x410314 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> FilePositionInBlocks);	// mov     ss:_ZN10CStreaming16ms_aInfoForModelE.FilePositionInBlocks[ebp], 0
		CPatch::PatchPointer(0x41033C + 2, &CStreaming__ms_aInfoForModel.gta_vc -> uiLoadFlag);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[esi], 0
		CPatch::PatchPointer(0x410343 + 2, CStreaming__ms_aInfoForModel.gta_vc);	// mov     dword ptr ds:_ZN10CStreaming16ms_aInfoForModelE.usNext[esi], 0; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x41034D + 2, &CStreaming__ms_aInfoForModel.gta_vc -> usPrev);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usPrev[esi], 0
		CPatch::PatchPointer(0x410357 + 3, &CStreaming__ms_aInfoForModel.gta_vc ->usNextOnCd);	// or      ds:_ZN10CStreaming16ms_aInfoForModelE.usNextOnCd[esi], 0FFFFh
		CPatch::PatchPointer(0x41035F + 2, &CStreaming__ms_aInfoForModel.gta_vc -> _unk_inCdImage);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE._unk_inCdImage[esi], 0
		CPatch::PatchPointer(0x410369 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> FilePositionInBlocks);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.FilePositionInBlocks[esi], 0
		CPatch::PatchPointer(0x41039E + 2, &CStreaming__ms_aInfoForModel.gta_vc -> uiLoadFlag);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[ecx], 0
		CPatch::PatchPointer(0x4103A5 + 2, CStreaming__ms_aInfoForModel.gta_vc);	// mov     dword ptr ds:_ZN10CStreaming16ms_aInfoForModelE.usNext[ecx], 0; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x4103AF + 2, &CStreaming__ms_aInfoForModel.gta_vc -> usPrev);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usPrev[ecx], 0
		CPatch::PatchPointer(0x4103B9 + 3, &CStreaming__ms_aInfoForModel.gta_vc ->usNextOnCd);	// or      ds:_ZN10CStreaming16ms_aInfoForModelE.usNextOnCd[ecx], 0FFFFh
		CPatch::PatchPointer(0x4103C1 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> _unk_inCdImage);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE._unk_inCdImage[ecx], 0
		CPatch::PatchPointer(0x4103CB + 2, &CStreaming__ms_aInfoForModel.gta_vc -> FilePositionInBlocks);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.FilePositionInBlocks[ecx], 0
		CPatch::PatchPointer(0x410400 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> uiLoadFlag);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[edi], 0
		CPatch::PatchPointer(0x410407 + 2, CStreaming__ms_aInfoForModel.gta_vc);	// mov     dword ptr ds:_ZN10CStreaming16ms_aInfoForModelE.usNext[edi], 0; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x410411 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> usPrev);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usPrev[edi], 0
		CPatch::PatchPointer(0x41041B + 3, &CStreaming__ms_aInfoForModel.gta_vc ->usNextOnCd);	// or      ds:_ZN10CStreaming16ms_aInfoForModelE.usNextOnCd[edi], 0FFFFh
		CPatch::PatchPointer(0x410423 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> _unk_inCdImage);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE._unk_inCdImage[edi], 0
		CPatch::PatchPointer(0x41042D + 2, &CStreaming__ms_aInfoForModel.gta_vc -> FilePositionInBlocks);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.FilePositionInBlocks[edi], 0
		CPatch::PatchPointer(0x410462 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> uiLoadFlag);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[ebx], 0
		CPatch::PatchPointer(0x410469 + 2, CStreaming__ms_aInfoForModel.gta_vc);	// mov     dword ptr ds:_ZN10CStreaming16ms_aInfoForModelE.usNext[ebx], 0; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x410473 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> usPrev);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usPrev[ebx], 0
		CPatch::PatchPointer(0x41047D + 3, &CStreaming__ms_aInfoForModel.gta_vc ->usNextOnCd);	// or      ds:_ZN10CStreaming16ms_aInfoForModelE.usNextOnCd[ebx], 0FFFFh
		CPatch::PatchPointer(0x410485 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> _unk_inCdImage);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE._unk_inCdImage[ebx], 0
		CPatch::PatchPointer(0x41048F + 2, &CStreaming__ms_aInfoForModel.gta_vc -> FilePositionInBlocks);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.FilePositionInBlocks[ebx], 0
		CPatch::PatchPointer(0x4104BA + 2, &CStreaming__ms_aInfoForModel.gta_vc -> uiLoadFlag);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[edx], 0
		CPatch::PatchPointer(0x4104C1 + 2, CStreaming__ms_aInfoForModel.gta_vc);	// mov     dword ptr ds:_ZN10CStreaming16ms_aInfoForModelE.usNext[edx], 0; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x4104CB + 2, &CStreaming__ms_aInfoForModel.gta_vc -> usPrev);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usPrev[edx], 0
		CPatch::PatchPointer(0x4104D5 + 3, &CStreaming__ms_aInfoForModel.gta_vc ->usNextOnCd);	// or      ds:_ZN10CStreaming16ms_aInfoForModelE.usNextOnCd[edx], 0FFFFh
		CPatch::PatchPointer(0x4104DD + 2, &CStreaming__ms_aInfoForModel.gta_vc -> _unk_inCdImage);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE._unk_inCdImage[edx], 0
		CPatch::PatchPointer(0x4104E7 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> FilePositionInBlocks);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.FilePositionInBlocks[edx], 0
		CPatch::PatchPointer(0x41050C + 2, &CStreaming__ms_aInfoForModel.gta_vc -> uiLoadFlag);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[eax], 0
		CPatch::PatchPointer(0x410513 + 2, CStreaming__ms_aInfoForModel.gta_vc);	// mov     dword ptr ds:_ZN10CStreaming16ms_aInfoForModelE.usNext[eax], 0; CStreaming::ms_aInfoForModel
		CPatch::PatchPointer(0x41051D + 2, &CStreaming__ms_aInfoForModel.gta_vc -> usPrev);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.usPrev[eax], 0
		CPatch::PatchPointer(0x410527 + 3, &CStreaming__ms_aInfoForModel.gta_vc ->usNextOnCd);	// or      ds:_ZN10CStreaming16ms_aInfoForModelE.usNextOnCd[eax], 0FFFFh
		CPatch::PatchPointer(0x41052F + 2, &CStreaming__ms_aInfoForModel.gta_vc -> _unk_inCdImage);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE._unk_inCdImage[eax], 0
		CPatch::PatchPointer(0x410539 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> FilePositionInBlocks);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.FilePositionInBlocks[eax], 0
		CPatch::PatchPointer(0x410545 + 2, &CStreaming__ms_aInfoForModel.gta_vc[240].uiLoadFlag);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+12C0h, 1
		CPatch::PatchPointer(0x410553 + 2, &CStreaming__ms_aInfoForModel.gta_vc[241].uiLoadFlag);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+12D4h, 1
		CPatch::PatchPointer(0x410564 + 2, &CStreaming__ms_aInfoForModel.gta_vc[242].uiLoadFlag);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+12E8h, 1
		CPatch::PatchPointer(0x41056B + 2, &CStreaming__ms_aInfoForModel.gta_vc[243].uiLoadFlag);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+12FCh, 1
		CPatch::PatchPointer(0x410572 + 2, &CStreaming__ms_aInfoForModel.gta_vc[244].uiLoadFlag);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+1310h, 1
		CPatch::PatchPointer(0x410579 + 2, &CStreaming__ms_aInfoForModel.gta_vc[245].uiLoadFlag);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+1324h, 1
		CPatch::PatchPointer(0x410580 + 2, &CStreaming__ms_aInfoForModel.gta_vc[246].uiLoadFlag);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+1338h, 1
		CPatch::PatchPointer(0x410587 + 2, &CStreaming__ms_aInfoForModel.gta_vc[247].uiLoadFlag);	// mov     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+134Ch, 1
		CPatch::PatchPointer(0x4105D8 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> uiLoadFlag);	// mov     ss:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[ebp], 1
		CPatch::PatchPointer(0x4105DF + 2, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// mov     ss:_ZN10CStreaming16ms_aInfoForModelE.flags[ebp], 1
		CPatch::PatchPointer(0x41064F + 2, &CStreaming__ms_aInfoForModel.gta_vc[GetBaseID(FILE_TYPE_TXD)].uiLoadFlag);	// mov     ds:(_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+1FBD0h)[edx], 1
		CPatch::PatchPointer(0x41CA74 + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_AMBULAN].uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+0B68h, 1
		CPatch::PatchPointer(0x41CA7F + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_MEDIC].uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+64h, 1
		CPatch::PatchPointer(0x41CAB2 + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_AMBULAN].uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+0B68h, 1
		CPatch::PatchPointer(0x41CB96 + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_FIRETRUK].uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+0AB4h, 1
		CPatch::PatchPointer(0x41CBA1 + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_FIREMAN].uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+78h, 1
		CPatch::PatchPointer(0x41CC07 + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_FIRETRUK].uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+0AB4h, 1
		CPatch::PatchPointer(0x426881 + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICECHEE].uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+1270h, 1
		CPatch::PatchPointer(0x4268A2 + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE1].uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+794h, 1; jumptable 0042689B case 0
		CPatch::PatchPointer(0x4268AB + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE2].uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+7A8h, 1
		CPatch::PatchPointer(0x4268BD + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE3].uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+7BCh, 1; jumptable 0042689B case 1
		CPatch::PatchPointer(0x4268C6 + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE4].uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+7D0h, 1
		CPatch::PatchPointer(0x4268D8 + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE5].uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+7E4h, 1; jumptable 0042689B case 2
		CPatch::PatchPointer(0x4268E1 + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE6].uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+7F8h, 1
		CPatch::PatchPointer(0x4268F3 + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE7].uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+80Ch, 1; jumptable 0042689B case 3
		CPatch::PatchPointer(0x4268FC + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_VICE8].uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+820h, 1
		CPatch::PatchPointer(0x426924 + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_ENFORCER].uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+0C44h, 1
		CPatch::PatchPointer(0x42692D + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_SWAT].uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+28h, 1
		CPatch::PatchPointer(0x4269A4 + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_FBIRANCH].uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+1130h, 1
		CPatch::PatchPointer(0x4269AD + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_FBI].uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+3Ch, 1
		CPatch::PatchPointer(0x4269D4 + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_RHINO].uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+0CA8h, 1
		CPatch::PatchPointer(0x4269DD + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_BARRACKS].uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+0CBCh, 1
		CPatch::PatchPointer(0x4269E6 + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_ARMY].uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+50h, 1
		CPatch::PatchPointer(0x426BD5 + 3, &CStreaming__ms_aInfoForModel.gta_vc -> uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[ecx*4], 1
		CPatch::PatchPointer(0x426BE8 + 3, &CStreaming__ms_aInfoForModel.gta_vc -> uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[edx*4], 1
		CPatch::PatchPointer(0x426D13 + 3, &CStreaming__ms_aInfoForModel.gta_vc -> uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[eax*4], 1
		CPatch::PatchPointer(0x4273A1 + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_PREDATOR].uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+0C80h, 1
		CPatch::PatchPointer(0x427502 + 3, &CStreaming__ms_aInfoForModel.gta_vc -> uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[eax*4], 1
		CPatch::PatchPointer(0x42AE8B + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_KAUFMAN].uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+10E0h, 1; jumptable 0042AE84 case 1
		CPatch::PatchPointer(0x42DF54 + 3, &CStreaming__ms_aInfoForModel.gta_vc -> uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[eax*4], 1
		CPatch::PatchPointer(0x431B54 + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_BOMB].uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+16BCh, 1
		CPatch::PatchPointer(0x44091B + 3, &CStreaming__ms_aInfoForModel.gta_vc -> uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[eax*4], 1
		CPatch::PatchPointer(0x4436FA + 3, &CStreaming__ms_aInfoForModel.gta_vc -> uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[eax*4], 1
		CPatch::PatchPointer(0x443B41 + 3, &CStreaming__ms_aInfoForModel.gta_vc -> uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[eax*4], 1
		CPatch::PatchPointer(0x44408A + 3, &CStreaming__ms_aInfoForModel.gta_vc -> uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[eax*4], 1
		CPatch::PatchPointer(0x44E761 + 2, &CStreaming__ms_aInfoForModel.gta_vc -> uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag, 1
		CPatch::PatchPointer(0x458A48 + 3, &CStreaming__ms_aInfoForModel.gta_vc -> uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[edx*4], 1
		CPatch::PatchPointer(0x4AE872 + 4, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// movzx   ebx, ds:_ZN10CStreaming16ms_aInfoForModelE.flags[eax*4]
		CPatch::PatchPointer(0x4AE922 + 3, &CStreaming__ms_aInfoForModel.gta_vc -> flags);	// movzx   ebx, ds:_ZN10CStreaming16ms_aInfoForModelE.flags[edi]
		CPatch::PatchPointer(0x4AE93A + 2, &CStreaming__ms_aInfoForModel.gta_vc -> uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[edi], 1
		CPatch::PatchPointer(0x4C75F9 + 3, &CStreaming__ms_aInfoForModel.gta_vc -> uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[eax*4], 1
		CPatch::PatchPointer(0x4FFC9C + 3, &CStreaming__ms_aInfoForModel.gta_vc -> uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[edi*4], 1
		CPatch::PatchPointer(0x4FFCAE + 3, &CStreaming__ms_aInfoForModel.gta_vc -> uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[esi*4], 1
		CPatch::PatchPointer(0x4FFD38 + 3, &CStreaming__ms_aInfoForModel.gta_vc -> uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[edi*4], 1
		CPatch::PatchPointer(0x4FFD4A + 3, &CStreaming__ms_aInfoForModel.gta_vc -> uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[esi*4], 1
		CPatch::PatchPointer(0x53C522 + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_COP].uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+14h, 1
		CPatch::PatchPointer(0x53C547 + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_FBI].uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+3Ch, 1
		CPatch::PatchPointer(0x53C55E + 3, &CStreaming__ms_aInfoForModel.gta_vc -> uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[eax*4], 1
		CPatch::PatchPointer(0x53C57E + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_SWAT].uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+28h, 1
		CPatch::PatchPointer(0x53C595 + 3, &CStreaming__ms_aInfoForModel.gta_vc -> uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[eax*4], 1
		CPatch::PatchPointer(0x53C5B7 + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_ARMY].uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+50h, 1
		CPatch::PatchPointer(0x53C5CE + 3, &CStreaming__ms_aInfoForModel.gta_vc -> uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[eax*4], 1
		CPatch::PatchPointer(0x53C5E6 + 3, &CStreaming__ms_aInfoForModel.gta_vc -> uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[eax*4], 1
		CPatch::PatchPointer(0x5A71F6 + 3, &CStreaming__ms_aInfoForModel.gta_vc -> uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[eax*4], 1
		CPatch::PatchPointer(0x5A72B3 + 3, &CStreaming__ms_aInfoForModel.gta_vc -> uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[eax*4], 1
		CPatch::PatchPointer(0x5ABE40 + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_SWAT].uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+28h, 1
		CPatch::PatchPointer(0x5ACDD6 + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_CHOPPER].uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+0CE4h, 1
		CPatch::PatchPointer(0x5ACE47 + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_CHOPPER].uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+0CE4h, 1
		CPatch::PatchPointer(0x5B18C0 + 3, &CStreaming__ms_aInfoForModel.gta_vc -> uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[eax*4], 1
		CPatch::PatchPointer(0x5B1DE2 + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_DEADDODO].uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+0E24h, 1
		CPatch::PatchPointer(0x5B1E8E + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_CHOPPER].uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+0CE4h, 1
		CPatch::PatchPointer(0x5B894B + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_CHROMEGUN].uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+15A4h, 1
		CPatch::PatchPointer(0x5B89AC + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_GOLFCLUB].uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+1464h, 1
		CPatch::PatchPointer(0x603CE0 + 3, &CStreaming__ms_aInfoForModel.gta_vc -> uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[eax*4], 1
		CPatch::PatchPointer(0x603E11 + 3, &CStreaming__ms_aInfoForModel.gta_vc -> uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[eax*4], 1
		CPatch::PatchPointer(0x604624 + 3, &CStreaming__ms_aInfoForModel.gta_vc -> uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[eax*4], 1
		CPatch::PatchPointer(0x604B63 + 3, &CStreaming__ms_aInfoForModel.gta_vc -> uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[eax*4], 1
		CPatch::PatchPointer(0x621381 + 3, &CStreaming__ms_aInfoForModel.gta_vc -> uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[edx*4], 1
		CPatch::PatchPointer(0x6215ED + 3, &CStreaming__ms_aInfoForModel.gta_vc -> uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[edi*4], 1
		CPatch::PatchPointer(0x62171C + 3, &CStreaming__ms_aInfoForModel.gta_vc -> uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[edx*4], 1
		CPatch::PatchPointer(0x623989 + 3, &CStreaming__ms_aInfoForModel.gta_vc -> uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[eax*4], 1
		CPatch::PatchPointer(0x62A018 + 3, &CStreaming__ms_aInfoForModel.gta_vc[GetBaseID(FILE_TYPE_COL)].flags);	// movzx   eax, ds:(_ZN10CStreaming16ms_aInfoForModelE.flags+26804h)[esi]
		CPatch::PatchPointer(0x62A307 + 2, &CStreaming__ms_aInfoForModel.gta_vc[GetBaseID(FILE_TYPE_COL)].uiLoadFlag);	// cmp     ds:(_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+26804h)[esi], 1
		CPatch::PatchPointer(0x63256A + 2, &CStreaming__ms_aInfoForModel.gta_vc[MODEL_SWAT].uiLoadFlag);	// cmp     ds:_ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag+28h, 1
	}
	#endif

	// Patch functions, so the last file ID doesn't have to be divisible by 6
	#if TRUE
	{
		// Patch CStreaming::Init
		if(true)
		{
			CPatch::RedirectCode(0x4100A8, &patch_CStreaming__Init_4100A8);

			CPatch::RedirectCodeOneByte(0x4102E7, 0x410321);
			CPatch::NOPinstructions(0x41032B, 7);
			CPatch::RedirectCodeOneByte(0x41033C, 0x410376);
			CPatch::NOPinstructions(0x410380, 3);
			CPatch::NOPinstructions(0x41038D, 7);
			CPatch::RedirectCodeOneByte(0x41039E, 0x4103D8);
			CPatch::NOPinstructions(0x4103E2, 3);
			CPatch::NOPinstructions(0x4103EF, 7);
			CPatch::RedirectCodeOneByte(0x410400, 0x41043A);
			CPatch::NOPinstructions(0x410444, 3);
			CPatch::NOPinstructions(0x410451, 7);
			CPatch::RedirectCodeOneByte(0x410462, 0x4104A2);
			CPatch::NOPinstructions(0x4104A9, 7);
			CPatch::RedirectCodeOneByte(0x4104BA, 0x4104F4);
			CPatch::NOPinstructions(0x4104FB, 7);
			CPatch::RedirectCodeOneByte(0x41050C, 0x410543);
		}

		// Patch LoadScene
		if(true)
		{
			CPatch::RedirectCode(0x40B3DA, &patch_LoadScene_40B3DA);
			CPatch::NOPinstructions(0x40B4B7, 8);
		}
	}
	#endif
	
	// Patch base IDs and multiplies
	#if TRUE
	{
		// 6500 (TXD index)
		{
			CPatch::PatchUINT32(0x401571 + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     ebp, 1964h
			CPatch::PatchUINT32(0x40998A + 1, GetBaseID(FILE_TYPE_TXD));	// cmp     eax, 1964h
			CPatch::PatchUINT32(0x409B25 + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     ebp, 1964h
			CPatch::PatchUINT32(0x409B9E + 1, GetBaseID(FILE_TYPE_TXD));	// mov     edx, 1964h
			CPatch::PatchUINT32(0x409BA7 + 2, GetBaseID(FILE_TYPE_TXD));	// lea     eax, [ecx+1964h]
			CPatch::PatchUINT32(0x409BC0 + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     edx, 1964h
			CPatch::PatchUINT32(0x409BC8 + 2, GetBaseID(FILE_TYPE_TXD));	// lea     ebp, [ecx+1964h]
			CPatch::PatchUINT32(0x409BD6 + 2, GetBaseID(FILE_TYPE_TXD));	// lea     edx, [ecx+1964h]
			CPatch::PatchUINT32(0x409C35 + 2, GetBaseID(FILE_TYPE_TXD));	// add     ecx, 1964h
			CPatch::PatchUINT32(0x409C88 + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     ecx, 1964h
			CPatch::PatchUINT32(0x409D08 + 1, GetBaseID(FILE_TYPE_TXD));	// mov     eax, 1964h
			CPatch::PatchUINT32(0x409D0D + 2, GetBaseID(FILE_TYPE_TXD));	// lea     ecx, [edx+1964h]
			CPatch::PatchUINT32(0x409D26 + 1, GetBaseID(FILE_TYPE_TXD));	// cmp     eax, 1964h
			CPatch::PatchUINT32(0x409D2D + 2, GetBaseID(FILE_TYPE_TXD));	// lea     eax, [edx+1964h]
			CPatch::PatchUINT32(0x409D3A + 2, GetBaseID(FILE_TYPE_TXD));	// lea     eax, [edx+1964h]
			CPatch::PatchUINT32(0x409D98 + 2, GetBaseID(FILE_TYPE_TXD));	// add     edx, 1964h
			CPatch::PatchUINT32(0x409DC7 + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     edx, 1964h
			CPatch::PatchUINT32(0x409E48 + 1, GetBaseID(FILE_TYPE_TXD));	// mov     eax, 1964h
			CPatch::PatchUINT32(0x409E4D + 2, GetBaseID(FILE_TYPE_TXD));	// lea     edx, [ecx+1964h]
			CPatch::PatchUINT32(0x409E66 + 1, GetBaseID(FILE_TYPE_TXD));	// cmp     eax, 1964h
			CPatch::PatchUINT32(0x409E6D + 2, GetBaseID(FILE_TYPE_TXD));	// lea     eax, [ecx+1964h]
			CPatch::PatchUINT32(0x409E7A + 2, GetBaseID(FILE_TYPE_TXD));	// lea     eax, [ecx+1964h]
			CPatch::PatchUINT32(0x409ED8 + 2, GetBaseID(FILE_TYPE_TXD));	// add     ecx, 1964h
			CPatch::PatchUINT32(0x409F07 + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     ecx, 1964h
			CPatch::PatchUINT32(0x409F88 + 1, GetBaseID(FILE_TYPE_TXD));	// mov     eax, 1964h
			CPatch::PatchUINT32(0x409F8D + 2, GetBaseID(FILE_TYPE_TXD));	// lea     ecx, [edx+1964h]
			CPatch::PatchUINT32(0x409FA6 + 1, GetBaseID(FILE_TYPE_TXD));	// cmp     eax, 1964h
			CPatch::PatchUINT32(0x409FAD + 2, GetBaseID(FILE_TYPE_TXD));	// lea     eax, [edx+1964h]
			CPatch::PatchUINT32(0x409FBA + 2, GetBaseID(FILE_TYPE_TXD));	// lea     eax, [edx+1964h]
			CPatch::PatchUINT32(0x40A018 + 2, GetBaseID(FILE_TYPE_TXD));	// add     edx, 1964h
			CPatch::PatchUINT32(0x40A0D6 + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     esi, 1964h
			CPatch::PatchUINT32(0x40A152 + 1, GetBaseID(FILE_TYPE_TXD));	// mov     eax, 1964h
			CPatch::PatchUINT32(0x40A157 + 2, GetBaseID(FILE_TYPE_TXD));	// lea     edx, [ecx+1964h]
			CPatch::PatchUINT32(0x40A170 + 1, GetBaseID(FILE_TYPE_TXD));	// cmp     eax, 1964h
			CPatch::PatchUINT32(0x40A177 + 2, GetBaseID(FILE_TYPE_TXD));	// lea     eax, [ecx+1964h]
			CPatch::PatchUINT32(0x40A184 + 2, GetBaseID(FILE_TYPE_TXD));	// lea     ebx, [ecx+1964h]
			CPatch::PatchUINT32(0x40A1E2 + 2, GetBaseID(FILE_TYPE_TXD));	// add     ecx, 1964h
			CPatch::PatchUINT32(0x40A39D + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     esi, 1964h
			CPatch::PatchUINT32(0x40A413 + 1, GetBaseID(FILE_TYPE_TXD));	// mov     ecx, 1964h
			CPatch::PatchUINT32(0x40A41C + 2, GetBaseID(FILE_TYPE_TXD));	// lea     eax, [edx+1964h]
			CPatch::PatchUINT32(0x40A435 + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     ecx, 1964h
			CPatch::PatchUINT32(0x40A43D + 2, GetBaseID(FILE_TYPE_TXD));	// lea     ebx, [edx+1964h]
			CPatch::PatchUINT32(0x40A44B + 2, GetBaseID(FILE_TYPE_TXD));	// lea     ebp, [edx+1964h]
			CPatch::PatchUINT32(0x40A4B0 + 2, GetBaseID(FILE_TYPE_TXD));	// add     edx, 1964h
			CPatch::PatchUINT32(0x40A610 + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     edx, 1964h
			CPatch::PatchUINT32(0x40A698 + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     ecx, 1964h
			CPatch::PatchUINT32(0x40A718 + 1, GetBaseID(FILE_TYPE_TXD));	// mov     eax, 1964h
			CPatch::PatchUINT32(0x40A71D + 2, GetBaseID(FILE_TYPE_TXD));	// lea     ecx, [edx+1964h]
			CPatch::PatchUINT32(0x40A736 + 1, GetBaseID(FILE_TYPE_TXD));	// cmp     eax, 1964h
			CPatch::PatchUINT32(0x40A73D + 2, GetBaseID(FILE_TYPE_TXD));	// lea     eax, [edx+1964h]
			CPatch::PatchUINT32(0x40A74A + 2, GetBaseID(FILE_TYPE_TXD));	// lea     eax, [edx+1964h]
			CPatch::PatchUINT32(0x40A7A8 + 2, GetBaseID(FILE_TYPE_TXD));	// add     edx, 1964h
			CPatch::PatchUINT32(0x40A7C6 + 1, GetBaseID(FILE_TYPE_TXD));	// mov     eax, 1964h
			CPatch::PatchUINT32(0x40A7CB + 2, GetBaseID(FILE_TYPE_TXD));	// lea     edx, [ecx+1964h]
			CPatch::PatchUINT32(0x40A7E4 + 1, GetBaseID(FILE_TYPE_TXD));	// cmp     eax, 1964h
			CPatch::PatchUINT32(0x40A7EB + 2, GetBaseID(FILE_TYPE_TXD));	// lea     eax, [ecx+1964h]
			CPatch::PatchUINT32(0x40A7F8 + 2, GetBaseID(FILE_TYPE_TXD));	// lea     eax, [ecx+1964h]
			CPatch::PatchUINT32(0x40A856 + 2, GetBaseID(FILE_TYPE_TXD));	// add     ecx, 1964h
			CPatch::PatchUINT32(0x40A8CB + 1, GetBaseID(FILE_TYPE_TXD));	// cmp     eax, 1964h
			CPatch::PatchUINT32(0x40A949 + 1, GetBaseID(FILE_TYPE_TXD));	// mov     eax, 1964h
			CPatch::PatchUINT32(0x40A94E + 2, GetBaseID(FILE_TYPE_TXD));	// lea     ecx, [edx+1964h]
			CPatch::PatchUINT32(0x40A967 + 1, GetBaseID(FILE_TYPE_TXD));	// cmp     eax, 1964h
			CPatch::PatchUINT32(0x40A96E + 2, GetBaseID(FILE_TYPE_TXD));	// lea     eax, [edx+1964h]
			CPatch::PatchUINT32(0x40A97B + 2, GetBaseID(FILE_TYPE_TXD));	// lea     ebp, [edx+1964h]
			CPatch::PatchUINT32(0x40A9E0 + 2, GetBaseID(FILE_TYPE_TXD));	// add     edx, 1964h
			CPatch::PatchUINT32(0x40AD77 + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     ecx, 1964h
			CPatch::PatchUINT32(0x40ADF4 + 1, GetBaseID(FILE_TYPE_TXD));	// mov     edx, 1964h
			CPatch::PatchUINT32(0x40AE04 + 2, GetBaseID(FILE_TYPE_TXD));	// lea     eax, [ecx+1964h]
			CPatch::PatchUINT32(0x40AE1D + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     edx, 1964h
			CPatch::PatchUINT32(0x40AE25 + 2, GetBaseID(FILE_TYPE_TXD));	// lea     edx, [ecx+1964h]
			CPatch::PatchUINT32(0x40AE33 + 2, GetBaseID(FILE_TYPE_TXD));	// lea     edx, [ecx+1964h]
			CPatch::PatchUINT32(0x40AE92 + 2, GetBaseID(FILE_TYPE_TXD));	// add     ecx, 1964h
			CPatch::PatchUINT32(0x40AEB7 + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     ecx, 1964h
			CPatch::PatchUINT32(0x40B815 + 3, GetBaseID(FILE_TYPE_TXD));	// cmp     [esp+20h+var_20], 1964h
			CPatch::PatchUINT32(0x40B92A + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     ebx, 1964h
			CPatch::PatchUINT32(0x40BA2D + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     ebx, 1964h
			CPatch::PatchUINT32(0x40BA62 + 2, GetBaseID(FILE_TYPE_TXD));	// add     edx, 1964h
			CPatch::PatchUINT32(0x40BAF0 + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     ebx, 1964h
			CPatch::PatchUINT32(0x40BD0D + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     ebx, 1964h
			CPatch::PatchUINT32(0x40BD3E + 2, GetBaseID(FILE_TYPE_TXD));	// add     edx, 1964h
			CPatch::PatchUINT32(0x40BF62 + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     ebp, 1964h
			CPatch::PatchUINT32(0x40C01F + 2, GetBaseID(FILE_TYPE_TXD));	// add     edi, 1964h
			CPatch::PatchUINT32(0x40C23D + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     ebp, 1964h
			CPatch::PatchUINT32(0x40C415 + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     ebp, 1964h
			CPatch::PatchUINT32(0x40C5E9 + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     ebp, 1964h
			CPatch::PatchUINT32(0x40C640 + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     ebp, 1964h
			CPatch::PatchUINT32(0x40C6B2 + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     ebp, 1964h
			CPatch::PatchUINT32(0x40C6D0 + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     ebp, 1964h
			CPatch::PatchUINT32(0x40C7A9 + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     ebx, 1964h
			CPatch::PatchUINT32(0x40C7EA + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     ebx, 1964h
			CPatch::PatchUINT32(0x40C857 + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     ebx, 1964h
			CPatch::PatchUINT32(0x40C904 + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     ebx, 1964h
			CPatch::PatchUINT32(0x40C920 + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     ebx, 1964h
			CPatch::PatchUINT32(0x40C9EF + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     ebp, 1964h
			CPatch::PatchUINT32(0x40CA2C + 1, GetBaseID(FILE_TYPE_TXD));	// cmp     eax, 1964h
			CPatch::PatchUINT32(0x40CA5C + 1, GetBaseID(FILE_TYPE_TXD));	// cmp     eax, 1964h
			CPatch::PatchUINT32(0x40CABD + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     ebp, 1964h
			CPatch::PatchUINT32(0x40CAFC + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     ebp, 1964h
			CPatch::PatchUINT32(0x40CB24 + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     ebx, 1964h
			CPatch::PatchUINT32(0x40D45C + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     edi, 1964h
			CPatch::PatchUINT32(0x40D5E5 + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     esi, 1964h
			CPatch::PatchUINT32(0x40D610 + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     esi, 1964h
			CPatch::PatchUINT32(0x40D706 + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     ebx, 1964h
			CPatch::PatchUINT32(0x40D720 + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     ebx, 1964h
			CPatch::PatchUINT32(0x40D824 + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     ebx, 1964h
			CPatch::PatchUINT32(0x40D833 + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     ebx, 1964h
			CPatch::PatchUINT32(0x40D940 + 1, GetBaseID(FILE_TYPE_TXD));	// add     eax, 1964h
			CPatch::PatchUINT32(0x40D98B + 1, GetBaseID(FILE_TYPE_TXD));	// add     eax, 1964h
			CPatch::PatchUINT32(0x40E37C + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     ecx, 1964h
			CPatch::PatchUINT32(0x40E429 + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     ecx, 1964h
			CPatch::PatchUINT32(0x40E43D + 1, GetBaseID(FILE_TYPE_TXD));	// add     eax, 1964h
			CPatch::PatchUINT32(0x40FCD9 + 4, GetBaseID(FILE_TYPE_TXD));	// add     [esp+40h+var_14], 1964h
			CPatch::PatchUINT32(0x40FFA9 + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     esi, 1964h
			CPatch::PatchUINT32(0x40FFDE + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     ebx, 1964h
			CPatch::PatchUINT32(0x410046 + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     ebx, 1964h
			CPatch::PatchUINT32(0x4105FD + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     esi, 1964h
			CPatch::PatchUINT16(0x45E8B8 + 3, GetBaseID(FILE_TYPE_TXD));	// cmp     bp, 1964h
			CPatch::PatchUINT32(0x48DD15 + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     ebx, 1964h
			CPatch::PatchUINT32(0x4C28BD + 1, GetBaseID(FILE_TYPE_TXD));	// add     eax, 1964h
			CPatch::PatchUINT32(0x4C2904 + 1, GetBaseID(FILE_TYPE_TXD));	// add     eax, 1964h
			CPatch::PatchUINT32(0x4C2959 + 1, GetBaseID(FILE_TYPE_TXD));	// add     eax, 1964h
			CPatch::PatchUINT32(0x4C61DD + 1, GetBaseID(FILE_TYPE_TXD));	// add     eax, 1964h
			CPatch::PatchUINT32(0x55F809 + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     edi, 1964h
			CPatch::PatchUINT32(0x56F3F3 + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     ecx, 1964h
			CPatch::PatchUINT32(0x56F3FB + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     ecx, 1964h
			CPatch::PatchUINT32(0x580A45 + 2, GetBaseID(FILE_TYPE_TXD));	// lea     eax, [ebx+1964h]
			CPatch::PatchUINT32(0x604797 + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     ebp, 1964h
			CPatch::PatchUINT32(0x60480D + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     edx, 1964h
			CPatch::PatchUINT32(0x604D3B + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     esi, 1964h
			CPatch::PatchUINT32(0x604DB6 + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     edi, 1964h
			CPatch::PatchUINT32(0x61DADF + 2, GetBaseID(FILE_TYPE_TXD));	// lea     eax, [ebx+1964h]
			CPatch::PatchUINT32(0x61DAF4 + 2, GetBaseID(FILE_TYPE_TXD));	// lea     eax, [ebx+1964h]
			CPatch::PatchUINT32(0x61DC3F + 2, GetBaseID(FILE_TYPE_TXD));	// add     ebx, 1964h
			CPatch::PatchUINT32(0x61DCE2 + 2, GetBaseID(FILE_TYPE_TXD));	// lea     eax, [ebx+1964h]
			CPatch::PatchUINT32(0x62A7A0 + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     ebp, 1964h
			CPatch::PatchUINT32(0x62A9C2 + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     esi, 1964h
		}

		// -6500 (- TXD base index)
		{
			CPatch::PatchUINT32(0x40B93A + 2, -GetBaseID(FILE_TYPE_TXD));	// lea     eax, [ebx-1964h]
			CPatch::PatchUINT32(0x40C439 + 2, -GetBaseID(FILE_TYPE_TXD));	// lea     eax, [ebp-1964h]
			CPatch::PatchUINT32(0x40C483 + 2, -GetBaseID(FILE_TYPE_TXD));	// lea     eax, [ebp-1964h]
			CPatch::PatchUINT32(0x40C4A0 + 2, -GetBaseID(FILE_TYPE_TXD));	// lea     eax, [ebp-1964h]
			CPatch::PatchUINT32(0x40C4BA + 2, -GetBaseID(FILE_TYPE_TXD));	// lea     eax, [ebp-1964h]
			CPatch::PatchUINT32(0x40C6E0 + 2, -GetBaseID(FILE_TYPE_TXD));	// lea     eax, [ebp-1964h]
			CPatch::PatchUINT32(0x40C867 + 2, -GetBaseID(FILE_TYPE_TXD));	// lea     eax, [ebx-1964h]
			CPatch::PatchUINT32(0x40C873 + 2, -GetBaseID(FILE_TYPE_TXD));	// lea     eax, [ebx-1964h]
			CPatch::PatchUINT32(0x40C888 + 2, -GetBaseID(FILE_TYPE_TXD));	// lea     eax, [ebx-1964h]
			CPatch::PatchUINT32(0x40C930 + 2, -GetBaseID(FILE_TYPE_TXD));	// lea     eax, [ebx-1964h]
			CPatch::PatchUINT32(0x40D475 + 2, -GetBaseID(FILE_TYPE_TXD));	// lea     eax, [edi-1964h]
			CPatch::PatchUINT32(0x40D620 + 2, -GetBaseID(FILE_TYPE_TXD));	// lea     eax, [esi-1964h]
			CPatch::PatchUINT32(0x40D631 + 2, -GetBaseID(FILE_TYPE_TXD));	// lea     eax, [esi-1964h]
			CPatch::PatchUINT32(0x40D730 + 2, -GetBaseID(FILE_TYPE_TXD));	// lea     eax, [ebx-1964h]
			CPatch::PatchUINT32(0x40D843 + 2, -GetBaseID(FILE_TYPE_TXD));	// lea     eax, [ebx-1964h]
		}


		// 7885 (COL base index)
		{
			CPatch::PatchUINT32(0x409994 + 1, GetBaseID(FILE_TYPE_COL));	// cmp     eax, 1ECDh
			CPatch::PatchUINT32(0x409B2D + 2, GetBaseID(FILE_TYPE_COL));	// cmp     ebp, 1ECDh
			CPatch::PatchUINT32(0x409BCE + 2, GetBaseID(FILE_TYPE_COL));	// cmp     ebp, 1ECDh
			CPatch::PatchUINT32(0x409C90 + 2, GetBaseID(FILE_TYPE_COL));	// cmp     ecx, 1ECDh
			CPatch::PatchUINT32(0x409D33 + 1, GetBaseID(FILE_TYPE_COL));	// cmp     eax, 1ECDh
			CPatch::PatchUINT32(0x409DCF + 2, GetBaseID(FILE_TYPE_COL));	// cmp     edx, 1ECDh
			CPatch::PatchUINT32(0x409E73 + 1, GetBaseID(FILE_TYPE_COL));	// cmp     eax, 1ECDh
			CPatch::PatchUINT32(0x409F0F + 2, GetBaseID(FILE_TYPE_COL));	// cmp     ecx, 1ECDh
			CPatch::PatchUINT32(0x409FB3 + 1, GetBaseID(FILE_TYPE_COL));	// cmp     eax, 1ECDh
			CPatch::PatchUINT32(0x40A0DE + 2, GetBaseID(FILE_TYPE_COL));	// cmp     esi, 1ECDh
			CPatch::PatchUINT32(0x40A17D + 1, GetBaseID(FILE_TYPE_COL));	// cmp     eax, 1ECDh
			CPatch::PatchUINT32(0x40A3A5 + 2, GetBaseID(FILE_TYPE_COL));	// cmp     esi, 1ECDh
			CPatch::PatchUINT32(0x40A443 + 2, GetBaseID(FILE_TYPE_COL));	// cmp     ebx, 1ECDh
			CPatch::PatchUINT32(0x40A618 + 2, GetBaseID(FILE_TYPE_COL));	// cmp     edx, 1ECDh
			CPatch::PatchUINT32(0x40A6A0 + 2, GetBaseID(FILE_TYPE_COL));	// cmp     ecx, 1ECDh
			CPatch::PatchUINT32(0x40A743 + 1, GetBaseID(FILE_TYPE_COL));	// cmp     eax, 1ECDh
			CPatch::PatchUINT32(0x40A7F1 + 1, GetBaseID(FILE_TYPE_COL));	// cmp     eax, 1ECDh
			CPatch::PatchUINT32(0x40A8D2 + 1, GetBaseID(FILE_TYPE_COL));	// cmp     eax, 1ECDh
			CPatch::PatchUINT32(0x40A974 + 1, GetBaseID(FILE_TYPE_COL));	// cmp     eax, 1ECDh
			CPatch::PatchUINT32(0x40AD7F + 2, GetBaseID(FILE_TYPE_COL));	// cmp     ecx, 1ECDh
			CPatch::PatchUINT32(0x40AE2B + 2, GetBaseID(FILE_TYPE_COL));	// cmp     edx, 1ECDh
			CPatch::PatchUINT32(0x40AEBF + 2, GetBaseID(FILE_TYPE_COL));	// cmp     ecx, 1ECDh
			CPatch::PatchUINT32(0x40B932 + 2, GetBaseID(FILE_TYPE_COL));	// cmp     ebx, 1ECDh
			CPatch::PatchUINT32(0x40C421 + 2, GetBaseID(FILE_TYPE_COL));	// cmp     ebp, 1ECDh
			CPatch::PatchUINT32(0x40C505 + 2, GetBaseID(FILE_TYPE_COL));	// cmp     ebp, 1ECDh
			CPatch::PatchUINT32(0x40C648 + 2, GetBaseID(FILE_TYPE_COL));	// cmp     ebp, 1ECDh
			CPatch::PatchUINT32(0x40C6D8 + 2, GetBaseID(FILE_TYPE_COL));	// cmp     ebp, 1ECDh
			CPatch::PatchUINT32(0x40C6F5 + 2, GetBaseID(FILE_TYPE_COL));	// cmp     ebp, 1ECDh
			CPatch::PatchUINT32(0x40C85F + 2, GetBaseID(FILE_TYPE_COL));	// cmp     ebx, 1ECDh
			CPatch::PatchUINT32(0x40C928 + 2, GetBaseID(FILE_TYPE_COL));	// cmp     ebx, 1ECDh
			CPatch::PatchUINT32(0x40C945 + 2, GetBaseID(FILE_TYPE_COL));	// cmp     ebx, 1ECDh
			CPatch::PatchUINT32(0x40D464 + 2, GetBaseID(FILE_TYPE_COL));	// cmp     edi, 1ECDh
			CPatch::PatchUINT32(0x40D618 + 2, GetBaseID(FILE_TYPE_COL));	// cmp     esi, 1ECDh
			CPatch::PatchUINT32(0x40D728 + 2, GetBaseID(FILE_TYPE_COL));	// cmp     ebx, 1ECDh
			CPatch::PatchUINT32(0x40D740 + 2, GetBaseID(FILE_TYPE_COL));	// cmp     ebx, 1ECDh
			CPatch::PatchUINT32(0x40D83B + 2, GetBaseID(FILE_TYPE_COL));	// cmp     ebx, 1ECDh
			CPatch::PatchUINT32(0x40D851 + 2, GetBaseID(FILE_TYPE_COL));	// cmp     ebx, 1ECDh
			CPatch::PatchUINT32(0x40FD17 + 4, GetBaseID(FILE_TYPE_COL));	// add     [esp+40h+var_14], 1ECDh
			CPatch::PatchUINT32(0x40FFE6 + 2, GetBaseID(FILE_TYPE_COL));	// cmp     ebx, 1ECDh
			CPatch::PatchUINT32(0x629FB8 + 2, GetBaseID(FILE_TYPE_COL));	// lea     eax, [ebx+1ECDh]
			CPatch::PatchUINT32(0x62A024 + 2, GetBaseID(FILE_TYPE_COL));	// lea     eax, [ebx+1ECDh]
			CPatch::PatchUINT32(0x62A1F7 + 2, GetBaseID(FILE_TYPE_COL));	// lea     eax, [ebp+1ECDh]
			CPatch::PatchUINT32(0x62A310 + 2, GetBaseID(FILE_TYPE_COL));	// lea     eax, [ebp+1ECDh]
			CPatch::PatchUINT32(0x62A64A + 2, GetBaseID(FILE_TYPE_COL));	// lea     eax, [ebx+1ECDh]
			CPatch::PatchUINT32(0x62A660 + 2, GetBaseID(FILE_TYPE_COL));	// lea     eax, [ebx+1ECDh]
		}

		// -7885 (negative COL base index)
		{
			CPatch::PatchUINT32(0x40C515 + 2, -GetBaseID(FILE_TYPE_COL));	// lea     eax, [ebp-1ECDh]
			CPatch::PatchUINT32(0x40C531 + 2, -GetBaseID(FILE_TYPE_COL));	// lea     eax, [ebp-1ECDh]
			CPatch::PatchUINT32(0x40C705 + 2, -GetBaseID(FILE_TYPE_COL));	// lea     eax, [ebp-1ECDh]
			CPatch::PatchUINT32(0x40C955 + 2, -GetBaseID(FILE_TYPE_COL));	// lea     eax, [ebx-1ECDh]
			CPatch::PatchUINT32(0x40D750 + 2, -GetBaseID(FILE_TYPE_COL));	// lea     eax, [ebx-1ECDh]
			CPatch::PatchUINT32(0x40D861 + 2, -GetBaseID(FILE_TYPE_COL));	// lea     eax, [ebx-1ECDh]
		}

		// 7916 (IFP base index)
		{
			CPatch::PatchUINT32(0x404969 + 2, GetBaseID(FILE_TYPE_IFP));	// lea     eax, [ecx+1EECh]
			CPatch::PatchUINT32(0x405B84 + 2, GetBaseID(FILE_TYPE_IFP));	// lea     eax, [ebx+1EECh]
			CPatch::PatchUINT32(0x40B94B + 2, GetBaseID(FILE_TYPE_IFP));	// cmp     ebx, 1EECh
			CPatch::PatchUINT32(0x40BA88 + 1, GetBaseID(FILE_TYPE_IFP));	// add     eax, 1EECh
			CPatch::PatchUINT32(0x40BD6A + 2, GetBaseID(FILE_TYPE_IFP));	// lea     edx, [eax+1EECh]
			CPatch::PatchUINT32(0x40BD88 + 1, GetBaseID(FILE_TYPE_IFP));	// add     eax, 1EECh
			CPatch::PatchUINT32(0x40BDA5 + 2, GetBaseID(FILE_TYPE_IFP));	// cmp     ebx, 1EECh
			CPatch::PatchUINT32(0x40C50D + 2, GetBaseID(FILE_TYPE_IFP));	// cmp     ebp, 1EECh
			CPatch::PatchUINT32(0x40C580 + 2, GetBaseID(FILE_TYPE_IFP));	// cmp     ebp, 1EECh
			CPatch::PatchUINT32(0x40C650 + 2, GetBaseID(FILE_TYPE_IFP));	// cmp     ebp, 1EECh
			CPatch::PatchUINT32(0x40C6FD + 2, GetBaseID(FILE_TYPE_IFP));	// cmp     ebp, 1EECh
			CPatch::PatchUINT32(0x40C720 + 2, GetBaseID(FILE_TYPE_IFP));	// cmp     ebp, 1EECh
			CPatch::PatchUINT32(0x40C94D + 2, GetBaseID(FILE_TYPE_IFP));	// cmp     ebx, 1EECh
			CPatch::PatchUINT32(0x40C970 + 2, GetBaseID(FILE_TYPE_IFP));	// cmp     ebx, 1EECh
			CPatch::PatchUINT32(0x40D650 + 2, GetBaseID(FILE_TYPE_IFP));	// cmp     esi, 1EECh
			CPatch::PatchUINT32(0x40D748 + 2, GetBaseID(FILE_TYPE_IFP));	// cmp     ebx, 1EECh
			CPatch::PatchUINT32(0x40D760 + 2, GetBaseID(FILE_TYPE_IFP));	// cmp     ebx, 1EECh
			CPatch::PatchUINT32(0x40D859 + 2, GetBaseID(FILE_TYPE_IFP));	// cmp     ebx, 1EECh
			CPatch::PatchUINT32(0x40D870 + 2, GetBaseID(FILE_TYPE_IFP));	// cmp     ebx, 1EECh
			CPatch::PatchUINT32(0x40E456 + 1, GetBaseID(FILE_TYPE_IFP));	// add     eax, 1EECh
			CPatch::PatchUINT32(0x40FD44 + 4, GetBaseID(FILE_TYPE_IFP));	// add     [esp+44h+var_14], 1EECh
			CPatch::PatchUINT32(0x51562A + 1, GetBaseID(FILE_TYPE_IFP));	// add     eax, 1EECh
			CPatch::PatchUINT32(0x515638 + 2, GetBaseID(FILE_TYPE_IFP));	// lea     eax, [ebx+1EECh]
			CPatch::PatchUINT32(0x515648 + 2, GetBaseID(FILE_TYPE_IFP));	// lea     eax, [ebp+1EECh]
			CPatch::PatchUINT32(0x515658 + 2, GetBaseID(FILE_TYPE_IFP));	// lea     eax, [esi+1EECh]
			CPatch::PatchUINT32(0x51566D + 1, GetBaseID(FILE_TYPE_IFP));	// add     eax, 1EECh
			CPatch::PatchUINT32(0x535DB0 + 1, GetBaseID(FILE_TYPE_IFP));	// add     eax, 1EECh
			CPatch::PatchUINT32(0x535F11 + 1, GetBaseID(FILE_TYPE_IFP));	// add     eax, 1EECh
			CPatch::PatchUINT32(0x631A11 + 1, GetBaseID(FILE_TYPE_IFP));	// add     eax, 1EECh
			CPatch::PatchUINT32(0x631A9E + 1, GetBaseID(FILE_TYPE_IFP));	// add     eax, 1EECh
		}

		// -7916 (negative IFP base index)
		{
			CPatch::PatchUINT32(0x40B953 + 2, -GetBaseID(FILE_TYPE_IFP));	// lea     eax, [ebx-1EECh]
			CPatch::PatchUINT32(0x40C594 + 2, -GetBaseID(FILE_TYPE_IFP));	// lea     eax, [ebp-1EECh]
			CPatch::PatchUINT32(0x40C728 + 2, -GetBaseID(FILE_TYPE_IFP));	// lea     eax, [ebp-1EECh]
			CPatch::PatchUINT32(0x40C978 + 2, -GetBaseID(FILE_TYPE_IFP));	// lea     eax, [ebx-1EECh]
			CPatch::PatchUINT32(0x40D658 + 2, -GetBaseID(FILE_TYPE_IFP));	// lea     eax, [esi-1EECh]
			CPatch::PatchUINT32(0x40D669 + 2, -GetBaseID(FILE_TYPE_IFP));	// lea     eax, [esi-1EECh]
			CPatch::PatchUINT32(0x40D768 + 2, -GetBaseID(FILE_TYPE_IFP));	// lea     eax, [ebx-1EECh]
			CPatch::PatchUINT32(0x40D878 + 2, -GetBaseID(FILE_TYPE_IFP));	// lea     eax, [ebx-1EECh]
		}

	}
	#endif

	// Patch limits itself
	#if TRUE
	{
		// TXD limit
		#if TRUE
		{
			// 1385
			{
				CPatch::PatchUINT32(0x41065D + 2, GetFileIDcurrentLimit(FILE_TYPE_TXD));	// cmp     ebx, 569h
				CPatch::PatchUINT32(0x580FD4 + 2, GetFileIDcurrentLimit(FILE_TYPE_TXD));	// cmp     ebx, 569h
				CPatch::PatchUINT32(0x58102C + 1, GetFileIDcurrentLimit(FILE_TYPE_TXD));	// push    569h
				CPatch::PatchUINT32(0x61D98B + 1, GetFileIDcurrentLimit(FILE_TYPE_TXD));	// push    569h
				CPatch::PatchUINT32(0x61D9E5 + 1, GetFileIDcurrentLimit(FILE_TYPE_TXD));	// push    569h
				CPatch::PatchUINT32(0x61DAA5 + 1, GetFileIDcurrentLimit(FILE_TYPE_TXD));	// push    569h
				CPatch::PatchUINT32(0x61DBA5 + 1, GetFileIDcurrentLimit(FILE_TYPE_TXD));	// push    569h
				CPatch::PatchUINT32(0x61DCFA + 2, GetFileIDcurrentLimit(FILE_TYPE_TXD));	// cmp     ebx, 569h
				CPatch::PatchUINT32(0x61DD39 + 1, GetFileIDcurrentLimit(FILE_TYPE_TXD));	// push    569h
			}
		}
		#endif

		// COL limit
		#if TRUE
		{
			tVariableValue value;
			value.Integer = GetFileIDcurrentLimit(FILE_TYPE_COL);

			CCodeMover::SetVariable("MAX_NUMBER_OF_COL_FILES", value);

			// 31 - default limit of COL files.
			{
				// Count of active addresses: 9
				CCodeMover::FixOnAddress(0x48DE71, 5, "\x01\x02\x81\xfb\x05\x04MAX_NUMBER_OF_COL_FILES\x00\x01\x02\x0F\x8C\x03\x14\xde\x48\x00\x00", 0x48DE76);	// line 1, 0x48D950 - sub_48D950 : cmp     ebx, 31
				CCodeMover::FixOnAddress(0x629FCC, 5, "\x01\x02\x81\xfb\x05\x04MAX_NUMBER_OF_COL_FILES\x00\x01\x02\x0F\x8C\x03\x90\x9f\x62\x00\x00", 0x629FD1);	// line 2, 0x629F80 - sub_629F80 : cmp     ebx, 1Fh
				CCodeMover::FixOnAddress(0x62A038, 5, "\x01\x02\x81\xfb\x05\x04MAX_NUMBER_OF_COL_FILES\x00\x01\x02\x0F\x8C\x03\xf3\x9f\x62\x00\x00", 0x62A03D);	// line 3, 0x629FE0 - sub_629FE0 : cmp     ebx, 1Fh
				CCodeMover::FixOnAddress(0x62A12A, 9, "\x01\x02\x81\xfb\x05\x04MAX_NUMBER_OF_COL_FILES\x00\x01\x02\x0F\x8C\x03\x63\xa0\x62\x00\x00", 0x62A133);	// line 4, 0x62A050 - sub_62A050 : cmp     ebx, 1Fh
				CCodeMover::FixOnAddress(0x62A20D, 9, "\x01\x02\x81\xfd\x05\x04MAX_NUMBER_OF_COL_FILES\x00\x01\x02\x0F\x8C\x03\x51\xa1\x62\x00\x00", 0x62A216);	// line 5, 0x62A140 - sub_62A140 : cmp     ebp, 1Fh
				CCodeMover::FixOnAddress(0x62A35A, 9, "\x01\x02\x81\xfd\x05\x04MAX_NUMBER_OF_COL_FILES\x00\x01\x02\x0F\x8C\x03\x60\xa2\x62\x00\x00", 0x62A363);	// line 6, 0x62A230 - sub_62A230 : cmp     ebp, 1Fh
				CCodeMover::FixOnAddress(0x62A673, 9, "\x01\x02\x81\xfb\x05\x04MAX_NUMBER_OF_COL_FILES\x00\x01\x02\x0F\x8C\x03\xa5\xa3\x62\x00\x00", 0x62A67C);	// line 7, 0x62A380 - sub_62A380 : cmp     ebx, 1Fh
				CCodeMover::FixOnAddress(0x62A9F8, 9, "\x01\x02\x81\xfb\x05\x04MAX_NUMBER_OF_COL_FILES\x00\x01\x02\x0F\x8C\x03\x47\xa9\x62\x00\x00", 0x62AA01);	// line 8, 0x62A940 - sub_62A940 : cmp     ebx, 1Fh
				CCodeMover::FixOnAddress(0x62AA55, 7, "\x01\x01\x68\x05\x04MAX_NUMBER_OF_COL_FILES\x00\x01\x01\xE8\x03\xd0\xab\x62\x00\x00", 0x62AA5C);	// line 9, 0x62AA30 - sub_62AA30 : push    1Fh
			}

			CCodeMover::DeleteVariable("MAX_NUMBER_OF_COL_FILES");
		}
		#endif

		// IFP limit
		#if TRUE
		{
			CPatch::PatchPointer(0x404937 + 3, &CAnimManager__ms_aAnimBlocks.gta_vc -> usageCount);	// movsx   eax, ds:_ZN12CAnimManager14ms_aAnimBlocksE.usageCount[eax]
			CPatch::PatchPointer(0x404947 + 3, &CAnimManager__ms_aAnimBlocks.gta_vc -> usageCount);	// dec     ds:_ZN12CAnimManager14ms_aAnimBlocksE.usageCount[eax]
			CPatch::PatchPointer(0x404959 + 1, CAnimManager__ms_aAnimBlocks.gta_vc);	// add     eax, offset _ZN12CAnimManager14ms_aAnimBlocksE; CAnimManager::ms_aAnimBlocks
			CPatch::PatchPointer(0x404987 + 3, &CAnimManager__ms_aAnimBlocks.gta_vc -> usageCount);	// inc     ds:_ZN12CAnimManager14ms_aAnimBlocksE.usageCount[eax]
			CPatch::PatchPointer(0x404999 + 2, CAnimManager__ms_aAnimBlocks.gta_vc);	// add     ebx, offset _ZN12CAnimManager14ms_aAnimBlocksE; CAnimManager::ms_aAnimBlocks
			CPatch::PatchPointer(0x404A12 + 2, &CAnimManager__ms_aAnimBlocks.gta_vc -> field_18);	// mov     eax, ds:_ZN12CAnimManager14ms_aAnimBlocksE.field_18[edx]
			CPatch::PatchPointer(0x404A22 + 2, &CAnimManager__ms_aAnimBlocks.gta_vc -> field_18);	// add     ecx, ds:_ZN12CAnimManager14ms_aAnimBlocksE.field_18[eax]
			CPatch::PatchPointer(0x404A3F + 2, &CAnimManager__ms_aAnimBlocks.gta_vc -> field_1C);	// cmp     ebx, ds:_ZN12CAnimManager14ms_aAnimBlocksE.field_1C[eax]
			CPatch::PatchPointer(0x404A47 + 2, &CAnimManager__ms_aAnimBlocks.gta_vc -> loaded);	// mov     ds:_ZN12CAnimManager14ms_aAnimBlocksE.loaded[eax], 0
			CPatch::PatchPointer(0x404AE6 + 1, CAnimManager__ms_aAnimBlocks.gta_vc);	// mov     ebp, offset _ZN12CAnimManager14ms_aAnimBlocksE; CAnimManager::ms_aAnimBlocks
			CPatch::PatchPointer(0x404B00 + 2, CAnimManager__ms_aAnimBlocks.gta_vc);	// add     esi, offset _ZN12CAnimManager14ms_aAnimBlocksE; CAnimManager::ms_aAnimBlocks
			CPatch::PatchPointer(0x404B36 + 4, CAnimManager__ms_aAnimBlocks.gta_vc);	// add     [esp+1F0h+var_1C8], offset _ZN12CAnimManager14ms_aAnimBlocksE; CAnimManager::ms_aAnimBlocks
			CPatch::PatchPointer(0x4053B2 + 1, CAnimManager__ms_aAnimBlocks.gta_vc);	// mov     ebx, offset _ZN12CAnimManager14ms_aAnimBlocksE; CAnimManager::ms_aAnimBlocks
			CPatch::PatchPointer(0x4053C7 + 2, CAnimManager__ms_aAnimBlocks.gta_vc);	// add     esi, offset _ZN12CAnimManager14ms_aAnimBlocksE; CAnimManager::ms_aAnimBlocks
			CPatch::PatchPointer(0x4053F3 + 2, CAnimManager__ms_aAnimBlocks.gta_vc);	// add     ebx, offset _ZN12CAnimManager14ms_aAnimBlocksE; CAnimManager::ms_aAnimBlocks
			CPatch::PatchPointer(0x40540C + 2, CAnimManager__ms_aAnimBlocks.gta_vc);	// sub     ebx, offset _ZN12CAnimManager14ms_aAnimBlocksE; CAnimManager::ms_aAnimBlocks
			CPatch::PatchPointer(0x405465 + 1, CAnimManager__ms_aAnimBlocks.gta_vc);	// mov     ebp, offset _ZN12CAnimManager14ms_aAnimBlocksE; CAnimManager::ms_aAnimBlocks
			CPatch::PatchPointer(0x405480 + 2, CAnimManager__ms_aAnimBlocks.gta_vc);	// add     edi, offset _ZN12CAnimManager14ms_aAnimBlocksE; CAnimManager::ms_aAnimBlocks
			CPatch::PatchPointer(0x4059F2 + 1, CAnimManager__ms_aAnimBlocks.gta_vc);	// mov     esi, offset _ZN12CAnimManager14ms_aAnimBlocksE; CAnimManager::ms_aAnimBlocks
			CPatch::PatchPointer(0x405A07 + 2, CAnimManager__ms_aAnimBlocks.gta_vc);	// add     ebx, offset _ZN12CAnimManager14ms_aAnimBlocksE; CAnimManager::ms_aAnimBlocks
			CPatch::PatchPointer(0x405A30 + 2, CAnimManager__ms_aAnimBlocks.gta_vc);	// sub     ebx, offset _ZN12CAnimManager14ms_aAnimBlocksE; CAnimManager::ms_aAnimBlocks
			CPatch::PatchPointer(0x405A62 + 1, CAnimManager__ms_aAnimBlocks.gta_vc);	// mov     esi, offset _ZN12CAnimManager14ms_aAnimBlocksE; CAnimManager::ms_aAnimBlocks
			CPatch::PatchPointer(0x405A78 + 2, CAnimManager__ms_aAnimBlocks.gta_vc);	// add     ebx, offset _ZN12CAnimManager14ms_aAnimBlocksE; CAnimManager::ms_aAnimBlocks
			CPatch::PatchPointer(0x405D8D + 1, CAnimManager__ms_aAnimBlocks.gta_vc);	// push    offset _ZN12CAnimManager14ms_aAnimBlocksE; CAnimManager::ms_aAnimBlocks
			CPatch::PatchPointer(0x40C2B2 + 2, &CAnimManager__ms_aAnimBlocks.gta_vc -> loaded);	// cmp     ds:_ZN12CAnimManager14ms_aAnimBlocksE.loaded[eax], 0
			CPatch::PatchPointer(0x40C731 + 1, CAnimManager__ms_aAnimBlocks.gta_vc);	// add     eax, offset _ZN12CAnimManager14ms_aAnimBlocksE; CAnimManager::ms_aAnimBlocks
			CPatch::PatchPointer(0x40C981 + 1, CAnimManager__ms_aAnimBlocks.gta_vc);	// add     eax, offset _ZN12CAnimManager14ms_aAnimBlocksE; CAnimManager::ms_aAnimBlocks
			CPatch::PatchPointer(0x535D90 + 2, CAnimManager__ms_aAnimBlocks.gta_vc);	// add     edi, offset _ZN12CAnimManager14ms_aAnimBlocksE; CAnimManager::ms_aAnimBlocks
		}
		#endif
	}
	#endif

	// finish
	CGenericLogStorage::WriteLineSeparator();
}
#endif

// GTA San Andreas
namespace Game_GTASA
{
	// 32-bit indices
	namespace FileInfoExtension
	{
	#define ENABLE_FILE_INFO_EXTENSION

		// Returns pointer to extended value
		void* realGetPointerToExtendedValue(uint32_t InfoForModelOffset, uint32_t memberOffset)
		{		
			#ifdef ENABLE_FILE_INFO_EXTENSION
			uint32_t fileID = InfoForModelOffset / sizeof(CStreamingInfo);

			CStreamingInfoExtension* pExtension = &g_fileIDlimits.CStreaming__ms_aInfoForModelExtension[fileID];

			switch (memberOffset)
			{
				case 0:
					return &pExtension->exUsNext;
				case 2:
					return &pExtension->exUsPrev;
				case 4:
					return &pExtension->exNextOnCd;
			}
			#else
			CStreamingInfo* pFileInfo = (CStreamingInfo*)((char*)(g_fileIDlimits.CStreaming__ms_aInfoForModel.gta_sa) + InfoForModelOffset);

			switch (memberOffset)
			{
			case 0:
				return &pFileInfo->usNext;
			case 2:
				return &pFileInfo->usPrev;
			case 4:
				return &pFileInfo->usNextOnCd;
			}
			#endif
			return NULL;
		}

		int32_t __cdecl realGetExtendedMemberByStreamingInfoMemOffset(uint32_t InfoForModelOffset, uint32_t memberOffset)
		{
#ifdef ENABLE_FILE_INFO_EXTENSION
			return *(int32_t*)realGetPointerToExtendedValue(InfoForModelOffset, memberOffset);
#else
			// DWORD result =  *((uint16_t*)((char*)FileIDlimit::CStreaming__ms_aInfoForModel + InfoForModelOffset + memberOffset));

			uint32_t result = *(uint16_t*)realGetPointerToExtendedValue(InfoForModelOffset, memberOffset);

			if (result == 0xFFFF)
				result = UNDEFINED_FILE_ID_INT32;

			return result;
#endif		
			

		}

		void __cdecl realSetExtendedMemberByStreamingInfoMemOffset(uint32_t InfoForModelOffset, uint32_t memberOffset, uint32_t newValue)
		{
#ifdef ENABLE_FILE_INFO_EXTENSION
			*(int32_t*)realGetPointerToExtendedValue(InfoForModelOffset, memberOffset) = newValue;
#else
			// *((uint16_t*)((char*)FileIDlimit::CStreaming__ms_aInfoForModel + InfoForModelOffset + memberOffset)) = (uint16_t)newValue;
			*(uint16_t*)realGetPointerToExtendedValue(InfoForModelOffset, memberOffset) = newValue;
#endif

			
		}

	#ifdef IS_PLATFORM_WIN_X86
		NAKED void GetExtendedMemberByStreamingInfoPtr(CStreamingInfo* pFileInfo, DWORD memberOffset, DWORD returnValue)
		{
			// stack:
			// 0x0: return address
			// 0x4: CStreamingInfo* pFileInfo
			// 0x8: DWORD memberOffset
			// 0xC: DWORD returnValue

			__asm
			{
				pushfd;
				pushad;
				
				mov eax, [esp + 0x24 + 0x8];
				push eax;	// memberOffset

				mov eax, [esp + 0x28 + 0x4];	// pFileInfo
				sub eax, ASM_ACCESS_LIMIT_VAR_32_BIT(g_fileIDlimits, FileIDlimit, CStreaming__ms_aInfoForModel);
				push eax;	// InfoForModelOffset

				call realGetExtendedMemberByStreamingInfoMemOffset;
				add esp, 0x8;

				mov[esp + 0x24 + 0xC], eax;

				popad;
				popfd;
				retn;
			}

		}

		NAKED void GetExtendedMemberByStreamingInfoMemOffset(DWORD InfoForModelOffset, DWORD memberOffset, DWORD returnValue)
		{
			// stack:
			// 0x0: return address
			// 0x4: DWORD InfoForModelOffset
			// 0x8: DWORD memberOffset
			// 0xC: DWORD returnValue

			__asm
			{
				pushfd;
				pushad;

				mov eax, [esp + 0x24 + 0x8];
				push eax;	// memberOffset

				mov eax, [esp + 0x28 + 0x4];	// pFileInfo
				push eax;	// InfoForModelOffset

				call realGetExtendedMemberByStreamingInfoMemOffset;
				add esp, 0x8;

				mov[esp + 0x24 + 0xC], eax;

				popad;
				popfd;
				retn;
			}
		}		
		
		NAKED void SetExtendedMemberByStreamingInfoPtr(CStreamingInfo* pFileInfo, DWORD memberOffset, DWORD newValue)
		{
			// stack:
			// 0x0: return address
			// 0x4: CStreamingInfo* pFileInfo
			// 0x8: DWORD memberOffset
			// 0xC: DWORD newValue

			__asm
			{
				pushfd;
				pushad;

				mov eax, [esp + 0x24 + 0xC];	// newValue
				push eax;

				mov eax, [esp + 0x28 + 0x8];
				push eax;	// memberOffset

				mov eax, [esp + 0x2C + 0x4];	// pFileInfo
				sub eax, ASM_ACCESS_LIMIT_VAR_32_BIT(g_fileIDlimits, FileIDlimit, CStreaming__ms_aInfoForModel);
				push eax;	// InfoForModelOffset

				call realSetExtendedMemberByStreamingInfoMemOffset;
				add esp, 0xC;

				popad;
				popfd;
				retn;
			}
		}

		NAKED void SetExtendedMemberByStreamingInfoMemOffset(DWORD InfoForModelOffset, DWORD memberOffset, DWORD returnValue)
		{
			// stack:
			// 0x0: return address
			// 0x4: DWORD InfoForModelOffset
			// 0x8: DWORD memberOffset
			// 0xC: DWORD newValue

			__asm
			{
				pushfd;
				pushad;

				mov eax, [esp + 0x24 + 0xC];	// newValue
				push eax;

				mov eax, [esp + 0x28 + 0x8];
				push eax;	// memberOffset

				mov eax, [esp + 0x2C + 0x4];
				push eax;	// InfoForModelOffset

				call realSetExtendedMemberByStreamingInfoMemOffset;
				add esp, 0xC;

				popad;
				popfd;
				retn;
			}
		}
	#endif
	}

#ifdef IS_PLATFORM_WIN_X86
	// 32-bit indices
	namespace FileInfoExtension
	{
		// Original code:	mov ax, [ecx]
		NAKED void patch_407520()
		{
			__asm
			{
				push 0;	// space for result
				push 0;	// member offset
				push ecx;	// CStreamingInfo*
				call GetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;

				mov eax, [esp - 4];
				retn;
			}
		}

		// Original code:	cmp ax, 0FFFFh
		NAKED void patch_407523()
		{
			__asm
			{
				cmp eax, UNDEFINED_FILE_ID_INT32;
				retn;
			}
		}

		// Original code:	mov ax, [ecx+2]
		NAKED void patch_407540()
		{
			__asm
			{
				push 0;	// space for result
				push 2;	// member offset
				push ecx;	// CStreamingInfo*
				call GetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;

				mov eax, [esp - 4];
				retn;
			}
		}

		// Original code:	cmp ax, 0FFFFh
		NAKED void patch_407544()
		{
			__asm
			{
				cmp eax, UNDEFINED_FILE_ID_INT32;
				retn;
			}
		}

		// Original code:	cmp word ptr [ecx], 0FFFFh
		NAKED void patch_407562()
		{
			__asm
			{
				push 0;	// space for result
				push 0;	// member offset
				push ecx;	// CStreamingInfo*
				call GetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;

				cmp dword ptr[esp - 4], UNDEFINED_FILE_ID_INT32;
				retn;
			}
		}

		// Original code:	mov ax, [eax]
		NAKED void patch_407AD5()
		{
			__asm
			{
				push 0;	// space for result
				push 0;	// member offset
				push eax;	// CStreamingInfo*
				call GetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;

				mov eax, [esp - 4];
				retn;
			}
		}

		// Original code:	cmp ax, 0FFFFh
		NAKED void patch_407AD8()
		{
			__asm
			{
				cmp eax, UNDEFINED_FILE_ID_INT32;
				retn;
			}
		}

		// Original code:	mov si, [esi]
		NAKED void patch_407B35()
		{
			__asm
			{
				push 0;	// space for result
				push 0;	// member offset
				push esi;	// CStreamingInfo*
				call GetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;

				mov esi, [esp - 4];
				retn;
			}
		}

		// Original code:	cmp si, 0FFFFh
		NAKED void patch_407B38()
		{
			__asm
			{
				cmp esi, UNDEFINED_FILE_ID_INT32;
				retn;
			}
		}

		// Original code:	cmp word ptr [esi], 0FFFFh
		NAKED void patch_40884E()
		{
			__asm
			{
				push 0;	// space for result
				push 0;	// member offset
				push esi;	// CStreamingInfo*
				call GetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;

				cmp dword ptr[esp - 4], UNDEFINED_FILE_ID_INT32;
				retn;
			}
		}

		// Original code:	movsx eax, word ptr [esi]
		NAKED void patch_40885E()
		{
			__asm
			{
				push 0;	// space for result
				push 0;	// member offset
				push esi;	// CStreamingInfo*
				call GetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;

				mov eax, [esp - 4];
				retn;
			}
		}

		// Original code:	mov cx, [esi+2]
		NAKED void patch_408861()
		{
			__asm
			{
				push 0;	// space for result
				push 2;	// member offset
				push esi;	// CStreamingInfo*
				call GetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;

				mov ecx, [esp - 4];
				retn;
			}
		}

		// Original code:	mov [edx+eax*4+2], cx
		NAKED void patch_40886E()
		{
			__asm
			{
				push ebx;
				lea ebx, [edx + eax * 4];

				push ecx;	// new value
				push 2;	// member offset
				push ebx;	// CStreamingInfo*
				call SetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;

				pop ebx;
				retn;
			}
		}

		// Original code:	movsx eax, word ptr [esi+2]
		NAKED void patch_408873()
		{
			__asm
			{
				push 0;	// space for result
				push 2;	// member offset
				push esi;	// CStreamingInfo*
				call GetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;

				mov eax, [esp - 4];
				retn;
			}
		}

		// Original code:	mov cx, [esi]
		NAKED void patch_408877()
		{
			__asm
			{
				push 0;	// space for result
				push 0;	// member offset
				push esi;	// CStreamingInfo*
				call GetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;

				mov ecx, [esp - 4];
				retn;
			}
		}

		// Original code:	mov [edx+eax*4], cx
		NAKED void patch_408883()
		{
			__asm
			{
				push ebx;
				lea ebx, [edx + eax * 4];

				push ecx;	// new value
				push 0;	// member offset
				push ebx;	// CStreamingInfo*
				call SetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;

				pop ebx;
				retn;
			}
		}

		// Original code:	mov word ptr [esi], 0FFFFh
		NAKED void patch_408887()
		{
			__asm
			{
				push UNDEFINED_FILE_ID_INT32;	// new value
				push 0;	// member offset
				push esi;	// CStreamingInfo*
				call SetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;
				retn;
			}
		}

		// Original code:	mov word ptr [esi+2], 0FFFFh
		NAKED void patch_40888C()
		{
			__asm
			{
				push UNDEFINED_FILE_ID_INT32;	// new value
				push 2;	// member offset
				push esi;	// CStreamingInfo*
				call SetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;
				retn;
			}
		}

		// Original code:	cmp word ptr ds:ms_aInfoForModel.usNext[edi], bp
		NAKED void patch_408ADA()
		{
			__asm
			{
				push 0;	// space for result
				push 0;	// member offset
				push edi;	// CStreaming::ms_aInfoForModel offset
				call GetExtendedMemberByStreamingInfoMemOffset;
				add esp, 0xC;

				cmp dword ptr[esp - 4], ebp;
				retn;
			}
		}

		// Original code:	movsx eax, word ptr ds:ms_aInfoForModel.usNext[edi]
		NAKED void patch_408B22()
		{
			__asm
			{
				push 0;	// space for result
				push 0;	// member offset
				push edi;	// CStreaming::ms_aInfoForModel offset
				call GetExtendedMemberByStreamingInfoMemOffset;
				add esp, 0xC;

				mov eax, [esp - 4];
				retn;
			}
		}

		// Original code:	mov ax, ds:ms_aInfoForModel.usPrev[edi]
		NAKED void patch_408B32()
		{
			__asm
			{
				push 0;	// space for result
				push 2;	// member offset
				push edi;	// CStreaming::ms_aInfoForModel offset
				call GetExtendedMemberByStreamingInfoMemOffset;
				add esp, 0xC;

				mov eax, [esp - 4];
				retn;
			}
		}

		// Original code:	mov [ecx+edx*4+2], ax
		NAKED void patch_408B39()
		{
			__asm
			{
				push ebx;
				lea ebx, [ecx + edx * 4];

				push eax;	// new value
				push 2;	// member offset
				push ebx;	// CStreamingInfo*
				call SetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;

				pop ebx;
				retn;
			}
		}

		// Original code:	movsx eax, ds:ms_aInfoForModel.usPrev[edi]
		NAKED void patch_408B3E()
		{
			__asm
			{
				push 0;	// space for result
				push 2;	// member offset
				push edi;	// CStreaming::ms_aInfoForModel offset
				call GetExtendedMemberByStreamingInfoMemOffset;
				add esp, 0xC;

				mov eax, [esp - 4];
				retn;
			}
		}

		// Original code:	mov ax, word ptr ds:ms_aInfoForModel.usNext[edi]
		NAKED void patch_408B4E()
		{
			__asm
			{
				push 0;	// space for result
				push 0;	// member offset
				push edi;	// CStreaming::ms_aInfoForModel offset
				call GetExtendedMemberByStreamingInfoMemOffset;
				add esp, 0xC;

				mov eax, [esp - 4];
				retn;
			}
		}

		// Original code:	mov [ecx+edx*4], ax
		NAKED void patch_408B55()
		{
			__asm
			{
				push ebx;
				lea ebx, [ecx + edx * 4];

				push eax;	// new value
				push 0;	// member offset
				push ebx;	// CStreamingInfo*
				call SetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;

				pop ebx;
				retn;
			}
		}

		// Original code:	mov word ptr ds:ms_aInfoForModel.usNext[edi], bp
		NAKED void patch_408B59()
		{
			__asm
			{
				push ebp;	// new value
				push 0;	// member offset
				push edi;	// CStreaming::ms_aInfoForModel offset
				call SetExtendedMemberByStreamingInfoMemOffset;
				add esp, 0xC;
				retn;
			}
		}

		// Original code:	mov ds:ms_aInfoForModel.usPrev[edi], bp
		NAKED void patch_408B60()
		{
			__asm
			{
				push ebp;	// new value
				push 2;	// member offset
				push edi;	// CStreaming::ms_aInfoForModel offset
				call SetExtendedMemberByStreamingInfoMemOffset;
				add esp, 0xC;
				retn;
			}
		}

		// Original code:	mov ax, [eax+CStreamingInfo.usNext]
		NAKED void patch_408E28()
		{
			__asm
			{
				push 0;	// space for result
				push 0;	// member offset
				push eax;	// CStreamingInfo*
				call GetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;

				mov eax, [esp - 4];
				retn;
			}
		}

		// Original code:	cmp ax, cx
		NAKED void patch_408E2E()
		{
			__asm
			{
				cmp eax, ecx;
				retn;
			}
		}

		// Original code:	mov ax, [esi]
		NAKED void patch_408E72()
		{
			__asm
			{
				push 0;	// space for result
				push 0;	// member offset
				push esi;	// CStreamingInfo*
				call GetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;

				mov eax, [esp - 4];
				retn;
			}
		}

		// Original code:	cmp ax, 0FFFFh
		NAKED void patch_408E7F()
		{
			__asm
			{
				cmp eax, UNDEFINED_FILE_ID_INT32;
				retn;
			}
		}

		// Original code:	cmp word ptr ds:ms_aInfoForModel.usNext[eax], 0FFFFh
		NAKED void patch_4098C8()
		{
			__asm
			{
				push 0;	// space for result
				push 0;	// member offset
				push eax;	// CStreaming::ms_aInfoForModel offset
				call GetExtendedMemberByStreamingInfoMemOffset;
				add esp, 0xC;

				cmp dword ptr[esp - 4], UNDEFINED_FILE_ID_INT32;
				retn;
			}
		}

		// Original code:	cmp ds:ms_aInfoForModel.usNext[ecx*4], 0FFFFh
		NAKED void patch_409978()
		{
			__asm
			{
				push 0;	// space for result
				push 0;	// member offset
				push ecx;	// CStreaming::ms_aInfoForModel offset
				shl [esp], 2;
				call GetExtendedMemberByStreamingInfoMemOffset;
				add esp, 0xC;

				cmp dword ptr[esp - 4], UNDEFINED_FILE_ID_INT32;
				retn;
			}
		}

		// Original code:	cmp ds:ms_aInfoForModel.usNext[ecx*4], 0FFFFh
		NAKED void patch_409A39()
		{
			__asm
			{
				push 0;	// space for result
				push 0;	// member offset
				push ecx;	// CStreaming::ms_aInfoForModel offset
				shl [esp], 2;
				call GetExtendedMemberByStreamingInfoMemOffset;
				add esp, 0xC;

				cmp dword ptr[esp - 4], UNDEFINED_FILE_ID_INT32;
				retn;
			}
		}

		// Original code:	cmp word ptr [ecx], 0FFFFh
		NAKED void patch_409C41()
		{
			__asm
			{
				push 0;	// space for result
				push 0;	// member offset
				push ecx;	// CStreamingInfo*
				call GetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;

				cmp dword ptr[esp - 4], UNDEFINED_FILE_ID_INT32;
				retn;
			}
		}

		// Original code:	cmp word ptr [ecx], 0FFFFh
		NAKED void patch_409CCC()
		{
			__asm
			{
				push 0;	// space for result
				push 0;	// member offset
				push ecx;	// CStreamingInfo*
				call GetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;

				cmp dword ptr[esp - 4], UNDEFINED_FILE_ID_INT32;
				retn;
			}
		}

		// Original code:	mov ds:ms_aInfoForModel.usNextOnCd[eax], 0FFFFh
		NAKED void patch_409FCA()
		{
			__asm
			{
				push UNDEFINED_FILE_ID_INT32;	// new value
				push 4;	// member offset
				push eax;	// CStreaming::ms_aInfoForModel offset
				call SetExtendedMemberByStreamingInfoMemOffset;
				add esp, 0xC;
				retn;
			}
		}

		// Original code:	cmp word ptr [ecx], 0FFFFh
		NAKED void patch_40A862()
		{
			__asm
			{
				push 0;	// space for result
				push 0;	// member offset
				push ecx;	// CStreamingInfo*
				call GetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;

				cmp dword ptr[esp - 4], UNDEFINED_FILE_ID_INT32;
				retn;
			}
		}

		// Original code:	mov dx, [edi]
		NAKED void patch_40A88B()
		{
			__asm
			{
				push 0;	// space for result
				push 0;	// member offset
				push edi;	// CStreamingInfo*
				call GetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;

				mov edx, [esp - 4];
				retn;
			}
		}

		// Original code:	mov [ecx], dx
		NAKED void patch_40A88E()
		{
			__asm
			{
				push edx;	// new value
				push 0;	// member offset
				push ecx;	// CStreamingInfo*
				call SetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;
				retn;
			}
		}

		// Original code:	mov [ecx+2], dx
		NAKED void patch_40A8AA()
		{
			__asm
			{
				push edx;	// new value
				push 2;	// member offset
				push ecx;	// CStreamingInfo*
				call SetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;
				retn;
			}
		}

		// Original code:	mov [edi], si
		NAKED void patch_40A8AE()
		{
			__asm
			{
				push esi;	// new value
				push 0;	// member offset
				push edi;	// CStreamingInfo*
				call SetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;
				retn;
			}
		}

		// Original code:	movsx eax, word ptr [ecx]
		NAKED void patch_40A8B1()
		{
			__asm
			{
				push 0;	// space for result
				push 0;	// member offset
				push ecx;	// CStreamingInfo*
				call GetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;

				mov eax, [esp - 4];
				retn;
			}
		}

		// Original code:	mov [edx+ecx*4+2], si
		NAKED void patch_40A8C1()
		{
			__asm
			{
				push ebx;
				lea ebx, [edx + ecx * 4];

				push esi;	// new value
				push 2;	// member offset
				push ebx;	// CStreamingInfo*
				call SetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;

				pop ebx;
				retn;
			}
		}

		// Original code:	cmp word ptr [ecx], 0FFFFh
		NAKED void patch_40A926()
		{
			__asm
			{
				push 0;	// space for result
				push 0;	// member offset
				push ecx;	// CStreamingInfo*
				call GetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;

				cmp dword ptr[esp - 4], UNDEFINED_FILE_ID_INT32;
				retn;
			}
		}

		// Original code:	mov dx, [edi]
		NAKED void patch_40A953()
		{
			__asm
			{
				push 0;	// space for result
				push 0;	// member offset
				push edi;	// CStreamingInfo*
				call GetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;

				mov edx, [esp - 4];
				retn;
			}
		}

		// Original code:	mov [ecx], dx
		NAKED void patch_40A956()
		{
			__asm
			{
				push edx;	// new value
				push 0;	// member offset
				push ecx;	// CStreamingInfo*
				call SetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;
				retn;
			}
		}

		// Original code:	mov [ecx+2], dx
		NAKED void patch_40A972()
		{
			__asm
			{
				push edx;	// new value
				push 2;	// member offset
				push ecx;	// CStreamingInfo*
				call SetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;
				retn;
			}
		}

		// Original code:	mov [edi], si
		NAKED void patch_40A976()
		{
			__asm
			{
				push esi;	// new value
				push 0;	// member offset
				push edi;	// CStreamingInfo*
				call SetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;
				retn;
			}
		}

		// Original code:	movsx eax, word ptr [ecx]
		NAKED void patch_40A979()
		{
			__asm
			{
				push 0;	// space for result
				push 0;	// member offset
				push ecx;	// CStreamingInfo*
				call GetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;

				mov eax, [esp - 4];
				retn;
			}
		}

		// Original code:	mov [edx+ecx*4+2], si
		NAKED void patch_40A989()
		{

			__asm
			{
				push ebx;
				lea ebx, [edx + ecx * 4];

				push esi;	// new value
				push 2;	// member offset
				push ebx;	// CStreamingInfo*
				call SetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;

				pop ebx;
				retn;
			}
		}

		// Original code:	cmp word ptr [ecx], 0FFFFh
		NAKED void patch_40ABEB()
		{
			__asm
			{
				push 0;	// space for result
				push 0;	// member offset
				push ecx;	// CStreamingInfo*
				call GetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;

				cmp dword ptr[esp - 4], UNDEFINED_FILE_ID_INT32;
				retn;
			}
		}

		// Original code:	cmp word ptr [esi], 0FFFFh
		NAKED void patch_40AC59()
		{
			__asm
			{
				push 0;	// space for result
				push 0;	// member offset
				push esi;	// CStreamingInfo*
				call GetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;

				cmp dword ptr[esp - 4], UNDEFINED_FILE_ID_INT32;
				retn;
			}
		}

		// Original code:	cmp word ptr [ecx], 0FFFFh
		NAKED void patch_40AE1F()
		{
			__asm
			{
				push 0;	// space for result
				push 0;	// member offset
				push ecx;	// CStreamingInfo*
				call GetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;

				cmp dword ptr[esp - 4], UNDEFINED_FILE_ID_INT32;
				retn;
			}
		}

		// Original code:	mov dx, [edi]
		NAKED void patch_40AE48()
		{
			__asm
			{
				push 0;	// space for result
				push 0;	// member offset
				push edi;	// CStreamingInfo*
				call GetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;

				mov edx, [esp - 4];
				retn;
			}
		}

		// Original code:	mov [ecx], dx
		NAKED void patch_40AE4B()
		{
			__asm
			{
				push edx;	// new value
				push 0;	// member offset
				push ecx;	// CStreamingInfo*
				call SetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;
				retn;
			}
		}

		// Original code:	mov [ecx+2], dx
		NAKED void patch_40AE63()
		{
			__asm
			{
				push edx;	// new value
				push 2;	// member offset
				push ecx;	// CStreamingInfo*
				call SetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;
				retn;
			}
		}

		// Original code:	mov [edi], si
		NAKED void patch_40AE67()
		{
			__asm
			{
				push esi;	// new value
				push 0;	// member offset
				push edi;	// CStreamingInfo*
				call SetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;
				retn;
			}
		}

		// Original code:	movsx eax, word ptr [ecx]
		NAKED void patch_40AE6A()
		{
			__asm
			{
				push 0;	// space for result
				push 0;	// member offset
				push ecx;	// CStreamingInfo*
				call GetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;

				mov eax, [esp - 4];
				retn;
			}
		}

		// Original code:	mov [edx+ecx*4+2], si
		NAKED void patch_40AE76()
		{

			__asm
			{
				push ebx;
				lea ebx, [edx + ecx * 4];

				push esi;	// new value
				push 2;	// member offset
				push ebx;	// CStreamingInfo*
				call SetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;

				pop ebx;
				retn;
			}
		}

		// Original code:	cmp word ptr [ecx], 0FFFFh
		NAKED void patch_40AEC8()
		{
			__asm
			{
				push 0;	// space for result
				push 0;	// member offset
				push ecx;	// CStreamingInfo*
				call GetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;

				cmp dword ptr[esp - 4], UNDEFINED_FILE_ID_INT32;
				retn;
			}
		}

		// Original code:	mov dx, [edi]
		NAKED void patch_40AEF1()
		{
			__asm
			{
				push 0;	// space for result
				push 0;	// member offset
				push edi;	// CStreamingInfo*
				call GetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;

				mov edx, [esp - 4];
				retn;
			}
		}

		// Original code:	mov [ecx], dx
		NAKED void patch_40AEF4()
		{
			__asm
			{
				push edx;	// new value
				push 0;	// member offset
				push ecx;	// CStreamingInfo*
				call SetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;
				retn;
			}
		}

		// Original code:	mov [ecx+2], dx
		NAKED void patch_40AF0C()
		{
			__asm
			{
				push edx;	// new value
				push 2;	// member offset
				push ecx;	// CStreamingInfo*
				call SetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;
				retn;
			}
		}

		// Original code:	mov [edi], si
		NAKED void patch_40AF10()
		{
			__asm
			{
				push esi;	// new value
				push 0;	// member offset
				push edi;	// CStreamingInfo*
				call SetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;
				retn;
			}
		}

		// Original code:	movsx eax, word ptr [ecx]
		NAKED void patch_40AF13()
		{
			__asm
			{
				push 0;	// space for result
				push 0;	// member offset
				push ecx;	// CStreamingInfo*
				call GetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;

				mov eax, [esp - 4];
				retn;
			}
		}

		// Original code:	mov [edx+ecx*4+2], si
		NAKED void patch_40AF1F()
		{
			__asm
			{
				push ebx;
				lea ebx, [edx + ecx * 4];

				push esi;	// new value
				push 2;	// member offset
				push ebx;	// CStreamingInfo*
				call SetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;

				pop ebx;
				retn;
			}
		}

		// Original code:	cmp word ptr [ecx], 0FFFFh
		NAKED void patch_40B952()
		{
			__asm
			{
				push 0;	// space for result
				push 0;	// member offset
				push ecx;	// CStreamingInfo*
				call GetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;

				cmp dword ptr[esp - 4], UNDEFINED_FILE_ID_INT32;
				retn;
			}
		}

		// Original code:	cmp word ptr [ecx], 0FFFFh
		NAKED void patch_40BAFA()
		{
			__asm
			{
				push 0;	// space for result
				push 0;	// member offset
				push ecx;	// CStreamingInfo*
				call GetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;

				cmp dword ptr[esp - 4], UNDEFINED_FILE_ID_INT32;
				retn;
			}
		}

		// Original code:	cmp word ptr [esi], 0FFFFh
		NAKED void patch_40BB5E()
		{
			__asm
			{
				push 0;	// space for result
				push 0;	// member offset
				push esi;	// CStreamingInfo*
				call GetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;

				cmp dword ptr[esp - 4], UNDEFINED_FILE_ID_INT32;
				retn;
			}
		}

		// Original code:	cmp word ptr [ecx], 0FFFFh
		NAKED void patch_40BEB9()
		{
			__asm
			{
				push 0;	// space for result
				push 0;	// member offset
				push ecx;	// CStreamingInfo*
				call GetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;

				cmp dword ptr[esp - 4], UNDEFINED_FILE_ID_INT32;
				retn;
			}
		}

		// Original code:	cmp word ptr [esi], 0FFFFh
		NAKED void patch_40BF19()
		{
			__asm
			{
				push 0;	// space for result
				push 0;	// member offset
				push esi;	// CStreamingInfo*
				call GetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;

				cmp dword ptr[esp - 4], UNDEFINED_FILE_ID_INT32;
				retn;
			}
		}

		// Original code:	mov ax, [eax+2]
		NAKED void patch_40C1E5()
		{
			__asm
			{
				push 0;	// space for result
				push 2;	// member offset
				push eax;	// CStreamingInfo*
				call GetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;

				mov eax, [esp - 4];
				retn;
			}
		}

		// Original code:	cmp ax, 0FFFFh
		NAKED void patch_40C1E9()
		{
			__asm
			{
				cmp eax, UNDEFINED_FILE_ID_INT32;
				retn;
			}
		}

		// Original code:	mov ax, [esi+2]
		NAKED void patch_40C229()
		{
			__asm
			{
				push 0;	// space for result
				push 2;	// member offset
				push esi;	// CStreamingInfo*
				call GetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;

				mov eax, [esp - 4];
				retn;
			}
		}

		// Original code:	cmp ax, 0FFFFh
		NAKED void patch_40C22D()
		{
			__asm
			{
				cmp eax, UNDEFINED_FILE_ID_INT32;
				retn;
			}
		}

		// Original code:	movsx esi, ds:ms_aInfoForModel.usNextOnCd[ebx]
		NAKED void patch_40CEDB()
		{
			__asm
			{
				push 0;	// space for result
				push 4;	// member offset
				push ebx;	// CStreaming::ms_aInfoForModel offset
				call GetExtendedMemberByStreamingInfoMemOffset;
				add esp, 0xC;

				mov esi, [esp - 4];
				retn;
			}
		}

		// Original code:	mov ax, [eax+2]
		NAKED void patch_40CFD5()
		{
			__asm
			{
				push 0;	// space for result
				push 2;	// member offset
				push eax;	// CStreamingInfo*
				call GetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;

				mov eax, [esp - 4];
				retn;
			}
		}

		// Original code:	cmp ax, 0FFFFh
		NAKED void patch_40CFD9()
		{
			__asm
			{
				cmp eax, UNDEFINED_FILE_ID_INT32;
				retn;
			}
		}

		// Original code:	mov di, [edi+2]
		NAKED void patch_40D030()
		{
			__asm
			{
				push 0;	// space for result
				push 2;	// member offset
				push edi;	// CStreamingInfo*
				call GetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;

				mov edi, [esp - 4];
				retn;
			}
		}

		// Original code:	cmp di, 0FFFFh
		NAKED void patch_40D034()
		{
			__asm
			{
				cmp edi, UNDEFINED_FILE_ID_INT32;
				retn;
			}
		}

		// Original code:	mov di, [edi+2]
		NAKED void patch_40D11B()
		{
			__asm
			{
				push 0;	// space for result
				push 2;	// member offset
				push edi;	// CStreamingInfo*
				call GetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;

				mov edi, [esp - 4];
				retn;
			}
		}

		// Original code:	cmp di, 0FFFFh
		NAKED void patch_40D11F()
		{
			__asm
			{
				cmp edi, UNDEFINED_FILE_ID_INT32;
				retn;
			}
		}

		// Original code:	mov ax, [eax+2]
		NAKED void patch_40D2F5()
		{
			__asm
			{
				push 0;	// space for result
				push 2;	// member offset
				push eax;	// CStreamingInfo*
				call GetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;

				mov eax, [esp - 4];
				retn;
			}
		}

		// Original code:	cmp ax, 0FFFFh
		NAKED void patch_40D2F9()
		{
			__asm
			{
				cmp eax, UNDEFINED_FILE_ID_INT32;
				retn;
			}
		}

		// Original code:	mov ax, [esi+2]
		NAKED void patch_40D349()
		{
			__asm
			{
				push 0;	// space for result
				push 2;	// member offset
				push esi;	// CStreamingInfo*
				call GetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;

				mov eax, [esp - 4];
				retn;
			}
		}

		// Original code:	cmp ax, 0FFFFh
		NAKED void patch_40D34D()
		{
			__asm
			{
				cmp eax, UNDEFINED_FILE_ID_INT32;
				retn;
			}
		}

		// Original code:	mov ax, [eax]
		NAKED void patch_40E4E5()
		{
			__asm
			{
				push 0;	// space for result
				push 0;	// member offset
				push eax;	// CStreamingInfo*
				call GetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;

				mov eax, [esp - 4];
				retn;
			}
		}

		// Original code:	cmp ax, 0FFFFh
		NAKED void patch_40E4E8()
		{
			__asm
			{
				cmp eax, UNDEFINED_FILE_ID_INT32;
				retn;
			}
		}

		// Original code:	mov ax, [ecx]
		NAKED void patch_40E510()
		{
			__asm
			{
				push 0;	// space for result
				push 0;	// member offset
				push ecx;	// CStreamingInfo*
				call GetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;

				mov eax, [esp - 4];
				retn;
			}
		}

		// Original code:	cmp ax, 0FFFFh
		NAKED void patch_40E513()
		{
			__asm
			{
				cmp eax, UNDEFINED_FILE_ID_INT32;
				retn;
			}
		}

		// Original code:	mov ax, [eax+2]
		NAKED void patch_40EA46()
		{
			__asm
			{
				push 0;	// space for result
				push 2;	// member offset
				push eax;	// CStreamingInfo*
				call GetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;

				mov eax, [esp - 4];
				retn;
			}
		}

		// Original code:	cmp ax, 0FFFFh
		NAKED void patch_40EA4A()
		{
			__asm
			{
				cmp eax, UNDEFINED_FILE_ID_INT32;
				retn;
			}
		}

		// Original code:	mov ax, [ecx+2]
		NAKED void patch_40EBD8()
		{
			__asm
			{
				push 0;	// space for result
				push 2;	// member offset
				push ecx;	// CStreamingInfo*
				call GetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;

				mov eax, [esp - 4];
				retn;
			}
		}

		// Original code:	cmp ax, 0FFFFh
		NAKED void patch_40EBDC()
		{
			__asm
			{
				cmp eax, UNDEFINED_FILE_ID_INT32;
				retn;
			}
		}

		// Original code:	mov ax, [esi+2]
		NAKED void patch_40EC0F()
		{
			__asm
			{
				push 0;	// space for result
				push 2;	// member offset
				push esi;	// CStreamingInfo*
				call GetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;

				mov eax, [esp - 4];
				retn;
			}
		}

		// Original code:	cmp ax, 0FFFFh
		NAKED void patch_40EC1D()
		{
			__asm
			{
				cmp eax, UNDEFINED_FILE_ID_INT32;
				retn;
			}
		}

		// Original code:	cmp word ptr [ecx], 0FFFFh
		NAKED void patch_1562A12()
		{
			__asm
			{
				push 0;	// space for result
				push 0;	// member offset
				push ecx;	// CStreamingInfo*
				call GetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;

				cmp dword ptr[esp - 4], UNDEFINED_FILE_ID_INT32;
				retn;
			}
		}

		// Original code:	cmp word ptr [esi], 0FFFFh
		NAKED void patch_1562A72()
		{
			__asm
			{
				push 0;	// space for result
				push 0;	// member offset
				push esi;	// CStreamingInfo*
				call GetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;

				cmp dword ptr[esp - 4], UNDEFINED_FILE_ID_INT32;
				retn;
			}
		}

		// Original code:	cmp word ptr [ecx], 0FFFFh
		NAKED void patch_1562B15()
		{
			__asm
			{
				push 0;	// space for result
				push 0;	// member offset
				push ecx;	// CStreamingInfo*
				call GetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;

				cmp dword ptr[esp - 4], UNDEFINED_FILE_ID_INT32;
				retn;
			}
		}

		// Original code:	mov dx, [edi]
		NAKED void patch_1562B3E()
		{
			__asm
			{
				push 0;	// space for result
				push 0;	// member offset
				push edi;	// CStreamingInfo*
				call GetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;

				mov edx, [esp - 4];
				retn;
			}
		}

		// Original code:	mov [ecx], dx
		NAKED void patch_1562B41()
		{
			__asm
			{
				push edx;	// new value
				push 0;	// member offset
				push ecx;	// CStreamingInfo*
				call SetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;
				retn;
			}
		}

		// Original code:	mov [ecx+2], dx
		NAKED void patch_1562B59()
		{
			__asm
			{
				push edx;	// new value
				push 2;	// member offset
				push ecx;	// CStreamingInfo*
				call SetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;
				retn;
			}
		}

		// Original code:	mov [edi], si
		NAKED void patch_1562B5D()
		{
			__asm
			{
				push esi;	// new value
				push 0;	// member offset
				push edi;	// CStreamingInfo*
				call SetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;
				retn;
			}
		}

		// Original code:	movsx eax, word ptr [ecx]
		NAKED void patch_1562B60()
		{
			__asm
			{
				push 0;	// space for result
				push 0;	// member offset
				push ecx;	// CStreamingInfo*
				call GetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;

				mov eax, [esp - 4];
				retn;
			}
		}

		// Original code:	mov [edx+ecx*4+2], si
		NAKED void patch_1562B70()
		{
			__asm
			{
				push ebx;
				lea ebx, [edx + ecx * 4];

				push esi;	// new value
				push 2;	// member offset
				push ebx;	// CStreamingInfo*
				call SetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;

				pop ebx;
				retn;
			}
		}

		// Original code:	cmp word ptr [ecx], 0FFFFh
		NAKED void patch_1562BD1()
		{
			__asm
			{
				push 0;	// space for result
				push 0;	// member offset
				push ecx;	// CStreamingInfo*
				call GetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;

				cmp dword ptr[esp - 4], UNDEFINED_FILE_ID_INT32;
				retn;
			}
		}

		// Original code:	mov dx, [edi]
		NAKED void patch_1562BFA()
		{
			__asm
			{
				push 0;	// space for result
				push 0;	// member offset
				push edi;	// CStreamingInfo*
				call GetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;

				mov edx, [esp - 4];
				retn;
			}
		}

		// Original code:	mov [ecx], dx
		NAKED void patch_1562BFD()
		{
			__asm
			{
				push edx;	// new value
				push 0;	// member offset
				push ecx;	// CStreamingInfo*
				call SetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;
				retn;
			}
		}

		// Original code:	mov [ecx+2], dx
		NAKED void patch_1562C15()
		{
			__asm
			{
				push edx;	// new value
				push 2;	// member offset
				push ecx;	// CStreamingInfo*
				call SetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;
				retn;
			}
		}

		// Original code:	mov [edi], si
		NAKED void patch_1562C19()
		{
			__asm
			{
				push esi;	// new value
				push 0;	// member offset
				push edi;	// CStreamingInfo*
				call SetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;
				retn;
			}
		}

		// Original code:	movsx eax, word ptr [ecx]
		NAKED void patch_1562C1C()
		{
			__asm
			{
				push 0;	// space for result
				push 0;	// member offset
				push ecx;	// CStreamingInfo*
				call GetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;

				mov eax, [esp - 4];
				retn;
			}
		}

		// Original code:	mov [edx+ecx*4+2], si
		NAKED void patch_1562C2C()
		{
			__asm
			{
				push ebx;
				lea ebx, [edx + ecx * 4];

				push esi;	// new value
				push 2;	// member offset
				push ebx;	// CStreamingInfo*
				call SetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;

				pop ebx;
				retn;
			}

		}

		// Original code:	mov ax, [eax]
		NAKED void patch_15664B5()
		{
			__asm
			{
				push 0;	// space for result
				push 0;	// member offset
				push eax;	// CStreamingInfo*
				call GetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;

				mov eax, [esp - 4];
				retn;
			}
		}

		// Original code:	cmp ax, 0FFFFh
		NAKED void patch_15664B8()
		{
			__asm
			{
				cmp eax, UNDEFINED_FILE_ID_INT32;
				retn;
			}
		}

		// Original code:	mov si, [esi]
		NAKED void patch_1566552()
		{
			__asm
			{
				push 0;	// space for result
				push 0;	// member offset
				push esi;	// CStreamingInfo*
				call GetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;

				mov esi, [esp - 4];
				retn;
			}
		}

		// Original code:	cmp si, 0FFFFh
		NAKED void patch_1566555()
		{
			__asm
			{
				cmp esi, UNDEFINED_FILE_ID_INT32;
				retn;
			}
		}

		// Original code:	mov ax, [edi]
		NAKED void patch_15674F9()
		{
			__asm
			{
				push 0;	// space for result
				push 0;	// member offset
				push edi;	// CStreamingInfo*
				call GetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;

				mov eax, [esp - 4];
				retn;
			}
		}

		// Original code:	mov [ecx], ax
		NAKED void patch_15674FC()
		{
			__asm
			{
				push eax;	// new value
				push 0;	// member offset
				push ecx;	// CStreamingInfo*
				call SetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;
				retn;
			}
		}

		// Original code:	mov [ecx+2], dx
		NAKED void patch_15674FF()
		{
			__asm
			{
				push edx;	// new value
				push 2;	// member offset
				push ecx;	// CStreamingInfo*
				call SetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;
				retn;
			}
		}

		// Original code:	mov [edi], si
		NAKED void patch_1567503()
		{
			__asm
			{
				push esi;	// new value
				push 0;	// member offset
				push edi;	// CStreamingInfo*
				call SetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;
				retn;
			}
		}

		// Original code:	movsx eax, word ptr [ecx]
		NAKED void patch_1567506()
		{
			__asm
			{
				push 0;	// space for result
				push 0;	// member offset
				push ecx;	// CStreamingInfo*
				call GetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;

				mov eax, [esp - 4];
				retn;
			}
		}

		// Original code:	mov [edx+ecx*4+2], si
		NAKED void patch_1567513()
		{
			__asm
			{
				push ebx;
				lea ebx, [edx + ecx * 4];

				push esi;	// new value
				push 2;	// member offset
				push ebx;	// CStreamingInfo*
				call SetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;

				pop ebx;
				retn;
			}

		}

		// Original code:	mov [ecx], dx
		NAKED void patch_1567528()
		{
			__asm
			{
				push edx;	// new value
				push 0;	// member offset
				push ecx;	// CStreamingInfo*
				call SetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;
				retn;
			}
		}

		// Original code:	mov [ecx+2], dx
		NAKED void patch_156752B()
		{
			__asm
			{
				push edx;	// new value
				push 2;	// member offset
				push ecx;	// CStreamingInfo*
				call SetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;
				retn;
			}
		}

		// Original code:	mov [ecx+4], dx
		NAKED void patch_156752F()
		{
			__asm
			{
				push edx;	// new value
				push 4;	// member offset
				push ecx;	// CStreamingInfo*
				call SetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;
				retn;
			}
		}

		// Original code:	cmp word ptr ds:ms_aInfoForModel.usNext[eax], 0
		NAKED void patch_1567B73()
		{
			__asm
			{
				push 0;	// space for result
				push 0;	// member offset
				push eax;	// CStreaming::ms_aInfoForModel offset
				call GetExtendedMemberByStreamingInfoMemOffset;
				add esp, 0xC;

				cmp dword ptr[esp - 4], 0;
				retn;
			}
		}

		// Original code:	cmp ds:ms_aInfoForModel.usPrev[eax], 0
		NAKED void patch_1567B7D()
		{
			__asm
			{
				push 0;	// space for result
				push 2;	// member offset
				push eax;	// CStreaming::ms_aInfoForModel offset
				call GetExtendedMemberByStreamingInfoMemOffset;
				add esp, 0xC;

				cmp dword ptr[esp - 4], 0;
				retn;
			}
		}

		// Original code:	movsx eax, word ptr [ecx]
		NAKED void patch_156CE90()
		{
			__asm
			{
				push 0;	// space for result
				push 0;	// member offset
				push ecx;	// CStreamingInfo*
				call GetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;

				mov eax, [esp - 4];
				retn;
			}
		}

		// Original code:	mov dx, [ecx+2]
		NAKED void patch_156CE93()
		{
			__asm
			{
				push 0;	// space for result
				push 2;	// member offset
				push ecx;	// CStreamingInfo*
				call GetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;

				mov edx, [esp - 4];
				retn;
			}
		}

		// Original code:	mov [esi+eax*4+2], dx
		NAKED void patch_156CEA1()
		{
			__asm
			{
				push ebx;
				lea ebx, [esi + eax * 4];

				push edx;	// new value
				push 2;	// member offset
				push ebx;	// CStreamingInfo*
				call SetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;

				pop ebx;
				retn;
			}
		}

		// Original code:	movsx eax, word ptr [ecx+2]
		NAKED void patch_156CEA6()
		{
			__asm
			{
				push 0;	// space for result
				push 2;	// member offset
				push ecx;	// CStreamingInfo*
				call GetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;

				mov eax, [esp - 4];
				retn;
			}
		}

		// Original code:	mov dx, [ecx]
		NAKED void patch_156CEB0()
		{
			__asm
			{
				push 0;	// space for result
				push 0;	// member offset
				push ecx;	// CStreamingInfo*
				call GetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;

				mov edx, [esp - 4];
				retn;
			}
		}

		// Original code:	mov [esi+eax*4], dx
		NAKED void patch_156CEB6()
		{
			__asm
			{
				push ebx;
				lea ebx, [esi + eax * 4];

				push edx;	// new value
				push 0;	// member offset
				push ebx;	// CStreamingInfo*
				call SetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;

				pop ebx;
				retn;
			}
		}

		// Original code:	mov word ptr [ecx], 0FFFFh
		NAKED void patch_156CEBA()
		{
			__asm
			{
				push UNDEFINED_FILE_ID_INT32;	// new value
				push 0;	// member offset
				push ecx;	// CStreamingInfo*
				call SetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;
				retn;
			}
		}

		// Original code:	mov word ptr [ecx+2], 0FFFFh
		NAKED void patch_156CEBF()
		{
			__asm
			{
				push UNDEFINED_FILE_ID_INT32;	// new value
				push 2;	// member offset
				push ecx;	// CStreamingInfo*
				call SetExtendedMemberByStreamingInfoPtr;
				add esp, 0xC;
				retn;
			}
		}

		static const tArrayItem ptrArray[] =
		{
			tArrayItem(0x407520, &patch_407520),
			tArrayItem(0x407523, &patch_407523),
			tArrayItem(0x407540, &patch_407540),
			tArrayItem(0x407544, &patch_407544),
			tArrayItem(0x407562, &patch_407562),
			tArrayItem(0x407AD5, &patch_407AD5),
			tArrayItem(0x407AD8, &patch_407AD8),
			tArrayItem(0x407B35, &patch_407B35),
			tArrayItem(0x407B38, &patch_407B38),
			tArrayItem(0x40884E, &patch_40884E),
			tArrayItem(0x40885E, &patch_40885E),
			tArrayItem(0x408861, &patch_408861),
			tArrayItem(0x40886E, &patch_40886E),
			tArrayItem(0x408873, &patch_408873),
			tArrayItem(0x408877, &patch_408877),
			tArrayItem(0x408883, &patch_408883),
			tArrayItem(0x408887, &patch_408887),
			tArrayItem(0x40888C, &patch_40888C),
			tArrayItem(0x408ADA, &patch_408ADA),
			tArrayItem(0x408B22, &patch_408B22),
			tArrayItem(0x408B32, &patch_408B32),
			tArrayItem(0x408B39, &patch_408B39),
			tArrayItem(0x408B3E, &patch_408B3E),
			tArrayItem(0x408B4E, &patch_408B4E),
			tArrayItem(0x408B55, &patch_408B55),
			tArrayItem(0x408B59, &patch_408B59),
			tArrayItem(0x408B60, &patch_408B60),
			tArrayItem(0x408E28, &patch_408E28),
			tArrayItem(0x408E2E, &patch_408E2E),
			tArrayItem(0x408E72, &patch_408E72),
			tArrayItem(0x408E7F, &patch_408E7F),
			tArrayItem(0x4098C8, &patch_4098C8),
			tArrayItem(0x409978, &patch_409978),
			tArrayItem(0x409A39, &patch_409A39),
			tArrayItem(0x409C41, &patch_409C41),
			tArrayItem(0x409CCC, &patch_409CCC),
			tArrayItem(0x409FCA, &patch_409FCA),
			tArrayItem(0x40A862, &patch_40A862),
			tArrayItem(0x40A88B, &patch_40A88B),
			tArrayItem(0x40A88E, &patch_40A88E),
			tArrayItem(0x40A8AA, &patch_40A8AA),
			tArrayItem(0x40A8AE, &patch_40A8AE),
			tArrayItem(0x40A8B1, &patch_40A8B1),
			tArrayItem(0x40A8C1, &patch_40A8C1),
			tArrayItem(0x40A926, &patch_40A926),
			tArrayItem(0x40A953, &patch_40A953),
			tArrayItem(0x40A956, &patch_40A956),
			tArrayItem(0x40A972, &patch_40A972),
			tArrayItem(0x40A976, &patch_40A976),
			tArrayItem(0x40A979, &patch_40A979),
			tArrayItem(0x40A989, &patch_40A989),
			tArrayItem(0x40ABEB, &patch_40ABEB),
			tArrayItem(0x40AC59, &patch_40AC59),
			tArrayItem(0x40AE1F, &patch_40AE1F),
			tArrayItem(0x40AE48, &patch_40AE48),
			tArrayItem(0x40AE4B, &patch_40AE4B),
			tArrayItem(0x40AE63, &patch_40AE63),
			tArrayItem(0x40AE67, &patch_40AE67),
			tArrayItem(0x40AE6A, &patch_40AE6A),
			tArrayItem(0x40AE76, &patch_40AE76),
			tArrayItem(0x40AEC8, &patch_40AEC8),
			tArrayItem(0x40AEF1, &patch_40AEF1),
			tArrayItem(0x40AEF4, &patch_40AEF4),
			tArrayItem(0x40AF0C, &patch_40AF0C),
			tArrayItem(0x40AF10, &patch_40AF10),
			tArrayItem(0x40AF13, &patch_40AF13),
			tArrayItem(0x40AF1F, &patch_40AF1F),
			tArrayItem(0x40B952, &patch_40B952),
			tArrayItem(0x40BAFA, &patch_40BAFA),
			tArrayItem(0x40BB5E, &patch_40BB5E),
			tArrayItem(0x40BEB9, &patch_40BEB9),
			tArrayItem(0x40BF19, &patch_40BF19),
			tArrayItem(0x40C1E5, &patch_40C1E5),
			tArrayItem(0x40C1E9, &patch_40C1E9),
			tArrayItem(0x40C229, &patch_40C229),
			tArrayItem(0x40C22D, &patch_40C22D),
			tArrayItem(0x40CEDB, &patch_40CEDB),
			tArrayItem(0x40CFD5, &patch_40CFD5),
			tArrayItem(0x40CFD9, &patch_40CFD9),
			tArrayItem(0x40D030, &patch_40D030),
			tArrayItem(0x40D034, &patch_40D034),
			tArrayItem(0x40D11B, &patch_40D11B),
			tArrayItem(0x40D11F, &patch_40D11F),
			tArrayItem(0x40D2F5, &patch_40D2F5),
			tArrayItem(0x40D2F9, &patch_40D2F9),
			tArrayItem(0x40D349, &patch_40D349),
			tArrayItem(0x40D34D, &patch_40D34D),
			tArrayItem(0x40E4E5, &patch_40E4E5),
			tArrayItem(0x40E4E8, &patch_40E4E8),
			tArrayItem(0x40E510, &patch_40E510),
			tArrayItem(0x40E513, &patch_40E513),
			tArrayItem(0x40EA46, &patch_40EA46),
			tArrayItem(0x40EA4A, &patch_40EA4A),
			tArrayItem(0x40EBD8, &patch_40EBD8),
			tArrayItem(0x40EBDC, &patch_40EBDC),
			tArrayItem(0x40EC0F, &patch_40EC0F),
			tArrayItem(0x40EC1D, &patch_40EC1D),
			tArrayItem(0x1562A12, &patch_1562A12),
			tArrayItem(0x1562A72, &patch_1562A72),
			tArrayItem(0x1562B15, &patch_1562B15),
			tArrayItem(0x1562B3E, &patch_1562B3E),
			tArrayItem(0x1562B41, &patch_1562B41),
			tArrayItem(0x1562B59, &patch_1562B59),
			tArrayItem(0x1562B5D, &patch_1562B5D),
			tArrayItem(0x1562B60, &patch_1562B60),
			tArrayItem(0x1562B70, &patch_1562B70),
			tArrayItem(0x1562BD1, &patch_1562BD1),
			tArrayItem(0x1562BFA, &patch_1562BFA),
			tArrayItem(0x1562BFD, &patch_1562BFD),
			tArrayItem(0x1562C15, &patch_1562C15),
			tArrayItem(0x1562C19, &patch_1562C19),
			tArrayItem(0x1562C1C, &patch_1562C1C),
			tArrayItem(0x1562C2C, &patch_1562C2C),
			tArrayItem(0x15664B5, &patch_15664B5),
			tArrayItem(0x15664B8, &patch_15664B8),
			tArrayItem(0x1566552, &patch_1566552),
			tArrayItem(0x1566555, &patch_1566555),
			tArrayItem(0x15674F9, &patch_15674F9),
			tArrayItem(0x15674FC, &patch_15674FC),
			tArrayItem(0x15674FF, &patch_15674FF),
			tArrayItem(0x1567503, &patch_1567503),
			tArrayItem(0x1567506, &patch_1567506),
			tArrayItem(0x1567513, &patch_1567513),
			tArrayItem(0x1567528, &patch_1567528),
			tArrayItem(0x156752B, &patch_156752B),
			tArrayItem(0x156752F, &patch_156752F),
			tArrayItem(0x1567B73, &patch_1567B73),
			tArrayItem(0x1567B7D, &patch_1567B7D),
			tArrayItem(0x156CE90, &patch_156CE90),
			tArrayItem(0x156CE93, &patch_156CE93),
			tArrayItem(0x156CEA1, &patch_156CEA1),
			tArrayItem(0x156CEA6, &patch_156CEA6),
			tArrayItem(0x156CEB0, &patch_156CEB0),
			tArrayItem(0x156CEB6, &patch_156CEB6),
			tArrayItem(0x156CEBA, &patch_156CEBA),
			tArrayItem(0x156CEBF, &patch_156CEBF),
		};
	}
#endif

#include "FileIDlimit\GTA SA 1.0 US HOODLUM WIN_X86\GTASA_ID_access_functions.h"

	// Finds total number of peds killed by player.
	int __cdecl CDarkel::FindTotalPedsKilledByPlayer(int playerID)
	{
		int16_t* pRegisteredKillsForThisPlayer = g_fileIDlimits.CDarkel__RegisteredKills.ptr + playerID;
		int16_t sum = 0;

		for(unsigned int i = 0; i < g_fileIDlimits.CDarkel__CountOfKillableModelIDs; i++)
			sum += *pRegisteredKillsForThisPlayer;
		
		return sum;
	}

	// Returns position and size of file
	bool CStreamingInfo::GetCdPosnAndSize(unsigned int &outBlockOffsetWithStreamHandle, unsigned int &outBlockCount)
	{
		return ((bool(__thiscall *)(CStreamingInfo*, unsigned int &, unsigned int &))0x4075A0)(
			this,
			outBlockOffsetWithStreamHandle,
			outBlockCount
			);
	}

	// Sets position and size of file
	void CStreamingInfo::SetCdPosnAndSize(unsigned int BlockOffset, unsigned int BlockCount)
	{
		return ((void(__thiscall *)(CStreamingInfo*, unsigned int, unsigned int))0x4075E0)(
			this,
			BlockOffset,
			BlockCount
			);
	}

	/*
	Crash documentation

	0x534134 (CEntity__GetBoundRect) - model ID = 3110, problem: modelPtrs[3110] -> CBaseModelInfo.m_pColModel = NULL

	model ID 3110 = md_billbolaeb
	model name hash by crc32FromUpcaseString = 0xea9bb8da

	modelPtrs[3110] = 0xB0B394 (objs)
	0xB0B394 + 0x14 = 0xB0B3A8 (CBaseModelInfo.m_pColModel)

	Where modelPtrs[3110] -> CBaseModelInfo.m_pColModel is set?
	0x4C4BD2 - function 0x4C4BC0
	called by 0x5B5195 - reads .col file
	called by 0x4106D0


	Name of COL file = levelxre.col
	ID of COL file = 15
	general ID of COL file = 25015


	file ID of COL file is placed on
	0x8E4AF8 [WORD]

	Written by 0x40CE3D

	Previous file position: 0xA0037A4

	45th file is wrong - 25020, but should be 25012

	Problem: CStreaming__ms_pStartRequestedList->usNext = 25020 when limit hacked.


	*/
	
	/*

		Save changes:

		Block 18: Models flags
			can be standard version or enhanced version:

			ENHANCED_FLAG_IDENTIFIER block header, without NULL.
			char version;	 // should have a value of 1
			uint32_t numberOfBytesFollowingThisByte;

			for (eFileType type = FILE_TYPE_DFF; type != FILE_TYPE_COUNT; ++type)
			{
				numberOfFilesOfThisType;
			}
	*/

	#ifdef IS_PLATFORM_WIN_X86
	// readIPLsFromSave patch for ID limit
	NAKED void patch_CIplStore__Load_5D5514()
	{
		__asm
		{
			mov eax, [esp + 10h];	// numberOfIplFiles

			cmp ebx, eax;

			// go to loop
			jae returnFromFunc;

			mov eax, 0x5D54C0;
			jmp eax;

		returnFromFunc:
			pop edi;
			pop esi;
			mov al, 1;
			pop ebx;
			add esp, 8;
			retn;
		}
	}
	#elif defined(IS_PLATFORM_ANDROID_ARM32)
	// patch for 0x48E46C
	extern "C"
	{
		uintptr_t Address_CIplStore__Load_48E426_thumb = 0;
		uintptr_t Address_CIplStore__Load_48E476_thumb = 0;
	}

	DLL_EXPORT TARGET_THUMB NAKED void patch_CIplStore__Load_48E46C()
	{
		__asm(
		".thumb\n"
			"ADDS R4, #1\n"
			"ADDS R6, #0x34\n"

			"LDR R0, [SP,#4]\n"
			"CMP R4, R0\n"
			"BNE 1f\n"
			ASM_JUMP_TO_ADDRESS_STORED_ON_SYMBOL(Address_CIplStore__Load_48E476_thumb)

			// Branches
			"1:\n"	// loc_48E426
			ASM_JUMP_TO_ADDRESS_STORED_ON_SYMBOL(Address_CIplStore__Load_48E426_thumb)
			);
	}
	#endif

	static const char ENHANCED_FLAG_IDENTIFIER[] = "fastman92_modelFlags";
	static const int ENHANCED_FLAG_IDENTIFIER_LEN = sizeof(ENHANCED_FLAG_IDENTIFIER) - 1;

	static const eFileType arrayOfFileTypes[] =
	{
		FILE_TYPE_DFF,
		FILE_TYPE_TXD,

		FILE_TYPE_COL,
		FILE_TYPE_IPL,
		FILE_TYPE_DAT,
		FILE_TYPE_IFP,
		FILE_TYPE_RRR,
		FILE_TYPE_SCM,
		FILE_TYPE_LOADED_START,
		FILE_TYPE_REQUESTED_START,
	};

	/*
		enhanced format of model flags
			version 1 - uses 16-bit counts of files
			version 2 - uses 32-bit counts of files
	*/

	// Patch for readModelFlagsFromSave
	bool __cdecl CStreaming__Load()
	{
		char buffer[ENHANCED_FLAG_IDENTIFIER_LEN + 1];

		for(unsigned int i = 0; i < ENHANCED_FLAG_IDENTIFIER_LEN; i++)
			CGenericGameStorage::LoadDataFromWorkBuffer(buffer + i, 1);
		
		if (strncmp(buffer, ENHANCED_FLAG_IDENTIFIER, ENHANCED_FLAG_IDENTIFIER_LEN))	// is standard version?
		{
			tFileTypeBaseID defaultCount = g_fileIDlimits.GetDefaultCountOfAllFileIDs();

			uint8_t* pFlags = new uint8_t[defaultCount];

			memcpy(pFlags, buffer, ENHANCED_FLAG_IDENTIFIER_LEN);

			// read the rest of data
			for (unsigned int i = 0; i < defaultCount - ENHANCED_FLAG_IDENTIFIER_LEN; i++)
				CGenericGameStorage::LoadDataFromWorkBuffer(pFlags + ENHANCED_FLAG_IDENTIFIER_LEN + i, 1);
			
			for (eFileType type : arrayOfFileTypes)
			{
				uint16_t numberOfFiles = g_fileIDlimits.GetFileIDdefaultLimit(type);

				if (g_fileIDlimits.GetFileIDcurrentLimit(type) < numberOfFiles)
					numberOfFiles = g_fileIDlimits.GetFileIDcurrentLimit(type);

				uint8_t* iFlags = pFlags + g_fileIDlimits.GetDefaultBaseID(type);
				CStreamingInfo* oInfoForModel = &g_fileIDlimits.CStreaming__ms_aInfoForModel.gta_sa
					[g_fileIDlimits.GetBaseID(type)];
				
				for (uint16_t i = 0; i < numberOfFiles; i++)
				{
					CStreamingInfo* file = oInfoForModel + i;

					if (file->uiLoadFlag == 1 && iFlags[i] != 0xFF)
						file->ucFlags |= iFlags[i];
				}

				/*
				printf_MessageBox("File type: %d Number of source IDs: %d Input ID: %d Output ID: %d",
				type,
				numberOfFiles,
				CLimitAdjusterMain::CStreaming__defaultBaseID[type],
				CLimitAdjusterMain::GetBaseID(type)
				);
				*/
			}

			delete[] pFlags;
		}
		else
		{
			char format;

			// enhanced version
			CGenericGameStorage::LoadDataFromWorkBuffer(&format, 1);

			if (format == 1 || format == 2)
			{
				uint32_t numberOfBytesFollowing;
				CGenericGameStorage::LoadDataFromWorkBuffer(&numberOfBytesFollowing, sizeof(numberOfBytesFollowing));

				for (eFileType type : arrayOfFileTypes)
				{
					uint32_t numberOfFilesToProcess = 0;
					uint16_t numberOfFilesToSkip = 0;
					
					if(format == 1)
						CGenericGameStorage::LoadDataFromWorkBuffer(&numberOfFilesToProcess, sizeof(uint16_t));
					else if(format == 2)
						CGenericGameStorage::LoadDataFromWorkBuffer(&numberOfFilesToProcess, sizeof(uint32_t));

					uint16_t numberOfFilesInGame = g_fileIDlimits.GetFileIDcurrentLimit(type);

					CStreamingInfo* oInfoForModel = &g_fileIDlimits.CStreaming__ms_aInfoForModel.gta_sa
						[g_fileIDlimits.GetBaseID(type)];

					if (numberOfFilesInGame < numberOfFilesToProcess)
					{
						numberOfFilesToSkip = numberOfFilesToProcess - numberOfFilesInGame;
						numberOfFilesToProcess = numberOfFilesInGame;
					}

					/*
					printf_MessageBox("File type: %d\nNumber of files to process: %d\nNumber of files to skip: %d\nOutput ID: %d",
					type,
					numberOfFilesToProcess,
					numberOfFilesToSkip,
					CLimitAdjusterMain::GetBaseID(type)
					);
					*/

					uint8_t flags;

					// Process file flags
					for (uint32_t i = 0; i < numberOfFilesToProcess; i++)
					{
						CGenericGameStorage::LoadDataFromWorkBuffer(&flags, 1);

						CStreamingInfo* file = oInfoForModel + i;

						if (file->uiLoadFlag == 1 && flags != 0xFF)
							file->ucFlags |= flags;
					}

					// Skip file flags
					while (numberOfFilesToSkip-- > 0)
						CGenericGameStorage::LoadDataFromWorkBuffer(&flags, 1);
				}
			}
		}
		
		return true;
	}

	bool __cdecl saveModelFlagsInSave_patch_standard()
	{
		for (eFileType type : arrayOfFileTypes)
		{
			tFileTypeBaseID defaultNumberOfFiles = g_fileIDlimits.GetFileIDdefaultLimit(type);

			tFileTypeBaseID numberOfFilesToProcess = g_fileIDlimits.GetFileIDcurrentLimit(type);
			tFileTypeBaseID numberOfFilesToSkip = defaultNumberOfFiles - numberOfFilesToProcess;

			CStreamingInfo* iInfoForModel = &g_fileIDlimits.CStreaming__ms_aInfoForModel.gta_sa
				[g_fileIDlimits.GetBaseID(type)];

			/*
			printf_MessageBox("File type: %d\nNumber of files to process: %d\nNumber of files to skip: %d\nInput ID: %d Output ID: %d",
			type,
			numberOfFilesToProcess,
			numberOfFilesToSkip,
			CLimitAdjusterMain::GetBaseID(type),
			CLimitAdjusterMain::CStreaming__defaultBaseID[type]
			);
			*/

			uint8_t flags;

			for (uint16_t i = 0; i < numberOfFilesToProcess; i++)
			{
				CStreamingInfo* file = iInfoForModel + i;

				flags = -1;

				if (file->uiLoadFlag == 1)
					flags = file->ucFlags;

				CGenericGameStorage::SaveDataToWorkBuffer(&flags, 1);
			}

			// Skip files
			flags = -1;

			while (numberOfFilesToSkip-- > 0)
				CGenericGameStorage::SaveDataToWorkBuffer(&flags, 1);
		}

		return true;
	}

	bool __cdecl saveModelFlagsInSave_patch_enhanced()
	{
		for (unsigned int i = 0; i < ENHANCED_FLAG_IDENTIFIER_LEN; i++)
			CGenericGameStorage::SaveDataToWorkBuffer(ENHANCED_FLAG_IDENTIFIER + i, 1);

		char format = 1;
		CGenericGameStorage::SaveDataToWorkBuffer(&format, 1);
		
		uint32_t numberOfBytesFollowing = _countof(arrayOfFileTypes) * sizeof(uint16_t);

		for (eFileType type : arrayOfFileTypes)
			numberOfBytesFollowing += g_fileIDlimits.GetFileIDcurrentLimit(type);		
		
		CGenericGameStorage::SaveDataToWorkBuffer(&numberOfBytesFollowing, sizeof(numberOfBytesFollowing));

		for (eFileType type : arrayOfFileTypes)
		{
			uint16_t numberOfFiles = g_fileIDlimits.GetFileIDcurrentLimit(type);

			CGenericGameStorage::SaveDataToWorkBuffer(&numberOfFiles, sizeof(numberOfFiles));

			CStreamingInfo* FileInfoBase = &g_fileIDlimits.CStreaming__ms_aInfoForModel.gta_sa
				[g_fileIDlimits.GetBaseID(type)];

			for (uint16_t i = 0; i < numberOfFiles; i++)
			{
				CStreamingInfo* file = FileInfoBase + i;

				uint8_t flags = -1;

				if (file->uiLoadFlag == 1)
					flags = file->ucFlags;

				CGenericGameStorage::SaveDataToWorkBuffer(&flags, 1);
			}
		}

		return true;
	}

	static void ApplyPatchForCStreaming__Save(FileIDlimit* pFileIDlimits, uintptr_t functionJumpAddress)
	{
		if (pFileIDlimits->IsAnyIDlimitIncreased())
		{
			CPatch::RedirectFunction(functionJumpAddress, (void*)&saveModelFlagsInSave_patch_enhanced);

			CGenericLogStorage::SaveText("Format of new savefiles will be different: patch for enhanced format of model flags block is applied!");
			g_SaveOfVariableLength.MakeSaveOfVariableLength();
		}
		else
			CPatch::RedirectFunction(functionJumpAddress, (void*)&saveModelFlagsInSave_patch_standard);
	}

	// Initializes prev, next for info for model.
	void InitPrevNextForInfoForModel(tFileTypeBaseID firstID, tFileTypeBaseID lastID, CStreamingInfo** ppStart, CStreamingInfo** ppEnd)
	{
		*ppStart = &g_fileIDlimits.CStreaming__ms_aInfoForModel.gta_sa[firstID];
		*ppEnd = &g_fileIDlimits.CStreaming__ms_aInfoForModel.gta_sa[lastID];

		tFileTypeBaseID curID = firstID;

		if (g_fileIDlimits.bIsIDbaseInt32)
		{
			do
			{
				if (curID == firstID)
					g_fileIDlimits.CStreaming__ms_aInfoForModelExtension[curID].exUsPrev = -1;
				else
					g_fileIDlimits.CStreaming__ms_aInfoForModelExtension[curID].exUsPrev = curID - 1;

				if (curID == lastID)
					g_fileIDlimits.CStreaming__ms_aInfoForModelExtension[curID].exUsNext = -1;
				else
					g_fileIDlimits.CStreaming__ms_aInfoForModelExtension[curID].exUsNext = curID + 1;

				curID++;

			} while (curID <= lastID);
		}

		curID = firstID;

		do
		{
			if (curID == firstID)
				g_fileIDlimits.CStreaming__ms_aInfoForModel.gta_sa[curID].usPrev.raw = -1;
			else
				g_fileIDlimits.CStreaming__ms_aInfoForModel.gta_sa[curID].usPrev.raw = curID - 1;

			if (curID == lastID)
				g_fileIDlimits.CStreaming__ms_aInfoForModel.gta_sa[curID].usNext.raw = -1;
			else
				g_fileIDlimits.CStreaming__ms_aInfoForModel.gta_sa[curID].usNext.raw = curID + 1;

			curID++;

		} while (curID <= lastID);		
	}

	extern "C"
	{
		// Initializes loaded & requested.
		void __cdecl patch_CStreaming__Init2_initializeLoadedAndRequested()
		{
			// Loaded
			InitPrevNextForInfoForModel(
				g_fileIDlimits.GetBaseID(FILE_TYPE_LOADED_START),
				g_fileIDlimits.GetLastIDforType(FILE_TYPE_LOADED_START),
				(CStreamingInfo**)g_fileIDlimits.CStreaming__ms_pStartLoadedList,
				(CStreamingInfo**)g_fileIDlimits.CStreaming__ms_pEndLoadedList
			);

			// Requested
			InitPrevNextForInfoForModel(
				g_fileIDlimits.GetBaseID(FILE_TYPE_REQUESTED_START),
				g_fileIDlimits.GetLastIDforType(FILE_TYPE_REQUESTED_START),
				(CStreamingInfo**)g_fileIDlimits.CStreaming__ms_pStartRequestedList,
				(CStreamingInfo**)g_fileIDlimits.CStreaming__ms_pEndRequestedList
			);

			/*
			CLimitAdjusterMain::CStreaming__ms_aInfoForModel[26312].usNext = 26313;
			CLimitAdjusterMain::CStreaming__ms_aInfoForModel[26312].usPrev = -1;
			CLimitAdjusterMain::CStreaming__ms_aInfoForModel[26313].usNext = -1;
			CLimitAdjusterMain::CStreaming__ms_aInfoForModel[26313].usPrev = 26312;
			CLimitAdjusterMain::CStreaming__ms_aInfoForModel[26314].usNext = 26315;
			CLimitAdjusterMain::CStreaming__ms_aInfoForModel[26314].usPrev = -1;
			CLimitAdjusterMain::CStreaming__ms_aInfoForModel[26315].usNext = -1;
			CLimitAdjusterMain::CStreaming__ms_aInfoForModel[26315].usPrev = 26314;
			*/
		}
	}

	void __cdecl patch_CStreaming__Init2_initalizeExtensionArray()
	{
		const FileIDlimit* pIDlimits = &g_fileIDlimits;
		tFileTypeBaseID countOfFiles = pIDlimits->GetCountOfAllFileIDs();

		if (pIDlimits->bIsFileInfoExtensionArrayEnabled)
		{
			for (tFileTypeBaseID i = 0; i < countOfFiles; i++)
				pIDlimits->CStreaming__ms_aInfoForModelExtension[i].Init();
		}

		/*
		if (pIDlimits->bIsIDbaseInt32)
		{
			
			for (tFileTypeBaseID i = 0; i < countOfFiles; i++)
			{
				const CStreamingInfo* pFileInfo = pIDlimits->CStreaming__ms_aInfoForModel.gta_sa + i;
				New_ID_from16bit(&pFileInfo->usNext);
				New_ID_from16bit(&pFileInfo->usPrev);
				New_ID_from16bit(&pFileInfo->usNextOnCd);
			}			
		}
		*/
	}

	#ifdef IS_PLATFORM_WIN_X86
	// Initializes CStreaming::ms_aInfoForModelExtension
	NAKED void patch_CStreaming__Init2_5B8AE7()
	{
		
		__asm
		{
			call patch_CStreaming__Init2_initalizeExtensionArray;

			// go back
			mov esi, ASM_ACCESS_LIMIT_VAR_32_BIT(g_fileIDlimits, FileIDlimit, CStreaming__ms_aInfoForModel);

			mov eax, 0x5B8AF0;
			jmp eax;
		}
	}

	// Initializes loaded & requested.
	NAKED void patch_CStreaming__Init2_5B8B12()
	{
		__asm
		{
			call patch_CStreaming__Init2_initializeLoadedAndRequested;

			mov eax, 0x5B8B7A;
			jmp eax;
		}
	}
	#endif
	
	extern "C"
	{
		void __cdecl SetCollisionFileIDinCollisionModel(CColModel* pColModel, int32_t COL_ID)
		{
			pColModel->COL_file_ID = COL_ID;
		}

		int __cdecl GetCollisionFileIDfromCollisionModel(const CColModel* pColModel)
		{
			return pColModel->COL_file_ID;
		}
	}

	#ifdef IS_PLATFORM_ANDROID_ARM32
	// patch for 0x46700E
	extern "C"
	{
		uintptr_t Address_CFileLoader__LoadCollisionFile_46701A = 0;
		uintptr_t Address_CFileLoader__LoadCollisionFile_467018_thumb = 0;
	}

	static TARGET_THUMB NAKED void patch_CFileLoader__LoadCollisionFile_46700E()
	{
		__asm(
			".thumb\n"
			"MOV R1, R0\n"	// ID
			"MOV R0, R11\n"
			"BLX SetCollisionFileIDinCollisionModel\n"

			"LDR R11, =("/* unk_9B2280 */"0x9B2280 - 0x46701A)\n"
			ASM_ADD_ADDRESS_STORED_ON_SYMBOL(R11, Address_CFileLoader__LoadCollisionFile_46701A)
			ASM_JUMP_TO_ADDRESS_STORED_ON_SYMBOL(Address_CFileLoader__LoadCollisionFile_467018_thumb)
			);
	}

	// patch for 0x452CBE
	extern "C"
	{
		uintptr_t Address_CObject__Init_452CE8_thumb = 0;
		uintptr_t Address_CObject__Init_452CCA_thumb = 0;
	}

	static TARGET_THUMB NAKED void patch_CObject__Init_452CBE()
	{
		__asm(
		".thumb\n"
			"LDRB R1, [R6,#0x29]\n"
			"LSLS R1, R1, #0x1F\n"
			"BEQ 1f\n"
			
			"MOV R0, R6\n"
			"BLX GetCollisionFileIDfromCollisionModel\n"
			ASM_JUMP_TO_ADDRESS_STORED_ON_SYMBOL(Address_CObject__Init_452CCA_thumb)

			// Branches
			"1:\n"	// loc_452CE8
			ASM_JUMP_TO_ADDRESS_STORED_ON_SYMBOL(Address_CObject__Init_452CE8_thumb)
			);
	}

	// patch for 0x3F0866
	extern "C"
	{
		uintptr_t Address_CFileLoader__LoadObjectInstance_3F08D6_thumb = 0;
		uintptr_t Address_CFileLoader__LoadObjectInstance_3F087A_thumb = 0;
	}

	static TARGET_THUMB NAKED void patch_CFileLoader__LoadObjectInstance_3F0866()
	{
		__asm(
			".thumb\n"

			"MOV R0, R4\n"
			"BLX GetCollisionFileIDfromCollisionModel\n"

			"CBZ R0, 1f\n"
			"LDR R0, [R5]\n"
			"MOV R1, R5\n"
			"LDR R2, [R0,#0x28]\n"
			"MOV R0, SP\n"
			"BLX R2\n"
			"MOV R0, R4\n"
			"BLX GetCollisionFileIDfromCollisionModel\n"
			ASM_JUMP_TO_ADDRESS_STORED_ON_SYMBOL(Address_CFileLoader__LoadObjectInstance_3F087A_thumb)

			// Branches
			"1:\n"	// loc_3F08D6
			ASM_JUMP_TO_ADDRESS_STORED_ON_SYMBOL(Address_CFileLoader__LoadObjectInstance_3F08D6_thumb)
			);
	}

	// patch for 0x3EF576
	extern "C"
	{
		uintptr_t Address_CFileLoader__LoadCollisionFile_3EF580_thumb = 0;
	}

	static TARGET_THUMB NAKED void patch_CFileLoader__LoadCollisionFile_3EF576()
	{
		__asm(
		".thumb\n"
			"LDR R0, [SP,#12]\n"

			"MOV R1, R0\n"	// ID
			"MOV R0, R5\n"	// pColModel
			"BLX SetCollisionFileIDinCollisionModel\n"

			"LDR R0, [R6]\n"
			"LDR R1, [R0,#0x14]\n"
			ASM_JUMP_TO_ADDRESS_STORED_ON_SYMBOL(Address_CFileLoader__LoadCollisionFile_3EF580_thumb)
			);
	}

	// patch for 0x46BA02
	extern "C"
	{
		uintptr_t Address_CFileLoader__LoadCollisionFileFirstTime_46BA0C_thumb = 0;
	}

	static TARGET_THUMB NAKED void patch_CFileLoader__LoadCollisionFileFirstTime_46BA02()
	{
		__asm(
		".thumb\n"
			"LDR R0, [SP,#8]\n"			
			
			"MOV R1, R0\n"	// ID
			"MOV R0, R5\n"	// pColModel
			"BLX SetCollisionFileIDinCollisionModel\n"

			"MOV R1, R5\n"
			"MOV R0, R6\n"
			ASM_JUMP_TO_ADDRESS_STORED_ON_SYMBOL(Address_CFileLoader__LoadCollisionFileFirstTime_46BA0C_thumb)
			);
	}

	// patch for 0x386810
	extern "C"
	{
		uintptr_t Address_CPedModelInfo__CreateHitColModelSkinned_386818_thumb = 0;
	}

	static TARGET_THUMB NAKED void patch_CPedModelInfo__CreateHitColModelSkinned_386810()
	{
		__asm(
			".thumb\n"

			"MOV R1, R4\n"	// ID
			"MOV R0, R5\n"	// pColModel
			"BLX SetCollisionFileIDinCollisionModel\n"

			"LDR R0, [SP,#16]\n"
			"STR R5, [R0,#0x4C]\n"
			ASM_JUMP_TO_ADDRESS_STORED_ON_SYMBOL(Address_CPedModelInfo__CreateHitColModelSkinned_386818_thumb)
			);
	}

	// patch for 0x2E2588
	extern "C"
	{
		uintptr_t Address_CColStore__RemoveCol_2E2596_thumb = 0;
		uintptr_t Address_CColStore__RemoveCol_2E2590_thumb = 0;
	}

	static TARGET_THUMB NAKED void patch_CColStore__RemoveCol_2E2588()
	{
		__asm(
			".thumb\n"

			"PUSH {R0, R1}\n"

			"BLX GetCollisionFileIDfromCollisionModel\n"
			"MOV R2, R0\n"

			"POP {R0, R1}\n"

			"CMP R2, R8\n"
			"BNE 1f\n"
			ASM_JUMP_TO_ADDRESS_STORED_ON_SYMBOL(Address_CColStore__RemoveCol_2E2590_thumb)

			// Branches
			"1:\n"	// loc_2E2596
			ASM_JUMP_TO_ADDRESS_STORED_ON_SYMBOL(Address_CColStore__RemoveCol_2E2596_thumb)
			);
	}
	#endif

	extern "C"
	{
		// Validates ped or vehicle model ID
		bool __cdecl ValidatePedOrVehicleID(int ID)
		{
			if (!PROJECT_USE_DEVELOPMENT_INI && ID > MAX_FILE_ID_UINT16)
			{
				const char* str = "Attempt to define a ped/vehicle ID %d higher than %d (MAX_FILE_ID_UINT16). It's not possible.";

				CGenericLogStorage::SaveFormattedTextLn(str, ID, MAX_FILE_ID_UINT16);
				CGenericLogStorage::SaveWorkBuffer();
				printf_MessageBox(str, ID, MAX_FILE_ID_UINT16);

				return false;
			}

			if (ID >= (int)g_fileIDlimits.CDarkel__CountOfKillableModelIDs)
			{
				const char* str = "Attempt to define a ped/vehicle ID %d exceeding the limit of killable model IDs which is %d\n"
					"Please increase this limit.";				

				CGenericLogStorage::SaveFormattedTextLn(
					str,
					ID,
					g_fileIDlimits.CDarkel__CountOfKillableModelIDs
				);

				CGenericLogStorage::SaveWorkBuffer();

				printf_MessageBox(
					str,
					ID,
					g_fileIDlimits.CDarkel__CountOfKillableModelIDs
				);

				return false;
			}

			return true;
		}
	}

	#ifdef IS_PLATFORM_WIN_X86
	NAKED void patch_CFileLoader__LoadCollisionFile_5B4FBD()
	{
		__asm
		{
			mov     edx, [esp + 40h + 8];

			push edx;
			push edi;
			call SetCollisionFileIDinCollisionModel;
			add esp, 8;			

			push 0x5B4FC4;
			retn;
		}
	}

	NAKED void patch_CObject__Init_59F8BE()
	{
		__asm
		{
			push edi;
			call GetCollisionFileIDfromCollisionModel;
			add esp, 4;

			push eax;

			push 0x59F8C3;
			retn;
		}
	}

	NAKED void patch_CFileLoader__LoadObjectInstance_5383D5()
	{
		__asm
		{
			push eax;
			call GetCollisionFileIDfromCollisionModel;
			add esp, 4;
			mov ebx, eax;
			test ebx, ebx;

			push 0x5383DA;
			retn;
		}
	}

	NAKED void patch_CFileLoader__LoadObjectInstance_5383E8()
	{
		_asm
		{
			push eax;
			push ebx;

			push 0x5383ED;
			retn;
		}
	}

	NAKED void patch_CFileLoader__LoadCollisionFile_538620()
	{
		__asm
		{
			mov eax, [esp + 44h + 0xC];

			push eax;
			push esi;
			call SetCollisionFileIDinCollisionModel;
			add esp, 8;

			push 0x538627;
			retn;
		}
	}

	NAKED void patch_CFileLoader__LoadCollisionFileFirstTime_5B5189()
	{
		__asm
		{
			mov     edx, [esp + 48h + 0xC];
			push edx;
			push esi;
			call SetCollisionFileIDinCollisionModel;
			add esp, 8;

			push 1;
			push esi;
			mov     ecx, edi;

			push 0x5B5195;
			retn;
		}
	}

	NAKED void patch_CPedModelInfo__CreateHitColModelSkinned_4C6F48()
	{
		__asm
		{
			push ecx;

			push 0;
			push esi;
			call SetCollisionFileIDinCollisionModel;
			add esp, 8;

			pop ecx;

			pop edi;

			push 0x4C6F4D;
			retn;
		}
	}

	NAKED void patch_CColStore__RemoveCol_1564EDE()
	{
		__asm
		{
			push ecx;

			push ecx;
			call GetCollisionFileIDfromCollisionModel;
			add esp, 4;

			cmp     eax, ebx;

			pop ecx;
			push 0x1564EE4;
			retn;
		}
	}

	NAKED void patch_CFileLoader__LoadVehicleObject_5B6FCC()
	{
		__asm
		{
			mov     ecx, [esp + 0F8h - 0x94];
			push ecx;
			call ValidatePedOrVehicleID;
			add esp, 4;

			mov     ecx, [esp + 0F8h - 0x94];
			push ecx;

			push 0x5B6FD1;
			retn;
		}
	}

	NAKED void patch_CFileLoader__LoadPedObject_5B74A2()
	{
		__asm
		{
			mov     eax, [esp + 178h - 0x128];
			push eax;
			call ValidatePedOrVehicleID;
			add esp, 4;

			mov     eax, [esp + 178h - 0x128];
			push eax;

			push 0x5B74A7;
			retn;
		}
	}

	// Pickups
	// Validates vehicle model ID
	bool __cdecl ValidatePickupModelID(int ID)
	{
		if (ID > MAX_FILE_ID_UINT16)
		{
			printf_MessageBox(
				"Attempt to define a pickup with model ID %d higher than %d (MAX_FILE_ID_UINT16). It's not possible.",
				ID,
				MAX_FILE_ID_UINT16
			);

			return false;
		}

		return true;
	}

	NAKED void patch_CPickups__GenerateNewOne_457290()
	{
		__asm
		{
			mov     eax, [esp + 10h + 10h];	// modelID

			push eax;
			call ValidatePickupModelID;
			add esp, 4;

			mov     ecx, [esp + 10h + 24h];	// message
			mov     eax, [esp + 10h + 10h];	// modelID
			push 0x457299;
			retn;
		}
	}

	// IPL ID
	
	// original: movzx   edx, [esi+CEntity.iplFile]
	NAKED void patch_CIplStore__RemoveIpl_404B85()
	{
		__asm
		{
			lea eax, [esi + 0x2E];	// IPL_ID
			push eax;
			call GetExtendedIDfrom8bitBefore;
			add esp, 4;

			mov edx, eax;		

			// FOR_DEBUG
			// movzx edx, byte ptr[eax + 0x2E];
			
			cmp     edx, [esp + 1Ch + 4];

			mov     eax, [esp + 1Ch - 0xC];
			mov     ecx, [esp + 1Ch - 8];	// CPools::ms_pBuildingPool

			push 0x404B8D;
			retn;
		}
	}

	// original: movzx   edx, [esi+CEntity.iplFile]
	NAKED void patch_CIplStore__RemoveIpl_404BCF()
	{
		__asm
		{
			lea eax, [esi + 0x2E];	// IPL_ID
			push eax;
			call GetExtendedIDfrom8bitBefore;
			add esp, 4;

			mov edx, eax;

			// FOR_DEBUG
			// movzx edx, byte ptr[eax + 0x2E];

			cmp     edx, [esp + 1Ch + 4];

			mov     eax, [esp + 1Ch - 0xC];

			push 0x404BD7;
			retn;
		}
	}

	// original: movzx   edx, [esi+CEntity.iplFile]
	NAKED void patch_CIplStore__RemoveIpl_404C33()
	{
		__asm
		{			
			lea eax, [esi + 0x2E];	// IPL_ID
			push eax;
			call GetExtendedIDfrom8bitBefore;
			add esp, 4;

			mov edx, eax;

			// FOR_DEBUG
			// movzx edx, byte ptr[eax + 0x2E];

			cmp     edx, [esp + 1Ch + 4];

			push 0x404C3B;
			retn;
		}
	}

	NAKED void patch_CEntity__CEntity_532AB9()
	{
		__asm
		{
			mov[esi + 0x2F], bl;	// m_nbInterior

			push ebx;	// new value
			lea eax, [esi + 0x2E];	// IPL_ID
			push eax;
			call SetExtendedIDfrom8bitBefore;
			add esp, 8;

			push 0x532ABF;
			retn;
		}
	}

	NAKED void patch_CIplStore__LoadIpl_4061ED()
	{
		__asm
		{
			mov esi, eax;
			add esp, 8;

			//////
			mov     edx, [esp + 18h + 4];	// IPL ID
			
			push edx;	// new value
			lea eax, [esi + 0x2E];	// IPL_ID
			push eax;
			call SetExtendedIDfrom8bitBefore;
			add esp, 8;

			//////

			mov     eax, [esi + 0x30];
			cmp     eax, 0FFFFFFFFh;

			push 0x4061FF;
			retn;
		}
	}

	NAKED void patch_CIplStore__LoadIpl_4062F9()
	{
		__asm
		{
			mov esi, eax;
			add esp, 4;

			//////
			// FOR_DEBUG
			mov[esi + 0x2E], bl;
			
			push ebx;	// new value
			lea eax, [esi + 0x2E];	// IPL_ID
			push eax;
			call SetExtendedIDfrom8bitBefore;
			add esp, 8;
			//////

			mov     eax, [esi + 0x30];
			cmp     eax, 0FFFFFFFFh;

			push 0x406305;
			retn;
		}
	}

	NAKED void patch_CIplStore__LoadIplBoundingBox_405E0F()
	{
		__asm
		{
			mov esi, eax;
			add esp, 4;

			//////
			mov     ecx, [esp + 24h + 4];
						
			push ecx;	// new value
			lea eax, [esi + 0x2E];	// IPL_ID
			push eax;
			call SetExtendedIDfrom8bitBefore;
			add esp, 8;
			//////

			mov     eax, [esi + 0x30];
			cmp     eax, 0FFFFFFFFh;

			push 0x405E21;
			retn;
		}
	}

	NAKED void patch_CIplStore__LoadIplBoundingBox_405C9E()
	{
		__asm
		{
			mov esi, eax;
			add esp, 8;

			//////
			mov     ecx, [esp + 28h + 4];
			
			push ecx;	// new value
			lea eax, [esi + 0x2E];	// IPL_ID
			push eax;
			call SetExtendedIDfrom8bitBefore;
			add esp, 8;
			//////

			mov     eax, [esi + 0x30];
			cmp     eax, 0FFFFFFFFh;

			push 0x405CB0;
			retn;
		}
	}

	NAKED void patch_CEntity__RegisterReference_571B85()
	{
		__asm
		{		
			lea eax, [ebp + 0x2E];	// IPL_ID
			push eax;
			call GetExtendedIDfrom8bitBefore;
			add esp, 4;

			test eax, eax;

			push 0x571B8A;
			retn;
		}
	}

	// original: mov     al, [edi+2Eh]
	NAKED void patch_CDummyObject__UpdateFromObject_59EBCA()
	{
		__asm
		{
			lea eax, [edi + 0x2E];	// IPL_ID
			push eax;
			call GetExtendedIDfrom8bitBefore;
			add esp, 4;

			test eax, eax;

			push 0x59EBCF;
			retn;
		}
	}

	// CObject::CObject(CDummyObject *)
	NAKED void patch_CObject__CObject_5A1E77()
	{
		__asm
		{
			lea eax, [edi + 0x2E];	// IPL_ID
			push eax;
			call GetExtendedIDfrom8bitBefore;
			add esp, 4;

			push eax;	// new value
			lea eax, [esi + 0x2E];	// IPL_ID
			push eax;
			call SetExtendedIDfrom8bitBefore;
			add esp, 8;

			push 0x5A1E7D;
			retn;
		}
	}

	// CDummyObject::CDummyObject
	NAKED void patch_CDummyObject__CDummyObject_59EA79()
	{
		__asm
		{
			lea eax, [edi + 0x2E];	// IPL_ID
			push eax;
			call GetExtendedIDfrom8bitBefore;
			add esp, 4;

			push eax;	// save IPL on stack

			push eax;	// new value
			lea eax, [esi + 0x2E];	// IPL_ID
			push eax;
			call SetExtendedIDfrom8bitBefore;
			add esp, 8;

			pop ecx;	// restore IPL from stack

			push 0x59EA82;
			retn;			
		}
	}


	/////////////////////////
	void NAKED patch_CAEVehicleAudioEntity__GetVehicleAudioSettings_4F5C10()
	{
		__asm
		{
			mov   eax, [esp + 8];

			push 0x4F5C15;
			retn;
		}
	}

	void NAKED patch_CRunningScript__ProcessCommands2500To2599_47A853()
	{
		__asm
		{
			push eax;
			push ecx;

			push UNDEFINED_FILE_ID_INT32;
			mov ecx, esi;
			call CEntity::SetModelID;

			pop ecx;
			pop eax;

			push 0x47A859;
			retn;
		}
	}

	NAKED void CBaseModelInfo__ClearTexDictionary()
	{
		_asm
		{
			push UNDEFINED_FILE_ID_INT32;
			lea eax, [ecx + 0Ah];
			push eax;
			call SetExtendedIDfrom16bitBefore;
			add esp, 8;
			retn;
		}
	}

	NAKED void patch_CPed__Dress_5E013A()
	{
		__asm
		{
			push ecx;
			push edx;

			push UNDEFINED_FILE_ID_INT32;
			mov ecx, ecx;
			call CEntity::SetModelID;

			pop edx;
			pop ecx;

			push 0x5E0140;
			retn;
		}
	}

	void NAKED patch_CBaseModelInfo__SetTexDictionary_4C4B51()
	{
		__asm
		{
			cmp eax, UNDEFINED_FILE_ID_INT32;
			je whenUndefinedID;

			push eax;
			lea eax, [esi + 0Ah];
			push eax;
			call SetExtendedIDfrom16bitBefore;
			add esp, 8;

			pop edi;
			pop esi;
			retn 4;

		whenUndefinedID:
			push 0x4C4B5B;
			retn;
		}
	}

	void NAKED patch_CBaseModelInfo__SetTexDictionary_4C4B61()
	{
		__asm
		{
			add     esp, 4;

			push eax;
			lea eax, [esi + 0Ah];
			push eax;
			call SetExtendedIDfrom16bitBefore;
			add esp, 8;

			push 0x4C4B68;
			retn;
		}
	}

	void CIplStore__IncludeEntity(int IPLid, CEntity *pEntity)
	{
		CPool_GTASA<CIplFile>* pIPLpool = ACCESS_DEF(g_fileIDlimits.CIplStore__ms_pPool.gta_sa);
		auto pBuildingPool = ACCESS_DEF(g_dynamicLimits.CPools__ms_pBuildingPool.gta_sa);
		auto pDummyPool = ACCESS_DEF(g_dynamicLimits.CPools__ms_pDummyPool.gta_sa);

		CIplFile* pFile = pIPLpool->GetAt(IPLid);

		int32_t entityIndex;

		int type = *(char*)(pEntity + 0x36) & 7;

		if (type == 1)
		{
			entityIndex = ((char *)pEntity - (char *)pBuildingPool->m_pObjects) / 56;

			if (entityIndex < GetExtendedIDfrom16bitBefore(&pFile->m_usBuildingsBegin))
				pFile->m_usBuildingsBegin = entityIndex;

			if (entityIndex > GetExtendedIDfrom16bitBefore(&pFile->m_usBuildingsEnd))
				pFile->m_usBuildingsEnd = entityIndex;
		}
		else if (type == 5)
		{
			entityIndex = ((char *)pEntity - (char *)pDummyPool->m_pObjects) / 56;

			if (entityIndex < GetExtendedIDfrom16bitBefore(&pFile->m_usDummyBegin))
				pFile->m_usDummyBegin = entityIndex;

			if (entityIndex > GetExtendedIDfrom16bitBefore(&pFile->m_usDummyEnd))
				pFile->m_usDummyEnd = entityIndex;
		}
	}


	void CColStore__IncludeModelIndex(int COLid, int32_t modelID)
	{
		CPool_GTASA<CCollisionFile>* pCOLpool = ACCESS_DEF(g_fileIDlimits.CColStore__ms_pColPool.gta_sa);

		CCollisionFile* pFile = pCOLpool->GetAt(COLid);

		if (modelID < GetExtendedIDfrom16bitBefore(&pFile->m_wLowestModelID))
			pFile->m_wLowestModelID = modelID;

		if (modelID > GetExtendedIDfrom16bitBefore(&pFile->m_wHighestModelID))
			pFile->m_wHighestModelID = modelID;
	}

	NAKED void patch_CVehicleModelInfo__CLinkedUpgradeList__FindOtherUpgrade_4C74D7()
	{
		__asm
		{
			mov edx, [esp + 4];
			push 0x4C74E0;
			retn;
		}
	}

	NAKED void patch_CVehicleModelInfo__CLinkedUpgradeList__FindOtherUpgrade_4C74F2()
	{
		__asm
		{
			or eax, UNDEFINED_FILE_ID_INT32;
			retn 4;
		}
	}

	NAKED void patch_CBuilding__ReplaceWithNewModel_403EE0()
	{
		__asm
		{
			mov     edx, [esp + 4 + 8];	// newModelIndex

			push 0x403EE5;
			retn;
		}
	}

	NAKED void patch_CBuilding__ReplaceWithNewModel_403EF0()
	{
		__asm
		{
			mov     eax, [esp + 4 + 8];	// newModelIndex

			push 0x403EF5;
			retn;
		}
	}

	NAKED void patch_CPickups__GivePlayerGoodiesWithPickUpMI_456535()
	{
		__asm
		{
			mov     ecx, [esp + 2Ch + 4];
			push 0x45653A;
			retn;
		}
	}

	NAKED void patch_CFileLoader__LoadTimeObject_5B3F94()
	{
		__asm
		{
			mov ecx, [esp + 4];

			push ecx;
			lea eax, [eax + 2];	// + CModelTimeinfo.pairedModel
			push eax;
			call SetExtendedIDfrom16bitBefore;
			add esp, 8;

			push 0x5B3F9D;
			retn;
		}
	}

	NAKED void patch_CVehicle__SetupRender_6D6521()
	{
		__asm
		{
			lea eax, [esi + 59Ah];	// m_wRemapTxd
			push eax;
			call GetExtendedIDfrom16bitBefore;
			add esp, 4;
			
			test eax, eax;

			push 0x6D652B;
			retn;
		}
	}
	
	NAKED void patch_CModelInfo__GetModelInfoUInt16_4C5A13()
	{
		__asm
		{
			mov edx, [esp];
			
			push edx;		
			push ecx;	// pIndex
			call SetExtendedIDfrom16bitBefore;
			add esp, 8;

			pop ecx;
			retn;
		}
	}

	void patch_CCam__GetBoatLook_L_R_HeightOffset_509CAA()
	{
		__asm
		{
			lea eax, [eax + 22h];
			push eax;
			call GetExtendedIDfrom16bitBefore;
			add esp, 4;

			mov ecx, ASM_ACCESS_LIMIT_VAR_32_BIT(g_fileIDlimits, FileIDlimit, CModelInfo__ms_modelInfoPtrs);

			push 0x509CB5;
			retn;
		}
	}

	// Function: sub_6A2B10
	// Original code:	mov     ax, [esi+22h]
	static NAKED void patch_CAutomobile__Render_6A2E98()
	{
		__asm
		{
			lea eax, [esi + 22h];
			push eax;
			call GetExtendedIDfrom16bitBefore;
			add esp, 12;

			push 0x6A2E9F;
			retn;
		}
	}

	static NAKED void patch_CAutomobile__PreRender_6AC730()
	{
		__asm
		{
			push eax;
			call [GetFullModelIDfrom16bitRegisterBefore];
			add esp, 4;

			mov ecx, ASM_ACCESS_LIMIT_VAR_32_BIT(g_fileIDlimits, FileIDlimit, CModelInfo__ms_modelInfoPtrs);

			mov eax, [ecx + eax * 4];

			ASM_JMP_STATIC_ADDRESS(0x6AC735);
		}
	}

	static NAKED void patch_CAutomobile__ProcessControl_6B2185()
	{
		__asm
		{			
			lea eax, [esi + 22h];
			push eax;
			call GetExtendedIDfrom16bitBefore;
			add esp, 4;

			mov ecx, ASM_ACCESS_LIMIT_VAR_32_BIT(g_fileIDlimits, FileIDlimit, CModelInfo__ms_modelInfoPtrs);
			mov ecx, [ecx + eax * 4];

			ASM_JMP_STATIC_ADDRESS(0x6B218B);
		}
	}
	
	// file binary search by name
	static NAKED void patch_CFileLoader__LoadVehicleObject_5B6FE8()
	{
		__asm
		{
			mov[ebp + 4], eax;
			add esp, 4;

			push eax;	// hash
			push dword ptr[esp + 0B8h - 0x94];// ID
			lea ecx, ASM_GET_OFFSET_OF_LIMIT_VAR(g_fileIDlimits, FileIDlimit, CModelInfo__ms_modelInfoIDhashStore);
			call CIDHashContainerBase::AllocateElement;

			push 0x5B6FEE;
			retn;
		}
	}
	
	static NAKED void patch_CFileLoader__LoadObject_5B3DBD()
	{
		__asm
		{
			mov[esi + 4], eax;


			push eax;	// hash
			push dword ptr[esp + 054h - 0x48];// ID
			lea ecx, ASM_GET_OFFSET_OF_LIMIT_VAR(g_fileIDlimits, FileIDlimit, CModelInfo__ms_modelInfoIDhashStore);
			call CIDHashContainerBase::AllocateElement;

			mov     ecx, esi;

			push 0x5B3DC2;
			retn;
		}
	}
	
	static NAKED void patch_CFileLoader__LoadPedObject_5B74C6()
	{
		__asm
		{
			mov[esi + 4], eax;

			push eax;	// hash
			push dword ptr[esp + 0x140 - 0x128];// ID
			lea ecx, ASM_GET_OFFSET_OF_LIMIT_VAR(g_fileIDlimits, FileIDlimit, CModelInfo__ms_modelInfoIDhashStore);
			call CIDHashContainerBase::AllocateElement;

			mov     ecx, esi;

			push 0x5B74CB;
			retn;
		}
	}

	static NAKED void patch_CFileLoader__LoadTimeObject_5B3F56()
	{
		__asm
		{
			mov[esi + 4], eax;
			add     esp, 4;

			push eax;	// hash
			push dword ptr[esp + 05Ch - 0x50];// ID
			lea ecx, ASM_GET_OFFSET_OF_LIMIT_VAR(g_fileIDlimits, FileIDlimit, CModelInfo__ms_modelInfoIDhashStore);
			call CIDHashContainerBase::AllocateElement;

			push 0x5B3F5C;
			retn;
		}
	}

	static NAKED void patch_CFileLoader__LoadAnimatedClumpObject_5B414C()
	{
		__asm
		{
			mov[ebx + 4], eax;
			add     esp, 8;

			push eax;	// hash
			push dword ptr[esp + 054h - 0x4C];// ID
			lea ecx, ASM_GET_OFFSET_OF_LIMIT_VAR(g_fileIDlimits, FileIDlimit, CModelInfo__ms_modelInfoIDhashStore);
			call CIDHashContainerBase::AllocateElement;

			push 0x5B4152;
			retn;
		}
	}

	static NAKED void patch_CFileLoader__LoadClumpObject_5B408F()
	{
		__asm
		{
			mov[esi + 4], eax;
			add esp, 8;


			push eax;	// hash
			push dword ptr[esp + 03Ch - 0x34];	// ID
			lea ecx, ASM_GET_OFFSET_OF_LIMIT_VAR(g_fileIDlimits, FileIDlimit, CModelInfo__ms_modelInfoIDhashStore);
			call CIDHashContainerBase::AllocateElement;

			push 0x5B4095;
			retn;
		}

	}

	static NAKED void patch_CFileLoader__LoadWeaponObject_5B3FFE()
	{
		__asm
		{
			mov[esi + 4], eax;

			push ecx;

			push eax;	// hash
			push dword ptr[esp + 058h - 0x4C];// ID
			lea ecx, ASM_GET_OFFSET_OF_LIMIT_VAR(g_fileIDlimits, FileIDlimit, CModelInfo__ms_modelInfoIDhashStore);
			call CIDHashContainerBase::AllocateElement;

			pop ecx;

			mov     eax, [esp + 50h - 0x48];

			push 0x5B4005;
			retn;
		}
	}

	static NAKED void patch_CStreaming__RequestSpecialModel_409E71()
	{
		__asm
		{
			mov[ebp + 4], eax;

			push eax;	// newKey
			push esi;	// oldKey
			push dword ptr[esp + 0x20 + 0x4]; // ID
			lea ecx, ASM_GET_OFFSET_OF_LIMIT_VAR(g_fileIDlimits, FileIDlimit, CModelInfo__ms_modelInfoIDhashStore);
			call CIDHashContainerBase::UpdateElement;

			xor     edx, edx;

			push 0x409E76;
			retn;
		}
	}

	// TXD files
	static void AddTxdSlot_registerHash(int32_t TXD_ID)
	{
		FileIDlimit* pIDlimits = &g_fileIDlimits;
		TxdDef* pTXDdef = ACCESS_DEF(pIDlimits->CTxdStore__pTxdPool.gta_sa)->GetAt(TXD_ID);
		uint32_t hash = pTXDdef->hash;
		pIDlimits->CTxdStore__pTxdHashStore.AllocateElement(TXD_ID, hash);		
	}

	static NAKED void patch_CTxdStore__AddTxdSlot_731CC8()
	{
		__asm
		{
			add     eax, edx;

			push eax;

			push eax;
			call AddTxdSlot_registerHash;
			add esp, 4;

			pop eax;

			pop     esi;
			retn;
		}
	}
	
	static NAKED void patch_CTxdStore__RemoveTxdSlot_731CF1()
	{
		__asm
		{
			push eax;
			push ecx;
			
			mov eax, [eax + TxdDef::hash];
			push eax;	// hash
			push esi;	// ID
			lea ecx, ASM_GET_OFFSET_OF_LIMIT_VAR(g_fileIDlimits, FileIDlimit, CTxdStore__pTxdHashStore);
			call CIDHashContainerAlwaysSorted::RemoveElement;

			pop ecx;
			pop eax;

			mov     eax, [eax + TxdDef::dictionary];
			test    eax, eax;
			jz loc_731D06;

			push 0x731CF7;
			retn;

		loc_731D06:
			push 0x731D06;
			retn;
		}
	}
	#endif

	#ifdef IS_PLATFORM_ANDROID_ARM32
	// patch for 0x469A7E
	extern "C"
	{
		uintptr_t Address_CFileLoader__LoadVehicleObject_193F14_arm = 0;
		uintptr_t Address_CFileLoader__LoadVehicleObject_469A88_thumb = 0;
	}

	static TARGET_THUMB NAKED void patch_CFileLoader__LoadVehicleObject_469A7E()
	{
		__asm(
			".thumb\n"
			"LDR R0, [SP,#80]\n"
			"BLX ValidatePedOrVehicleID\n"

			"LDR R0, [SP,#80]\n"
			"BL 1f\n"
			"MOV R4, R0\n"
			"MOV R0, R6\n"
			ASM_JUMP_TO_ADDRESS_STORED_ON_SYMBOL(Address_CFileLoader__LoadVehicleObject_469A88_thumb)

			// Branches
			"1:\n"	// j__ZN10CModelInfo15AddVehicleModelEi
			ASM_JUMP_TO_ADDRESS_STORED_ON_SYMBOL(Address_CFileLoader__LoadVehicleObject_193F14_arm)
			);
	}

	// patch for 0x469EB4
	extern "C"
	{
		uintptr_t Address_CFileLoader__LoadPedObject_1A1440_arm = 0;
		uintptr_t Address_CFileLoader__LoadPedObject_469EBC_thumb = 0;
	}

	static TARGET_THUMB NAKED void patch_CFileLoader__LoadPedObject_469EB4()
	{
		__asm(
			".thumb\n"
			"LDR R0, [SP,#72]\n"
			"BLX ValidatePedOrVehicleID\n"

			"LDR R0, [SP,#72]\n"
			"BL 1f\n"
			"MOV R8, R0\n"
			ASM_JUMP_TO_ADDRESS_STORED_ON_SYMBOL(Address_CFileLoader__LoadPedObject_469EBC_thumb)

			// Branches
			"1:\n"	// j__ZN10CModelInfo11AddPedModelEi
			ASM_JUMP_TO_ADDRESS_STORED_ON_SYMBOL(Address_CFileLoader__LoadPedObject_1A1440_arm)
			);
	}
	#endif

	// CModelInfo::GetModelInfo using binary search
	static CBaseModelInfo_generic *__cdecl CModelInfo__GetModelInfo(const char *modelName, int *pIndex)
	{
		FileIDlimit* pIDlimits = &g_fileIDlimits;
		uint32_t hash = CKeyGen::GetUppercaseKey(modelName);
		int32_t ID = pIDlimits->CModelInfo__ms_modelInfoIDhashStore.GetIDbyHashBinarySearch(hash);

		if (pIndex)
			*pIndex = ID;

	#if 0
		if (ID == UNDEFINED_FILE_ID_INT32)
		{
			/*
			for (auto it = pIDlimits->CModelInfo__ms_modelInfoIDhashStore.array.begin();
				it != pIDlimits->CModelInfo__ms_modelInfoIDhashStore.array.end();
				++it)
				CGenericLogStorage::SaveFormattedTextLn("Hash: 0x%08X ID: %d", it->hash, it->ID);

			CGenericLogStorage::SaveWorkBuffer();
			*/
					
			CGenericLogStorage::SaveFormattedTextLn("not found: %s hash: 0x%X",
				modelName,
				hash);

			CGenericLogStorage::SaveWorkBuffer();				
		}
	#endif

		return ID != UNDEFINED_FILE_ID_INT32 ? pIDlimits->CModelInfo__ms_modelInfoPtrs.generic[ID] : 0;
	}

	#ifdef IS_PLATFORM_WIN_X86
	static NAKED CBaseModelInfo_generic *__cdecl CModelInfo__GetModelInfoLinearSearch(const char *modelName, int *pIndex)
	{
		__asm
		{
			mov     eax, [esp + 4];
			push esi;

			push 0x4C5945;
			retn;
		}
	}
	#endif

	static CBaseModelInfo_generic *__cdecl CModelInfo__GetModelInfoFromHashKey(uint32_t hash, int *pIndex)
	{
		FileIDlimit* pIDlimits = &g_fileIDlimits;
		int32_t ID = pIDlimits->CModelInfo__ms_modelInfoIDhashStore.GetIDbyHashBinarySearch(hash);

		if (pIndex)
			*pIndex = ID;

		return ID != UNDEFINED_FILE_ID_INT32 ? pIDlimits->CModelInfo__ms_modelInfoPtrs.generic[ID] : 0;
	}

	class CTimeInfo_patch_4C47E0 : public CTimeInfo
	{
	public:
		CTimeInfo* FindOtherTimeModel(const char *modelName)
		{
			int l = strlen(modelName);

			if (l < 3)
				return 0;

			char theModelName[64];

			if (!memcmp(modelName + l - 3, "_dy", 3))
			{
				memcpy(theModelName, modelName, l - 3);
				memcpy(theModelName + l - 3, "_nt", 4);
			}
			else if (!memcmp(modelName + l - 3, "_nt", 3))
			{
				memcpy(theModelName, modelName, l - 3);
				memcpy(theModelName + l - 3, "_dy", 4);
			}
			else
				return 0;


			FileIDlimit* pIDlimits = &g_fileIDlimits;
			uint32_t hash = CKeyGen::GetUppercaseKey(theModelName);
			int32_t ID = pIDlimits->CModelInfo__ms_modelInfoIDhashStore.GetIDbyHashLinearSearch(hash);
			
			if (ID == UNDEFINED_FILE_ID_INT32)
				return 0;

			CBaseModelInfo_generic* pModelInfo = pIDlimits->CModelInfo__ms_modelInfoPtrs.generic[ID];
			
			// GetTimeInfo
			CTimeInfo* pTimeInfo = (*(CTimeInfo*(__thiscall **)(void *))(*(char**)pModelInfo + 20))(pModelInfo);

			if (!pTimeInfo)
				return 0;

			pTimeInfo->m_wOtherTimeModel = ID;

			return pTimeInfo;
		}
	};

	// TXD
	int32_t __cdecl CTxdStore__FindTxdSlot(const char *TXDname)
	{
		uint32_t hash = CKeyGen::GetUppercaseKey(TXDname);
		int32_t ID = g_fileIDlimits.CTxdStore__pTxdHashStore.GetIDbyHashBinarySearch(hash);

		return ID;
	}
}

#ifdef IS_PLATFORM_WIN_X86
// Patches ID limit if neccessary.
void FileIDlimit::PatchFileIDlimits_GTA_SA_PC_1_0_HOODLUM()
{
	using namespace Game_GTASA;

	// Set variables
	CStreamedScriptInfo* CTheScripts__StreamedScriptsArray = (CStreamedScriptInfo*)CTheScripts__StreamedScripts.ptr;
	CStreamedScripts_Footer* CTheScripts__StreamedScriptsFooter =
		(CStreamedScripts_Footer*)(
			CTheScripts__StreamedScripts.ptr + GetFileIDcurrentLimit(FILE_TYPE_SCM) * sizeof(CStreamedScriptInfo));	

	// Patch pointers to CModelInfo::ms_modelInfoPtrs
	#if TRUE
	{
		CPatch::PatchPointer(0x40122A + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; _ZNK7CEntity13GetIsTouchingERK7CVectorf
		CPatch::PatchPointer(0x403DA4 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_403DA0
		CPatch::PatchPointer(0x403ECC + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_403EC0
		CPatch::PatchPointer(0x405CDF + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_405C00
		CPatch::PatchPointer(0x405E50 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_405C00
		CPatch::PatchPointer(0x4063F1 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     esi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_49CCE0
		CPatch::PatchPointer(0x406B11 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_5534B0
		CPatch::PatchPointer(0x407B21 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CStreaming__AreAnimsUsedByRequestedModels2
		CPatch::PatchPointer(0x407B72 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CStreaming__AreAnimsUsedByRequestedModels2
		CPatch::PatchPointer(0x407B92 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CStreaming__AreAnimsUsedByRequestedModels2
		CPatch::PatchPointer(0x4080B4 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; 
		CPatch::PatchPointer(0x4082F4 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_4082F0
		CPatch::PatchPointer(0x4083A4 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_4083A0
		CPatch::PatchPointer(0x408894 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CStreaming__RequestModel
		CPatch::PatchPointer(0x4088A8 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CStreaming__RequestModel
		CPatch::PatchPointer(0x408903 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ebp, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CStreaming__RequestModel
		CPatch::PatchPointer(0x4089CF + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ebx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CStreaming__RemoveModel
		CPatch::PatchPointer(0x408CD0 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; sub_408CB0
		CPatch::PatchPointer(0x408D1D + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; sub_408CB0
		CPatch::PatchPointer(0x408EC0 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CStreaming__GetNextFileOnCd
		CPatch::PatchPointer(0x4093F4 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ebx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_4093B0
		CPatch::PatchPointer(0x4097B3 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; sub_409760
		CPatch::PatchPointer(0x4098A6 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_409760
		CPatch::PatchPointer(0x4098FC + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; sub_409760
		CPatch::PatchPointer(0x40999E + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; sub_409940
		CPatch::PatchPointer(0x409A4B + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; sub_4099E0
		CPatch::PatchPointer(0x409C74 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_409C70
		CPatch::PatchPointer(0x409CF5 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; sub_409C90
		CPatch::PatchPointer(0x409D19 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ebp, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; sub_409D10
		CPatch::PatchPointer(0x409FFD + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; sub_409FF0
		CPatch::PatchPointer(0x40A2B8 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_40A2A0
		CPatch::PatchPointer(0x40A2E0 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_40A2A0
		CPatch::PatchPointer(0x40A308 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_40A2A0
		CPatch::PatchPointer(0x40A330 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_40A2A0
		CPatch::PatchPointer(0x40A358 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_40A2A0
		CPatch::PatchPointer(0x40A380 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_40A2A0
		CPatch::PatchPointer(0x40A5B2 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_40A560
		CPatch::PatchPointer(0x40A6CC + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; sub_40A560
		CPatch::PatchPointer(0x40A731 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; sub_40A560
		CPatch::PatchPointer(0x40A8E5 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_40A560
		CPatch::PatchPointer(0x40AC1C + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_40AA10
		CPatch::PatchPointer(0x40AE8B + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; sub_40AA10
		CPatch::PatchPointer(0x40BB22 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_40BAA0
		CPatch::PatchPointer(0x40BEDD + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_40BE60
		CPatch::PatchPointer(0x40C04B + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_40C020
		CPatch::PatchPointer(0x40C0BC + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_40C020
		CPatch::PatchPointer(0x40C12A + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_40C020
		CPatch::PatchPointer(0x40C300 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ebx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_40C270
		CPatch::PatchPointer(0x40C4C2 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_40C450
		CPatch::PatchPointer(0x40C6FE + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ebp, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CStreaming__ConvertBufferToObject
		CPatch::PatchPointer(0x40CD64 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CStreaming__RequestModelStream
		CPatch::PatchPointer(0x40CE6E + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CStreaming__RequestModelStream
		CPatch::PatchPointer(0x40CE83 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CStreaming__RequestModelStream
		CPatch::PatchPointer(0x40D05D + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; sub_40CFD0
		CPatch::PatchPointer(0x40E207 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; sub_40E170
		CPatch::PatchPointer(0x40E23B + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; sub_40E170
		CPatch::PatchPointer(0x41A8DE + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_41A820
		CPatch::PatchPointer(0x41AA49 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_41A990
		CPatch::PatchPointer(0x41AA79 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_41A990
		CPatch::PatchPointer(0x41AB98 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_41AB20
		CPatch::PatchPointer(0x41D77E + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_41D660
		CPatch::PatchPointer(0x42145A + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; sub_401438
		CPatch::PatchPointer(0x42187B + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; sub_4217C0
		CPatch::PatchPointer(0x424204 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_424200
		CPatch::PatchPointer(0x425467 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_425440
		CPatch::PatchPointer(0x425506 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_425440
		CPatch::PatchPointer(0x425B82 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_425B30
		CPatch::PatchPointer(0x425BAD + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_425B30
		CPatch::PatchPointer(0x425FA9 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_425F70
		CPatch::PatchPointer(0x425FBE + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_425F70
		CPatch::PatchPointer(0x4262BE + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_426220
		CPatch::PatchPointer(0x4262CC + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_426220
		CPatch::PatchPointer(0x426498 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_426350
		CPatch::PatchPointer(0x4264F4 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_426350
		CPatch::PatchPointer(0x426561 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_426350
		CPatch::PatchPointer(0x4265C7 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_426350
		CPatch::PatchPointer(0x426662 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_426350
		CPatch::PatchPointer(0x4266BA + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_426350
		CPatch::PatchPointer(0x426727 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_426350
		CPatch::PatchPointer(0x426790 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_426350
		CPatch::PatchPointer(0x4267F4 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_426350
		CPatch::PatchPointer(0x426806 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_426350
		CPatch::PatchPointer(0x426A2E + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_426970
		CPatch::PatchPointer(0x426C50 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_426BC0
		CPatch::PatchPointer(0x426D6D + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_426BC0
		CPatch::PatchPointer(0x42C48D + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_42C2B0
		CPatch::PatchPointer(0x42C797 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_42C740
		CPatch::PatchPointer(0x42DB81 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_42DAB0
		CPatch::PatchPointer(0x42DC2A + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_42DAB0
		CPatch::PatchPointer(0x42DD34 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_42DAB0
		CPatch::PatchPointer(0x42DD97 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_42DAB0
		CPatch::PatchPointer(0x42DDED + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_42DAB0
		CPatch::PatchPointer(0x42ECA2 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_42EC90
		CPatch::PatchPointer(0x430633 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; sub_403AEE
		CPatch::PatchPointer(0x430735 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_403AEE
		CPatch::PatchPointer(0x430947 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_403AEE
		CPatch::PatchPointer(0x431B46 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_403AEE
		CPatch::PatchPointer(0x431BD2 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_403AEE
		CPatch::PatchPointer(0x431F0E + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_403AEE
		CPatch::PatchPointer(0x436B31 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_436A90
		CPatch::PatchPointer(0x436B42 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_436A90
		CPatch::PatchPointer(0x43772C + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_436A90
		CPatch::PatchPointer(0x43773D + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_436A90
		CPatch::PatchPointer(0x437797 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_436A90
		CPatch::PatchPointer(0x4377A8 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_436A90
		CPatch::PatchPointer(0x4377E7 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_436A90
		CPatch::PatchPointer(0x4377F8 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_436A90
		CPatch::PatchPointer(0x437842 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_436A90
		CPatch::PatchPointer(0x437853 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_436A90
		CPatch::PatchPointer(0x43A162 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; _spawnCarAtPlayerLocation
		CPatch::PatchPointer(0x43A345 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; _spawnCarAtPlayerLocation
		CPatch::PatchPointer(0x43B3E3 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_43B0F0
		CPatch::PatchPointer(0x43D546 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_43D3B0
		CPatch::PatchPointer(0x43D7D3 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_43D7A0
		CPatch::PatchPointer(0x43D883 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_43D7A0
		CPatch::PatchPointer(0x43D89E + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_43D7A0
		CPatch::PatchPointer(0x447EEB + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_447E40
		CPatch::PatchPointer(0x448AF9 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     esi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_448AF0
		CPatch::PatchPointer(0x448EEE + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_448EE0
		CPatch::PatchPointer(0x449C34 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; 
		CPatch::PatchPointer(0x44A52A + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_44A4D0
		CPatch::PatchPointer(0x44B11E + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_44AA50
		CPatch::PatchPointer(0x453173 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_453060
		CPatch::PatchPointer(0x4531AE + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_453060
		CPatch::PatchPointer(0x4531E4 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_453060
		CPatch::PatchPointer(0x453220 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_453060
		CPatch::PatchPointer(0x453490 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_453370
		CPatch::PatchPointer(0x4534CB + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_453370
		CPatch::PatchPointer(0x453501 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_453370
		CPatch::PatchPointer(0x45353D + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_453370
		CPatch::PatchPointer(0x454B1E + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     esi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_454AE0
		CPatch::PatchPointer(0x455C0E + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_455720
		CPatch::PatchPointer(0x456802 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_4567E0
		CPatch::PatchPointer(0x456816 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_4567E0
		CPatch::PatchPointer(0x4613FD + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_461170
		CPatch::PatchPointer(0x461C2D + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; sub_4619C7
		CPatch::PatchPointer(0x4624F5 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; sub_4619C7
		CPatch::PatchPointer(0x4657A2 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_465680
		CPatch::PatchPointer(0x46978D + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ebx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_469390
		CPatch::PatchPointer(0x46A773 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_46A760
		CPatch::PatchPointer(0x46B5F7 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     esi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_46B460
		CPatch::PatchPointer(0x472950 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_472310
		CPatch::PatchPointer(0x474806 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     esi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_474800
		CPatch::PatchPointer(0x474AAD + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_474900
		CPatch::PatchPointer(0x475D91 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_474D49
		CPatch::PatchPointer(0x478505 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_478000
		CPatch::PatchPointer(0x478552 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_478000
		CPatch::PatchPointer(0x4788D7 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_478000
		CPatch::PatchPointer(0x4794DF + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_478000
		CPatch::PatchPointer(0x47AEA3 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_47A760
		CPatch::PatchPointer(0x47F50F + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ebx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_47F370
		CPatch::PatchPointer(0x47FDC0 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_47FA30
		CPatch::PatchPointer(0x47FDFC + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_47FA30
		CPatch::PatchPointer(0x47FE52 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_47FA30
		CPatch::PatchPointer(0x481E29 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_481300
		CPatch::PatchPointer(0x48229F + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_481300
		CPatch::PatchPointer(0x482560 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_481300
		CPatch::PatchPointer(0x4829D6 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_481300
		CPatch::PatchPointer(0x483365 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; sub_481300
		CPatch::PatchPointer(0x48337F + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; sub_481300
		CPatch::PatchPointer(0x484D39 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; sub_483BD0
		CPatch::PatchPointer(0x486B1D + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_486B00
		CPatch::PatchPointer(0x48ACA2 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_48A320
		CPatch::PatchPointer(0x494EDC + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     esi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_49485A
		CPatch::PatchPointer(0x49835E + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_496E00
		CPatch::PatchPointer(0x4983A6 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_496E00
		CPatch::PatchPointer(0x4986FB + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_496E00
		CPatch::PatchPointer(0x499A10 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_4998A0
		CPatch::PatchPointer(0x49B0EC + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ebp, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; sub_49B0E0
		CPatch::PatchPointer(0x49B0F9 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ebx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_49B0E0
		CPatch::PatchPointer(0x49DB82 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_49DAB0
		CPatch::PatchPointer(0x4AF46B + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_4AF460
		CPatch::PatchPointer(0x4B313C + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_4B3120
		CPatch::PatchPointer(0x4B49B3 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_4B4950
		CPatch::PatchPointer(0x4C0C13 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_4C0BD0
		CPatch::PatchPointer(0x4C4860 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     esi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; sub_4C47E0
		CPatch::PatchPointer(0x4C4BED + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; _ZN14CBaseModelInfo11SetColModelEP9CColModelb
		CPatch::PatchPointer(0x4C5538 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4], eax; sub_4C5530
		CPatch::PatchPointer(0x4C5960 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_4C5940
		CPatch::PatchPointer(0x4C5980 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_4C5940
		CPatch::PatchPointer(0x4C59C0 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_4C59B0
		CPatch::PatchPointer(0x4C5A40 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_4C5A20
		CPatch::PatchPointer(0x4C5A75 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; sub_4C5A70
		CPatch::PatchPointer(0x4C5A89 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; sub_4C5A70
		CPatch::PatchPointer(0x4C5AA5 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; sub_4C5AA0
		CPatch::PatchPointer(0x4C5AB9 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; sub_4C5AA0
		CPatch::PatchPointer(0x4C5AD5 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; 
		CPatch::PatchPointer(0x4C5AE9 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; 
		CPatch::PatchPointer(0x4C5B05 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; sub_4C5B00
		CPatch::PatchPointer(0x4C5B19 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; sub_4C5B00
		CPatch::PatchPointer(0x4C5B35 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; sub_4C5B30
		CPatch::PatchPointer(0x4C5B49 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; sub_4C5B30
		CPatch::PatchPointer(0x4C5B65 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; 
		CPatch::PatchPointer(0x4C5B79 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; 
		CPatch::PatchPointer(0x4C5B95 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; 
		CPatch::PatchPointer(0x4C5BA9 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; 
		CPatch::PatchPointer(0x4C5BC5 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; 
		CPatch::PatchPointer(0x4C5BD9 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; 
		CPatch::PatchPointer(0x4C5BF5 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; 
		CPatch::PatchPointer(0x4C5C09 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; 
		CPatch::PatchPointer(0x4C5C25 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; 
		CPatch::PatchPointer(0x4C5C39 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; 
		CPatch::PatchPointer(0x4C5C55 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; 
		CPatch::PatchPointer(0x4C5C69 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; 
		CPatch::PatchPointer(0x4C5C90 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     esi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; 
		CPatch::PatchPointer(0x4C63BB + 1, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     esi, offset _ZN10CModelInfo16ms_modelInfoPtrsE; 
		CPatch::PatchPointer(0x4C6642 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4], esi; sub_4C6620
		CPatch::PatchPointer(0x4C6671 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4], esi; sub_4C6650
		CPatch::PatchPointer(0x4C66A1 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4], esi; sub_4C6680
		CPatch::PatchPointer(0x4C66D1 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4], esi; sub_4C66B0
		CPatch::PatchPointer(0x4C6701 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4], esi; sub_4C66E0
		CPatch::PatchPointer(0x4C6731 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4], esi; sub_4C6710
		CPatch::PatchPointer(0x4C6761 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4], esi; sub_4C6740
		CPatch::PatchPointer(0x4C6795 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4], esi; sub_4C6770
		CPatch::PatchPointer(0x4C67C2 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4], esi; sub_4C67A0
		CPatch::PatchPointer(0x4C681A + 1, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edi, offset _ZN10CModelInfo16ms_modelInfoPtrsE; _ZN10CModelInfo10InitialiseEv
		CPatch::PatchPointer(0x4C85D6 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// cmp     ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4], esi; sub_4C8500
		CPatch::PatchPointer(0x4C89B8 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; sub_4C89B0
		CPatch::PatchPointer(0x4D3D52 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_4D3CC0
		CPatch::PatchPointer(0x4DACAE + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_506ED0
		CPatch::PatchPointer(0x4E4204 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; 
		CPatch::PatchPointer(0x4E58C4 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; 
		CPatch::PatchPointer(0x4E693A + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; sub_4E68D0
		CPatch::PatchPointer(0x4E6970 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; sub_4E68D0
		CPatch::PatchPointer(0x509CAE + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_509CA0
		CPatch::PatchPointer(0x50D59F + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     esi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_50D590
		CPatch::PatchPointer(0x5135A2 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_513510
		CPatch::PatchPointer(0x513634 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_513510
		CPatch::PatchPointer(0x514C80 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_514B80
		CPatch::PatchPointer(0x516E1A + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_516B20
		CPatch::PatchPointer(0x5172F2 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_517130
		CPatch::PatchPointer(0x5180EF + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_517EA0
		CPatch::PatchPointer(0x51D233 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_51D100
		CPatch::PatchPointer(0x52034F + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_520190
		CPatch::PatchPointer(0x532A58 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_532A50
		CPatch::PatchPointer(0x532B7B + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_532B70
		CPatch::PatchPointer(0x532D01 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_532B70
		CPatch::PatchPointer(0x532D44 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_532D40
		CPatch::PatchPointer(0x533162 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_533150
		CPatch::PatchPointer(0x533247 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_533240
		CPatch::PatchPointer(0x533389 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ebp, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_533380
		CPatch::PatchPointer(0x53341D + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_533410
		CPatch::PatchPointer(0x533706 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_533700
		CPatch::PatchPointer(0x5337AA + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ebx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_533790
		CPatch::PatchPointer(0x533BFA + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; _ZN7CEntity14DestroyEffectsEv
		CPatch::PatchPointer(0x533D46 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ebx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_533D30
		CPatch::PatchPointer(0x533F30 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_533ED0
		CPatch::PatchPointer(0x533F86 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_533ED0
		CPatch::PatchPointer(0x533FBE + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_533FB0
		CPatch::PatchPointer(0x533FDF + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_533FB0
		CPatch::PatchPointer(0x5340A5 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_534030
		CPatch::PatchPointer(0x5340D7 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_534030
		CPatch::PatchPointer(0x53412A + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CEntity__GetBoundRect
		CPatch::PatchPointer(0x534254 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_534250
		CPatch::PatchPointer(0x534294 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_534290
		CPatch::PatchPointer(0x5342BD + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_5342B0
		CPatch::PatchPointer(0x5343FA + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_5343F0
		CPatch::PatchPointer(0x53441D + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_5343F0
		CPatch::PatchPointer(0x53443B + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_5343F0
		CPatch::PatchPointer(0x534449 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_5343F0
		CPatch::PatchPointer(0x5344D8 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; _ZNK7CEntity13GetIsTouchingERK7CVectorf
		CPatch::PatchPointer(0x53454A + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_534540
		CPatch::PatchPointer(0x534568 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_534540
		CPatch::PatchPointer(0x5345F2 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_5345D0
		CPatch::PatchPointer(0x534FB3 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_534E90
		CPatch::PatchPointer(0x535326 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_535300
		CPatch::PatchPointer(0x5354C8 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_535450
		CPatch::PatchPointer(0x535FB0 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_535FA0
		CPatch::PatchPointer(0x536C0B + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_536BE0
		CPatch::PatchPointer(0x537155 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_537150
		CPatch::PatchPointer(0x5371F6 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; sub_5371F0
		CPatch::PatchPointer(0x5372D7 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     esi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; sub_5372D0
		CPatch::PatchPointer(0x5373F4 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; 
		CPatch::PatchPointer(0x537457 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; sub_537450
		CPatch::PatchPointer(0x53748D + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; sub_537450
		CPatch::PatchPointer(0x5380B2 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ebp, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_538090
		CPatch::PatchPointer(0x538124 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_538090
		CPatch::PatchPointer(0x5383BB + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_538090
		CPatch::PatchPointer(0x5384FF + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ebx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_538440
		CPatch::PatchPointer(0x53A70D + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_53A570
		CPatch::PatchPointer(0x53C449 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_53C440
		CPatch::PatchPointer(0x5449C4 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_5449B0
		CPatch::PatchPointer(0x546680 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_546670
		CPatch::PatchPointer(0x546D28 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_546D00
		CPatch::PatchPointer(0x547FDB + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_547B80
		CPatch::PatchPointer(0x5483E7 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_5483D0
		CPatch::PatchPointer(0x54A998 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_54A2C6
		CPatch::PatchPointer(0x54BA95 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_54BA60
		CPatch::PatchPointer(0x54C6C2 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_54BA60
		CPatch::PatchPointer(0x54C817 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_54BA60
		CPatch::PatchPointer(0x54D030 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_54CFF0
		CPatch::PatchPointer(0x5537F0 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_553770
		CPatch::PatchPointer(0x553898 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_553770
		CPatch::PatchPointer(0x553A64 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_553A10
		CPatch::PatchPointer(0x553B14 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_553AA0
		CPatch::PatchPointer(0x55423D + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_554230
		CPatch::PatchPointer(0x554463 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; sub_554230
		CPatch::PatchPointer(0x5544BD + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; sub_554230
		CPatch::PatchPointer(0x554663 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_554650
		CPatch::PatchPointer(0x5546B9 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; sub_554650
		CPatch::PatchPointer(0x5546F5 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; sub_554650
		CPatch::PatchPointer(0x5549F3 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_554840
		CPatch::PatchPointer(0x554ED8 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_554EB0
		CPatch::PatchPointer(0x554F39 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_554EB0
		CPatch::PatchPointer(0x556768 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_556760
		CPatch::PatchPointer(0x5569F7 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_5569C0
		CPatch::PatchPointer(0x558164 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_557530
		CPatch::PatchPointer(0x558406 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_557530
		CPatch::PatchPointer(0x5584CA + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_557530
		CPatch::PatchPointer(0x558688 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_557530
		CPatch::PatchPointer(0x5588E0 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_557530
		CPatch::PatchPointer(0x558961 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_557530
		CPatch::PatchPointer(0x5589D6 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_557530
		CPatch::PatchPointer(0x558A60 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_557530
		CPatch::PatchPointer(0x558B4E + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_557530
		CPatch::PatchPointer(0x55C011 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_55BF20
		CPatch::PatchPointer(0x56508C + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_565000
		CPatch::PatchPointer(0x56512C + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     esi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; sub_5650E0
		CPatch::PatchPointer(0x565266 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_565200
		CPatch::PatchPointer(0x56537E + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     esi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; sub_565300
		CPatch::PatchPointer(0x5662D2 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_566147
		CPatch::PatchPointer(0x56709E + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_566EE0
		CPatch::PatchPointer(0x5683CD + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_567750
		CPatch::PatchPointer(0x569999 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ebx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; sub_569850
		CPatch::PatchPointer(0x569BB3 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_569850
		CPatch::PatchPointer(0x569BDE + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; sub_569850
		CPatch::PatchPointer(0x569CBB + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_569850
		CPatch::PatchPointer(0x569DC6 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_569850
		CPatch::PatchPointer(0x571EAA + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_571EA0
		CPatch::PatchPointer(0x57206F + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_5720A0
		CPatch::PatchPointer(0x585EC7 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_585BF0
		CPatch::PatchPointer(0x58D80B + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_58D7D0
		CPatch::PatchPointer(0x58E855 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_58E020
		CPatch::PatchPointer(0x593508 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; _ZN10Interior_c11PlaceObjectEhP11Furniture_cffff
		CPatch::PatchPointer(0x594CCE + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; sub_594C10
		CPatch::PatchPointer(0x594FC0 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_594E90
		CPatch::PatchPointer(0x595390 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_595380
		CPatch::PatchPointer(0x596756 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_5965E0
		CPatch::PatchPointer(0x59843C + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ebp, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_598430
		CPatch::PatchPointer(0x5986A5 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_598690
		CPatch::PatchPointer(0x598DFC + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_598D80
		CPatch::PatchPointer(0x59E7FE + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_59E750
		CPatch::PatchPointer(0x59E826 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_59E750
		CPatch::PatchPointer(0x59EF37 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_59EEB7
		CPatch::PatchPointer(0x59EFB1 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; sub_59EEB7
		CPatch::PatchPointer(0x59F1B4 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_59F180
		CPatch::PatchPointer(0x59F33B + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_59F330
		CPatch::PatchPointer(0x59F6F2 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_59F660
		CPatch::PatchPointer(0x59F74D + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_59F660
		CPatch::PatchPointer(0x59F8AA + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_59F840
		CPatch::PatchPointer(0x59F8E0 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_59F840
		CPatch::PatchPointer(0x59F9B9 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_59F840
		CPatch::PatchPointer(0x59FA3D + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_59F840
		CPatch::PatchPointer(0x59FA53 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_59F840
		CPatch::PatchPointer(0x59FB5A + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_59FB50
		CPatch::PatchPointer(0x5A04A2 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_5A02E0
		CPatch::PatchPointer(0x5A06D4 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_5A02E0
		CPatch::PatchPointer(0x5A1B84 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     esi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_5A1B60
		CPatch::PatchPointer(0x5A1BB5 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_5A1B60
		CPatch::PatchPointer(0x5A213D + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_5A2130
		CPatch::PatchPointer(0x5A2D06 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; sub_5A2D00
		CPatch::PatchPointer(0x5A2DEA + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; sub_5A2D00
		CPatch::PatchPointer(0x5A332F + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_5A32D0
		CPatch::PatchPointer(0x5A3525 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_5A32D0
		CPatch::PatchPointer(0x5A41C6 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; sub_5A41C0
		CPatch::PatchPointer(0x5A81EB + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     esi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_5A81E0
		CPatch::PatchPointer(0x5B02D0 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; sub_5B02A0
		CPatch::PatchPointer(0x5B02EF + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; sub_5B02A0
		CPatch::PatchPointer(0x5B0610 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; sub_5B05A0
		CPatch::PatchPointer(0x5B062B + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ebp, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; sub_5B05A0
		CPatch::PatchPointer(0x5B1AE5 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_5B1AB0
		CPatch::PatchPointer(0x5B1B69 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; sub_5B1B20
		CPatch::PatchPointer(0x5B1B7C + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; sub_5B1B20
		CPatch::PatchPointer(0x5B2D09 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_5B2CC0
		CPatch::PatchPointer(0x5B2F7A + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_5B2F10
		CPatch::PatchPointer(0x5B2F9F + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_5B2F10
		CPatch::PatchPointer(0x5B2FCA + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     esi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_5B2F10
		CPatch::PatchPointer(0x5B2FD1 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_5B2F10
		CPatch::PatchPointer(0x5B3088 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_5B3040
		CPatch::PatchPointer(0x5B30D0 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; sub_5B3040
		CPatch::PatchPointer(0x5B30D7 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     esi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; sub_5B3040
		CPatch::PatchPointer(0x5B50A1 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_5B5000
		CPatch::PatchPointer(0x5B5270 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_5B51E0
		CPatch::PatchPointer(0x5B529D + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ebp, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_5B51E0
		CPatch::PatchPointer(0x5B52AA + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_5B51E0
		CPatch::PatchPointer(0x5B76BD + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_5B7670
		CPatch::PatchPointer(0x5B8C18 + 1, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edi, offset _ZN10CModelInfo16ms_modelInfoPtrsE; sub_5B8AD0
		CPatch::PatchPointer(0x5B9299 + 1, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     esi, offset _ZN10CModelInfo16ms_modelInfoPtrsE; sub_5B9030
		CPatch::PatchPointer(0x5BEC46 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_5BE670
		CPatch::PatchPointer(0x5C010B + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; _ZN19FurnitureSubGroup_c12AddFurnitureEtshhh
		CPatch::PatchPointer(0x5C0157 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; _ZN19FurnitureSubGroup_c12AddFurnitureEtshhh
		CPatch::PatchPointer(0x5D2A8F + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; 
		CPatch::PatchPointer(0x5DC7F7 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_5DC510
		CPatch::PatchPointer(0x5DF949 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_5DF910
		CPatch::PatchPointer(0x5DFD98 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_5DFD90
		CPatch::PatchPointer(0x5E255F + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_5E2530
		CPatch::PatchPointer(0x5E39E3 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_5E3990
		CPatch::PatchPointer(0x5E3A9D + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_5E3A90
		CPatch::PatchPointer(0x5E48B2 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_5E4880
		CPatch::PatchPointer(0x5E5F07 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; sub_5E5ED0
		CPatch::PatchPointer(0x5E72F2 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     esi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_5E65A0
		CPatch::PatchPointer(0x5E8EF9 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_5E8CD0
		CPatch::PatchPointer(0x5E8F0A + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_5E8CD0
		CPatch::PatchPointer(0x5F201F + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_403AA0
		CPatch::PatchPointer(0x5F2F20 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_5F2F00
		CPatch::PatchPointer(0x5F336B + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_5F32D0
		CPatch::PatchPointer(0x5F38B2 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_5F3880
		CPatch::PatchPointer(0x5F39C1 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_5F3970
		CPatch::PatchPointer(0x5F5720 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_5F3DD0
		CPatch::PatchPointer(0x5F5733 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_5F3DD0
		CPatch::PatchPointer(0x5FCADE + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_5FD810
		CPatch::PatchPointer(0x5FCBB9 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_5FD810
		CPatch::PatchPointer(0x5FD093 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; sub_5FCE80
		CPatch::PatchPointer(0x5FD546 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; sub_5FD330
		CPatch::PatchPointer(0x60039A + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ebx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_6002F0
		CPatch::PatchPointer(0x603509 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ebp, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_6034B0
		CPatch::PatchPointer(0x605474 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_605420
		CPatch::PatchPointer(0x606A5B + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_606890
		CPatch::PatchPointer(0x609A2C + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_6098F0
		CPatch::PatchPointer(0x60FFB5 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_60FBD0
		CPatch::PatchPointer(0x6100EF + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_60FFD0
		CPatch::PatchPointer(0x610154 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_610150
		CPatch::PatchPointer(0x610738 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_610720
		CPatch::PatchPointer(0x6110C4 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; 
		CPatch::PatchPointer(0x6110F3 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; sub_6110E0
		CPatch::PatchPointer(0x611206 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; sub_6110E0
		CPatch::PatchPointer(0x611454 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; 
		CPatch::PatchPointer(0x611474 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; 
		CPatch::PatchPointer(0x611494 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; 
		CPatch::PatchPointer(0x611764 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_611760
		CPatch::PatchPointer(0x611794 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; 
		CPatch::PatchPointer(0x6117B4 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; 
		CPatch::PatchPointer(0x611CC7 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_611C50
		CPatch::PatchPointer(0x611D04 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_611C50
		CPatch::PatchPointer(0x611D65 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; sub_611C50
		CPatch::PatchPointer(0x611D92 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; sub_611C50
		CPatch::PatchPointer(0x611DD0 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; sub_611C50
		CPatch::PatchPointer(0x611E49 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ebp, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; sub_611E10
		CPatch::PatchPointer(0x611E53 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ebx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; sub_611E10
		CPatch::PatchPointer(0x611ED4 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ebx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_611E90
		CPatch::PatchPointer(0x611EEF + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_611E90
		CPatch::PatchPointer(0x61235C + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_612320
		CPatch::PatchPointer(0x6123AF + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_6123A0
		CPatch::PatchPointer(0x61259A + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_6123A0
		CPatch::PatchPointer(0x612EEA + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_612CD0
		CPatch::PatchPointer(0x61300B + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; sub_612F90
		CPatch::PatchPointer(0x61304C + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; sub_612F90
		CPatch::PatchPointer(0x613065 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; sub_612F90
		CPatch::PatchPointer(0x61307E + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; sub_612F90
		CPatch::PatchPointer(0x613093 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; sub_612F90
		CPatch::PatchPointer(0x6130AB + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; sub_612F90
		CPatch::PatchPointer(0x6130DA + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; sub_612F90
		CPatch::PatchPointer(0x6130F3 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; sub_612F90
		CPatch::PatchPointer(0x6131F2 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_613180
		CPatch::PatchPointer(0x61321D + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_613180
		CPatch::PatchPointer(0x613227 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_613180
		CPatch::PatchPointer(0x61326D + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_613260
		CPatch::PatchPointer(0x6132B1 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; sub_613260
		CPatch::PatchPointer(0x61331D + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; sub_613260
		CPatch::PatchPointer(0x61333F + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; sub_613260
		CPatch::PatchPointer(0x613B43 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_613A00
		CPatch::PatchPointer(0x613B5A + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_613A00
		CPatch::PatchPointer(0x613CF7 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; 
		CPatch::PatchPointer(0x613E15 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; 
		CPatch::PatchPointer(0x613E9B + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; 
		CPatch::PatchPointer(0x613EED + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; 
		CPatch::PatchPointer(0x61417E + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; 
		CPatch::PatchPointer(0x6142E8 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_614210
		CPatch::PatchPointer(0x61451B + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_6144B0
		CPatch::PatchPointer(0x61454B + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_6144B0
		CPatch::PatchPointer(0x6145DD + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_614580
		CPatch::PatchPointer(0x6146AD + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_614670
		CPatch::PatchPointer(0x6146D5 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_614670
		CPatch::PatchPointer(0x614A90 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; sub_614720
		CPatch::PatchPointer(0x614C6D + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; sub_614720
		CPatch::PatchPointer(0x614CF4 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; sub_614720
		CPatch::PatchPointer(0x614D42 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; sub_614720
		CPatch::PatchPointer(0x614FE9 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_614720
		CPatch::PatchPointer(0x61511B + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_614720
		CPatch::PatchPointer(0x6152C3 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_614720
		CPatch::PatchPointer(0x616A9D + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_616A40
		CPatch::PatchPointer(0x616AAE + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_616A40
		CPatch::PatchPointer(0x616AE8 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_616A40
		CPatch::PatchPointer(0x616AF9 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_616A40
		CPatch::PatchPointer(0x62428B + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_6240B0
		CPatch::PatchPointer(0x62439A + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_6240B0
		CPatch::PatchPointer(0x627D2B + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_627CC0
		CPatch::PatchPointer(0x628A3B + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_6289F0
		CPatch::PatchPointer(0x6291BB + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_629080
		CPatch::PatchPointer(0x64087E + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_640730
		CPatch::PatchPointer(0x640B3B + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_6408D0
		CPatch::PatchPointer(0x642431 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_642390
		CPatch::PatchPointer(0x642C41 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_642AE0
		CPatch::PatchPointer(0x6432BF + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_6432A0
		CPatch::PatchPointer(0x643357 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_6432A0
		CPatch::PatchPointer(0x64916D + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_649160
		CPatch::PatchPointer(0x64D04D + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_64CF40
		CPatch::PatchPointer(0x64E74F + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_64E740
		CPatch::PatchPointer(0x64FC21 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_64FC10
		CPatch::PatchPointer(0x6504DE + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_6504C0
		CPatch::PatchPointer(0x6509BC + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_6509B0
		CPatch::PatchPointer(0x650E09 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_650BB0
		CPatch::PatchPointer(0x656573 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_6564A0
		CPatch::PatchPointer(0x65D35A + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_402333
		CPatch::PatchPointer(0x6641B5 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_6641A0
		CPatch::PatchPointer(0x66477A + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_664770
		CPatch::PatchPointer(0x664BF2 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_664BA0
		CPatch::PatchPointer(0x66F4D1 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_66F4C0
		CPatch::PatchPointer(0x66F92A + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_66F6C0
		CPatch::PatchPointer(0x66F99E + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_66F6C0
		CPatch::PatchPointer(0x66FA11 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_66F6C0
		CPatch::PatchPointer(0x66FA85 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_66F6C0
		CPatch::PatchPointer(0x66FAF9 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_66F6C0
		CPatch::PatchPointer(0x66FB6D + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_66F6C0
		CPatch::PatchPointer(0x66FBE1 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_66F6C0
		CPatch::PatchPointer(0x66FC55 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_66F6C0
		CPatch::PatchPointer(0x67C82F + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_67C770
		CPatch::PatchPointer(0x67C83D + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_67C770
		CPatch::PatchPointer(0x67D5C6 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_67D590
		CPatch::PatchPointer(0x67DFCE + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_67DE15
		CPatch::PatchPointer(0x68D9F5 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_68D911
		CPatch::PatchPointer(0x69368A + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_693610
		CPatch::PatchPointer(0x69598C + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_6958F0
		CPatch::PatchPointer(0x6969C5 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_696940
		CPatch::PatchPointer(0x696AF3 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_696940
		CPatch::PatchPointer(0x69912E + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ebx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_699120
		CPatch::PatchPointer(0x699268 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_699230
		CPatch::PatchPointer(0x6995EC + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_699570
		CPatch::PatchPointer(0x699701 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_699570
		CPatch::PatchPointer(0x69A77E + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_69A770
		CPatch::PatchPointer(0x69B909 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_69B860
		CPatch::PatchPointer(0x6A07AB + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_6A07A0
		CPatch::PatchPointer(0x6A1577 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_6A1460
		CPatch::PatchPointer(0x6A26C2 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_6A2530
		CPatch::PatchPointer(0x6A2792 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_6A2530
		CPatch::PatchPointer(0x6A2871 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_6A27F0
		CPatch::PatchPointer(0x6A2936 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_6A27F0
		CPatch::PatchPointer(0x6A2F88 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_6A2F80
		CPatch::PatchPointer(0x6A307F + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_6A3060
		CPatch::PatchPointer(0x6A3BDB + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_6A3BD0
		CPatch::PatchPointer(0x6A452A + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_6A4520
		CPatch::PatchPointer(0x6A4805 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_6A47F0
		CPatch::PatchPointer(0x6A609E + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_6A6090
		CPatch::PatchPointer(0x6A65DD + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_6A65D0
		CPatch::PatchPointer(0x6A7178 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_6A7090
		CPatch::PatchPointer(0x6A8B6E + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_406044
		CPatch::PatchPointer(0x6AA2E2 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_6AA291
		CPatch::PatchPointer(0x6ABD16 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_408C8C
		CPatch::PatchPointer(0x6AE28B + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_6AE0D0
		CPatch::PatchPointer(0x6AE2CE + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_6AE0D0
		CPatch::PatchPointer(0x6AE302 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_6AE0D0
		CPatch::PatchPointer(0x6AE332 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_6AE0D0
		CPatch::PatchPointer(0x6AE5D5 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_6AE0D0
		CPatch::PatchPointer(0x6AE60E + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_6AE0D0
		CPatch::PatchPointer(0x6AF1EA + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_6AF1D0
		CPatch::PatchPointer(0x6AF320 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_6AF250
		CPatch::PatchPointer(0x6AF37E + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_6AF250
		CPatch::PatchPointer(0x6B0581 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_6AFB10
		CPatch::PatchPointer(0x6B0B22 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ebp, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_6B0A97
		CPatch::PatchPointer(0x6B1386 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_6B1350
		CPatch::PatchPointer(0x6B14B6 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_6B1480
		CPatch::PatchPointer(0x6B16A7 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_6B1600
		CPatch::PatchPointer(0x6B16EF + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_6B1600
		CPatch::PatchPointer(0x6B3F7D + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_6B3F70
		CPatch::PatchPointer(0x6B4100 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     esi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_6B3F70
		CPatch::PatchPointer(0x6B67EA + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_6B67E0
		CPatch::PatchPointer(0x6B6965 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_6B6950
		CPatch::PatchPointer(0x6B804C + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_6B7F90
		CPatch::PatchPointer(0x6B89BC + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     esi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_6B89B0
		CPatch::PatchPointer(0x6B927D + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_6B9250
		CPatch::PatchPointer(0x6BBC76 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_6B9250
		CPatch::PatchPointer(0x6BC93C + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ebp, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_6BC930
		CPatch::PatchPointer(0x6BD683 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_6BD090
		CPatch::PatchPointer(0x6BE009 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_6BDEA0
		CPatch::PatchPointer(0x6BF48A + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ebp, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_6BF437
		CPatch::PatchPointer(0x6BF857 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; sub_6BF820
		CPatch::PatchPointer(0x6C0B3F + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_6C0810
		CPatch::PatchPointer(0x6C1483 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_6C1470
		CPatch::PatchPointer(0x6C440E + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_6C4400
		CPatch::PatchPointer(0x6C5446 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ebx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_6C5420
		CPatch::PatchPointer(0x6C7E83 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_6C7DE0
		CPatch::PatchPointer(0x6C7FBC + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_6C7FB0
		CPatch::PatchPointer(0x6C8E60 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; sub_6C8E20
		CPatch::PatchPointer(0x6C94D5 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ebp, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_6C94A0
		CPatch::PatchPointer(0x6CE3B9 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ebp, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; sub_6CE377
		CPatch::PatchPointer(0x6CEAF5 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_6CEAD0
		CPatch::PatchPointer(0x6CEEA4 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_6CEEA0
		CPatch::PatchPointer(0x6CF1AA + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_6CF1A0
		CPatch::PatchPointer(0x6CFAE5 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_6CFAC0
		CPatch::PatchPointer(0x6CFD92 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_6CFD60
		CPatch::PatchPointer(0x6D0B79 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_6D0B70
		CPatch::PatchPointer(0x6D0C4C + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_6D0C00
		CPatch::PatchPointer(0x6D0E49 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_6D0E20
		CPatch::PatchPointer(0x6D0E76 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_6D0E60
		CPatch::PatchPointer(0x6D2315 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_6D22F0
		CPatch::PatchPointer(0x6D2B02 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_6D2A80
		CPatch::PatchPointer(0x6D31EC + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_6D30E0
		CPatch::PatchPointer(0x6D32DF + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_6D3210
		CPatch::PatchPointer(0x6D3875 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ebp, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_408121
		CPatch::PatchPointer(0x6D3971 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_408121
		CPatch::PatchPointer(0x6D3A21 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_6D39E0
		CPatch::PatchPointer(0x6D429A + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_6D4290
		CPatch::PatchPointer(0x6D46E9 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; sub_6D46E0
		CPatch::PatchPointer(0x6D5431 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_6D5400
		CPatch::PatchPointer(0x6D63F4 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_6D63F0
		CPatch::PatchPointer(0x6D649D + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_6D6480
		CPatch::PatchPointer(0x6D64D4 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_6D6480
		CPatch::PatchPointer(0x6D64FE + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_6D64F0
		CPatch::PatchPointer(0x6D6A4E + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CVehicle__SetModelIndex
		CPatch::PatchPointer(0x6D7ED5 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_6D7C97
		CPatch::PatchPointer(0x6DB171 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_6DAF2D
		CPatch::PatchPointer(0x6DE2EA + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_6DE240
		CPatch::PatchPointer(0x6DF455 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_6DF3D0
		CPatch::PatchPointer(0x6DF936 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     esi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; sub_6DF930
		CPatch::PatchPointer(0x6DF9C7 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ebx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; sub_6DF930
		CPatch::PatchPointer(0x6DFA2F + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ebp, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_6DFA20
		CPatch::PatchPointer(0x6DFA3A + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_6DFA20
		CPatch::PatchPointer(0x6DFB48 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_6DFA20
		CPatch::PatchPointer(0x6DFB8A + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_6DFB70
		CPatch::PatchPointer(0x6DFBFA + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_6DFBE0
		CPatch::PatchPointer(0x6E0A62 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_6E0A50
		CPatch::PatchPointer(0x6E0E2D + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_6E0E20
		CPatch::PatchPointer(0x6E144B + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_6E1440
		CPatch::PatchPointer(0x6E1607 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_6E1600
		CPatch::PatchPointer(0x6E17B9 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_6E1780
		CPatch::PatchPointer(0x6E1F35 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_6E1A60
		CPatch::PatchPointer(0x6E22C6 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_6E1A60
		CPatch::PatchPointer(0x6E3296 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     esi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; sub_6E3290
		CPatch::PatchPointer(0x6E3312 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_6E3290
		CPatch::PatchPointer(0x6E3396 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; sub_6E3290
		CPatch::PatchPointer(0x6F0D46 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_6F0D00
		CPatch::PatchPointer(0x6F2987 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ebp, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_6F2947
		CPatch::PatchPointer(0x6F2FF1 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_6F2F40
		CPatch::PatchPointer(0x6F32EA + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     esi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_6F32E0
		CPatch::PatchPointer(0x6F3398 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_6F32E0
		CPatch::PatchPointer(0x6F33B2 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_6F32E0
		CPatch::PatchPointer(0x6F3596 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; _ZN13CCarGenerator20DoInternalProcessingEv
		CPatch::PatchPointer(0x6F3753 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; _ZN13CCarGenerator20DoInternalProcessingEv
		CPatch::PatchPointer(0x6F376B + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; _ZN13CCarGenerator20DoInternalProcessingEv
		CPatch::PatchPointer(0x6F3990 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; _ZN13CCarGenerator20DoInternalProcessingEv
		CPatch::PatchPointer(0x6F6073 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ebp, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_6F6030
		CPatch::PatchPointer(0x6F6FBB + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CTrain__RecalcOnRailDistance
		CPatch::PatchPointer(0x6F7735 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CTrain__CreateTrain
		CPatch::PatchPointer(0x6F7761 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CTrain__CreateTrain
		CPatch::PatchPointer(0x6F9132 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; _ZN6CTrain14ProcessControlEv
		CPatch::PatchPointer(0x6F9380 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; _ZN6CTrain14ProcessControlEv
		CPatch::PatchPointer(0x6F9953 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; _ZN6CTrain14ProcessControlEv
		CPatch::PatchPointer(0x6FC805 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_6FC7A0
		CPatch::PatchPointer(0x7086E9 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_7086B0
		CPatch::PatchPointer(0x70A068 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_70A040
		CPatch::PatchPointer(0x70A586 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_70A470
		CPatch::PatchPointer(0x70A73C + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_70A630
		CPatch::PatchPointer(0x70A8C5 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_70A7E0
		CPatch::PatchPointer(0x711BDA + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_711760
		CPatch::PatchPointer(0x711CF6 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_711760
		CPatch::PatchPointer(0x7179E3 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_717970
		CPatch::PatchPointer(0x71AFF4 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_71AFC0
		CPatch::PatchPointer(0x71BC64 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_71BC40
		CPatch::PatchPointer(0x71C0E9 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_71C0D0
		CPatch::PatchPointer(0x71CFF8 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_71CF50
		CPatch::PatchPointer(0x71D032 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_71CF50
		CPatch::PatchPointer(0x71FB44 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_71FAE0
		CPatch::PatchPointer(0x71FB88 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_71FAE0
		CPatch::PatchPointer(0x71FC9A + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_71FAE0
		CPatch::PatchPointer(0x71FD47 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_71FAE0
		CPatch::PatchPointer(0x71FE09 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_71FAE0
		CPatch::PatchPointer(0x71FEDA + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_71FAE0
		CPatch::PatchPointer(0x71FFC2 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_71FAE0
		CPatch::PatchPointer(0x73227A + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_732260
		CPatch::PatchPointer(0x732B55 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_732B40
		CPatch::PatchPointer(0x736F18 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_736A50
		CPatch::PatchPointer(0x738409 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_737C87
		CPatch::PatchPointer(0x73A398 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_73A380
		CPatch::PatchPointer(0x73A3B9 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_73A380
		CPatch::PatchPointer(0x73B525 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; sub_73B4A0
		CPatch::PatchPointer(0x73B537 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; sub_73B4A0
		CPatch::PatchPointer(0x73D5D0 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_73D1E0
		CPatch::PatchPointer(0x73EC5D + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_73EC40
		CPatch::PatchPointer(0x74103B + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_73FB10
		CPatch::PatchPointer(0x1562A35 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_15629D0
		CPatch::PatchPointer(0x1562B94 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_15629D0
		CPatch::PatchPointer(0x1564612 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; sub_15645F0
		CPatch::PatchPointer(0x1564EC2 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; sub_1564E90
		CPatch::PatchPointer(0x156526C + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_1565260
		CPatch::PatchPointer(0x15662F5 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; _ZN10CStreaming31AddToLoadedVehiclesList_HOODLUMEi
		CPatch::PatchPointer(0x156650D + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_15664B0
		CPatch::PatchPointer(0x156658A + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_15664B0
		CPatch::PatchPointer(0x15665C9 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_15664B0
		CPatch::PatchPointer(0x1566F25 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_1566EF0
		CPatch::PatchPointer(0x15691DE + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_15691D0
		CPatch::PatchPointer(0x15691E5 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; sub_15691D0
		CPatch::PatchPointer(0x156C247 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_156C100
		CPatch::PatchPointer(0x156C275 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_156C100
		CPatch::PatchPointer(0x156C5B5 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; sub_156C5B0
		CPatch::PatchPointer(0x156CA5D + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_156CA30
		CPatch::PatchPointer(0x156CAA2 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_156CA30
		CPatch::PatchPointer(0x156D089 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; _ZN10CStreaming14ReInit_HOODLUMEv
		CPatch::PatchPointer(0x156D0B6 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; _ZN10CStreaming14ReInit_HOODLUMEv
		CPatch::PatchPointer(0x156D0E6 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edi, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; _ZN10CStreaming14ReInit_HOODLUMEv
		CPatch::PatchPointer(0x1570248 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_1570230
		CPatch::PatchPointer(0x1570270 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_1570230
		CPatch::PatchPointer(0x1570298 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_1570230
		CPatch::PatchPointer(0x15702C0 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     edx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; sub_1570230
		CPatch::PatchPointer(0x15702E8 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_1570230
		CPatch::PatchPointer(0x1570310 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_1570230
		CPatch::PatchPointer(0x1570404 + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     ecx, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; sub_15703E0
		CPatch::PatchPointer(0x157045D + 3, CModelInfo__ms_modelInfoPtrs.gta_sa);	// mov     eax, ds:_ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; sub_15703E0

		// modelPtrs + 1
		CPatch::PatchPointer(0x409E76 + 1, CModelInfo__ms_modelInfoPtrs.gta_sa + 1);
		CPatch::PatchPointer(0x5B36C8 + 1, CModelInfo__ms_modelInfoPtrs.gta_sa + 1);

		// modelPtrs + countof(modelPtrs)
		CPatch::PatchPointer(0x4C63CE + 2, CModelInfo__ms_modelInfoPtrs.gta_sa + GetBaseID(FILE_TYPE_TXD));
		CPatch::PatchPointer(0x5B92AE + 2, CModelInfo__ms_modelInfoPtrs.gta_sa + GetBaseID(FILE_TYPE_TXD));

		// different
		CPatch::PatchPointer(0x409F0E + 1, CModelInfo__ms_modelInfoPtrs.gta_sa + 1001);
		CPatch::PatchPointer(0x48338D + 2, CModelInfo__ms_modelInfoPtrs.gta_sa + MODEL_MALE01);

		// Initialization of special ped models.
		CPatch::PatchPointer(0x4C687B + 2, CModelInfo__ms_modelInfoPtrs.gta_sa + MODEL_CAR_DOOR);
		CPatch::PatchPointer(0x4C68C8 + 2, CModelInfo__ms_modelInfoPtrs.gta_sa + MODEL_CAR_BUMPER);
		CPatch::PatchPointer(0x4C690C + 2, CModelInfo__ms_modelInfoPtrs.gta_sa + MODEL_CAR_PANEL);
		CPatch::PatchPointer(0x4C694B + 2, CModelInfo__ms_modelInfoPtrs.gta_sa + MODEL_CAR_BONNET);
		CPatch::PatchPointer(0x4C698A + 2, CModelInfo__ms_modelInfoPtrs.gta_sa + MODEL_CAR_BOOT);
		CPatch::PatchPointer(0x4C69C9 + 2, CModelInfo__ms_modelInfoPtrs.gta_sa + MODEL_CAR_WHEEL);
		CPatch::PatchPointer(0x4C6A08 + 2, CModelInfo__ms_modelInfoPtrs.gta_sa + MODEL_BODY_PART_A);
		CPatch::PatchPointer(0x4C6A47 + 2, CModelInfo__ms_modelInfoPtrs.gta_sa + MODEL_BODY_PART_B);

		// CTaskComplexHandsignalAnim::controlSubTask
		CPatch::PatchPointer(0x61B59B + 2, CModelInfo__ms_modelInfoPtrs.gta_sa + MODEL_FHANDL);
		CPatch::PatchPointer(0x61B5CA + 2, CModelInfo__ms_modelInfoPtrs.gta_sa + MODEL_FHANDR);
		CPatch::PatchPointer(0x61B5F1 + 2, CModelInfo__ms_modelInfoPtrs.gta_sa + MODEL_SHANDL);
		CPatch::PatchPointer(0x61B620 + 2, CModelInfo__ms_modelInfoPtrs.gta_sa + MODEL_SHANDR);

		// CTaskComplexHandsignalAnim::~CTaskComplexHandsignalAnim
		CPatch::PatchPointer(0x61BE24 + 2, CModelInfo__ms_modelInfoPtrs.gta_sa + MODEL_FHANDL);
		CPatch::PatchPointer(0x61BE2C + 2, CModelInfo__ms_modelInfoPtrs.gta_sa + MODEL_SHANDL);
		CPatch::PatchPointer(0x61BE40 + 2, CModelInfo__ms_modelInfoPtrs.gta_sa + MODEL_FHANDR);
		CPatch::PatchPointer(0x61BE48 + 2, CModelInfo__ms_modelInfoPtrs.gta_sa + MODEL_SHANDR);
		CPatch::PatchPointer(0x61BE6B + 2, CModelInfo__ms_modelInfoPtrs.gta_sa + MODEL_FHANDL);
		CPatch::PatchPointer(0x61BE78 + 2, CModelInfo__ms_modelInfoPtrs.gta_sa + MODEL_FHANDR);
		CPatch::PatchPointer(0x61BE85 + 1, CModelInfo__ms_modelInfoPtrs.gta_sa + MODEL_SHANDL);
		CPatch::PatchPointer(0x61BE91 + 2, CModelInfo__ms_modelInfoPtrs.gta_sa + MODEL_SHANDR);

		// different
		CPatch::PatchPointer(0x5A55B2 + 1, CModelInfo__ms_modelInfoPtrs.gta_sa + MODEL_CLOTHES01_ID384 - 1);
		CPatch::PatchPointer(0x613B68 + 2, CModelInfo__ms_modelInfoPtrs.gta_sa + MODEL_MALE01);
		CPatch::PatchPointer(0x67B5E5 + 2, CModelInfo__ms_modelInfoPtrs.gta_sa + MODEL_JETPACK);
		CPatch::PatchPointer(0x6AC730 + 1, CModelInfo__ms_modelInfoPtrs.gta_sa + MODEL_FORKLIFT);	// CAutomobile::PreRender
		CPatch::PatchPointer(0x6B2185 + 2, CModelInfo__ms_modelInfoPtrs.gta_sa + MODEL_SKIMMER);	// CAutomobile::ProcessControl
		CPatch::PatchPointer(0x6CC3DC + 1, CModelInfo__ms_modelInfoPtrs.gta_sa + MODEL_HUNTER);	// CPlane::ProcessFlyingCarStuff
		CPatch::PatchPointer(0x612DBC + 1, CModelInfo__ms_modelInfoPtrs.gta_sa + MODEL_MALE01);		
	}
	#endif

	// Patch pointers to CStreaming::ms_aInfoForModel
	#if TRUE
	{
		// Generated automatically, non-modified functions on HOODLUM
		CPatch::PatchPointer(0x405451 + 1, &CStreaming__ms_aInfoForModel.gta_sa[GetBaseID(FILE_TYPE_IPL) + 1].uiLoadFlag);
		CPatch::PatchPointer(0x405509 + 2, &CStreaming__ms_aInfoForModel.gta_sa[GetBaseID(FILE_TYPE_DAT)].uiLoadFlag);
		CPatch::PatchPointer(0x405733 + 1, &CStreaming__ms_aInfoForModel.gta_sa[GetBaseID(FILE_TYPE_IPL) + 1].ucFlags);
		CPatch::PatchPointer(0x405773 + 2, &CStreaming__ms_aInfoForModel.gta_sa[GetBaseID(FILE_TYPE_DAT)].ucFlags);
		CPatch::PatchPointer(0x407C73 + 3, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);
		CPatch::PatchPointer(0x407C84 + 3, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);
		CPatch::PatchPointer(0x407C9E + 3, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);
		CPatch::PatchPointer(0x407CAA + 3, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);
		CPatch::PatchPointer(0x407CDA + 3, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);
		CPatch::PatchPointer(0x407CED + 3, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);
		CPatch::PatchPointer(0x407D62 + 3, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);
		CPatch::PatchPointer(0x407FDD + 3, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);
		CPatch::PatchPointer(0x4089AC + 2, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);
		CPatch::PatchPointer(0x408AC3 + 2, &CStreaming__ms_aInfoForModel.gta_sa->iBlockCount);
		CPatch::PatchPointer(0x408ADA + 3, CStreaming__ms_aInfoForModel.gta_sa);
		CPatch::PatchPointer(0x408AE1 + 2, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);
		CPatch::PatchPointer(0x408AF7 + 2, &CStreaming__ms_aInfoForModel.gta_sa->ucFlags);
		CPatch::PatchPointer(0x408B08 + 2, &CStreaming__ms_aInfoForModel.gta_sa->ucFlags);
		CPatch::PatchPointer(0x408B17 + 2, &CStreaming__ms_aInfoForModel.gta_sa->ucFlags);
		CPatch::PatchPointer(0x408B22 + 3, CStreaming__ms_aInfoForModel.gta_sa);
		CPatch::PatchPointer(0x408B32 + 3, &CStreaming__ms_aInfoForModel.gta_sa->usPrev);
		CPatch::PatchPointer(0x408B3E + 3, &CStreaming__ms_aInfoForModel.gta_sa->usPrev);
		CPatch::PatchPointer(0x408B4E + 3, CStreaming__ms_aInfoForModel.gta_sa);
		CPatch::PatchPointer(0x408B59 + 3, CStreaming__ms_aInfoForModel.gta_sa);
		CPatch::PatchPointer(0x408B60 + 3, &CStreaming__ms_aInfoForModel.gta_sa->usPrev);
		CPatch::PatchPointer(0x408B90 + 2, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);
		CPatch::PatchPointer(0x408BAB + 2, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);
		CPatch::PatchPointer(0x408BCC + 2, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);
		CPatch::PatchPointer(0x408BED + 2, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);
		CPatch::PatchPointer(0x408C0E + 2, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);
		CPatch::PatchPointer(0x408C37 + 2, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);
		CPatch::PatchPointer(0x408C5A + 2, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);
		CPatch::PatchPointer(0x408CBF + 2, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);
		CPatch::PatchPointer(0x408CE4 + 2, &CStreaming__ms_aInfoForModel.gta_sa->iBlockCount);
		CPatch::PatchPointer(0x408DC6 + 2, &CStreaming__ms_aInfoForModel.gta_sa->iBlockCount);
		CPatch::PatchPointer(0x408DDD + 2, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);
		CPatch::PatchPointer(0x408DF3 + 3, &CStreaming__ms_aInfoForModel.gta_sa->ucFlags);
		CPatch::PatchPointer(0x408E65 + 2, CStreaming__ms_aInfoForModel.gta_sa);
		CPatch::PatchPointer(0x408ECE + 4, &CStreaming__ms_aInfoForModel.gta_sa[GetBaseID(FILE_TYPE_TXD)].uiLoadFlag);
		CPatch::PatchPointer(0x408EE3 + 4, &CStreaming__ms_aInfoForModel.gta_sa->ucFlags);
		CPatch::PatchPointer(0x408F17 + 4, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);
		CPatch::PatchPointer(0x408F7F + 4, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);
		CPatch::PatchPointer(0x408FB9 + 2, &CStreaming__ms_aInfoForModel.gta_sa[7].uiLoadFlag);
		CPatch::PatchPointer(0x408FC5 + 4, &CStreaming__ms_aInfoForModel.gta_sa->ucImgId);
		CPatch::PatchPointer(0x408FCD + 3, &CStreaming__ms_aInfoForModel.gta_sa->usNext);
		CPatch::PatchPointer(0x40907A + 3, &CStreaming__ms_aInfoForModel.gta_sa->ucFlags);
		CPatch::PatchPointer(0x4098C8 + 3, CStreaming__ms_aInfoForModel.gta_sa);
		CPatch::PatchPointer(0x4098D6 + 2, &CStreaming__ms_aInfoForModel.gta_sa->ucFlags);
		CPatch::PatchPointer(0x409A39 + 4, CStreaming__ms_aInfoForModel.gta_sa);
		CPatch::PatchPointer(0x409C1B + 2, &CStreaming__ms_aInfoForModel.gta_sa->ucFlags);
		CPatch::PatchPointer(0x409C21 + 2, CStreaming__ms_aInfoForModel.gta_sa);
		CPatch::PatchPointer(0x409C2D + 2, &CStreaming__ms_aInfoForModel.gta_sa->ucFlags);
		CPatch::PatchPointer(0x409C38 + 2, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);
		CPatch::PatchPointer(0x409CA6 + 2, &CStreaming__ms_aInfoForModel.gta_sa->ucFlags);
		CPatch::PatchPointer(0x409CAC + 2, CStreaming__ms_aInfoForModel.gta_sa);
		CPatch::PatchPointer(0x409CB8 + 2, &CStreaming__ms_aInfoForModel.gta_sa->ucFlags);
		CPatch::PatchPointer(0x409CC3 + 2, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);
		CPatch::PatchPointer(0x409D39 + 3, &CStreaming__ms_aInfoForModel.gta_sa->iBlockCount);
		CPatch::PatchPointer(0x409D42 + 3, CStreaming__ms_aInfoForModel.gta_sa);
		CPatch::PatchPointer(0x409FB1 + 2, &CStreaming__ms_aInfoForModel.gta_sa->iBlockOffset);
		CPatch::PatchPointer(0x409FBB + 2, &CStreaming__ms_aInfoForModel.gta_sa->iBlockCount);
		CPatch::PatchPointer(0x409FC4 + 2, &CStreaming__ms_aInfoForModel.gta_sa->ucImgId);
		CPatch::PatchPointer(0x409FCA + 3, &CStreaming__ms_aInfoForModel.gta_sa->usNextOnCd);
		CPatch::PatchPointer(0x40A052 + 2, &CStreaming__ms_aInfoForModel.gta_sa->iBlockOffset);
		CPatch::PatchPointer(0x40A05F + 2, &CStreaming__ms_aInfoForModel.gta_sa->iBlockCount);
		CPatch::PatchPointer(0x40A065 + 2, &CStreaming__ms_aInfoForModel.gta_sa->ucImgId);
		CPatch::PatchPointer(0x40A3D8 + 3, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);
		CPatch::PatchPointer(0x40A3EE + 3, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);
		CPatch::PatchPointer(0x40ABC5 + 2, &CStreaming__ms_aInfoForModel.gta_sa->ucFlags);
		CPatch::PatchPointer(0x40ABCB + 2, CStreaming__ms_aInfoForModel.gta_sa);
		CPatch::PatchPointer(0x40ABD7 + 2, &CStreaming__ms_aInfoForModel.gta_sa->ucFlags);
		CPatch::PatchPointer(0x40ABE2 + 2, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);
		CPatch::PatchPointer(0x40AC33 + 2, &CStreaming__ms_aInfoForModel.gta_sa->ucFlags);
		CPatch::PatchPointer(0x40AC39 + 2, CStreaming__ms_aInfoForModel.gta_sa);
		CPatch::PatchPointer(0x40AC45 + 2, &CStreaming__ms_aInfoForModel.gta_sa->ucFlags);
		CPatch::PatchPointer(0x40AC50 + 2, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);
		CPatch::PatchPointer(0x40ACEB + 3, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);
		CPatch::PatchPointer(0x40ADF9 + 2, &CStreaming__ms_aInfoForModel.gta_sa->ucFlags);
		CPatch::PatchPointer(0x40ADFF + 2, CStreaming__ms_aInfoForModel.gta_sa);
		CPatch::PatchPointer(0x40AE0B + 2, &CStreaming__ms_aInfoForModel.gta_sa->ucFlags);
		CPatch::PatchPointer(0x40AE16 + 2, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);
		CPatch::PatchPointer(0x40AEA2 + 2, &CStreaming__ms_aInfoForModel.gta_sa->ucFlags);
		CPatch::PatchPointer(0x40AEA8 + 2, CStreaming__ms_aInfoForModel.gta_sa);
		CPatch::PatchPointer(0x40AEB4 + 2, &CStreaming__ms_aInfoForModel.gta_sa->ucFlags);
		CPatch::PatchPointer(0x40AEBF + 2, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);
		CPatch::PatchPointer(0x40B747 + 1, &CStreaming__ms_aInfoForModel.gta_sa[427].uiLoadFlag);
		CPatch::PatchPointer(0x40B79C + 1, &CStreaming__ms_aInfoForModel.gta_sa[490].uiLoadFlag);
		CPatch::PatchPointer(0x40B807 + 1, &CStreaming__ms_aInfoForModel.gta_sa[433].uiLoadFlag);
		CPatch::PatchPointer(0x40B813 + 2, &CStreaming__ms_aInfoForModel.gta_sa[432].uiLoadFlag);
		CPatch::PatchPointer(0x40B92C + 2, &CStreaming__ms_aInfoForModel.gta_sa->ucFlags);
		CPatch::PatchPointer(0x40B932 + 2, CStreaming__ms_aInfoForModel.gta_sa);
		CPatch::PatchPointer(0x40B93E + 2, &CStreaming__ms_aInfoForModel.gta_sa->ucFlags);
		CPatch::PatchPointer(0x40B949 + 2, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);
		CPatch::PatchPointer(0x40BAD5 + 2, &CStreaming__ms_aInfoForModel.gta_sa->ucFlags);
		CPatch::PatchPointer(0x40BADB + 2, CStreaming__ms_aInfoForModel.gta_sa);
		CPatch::PatchPointer(0x40BAE6 + 2, &CStreaming__ms_aInfoForModel.gta_sa->ucFlags);
		CPatch::PatchPointer(0x40BAF1 + 2, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);
		CPatch::PatchPointer(0x40BB39 + 2, &CStreaming__ms_aInfoForModel.gta_sa->ucFlags);
		CPatch::PatchPointer(0x40BB3F + 2, CStreaming__ms_aInfoForModel.gta_sa);
		CPatch::PatchPointer(0x40BB4A + 2, &CStreaming__ms_aInfoForModel.gta_sa->ucFlags);
		CPatch::PatchPointer(0x40BB55 + 2, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);
		CPatch::PatchPointer(0x40BE94 + 2, &CStreaming__ms_aInfoForModel.gta_sa->ucFlags);
		CPatch::PatchPointer(0x40BE9A + 2, CStreaming__ms_aInfoForModel.gta_sa);
		CPatch::PatchPointer(0x40BEA5 + 2, &CStreaming__ms_aInfoForModel.gta_sa->ucFlags);
		CPatch::PatchPointer(0x40BEB0 + 2, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);
		CPatch::PatchPointer(0x40BEF4 + 2, &CStreaming__ms_aInfoForModel.gta_sa->ucFlags);
		CPatch::PatchPointer(0x40BEFA + 2, CStreaming__ms_aInfoForModel.gta_sa);
		CPatch::PatchPointer(0x40BF05 + 2, &CStreaming__ms_aInfoForModel.gta_sa->ucFlags);
		CPatch::PatchPointer(0x40BF10 + 2, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);
		CPatch::PatchPointer(0x40C05F + 2, &CStreaming__ms_aInfoForModel.gta_sa->ucFlags);
		CPatch::PatchPointer(0x40C067 + 2, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);
		CPatch::PatchPointer(0x40C0D0 + 2, &CStreaming__ms_aInfoForModel.gta_sa->ucFlags);
		CPatch::PatchPointer(0x40C0D8 + 2, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);
		CPatch::PatchPointer(0x40C13E + 2, &CStreaming__ms_aInfoForModel.gta_sa->ucFlags);
		CPatch::PatchPointer(0x40C146 + 2, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);
		CPatch::PatchPointer(0x40C212 + 2, CStreaming__ms_aInfoForModel.gta_sa);
		CPatch::PatchPointer(0x40C2B9 + 3, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);
		CPatch::PatchPointer(0x40C491 + 3, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);
		CPatch::PatchPointer(0x40C6CC + 2, &CStreaming__ms_aInfoForModel.gta_sa->iBlockCount);
		CPatch::PatchPointer(0x40C81B + 2, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);
		CPatch::PatchPointer(0x40C876 + 3, &CStreaming__ms_aInfoForModel.gta_sa->ucFlags);
		CPatch::PatchPointer(0x40C8E4 + 2, &CStreaming__ms_aInfoForModel.gta_sa->ucFlags);
		CPatch::PatchPointer(0x40C92F + 2, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);
		CPatch::PatchPointer(0x40C94E + 3, &CStreaming__ms_aInfoForModel.gta_sa->ucFlags);
		CPatch::PatchPointer(0x40C99F + 3, &CStreaming__ms_aInfoForModel.gta_sa->ucFlags);
		CPatch::PatchPointer(0x40C9F0 + 3, &CStreaming__ms_aInfoForModel.gta_sa->ucFlags);
		CPatch::PatchPointer(0x40CA3D + 2, &CStreaming__ms_aInfoForModel.gta_sa->ucFlags);
		CPatch::PatchPointer(0x40CAFA + 2, &CStreaming__ms_aInfoForModel.gta_sa->ucFlags);
		CPatch::PatchPointer(0x40CB10 + 2, &CStreaming__ms_aInfoForModel.gta_sa->ucFlags);
		CPatch::PatchPointer(0x40CB18 + 2, CStreaming__ms_aInfoForModel.gta_sa);
		CPatch::PatchPointer(0x40CB49 + 2, &CStreaming__ms_aInfoForModel.gta_sa->ucFlags);
		CPatch::PatchPointer(0x40CB58 + 2, CStreaming__ms_aInfoForModel.gta_sa);
		CPatch::PatchPointer(0x40CB63 + 2, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);
		CPatch::PatchPointer(0x40CB71 + 2, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);
		CPatch::PatchPointer(0x40CB78 + 2, &CStreaming__ms_aInfoForModel.gta_sa->iBlockCount);
		CPatch::PatchPointer(0x40CBD8 + 2, &CStreaming__ms_aInfoForModel.gta_sa->ucFlags);
		CPatch::PatchPointer(0x40CC38 + 2, &CStreaming__ms_aInfoForModel.gta_sa->iBlockCount);
		CPatch::PatchPointer(0x40CC45 + 3, &CStreaming__ms_aInfoForModel.gta_sa->ucImgId);
		CPatch::PatchPointer(0x40CC58 + 2, &CStreaming__ms_aInfoForModel.gta_sa->iBlockOffset);
		CPatch::PatchPointer(0x40CC8E + 2, &CStreaming__ms_aInfoForModel.gta_sa->ucFlags);
		CPatch::PatchPointer(0x40CCA6 + 3, &CStreaming__ms_aInfoForModel.gta_sa->iBlockCount);
		CPatch::PatchPointer(0x40CCAF + 3, &CStreaming__ms_aInfoForModel.gta_sa->usNext);
		CPatch::PatchPointer(0x40CD1F + 2, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);
		CPatch::PatchPointer(0x40CD2C + 2, &CStreaming__ms_aInfoForModel.gta_sa->iBlockCount);
		CPatch::PatchPointer(0x40CD34 + 2, CStreaming__ms_aInfoForModel.gta_sa);
		CPatch::PatchPointer(0x40CD49 + 2, &CStreaming__ms_aInfoForModel.gta_sa->ucFlags);
		CPatch::PatchPointer(0x40CD9F + 3, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);
		CPatch::PatchPointer(0x40CDC6 + 3, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);
		CPatch::PatchPointer(0x40CDF8 + 2, &CStreaming__ms_aInfoForModel.gta_sa[7].uiLoadFlag);
		CPatch::PatchPointer(0x40CEB2 + 2, &CStreaming__ms_aInfoForModel.gta_sa->ucFlags);
		CPatch::PatchPointer(0x40CEDB + 3, &CStreaming__ms_aInfoForModel.gta_sa->usNextOnCd);
		CPatch::PatchPointer(0x40D012 + 2, CStreaming__ms_aInfoForModel.gta_sa);
		CPatch::PatchPointer(0x40D332 + 2, CStreaming__ms_aInfoForModel.gta_sa);
		CPatch::PatchPointer(0x40E1F7 + 2, &CStreaming__ms_aInfoForModel.gta_sa->iBlockCount);
		CPatch::PatchPointer(0x40E232 + 2, &CStreaming__ms_aInfoForModel.gta_sa->ucFlags);
		CPatch::PatchPointer(0x40E250 + 2, &CStreaming__ms_aInfoForModel.gta_sa->ucFlags);
		CPatch::PatchPointer(0x40E2CA + 2, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);
		CPatch::PatchPointer(0x40E52C + 2, CStreaming__ms_aInfoForModel.gta_sa);
		CPatch::PatchPointer(0x40EC02 + 2, CStreaming__ms_aInfoForModel.gta_sa);
		CPatch::PatchPointer(0x40ED3E + 1, &CStreaming__ms_aInfoForModel.gta_sa->ucFlags);
		CPatch::PatchPointer(0x410B31 + 1, &CStreaming__ms_aInfoForModel.gta_sa[GetBaseID(FILE_TYPE_COL) + 1].uiLoadFlag);
		CPatch::PatchPointer(0x410BDE + 2, &CStreaming__ms_aInfoForModel.gta_sa[GetBaseID(FILE_TYPE_IPL)].uiLoadFlag);
		CPatch::PatchPointer(0x410E13 + 1, &CStreaming__ms_aInfoForModel.gta_sa[GetBaseID(FILE_TYPE_COL) + 1].ucFlags);
		CPatch::PatchPointer(0x410E53 + 2, &CStreaming__ms_aInfoForModel.gta_sa[GetBaseID(FILE_TYPE_IPL)].ucFlags);
		CPatch::PatchPointer(0x42194F + 3, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);
		CPatch::PatchPointer(0x421998 + 2, &CStreaming__ms_aInfoForModel.gta_sa[427].uiLoadFlag);
		CPatch::PatchPointer(0x4219A0 + 2, &CStreaming__ms_aInfoForModel.gta_sa[285].uiLoadFlag);
		CPatch::PatchPointer(0x4219D5 + 2, &CStreaming__ms_aInfoForModel.gta_sa[490].uiLoadFlag);
		CPatch::PatchPointer(0x4219DD + 2, &CStreaming__ms_aInfoForModel.gta_sa[286].uiLoadFlag);
		CPatch::PatchPointer(0x421A01 + 2, &CStreaming__ms_aInfoForModel.gta_sa[432].uiLoadFlag);
		CPatch::PatchPointer(0x421A09 + 2, &CStreaming__ms_aInfoForModel.gta_sa[433].uiLoadFlag);
		CPatch::PatchPointer(0x421A11 + 2, &CStreaming__ms_aInfoForModel.gta_sa[287].uiLoadFlag);
		CPatch::PatchPointer(0x424D8E + 3, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);
		CPatch::PatchPointer(0x424F66 + 3, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);
		CPatch::PatchPointer(0x42C334 + 3, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);
		CPatch::PatchPointer(0x43A5B1 + 1, &CStreaming__ms_aInfoForModel.gta_sa[584].uiLoadFlag);
		CPatch::PatchPointer(0x443A68 + 3, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);
		CPatch::PatchPointer(0x445961 + 3, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);
		CPatch::PatchPointer(0x447E98 + 3, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);
		CPatch::PatchPointer(0x44B5DD + 2, &CStreaming__ms_aInfoForModel.gta_sa[364].uiLoadFlag);
		CPatch::PatchPointer(0x4614F0 + 3, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);
		CPatch::PatchPointer(0x46CE45 + 3, &CStreaming__ms_aInfoForModel.gta_sa[GetBaseID(FILE_TYPE_SCM)].uiLoadFlag);
		CPatch::PatchPointer(0x46CFB6 + 3, &CStreaming__ms_aInfoForModel.gta_sa[GetBaseID(FILE_TYPE_SCM)].uiLoadFlag);
		CPatch::PatchPointer(0x47EFFF + 3, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);
		CPatch::PatchPointer(0x4D5C57 + 4, &CStreaming__ms_aInfoForModel.gta_sa->ucFlags);
		CPatch::PatchPointer(0x55497F + 3, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);
		CPatch::PatchPointer(0x56EC4D + 2, &CStreaming__ms_aInfoForModel.gta_sa[371].uiLoadFlag);
		CPatch::PatchPointer(0x5B0025 + 4, &CStreaming__ms_aInfoForModel.gta_sa->ucFlags);
		CPatch::PatchPointer(0x5B11E8 + 3, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);
		CPatch::PatchPointer(0x5B8AE7 + 1, CStreaming__ms_aInfoForModel.gta_sa);
		CPatch::PatchPointer(0x5B8B08 + 6, CStreaming__ms_aInfoForModel.gta_sa);

		CPatch::PatchPointer(0x5B8B86 + 2, &CStreaming__ms_aInfoForModel.gta_sa[374].uiLoadFlag);
		CPatch::PatchPointer(0x5B8B8D + 2, &CStreaming__ms_aInfoForModel.gta_sa[375].uiLoadFlag);
		CPatch::PatchPointer(0x5B8B94 + 2, &CStreaming__ms_aInfoForModel.gta_sa[376].uiLoadFlag);
		CPatch::PatchPointer(0x5B8B9B + 2, &CStreaming__ms_aInfoForModel.gta_sa[377].uiLoadFlag);
		CPatch::PatchPointer(0x5B8BA2 + 2, &CStreaming__ms_aInfoForModel.gta_sa[378].uiLoadFlag);
		CPatch::PatchPointer(0x5B8BA9 + 2, &CStreaming__ms_aInfoForModel.gta_sa[379].uiLoadFlag);
		CPatch::PatchPointer(0x5B8BB0 + 2, &CStreaming__ms_aInfoForModel.gta_sa[380].uiLoadFlag);
		CPatch::PatchPointer(0x5B8BB7 + 2, &CStreaming__ms_aInfoForModel.gta_sa[381].uiLoadFlag);
		CPatch::PatchPointer(0x5B8C1D + 1, &CStreaming__ms_aInfoForModel.gta_sa->ucFlags);
		CPatch::PatchPointer(0x5B8C4F + 2, &CStreaming__ms_aInfoForModel.gta_sa[GetBaseID(FILE_TYPE_TXD)].ucFlags);
		CPatch::PatchPointer(0x5B8C63 + 1, &CStreaming__ms_aInfoForModel.gta_sa[GetBaseID(FILE_TYPE_TXD) + 1].uiLoadFlag);
		CPatch::PatchPointer(0x5DE619 + 3, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);
		CPatch::PatchPointer(0x5E896B + 3, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);
		CPatch::PatchPointer(0x5E897E + 3, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);
		CPatch::PatchPointer(0x60FED6 + 3, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);
		CPatch::PatchPointer(0x61329E + 3, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);
		CPatch::PatchPointer(0x613A5A + 3, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);
		CPatch::PatchPointer(0x61B58E + 2, &CStreaming__ms_aInfoForModel.gta_sa[396].uiLoadFlag);
		CPatch::PatchPointer(0x61B5BB + 2, &CStreaming__ms_aInfoForModel.gta_sa[397].uiLoadFlag);
		CPatch::PatchPointer(0x61B5E4 + 2, &CStreaming__ms_aInfoForModel.gta_sa[394].uiLoadFlag);
		CPatch::PatchPointer(0x61B611 + 2, &CStreaming__ms_aInfoForModel.gta_sa[395].uiLoadFlag);
		CPatch::PatchPointer(0x61B66A + 3, &CStreaming__ms_aInfoForModel.gta_sa[GetBaseID(FILE_TYPE_IFP)].uiLoadFlag);
		CPatch::PatchPointer(0x64D0D8 + 3, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);
		CPatch::PatchPointer(0x67B5C0 + 2, &CStreaming__ms_aInfoForModel.gta_sa[370].uiLoadFlag);
		CPatch::PatchPointer(0x6C6A24 + 2, &CStreaming__ms_aInfoForModel.gta_sa[285].uiLoadFlag);
		CPatch::PatchPointer(0x6C7A96 + 2, &CStreaming__ms_aInfoForModel.gta_sa[497].uiLoadFlag);
		CPatch::PatchPointer(0x6C7AEE + 2, &CStreaming__ms_aInfoForModel.gta_sa[488].uiLoadFlag);
		CPatch::PatchPointer(0x6CD352 + 3, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);
		CPatch::PatchPointer(0x6D19BC + 2, &CStreaming__ms_aInfoForModel.gta_sa[349].uiLoadFlag);
		CPatch::PatchPointer(0x6D1A04 + 2, &CStreaming__ms_aInfoForModel.gta_sa[333].uiLoadFlag);
		CPatch::PatchPointer(0x6D6537 + 3, &CStreaming__ms_aInfoForModel.gta_sa[GetBaseID(FILE_TYPE_TXD)].uiLoadFlag);
		CPatch::PatchPointer(0x6F3549 + 1, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);
		CPatch::PatchPointer(0x6F3639 + 3, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);
		CPatch::PatchPointer(0x6F3654 + 2, &CStreaming__ms_aInfoForModel.gta_sa[168].uiLoadFlag);
		CPatch::PatchPointer(0x6F3CC0 + 2, &CStreaming__ms_aInfoForModel.gta_sa[168].uiLoadFlag);
		CPatch::PatchPointer(0x6F7B26 + 3, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);

		// Added manually				
		{
			// _isModelExists
			CPatch::PatchPointer(0x1565177 + 3, &CStreaming__ms_aInfoForModel.gta_sa->iBlockCount);

			// CStreaming__HasVehicleUpgradeLoaded
			CPatch::PatchPointer(0x156A107 + 3, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);
			CPatch::PatchPointer(0x156A12D + 3, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);

			// sub_407A40
			CPatch::PatchPointer(0x1566496 + 1, &CStreaming__ms_aInfoForModel.gta_sa->ucFlags);

			// CStreaming__AreAnimsUsedByRequestedModels
			CPatch::PatchPointer(0x407B02 + 2, CStreaming__ms_aInfoForModel.gta_sa);

			// sub_407C00
			CPatch::PatchPointer(0x156C9F0 + 3, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);
			CPatch::PatchPointer(0x156CA0A + 3, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);

			// CStreaming::isSpecialActorLoaded
			CPatch::PatchPointer(0x15642C7 + 3, &CStreaming__ms_aInfoForModel.gta_sa[290].uiLoadFlag);

			// sub_407F80
			CPatch::PatchPointer(0x1567B6A + 2, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);
			CPatch::PatchPointer(0x1567B73 + 3, &CStreaming__ms_aInfoForModel.gta_sa->usNext);
			CPatch::PatchPointer(0x1567B7D + 3, &CStreaming__ms_aInfoForModel.gta_sa->usPrev);

			// CStreaming__RequestModel
			CPatch::PatchPointer(0x404C95 + 2, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);
			CPatch::PatchPointer(0x408802 + 2, &CStreaming__ms_aInfoForModel.gta_sa->ucFlags);
			CPatch::PatchPointer(0x408811 + 2, &CStreaming__ms_aInfoForModel.gta_sa->ucFlags);
			CPatch::PatchPointer(0x408820 + 2, &CStreaming__ms_aInfoForModel.gta_sa->ucFlags);
			CPatch::PatchPointer(0x40882F + 2, &CStreaming__ms_aInfoForModel.gta_sa->ucFlags);
			CPatch::PatchPointer(0x408835 + 2, &CStreaming__ms_aInfoForModel.gta_sa->usNext);
			CPatch::PatchPointer(0x408840 + 2, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);
			CPatch::PatchPointer(0x4088BC + 2, &CStreaming__ms_aInfoForModel.gta_sa->ucFlags);
			CPatch::PatchPointer(0x40897F + 2, &CStreaming__ms_aInfoForModel.gta_sa->ucFlags);
			CPatch::PatchPointer(0x408989 + 2, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);

			// CStreaming__DeleteRwObjectsNotInFrustumInSectorList
			CPatch::PatchPointer(0x409978 + 4, &CStreaming__ms_aInfoForModel.gta_sa->usNext);

			// CStreaming::AreTexturesUsedByRequestedModels
			CPatch::PatchPointer(0x15664EE + 2, CStreaming__ms_aInfoForModel.gta_sa);

			// CStreaming__InitObjectInfoAndRequestModel
			CPatch::PatchPointer(0x15663C9 + 2, &CStreaming__ms_aInfoForModel.gta_sa->iBlockCount);
			CPatch::PatchPointer(0x15663D8 + 3, &CStreaming__ms_aInfoForModel.gta_sa->ucImgId);
			CPatch::PatchPointer(0x15663EB + 2, &CStreaming__ms_aInfoForModel.gta_sa->iBlockOffset);
			CPatch::PatchPointer(0x156643C + 2, &CStreaming__ms_aInfoForModel.gta_sa->iBlockOffset);
			CPatch::PatchPointer(0x1566442 + 2, &CStreaming__ms_aInfoForModel.gta_sa->iBlockCount);
			CPatch::PatchPointer(0x1566449 + 2, &CStreaming__ms_aInfoForModel.gta_sa->ucImgId);

			// CStreaming__ReloadCopModel
			CPatch::PatchPointer(0x156C202 + 3, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);
			CPatch::PatchPointer(0x156C21D + 3, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);

			// CStreaming__ReloadFiremanModels
			CPatch::PatchPointer(0x15703AA + 3, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);
			CPatch::PatchPointer(0x15703C0 + 3, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);

			// sub_40A560
			CPatch::PatchPointer(0x40A61A + 3, &CStreaming__ms_aInfoForModel.gta_sa->ucFlags);
			CPatch::PatchPointer(0x40A621 + 3, CStreaming__ms_aInfoForModel.gta_sa);
			CPatch::PatchPointer(0x40A70A + 3, &CStreaming__ms_aInfoForModel.gta_sa->ucFlags);
			CPatch::PatchPointer(0x40A711 + 3, CStreaming__ms_aInfoForModel.gta_sa);
			CPatch::PatchPointer(0x40A838 + 2, &CStreaming__ms_aInfoForModel.gta_sa->ucFlags);
			CPatch::PatchPointer(0x40A83E + 2, CStreaming__ms_aInfoForModel.gta_sa);
			CPatch::PatchPointer(0x40A84A + 2, &CStreaming__ms_aInfoForModel.gta_sa->ucFlags);
			CPatch::PatchPointer(0x40A859 + 2, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);
			CPatch::PatchPointer(0x40A8FC + 2, &CStreaming__ms_aInfoForModel.gta_sa->ucFlags);
			CPatch::PatchPointer(0x40A902 + 2, CStreaming__ms_aInfoForModel.gta_sa);
			CPatch::PatchPointer(0x40A90E + 2, &CStreaming__ms_aInfoForModel.gta_sa->ucFlags);
			CPatch::PatchPointer(0x40A91D + 2, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);

			// sub_40B080
			CPatch::PatchPointer(0x15629EC + 2, &CStreaming__ms_aInfoForModel.gta_sa->ucFlags);
			CPatch::PatchPointer(0x15629F2 + 2, &CStreaming__ms_aInfoForModel.gta_sa->usNext);
			CPatch::PatchPointer(0x15629FE + 2, &CStreaming__ms_aInfoForModel.gta_sa->ucFlags);
			CPatch::PatchPointer(0x1562A09 + 2, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);
			CPatch::PatchPointer(0x1562A4C + 2, &CStreaming__ms_aInfoForModel.gta_sa->ucFlags);
			CPatch::PatchPointer(0x1562A52 + 2, &CStreaming__ms_aInfoForModel.gta_sa->usNext);
			CPatch::PatchPointer(0x1562A5E + 2, &CStreaming__ms_aInfoForModel.gta_sa->ucFlags);
			CPatch::PatchPointer(0x1562A69 + 2, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);
			CPatch::PatchPointer(0x1562AEF + 2, &CStreaming__ms_aInfoForModel.gta_sa->ucFlags);
			CPatch::PatchPointer(0x1562AF5 + 2, &CStreaming__ms_aInfoForModel.gta_sa->usNext);
			CPatch::PatchPointer(0x1562B01 + 2, &CStreaming__ms_aInfoForModel.gta_sa->ucFlags);
			CPatch::PatchPointer(0x1562B0C + 2, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);
			CPatch::PatchPointer(0x1562BAB + 2, &CStreaming__ms_aInfoForModel.gta_sa->ucFlags);
			CPatch::PatchPointer(0x1562BB1 + 2, &CStreaming__ms_aInfoForModel.gta_sa->usNext);
			CPatch::PatchPointer(0x1562BBD + 2, &CStreaming__ms_aInfoForModel.gta_sa->ucFlags);
			CPatch::PatchPointer(0x1562BC8 + 2, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);

			// sub_40B340
			CPatch::PatchPointer(0x1566F13 + 2, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);
			CPatch::PatchPointer(0x1566F1C + 2, &CStreaming__ms_aInfoForModel.gta_sa->ucFlags);

			// _ZN10CStreaming15StreamOneNewCarEv
			CPatch::PatchPointer(0x1563C12 + 2, &CStreaming__ms_aInfoForModel.gta_sa[420].uiLoadFlag);
			CPatch::PatchPointer(0x1563C22 + 2, &CStreaming__ms_aInfoForModel.gta_sa[438].uiLoadFlag);

			// sub_40CF80
			CPatch::PatchPointer(0x1564604 + 1, &CStreaming__ms_aInfoForModel.gta_sa[1000].ucFlags);

			// sub_40E560
			CPatch::PatchPointer(0x156D020 + 1, &CStreaming__ms_aInfoForModel.gta_sa->ucFlags);

			// sub_42FBC0
			CPatch::PatchPointer(0x1570109 + 3, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);

			// _ZN8CCarCtrl20GenerateOneRandomCarEv
			CPatch::PatchPointer(0x4304E6 + 2, &CStreaming__ms_aInfoForModel.gta_sa[430].uiLoadFlag);
			CPatch::PatchPointer(0x430535 + 3, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);

			// _spawnCarAtPlayerLocation
			CPatch::PatchPointer(0x43A128 + 3, &CStreaming__ms_aInfoForModel.gta_sa->ucFlags);
			CPatch::PatchPointer(0x43A13C + 2, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);

			// sub_441D30
			CPatch::PatchPointer(0x1560FCF + 3, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);
			CPatch::PatchPointer(0x1560FE0 + 3, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);

			// CGangWars__HasAnyPedFromGangLoaded
			CPatch::PatchPointer(0x443A06 + 3, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);

			// CGangWars__InitWave
			CPatch::PatchPointer(0x444A33 + 3, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);
			CPatch::PatchPointer(0x444A44 + 3, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);
			CPatch::PatchPointer(0x4450F4 + 3, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);

			// ?processPickup
			CPatch::PatchPointer(0x457A9C + 3, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);

			// sub_45CA70
			CPatch::PatchPointer(0x45CC8E + 3, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);

			// sub_45CEA0
			CPatch::PatchPointer(0x156EDFA + 3, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);

			// sub_45F380
			CPatch::PatchPointer(0x45F406 + 3, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);
			CPatch::PatchPointer(0x45F674 + 3, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);
			CPatch::PatchPointer(0x45F7B2 + 3, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);
			CPatch::PatchPointer(0x45F8E5 + 3, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);
			CPatch::PatchPointer(0x45FA18 + 3, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);
			CPatch::PatchPointer(0x45FB5F + 3, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);


			// sub_4619C0
			CPatch::PatchPointer(0x461BFC + 3, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);

			// _opcode_handler_00
			CPatch::PatchPointer(0x4668E7 + 2, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);

			// sub_46AA80
			CPatch::PatchPointer(0x156D9A4 + 3, &CStreaming__ms_aInfoForModel.gta_sa[GetBaseID(FILE_TYPE_SCM)].uiLoadFlag);

			// sub_46AB20
			CPatch::PatchPointer(0x1561671 + 3, &CStreaming__ms_aInfoForModel.gta_sa[GetBaseID(FILE_TYPE_SCM)].uiLoadFlag);

			// _opcode_handler_22
			CPatch::PatchPointer(0x475071 + 3, &CStreaming__ms_aInfoForModel.gta_sa[GetBaseID(FILE_TYPE_SCM)].uiLoadFlag);

			// CCutSceneMgr__LoadCutsceneFile_preload
			CPatch::PatchPointer(0x5B0606 + 1, &CStreaming__ms_aInfoForModel.gta_sa[300].uiLoadFlag);
			CPatch::PatchPointer(0x5B0646 + 2, &CStreaming__ms_aInfoForModel.gta_sa[319].uiLoadFlag);
			CPatch::PatchPointer(0x5B112F + 3, &CStreaming__ms_aInfoForModel.gta_sa[300].uiLoadFlag);
			CPatch::PatchPointer(0x5B1139 + 3, &CStreaming__ms_aInfoForModel.gta_sa[300].uiLoadFlag);

			// CStreaming__LoadCdDirectory2
			CPatch::PatchPointer(0x409BBC + 3, CStreaming__ms_aInfoForModel.gta_sa);
			CPatch::PatchPointer(0x5B6468 + 2, &CStreaming__ms_aInfoForModel.gta_sa->ucImgId);
			CPatch::PatchPointer(0x5B6480 + 2, CStreaming__ms_aInfoForModel.gta_sa);
			CPatch::PatchPointer(0x5B6495 + 3, &CStreaming__ms_aInfoForModel.gta_sa->ucFlags);
			CPatch::PatchPointer(0x5B64A2 + 4, &CStreaming__ms_aInfoForModel.gta_sa->usNextOnCd);

			// sub_612F90
			CPatch::PatchPointer(0x612FFD + 3, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);

			// CPopulation::AddToPopulation
			CPatch::PatchPointer(0x615076 + 2, &CStreaming__ms_aInfoForModel.gta_sa[286].uiLoadFlag);
			CPatch::PatchPointer(0x615092 + 3, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);
			CPatch::PatchPointer(0x6150AE + 2, &CStreaming__ms_aInfoForModel.gta_sa[285].uiLoadFlag);
			CPatch::PatchPointer(0x6150CA + 3, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);
			CPatch::PatchPointer(0x6150E4 + 2, &CStreaming__ms_aInfoForModel.gta_sa[287].uiLoadFlag);
			CPatch::PatchPointer(0x615100 + 3, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);

			// Later added

			// _sub_4044C0
			CPatch::PatchPointer(0x4044C7 + 3, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);

			// sub_4044E0
			CPatch::PatchPointer(0x4044E7 + 4, &CStreaming__ms_aInfoForModel.gta_sa->ucFlags);

			// _sub_404530
			CPatch::PatchPointer(0x404537 + 4, &CStreaming__ms_aInfoForModel.gta_sa[GetBaseID(FILE_TYPE_IPL)].ucFlags);

			// _sub_4080B0
			CPatch::PatchPointer(0x4080C8 + 2, &CStreaming__ms_aInfoForModel.gta_sa->ucFlags);
			CPatch::PatchPointer(0x4080D1 + 2, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);

			// _sub_410380
			CPatch::PatchPointer(0x410387 + 4, &CStreaming__ms_aInfoForModel.gta_sa[GetBaseID(FILE_TYPE_COL)].ucFlags);

			// _sub_46A590
			CPatch::PatchPointer(0x46A597 + 3, &CStreaming__ms_aInfoForModel.gta_sa[GetBaseID(FILE_TYPE_SCM)].uiLoadFlag);

			// _sub_6E2900
			CPatch::PatchPointer(0x6E298A + 3, &CStreaming__ms_aInfoForModel.gta_sa->uiLoadFlag);
		}

		// offset _ZN14CStreamingInfo13ms_pArrayBaseE+2
		{
			// sub_407A40
			CPatch::PatchPointer(0x15664A5 + 1, &CStreaming__ms_aInfoForModel.gta_sa[GetCountOfAllFileIDs()].ucFlags);

			

			// sub_40E560
			CPatch::PatchPointer(0x156D037 + 2, &CStreaming__ms_aInfoForModel.gta_sa[GetCountOfAllFileIDs()].ucFlags);

			// sub_40EB70
			CPatch::PatchPointer(0x40ED49 + 1, &CStreaming__ms_aInfoForModel.gta_sa[GetCountOfAllFileIDs()].ucFlags);
		}

		// References to of ms_aInfoForModel + countof(ms_aInfoForModel)
		{
			CPatch::PatchPointer(0x5B8AFA + 2, CStreaming__ms_aInfoForModel.gta_sa + GetCountOfAllFileIDs());
		}
	}
	#endif

	// CStreaming::ms_channel buggy code by R*?
	#if TRUE
	{
		CStreamingChannel* ms_channel = CStreamingAddition::ms_channel;

		CPatch::PatchPointer(0x1566597 + 2, &ms_channel[1].readbuf_id[GetBaseID(FILE_TYPE_TXD)]);
		CPatch::PatchPointer(0x156659F + 2, &ms_channel[1].readbuf_id[GetBaseID(FILE_TYPE_COL)]);
		CPatch::PatchPointer(0x15665DA + 2, &ms_channel[1].readbuf_id[GetBaseID(FILE_TYPE_TXD)]);
		CPatch::PatchPointer(0x15665E2 + 2, &ms_channel[1].readbuf_id[GetBaseID(FILE_TYPE_COL)]);
	}
	#endif

	// Patch base IDs
	#if TRUE
	{
		// TXD
		{
			// Function to consider:
			// sub_566510

			// Limit should be aligned to 10
			// _sub_5B36C0

			// 20000
			{
				// sub_566510, unfinished.
				CPatch::PatchUINT32(0x566516 + 1, GetBaseID(FILE_TYPE_TXD) * 4);
				CPatch::PatchUINT32(0x566532 + 1, GetBaseID(FILE_TYPE_TXD));	// mov     ecx, 4E20h


				// 20000
				CPatch::PatchUINT32(0x407104 + 2, GetBaseID(FILE_TYPE_TXD));	// add     ecx, 4E20h
				CPatch::PatchUINT32(0x407B19 + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     edx, 4E20h
				CPatch::PatchUINT32(0x407B6B + 1, GetBaseID(FILE_TYPE_TXD));	// cmp     eax, 4E20h
				CPatch::PatchUINT32(0x407B8B + 1, GetBaseID(FILE_TYPE_TXD));	// cmp     eax, 4E20h
				CPatch::PatchUINT32(0x408858 + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     ebp, 4E20h
				CPatch::PatchUINT32(0x4088FB + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     ebp, 4E20h
				CPatch::PatchUINT32(0x4089C6 + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     esi, 4E20h
				CPatch::PatchUINT32(0x408B9E + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     esi, 4E20h
				CPatch::PatchUINT32(0x408CC8 + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     esi, 4E20h
				CPatch::PatchUINT32(0x408D13 + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     esi, 20000
				CPatch::PatchUINT32(0x408EB4 + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     edi, 4E20h
				CPatch::PatchUINT32(0x408EEC + 1, GetBaseID(FILE_TYPE_TXD));	// add     eax, 4E20h
				CPatch::PatchUINT32(0x408F77 + 1, GetBaseID(FILE_TYPE_TXD));	// add     eax, 4E20h
				CPatch::PatchUINT32(0x409C7F + 2, GetBaseID(FILE_TYPE_TXD));	// add     edx, 4E20h
				CPatch::PatchUINT32(0x409CED + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     esi, 4E20h
				CPatch::PatchUINT32(0x409D00 + 2, GetBaseID(FILE_TYPE_TXD));	// add     esi, 4E20h
				CPatch::PatchUINT32(0x40A2C3 + 1, GetBaseID(FILE_TYPE_TXD));	// add     eax, 4E20h
				CPatch::PatchUINT32(0x40A2EB + 2, GetBaseID(FILE_TYPE_TXD));	// add     ecx, 4E20h
				CPatch::PatchUINT32(0x40A313 + 2, GetBaseID(FILE_TYPE_TXD));	// add     edx, 4E20h
				CPatch::PatchUINT32(0x40A33B + 1, GetBaseID(FILE_TYPE_TXD));	// add     eax, 4E20h
				CPatch::PatchUINT32(0x40A363 + 2, GetBaseID(FILE_TYPE_TXD));	// add     ecx, 4E20h
				CPatch::PatchUINT32(0x40A38B + 2, GetBaseID(FILE_TYPE_TXD));	// add     edx, 4E20h
				CPatch::PatchUINT32(0x40A5BD + 1, GetBaseID(FILE_TYPE_TXD));	// add     eax, 4E20h
				CPatch::PatchUINT32(0x40A73C + 2, GetBaseID(FILE_TYPE_TXD));	// add     edx, 4E20h
				CPatch::PatchUINT32(0x40A8F0 + 2, GetBaseID(FILE_TYPE_TXD));	// add     esi, 4E20h
				CPatch::PatchUINT32(0x40AC27 + 2, GetBaseID(FILE_TYPE_TXD));	// add     edx, 4E20h
				CPatch::PatchUINT32(0x40AE96 + 2, GetBaseID(FILE_TYPE_TXD));	// add     esi, 4E20h
				CPatch::PatchUINT32(0x40BB2D + 2, GetBaseID(FILE_TYPE_TXD));	// add     edx, 4E20h
				CPatch::PatchUINT32(0x40BEE8 + 2, GetBaseID(FILE_TYPE_TXD));	// add     edx, 4E20h
				CPatch::PatchUINT32(0x40C184 + 1, GetBaseID(FILE_TYPE_TXD));	// add     eax, 4E20h
				CPatch::PatchUINT32(0x40C6F0 + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     esi, 4E20h
				CPatch::PatchUINT32(0x40CACA + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     esi, 4E20h
				CPatch::PatchUINT32(0x40CBE8 + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     esi, 4E20h
				CPatch::PatchUINT32(0x40CD56 + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     esi, 4E20h
				CPatch::PatchUINT32(0x40CD97 + 1, GetBaseID(FILE_TYPE_TXD));	// add     eax, 4E20h
				CPatch::PatchUINT32(0x40CE62 + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     esi, 4E20h
				CPatch::PatchUINT32(0x40D055 + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     esi, 4E20h
				CPatch::PatchUINT32(0x40D07C + 2, GetBaseID(FILE_TYPE_TXD));	// sub     esi, 4E20h
				CPatch::PatchUINT32(0x40D0A0 + 2, GetBaseID(FILE_TYPE_TXD));	// add     esi, 4E20h
				CPatch::PatchUINT32(0x40D2B2 + 2, GetBaseID(FILE_TYPE_TXD));	// lea     eax, [esi+4E20h]
				CPatch::PatchUINT32(0x40D368 + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     edi, 4E20h
				CPatch::PatchUINT32(0x40D37E + 2, GetBaseID(FILE_TYPE_TXD));	// sub     edi, 4E20h
				CPatch::PatchUINT32(0x40D391 + 2, GetBaseID(FILE_TYPE_TXD));	// add     edi, 4E20h
				CPatch::PatchUINT32(0x40E1F1 + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     esi, 4E20h

				CPatch::PatchUINT32(0x4C487C + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     edi, 4E20h
				CPatch::PatchUINT32(0x4C4884 + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     edi, 4E20h
				CPatch::PatchUINT32(0x4C5956 + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     ecx, 4E20h
				CPatch::PatchUINT32(0x4C5971 + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     ecx, 4E20h
				CPatch::PatchUINT32(0x4C59D1 + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     ecx, 4E20h
				CPatch::PatchUINT32(0x4C5C84 + 1, GetBaseID(FILE_TYPE_TXD));	// cmp     eax, 4E20h
				CPatch::PatchUINT32(0x4C6815 + 1, GetBaseID(FILE_TYPE_TXD));	// mov     ecx, 4E20h

				CPatch::PatchUINT32(0x5384F4 + 1, GetBaseID(FILE_TYPE_TXD));	// cmp     eax, 4E20h
				CPatch::PatchUINT32(0x584B8D + 1, GetBaseID(FILE_TYPE_TXD));	// add     eax, 4E20h
				CPatch::PatchUINT32(0x584C24 + 1, GetBaseID(FILE_TYPE_TXD));	// add     eax, 4E20h
				CPatch::PatchUINT32(0x584CC3 + 1, GetBaseID(FILE_TYPE_TXD));	// add     eax, 4E20h
				CPatch::PatchUINT32(0x584CF9 + 1, GetBaseID(FILE_TYPE_TXD));	// add     eax, 4E20h
				CPatch::PatchUINT32(0x5A4295 + 2, GetBaseID(FILE_TYPE_TXD));	// lea     eax, [esi+4E20h]
				CPatch::PatchUINT32(0x5A5FF4 + 2, GetBaseID(FILE_TYPE_TXD));	// add     edi, 4E20h
				CPatch::PatchUINT32(0x5A6184 + 2, GetBaseID(FILE_TYPE_TXD));	// add     ecx, 4E20h
				CPatch::PatchUINT32(0x5A625D + 2, GetBaseID(FILE_TYPE_TXD));	// add     edi, 4E20h
				CPatch::PatchUINT32(0x5A6391 + 2, GetBaseID(FILE_TYPE_TXD));	// add     ecx, 4E20h
				CPatch::PatchUINT32(0x5A64EE + 2, GetBaseID(FILE_TYPE_TXD));	// add     edi, 4E20h
				CPatch::PatchUINT32(0x5B384D + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     esi, 4E20h
				CPatch::PatchUINT32(0x5B5096 + 1, GetBaseID(FILE_TYPE_TXD));	// cmp     eax, 4E20h
				CPatch::PatchUINT32(0x5B62CE + 1, GetBaseID(FILE_TYPE_TXD));	// add     eax, 4E20h

				CPatch::PatchUINT32(0x6D65D2 + 1, GetBaseID(FILE_TYPE_TXD));	// add     eax, 4E20h
				CPatch::PatchUINT32(0x731A59 + 2, GetBaseID(FILE_TYPE_TXD));	// add     ecx, 4E20h
				CPatch::PatchUINT32(0x731F01 + 1, GetBaseID(FILE_TYPE_TXD));	// add     eax, 4E20h
				CPatch::PatchUINT32(0x1562A40 + 2, GetBaseID(FILE_TYPE_TXD));	// add     edx, 4E20h
				CPatch::PatchUINT32(0x1562B9F + 2, GetBaseID(FILE_TYPE_TXD));	// add     esi, 4E20h
				CPatch::PatchUINT32(0x1564630 + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     esi, 4E20h
				CPatch::PatchUINT32(0x1566505 + 2, GetBaseID(FILE_TYPE_TXD));	// cmp     edx, 4E20h
				CPatch::PatchUINT32(0x1566583 + 1, GetBaseID(FILE_TYPE_TXD));	// cmp     eax, 4E20h
				CPatch::PatchUINT32(0x15665C2 + 1, GetBaseID(FILE_TYPE_TXD));	// cmp     eax, 4E20h
				CPatch::PatchUINT32(0x156C252 + 2, GetBaseID(FILE_TYPE_TXD));	// add     ecx, 4E20h
				CPatch::PatchUINT32(0x156C280 + 2, GetBaseID(FILE_TYPE_TXD));	// add     edx, 4E20h
				CPatch::PatchUINT32(0x156CA68 + 1, GetBaseID(FILE_TYPE_TXD));	// add     eax, 4E20h
				CPatch::PatchUINT32(0x156CAAD + 1, GetBaseID(FILE_TYPE_TXD));	// add     eax, 4E20h
				CPatch::PatchUINT32(0x1570253 + 1, GetBaseID(FILE_TYPE_TXD));	// add     eax, 4E20h
				CPatch::PatchUINT32(0x157027B + 2, GetBaseID(FILE_TYPE_TXD));	// add     ecx, 4E20h
				CPatch::PatchUINT32(0x15702A3 + 2, GetBaseID(FILE_TYPE_TXD));	// add     edx, 4E20h
				CPatch::PatchUINT32(0x15702CB + 1, GetBaseID(FILE_TYPE_TXD));	// add     eax, 4E20h
				CPatch::PatchUINT32(0x15702F3 + 2, GetBaseID(FILE_TYPE_TXD));	// add     ecx, 4E20h
				CPatch::PatchUINT32(0x157031B + 2, GetBaseID(FILE_TYPE_TXD));	// add     edx, 4E20h
				CPatch::PatchUINT32(0x157040F + 2, GetBaseID(FILE_TYPE_TXD));	// add     edx, 4E20h
				CPatch::PatchUINT32(0x1570468 + 2, GetBaseID(FILE_TYPE_TXD));	// add     ecx, 4E20h
			}

			// -20000
			{
				CPatch::PatchUINT32(0x40893C + 2, -GetBaseID(FILE_TYPE_TXD));	// add     ebp, -20000
				CPatch::PatchUINT32(0x408A30 + 2, -GetBaseID(FILE_TYPE_TXD));	// lea     ecx, [esi-4E20h]
				CPatch::PatchUINT32(0x408BBD + 2, -GetBaseID(FILE_TYPE_TXD));	// add     esi, 0FFFFB1E0h
				CPatch::PatchUINT32(0x408D91 + 2, -GetBaseID(FILE_TYPE_TXD));	// lea     ebx, [esi-4E20h]
				CPatch::PatchUINT32(0x408DA6 + 2, -GetBaseID(FILE_TYPE_TXD));	// lea     eax, [esi-4E20h]
				CPatch::PatchUINT32(0x408F53 + 3, -GetBaseID(FILE_TYPE_TXD));	// mov     cl, [edi+edx-4E20h]
				CPatch::PatchUINT32(0x40C8B0 + 3, -GetBaseID(FILE_TYPE_TXD));	// mov     dl, [ecx+esi-4E20h]
				CPatch::PatchUINT32(0x40C8B9 + 2, -GetBaseID(FILE_TYPE_TXD));	// lea     ebp, [esi-4E20h]
				CPatch::PatchUINT32(0x40CBF8 + 2, -GetBaseID(FILE_TYPE_TXD));	// lea     eax, [esi-4E20h]
				CPatch::PatchUINT32(0x1566529 + 3, -GetBaseID(FILE_TYPE_TXD));	// mov     cl, [eax+edx-20000]
				CPatch::PatchUINT32(0x15665A7 + 1, -GetBaseID(FILE_TYPE_TXD));	// add     eax, 0FFFFB1E0h
				CPatch::PatchUINT32(0x15665EA + 1, -GetBaseID(FILE_TYPE_TXD));	// add     eax, 0FFFFB1E0h
			}

			// 60000
			{
				CPatch::PatchUINT32(0x40D2D4 + 2, (GetBaseID(FILE_TYPE_COL) - GetBaseID(FILE_TYPE_TXD)) * sizeof(TxdDef));
				CPatch::PatchUINT32(0x73209C + 2, (GetBaseID(FILE_TYPE_COL) - GetBaseID(FILE_TYPE_TXD)) * sizeof(TxdDef));
			}

			// -60000
			{
				CPatch::PatchUINT32(0x408F64 + 3, -GetBaseID(FILE_TYPE_TXD) * 3);	// lea     ecx, [edi+edi*2-60000]
				CPatch::PatchUINT32(0x1566538 + 3, -GetBaseID(FILE_TYPE_TXD) * 3);	// lea     ecx, [edx+edx*2-60000]
			}

			// 60036
			{
				CPatch::PatchUINT32(0x5B8D8C + 2, ((GetBaseID(FILE_TYPE_COL) - GetBaseID(FILE_TYPE_TXD)) + 3) * sizeof(TxdDef));	// cmp     ecx, 60036
			}
		}

		// COL
		{
			// 25000
			{
				CPatch::PatchUINT32(0x408934 + 2, GetBaseID(FILE_TYPE_COL));	// cmp     ebp, 25000
				CPatch::PatchUINT32(0x408A28 + 2, GetBaseID(FILE_TYPE_COL));	// cmp     esi, 25000
				CPatch::PatchUINT32(0x408BB5 + 2, GetBaseID(FILE_TYPE_COL));	// cmp     esi, 61A8h
				CPatch::PatchUINT32(0x408D89 + 2, GetBaseID(FILE_TYPE_COL));	// cmp     esi, 25000
				CPatch::PatchUINT32(0x408F43 + 2, GetBaseID(FILE_TYPE_COL));	// cmp     edi, 25000
				CPatch::PatchUINT32(0x40C89C + 2, GetBaseID(FILE_TYPE_COL));	// cmp     esi, 25000
				CPatch::PatchUINT32(0x40CB29 + 2, GetBaseID(FILE_TYPE_COL));	// cmp     esi, 25000
				CPatch::PatchUINT32(0x40CBF0 + 2, GetBaseID(FILE_TYPE_COL));	// cmp     esi, 25000
				CPatch::PatchUINT32(0x40D074 + 2, GetBaseID(FILE_TYPE_COL));	// cmp     esi, 25000
				CPatch::PatchUINT32(0x40D370 + 2, GetBaseID(FILE_TYPE_COL));	// cmp     edi, 25000
				CPatch::PatchUINT32(0x410344 + 2, GetBaseID(FILE_TYPE_COL));	// add     ecx, 61A8h
				CPatch::PatchUINT32(0x410364 + 1, GetBaseID(FILE_TYPE_COL));	// add     eax, 61A8h
				CPatch::PatchUINT32(0x410A73 + 2, GetBaseID(FILE_TYPE_COL));	// lea     eax, [edi+25000]
				CPatch::PatchUINT32(0x410A8B + 2, GetBaseID(FILE_TYPE_COL));	// lea     ecx, [edi+25000]
				CPatch::PatchUINT32(0x410B8A + 2, GetBaseID(FILE_TYPE_COL));	// lea     edx, [edi+61A8h]
				CPatch::PatchUINT32(0x410CA2 + 2, GetBaseID(FILE_TYPE_COL));	// lea     eax, [ebp+25000]
				CPatch::PatchUINT32(0x410E38 + 2, GetBaseID(FILE_TYPE_COL));	// lea     eax, [edi+61A8h]
				CPatch::PatchUINT32(0x410E86 + 2, GetBaseID(FILE_TYPE_COL));	// lea     esi, [edi+61A8h]
				CPatch::PatchUINT32(0x5B6313 + 1, GetBaseID(FILE_TYPE_COL));	// add     eax, 61A8h
				CPatch::PatchUINT32(0x156651E + 2, GetBaseID(FILE_TYPE_COL));	// cmp     edx, 25000
			}

			// -25000
			{
				CPatch::PatchUINT32(0x408A4C + 2, -GetBaseID(FILE_TYPE_COL));	// lea     edx, [esi-61A8h]
				CPatch::PatchUINT32(0x408BDE + 2, -GetBaseID(FILE_TYPE_COL));	// add     esi, 0FFFF9E58h
				CPatch::PatchUINT32(0x40C982 + 2, -GetBaseID(FILE_TYPE_COL));	// lea     ecx, [esi-25000]
			}

			// 11220
			{
				uint32_t sizeOfAllCollisionFiles = (GetBaseID(FILE_TYPE_IPL) - GetBaseID(FILE_TYPE_COL)) * sizeof(CCollisionFile);

				CPatch::PatchUINT32(0x410AA4 + 2, sizeOfAllCollisionFiles);	// cmp     ebp, 11220
				CPatch::PatchUINT32(0x410CC4 + 2, sizeOfAllCollisionFiles);	// cmp     edi, 11220
				CPatch::PatchUINT32(0x410DBB + 2, sizeOfAllCollisionFiles);	// cmp     esi, 11220
				CPatch::PatchUINT32(0x410DE5 + 2, sizeOfAllCollisionFiles);	// cmp     ecx, 11220
				CPatch::PatchUINT32(0x410EAD + 2, sizeOfAllCollisionFiles);	// cmp     ebx, 11220
				CPatch::PatchUINT32(0x410F68 + 2, sizeOfAllCollisionFiles);	// cmp     ebp, 11220
				CPatch::PatchUINT32(0x411505 + 2, sizeOfAllCollisionFiles);	// cmp     edi, 11220
			}

			// 255
			{
				CPatch::PatchUINT32(0x410D90 + 2, GetBaseID(FILE_TYPE_IPL) - GetBaseID(FILE_TYPE_COL));	// cmp     edx, 0FFh
			}
		}


		// IPL
		{
			// 25255
			{
				CPatch::PatchUINT32(0x4044F4 + 2, GetBaseID(FILE_TYPE_IPL));	// add     ecx, 25255
				CPatch::PatchUINT32(0x404514 + 1, GetBaseID(FILE_TYPE_IPL));	// add     eax, 25255
				CPatch::PatchUINT32(0x4050BA + 2, GetBaseID(FILE_TYPE_IPL));	// lea     eax, [edi+25255]
				CPatch::PatchUINT32(0x405147 + 2, GetBaseID(FILE_TYPE_IPL));	// lea     eax, [esi+25255]
				CPatch::PatchUINT32(0x40538C + 2, GetBaseID(FILE_TYPE_IPL));	// lea     ecx, [edi+25255]
				CPatch::PatchUINT32(0x4053A7 + 2, GetBaseID(FILE_TYPE_IPL));	// lea     edx, [edi+25255]
				CPatch::PatchUINT32(0x4054B5 + 2, GetBaseID(FILE_TYPE_IPL));	// lea     edx, [edi+25255]
				CPatch::PatchUINT32(0x4055C2 + 2, GetBaseID(FILE_TYPE_IPL));	// lea     eax, [ebp+25255]
				CPatch::PatchUINT32(0x405758 + 2, GetBaseID(FILE_TYPE_IPL));	// lea     eax, [edi+25255]
				CPatch::PatchUINT32(0x405807 + 2, GetBaseID(FILE_TYPE_IPL));	// lea     esi, [ebx+25255]
				CPatch::PatchUINT32(0x408A44 + 2, GetBaseID(FILE_TYPE_IPL));	// cmp     esi, 25255
				CPatch::PatchUINT32(0x408BD6 + 2, GetBaseID(FILE_TYPE_IPL));	// cmp     esi, 25255
				CPatch::PatchUINT32(0x40C974 + 2, GetBaseID(FILE_TYPE_IPL));	// cmp     esi, 25255
				CPatch::PatchUINT32(0x40E293 + 2, GetBaseID(FILE_TYPE_IPL));	// cmp     esi, 25255
				CPatch::PatchUINT32(0x5B6358 + 1, GetBaseID(FILE_TYPE_IPL));	// add     eax, 25255
				CPatch::PatchUINT32(0x15651B1 + 1, GetBaseID(FILE_TYPE_IPL));	// add     eax, 25255
				CPatch::PatchUINT32(0x156A161 + 1, GetBaseID(FILE_TYPE_IPL));	// add     eax, 25255
			}

			// -25255
			{
				CPatch::PatchUINT32(0x408A65 + 2, -GetBaseID(FILE_TYPE_IPL));	// lea     eax, [esi-25255]
				CPatch::PatchUINT32(0x408BFF + 2, -GetBaseID(FILE_TYPE_IPL));	// add     esi, 0FFFF9D59h
				CPatch::PatchUINT32(0x40C9D3 + 2, -GetBaseID(FILE_TYPE_IPL));	// lea     edx, [esi-25255]
			}

			uint32_t sizeOfAllIPLfiles = (GetBaseID(FILE_TYPE_DAT) - GetBaseID(FILE_TYPE_IPL)) * sizeof(CIplFile);

			// 13312
			{
				CPatch::PatchUINT32(0x4053C5 + 2, sizeOfAllIPLfiles);	// cmp     ebp, 3400h
				CPatch::PatchUINT32(0x4055E4 + 2, sizeOfAllIPLfiles);	// cmp     edi, 3400h
				CPatch::PatchUINT32(0x4056DC + 2, sizeOfAllIPLfiles);	// cmp     esi, 3400h
				CPatch::PatchUINT32(0x405707 + 2, sizeOfAllIPLfiles);	// cmp     ecx, 3400h
				CPatch::PatchUINT32(0x405829 + 2, sizeOfAllIPLfiles);	// cmp     edi, 3400h


				CPatch::PatchUINT32(0x5B2B3D + 1, sizeOfAllIPLfiles);	// push    3400h		


				CPatch::PatchUINT32(0x5D548F + 2, sizeOfAllIPLfiles);	// cmp     esi, 3400h				

				CPatch::PatchUINT32(0x15610E9 + 2, sizeOfAllIPLfiles);	// cmp     edi, 3400h
			}

			// 256 ( number of IPL files )
			{
				CPatch::PatchUINT32(0x5D542C + 4, GetBaseID(FILE_TYPE_DAT) - GetBaseID(FILE_TYPE_IPL));	// mov     [esp+18h+var_4], 256

				CPatch::PatchUINT32(0x4056B3 + 2, GetBaseID(FILE_TYPE_DAT) - GetBaseID(FILE_TYPE_IPL));	// cmp     edx, 100h
			}

			// CColAccel__startCache
			{
				CPatch::PatchUINT32(0x5B31AA + 1, GetBaseID(FILE_TYPE_DAT) - GetBaseID(FILE_TYPE_IPL));	// push    256
				CPatch::PatchUINT32(0x5B31BA + 1, sizeOfAllIPLfiles);	// push    3400h
				CPatch::PatchUINT32(0x5B31DD + 1, GetBaseID(FILE_TYPE_DAT) - GetBaseID(FILE_TYPE_IPL));	// mov     edx, 100h
				CPatch::PatchUINT32(0x5B31F7 + 1, (GetBaseID(FILE_TYPE_DAT) - GetBaseID(FILE_TYPE_IPL)) * 24);	// push    1800h
				CPatch::PatchUINT32(0x5B3210 + 1, GetBaseID(FILE_TYPE_DAT) - GetBaseID(FILE_TYPE_IPL));	// mov     edx, 256
			}
		}

		// DAT
		{
			// 25511
			{
				CPatch::PatchUINT32(0x408A5D + 2, GetBaseID(FILE_TYPE_DAT));	// cmp     esi, 25511
				CPatch::PatchUINT32(0x408BF7 + 2, GetBaseID(FILE_TYPE_DAT));	// cmp     esi, 25511
				CPatch::PatchUINT32(0x40C9C5 + 2, GetBaseID(FILE_TYPE_DAT));	// cmp     esi, 25511
				CPatch::PatchUINT32(0x40E29B + 2, GetBaseID(FILE_TYPE_DAT));	// cmp     esi, 25511

				CPatch::PatchUINT32(0x44E4C4 + 1, GetBaseID(FILE_TYPE_DAT));	// add     eax, 25511
				CPatch::PatchUINT32(0x450CF2 + 2, GetBaseID(FILE_TYPE_DAT));	// lea     edi, [ebp+25511]
				CPatch::PatchUINT32(0x5B6395 + 1, GetBaseID(FILE_TYPE_DAT));	// add     eax, 25511
				CPatch::PatchUINT32(0x1562761 + 2, GetBaseID(FILE_TYPE_DAT));	// lea     edi, [ebp+25511]
				CPatch::PatchUINT32(0x15656D2 + 2, GetBaseID(FILE_TYPE_DAT));	// lea     edi, [ebp+25511]
			}

			// -25511
			{
				CPatch::PatchUINT32(0x408A7E + 2, -GetBaseID(FILE_TYPE_DAT));	// lea     ecx, [esi-25511]
				CPatch::PatchUINT32(0x40CA1E + 2, -GetBaseID(FILE_TYPE_DAT));	// lea     edx, [esi-25511]
			}
		}

		// IFP
		{
			// 25575
			{
				CPatch::PatchUINT32(0x407124 + 2, GetBaseID(FILE_TYPE_IFP));	// add     ecx, 25575
				CPatch::PatchUINT32(0x408925 + 1, GetBaseID(FILE_TYPE_IFP));	// add     eax, 25575
				CPatch::PatchUINT32(0x408A76 + 2, GetBaseID(FILE_TYPE_IFP));	// cmp     esi, 25575
				CPatch::PatchUINT32(0x408C18 + 2, GetBaseID(FILE_TYPE_IFP));	// cmp     esi, 25575
				CPatch::PatchUINT32(0x408F0F + 1, GetBaseID(FILE_TYPE_IFP));	// add     eax, 25575
				CPatch::PatchUINT32(0x408FA0 + 2, GetBaseID(FILE_TYPE_IFP));	// cmp     edi, 25575
				CPatch::PatchUINT32(0x40C1A4 + 1, GetBaseID(FILE_TYPE_IFP));	// add     eax, 25575
				CPatch::PatchUINT32(0x40CA16 + 2, GetBaseID(FILE_TYPE_IFP));	// cmp     esi, 25575
				CPatch::PatchUINT32(0x40CB31 + 2, GetBaseID(FILE_TYPE_IFP));	// cmp     esi, 25575
				CPatch::PatchUINT32(0x40CC0B + 2, GetBaseID(FILE_TYPE_IFP));	// cmp     esi, 25575
				CPatch::PatchUINT32(0x40CDBE + 1, GetBaseID(FILE_TYPE_IFP));	// add     eax, 25575
				CPatch::PatchUINT32(0x40CDDB + 2, GetBaseID(FILE_TYPE_IFP));	// cmp     esi, 25575
				CPatch::PatchUINT32(0x40D0B5 + 2, GetBaseID(FILE_TYPE_IFP));	// cmp     esi, 25575
				CPatch::PatchUINT32(0x40D0C5 + 2, GetBaseID(FILE_TYPE_IFP));	// sub     esi, 25575
				CPatch::PatchUINT32(0x40D0E5 + 2, GetBaseID(FILE_TYPE_IFP));	// add     esi, 25575

				CPatch::PatchUINT32(0x48B574 + 1, GetBaseID(FILE_TYPE_IFP));	// add     eax, 25575
				CPatch::PatchUINT32(0x48C36B + 2, GetBaseID(FILE_TYPE_IFP));	// lea     eax, [edi+25575]


				CPatch::PatchUINT32(0x49B907 + 1, GetBaseID(FILE_TYPE_IFP));	// add     eax, 25575
				CPatch::PatchUINT32(0x49BB99 + 1, GetBaseID(FILE_TYPE_IFP));	// add     eax, 25575
				CPatch::PatchUINT32(0x4D4133 + 2, GetBaseID(FILE_TYPE_IFP));	// lea     eax, [esi+25575]
				CPatch::PatchUINT32(0x59512C + 1, GetBaseID(FILE_TYPE_IFP));	// add     eax, 25575
				CPatch::PatchUINT32(0x5A814C + 2, GetBaseID(FILE_TYPE_IFP));	// add     esi, 25575
				CPatch::PatchUINT32(0x5A8158 + 2, GetBaseID(FILE_TYPE_IFP));	// add     edi, 25575
				CPatch::PatchUINT32(0x5A816B + 2, GetBaseID(FILE_TYPE_IFP));	// add     edi, 25575
				CPatch::PatchUINT32(0x5A817E + 2, GetBaseID(FILE_TYPE_IFP));	// add     esi, 25575
				CPatch::PatchUINT32(0x5A8197 + 2, GetBaseID(FILE_TYPE_IFP));	// add     esi, 25575
				CPatch::PatchUINT32(0x5B63C4 + 1, GetBaseID(FILE_TYPE_IFP));	// add     eax, 25575
				CPatch::PatchUINT32(0x609AFF + 2, GetBaseID(FILE_TYPE_IFP));	// add     ecx, 25575
				CPatch::PatchUINT32(0x609BAA + 2, GetBaseID(FILE_TYPE_IFP));	// add     ecx, 25575
				CPatch::PatchUINT32(0x61B68D + 1, GetBaseID(FILE_TYPE_IFP));	// add     eax, 25575
				CPatch::PatchUINT32(0x61BEBA + 2, GetBaseID(FILE_TYPE_IFP));	// add     ecx, 25575
				CPatch::PatchUINT32(0x61C97A + 2, GetBaseID(FILE_TYPE_IFP));	// add     ecx, 25575
				CPatch::PatchUINT32(0x623D73 + 2, GetBaseID(FILE_TYPE_IFP));	// add     ecx, 25575
				CPatch::PatchUINT32(0x627B58 + 2, GetBaseID(FILE_TYPE_IFP));	// add     ecx, 25575
				CPatch::PatchUINT32(0x629B22 + 2, GetBaseID(FILE_TYPE_IFP));	// add     ecx, 25575
				CPatch::PatchUINT32(0x63175E + 1, GetBaseID(FILE_TYPE_IFP));	// add     eax, 25575
				CPatch::PatchUINT32(0x6381D6 + 2, GetBaseID(FILE_TYPE_IFP));	// add     edx, 25575
				CPatch::PatchUINT32(0x638218 + 2, GetBaseID(FILE_TYPE_IFP));	// add     edx, 25575
				CPatch::PatchUINT32(0x639D2F + 2, GetBaseID(FILE_TYPE_IFP));	// add     edx, 25575
				CPatch::PatchUINT32(0x639D4D + 2, GetBaseID(FILE_TYPE_IFP));	// add     ecx, 25575
				CPatch::PatchUINT32(0x64F88B + 2, GetBaseID(FILE_TYPE_IFP));	// lea     eax, [ebx+25575]
				CPatch::PatchUINT32(0x64F899 + 2, GetBaseID(FILE_TYPE_IFP));	// lea     ecx, [edi+25575]
				CPatch::PatchUINT32(0x6600F1 + 1, GetBaseID(FILE_TYPE_IFP));	// add     eax, 25575
				CPatch::PatchUINT32(0x660647 + 1, GetBaseID(FILE_TYPE_IFP));	// add     eax, 25575
				CPatch::PatchUINT32(0x660DB3 + 1, GetBaseID(FILE_TYPE_IFP));	// add     eax, 25575
				CPatch::PatchUINT32(0x6619FD + 1, GetBaseID(FILE_TYPE_IFP));	// add     eax, 25575
				CPatch::PatchUINT32(0x6623CC + 1, GetBaseID(FILE_TYPE_IFP));	// add     eax, 25575
				CPatch::PatchUINT32(0x662E5D + 1, GetBaseID(FILE_TYPE_IFP));	// add     eax, 25575
				CPatch::PatchUINT32(0x66384B + 1, GetBaseID(FILE_TYPE_IFP));	// add     eax, 25575
				CPatch::PatchUINT32(0x665002 + 1, GetBaseID(FILE_TYPE_IFP));	// add     eax, 25575
				CPatch::PatchUINT32(0x68234C + 1, GetBaseID(FILE_TYPE_IFP));	// add     eax, 25575
				CPatch::PatchUINT32(0x6823EF + 1, GetBaseID(FILE_TYPE_IFP));	// add     eax, 25575
				CPatch::PatchUINT32(0x6825DF + 1, GetBaseID(FILE_TYPE_IFP));	// add     eax, 25575
				CPatch::PatchUINT32(0x6873A2 + 1, GetBaseID(FILE_TYPE_IFP));	// add     eax, 25575
				CPatch::PatchUINT32(0x6874CA + 1, GetBaseID(FILE_TYPE_IFP));	// add     eax, 25575
				CPatch::PatchUINT32(0x689B8C + 2, GetBaseID(FILE_TYPE_IFP));	// add     ecx, 25575
				CPatch::PatchUINT32(0x691DA7 + 2, GetBaseID(FILE_TYPE_IFP));	// add     ecx, 25575
				CPatch::PatchUINT32(0x6930CC + 1, GetBaseID(FILE_TYPE_IFP));	// add     eax, 25575
				CPatch::PatchUINT32(0x1562411 + 1, GetBaseID(FILE_TYPE_IFP));	// add     eax, 25575
				CPatch::PatchUINT32(0x15624F1 + 1, GetBaseID(FILE_TYPE_IFP));	// add     eax, 25575
			}

			// -25775
			{
				CPatch::PatchUINT32(0x408A99 + 2, -GetBaseID(FILE_TYPE_IFP));	// lea     edx, [esi-63E7h]
				CPatch::PatchUINT32(0x408C28 + 2, -GetBaseID(FILE_TYPE_IFP));	// add     esi, 0FFFF9C19h
				CPatch::PatchUINT32(0x40CA46 + 2, -GetBaseID(FILE_TYPE_IFP));	// lea     eax, [esi-25575]
				CPatch::PatchUINT32(0x40CC1F + 2, -GetBaseID(FILE_TYPE_IFP));	// lea     ecx, [esi-25575]
			}

			// 180
			{
				CPatch::PatchUINT32(0x4D4143 + 2, GetFileIDcurrentLimit(FILE_TYPE_IFP));		// cmp     esi, 180
				CPatch::PatchUINT32(0x84C005 + 1, GetFileIDcurrentLimit(FILE_TYPE_IFP));		// mov     ecx, 180
			}
		}

		// RRR
		{
			// 25755
			{
				CPatch::PatchUINT32(0x408A91 + 2, GetBaseID(FILE_TYPE_RRR));	// cmp     esi, 25755
				CPatch::PatchUINT32(0x408C20 + 2, GetBaseID(FILE_TYPE_RRR));	// cmp     esi, 25755
				CPatch::PatchUINT32(0x408FA8 + 2, GetBaseID(FILE_TYPE_RRR));	// cmp     edi, 25755
				CPatch::PatchUINT32(0x40CA35 + 2, GetBaseID(FILE_TYPE_RRR));	// cmp     esi, 25755
				CPatch::PatchUINT32(0x40CB39 + 2, GetBaseID(FILE_TYPE_RRR));	// cmp     esi, 25755
				CPatch::PatchUINT32(0x40CC17 + 2, GetBaseID(FILE_TYPE_RRR));	// cmp     esi, 25755
				CPatch::PatchUINT32(0x40CDE3 + 2, GetBaseID(FILE_TYPE_RRR));	// cmp     esi, 25755
				CPatch::PatchUINT32(0x40D0BD + 2, GetBaseID(FILE_TYPE_RRR));	// cmp     esi, 25755

				CPatch::PatchUINT32(0x4594A1 + 2, GetBaseID(FILE_TYPE_RRR));	// add     edi, 25755
				CPatch::PatchUINT32(0x45A5F2 + 2, GetBaseID(FILE_TYPE_RRR));	// add     esi, 25755
				CPatch::PatchUINT32(0x5B63F0 + 1, GetBaseID(FILE_TYPE_RRR));	// add     eax, 25755
				CPatch::PatchUINT32(0x1564118 + 2, GetBaseID(FILE_TYPE_RRR));	// lea     eax, [edi+25755]
				CPatch::PatchUINT32(0x156890F + 2, GetBaseID(FILE_TYPE_RRR));	// lea     eax, [edi+25755]
				CPatch::PatchUINT32(0x156D134 + 1, GetBaseID(FILE_TYPE_RRR));	// add     eax, 25755
				CPatch::PatchUINT32(0x156E788 + 2, GetBaseID(FILE_TYPE_RRR));	// lea     eax, [edi+25755]
			}

			// -25755
			{
				CPatch::PatchUINT32(0x40CA94 + 2, -GetBaseID(FILE_TYPE_RRR));	// lea     edx, [esi-25755]
			}
		}

		// SCM
		{
			// 26230
			{
				CPatch::PatchUINT32(0x408AAA + 2, GetBaseID(FILE_TYPE_SCM));	// cmp     esi, 26230
				CPatch::PatchUINT32(0x408C41 + 2, GetBaseID(FILE_TYPE_SCM));	// cmp     esi, 26230
				CPatch::PatchUINT32(0x40C1C4 + 1, GetBaseID(FILE_TYPE_SCM));	// add     eax, 26230
				CPatch::PatchUINT32(0x40CA8B + 2, GetBaseID(FILE_TYPE_SCM));	// cmp     esi, 26230
				CPatch::PatchUINT32(0x40CB41 + 2, GetBaseID(FILE_TYPE_SCM));	// cmp     esi, 26230
				CPatch::PatchUINT32(0x40D0FA + 2, GetBaseID(FILE_TYPE_SCM));	// cmp     esi, 26230
				CPatch::PatchUINT32(0x40D102 + 2, GetBaseID(FILE_TYPE_SCM));	// sub     esi, 26230
				CPatch::PatchUINT32(0x40D25A + 2, GetBaseID(FILE_TYPE_SCM));	// add     esi, 26230
				CPatch::PatchUINT32(0x46A574 + 2, GetBaseID(FILE_TYPE_SCM));	// add     esi, 26230

				CPatch::PatchUINT32(0x46CE60 + 1, GetBaseID(FILE_TYPE_SCM));	// add     eax, 26230
				CPatch::PatchUINT32(0x46CFD4 + 1, GetBaseID(FILE_TYPE_SCM));	// add     eax, 26230

				CPatch::PatchUINT32(0x4700E4 + 1, GetBaseID(FILE_TYPE_SCM));	// add     eax, 26230
				CPatch::PatchUINT32(0x475032 + 1, GetBaseID(FILE_TYPE_SCM));	// add     eax, 26230
				CPatch::PatchUINT32(0x59F6AC + 1, GetBaseID(FILE_TYPE_SCM));	// add     eax, 26230
				CPatch::PatchUINT32(0x5B641E + 1, GetBaseID(FILE_TYPE_SCM));	// add     eax, 26230
				CPatch::PatchUINT32(0x5E8675 + 1, GetBaseID(FILE_TYPE_SCM));	// add     eax, 26230


				CPatch::PatchUINT32(0x156BFDF + 2, GetBaseID(FILE_TYPE_SCM));	// add     ebp, 26230
				CPatch::PatchUINT32(0x156D041 + 2, GetBaseID(FILE_TYPE_SCM));	// lea     eax, [esi+26230]
				CPatch::PatchUINT32(0x156D9AE + 1, GetBaseID(FILE_TYPE_SCM));	// add     eax, 26230
			}

			// -26230
			{
				CPatch::PatchUINT32(0x408AB2 + 2, -GetBaseID(FILE_TYPE_SCM));	// lea     eax, [esi-6676h]
				CPatch::PatchUINT32(0x408C49 + 2, -GetBaseID(FILE_TYPE_SCM));	// add     esi, 0FFFF998Ah
				CPatch::PatchUINT32(0x40CAA6 + 2, -GetBaseID(FILE_TYPE_SCM));	// lea     eax, [esi-6676h]
			}
		}
	}
	#endif

	// Patch limits for different types of files.
	#if TRUE
	{
		// TXD
		#if TRUE
		{
			// CTxdStore__Init
			CPatch::PatchUINT32(0x731F5F + 1, GetBaseID(FILE_TYPE_COL) - GetBaseID(FILE_TYPE_TXD));
		}
		#endif

		// COL
		#if TRUE
		{
			// CColFileMgr__initialize
			CPatch::PatchUINT32(0x411457 + 1, GetBaseID(FILE_TYPE_IPL) - GetBaseID(FILE_TYPE_COL));
		}
		#endif

		// IPL
		#if TRUE
		{
			CPatch::PatchUINT32(0x405F25 + 1, GetBaseID(FILE_TYPE_DAT) - GetBaseID(FILE_TYPE_IPL));
		}
		#endif

		// IFP
		#if TRUE
		{
			CPatch::PatchPointer(0x40C745 + 2, &CAnimManager__ms_aAnimBlocks.gta_sa->loaded);	// mov     dl, ds:_ZN12CAnimManager14ms_aAnimBlocksE.loaded[eax]

			CPatch::PatchPointer(0x4C51CD + 1, CAnimManager__ms_aAnimBlocks.gta_sa);	// add     eax, offset _ZN12CAnimManager14ms_aAnimBlocksE; CAnimManager::ms_aAnimBlocks
			CPatch::PatchPointer(0x4D3952 + 1, CAnimManager__ms_aAnimBlocks.gta_sa);	// mov     edi, offset _ZN12CAnimManager14ms_aAnimBlocksE; CAnimManager::ms_aAnimBlocks
			CPatch::PatchPointer(0x4D397F + 1, CAnimManager__ms_aAnimBlocks.gta_sa);	// add     eax, offset _ZN12CAnimManager14ms_aAnimBlocksE; CAnimManager::ms_aAnimBlocks
			CPatch::PatchPointer(0x4D39A5 + 1, CAnimManager__ms_aAnimBlocks.gta_sa);	// sub     eax, offset _ZN12CAnimManager14ms_aAnimBlocksE; CAnimManager::ms_aAnimBlocks
			CPatch::PatchPointer(0x4D3CEF + 1, CAnimManager__ms_aAnimBlocks.gta_sa);	// mov     ebx, offset _ZN12CAnimManager14ms_aAnimBlocksE; CAnimManager::ms_aAnimBlocks
			CPatch::PatchPointer(0x4D3D1A + 2, CAnimManager__ms_aAnimBlocks.gta_sa);	// add     edi, offset _ZN12CAnimManager14ms_aAnimBlocksE; CAnimManager::ms_aAnimBlocks
			CPatch::PatchPointer(0x4D3E71 + 2, CAnimManager__ms_aAnimBlocks.gta_sa);	// add     esi, offset _ZN12CAnimManager14ms_aAnimBlocksE; CAnimManager::ms_aAnimBlocks
			CPatch::PatchPointer(0x4D3E99 + 1, CAnimManager__ms_aAnimBlocks.gta_sa);	// sub     eax, offset _ZN12CAnimManager14ms_aAnimBlocksE; CAnimManager::ms_aAnimBlocks
			CPatch::PatchPointer(0x4D3EDC + 2, &CAnimManager__ms_aAnimBlocks.gta_sa->startAnimation);	// mov     edx, ds:_ZN12CAnimManager14ms_aAnimBlocksE.startAnimation[eax]
			CPatch::PatchPointer(0x4D3EE8 + 2, &CAnimManager__ms_aAnimBlocks.gta_sa->animationCount);	// mov     edx, ds:_ZN12CAnimManager14ms_aAnimBlocksE.animationCount[eax]
			CPatch::PatchPointer(0x4D3F00 + 2, &CAnimManager__ms_aAnimBlocks.gta_sa->startAnimation);	// mov     eax, ds:_ZN12CAnimManager14ms_aAnimBlocksE.startAnimation[eax]
			CPatch::PatchPointer(0x4D3F22 + 2, &CAnimManager__ms_aAnimBlocks.gta_sa->animationCount);	// mov     edx, ds:_ZN12CAnimManager14ms_aAnimBlocksE.animationCount[eax]
			CPatch::PatchPointer(0x4D3F30 + 2, &CAnimManager__ms_aAnimBlocks.gta_sa->loaded);	// mov     ds:_ZN12CAnimManager14ms_aAnimBlocksE.loaded[ecx], 0
			CPatch::PatchPointer(0x4D3F51 + 2, CAnimManager__ms_aAnimBlocks.gta_sa);	// add     esi, offset _ZN12CAnimManager14ms_aAnimBlocksE; CAnimManager::ms_aAnimBlocks
			CPatch::PatchPointer(0x4D3FB7 + 1, CAnimManager__ms_aAnimBlocks.gta_sa);	// add     eax, offset _ZN12CAnimManager14ms_aAnimBlocksE; CAnimManager::ms_aAnimBlocks
			CPatch::PatchPointer(0x4D3FD7 + 1, CAnimManager__ms_aAnimBlocks.gta_sa);	// add     eax, offset _ZN12CAnimManager14ms_aAnimBlocksE; CAnimManager::ms_aAnimBlocks
			CPatch::PatchPointer(0x4D3FF7 + 1, CAnimManager__ms_aAnimBlocks.gta_sa);	// add     eax, offset _ZN12CAnimManager14ms_aAnimBlocksE; CAnimManager::ms_aAnimBlocks
			CPatch::PatchPointer(0x4D4017 + 3, &CAnimManager__ms_aAnimBlocks.gta_sa->usageCount);	// movsx   eax, ds:_ZN12CAnimManager14ms_aAnimBlocksE.usageCount[eax]
			CPatch::PatchPointer(0x4D48D4 + 2, CAnimManager__ms_aAnimBlocks.gta_sa);	// add     esi, offset _ZN12CAnimManager14ms_aAnimBlocksE; CAnimManager::ms_aAnimBlocks
			CPatch::PatchPointer(0x4D4947 + 1, CAnimManager__ms_aAnimBlocks.gta_sa);	// sub     eax, offset _ZN12CAnimManager14ms_aAnimBlocksE; CAnimManager::ms_aAnimBlocks
			CPatch::PatchPointer(0x4D51DA + 2, CAnimManager__ms_aAnimBlocks.gta_sa);	// add     esi, offset _ZN12CAnimManager14ms_aAnimBlocksE; CAnimManager::ms_aAnimBlocks
			CPatch::PatchPointer(0x4D5244 + 2, CAnimManager__ms_aAnimBlocks.gta_sa);	// sub     esi, offset _ZN12CAnimManager14ms_aAnimBlocksE; CAnimManager::ms_aAnimBlocks

			CPatch::PatchPointer(0x595123 + 2, &CAnimManager__ms_aAnimBlocks.gta_sa->loaded);	// cmp     ds:_ZN12CAnimManager14ms_aAnimBlocksE.loaded[ecx], 1
			CPatch::PatchPointer(0x598105 + 2, &CAnimManager__ms_aAnimBlocks.gta_sa->loaded);	// mov     cl, ds:_ZN12CAnimManager14ms_aAnimBlocksE.loaded[eax]
			CPatch::PatchPointer(0x598129 + 2, &CAnimManager__ms_aAnimBlocks.gta_sa->loaded);	// mov     cl, ds:_ZN12CAnimManager14ms_aAnimBlocksE.loaded[eax]
			CPatch::PatchPointer(0x59814D + 2, &CAnimManager__ms_aAnimBlocks.gta_sa->loaded);	// mov     cl, ds:_ZN12CAnimManager14ms_aAnimBlocksE.loaded[eax]
			CPatch::PatchPointer(0x598166 + 2, &CAnimManager__ms_aAnimBlocks.gta_sa->loaded);	// cmp     ds:_ZN12CAnimManager14ms_aAnimBlocksE.loaded[eax], 1
			CPatch::PatchPointer(0x6011F9 + 1, CAnimManager__ms_aAnimBlocks.gta_sa);	// add     eax, offset _ZN12CAnimManager14ms_aAnimBlocksE; CAnimManager::ms_aAnimBlocks
			CPatch::PatchPointer(0x6093F6 + 1, CAnimManager__ms_aAnimBlocks.gta_sa);	// sub     eax, offset _ZN12CAnimManager14ms_aAnimBlocksE; CAnimManager::ms_aAnimBlocks
			CPatch::PatchPointer(0x60942E + 1, CAnimManager__ms_aAnimBlocks.gta_sa);	// sub     eax, offset _ZN12CAnimManager14ms_aAnimBlocksE; CAnimManager::ms_aAnimBlocks
			CPatch::PatchPointer(0x609AF1 + 2, CAnimManager__ms_aAnimBlocks.gta_sa);	// sub     ecx, offset _ZN12CAnimManager14ms_aAnimBlocksE; CAnimManager::ms_aAnimBlocks
			CPatch::PatchPointer(0x609B98 + 2, CAnimManager__ms_aAnimBlocks.gta_sa);	// sub     ecx, offset _ZN12CAnimManager14ms_aAnimBlocksE; CAnimManager::ms_aAnimBlocks
			CPatch::PatchPointer(0x609BD8 + 1, CAnimManager__ms_aAnimBlocks.gta_sa);	// sub     eax, offset _ZN12CAnimManager14ms_aAnimBlocksE; CAnimManager::ms_aAnimBlocks
			CPatch::PatchPointer(0x609C14 + 2, CAnimManager__ms_aAnimBlocks.gta_sa);	// sub     edx, offset _ZN12CAnimManager14ms_aAnimBlocksE; CAnimManager::ms_aAnimBlocks
			CPatch::PatchPointer(0x61BEAE + 1, CAnimManager__ms_aAnimBlocks.gta_sa);	// add     eax, offset _ZN12CAnimManager14ms_aAnimBlocksE; CAnimManager::ms_aAnimBlocks
			CPatch::PatchPointer(0x61C5AC + 1, CAnimManager__ms_aAnimBlocks.gta_sa);	// sub     eax, offset _ZN12CAnimManager14ms_aAnimBlocksE; CAnimManager::ms_aAnimBlocks
			CPatch::PatchPointer(0x61C831 + 1, CAnimManager__ms_aAnimBlocks.gta_sa);	// sub     eax, offset _ZN12CAnimManager14ms_aAnimBlocksE; CAnimManager::ms_aAnimBlocks
			CPatch::PatchPointer(0x61C90B + 2, CAnimManager__ms_aAnimBlocks.gta_sa);	// sub     ecx, offset _ZN12CAnimManager14ms_aAnimBlocksE; CAnimManager::ms_aAnimBlocks
			CPatch::PatchPointer(0x61C956 + 2, CAnimManager__ms_aAnimBlocks.gta_sa);	// sub     ecx, offset _ZN12CAnimManager14ms_aAnimBlocksE; CAnimManager::ms_aAnimBlocks
			CPatch::PatchPointer(0x621901 + 1, CAnimManager__ms_aAnimBlocks.gta_sa);	// sub     eax, offset _ZN12CAnimManager14ms_aAnimBlocksE; CAnimManager::ms_aAnimBlocks
			CPatch::PatchPointer(0x623CC4 + 2, CAnimManager__ms_aAnimBlocks.gta_sa);	// sub     edx, offset _ZN12CAnimManager14ms_aAnimBlocksE; CAnimManager::ms_aAnimBlocks
			CPatch::PatchPointer(0x623D12 + 2, CAnimManager__ms_aAnimBlocks.gta_sa);	// sub     ecx, offset _ZN12CAnimManager14ms_aAnimBlocksE; CAnimManager::ms_aAnimBlocks
			CPatch::PatchPointer(0x627B4A + 2, CAnimManager__ms_aAnimBlocks.gta_sa);	// sub     ecx, offset _ZN12CAnimManager14ms_aAnimBlocksE; CAnimManager::ms_aAnimBlocks
			CPatch::PatchPointer(0x629AB5 + 1, CAnimManager__ms_aAnimBlocks.gta_sa);	// sub     eax, offset _ZN12CAnimManager14ms_aAnimBlocksE; CAnimManager::ms_aAnimBlocks
			CPatch::PatchPointer(0x629B02 + 2, CAnimManager__ms_aAnimBlocks.gta_sa);	// sub     ecx, offset _ZN12CAnimManager14ms_aAnimBlocksE; CAnimManager::ms_aAnimBlocks

			CPatch::PatchPointer(0x631750 + 2, &CAnimManager__ms_aAnimBlocks.gta_sa->loaded);	// mov     dl, ds:_ZN12CAnimManager14ms_aAnimBlocksE.loaded[ecx]
			CPatch::PatchPointer(0x6600E0 + 2, CAnimManager__ms_aAnimBlocks.gta_sa);	// add     ecx, offset _ZN12CAnimManager14ms_aAnimBlocksE; CAnimManager::ms_aAnimBlocks
			CPatch::PatchPointer(0x660636 + 2, CAnimManager__ms_aAnimBlocks.gta_sa);	// add     ecx, offset _ZN12CAnimManager14ms_aAnimBlocksE; CAnimManager::ms_aAnimBlocks
			CPatch::PatchPointer(0x660DA5 + 2, &CAnimManager__ms_aAnimBlocks.gta_sa->loaded);	// mov     dl, ds:_ZN12CAnimManager14ms_aAnimBlocksE.loaded[ecx]
			CPatch::PatchPointer(0x6619EE + 2, &CAnimManager__ms_aAnimBlocks.gta_sa->loaded);	// mov     dl, ds:_ZN12CAnimManager14ms_aAnimBlocksE.loaded[ecx]
			CPatch::PatchPointer(0x6623BB + 2, CAnimManager__ms_aAnimBlocks.gta_sa);	// add     ecx, offset _ZN12CAnimManager14ms_aAnimBlocksE; CAnimManager::ms_aAnimBlocks
			CPatch::PatchPointer(0x662E4C + 2, CAnimManager__ms_aAnimBlocks.gta_sa);	// add     ecx, offset _ZN12CAnimManager14ms_aAnimBlocksE; CAnimManager::ms_aAnimBlocks
			CPatch::PatchPointer(0x66383A + 2, CAnimManager__ms_aAnimBlocks.gta_sa);	// add     ecx, offset _ZN12CAnimManager14ms_aAnimBlocksE; CAnimManager::ms_aAnimBlocks
			CPatch::PatchPointer(0x664FF4 + 2, &CAnimManager__ms_aAnimBlocks.gta_sa->loaded);	// mov     dl, ds:_ZN12CAnimManager14ms_aAnimBlocksE.loaded[ecx]
			CPatch::PatchPointer(0x68233E + 2, &CAnimManager__ms_aAnimBlocks.gta_sa->loaded);	// mov     dl, ds:_ZN12CAnimManager14ms_aAnimBlocksE.loaded[ecx]
			CPatch::PatchPointer(0x6823E1 + 2, &CAnimManager__ms_aAnimBlocks.gta_sa->loaded);	// mov     dl, ds:_ZN12CAnimManager14ms_aAnimBlocksE.loaded[ecx]
			CPatch::PatchPointer(0x6825D1 + 2, &CAnimManager__ms_aAnimBlocks.gta_sa->loaded);	// mov     dl, ds:_ZN12CAnimManager14ms_aAnimBlocksE.loaded[ecx]
			CPatch::PatchPointer(0x687386 + 2, CAnimManager__ms_aAnimBlocks.gta_sa);	// add     esi, offset _ZN12CAnimManager14ms_aAnimBlocksE; CAnimManager::ms_aAnimBlocks
			CPatch::PatchPointer(0x688A3D + 1, CAnimManager__ms_aAnimBlocks.gta_sa);	// sub     eax, offset _ZN12CAnimManager14ms_aAnimBlocksE; CAnimManager::ms_aAnimBlocks
			CPatch::PatchPointer(0x689B7F + 2, CAnimManager__ms_aAnimBlocks.gta_sa);	// sub     ecx, offset _ZN12CAnimManager14ms_aAnimBlocksE; CAnimManager::ms_aAnimBlocks
			CPatch::PatchPointer(0x69152A + 1, CAnimManager__ms_aAnimBlocks.gta_sa);	// sub     eax, offset _ZN12CAnimManager14ms_aAnimBlocksE; CAnimManager::ms_aAnimBlocks
			CPatch::PatchPointer(0x6915FA + 1, CAnimManager__ms_aAnimBlocks.gta_sa);	// sub     eax, offset _ZN12CAnimManager14ms_aAnimBlocksE; CAnimManager::ms_aAnimBlocks
			CPatch::PatchPointer(0x6916A6 + 1, CAnimManager__ms_aAnimBlocks.gta_sa);	// sub     eax, offset _ZN12CAnimManager14ms_aAnimBlocksE; CAnimManager::ms_aAnimBlocks
			CPatch::PatchPointer(0x691AA1 + 1, CAnimManager__ms_aAnimBlocks.gta_sa);	// sub     eax, offset _ZN12CAnimManager14ms_aAnimBlocksE; CAnimManager::ms_aAnimBlocks
			CPatch::PatchPointer(0x691D8B + 2, CAnimManager__ms_aAnimBlocks.gta_sa);	// sub     ecx, offset _ZN12CAnimManager14ms_aAnimBlocksE; CAnimManager::ms_aAnimBlocks
			CPatch::PatchPointer(0x69305A + 1, CAnimManager__ms_aAnimBlocks.gta_sa);	// sub     eax, offset _ZN12CAnimManager14ms_aAnimBlocksE; CAnimManager::ms_aAnimBlocks
			CPatch::PatchPointer(0x6BF4AF + 2, &CAnimManager__ms_aAnimBlocks.gta_sa->animationStyle);	// mov     eax, ds:_ZN12CAnimManager14ms_aAnimBlocksE.animationStyle[eax]
			CPatch::PatchPointer(0x6BF876 + 2, &CAnimManager__ms_aAnimBlocks.gta_sa->animationStyle);	// mov     eax, ds:_ZN12CAnimManager14ms_aAnimBlocksE.animationStyle[eax]
			CPatch::PatchPointer(0x84C000 + 1, CAnimManager__ms_aAnimBlocks.gta_sa);	// mov     eax, offset _ZN12CAnimManager14ms_aAnimBlocksE; CAnimManager::ms_aAnimBlocks
		}
		#endif

		// RRR
		#if TRUE
		{
			CPatch::PatchPointer(0x4593D9 + 1, &CVehicleRecording__StreamingArray.gta_sa->usageCount);	// mov     eax, offset _ZN17CVehicleRecording14StreamingArrayE.usageCount
			CPatch::PatchPointer(0x4593E9 + 1, &CVehicleRecording__StreamingArray.gta_sa[GetFileIDcurrentLimit(FILE_TYPE_RRR)].usageCount);	// cmp     eax, offset flt_97F63C	; sub_459390
			CPatch::PatchPointer(0x45948D + 2, &CVehicleRecording__StreamingArray.gta_sa->usageCount);	// dec     ds:_ZN17CVehicleRecording14StreamingArrayE.usageCount[esi]
			CPatch::PatchPointer(0x459495 + 2, &CVehicleRecording__StreamingArray.gta_sa->pData);	// mov     eax, ds:_ZN17CVehicleRecording14StreamingArrayE.pData[esi]
			CPatch::PatchPointer(0x4594A8 + 2, &CVehicleRecording__StreamingArray.gta_sa->pData);	// mov     ds:_ZN17CVehicleRecording14StreamingArrayE.pData[esi], ebx
			CPatch::PatchPointer(0x45A0F8 + 2, &CVehicleRecording__StreamingArray.gta_sa->field_8);	// mov     ecx, ds:_ZN17CVehicleRecording14StreamingArrayE.field_8[edi]
			CPatch::PatchPointer(0x45A0FE + 2, &CVehicleRecording__StreamingArray.gta_sa->pData);	// mov     eax, ds:_ZN17CVehicleRecording14StreamingArrayE.pData[edi]
			CPatch::PatchPointer(0x45A13D + 2, &CVehicleRecording__StreamingArray.gta_sa->field_8);	// mov     ecx, ds:_ZN17CVehicleRecording14StreamingArrayE.field_8[edi]
			CPatch::PatchPointer(0x45A1B0 + 1, &CVehicleRecording__StreamingArray.gta_sa->pData);	// mov     eax, offset _ZN17CVehicleRecording14StreamingArrayE.pData
			CPatch::PatchPointer(0x45A1C9 + 1, &CVehicleRecording__StreamingArray.gta_sa[GetFileIDcurrentLimit(FILE_TYPE_RRR)].pData);	// cmp     eax, offset flt_97F634	; sub_45A1B0
			CPatch::PatchPointer(0x45A5DE + 2, &CVehicleRecording__StreamingArray.gta_sa->usageCount);	// dec     ds:_ZN17CVehicleRecording14StreamingArrayE.usageCount[edi]
			CPatch::PatchPointer(0x45A5E6 + 2, &CVehicleRecording__StreamingArray.gta_sa->pData);	// mov     eax, ds:_ZN17CVehicleRecording14StreamingArrayE.pData[edi]
			CPatch::PatchPointer(0x45A5F9 + 2, &CVehicleRecording__StreamingArray.gta_sa->pData);	// mov     ds:_ZN17CVehicleRecording14StreamingArrayE.pData[edi], ebx
			CPatch::PatchPointer(0x1560EA1 + 1, CVehicleRecording__StreamingArray.gta_sa);	// mov     ecx, offset _ZN17CVehicleRecording14StreamingArrayE; CVehicleRecording::StreamingArray
			CPatch::PatchPointer(0x1560EB7 + 2, &CVehicleRecording__StreamingArray.gta_sa->pData);	// mov     edx, ds:_ZN17CVehicleRecording14StreamingArrayE.pData[eax]
			CPatch::PatchPointer(0x15640F9 + 1, &CVehicleRecording__StreamingArray.gta_sa->pData);	// mov     esi, offset _ZN17CVehicleRecording14StreamingArrayE.pData
			CPatch::PatchPointer(0x1564131 + 2, &CVehicleRecording__StreamingArray.gta_sa[GetFileIDcurrentLimit(FILE_TYPE_RRR)].pData);	// cmp     esi, offset flt_97F634	; sub_15640F0
			CPatch::PatchPointer(0x1565F24 + 1, CVehicleRecording__StreamingArray.gta_sa);	// mov     ecx, offset _ZN17CVehicleRecording14StreamingArrayE; CVehicleRecording::StreamingArray
			CPatch::PatchPointer(0x1566005 + 2, &CVehicleRecording__StreamingArray.gta_sa->pData);	// mov     edx, ds:_ZN17CVehicleRecording14StreamingArrayE.pData[eax]
			CPatch::PatchPointer(0x1566012 + 2, &CVehicleRecording__StreamingArray.gta_sa->field_8);	// mov     edx, ds:_ZN17CVehicleRecording14StreamingArrayE.field_8[eax]
			CPatch::PatchPointer(0x1566021 + 2, &CVehicleRecording__StreamingArray.gta_sa->usageCount);	// inc     ds:_ZN17CVehicleRecording14StreamingArrayE.usageCount[eax]
			CPatch::PatchPointer(0x15688F4 + 1, &CVehicleRecording__StreamingArray.gta_sa -> pData);	// mov     esi, offset _ZN17CVehicleRecording14StreamingArrayE.pData	; sub_15688F0
			CPatch::PatchPointer(0x1568928 + 2, &CVehicleRecording__StreamingArray.gta_sa[GetFileIDcurrentLimit(FILE_TYPE_RRR)].pData);	// cmp     esi, offset flt_97F634	; sub_15688F0
			CPatch::PatchPointer(0x156D121 + 1, CVehicleRecording__StreamingArray.gta_sa);	// mov     ecx, offset _ZN17CVehicleRecording14StreamingArrayE; CVehicleRecording::StreamingArray
			CPatch::PatchPointer(0x156E774 + 1, &CVehicleRecording__StreamingArray.gta_sa->pData);	// mov     esi, offset _ZN17CVehicleRecording14StreamingArrayE.pData
			CPatch::PatchPointer(0x156E7A1 + 2, &CVehicleRecording__StreamingArray.gta_sa[GetFileIDcurrentLimit(FILE_TYPE_RRR)].pData);	// cmp     esi, offset flt_97F634	; sub_156E770
			CPatch::PatchPointer(0x156F157 + 2, &CVehicleRecording__StreamingArray.gta_sa->number);	// mov     dword ptr ds:_ZN17CVehicleRecording14StreamingArrayE.number[eax], edx; CVehicleRecording::StreamingArray
			CPatch::PatchPointer(0x156F15E + 2, &CVehicleRecording__StreamingArray.gta_sa->pData);	// mov     ds:_ZN17CVehicleRecording14StreamingArrayE.pData[eax], 0
			CPatch::PatchPointer(0x156F533 + 2, &CVehicleRecording__StreamingArray.gta_sa->pData);	// mov     ds:_ZN17CVehicleRecording14StreamingArrayE.pData[esi], eax
			CPatch::PatchPointer(0x156F541 + 2, &CVehicleRecording__StreamingArray.gta_sa->field_8);	// mov     ds:_ZN17CVehicleRecording14StreamingArrayE.field_8[esi], eax
			CPatch::PatchPointer(0x156F54C + 2, &CVehicleRecording__StreamingArray.gta_sa->field_8);	// mov     ecx, ds:_ZN17CVehicleRecording14StreamingArrayE.field_8[esi]
			CPatch::PatchPointer(0x156F55B + 2, &CVehicleRecording__StreamingArray.gta_sa->pData);	// mov     ecx, ds:_ZN17CVehicleRecording14StreamingArrayE.pData[esi]
			CPatch::PatchPointer(0x156F573 + 2, &CVehicleRecording__StreamingArray.gta_sa->field_8);	// mov     ds:_ZN17CVehicleRecording14StreamingArrayE.field_8[esi], eax
			CPatch::PatchPointer(0x156F579 + 2, &CVehicleRecording__StreamingArray.gta_sa->field_8);	// mov     edx, ds:_ZN17CVehicleRecording14StreamingArrayE.field_8[esi]
		}
		#endif

		// SCM
		#if TRUE
		{
			// Patch pointers to CTheScripts::StreamedScripts
			{
				CPatch::PatchPointer(0x408AB9 + 1, CTheScripts__StreamedScripts.ptr);	// mov     ecx, offset _ZN11CTheScripts15StreamedScriptsE; CTheScripts::StreamedScripts
				CPatch::PatchPointer(0x408C50 + 1, CTheScripts__StreamedScripts.ptr);	// mov     ecx, offset _ZN11CTheScripts15StreamedScriptsE; CTheScripts::StreamedScripts
				CPatch::PatchPointer(0x40CAAE + 1, CTheScripts__StreamedScripts.ptr);	// mov     ecx, offset _ZN11CTheScripts15StreamedScriptsE; CTheScripts::StreamedScripts
				CPatch::PatchPointer(0x40D10D + 2, &CTheScripts__StreamedScriptsArray->status);	// mov     al, ds:_ZN11CTheScripts15StreamedScriptsE.scripts.status[ecx]


				CPatch::PatchPointer(0x465ACF + 1, CTheScripts__StreamedScripts.ptr);	// mov     ecx, offset _ZN11CTheScripts15StreamedScriptsE; CTheScripts::StreamedScripts
				CPatch::PatchPointer(0x465ADC + 2, &CTheScripts__StreamedScriptsArray->status);	// lea     ecx, _ZN11CTheScripts15StreamedScriptsE.scripts.status[eax]

				CPatch::PatchPointer(0x468F92 + 1, CTheScripts__StreamedScripts.ptr);	// mov     ecx, offset _ZN11CTheScripts15StreamedScriptsE; CTheScripts::StreamedScripts
				CPatch::PatchPointer(0x46B281 + 1, CTheScripts__StreamedScripts.ptr);	// mov     ecx, offset _ZN11CTheScripts15StreamedScriptsE; CTheScripts::StreamedScripts
				CPatch::PatchPointer(0x471CA1 + 1, CTheScripts__StreamedScripts.ptr);	// mov     ecx, offset _ZN11CTheScripts15StreamedScriptsE; CTheScripts::StreamedScripts
				CPatch::PatchPointer(0x47237D + 1, CTheScripts__StreamedScripts.ptr);	// mov     ecx, offset _ZN11CTheScripts15StreamedScriptsE; CTheScripts::StreamedScripts
				CPatch::PatchPointer(0x475020 + 1, CTheScripts__StreamedScripts.ptr);	// mov     ecx, offset _ZN11CTheScripts15StreamedScriptsE; CTheScripts::StreamedScripts
				CPatch::PatchPointer(0x47505B + 1, CTheScripts__StreamedScripts.ptr);	// mov     ecx, offset _ZN11CTheScripts15StreamedScriptsE; CTheScripts::StreamedScripts
				CPatch::PatchPointer(0x47669B + 1, CTheScripts__StreamedScripts.ptr);	// mov     ecx, offset _ZN11CTheScripts15StreamedScriptsE; CTheScripts::StreamedScripts
				CPatch::PatchPointer(0x4766D8 + 1, CTheScripts__StreamedScripts.ptr);	// mov     ecx, offset _ZN11CTheScripts15StreamedScriptsE; CTheScripts::StreamedScripts
				CPatch::PatchPointer(0x476754 + 1, CTheScripts__StreamedScripts.ptr);	// mov     ecx, offset _ZN11CTheScripts15StreamedScriptsE; CTheScripts::StreamedScripts
				CPatch::PatchPointer(0x476763 + 1, CTheScripts__StreamedScripts.ptr);	// mov     ecx, offset _ZN11CTheScripts15StreamedScriptsE; CTheScripts::StreamedScripts
				CPatch::PatchPointer(0x476D23 + 1, CTheScripts__StreamedScripts.ptr);	// mov     ecx, offset _ZN11CTheScripts15StreamedScriptsE; CTheScripts::StreamedScripts
				CPatch::PatchPointer(0x476D33 + 3, &CTheScripts__StreamedScriptsArray->status);	// movzx   edx, ds:_ZN11CTheScripts15StreamedScriptsE.scripts.status[eax]
				CPatch::PatchPointer(0x476D50 + 1, CTheScripts__StreamedScripts.ptr);	// mov     ecx, offset _ZN11CTheScripts15StreamedScriptsE; CTheScripts::StreamedScripts
				CPatch::PatchPointer(0x476DAA + 1, CTheScripts__StreamedScripts.ptr);	// mov     ecx, offset _ZN11CTheScripts15StreamedScriptsE; CTheScripts::StreamedScripts



				CPatch::PatchPointer(0x5B6414 + 1, CTheScripts__StreamedScripts.ptr);	// mov     ecx, offset _ZN11CTheScripts15StreamedScriptsE; CTheScripts::StreamedScripts
				CPatch::PatchPointer(0x5B8DEF + 1, CTheScripts__StreamedScripts.ptr);	// mov     ecx, offset _ZN11CTheScripts15StreamedScriptsE; CTheScripts::StreamedScripts

				CPatch::PatchPointer(0x156D003 + 1, CTheScripts__StreamedScripts.ptr);	// mov     ecx, offset _ZN11CTheScripts15StreamedScriptsE; CTheScripts::StreamedScripts
				CPatch::PatchPointer(0x156D961 + 2, &CTheScripts__StreamedScriptsArray->status);	// mov     al, ds:_ZN11CTheScripts15StreamedScriptsE.scripts.status[ecx]
				CPatch::PatchPointer(0x156D969 + 2, &CTheScripts__StreamedScriptsArray->status);	// lea     ecx, _ZN11CTheScripts15StreamedScriptsE.scripts.status[ecx]
				CPatch::PatchPointer(0x156D9D4 + 2, &CTheScripts__StreamedScriptsArray->status);	// mov     cl, ds:_ZN11CTheScripts15StreamedScriptsE.scripts.status[eax]
				CPatch::PatchPointer(0x156D9DA + 2, &CTheScripts__StreamedScriptsArray->status);	// lea     eax, _ZN11CTheScripts15StreamedScriptsE.scripts.status[eax]
			}

			// Patch functions
			{
				// CStreaming::ReInit_HOODLUM
				CPatch::PatchUINT8(0x156D051 + 2, GetFileIDcurrentLimit(FILE_TYPE_SCM));

				// CStreamedScripts::ReInitialise
				CPatch::PatchUINT8(0x1565AD3 + 1, GetFileIDcurrentLimit(FILE_TYPE_SCM));

				// CStreamedScripts::FindStreamedScriptQuiet
				{
					CPatch::PatchUINT32(0x4706F7 + 3, (int)&CTheScripts__StreamedScriptsFooter->countOfScripts - (int)CTheScripts__StreamedScriptsArray);	// cmp     [ebx+0A44h], si
					CPatch::PatchUINT32(0x470716 + 3, (int)&CTheScripts__StreamedScriptsFooter->countOfScripts - (int)CTheScripts__StreamedScriptsArray);	// movzx   eax, [ebx+CStreamedScripts.countOfScripts]
				}

				// CStreamedScripts::GetProperIndexFromIndexUsedByScript_HOODLUM
				CPatch::PatchUINT16(0x47082E + 2, GetFileIDcurrentLimit(FILE_TYPE_SCM));

				// CStreamedScripts::GetStreamedScriptWithThisStartAddress
				CPatch::PatchUINT16(0x470917 + 2, GetFileIDcurrentLimit(FILE_TYPE_SCM));

				// CStreamedScripts::Initialise_HOODLUM
				{
					CPatch::PatchUINT32(0x1562434 + 1, GetFileIDcurrentLimit(FILE_TYPE_SCM));
					CPatch::PatchUINT32(0x1562457 + 2, (int)&CTheScripts__StreamedScriptsFooter->largestExternalSize - (int)CTheScripts__StreamedScriptsArray);
					CPatch::PatchUINT32(0x156245D + 3, (int)&CTheScripts__StreamedScriptsFooter->countOfScripts - (int)CTheScripts__StreamedScriptsArray);
				}

				// CStreamedScripts::ReadStreamedScriptData_HOODLUM
				{
					CPatch::PatchUINT32(0x156F473 + 3, (int)&CTheScripts__StreamedScriptsFooter->countOfScripts - (int)CTheScripts__StreamedScriptsArray);	// cmp     [edi+CStreamedScripts.countOfScripts], 0
					CPatch::PatchUINT32(0x156F47B + 2, (int)&CTheScripts__StreamedScriptsFooter->largestExternalSize - (int)CTheScripts__StreamedScriptsArray);	// mov     [edi+CStreamedScripts.largestExternalSize], edx
					CPatch::PatchUINT32(0x156F4A6 + 3, (int)&CTheScripts__StreamedScriptsFooter->countOfScripts - (int)CTheScripts__StreamedScriptsArray);	// cmp     [edi+CStreamedScripts.countOfScripts], bx
					CPatch::PatchUINT32(0x156F4C4 + 3, (int)&CTheScripts__StreamedScriptsFooter->countOfScripts - (int)CTheScripts__StreamedScriptsArray);	// movzx   ecx, [edi+CStreamedScripts.countOfScripts]
					CPatch::PatchUINT32(0x156F4F0 + 3, (int)&CTheScripts__StreamedScriptsFooter->countOfScripts - (int)CTheScripts__StreamedScriptsArray);	//  cmp     cx, [edi+CStreamedScripts.countOfScripts]				
				}

				// CStreamedScripts__RegisterScript_HOODLUM
				{
					CPatch::PatchUINT32(0x156B7A0 + 3, (int)&CTheScripts__StreamedScriptsFooter->countOfScripts - (int)CTheScripts__StreamedScriptsArray);	// movzx   eax, [ecx+CStreamedScripts.countOfScripts]
					CPatch::PatchUINT32(0x156B7BD + 3, (int)&CTheScripts__StreamedScriptsFooter->countOfScripts - (int)CTheScripts__StreamedScriptsArray);	// inc     [ecx+CStreamedScripts.countOfScripts]
					CPatch::PatchUINT32(0x156B7C4 + 3, (int)&CTheScripts__StreamedScriptsFooter->countOfScripts - (int)CTheScripts__StreamedScriptsArray);	// movzx   eax, [ecx+CStreamedScripts.countOfScripts]
				}
			}
		}
		#endif

		// Loaded & Requested
		#if TRUE
		{
			/*
			// useless if there's a function for this
			CPatch::PatchPointer(0x5B8B12 + 6, &CStreaming__ms_aInfoForModel.gta_sa[GetBaseID(FILE_TYPE_LOADED_START)].usNext);
			CPatch::PatchPointer(0x5B8B1C + 6, &CStreaming__ms_aInfoForModel.gta_sa[GetBaseID(FILE_TYPE_REQUESTED_START) - 1].usNext);
			CPatch::PatchPointer(0x5B8B26 + 6, &CStreaming__ms_aInfoForModel.gta_sa[GetBaseID(FILE_TYPE_REQUESTED_START)].usNext);
			CPatch::PatchPointer(0x5B8B30 + 6, &CStreaming__ms_aInfoForModel.gta_sa[GetCountOfAllFileIDs() - 1].usNext);
			CPatch::PatchPointer(0x5B8B3A + 3, &CStreaming__ms_aInfoForModel.gta_sa[GetBaseID(FILE_TYPE_LOADED_START)].usNext);
			CPatch::PatchPointer(0x5B8B43 + 3, &CStreaming__ms_aInfoForModel.gta_sa[GetBaseID(FILE_TYPE_LOADED_START)].usPrev);
			CPatch::PatchPointer(0x5B8B4A + 3, &CStreaming__ms_aInfoForModel.gta_sa[GetBaseID(FILE_TYPE_REQUESTED_START) - 1].usNext);
			CPatch::PatchPointer(0x5B8B51 + 3, &CStreaming__ms_aInfoForModel.gta_sa[GetBaseID(FILE_TYPE_REQUESTED_START) - 1].usPrev);
			CPatch::PatchPointer(0x5B8B5A + 3, &CStreaming__ms_aInfoForModel.gta_sa[GetBaseID(FILE_TYPE_REQUESTED_START)].usNext);
			CPatch::PatchPointer(0x5B8B63 + 3, &CStreaming__ms_aInfoForModel.gta_sa[GetBaseID(FILE_TYPE_REQUESTED_START)].usPrev);
			CPatch::PatchPointer(0x5B8B6A + 3, &CStreaming__ms_aInfoForModel.gta_sa[GetCountOfAllFileIDs() - 1].usNext);
			CPatch::PatchPointer(0x5B8B71 + 3, &CStreaming__ms_aInfoForModel.gta_sa[GetCountOfAllFileIDs() - 1].usPrev);

			CPatch::PatchUINT16(0x5B8B3A + 7, GetBaseID(FILE_TYPE_REQUESTED_START) - 1);
			CPatch::PatchUINT16(0x5B8B51 + 7, GetBaseID(FILE_TYPE_LOADED_START));
			CPatch::PatchUINT16(0x5B8B5A + 7, GetCountOfAllFileIDs() - 1);
			CPatch::PatchUINT16(0x5B8B71 + 7, GetBaseID(FILE_TYPE_REQUESTED_START));
			*/

			// proper patch
			CPatch::RedirectFunction(0x5B8B12, &patch_CStreaming__Init2_5B8B12);
		}
		#endif

		// Model flags
		#if TRUE
		{
			/*
			// readModelFlagsFromSave
			CPatch::PatchPointer(0x5D29E2 + 1, &CStreaming__ms_aInfoForModel->ucFlags);
			CPatch::PatchPointer(0x5D2A0A + 2, &CStreaming__ms_aInfoForModel[GetCountOfAllFileIDs()].ucFlags);


			// saveModelFlagsInSave
			CPatch::PatchPointer(0x5D29A2 + 1, &CStreaming__ms_aInfoForModel->ucFlags);
			CPatch::PatchPointer(0x5D29D3 + 2, &CStreaming__ms_aInfoForModel[GetCountOfAllFileIDs()].ucFlags);
			*/
			// 0x5D29A0 - saveModelFlagsInSave
			ApplyPatchForCStreaming__Save(this, 0x5D29A0);
		}
		#endif
	}
	#endif

	// Define variables for automatic patch
	#if TRUE
	{
		CCodeMover::SetVariable("IS_ID_UNSIGNED", this->bIsIDunsigned);
		CCodeMover::SetVariable("IS_BASE_ID_INT32", this->bIsIDbaseInt32);
		CCodeMover::SetVariable("ARE_DIFFICULT_FILE_TYPES_INT32", this->bAreDifficultFileTypesInt32);

		CCodeMover::SetVariable("MODEL_SPECIAL_FEATURES_ENABLED", g_modelSpecialFeatures.AreSpecialFeaturesEnabled());
		
		tVariableValue variable;
		variable.Array.pItems = FileInfoExtension::ptrArray;
		variable.Array.count = _countof(FileInfoExtension::ptrArray);
		CCodeMover::SetVariable("MINFO_PTR_ARRAY", variable);

		variable.Array.pItems = IDaccessPatches::ptrArray;
		variable.Array.count = _countof(IDaccessPatches::ptrArray);
		CCodeMover::SetVariable("ID_ACCESS_PTR_ARRAY", variable);
	}
	#endif

	// Make ID count uint16_t.
	#if TRUE
	if (this->bIsIDunsigned || g_modelSpecialFeatures.AreSpecialFeaturesEnabled())
	{
		// Manually added
		#if TRUE
		if(this->bIsIDunsigned)
		{
			// CColStore
			#if TRUE
			{
				// CColStore::AddColSlot
				CPatch::PatchUINT16(0x15632CF + 4, (uint16_t)MAX_FILE_ID_UINT16);	// mov     word ptr [esi+22h], 32767	; CCollisionFile.lowestModelID
				CPatch::PatchUINT16(0x15632D5 + 4, 0);		// mov     word ptr [esi+24h], -32768	; CCollisionFile.highestModelID

				// CColStore::LoadCol
				CPatch::PatchUINT8(0x15613CA, 0x76);	// make unsigned JBE

				// CColStore::RemoveCol
				CPatch::PatchUINT8(0x1564EC0, 0x77);	// make unsigned JA
				CPatch::PatchUINT8(0x1564EF2, 0x76);	// make unsigned JBE

				// CColStore::IncludeModelIndex
				CPatch::PatchUINT8(0x156A288, 0x73);
				CPatch::PatchUINT8(0x156A294, 0x76);	// make unsigned JBE
			}
			#endif

			// CIplStore
			#if TRUE
			{
				// CIplStore::AddIplSlot
				CPatch::PatchUINT32(0x156C488 + 1, 0);	// mov     ecx, -32768
				CPatch::PatchUINT32(0x156C48F + 1, (uint16_t)MAX_FILE_ID_UINT16);	// mov     edx, 32767

				// CIplStore::RemoveIpl
				CPatch::PatchUINT8(0x404B6F, 0x77);	// make unsigned JA
				CPatch::PatchUINT8(0x404BB2, 0x76);	// make unsigned JBE

				// CIplStore::RemoveIpl
				CPatch::PatchUINT8(0x404C19, 0x77);	// make unsigned JA
				CPatch::PatchUINT8(0X404C58, 0x76);	// make unsigned JBE
			}
			#endif

			// CLoadedCarGroup::PickRandomCar
			CPatch::PatchUINT16(0x611D62, 0xB70F);	// movsx   esi, word ptr [edx]
			CPatch::PatchUINT16(0x611D8F, 0xB70F);	// movsx   esi, word ptr [edx]

			// CPopulation::PickRiotRoadBlockCar
			CPatch::PatchUINT16(0x614512, 0xB70F);	// movsx   eax, word ptr _ZN11CPopulation23m_AppropriateLoadedCarsE.m_Array[edx]
			CPatch::PatchUINT16(0x614542, 0xB70F);	// movsx   eax, word ptr _ZN11CPopulation25m_InAppropriateLoadedCarsE.m_Array[edx]

			// vehicle audio loader
			if(!this->bAreDifficultFileTypesInt32)
				CPatch::PatchUINT16(0x4F77B8, 0xB70F);	// movsx   eax, ax
			
			// CTheScripts::ProcessWaitingForScriptBrainArray
			CPatch::PatchUINT16(0x46CFAB, 0xB70F);	// movsx   eax, _ZN11CTheScripts16ScriptsForBrainsE.SCM_script_index[eax*4]

			// CObject::~CObject
			CPatch::PatchUINT16(0x59F6A4, 0xB70F);	// movsx   eax, _ZN11CTheScripts16ScriptsForBrainsE.SCM_script_index[eax*4]

			// CPed::~CPed
			CPatch::PatchUINT16(0x5E866D, 0xB70F);	// movsx   eax, _ZN11CTheScripts16ScriptsForBrainsE.SCM_script_index[eax*4]

			// CCarGenerator::DoInternalProcessing
			CPatch::PatchUINT16(0x6F3608, 0xB70F);	// movsx   ecx, word ptr [edi]
			CPatch::PatchUINT16(0x6F3616, 0xB70F);	// movsx   ebp, ax

			// unk
			CPatch::PatchUINT16(0x611C24, 0xB70F);

			// CLoadedCarGroup::RemoveMember
			CPatch::PatchUINT16(0x611BE0, 0xB70F);	// movsx   esi, word ptr [edx+eax*2]

			// CLoadedCarGroup::PickRandomCar
			CPatch::PatchUINT16(0x611CC4, 0xB70F);	// movsx   eax, cx
			CPatch::PatchUINT16(0x611D00, 0xB70F);	// movsx   ecx, word ptr [edx+eax*2]
			CPatch::PatchUINT16(0x611D5B, 0xB70F);	// movsx   esi, [esp+4Ch+var_30.m_Array]
			CPatch::PatchUINT16(0x611D89, 0xB70F);	// movsx   esi, word ptr [edi+edx]

			// CLoadedCarGroup::SortBasedOnUsage
			CPatch::PatchUINT16(0x611E45, 0xB70F);	// movsx   edi, word ptr [ecx+eax*2]
			CPatch::PatchUINT16(0x611E50, 0xB70F);	// movsx   ebx, si

			// CReplay::ProcessPedUpdate
			CPatch::PatchUINT16(0x45CC88, 0xB70F);	// movsx   eax, di

			// CReplay::PlayBackThisFrameInterpolation
			CPatch::PatchUINT16(0x45F3FF, 0xB70F);	// movsx   ecx, word ptr [esi+1Ah]	; sub_45F380
			CPatch::PatchUINT16(0x45F451, 0xB70F);	// movsx   ecx, word ptr [esi+1Ah]	; sub_45F380
			CPatch::PatchUINT16(0x45F48D, 0xB70F);	// movsx   edx, word ptr [esi+1Ah]	; sub_45F380
			CPatch::PatchUINT16(0x45F4C9, 0xB70F);	// movsx   ecx, word ptr [esi+1Ah]	; sub_45F380
			CPatch::PatchUINT16(0x45F503, 0xB70F);	// movsx   edx, word ptr [esi+1Ah]	; sub_45F380
			CPatch::PatchUINT16(0x45F538, 0xB70F);	// movsx   ecx, word ptr [esi+1Ah]	; sub_45F380
			CPatch::PatchUINT16(0x45F56D, 0xB70F);	// movsx   edx, word ptr [esi+1Ah]	; sub_45F380
			CPatch::PatchUINT16(0x45F66D, 0xB70F);	// movsx   ecx, word ptr [esi+1Ah]	; sub_45F380
			CPatch::PatchUINT16(0x45F6A7, 0xB70F);	// movsx   ecx, word ptr [esi+1Ah]	; sub_45F380
			CPatch::PatchUINT16(0x45F7AB, 0xB70F);	// movsx   ecx, word ptr [esi+1Ah]	; sub_45F380
			CPatch::PatchUINT16(0x45F7E5, 0xB70F);	// movsx   ecx, word ptr [esi+1Ah]	; sub_45F380
			CPatch::PatchUINT16(0x45F8DE, 0xB70F);	// movsx   ecx, word ptr [esi+1Ah]	; sub_45F380
			CPatch::PatchUINT16(0x45F918, 0xB70F);	// movsx   ecx, word ptr [esi+1Ah]	; sub_45F380
			CPatch::PatchUINT16(0x45FA11, 0xB70F);	// movsx   ecx, word ptr [esi+1Ah]	; sub_45F380
			CPatch::PatchUINT16(0x45FA4B, 0xB70F);	// movsx   ecx, word ptr [esi+1Ah]	; sub_45F380
			CPatch::PatchUINT16(0x45FB58, 0xB70F);	// movsx   ecx, word ptr [esi+1Ah]	; sub_45F380
			CPatch::PatchUINT16(0x45FB92, 0xB70F);	// movsx   ecx, word ptr [esi+1Ah]	; sub_45F380

			// CVehicle::SetupRender
			CPatch::RedirectCode(0x6D6521, &patch_CVehicle__SetupRender_6D6521);

			// CCam::GetBoatLook_L_R_HeightOffset
			CPatch::PatchUINT16(0x509CAA, 0xB70F);	// movsx   eax, word ptr[eax + 22h]
		}
		#endif

		// Automatically generated
		#include "FileIDlimit\GTA SA 1.0 US HOODLUM WIN_X86\GTASA_ID_access_patches.h"
	}
	#endif
		
	// Make base ID to be int32_t value.
	#if TRUE
	if (this->bIsIDbaseInt32)
	{
		// Apply automatic patch
		#include "FileIDlimit\GTA SA 1.0 US HOODLUM WIN_X86\GTASA_int32_base_movsx_patches.h"
	}
	#endif

	// Make COL ID to be uint32_t
	#if TRUE
	if(this->bIsCOLIDsizeIncreased)
	{
		CPatch::RedirectCode(0x5B4FBD, &patch_CFileLoader__LoadCollisionFile_5B4FBD);
		CPatch::RedirectCode(0x59F8BE, &patch_CObject__Init_59F8BE);

		CPatch::RedirectCode(0x5383D5, &patch_CFileLoader__LoadObjectInstance_5383D5);
		CPatch::RedirectCode(0x5383E8, &patch_CFileLoader__LoadObjectInstance_5383E8);

		// CFileLoader::LoadCollisionFile
		{
			static const unsigned char data[] = { 0x8B, 0x44, 0x24, 0x50, 0x90 };	// mov eax,[esp+50h]; nop;
			CPatch::PatchMemoryData(0x53851E, data, sizeof(data));

			CPatch::RedirectCode(0x538620, &patch_CFileLoader__LoadCollisionFile_538620);
		}

		// CFileLoader::LoadCollisionFileFirstTime
		{
			static const unsigned char data[] = { 0x8B, 0x4C, 0x24, 0x54, 0x90 };	// mov ecx,[esp+54h]; nop;
			CPatch::PatchMemoryData(0x5B50F4, data, sizeof(data));

			CPatch::RedirectCode(0x5B5189, &patch_CFileLoader__LoadCollisionFileFirstTime_5B5189);
		}

		CPatch::RedirectCode(0x4C6F48, &patch_CPedModelInfo__CreateHitColModelSkinned_4C6F48);
		CPatch::RedirectCode(0x1564EDE, &patch_CColStore__RemoveCol_1564EDE);
		
	}
	#endif

	// Make IPL ID to be uint32_t
	#if TRUE
	if(this->bIsIPLIDsizeIncreased)
	{
		// Set
		#if TRUE
		{
			// CEntity::CEntity
			CPatch::RedirectCode(0x532AB9, &patch_CEntity__CEntity_532AB9);

			// CIplStore::LoadIpl
			CPatch::RedirectCode(0x4061ED, &patch_CIplStore__LoadIpl_4061ED);
			CPatch::RedirectCode(0x4062F9, &patch_CIplStore__LoadIpl_4062F9);

			// CIplStore::LoadIplBoundingBox
			CPatch::RedirectCode(0x405E0F, &patch_CIplStore__LoadIplBoundingBox_405E0F);
			CPatch::RedirectCode(0x405C9E, &patch_CIplStore__LoadIplBoundingBox_405C9E);

			// CObject::CObject(CDummyObject *)
			CPatch::RedirectCode(0x5A1E77, &patch_CObject__CObject_5A1E77);

			// CDummyObject::CDummyObject
			CPatch::RedirectCode(0x59EA79, &patch_CDummyObject__CDummyObject_59EA79);
		}
		#endif

		// Get
		#if TRUE
		{
			// CIplStore::RemoveIpl
			CPatch::RedirectCode(0x404B85, &patch_CIplStore__RemoveIpl_404B85);	// buildings
			CPatch::RedirectCode(0x404BCF, &patch_CIplStore__RemoveIpl_404BCF);	// objects
			CPatch::RedirectCode(0x404C33, &patch_CIplStore__RemoveIpl_404C33);	// dummies

			// CEntity::RegisterReference
			CPatch::RedirectCode(0x571B85, &patch_CEntity__RegisterReference_571B85);

			// CDummyObject::UpdateFromObject
			CPatch::RedirectCode(0x59EBCA, &patch_CDummyObject__UpdateFromObject_59EBCA);
		}
		#endif
	}
	#endif

	// Manual patches for difficult file types
	#if TRUE
	if (this->bAreDifficultFileTypesInt32)
	{
		CPatch::RedirectCode(0x4F5C10, &patch_CAEVehicleAudioEntity__GetVehicleAudioSettings_4F5C10);
		CPatch::RedirectCode(0x47A853, &patch_CRunningScript__ProcessCommands2500To2599_47A853);
		CPatch::RedirectFunction(0x4C4B70, &CBaseModelInfo__ClearTexDictionary);
		CPatch::RedirectCode(0x5E013A, &patch_CPed__Dress_5E013A);

		CPatch::RedirectCode(0x4C4B51, &patch_CBaseModelInfo__SetTexDictionary_4C4B51);
		CPatch::RedirectCode(0x4C4B61, &patch_CBaseModelInfo__SetTexDictionary_4C4B61);
		CPatch::RedirectFunction(0x404C90, &CIplStore__IncludeEntity);
		CPatch::RedirectFunction(0x410820, &CColStore__IncludeModelIndex);

		CPatch::RedirectCode(0x4C74D7, &patch_CVehicleModelInfo__CLinkedUpgradeList__FindOtherUpgrade_4C74D7);
		CPatch::RedirectCode(0x4C74F2, &patch_CVehicleModelInfo__CLinkedUpgradeList__FindOtherUpgrade_4C74F2);

		CPatch::RedirectCode(0x403EE0, &patch_CBuilding__ReplaceWithNewModel_403EE0);
		CPatch::RedirectCode(0x403EF0, &patch_CBuilding__ReplaceWithNewModel_403EF0);
		CPatch::NOPinstructions(0x53BF4E, 5);	// Disable CCover::Update, which is buggy in GTA SA

		CPatch::RedirectCode(0x456535, &patch_CPickups__GivePlayerGoodiesWithPickUpMI_456535);

		// CVisibilityPlugins::SetModelInfoIndex
		{
			static const unsigned char data[] = { 0x8B, 0x44, 0x24, 0x08, 0x90 };	// mov eax,[esp+8]; nop;
			CPatch::PatchMemoryData(0x732230, data, sizeof(data));
		}

		// CFileLoader::LoadTimeObject
		{
			CPatch::RedirectCode(0x5B3F94, &patch_CFileLoader__LoadTimeObject_5B3F94);
		}

		// FurnitureSubGroup_c::AddFurniture
		{
			static const unsigned char data[] = { 0x8B, 0x44, 0x24, 0x24, 0x90 };	// mov eax,[esp+24h]; nop;
			CPatch::PatchMemoryData(0x5C00E1, data, sizeof(data));
		}

		// InitModelIndices
		CPatch::PatchUINT32(0x5BCA10 + 1, UNDEFINED_FILE_ID_INT32);
		
		// CModelInfo::GetModelInfoUInt16
		CPatch::RedirectCode(0x4C5A13, &patch_CModelInfo__GetModelInfoUInt16_4C5A13);

		// CCam::GetBoatLook_L_R_HeightOffset
		CPatch::RedirectCode(0x509CAA, &patch_CCam__GetBoatLook_L_R_HeightOffset_509CAA);
		CPatch::RedirectCode(0x6A2E98, &patch_CAutomobile__Render_6A2E98);

		// CVehicleModelInfo::AddRemap
		{
			static const unsigned char data[] = { 0x8B, 0x54, 0x24, 0x04, 0x90 };	// mov eax,[esp+04h]; nop;
			CPatch::PatchMemoryData(0x4C86EF, data, sizeof(data));
		}
	}
	#endif

	// Manual patches for special features
	#if TRUE
	if(g_modelSpecialFeatures.AreSpecialFeaturesEnabled())
	{
		// MODEL_FORKLIFT
		CPatch::RedirectCode(0x6AC730, &patch_CAutomobile__PreRender_6AC730);

		// MODEL_SKIMMER
		CPatch::RedirectCode(0x6B2185, &patch_CAutomobile__ProcessControl_6B2185);
	}
	#endif

	// Undefine variables of automatic patch
	#if TRUE
	{
		CCodeMover::DeleteVariable("IS_ID_UNSIGNED");
		CCodeMover::DeleteVariable("IS_BASE_ID_INT32");
		CCodeMover::DeleteVariable("ARE_DIFFICULT_FILE_TYPES_INT32");
		CCodeMover::DeleteVariable("MODEL_SPECIAL_FEATURES_ENABLED");
		CCodeMover::DeleteVariable("MINFO_PTR_ARRAY");
		CCodeMover::DeleteVariable("ID_ACCESS_PTR_ARRAY");
	}
	#endif

	g_modelSpecialFeatures.ApplyLoader();
}
#endif

#ifdef IS_PLATFORM_ANDROID_ARM32
extern "C"
{
	// Addresses
	uintptr_t Address_CStreaming__ms_aInfoForModel = 0;
	uintptr_t Address_CStreamingInfo__ms_pArrayBase = 0;
	uintptr_t Address_CStreaming__ms_channel = 0;
	///////////

	#define DEFINE_VARS_FOR_FILE_TYPE(member, ...) uint32_t member##_BASE_ID; \
	uint32_t member##_BASE_ID_PLUS_ONE; \
	uint32_t member##_LAST_ID; \
	uint32_t member##_COUNT; \
	uint32_t member##_COUNT_MINUS_ONE;

	uint32_t FILE_TYPE_IPL_COUNT_MINUS_TWO;

	#define SET_FILE_TYPE_INFO_IN_VARIABLES(member, ...) member##_BASE_ID = GetBaseID(member); \
			member##_BASE_ID_PLUS_ONE = member##_BASE_ID + 1; \
			member##_LAST_ID = GetLastIDforType(member); \
			member##_COUNT = GetFileIDcurrentLimit(member); \
			member##_COUNT_MINUS_ONE = member##_COUNT - 1; 

	MAKE_FILE_TYPE_LIST(DEFINE_VARS_FOR_FILE_TYPE);

	uint32_t FILE_TYPE_TXD_BASE_ID_multiplied_by_sizeof_CStreamingInfo_minus_one = 0;	// GTA SA: 399999
	uint32_t FILE_TYPE_COL_BASE_ID_multiplied_by_sizeof_CStreamingInfo_minus_one = 0;	// GTA SA: 499999
	uint32_t negated_file_ID_offset_to_array_TxdDef_parent_index = 0;	// GTA SA: 0x1116FA

	uint32_t negated_FILE_TYPE_DAT_LAST_ID_multiplied_by_sizeof_CStreamingInfo_plus_one = 0; // GTA SA: -0x7CDF9
	uint32_t FILE_TYPE_IFP_COUNT_plus_one_then_multiplied_by_sizeof_CStreamingInfo_then_minus_two = 0; // GTA SA: 0xE22
	uint32_t FILE_TYPE_RRR_LAST_ID_multiplied_by_sizeof_CStreamingInfo_then_plus_one = 0; // GTA SA: 0x80125

	uint32_t negated_file_ID_offset_to_array_to_CStreamedScriptInfo_status = 0;	// GTA SA: 0xCCEBC

	uint32_t FILE_TYPE_TXD_COUNT_multiplied_by_sizeof_TxdDef = 0;	// GTA SA: 280000
	uint32_t FILE_TYPE_TXD_BASE_ID_multiplied_by_sizeof_CStreamingInfo_plus_uiLoadFlag;	// GTA SA: 0x61000 + 0xA90 = 0x61A90
	uint32_t FILE_TYPE_IFP_BASE_ID_multiplied_by_sizeof_CStreamingInfo_plus_uiLoadFlag;	// GTA SA: 0x7CE1C
	uint32_t FILE_TYPE_IPL_BASE_ID_PLUS_ONE_then_multiplied_by_sizeof_CStreamingInfo_plus_ucFlags; // GTA SA: 0x7B526
	uint32_t FILE_TYPE_IPL_COUNT_multiplied_by_sizeof_IplDef = 0;	// 0x3400
	uint32_t FILE_TYPE_COL_COUNT_multiplied_by_sizeof_CCollisionFile;	// 0x2BD4
	uint32_t FILE_TYPE_SCM_BASE_ID_multiplied_by_sizeof_CStreamingInfo_plus_uiLoadFlag;	// GTA SA: 0x80148
	uint32_t FILE_TYPE_COUNT_OF_IDS = 0;

	// see also: 0x46BC3C
}

// GTA San Andreas
namespace Game_GTASA
{
	// Automatically generated patches
	#include "FileIDlimit/GTA SA 2.0 ANDROID_ARM32/functions.cpp"

	//////////////////////////////////

	// Manually written code
	#if TRUE
	extern "C"
	{
		uintptr_t Address_CStreaming__LoadCdDirectory_46C140_thumb = 0;
		uintptr_t Address_CStreaming__LoadCdDirectory_18D8E0_thumb = 0;
		uintptr_t Address_CStreaming__LoadCdDirectory_46C144_thumb = 0;
	}

	static TARGET_THUMB NAKED void patch_CStreaming__LoadCdDirectory_46C132()
	{
		__asm(
		".thumb\n"
			"BNE 1f\n"
			"LDR R1, [SP,#12]\n"
			"MOV R0, R10\n"
			"MOVS R2, #0\n"
			"BLX 2f\n"
			"STR R0, [SP,#80]\n"
			ASM_LOAD_4BYTE_UNSIGNED_VALUE_STORED_ON_SYMBOL(R1, FILE_TYPE_TXD_BASE_ID)
			ASM_JUMP_TO_ADDRESS_STORED_ON_SYMBOL(Address_CStreaming__LoadCdDirectory_46C144_thumb)

			"1:\n"	// loc_46C140
			ASM_JUMP_TO_ADDRESS_STORED_ON_SYMBOL(Address_CStreaming__LoadCdDirectory_46C140_thumb)

			"2:\n"	// j__ZN9CTxdStore10AddTxdSlotEPKcS1_b
			ASM_JUMP_TO_ADDRESS_STORED_ON_SYMBOL(Address_CStreaming__LoadCdDirectory_18D8E0_thumb));
	}

	// CPathFind::Shutdown
	// patch for 0x3191F0
	extern "C"
	{
		uintptr_t Address_CPathFind__Shutdown_319204_thumb_ = 0;
	}

	static TARGET_THUMB NAKED void patch_CPathFind__Shutdown_3191F0()
	{
		__asm(
			".thumb\n"
			ASM_LOAD_4BYTE_UNSIGNED_VALUE_STORED_ON_SYMBOL(R10, FILE_TYPE_DAT_BASE_ID)
			"ADD R9, R10, #8\n"
			"ADD R8, R10, #8*2\n"
			"ADD R6, R10, #8*3\n"
			"ADD R11, R10, #8*4\n"
			ASM_JUMP_TO_ADDRESS_STORED_ON_SYMBOL(Address_CPathFind__Shutdown_319204_thumb_)
			);
	}

	// patch for 0x46BAB6
	extern "C"
	{
		uintptr_t Address_CStreaming__Init2_46BB82_thumb = 0;
		uintptr_t Address_CStreaming__m_bHarvesterModelsRequested = 0;
	}

	// use patch_CStreaming__Init2_initializeLoadedAndRequested
	static TARGET_THUMB NAKED void patch_CStreaming__Init2_46BAB6()
	{
		__asm(
			".thumb\n"
			"BLX patch_CStreaming__Init2_initializeLoadedAndRequested\n"

			ASM_LOAD_ADDRESS_STORED_ON_SYMBOL(R6, Address_CStreaming__ms_aInfoForModel)
			ASM_LOAD_ADDRESS_STORED_ON_SYMBOL(R5, Address_CStreamingInfo__ms_pArrayBase)
			"STR             R6, [R5]\n"

			ASM_LOAD_ADDRESS_STORED_ON_SYMBOL(R8, Address_CStreaming__ms_channel)
			ASM_LOAD_ADDRESS_STORED_ON_SYMBOL(R0, Address_CStreaming__m_bHarvesterModelsRequested)

			ASM_JUMP_TO_ADDRESS_STORED_ON_SYMBOL(Address_CStreaming__Init2_46BB82_thumb)
			);
	}
	#endif

	// patch for 0x46BC54
	extern "C"
	{
		uintptr_t Address_CStreaming__Init2_46BC5E_thumb = 0;
	}
	
	static TARGET_THUMB NAKED void patch_CStreaming__Init2_46BC54()
	{
		__asm(
		".thumb\n"
			ASM_ADD_4BYTE_VALUE_STORED_ON_SYMBOL(R0, FILE_TYPE_TXD_BASE_ID_multiplied_by_sizeof_CStreamingInfo_plus_uiLoadFlag)

			"LDR R1, [R2]\n"

			ASM_JUMP_TO_ADDRESS_STORED_ON_SYMBOL(Address_CStreaming__Init2_46BC5E_thumb)
			);
	}

	///// Shopping //////
	// patch for 0x35F9B2
	extern "C"
	{
		uintptr_t Address_CShopping__ShutdownForRestart_19BC1C_arm = 0;
		uintptr_t Address_CShopping__ShutdownForRestart_35F9BE_thumb = 0;
	}

	static TARGET_THUMB NAKED void patch_CShopping__ShutdownForRestart_35F9B2()
	{
		__asm(
		".thumb\n"

			"BEQ 10f\n"
			ASM_ADD_4BYTE_VALUE_STORED_ON_SYMBOL(R0, FILE_TYPE_IFP_BASE_ID)
			"BL 1f\n"

			"10:\n"
			ASM_JUMP_TO_ADDRESS_STORED_ON_SYMBOL(Address_CShopping__ShutdownForRestart_35F9BE_thumb)

			// Branches
			"1:\n"	// j__ZN10CStreaming19SetModelIsDeletableEi
			ASM_JUMP_TO_ADDRESS_STORED_ON_SYMBOL(Address_CShopping__ShutdownForRestart_19BC1C_arm)
			);
	}

	// patch for 0x35FA1E
	extern "C"
	{
		uintptr_t Address_CShopping__RemoveLoadedShop_19BC1C_arm = 0;
		uintptr_t Address_CShopping__RemoveLoadedShop_35FA2A_thumb = 0;
	}

	static TARGET_THUMB NAKED void patch_CShopping__RemoveLoadedShop_35FA1E()
	{
		__asm(
		".thumb\n"
			"BEQ 10f\n"
			ASM_ADD_4BYTE_VALUE_STORED_ON_SYMBOL(R0, FILE_TYPE_IFP_BASE_ID)
			"BL 1f\n"

			"10:\n"
			ASM_JUMP_TO_ADDRESS_STORED_ON_SYMBOL(Address_CShopping__RemoveLoadedShop_35FA2A_thumb)

			// Branches
			"1:\n"	// j__ZN10CStreaming19SetModelIsDeletableEi
			ASM_JUMP_TO_ADDRESS_STORED_ON_SYMBOL(Address_CShopping__RemoveLoadedShop_19BC1C_arm)
			);
	}

	// patch for 0x35FCAE
	extern "C"
	{
		uintptr_t Address_CShopping__LoadPrices_19BC1C_arm = 0;
		uintptr_t Address_CShopping__LoadPrices_35FCBA_thumb = 0;
	}

	static TARGET_THUMB NAKED void patch_CShopping__LoadPrices_35FCAE()
	{
		__asm(
		".thumb\n"
			"BEQ 10f\n"
			ASM_ADD_4BYTE_VALUE_STORED_ON_SYMBOL(R0, FILE_TYPE_IFP_BASE_ID)
			"BL 1f\n"

			"10:\n"
			ASM_JUMP_TO_ADDRESS_STORED_ON_SYMBOL(Address_CShopping__LoadPrices_35FCBA_thumb)

			// Branches
			"1:\n"	// j__ZN10CStreaming19SetModelIsDeletableEi
			ASM_JUMP_TO_ADDRESS_STORED_ON_SYMBOL(Address_CShopping__LoadPrices_19BC1C_arm)
			);
	}

	// patch for 0x360040
	extern "C"
	{
		uintptr_t Address_CShopping__RemoveLoadedPrices_19BC1C_arm = 0;
		uintptr_t Address_CShopping__RemoveLoadedPrices_36004C_thumb = 0;
	}

	static TARGET_THUMB NAKED void patch_CShopping__RemoveLoadedPrices_360040()
	{
		__asm(
		".thumb\n"
			"BEQ 10f\n"
			ASM_ADD_4BYTE_VALUE_STORED_ON_SYMBOL(R0, FILE_TYPE_IFP_BASE_ID)
			"BL 1f\n"

			"10:\n"
			ASM_JUMP_TO_ADDRESS_STORED_ON_SYMBOL(Address_CShopping__RemoveLoadedPrices_36004C_thumb)

			// Branches
			"1:\n"	// j__ZN10CStreaming19SetModelIsDeletableEi
			ASM_JUMP_TO_ADDRESS_STORED_ON_SYMBOL(Address_CShopping__RemoveLoadedPrices_19BC1C_arm)
			);
	}
}
#endif

#ifdef IS_PLATFORM_ANDROID_ARM32
// Patches ID limits if neccessary.
void FileIDlimit::PatchFileIDlimits_GTA_SA_2_0_ANDROID_ARM32()
{
	using namespace Game_GTASA;

	// Set variables
	#if TRUE
	{
		MAKE_FILE_TYPE_LIST(SET_FILE_TYPE_INFO_IN_VARIABLES);
		FILE_TYPE_IPL_COUNT_MINUS_TWO = FILE_TYPE_IPL_COUNT_MINUS_ONE - 1;

		Address_CStreaming__ms_aInfoForModel = (uintptr_t)this->CStreaming__ms_aInfoForModel.generic;
		Address_CStreaming__ms_channel = (uintptr_t)Library::GetSymbolAddress(&g_LimitAdjuster.hModule_of_game,
			"_ZN10CStreaming10ms_channelE");

		Address_CStreamingInfo__ms_pArrayBase = (uintptr_t)Library::GetSymbolAddress(&g_LimitAdjuster.hModule_of_game,
			"_ZN14CStreamingInfo13ms_pArrayBaseE");

		FILE_TYPE_TXD_BASE_ID_multiplied_by_sizeof_CStreamingInfo_minus_one = GetBaseID(FILE_TYPE_TXD) * TypeDetails::SizeOf<CStreamingInfo>() - 1;
		FILE_TYPE_COL_BASE_ID_multiplied_by_sizeof_CStreamingInfo_minus_one = GetBaseID(FILE_TYPE_COL) * TypeDetails::SizeOf<CStreamingInfo>() - 1;
		negated_file_ID_offset_to_array_TxdDef_parent_index = -(TypeDetails::SizeOf<TxdDef_Android>() * FILE_TYPE_TXD_BASE_ID - offsetof(TxdDef_Android, parent_index));
		negated_FILE_TYPE_DAT_LAST_ID_multiplied_by_sizeof_CStreamingInfo_plus_one = -(GetLastIDforType(FILE_TYPE_DAT) * TypeDetails::SizeOf<CStreamingInfo>() + 1);
		FILE_TYPE_IFP_COUNT_plus_one_then_multiplied_by_sizeof_CStreamingInfo_then_minus_two = (GetFileIDcurrentLimit(FILE_TYPE_IFP) + 1) * TypeDetails::SizeOf<CStreamingInfo>() - 2;
		FILE_TYPE_RRR_LAST_ID_multiplied_by_sizeof_CStreamingInfo_then_plus_one = GetLastIDforType(FILE_TYPE_RRR) * TypeDetails::SizeOf<CStreamingInfo>() + 1;
		negated_file_ID_offset_to_array_to_CStreamedScriptInfo_status = -(TypeDetails::SizeOf<CStreamedScriptInfo>() * FILE_TYPE_SCM_BASE_ID - offsetof(CStreamedScriptInfo, status));
		FILE_TYPE_TXD_COUNT_multiplied_by_sizeof_TxdDef = FILE_TYPE_TXD_COUNT * TypeDetails::SizeOf<TxdDef_Android>();
		FILE_TYPE_TXD_BASE_ID_multiplied_by_sizeof_CStreamingInfo_plus_uiLoadFlag = GetBaseID(FILE_TYPE_TXD) * TypeDetails::SizeOf<CStreamingInfo>() + offsetof(CStreamingInfo, uiLoadFlag);
		FILE_TYPE_IFP_BASE_ID_multiplied_by_sizeof_CStreamingInfo_plus_uiLoadFlag = GetBaseID(FILE_TYPE_IFP) * TypeDetails::SizeOf<CStreamingInfo>() + offsetof(CStreamingInfo, uiLoadFlag);
		FILE_TYPE_IPL_BASE_ID_PLUS_ONE_then_multiplied_by_sizeof_CStreamingInfo_plus_ucFlags = FILE_TYPE_IPL_BASE_ID_PLUS_ONE * TypeDetails::SizeOf<CStreamingInfo>() + offsetof(CStreamingInfo, ucFlags);
		FILE_TYPE_IPL_COUNT_multiplied_by_sizeof_IplDef = FILE_TYPE_IPL_COUNT * TypeDetails::SizeOf <CIplFile>();
		FILE_TYPE_COL_COUNT_multiplied_by_sizeof_CCollisionFile = FILE_TYPE_COL_COUNT * TypeDetails::SizeOf <CCollisionFile>();
		FILE_TYPE_SCM_BASE_ID_multiplied_by_sizeof_CStreamingInfo_plus_uiLoadFlag = GetBaseID(FILE_TYPE_SCM) * TypeDetails::SizeOf<CStreamingInfo>() + offsetof(CStreamingInfo, uiLoadFlag);
		FILE_TYPE_COUNT_OF_IDS = GetCountOfAllFileIDs();
	}
	#endif
	
	// Patch pointers to CStreaming::ms_aInfoForModel
	#if TRUE
	{
		CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x677DD8), CStreaming__ms_aInfoForModel.gta_sa);
	}
	#endif

	// Patch limits itself
	#if TRUE
	{
		// Model DFF limit, patch pointers to CModelInfo::ms_modelInfoPtrs
		#if TRUE
		{
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x6796D4), CModelInfo__ms_modelInfoPtrs.gta_sa);
		}
		#endif

		// IFP
		#if TRUE
		{
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x6793B4), CAnimManager__ms_aAnimBlocks.gta_sa);
		}
		#endif

		// RRR
		#if TRUE
		{
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x679B70), CVehicleRecording__StreamingArray.gta_sa);
		}
		#endif

		// Model flags
		#if TRUE
		{
			uintptr_t pFunc = (uintptr_t)Library::GetSymbolAddress(&g_LimitAdjuster.hModule_of_game,
				"_ZN10CStreaming4SaveEv");

			ApplyPatchForCStreaming__Save(this, pFunc);
		}
		#endif
	}
	#endif

	// Automatic patches
	#if TRUE
	{
		#include "FileIDlimit/GTA SA 2.0 ANDROID_ARM32/changes.cpp"
	}
	#endif

	// Make COL ID to be uint32_t
	#if TRUE
	if (this->bIsCOLIDsizeIncreased)
	{
		Address_CFileLoader__LoadCollisionFile_46701A = g_mCalc.GetCurrentVAbyPreferedVA(0x46701A);
		Address_CFileLoader__LoadCollisionFile_467018_thumb = g_mCalc.GetCurrentVAbyPreferedVA(ASM_GET_THUMB_ADDRESS_FOR_JUMP(0x467018));
		CPatch::RedirectCodeEx(INSTRUCTION_SET_THUMB, g_mCalc.GetCurrentVAbyPreferedVA(0x46700E),
			(void*)&patch_CFileLoader__LoadCollisionFile_46700E
		);

		Address_CObject__Init_452CE8_thumb = g_mCalc.GetCurrentVAbyPreferedVA(ASM_GET_THUMB_ADDRESS_FOR_JUMP(0x452CE8));
		Address_CObject__Init_452CCA_thumb = g_mCalc.GetCurrentVAbyPreferedVA(ASM_GET_THUMB_ADDRESS_FOR_JUMP(0x452CCA));
		CPatch::RedirectCodeEx(INSTRUCTION_SET_THUMB, g_mCalc.GetCurrentVAbyPreferedVA(0x452CBE),
			(void*)&patch_CObject__Init_452CBE
		);

		Address_CFileLoader__LoadObjectInstance_3F08D6_thumb = g_mCalc.GetCurrentVAbyPreferedVA(ASM_GET_THUMB_ADDRESS_FOR_JUMP(0x3F08D6));
		Address_CFileLoader__LoadObjectInstance_3F087A_thumb = g_mCalc.GetCurrentVAbyPreferedVA(ASM_GET_THUMB_ADDRESS_FOR_JUMP(0x3F087A));
		CPatch::RedirectCodeEx(INSTRUCTION_SET_THUMB, g_mCalc.GetCurrentVAbyPreferedVA(0x3F0866),
			(void*)&patch_CFileLoader__LoadObjectInstance_3F0866
		);
		

		Address_CFileLoader__LoadCollisionFile_3EF580_thumb = g_mCalc.GetCurrentVAbyPreferedVA(ASM_GET_THUMB_ADDRESS_FOR_JUMP(0x3EF580));
		CPatch::RedirectCodeEx(INSTRUCTION_SET_THUMB, g_mCalc.GetCurrentVAbyPreferedVA(0x3EF576),
			(void*)&patch_CFileLoader__LoadCollisionFile_3EF576
		);
		

		Address_CFileLoader__LoadCollisionFileFirstTime_46BA0C_thumb = g_mCalc.GetCurrentVAbyPreferedVA(ASM_GET_THUMB_ADDRESS_FOR_JUMP(0x46BA0C));
		CPatch::RedirectCodeEx(INSTRUCTION_SET_THUMB, g_mCalc.GetCurrentVAbyPreferedVA(0x46BA02),
			(void*)&patch_CFileLoader__LoadCollisionFileFirstTime_46BA02
		);		

		Address_CPedModelInfo__CreateHitColModelSkinned_386818_thumb = g_mCalc.GetCurrentVAbyPreferedVA(ASM_GET_THUMB_ADDRESS_FOR_JUMP(0x386818));
		CPatch::RedirectCodeEx(INSTRUCTION_SET_THUMB, g_mCalc.GetCurrentVAbyPreferedVA(0x386810),
			(void*)&patch_CPedModelInfo__CreateHitColModelSkinned_386810
		);		

		Address_CColStore__RemoveCol_2E2596_thumb = g_mCalc.GetCurrentVAbyPreferedVA(ASM_GET_THUMB_ADDRESS_FOR_JUMP(0x2E2596));
		Address_CColStore__RemoveCol_2E2590_thumb = g_mCalc.GetCurrentVAbyPreferedVA(ASM_GET_THUMB_ADDRESS_FOR_JUMP(0x2E2590));
		CPatch::RedirectCodeEx(INSTRUCTION_SET_THUMB, g_mCalc.GetCurrentVAbyPreferedVA(0x2E2588),
			(void*)&patch_CColStore__RemoveCol_2E2588
		);		
	}
	#endif

	// Manual patches
	#if TRUE
	{
		// MOVW            R1, #20000
		Address_CStreaming__LoadCdDirectory_46C140_thumb = g_mCalc.GetCurrentVAbyPreferedVA(ASM_GET_THUMB_ADDRESS_FOR_JUMP(0x46C140));
		Address_CStreaming__LoadCdDirectory_18D8E0_thumb = g_mCalc.GetCurrentVAbyPreferedVA(ASM_GET_THUMB_ADDRESS_FOR_JUMP(0x18D8E0));
		Address_CStreaming__LoadCdDirectory_46C144_thumb = g_mCalc.GetCurrentVAbyPreferedVA(ASM_GET_THUMB_ADDRESS_FOR_JUMP(0x46C144));
		CPatch::RedirectCodeEx(INSTRUCTION_SET_THUMB, g_mCalc.GetCurrentVAbyPreferedVA(0x46C132),
			(void*)&patch_CStreaming__LoadCdDirectory_46C132
		);

		// CPathFind::Shutdown
		Address_CPathFind__Shutdown_319204_thumb_ = g_mCalc.GetCurrentVAbyPreferedVA(ASM_GET_THUMB_ADDRESS_FOR_JUMP(0x319204));
		CPatch::RedirectCodeEx(INSTRUCTION_SET_THUMB, g_mCalc.GetCurrentVAbyPreferedVA(0x3191F0),
			(void*)&patch_CPathFind__Shutdown_3191F0
		);
		
		// CPathFind::LoadSceneForPathNodes 
		CPatch::RedirectMethod((uintptr_t)Library::GetSymbolAddress(
			&g_LimitAdjuster.hModule_of_game,
			"_ZN9CPathFind8ShutdownEv"
		), &CPathFind_extended::Shutdown);

		// CPathFind::LoadSceneForPathNodes 
		CPatch::RedirectMethod((uintptr_t)Library::GetSymbolAddress(
			&g_LimitAdjuster.hModule_of_game,
			"_ZN9CPathFind21LoadSceneForPathNodesE7CVector"
		), &CPathFind_extended::LoadSceneForPathNodes);

		// CStreaming::Init2
		Address_CStreaming__m_bHarvesterModelsRequested = (uintptr_t)Library::GetSymbolAddress(&g_LimitAdjuster.hModule_of_game,
			"_ZN10CStreaming27m_bHarvesterModelsRequestedE");

		Address_CStreaming__Init2_46BB82_thumb = g_mCalc.GetCurrentVAbyPreferedVA(ASM_GET_THUMB_ADDRESS_FOR_JUMP(0x46BB82));
		CPatch::RedirectCodeEx(INSTRUCTION_SET_THUMB, g_mCalc.GetCurrentVAbyPreferedVA(0x46BAB6),
			(void*)&patch_CStreaming__Init2_46BAB6
		);

		CPatch::NOPinstructions(INSTRUCTION_SET_THUMB, g_mCalc.GetCurrentVAbyPreferedVA(0x46BB84), 2);
		CPatch::NOPinstructions(INSTRUCTION_SET_THUMB, g_mCalc.GetCurrentVAbyPreferedVA(0x46BB8A), 2);
		CPatch::NOPinstructions(INSTRUCTION_SET_THUMB, g_mCalc.GetCurrentVAbyPreferedVA(0x46BB92), 4);
		CPatch::NOPinstructions(INSTRUCTION_SET_THUMB, g_mCalc.GetCurrentVAbyPreferedVA(0x46BB98), 2);
		CPatch::NOPinstructions(INSTRUCTION_SET_THUMB, g_mCalc.GetCurrentVAbyPreferedVA(0x46BBA0), 2);
		CPatch::NOPinstructions(INSTRUCTION_SET_THUMB, g_mCalc.GetCurrentVAbyPreferedVA(0x46BBA4), 2);

		
		Address_CStreaming__Init2_46BC5E_thumb = g_mCalc.GetCurrentVAbyPreferedVA(ASM_GET_THUMB_ADDRESS_FOR_JUMP(0x46BC5E));
		CPatch::RedirectCodeEx(INSTRUCTION_SET_THUMB, g_mCalc.GetCurrentVAbyPreferedVA(0x46BC54),
			(void*)&patch_CStreaming__Init2_46BC54
		);

		///// Shopping /////
		Address_CShopping__ShutdownForRestart_19BC1C_arm = g_mCalc.GetCurrentVAbyPreferedVA(ASM_GET_ARM_ADDRESS_FOR_JUMP(0x19BC1C));
		Address_CShopping__ShutdownForRestart_35F9BE_thumb = g_mCalc.GetCurrentVAbyPreferedVA(ASM_GET_THUMB_ADDRESS_FOR_JUMP(0x35F9BE));
		CPatch::RedirectCodeEx(INSTRUCTION_SET_THUMB, g_mCalc.GetCurrentVAbyPreferedVA(0x35F9B2),
			(void*)&patch_CShopping__ShutdownForRestart_35F9B2
		);

		Address_CShopping__RemoveLoadedShop_19BC1C_arm = g_mCalc.GetCurrentVAbyPreferedVA(ASM_GET_ARM_ADDRESS_FOR_JUMP(0x19BC1C));
		Address_CShopping__RemoveLoadedShop_35FA2A_thumb = g_mCalc.GetCurrentVAbyPreferedVA(ASM_GET_THUMB_ADDRESS_FOR_JUMP(0x35FA2A));
		CPatch::RedirectCodeEx(INSTRUCTION_SET_THUMB, g_mCalc.GetCurrentVAbyPreferedVA(0x35FA1E),
			(void*)&patch_CShopping__RemoveLoadedShop_35FA1E
		);

		Address_CShopping__LoadPrices_19BC1C_arm = g_mCalc.GetCurrentVAbyPreferedVA(ASM_GET_ARM_ADDRESS_FOR_JUMP(0x19BC1C));
		Address_CShopping__LoadPrices_35FCBA_thumb = g_mCalc.GetCurrentVAbyPreferedVA(ASM_GET_THUMB_ADDRESS_FOR_JUMP(0x35FCBA));
		CPatch::RedirectCodeEx(INSTRUCTION_SET_THUMB, g_mCalc.GetCurrentVAbyPreferedVA(0x35FCAE),
			(void*)&patch_CShopping__LoadPrices_35FCAE
		);

		Address_CShopping__RemoveLoadedPrices_19BC1C_arm = g_mCalc.GetCurrentVAbyPreferedVA(ASM_GET_ARM_ADDRESS_FOR_JUMP(0x19BC1C));
		Address_CShopping__RemoveLoadedPrices_36004C_thumb = g_mCalc.GetCurrentVAbyPreferedVA(ASM_GET_THUMB_ADDRESS_FOR_JUMP(0x36004C));
		CPatch::RedirectCodeEx(INSTRUCTION_SET_THUMB, g_mCalc.GetCurrentVAbyPreferedVA(0x360040),
			(void*)&patch_CShopping__RemoveLoadedPrices_360040
		);
	}
	#endif
}
#endif

// Patches ID limits if neccessary.
void FileIDlimit::PatchFileIDlimits_GTA_SA()
{
	MAKE_VAR_GAME_VERSION();
	using namespace Game_GTASA;
	
	if(!WriteToLogIfLimitsAreChanged(true))
		return;

	// ID limit checks
	#if TRUE

	// See if count is unsigned
	if (GetCountOfAllFileIDs() >= 32768)
	{
		if (gameVersion == GAME_VERSION_GTA_SA_1_0_US_HOODLUM_WIN_X86)
		{
			this->bIsIDunsigned = true;

			CGenericLogStorage::SaveFormattedTextLn("Count of file IDs is over 32767 and requires unsigned ID. Applying unsigned ID patches.");
		}
		else
			throw f92_runtime_error("Count of file IDs should be lower than 32768."
				"\nUnsigned ID is not supported yet."
				"\nCount of IDs: %d", GetCountOfAllFileIDs()
			);
	}

	// See if base ID of files should be stored in int32_t
	if (GetCountOfAllFileIDs() > MAX_FILE_ID_UINT16)
	{
		this->bIsIDbaseInt32 = true;
		this->EnableFileInfoExtensionArray();

		CGenericLogStorage::SaveFormattedTextLn(
			"Count of file IDs is over %d and base ID of CStreamingInfo requires int32_t. Applying int32_t base ID patches.",
			MAX_FILE_ID_UINT16
		);
	}

	// See if DFF or TXD ID should be stored in int32_t
	if (GetFileIDcurrentLimit(FILE_TYPE_DFF) > MAX_FILE_ID_UINT16 || GetFileIDcurrentLimit(FILE_TYPE_TXD) > MAX_FILE_ID_UINT16)
	{
		this->bAreDifficultFileTypesInt32 = true;
		GetFullModelIDfrom16bitRegisterBefore = &GetFullModelIDfrom16bitRegisterBefore_32_bit_value;

		CGenericLogStorage::SaveFormattedTextLn(
			"Count of DFF/TXD IDs is over %d and requires int32_t. Applying int32_t DFF+TXD ID patches.",
			MAX_FILE_ID_UINT16);
	}

	// See if DFF limit is a multiply of 10
	if (GetFileIDcurrentLimit(FILE_TYPE_DFF) % 10 != 0)
		throw f92_runtime_error("DFF limit should be a multiply of 10.\n%d models isn't a multiply of 10!",
			GetFileIDcurrentLimit(FILE_TYPE_DFF)
			);

	// See if TXD limit is a  multiply of 10
	if (GetFileIDcurrentLimit(FILE_TYPE_TXD) % 10 != 0)
		throw f92_runtime_error("TXD limit should be a multiply of 10.\n%d models isn't a multiply of 10!",
			GetFileIDcurrentLimit(FILE_TYPE_TXD)
			);
	
	if (GetFileIDcurrentLimit(FILE_TYPE_COL) > 256)
	{
		this->bIsCOLIDsizeIncreased = true;	

		CGenericLogStorage::SaveFormattedTextLn(
			"COL ID limit is over 256 and requires something more than uint8_t."
			" Applying COL ID uint32_t patches."
			);
	}

	if (GetFileIDcurrentLimit(FILE_TYPE_IPL) > 256)
	{
		this->bIsIPLIDsizeIncreased = true;
		
		CGenericLogStorage::SaveFormattedTextLn(
			"IPL ID limit is over 256 and requires something more than uint8_t."
			" Applying IPL ID uint32_t patches."
		);
	}

	// See if SCM limit can be packed into one signed byte.
	if (GetFileIDcurrentLimit(FILE_TYPE_SCM) >= 128)
		throw f92_runtime_error("SCM limit should be less than 128 files.\nCurrent value: %d.",
			GetFileIDcurrentLimit(FILE_TYPE_SCM)
			);
	#endif

	// Print new limits to log file
	#if TRUE
	{
		CGenericLogStorage::SaveFormattedTextLn(
			"\nNew ID limits:\n"
			"%d - %d (%d) - DFF models defined within IDE files\n"
			"%d - %d (%d) - TXD texture archives.\n"
			"%d - %d (%d) - COL collision archives.\n"
			"%d - %d (%d) - IPL Binary IPL files.\n"
			"%d - %d (%d) - DAT files limited to nodes*.dat\n"
			"%d - %d (%d) - IFP animation archives.\n"
			"%d - %d (%d) - RRR car recordings, carrec*.rrr files\n"
			"%d - %d (%d) - SCM scripts\n"
			"%d - %d (%d) - Loaded list\n"
			"%d - %d (%d) - Requested list\n"
			"%d - count of all file IDs\n"
			,

			GetBaseID(FILE_TYPE_DFF), GetLastIDforType(FILE_TYPE_DFF), GetFileIDcurrentLimit(FILE_TYPE_DFF),
			GetBaseID(FILE_TYPE_TXD), GetLastIDforType(FILE_TYPE_TXD) , GetFileIDcurrentLimit(FILE_TYPE_TXD),
			GetBaseID(FILE_TYPE_COL), GetLastIDforType(FILE_TYPE_COL) , GetFileIDcurrentLimit(FILE_TYPE_COL),
			GetBaseID(FILE_TYPE_IPL), GetLastIDforType(FILE_TYPE_IPL) , GetFileIDcurrentLimit(FILE_TYPE_IPL),
			GetBaseID(FILE_TYPE_DAT), GetLastIDforType(FILE_TYPE_DAT) , GetFileIDcurrentLimit(FILE_TYPE_DAT),
			GetBaseID(FILE_TYPE_IFP), GetLastIDforType(FILE_TYPE_IFP) , GetFileIDcurrentLimit(FILE_TYPE_IFP),
			GetBaseID(FILE_TYPE_RRR), GetLastIDforType(FILE_TYPE_RRR) , GetFileIDcurrentLimit(FILE_TYPE_RRR),
			GetBaseID(FILE_TYPE_SCM), GetLastIDforType(FILE_TYPE_SCM) , GetFileIDcurrentLimit(FILE_TYPE_SCM),
			GetBaseID(FILE_TYPE_LOADED_START), GetLastIDforType(FILE_TYPE_LOADED_START) , GetFileIDcurrentLimit(FILE_TYPE_LOADED_START),
			GetBaseID(FILE_TYPE_REQUESTED_START), GetLastIDforType(FILE_TYPE_REQUESTED_START) , GetFileIDcurrentLimit(FILE_TYPE_REQUESTED_START),

			GetCountOfAllFileIDs()
			);
	}
	#endif	
	
	// Allocate variables
	#if TRUE
	if (!CPatch::IsDebugModeActive())
	{
		// DFF
		if (GetFileIDcurrentLimit(FILE_TYPE_DFF) > GetFileIDdefaultLimit(FILE_TYPE_DFF))
		{
			CModelInfo__ms_modelInfoPtrs.gta_sa = new CBaseModelInfo*[GetBaseID(FILE_TYPE_TXD)];
			memset(CModelInfo__ms_modelInfoPtrs.gta_sa, NULL, GetBaseID(FILE_TYPE_TXD) * sizeof(void**));
			CModelInfo__ms_modelInfoPtrs.bIsAllocated.Set(true);
		}

		// CStreaming__ms_aInfoForModel
		if (GetCountOfAllFileIDs() > GetDefaultCountOfAllFileIDs())
		{
			CStreaming__ms_aInfoForModel.gta_sa = (CStreamingInfo*)new char[sizeof(CStreamingInfo) * GetCountOfAllFileIDs()];
			memset(CStreaming__ms_aInfoForModel.gta_sa, NULL, GetCountOfAllFileIDs() * sizeof(CStreamingInfo));
			CStreaming__ms_aInfoForModel.bIsAllocated.Set(true);
		}

		// IFP files
		if (GetFileIDcurrentLimit(FILE_TYPE_IFP) > GetFileIDdefaultLimit(FILE_TYPE_IFP))
		{
			CAnimManager__ms_aAnimBlocks.gta_sa = new CAnimBlock[GetFileIDcurrentLimit(FILE_TYPE_IFP)];
			memset(CAnimManager__ms_aAnimBlocks.gta_sa, NULL, GetFileIDcurrentLimit(FILE_TYPE_IFP) * sizeof(CAnimBlock));
			CAnimManager__ms_aAnimBlocks.bIsAllocated.Set(true);
		}

		// RRR
		if (GetFileIDcurrentLimit(FILE_TYPE_RRR) > GetFileIDdefaultLimit(FILE_TYPE_RRR))
		{
			CVehicleRecording__StreamingArray.gta_sa = new CVehicleRecordingFileInfo[GetFileIDcurrentLimit(FILE_TYPE_RRR)];
			CVehicleRecording__StreamingArray.bIsAllocated.Set(true);
		}

		// SCM
		if (GetFileIDcurrentLimit(FILE_TYPE_SCM) > GetFileIDdefaultLimit(FILE_TYPE_SCM))
		{
			CTheScripts__StreamedScripts.ptr = new char[GetFileIDcurrentLimit(FILE_TYPE_SCM) * sizeof(CStreamedScriptInfo) + sizeof(CStreamedScripts_Footer)];
			CTheScripts__StreamedScripts.bIsAllocated.Set(true);
		}
	}
	#endif

	CStreamedScriptInfo* CTheScripts__StreamedScriptsArray = (CStreamedScriptInfo*)CTheScripts__StreamedScripts.ptr;
	CStreamedScripts_Footer* CTheScripts__StreamedScriptsFooter =
		(CStreamedScripts_Footer*)(
		CTheScripts__StreamedScripts.ptr + GetFileIDcurrentLimit(FILE_TYPE_SCM) * sizeof(CStreamedScriptInfo));

	CGenericLogStorage::SaveFormattedTextLn("\nAddresses:\n"
		"CModelInfo::ms_modelInfoPtrs: 0x%X\n"
		"CStreaming::ms_aInfoForModel: 0x%X\n"
		"CAnimManager::ms_aAnimBlocks: 0x%X\n"
		"CVehicleRecording::StreamingArray: 0x%X\n"
		"CTheScripts::StreamedScripts: 0x%X\n\n"

		,
		CModelInfo__ms_modelInfoPtrs.gta_sa,
		CStreaming__ms_aInfoForModel.gta_sa,
		CAnimManager__ms_aAnimBlocks.gta_sa,
		CVehicleRecording__StreamingArray.gta_sa,
		CTheScripts__StreamedScripts.ptr
		);

	if (GetFileIDdefaultLimit(FILE_TYPE_IPL) != GetFileIDcurrentLimit(FILE_TYPE_IPL))
		CGenericLogStorage::SaveFormattedTextLn("Format of new savefiles will be different: patch for block of IPL flags is applied! Number of IPL flags will depend on current IPL ID limit.\n");

	/////////////////////////////////
	#ifdef IS_PLATFORM_WIN_X86
	if (gameVersion == GAME_VERSION_GTA_SA_1_0_US_HOODLUM_WIN_X86)
		this->PatchFileIDlimits_GTA_SA_PC_1_0_HOODLUM();
	#elif defined(IS_PLATFORM_ANDROID_ARM32)
	if(gameVersion == GAME_VERSION_GTA_SA_2_00_ANDROID_ARM32)
		this->PatchFileIDlimits_GTA_SA_2_0_ANDROID_ARM32();
	#endif

	/////////////////////////////////
	CGenericLogStorage::WriteLineSeparator();
}

#if IS_ARCHITECTURE_32_BIT
// Patches ID limits if neccessary.
void FileIDlimit::PatchFileIDlimits_GTA_IV_or_EFLC()
{
	MAKE_VAR_GAME_VERSION();
	using namespace Game_GTAIV;
			
	if(!WriteToLogIfLimitsAreChanged())
		return;

	// Print new limits to file
	#if TRUE
	{
		CGenericLogStorage::SaveFormattedTextLn(
			"\nNew ID limits:\n"
			"%d - DFF / WDR / WFT / LOD / NTH models defined within IDE files\n"
			"%d - WTD texture archives\n"
			"%d - MAP_PLACEMENT, IPL / WPL placement files\n"
			"%d - WAD, animation archives\n"
			"%d - RRR car recordings, carrec*.rrr files\n"
			"%d - count of all file IDs\n"
			,

			GetFileIDcurrentLimit(FILE_TYPE_MODEL),
			GetFileIDcurrentLimit(FILE_TYPE_WTD),
			GetFileIDcurrentLimit(FILE_TYPE_MAP_PLACEMENT),
			GetFileIDcurrentLimit(FILE_TYPE_WAD),
			GetFileIDcurrentLimit(FILE_TYPE_RRR),

			GetCountOfAllFileIDs()
			);
	}
	#endif

	// ID limit checks
	#if TRUE
	{
		for (int i = 0; i < g_fileIDlimits.GetCountOfTypes(); i++)
		{
			eFileType type = g_fileIDlimits.GetFileTypeInfoByIndex(i)->fileType;

			if(GetFileIDcurrentLimit(type) >= 32768)
				throw f92_runtime_error("Count of IDs of any type should be less than 32768."
					"\nUnsigned IDs for GTA IV file types are not supported yet."
					"\nCount of IDs for type %s: %d",
					CFileTypeInfo::GetEnumNameByMember(type),
					GetCountOfAllFileIDs()
				);
		}
	}
	#endif

	// Allocate variables
	#if TRUE
	if (!CPatch::IsDebugModeActive())
	{
		// g_pModelPointers
		if(GetFileIDcurrentLimit(FILE_TYPE_MODEL) > FileIDlimit::GetFileIDdefaultLimit(FILE_TYPE_MODEL))
		{
			CModelInfo__ms_modelInfoPtrs.gta_iv = new CBaseModelInfo*[GetFileIDcurrentLimit(FILE_TYPE_MODEL)];
			memset(CModelInfo__ms_modelInfoPtrs.gta_iv, NULL, GetFileIDcurrentLimit(FILE_TYPE_MODEL) * sizeof(void*));
			CModelInfo__ms_modelInfoPtrs.bIsAllocated.Set(true);
		}

		// RRR
		if (GetFileIDcurrentLimit(FILE_TYPE_RRR) > FileIDlimit::GetFileIDdefaultLimit(FILE_TYPE_RRR))
		{
			CVehicleRecording__StreamingArray.gta_iv = new CVehicleRecordingFileInfo[GetFileIDcurrentLimit(FILE_TYPE_RRR)];

			memset(
				CVehicleRecording__StreamingArray.gta_iv,
				NULL,
				GetFileIDcurrentLimit(FILE_TYPE_RRR) * sizeof(CVehicleRecordingFileInfo)
				);

			CVehicleRecording__StreamingArray.bIsAllocated.Set(true);
		}
	}
	#endif

	#ifdef IS_PLATFORM_WIN_X86
	if (gameVersion == GAME_VERSION_GTA_IV_1_0_7_0_WIN_X86)
		this->PatchFileIDlimits_GTA_IV_1_0_7_0();
	else if (gameVersion == GAME_VERSION_GTA_IV_1_0_8_0_WIN_X86)
		this->PatchFileIDlimits_GTA_IV_1_0_8_0();
	else if (gameVersion == GAME_VERSION_GTA_EFLC_1_1_2_0_WIN_X86)
		this->PatchFileIDlimits_GTA_EFLC_1_1_2_0();
	#endif

	CGenericLogStorage::WriteLineSeparator();
}
#endif

#ifdef IS_PLATFORM_WIN_X86
// Patches ID limits if neccessary.
void FileIDlimit::PatchFileIDlimits_GTA_IV_1_0_7_0()
{
	// Patch limit of models
	#if TRUE
	{
		// Patch pointers to CModelInfo__ms_modelInfoPtrs.gta_iv
		#if TRUE
		{
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x422252 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x4258A6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x425928 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x4261B6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x4296BA + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x42A0F4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x42A2A4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x42FFC4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x4304A8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x430CB0 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[esi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x430D25 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[esi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x43127E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[esi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x4312F3 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[esi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x4329DC + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x432A12 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[ebx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x434AB6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[esi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x434ADE + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[esi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x434B8C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[esi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x434BD8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[esi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x43A01C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x443D66 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x4453CB + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x4453D2 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x448255 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x45B88E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x490DE5 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x498522 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x498644 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x4D9E91 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x4DE016 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x4DE171 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x4DE25F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x4DE51D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x4DE829 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x4DEA95 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x4DEE04 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x4E4E24 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x4E5A0E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x4E5A4C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x4E6A95 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x4EADAD + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x4EBFD9 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x4F106D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x4F143B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x4F145B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x4F1531 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x4F15F6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x4F174A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x4F1C6B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x4F1C8B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x4F1D42 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x4F5060 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x4F530F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x4F6DBA + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x4F6F78 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x4FB02E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x4FB8E1 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x4FB91E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x4FB945 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x797F29 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x7A64C2 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebp), CModelInfo__ms_modelInfoPtrs.gta_iv[ebp*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x7B7138 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x7B72DF + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x7BE0B9 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x7BE197 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x7BE2D9 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x7BE542 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x7BE618 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x7BF5D8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x7BF713 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x7BF773 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x7BFB51 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x7C1752 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x7C25C2 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x7C2CCD + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x7C2DE8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ebx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x7C2ECC + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x7C2FF3 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x7D7840 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x7D7E8C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x7D7F8C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x7D81BA + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x7D8607 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x7D8B8E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x7D99C2 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x7D9BC6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x7D9C85 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x7DF7B4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebp), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x7DF991 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x7DFB60 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x7DFF10 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebp), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x7DFF9F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x7E0085 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x7E01ED + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x7E26D5 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x803828 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x80388B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8045FF + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x809C58 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x80A91E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x80EE83 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x80F028 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x80F04A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x80F348 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x81756F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x81A59C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x81EC6B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x81ED1B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x81EDCB + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x820742 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x820822 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x83AF42 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x83AF74 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x83B318 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x83B8C8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x83BC99 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x83DA84 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[esi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x83DB17 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[esi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x83DB84 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[esi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x83DC56 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x83E119 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x83E178 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x83E2BB + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x83E766 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x83EDB1 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x83F5B6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x83FC7A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x84019B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[esi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8402DA + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi), CModelInfo__ms_modelInfoPtrs.gta_iv[esi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x840410 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[esi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8405DA + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8415F5 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx), CModelInfo__ms_modelInfoPtrs.gta_iv[edi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x84181F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[edi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x84E07F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x854D56 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x866723 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x877806 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x87A5CD + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x88CD46 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x88F790 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x88F91C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x89144C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x891498 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x89529D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x895A92 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x89C3BB + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x89C488 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x89C6FB + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x89DAA3 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x89DB0B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x89E1B8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x89E220 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x89E2D4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x89E50E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x89F1A3 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x89FF1F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8A05C2 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8A31F1 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8A3881 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8A3B5A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8A4098 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi), CModelInfo__ms_modelInfoPtrs.gta_iv[esi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8A415F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8A421D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8A4272 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8A42BF + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8A4352 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8A43DC + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8A442D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8A44AC + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8A4518 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8A47A2 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8A4904 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8A50B7 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8A50D8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8A5A6E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8A5A96 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8A6794 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8A6EF4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8A6F59 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8A6FC1 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8A73F9 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8A7462 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8A76D9 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8A84CA + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8A96C2 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8AAA85 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8AAB1D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8AB2C2 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8AC120 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8ADE4B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8AEA90 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8AEB8F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8AECDA + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8AED36 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8AED5F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8AED89 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8AEDB3 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8B1120 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8B116F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8B11F9 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8B1241 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8B1608 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8B172A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8B17CD + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8B202A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8B2319 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8B23AA + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8B25B4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8B2674 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8B297A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8B2B5A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8B2D50 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8B2F64 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8B876E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8B87B2 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8B87FC + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8B8843 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8B9086 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8B90FD + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8B9148 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8B9184 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8B91C9 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8B9205 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8BA6F1 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8BA9D2 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8BAF40 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8BB3CE + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8BD058 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8BD096 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8BD244 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8BD47E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8C99D6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8CC614 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// cmp     CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4], 0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8CC98F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8CE255 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8D13A3 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8D4847 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi), CModelInfo__ms_modelInfoPtrs.gta_iv[edi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8D4C65 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8D56EA + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8D62C6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8D813D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8D8270 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8D82A6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8DAC8D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8DB8B9 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8DBE2D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8DC2C7 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8DC499 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8DC538 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8DC5A0 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8DC70A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8DC7DD + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8DCE67 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8DD219 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8DD571 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8DF5D5 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi), CModelInfo__ms_modelInfoPtrs.gta_iv[esi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8E0188 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8E0762 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8E08A4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8E1BAB + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8E2211 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8E34BB + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8E3AE3 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8E4738 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8E4781 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8E4A09 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8E4C9C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8E517B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebp), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8E550C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8E6E30 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8E6EA0 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8EACBC + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8ED848 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8ED8F4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8EE674 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8EF4F5 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[esi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8EF7DB + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[esi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8EF86A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[esi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8EFA3C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[esi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8EFAF5 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[esi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8EFBC1 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8EFC42 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8EFDAB + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8EFE3E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi), CModelInfo__ms_modelInfoPtrs.gta_iv[esi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8EFF43 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8EFF72 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[ebp*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8EFFF3 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F0022 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[ebp*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F01A8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F0276 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F02FA + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F036B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[esi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F0416 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F0476 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F0556 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F05C6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F064D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F06DD + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F0756 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F07CF + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F0878 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F0F16 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F0F40 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F0F54 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F0F6C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F0F7F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F0FC2 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F0FE9 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F1010 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F1192 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[esi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F11F3 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F13A0 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F14CA + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx), CModelInfo__ms_modelInfoPtrs.gta_iv[edi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F15E0 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F192B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[esi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F194F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F1974 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F19F8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F1A60 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F1B1F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[esi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F1F18 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[esi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F1F76 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[esi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F200F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F20E5 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F2199 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F2268 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F22D8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F2383 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F2403 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F262A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F268E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F2FB1 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[esi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F3135 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi), CModelInfo__ms_modelInfoPtrs.gta_iv[esi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F3252 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[esi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F349D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ebp*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F34E8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[esi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F3507 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[esi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F3539 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[esi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F354D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[esi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F3858 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[esi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F3877 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[esi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F3903 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[esi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F39C0 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[esi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F3E68 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F3E6F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F40C8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F40CF + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F41CF + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi), CModelInfo__ms_modelInfoPtrs.gta_iv[ebp*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F4258 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F4303 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F4380 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[edi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F4AFD + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F4D5A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F4F49 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F5018 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F507D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F5152 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F9E34 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8FCED3 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x901950 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x902B46 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x902B70 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x905D37 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x905D77 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x905F33 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x906075 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx), CModelInfo__ms_modelInfoPtrs.gta_iv[edi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9061B9 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x907AC6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x908436 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x908AEA + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x90F976 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x90FF47 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x917B7E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x91809A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x91D6EA + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x91D7CD + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x91DE45 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x91DF9E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x91EF43 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebp), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x921554 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x926EDA + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x92D210 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x932FC8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9335AF + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi), CModelInfo__ms_modelInfoPtrs.gta_iv[esi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9335DA + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi), CModelInfo__ms_modelInfoPtrs.gta_iv[esi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x933608 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi), CModelInfo__ms_modelInfoPtrs.gta_iv[esi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x933636 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi), CModelInfo__ms_modelInfoPtrs.gta_iv[esi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x933664 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi), CModelInfo__ms_modelInfoPtrs.gta_iv[esi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x933692 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi), CModelInfo__ms_modelInfoPtrs.gta_iv[esi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9336C0 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi), CModelInfo__ms_modelInfoPtrs.gta_iv[esi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9336EE + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi), CModelInfo__ms_modelInfoPtrs.gta_iv[esi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x93371C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi), CModelInfo__ms_modelInfoPtrs.gta_iv[esi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x93374A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi), CModelInfo__ms_modelInfoPtrs.gta_iv[esi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x933AF2 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x934C65 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9351A1 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9364BE + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9364D6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9366D8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x936700 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x948FE4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x94946E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9494AF + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9494F6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x94953D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x949DFA + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x949EC1 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x94A03E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x94F077 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x94F17E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x94F198 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x94F1B7 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x94F1D2 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x94F216 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x94F238 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x94F259 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x94F36D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x94F3E5 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x94F407 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x94F428 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x94F449 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x94F4A5 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x94F564 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x94F610 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x94F654 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x94F76A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x94F813 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x94FA65 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x94FA7C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x94FAA2 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x94FAB9 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x94FE3A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x94FE51 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x950443 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x950464 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x950679 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x950698 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x950F27 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x950F3F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x950F5A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x950F75 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x950F94 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x952408 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9525E5 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x952663 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x952690 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x952778 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9527D3 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x952811 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x95283E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9528CA + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x952925 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9529FA + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x952ABB + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x952B5D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x953202 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x95321D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x953238 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x953251 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x953268 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x953283 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9539E8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x953A68 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x953BEE + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x953D13 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x953DB0 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x953E4E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x953EC8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x953F3D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x95407A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9540DC + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9541DF + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x954A08 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x954CE3 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x955777 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x95579C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9558EB + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x955914 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x955938 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x95595C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x955EDD + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x955F0F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x955F44 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x955FC5 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x956111 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x95615F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x956178 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x956198 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9561B1 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x956231 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x95627F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x956296 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9562B8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9562CF + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9564F9 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9565C5 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x956DDE + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x956E12 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x957587 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x95762B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x957751 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x957841 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9578B7 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x957912 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x95796D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x957984 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9579A6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9579BD + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9579DF + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9579F8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x957A18 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x957A2F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x957A51 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x957A71 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x957C94 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x957CFD + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x957F59 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x958619 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x95863B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x95865B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x958B25 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x958B45 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x958B69 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9591A1 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9591BF + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9591F4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x95920B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x959387 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9598C4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9598DB + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9598F6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x95990D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x959971 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x959988 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9599A3 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9599BA + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x959C76 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x95A044 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x95A571 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x95AA48 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x95AB35 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x95ABDD + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x95AF13 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x95BFD6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x969733 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96EC4A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96FB0A + 1), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, offset CModelInfo__ms_modelInfoPtrs.gta_iv
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96FBB6 + 2), CModelInfo__ms_modelInfoPtrs.gta_iv + GetFileIDcurrentLimit(FILE_TYPE_MODEL));	// cmp     esi, offset g_escalatorAttrStore
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x97477A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x974794 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebp), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x975FCB + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9762BE + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x977D6A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x978836 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x978BF3 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x97A2B8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x97A3B2 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x97A5E4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x97B1F7 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x97EC01 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x97EC87 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x97ECA6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x97ED15 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x97EFBF + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x97F0D0 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x97F8F7 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x98038F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9803DC + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x981075 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x98293B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x982A11 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x983056 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9842C8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x98465D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x986D03 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x98711E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9871D6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9872AA + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x988445 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x988475 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9884A4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9884B4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9884C4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9884D7 + 1), CModelInfo__ms_modelInfoPtrs.gta_iv);	// push    offset CModelInfo__ms_modelInfoPtrs.gta_iv
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x988504 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9885B4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9885CB + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[esi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x98916B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9896F4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi), CModelInfo__ms_modelInfoPtrs.gta_iv[edi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x989766 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi), CModelInfo__ms_modelInfoPtrs.gta_iv[edi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9897D6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi), CModelInfo__ms_modelInfoPtrs.gta_iv[ebx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x989832 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi), CModelInfo__ms_modelInfoPtrs.gta_iv[ebx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x989946 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi), CModelInfo__ms_modelInfoPtrs.gta_iv[edi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9899C2 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi), CModelInfo__ms_modelInfoPtrs.gta_iv[ebx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x98A8E5 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x98AAD3 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x98AB29 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x98ABD6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi), CModelInfo__ms_modelInfoPtrs.gta_iv[edi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x98AC9A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4], esi
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x98AD09 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4], esi
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x98AD79 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4], esi
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x98ADE9 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4], esi
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x98AE5C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4], esi
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x98AECC + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4], esi
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x98AF39 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4], esi
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x98CFDE + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x98D019 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x98D090 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x98D0C8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x98E590 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi), CModelInfo__ms_modelInfoPtrs.gta_iv[edi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x99BEC4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9A7414 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9A9CB5 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9A9F84 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9AA1E0 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9AC590 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9B6320 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9B639F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9B6D95 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9B6F24 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9B8B35 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9B8BE6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9B8C67 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9B8C97 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9B94C9 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9BA095 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9BA13E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9BD200 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9C0AA6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9C5642 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9C5E55 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9C5FD4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9C6390 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9C653B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9C6564 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9C65C7 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[esi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9C65FC + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[esi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9C6676 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9C6AFB + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9C6B14 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9C6B24 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9C80FB + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9C817C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9C821A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9C82C3 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9C83ED + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9C8670 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9C8726 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9C87A7 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9C8B44 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9C8BA4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9C8BF5 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9CA966 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9CB478 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9CC2E6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9CC8B7 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9CC927 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9CC9D5 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9D0F1B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9D2A2F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9D2EC0 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9D2F24 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9D2F87 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9D2FEB + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9D304E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9D3114 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9D3C09 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9D3CB9 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9D3DEA + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9D3F64 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9D41F9 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9D42DB + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi), CModelInfo__ms_modelInfoPtrs.gta_iv[ebp*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9D4369 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ebp*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9D44F7 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9D4528 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9D45F7 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebp), CModelInfo__ms_modelInfoPtrs.gta_iv[ebx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9D49D9 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9D4A14 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9D4B44 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9D4BF2 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9D4FB4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9D5218 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9D5258 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9D5DC2 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9D5F0C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9D677A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9D67BE + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9D7B8B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9D7DEC + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9D7E01 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9D93F4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9DA17C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9DCC93 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9E04BD + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9E06DF + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9E0C0F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9E1696 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9E1D38 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9E1E81 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9E27D9 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9E2819 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9E311A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9E36C2 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9E39A4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9E68AA + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9E6AA0 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9E6AE4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9E6AF4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9E6B04 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9E6B15 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9E6DE0 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9E6DF4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9E6E32 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9E6FC5 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9E704E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9E71A4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9E71BB + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9E72A1 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9E7321 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9E73AF + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9E7AFA + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9E7C2D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi), CModelInfo__ms_modelInfoPtrs.gta_iv[ebx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9E7D0C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9E7ED8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9E828D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9E839D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9E84A0 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9E858F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9E8702 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9E8751 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9E87F5 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9E88A1 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9E895F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9E8A06 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9E8DBF + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9E95E8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9E9BAB + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebp), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9E9DA6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9EA116 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9EA58F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9EAB90 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9EAD1F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9F715E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9F77B6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebp), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9F7A17 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[esi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9F7BF4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9F7C14 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9F7C44 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9F92D8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9F9619 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9F9C8C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9F9DC2 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi), CModelInfo__ms_modelInfoPtrs.gta_iv[edi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9FA99A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9FAF6B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9FB00F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[esi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9FB01C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9FB8F8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9FBBC3 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9FC063 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9FC3C3 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9FCB9F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9FEF8B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi), CModelInfo__ms_modelInfoPtrs.gta_iv[edi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9FF02D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi), CModelInfo__ms_modelInfoPtrs.gta_iv[edi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA0ACC8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA1EB62 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA1F2AF + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA2178B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA22A4F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA23D11 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA240C1 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA24274 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA26699 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA26887 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA2C016 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA2C996 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA2D1EF + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA2DFA7 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA2FFC7 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA2FFE7 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA3075E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA309E2 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA30AB3 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA30E1C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA31DF7 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA31EEF + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA32DAE + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA32DF6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA32E88 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA33BF5 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA33BFC + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA33DC6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA33DCD + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA33E87 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA33E92 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA34136 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA3413D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA3BA2B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA3C1D2 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA3C664 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA3C6E4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA3C714 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA3C744 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA3C774 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA3C7A4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA3C7D4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA3C7F4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA3C824 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA3C854 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA3C874 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA3C8C4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA3C904 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA3C944 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA3C984 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA3C9C4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA3CA04 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA3CA44 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA3CA64 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA3D0B9 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA3D313 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA3D43D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA3DEF4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA3EBE1 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA3F587 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA3F7CF + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA3F8C4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA3F9A1 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA3FA0B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA40011 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA403D3 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA40885 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA40D77 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA4118C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA485F4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA489DC + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA48B36 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA4AF33 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA50248 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA508E8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA551E9 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA56147 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA594A8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA5DE36 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA5DEA1 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA5DEC9 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA5DF4F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA5E1F4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA5E81D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA61F22 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA620F7 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA6A277 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA6A29B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA6A4AB + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA6ABBC + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA733A9 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA733EA + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA734CD + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA7359D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA7A565 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA7A74C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA85C5C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA85CF6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA8624D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA88CE1 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA938CC + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA975EC + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA979EE + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA97A5C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA98395 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA984A0 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA99506 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA9991D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA9A32C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA9C285 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA9C4E3 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA9C6BE + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA9D384 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA9D715 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA9D75C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA9D794 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA9DC47 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA9DE89 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA9E29D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA9E5B4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA9E6E7 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA9E8F3 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA9E92E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA9E964 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA9E9DE + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA9EAF2 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA9EC7D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA9F16E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA9F191 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA9F1C3 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA9F36C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA9F69A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA9F6C5 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA9F6F1 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA9F71E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA9F840 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAA01D2 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAA02F9 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAA0379 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAA05BB + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAA09AB + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAA0C6E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAA0CE0 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAA101A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAA1764 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAA7114 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAA719A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAA7274 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAA72A8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAA73A4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAA7434 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAA7519 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAA75BE + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAA8C2B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAA964D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAA9B94 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAADEC4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAAE0CA + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAB6D00 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xABA0C7 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xABA9EC + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xABBDFE + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xABECD6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xABFDBC + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAC1A22 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ebx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAC1CFE + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAD506B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAD5191 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAD90D6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xADB8D6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xADE549 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xADE61A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAE1C73 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAE4569 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAEA48A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAEC854 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAEC94B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAEEB38 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAEF73A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAF0626 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAF0B12 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB09C24 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB0ED44 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB0F765 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB0FA0B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB0FA28 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB0FA68 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB0FA85 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB0FABB + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB0FE2E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB0FF87 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB1071C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB10D2A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB10E5D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB11F88 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB12A0C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB13ECB + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB1405B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB15035 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB15108 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB153DC + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB1CFB3 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB1E79A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB238DD + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB26E1C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB28902 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB28F3E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB28FA1 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB29013 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB29420 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB29DA9 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB2B274 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB2B436 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB2B535 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB2B5F1 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB2EE59 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB374D2 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB37B78 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB37BDC + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB38C4B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB38F41 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB39427 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB39960 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB3B77C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB3D320 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB3D3A3 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[esi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB3D3FC + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[esi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB3D521 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB3DA7D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB3E55D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB3E756 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[esi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB3E98D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB3EA47 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB3F748 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB3F886 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB3F8B5 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB43289 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB432FF + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB448CD + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB48E35 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB4906C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB4929B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB4931B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB4A753 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB4A800 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB51031 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB54CCA + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB59E23 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB6CFD5 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB6E760 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB6E777 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB70A2D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB70E9B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB72DC4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB72FC0 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB73219 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB7349D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB73703 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB7371A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB73813 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB73C45 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB76511 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB784A7 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB784D5 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB78509 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB78525 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB7858B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB7ABC2 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB7B39E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB7C884 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB7C8BA + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB7DBE4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB7DE6F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB8C857 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB92435 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB954F5 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB9552A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB95D4B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB95DD2 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB9B644 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB9BD46 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB9CDF4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB9CE54 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBC1A39 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBC1A9A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBC4A35 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBC4E6C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBC5E99 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ebx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBC60E5 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[esi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBC63AC + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBC6DDD + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBC6F6E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBC746D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBC748E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBC7673 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBC8794 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBC8FD4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBD778F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBD8FE8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBDA089 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBDA1EA + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBDA206 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBE160C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBE194B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBE37ED + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBE6033 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBE6A64 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBEBBBE + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBEBD5B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBEBE40 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBEBFD8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBEC48C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBECB44 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBED320 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBF48D7 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBF632A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBF78B2 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBF916B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBF919C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBFCED1 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBFD026 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebp), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBFD047 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBFD0DD + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBFD18D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBFD22E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBFD934 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBFDC6E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBFE0E3 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBFE137 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBFE1AC + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBFE1F6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBFE23E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBFE28B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC00464 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC00BD8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC01FEF + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC033BE + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edi*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC0343B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC0351C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC04FFF + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC05031 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC05113 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC0880F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC283C4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC283E4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC287B8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC2A234 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC2A295 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC2B145 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC2B1EE + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC2B2DB + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC2B700 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC2D7F6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC40CC1 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC4C8D6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC50050 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC50AA0 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC51058 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC58D3C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC58D74 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC59B62 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC5A680 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC5AC4D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC6661C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC66690 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC69DD6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC6A996 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC6CB26 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC6D090 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC6D0B7 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC6D12F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC6D6D2 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC74B02 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC795A1 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC798E4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC79E58 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC79E78 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC79E98 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC79EB8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC79F4F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC7A26F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC7A58F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC7A82F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC7AA74 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC7BB2E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC7C3BE + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC7DE49 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC7E3C6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC7F2B6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC7F2DF + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC7F4F9 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC7F532 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC7F562 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC7F5B8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC7FF02 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC80093 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC8CD73 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC8D5A2 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC91ACC + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC934B4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC9E048 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xCA792A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xCA858E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xCA89BA + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xCA9ABC + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xCA9B24 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xCA9BEC + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xCA9C20 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xCAE1D0 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xCAE4C0 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xCAE695 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xCAEB92 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xCAFBF9 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xCB098E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xCB0C12 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xCBA6E3 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xCBA84B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xCBB1A3 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xCBBED6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xCBC6BD + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xCC4D5B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xCC577F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xCC5B92 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xCC6684 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xCCD111 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xCCE3FF + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx), CModelInfo__ms_modelInfoPtrs.gta_iv[eax*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xCCF6FA + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[ecx*4]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xCCFEE1 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax), CModelInfo__ms_modelInfoPtrs.gta_iv[edx*4]
		}
		#endif

		// Patch values
		#if TRUE
		{
			// 31000 - count of models
			{
				CPatch::PatchUINT32(g_mCalc.GetCurrentVAbyPreferedVA(0x989751 + 2), GetFileIDcurrentLimit(FILE_TYPE_MODEL));	// cmp     edi, 31000
				CPatch::PatchUINT32(g_mCalc.GetCurrentVAbyPreferedVA(0x98A956 + 2), GetFileIDcurrentLimit(FILE_TYPE_MODEL));	// cmp     edi, 31000
				CPatch::PatchUINT32(g_mCalc.GetCurrentVAbyPreferedVA(0x98B03C + 1), GetFileIDcurrentLimit(FILE_TYPE_MODEL));	// push    31000
				CPatch::PatchUINT32(g_mCalc.GetCurrentVAbyPreferedVA(0x98E5B2 + 2), GetFileIDcurrentLimit(FILE_TYPE_MODEL));	// cmp     edi, 31000
				CPatch::PatchUINT32(g_mCalc.GetCurrentVAbyPreferedVA(0x98E5BA + 2), GetFileIDcurrentLimit(FILE_TYPE_MODEL));	// cmp     edi, 31000
			}
		}
		#endif
	}
	#endif

	// Patch WTD limit
	#if TRUE
	{
		CPatch::PatchUINT32(g_mCalc.GetCurrentVAbyPreferedVA(0x820DB6 + 1), GetFileIDcurrentLimit(FILE_TYPE_WTD));	// push    6000
		CPatch::PatchUINT32(g_mCalc.GetCurrentVAbyPreferedVA(0x820E2E + 1), GetFileIDcurrentLimit(FILE_TYPE_WTD));	// push    6000
	}
	#endif
	
	// Patch IPL/WPL limit
	#if TRUE
	{
		CPatch::PatchUINT32(g_mCalc.GetCurrentVAbyPreferedVA(0xB276F1 + 1), GetFileIDcurrentLimit(FILE_TYPE_MAP_PLACEMENT));	// push    500
		CPatch::PatchUINT32(g_mCalc.GetCurrentVAbyPreferedVA(0xB27742 + 1), GetFileIDcurrentLimit(FILE_TYPE_MAP_PLACEMENT));	// push    500
	}
	#endif

	// Patch WAD limit
	#if TRUE
	{
		CPatch::PatchUINT32(g_mCalc.GetCurrentVAbyPreferedVA(0x9098B3 + 1), GetFileIDcurrentLimit(FILE_TYPE_WAD));	// push    1500
		CPatch::PatchUINT32(g_mCalc.GetCurrentVAbyPreferedVA(0x909904 + 1), GetFileIDcurrentLimit(FILE_TYPE_WAD));	// push    1500
	}
	#endif

	// Patch RRR limit
	#if TRUE
	{
		// Patch pointers to CVehicleRecording::StreamingArray
		#if TRUE
		{
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96A1BB + 1), &CVehicleRecording__StreamingArray.gta_iv->usageCount);	// mov     eax, offset _ZN17CVehicleRecording14StreamingArrayE.usageCount
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96A1C8 + 1), &CVehicleRecording__StreamingArray.gta_iv[GetFileIDcurrentLimit(FILE_TYPE_RRR)].usageCount);	// cmp     eax, offset unk_15E32FC
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96A1E4 + 1), &CVehicleRecording__StreamingArray.gta_iv->pData);	// mov     esi, offset _ZN17CVehicleRecording14StreamingArrayE.pData
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96A20A + 2), &CVehicleRecording__StreamingArray.gta_iv[GetFileIDcurrentLimit(FILE_TYPE_RRR)].pData);	// cmp     esi, offset unk_15E32F4
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96A328 + 2), &CVehicleRecording__StreamingArray.gta_iv->usageCount);	// lea     eax, _ZN17CVehicleRecording14StreamingArrayE.usageCount[ecx]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96A3FD + 2), &CVehicleRecording__StreamingArray.gta_iv->number);	// mov     ecx, dword ptr _ZN17CVehicleRecording14StreamingArrayE.number[eax]; CVehicleRecording::StreamingArray
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96AAA8 + 1), CVehicleRecording__StreamingArray.gta_iv);	// mov     edx, offset _ZN17CVehicleRecording14StreamingArrayE; CVehicleRecording::StreamingArray
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96AAC3 + 2), &CVehicleRecording__StreamingArray.gta_iv->number);	// mov     dword ptr _ZN17CVehicleRecording14StreamingArrayE.number[eax], esi; CVehicleRecording::StreamingArray
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96AACC + 2), &CVehicleRecording__StreamingArray.gta_iv->pData);	// mov     _ZN17CVehicleRecording14StreamingArrayE.pData[eax], 0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96AB01 + 1), CVehicleRecording__StreamingArray.gta_iv);	// mov     ecx, offset _ZN17CVehicleRecording14StreamingArrayE; CVehicleRecording::StreamingArray
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96AB52 + 2), &CVehicleRecording__StreamingArray.gta_iv->pData);	// cmp     _ZN17CVehicleRecording14StreamingArrayE.pData[eax], ecx
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96AB69 + 1), &CVehicleRecording__StreamingArray.gta_iv->pData);	// mov     esi, offset _ZN17CVehicleRecording14StreamingArrayE.pData
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96AB95 + 2), &CVehicleRecording__StreamingArray.gta_iv[GetFileIDcurrentLimit(FILE_TYPE_RRR)].pData);	// cmp     esi, offset unk_15E32F4
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96ABB8 + 2), &CVehicleRecording__StreamingArray.gta_iv->pData);	// mov     eax, _ZN17CVehicleRecording14StreamingArrayE.pData[esi]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96ABBE + 2), &CVehicleRecording__StreamingArray.gta_iv->pData);	// lea     esi, _ZN17CVehicleRecording14StreamingArrayE.pData[esi]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96ABE7 + 2), &CVehicleRecording__StreamingArray.gta_iv->field_8);	// mov     ecx, _ZN17CVehicleRecording14StreamingArrayE.field_8[edx]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96ABED + 2), &CVehicleRecording__StreamingArray.gta_iv->pData);	// mov     eax, _ZN17CVehicleRecording14StreamingArrayE.pData[edx]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96AC4D + 2), &CVehicleRecording__StreamingArray.gta_iv->field_8);	// mov     edi, _ZN17CVehicleRecording14StreamingArrayE.field_8[edx]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96AD34 + 1), &CVehicleRecording__StreamingArray.gta_iv->usageCount);	// mov     esi, offset _ZN17CVehicleRecording14StreamingArrayE.usageCount
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96AD60 + 2), &CVehicleRecording__StreamingArray.gta_iv[GetFileIDcurrentLimit(FILE_TYPE_RRR)].usageCount);	// cmp     esi, offset unk_15E32FC
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96B984 + 2), &CVehicleRecording__StreamingArray.gta_iv->pData);	// mov     eax, _ZN17CVehicleRecording14StreamingArrayE.pData[esi]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96B9A1 + 2), &CVehicleRecording__StreamingArray.gta_iv->field_8);	// cmp     _ZN17CVehicleRecording14StreamingArrayE.field_8[esi], 0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96B9E4 + 2), &CVehicleRecording__StreamingArray.gta_iv->field_8);	// cmp     eax, _ZN17CVehicleRecording14StreamingArrayE.field_8[esi]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96BA0E + 2), &CVehicleRecording__StreamingArray.gta_iv->field_8);	// mov     ecx, _ZN17CVehicleRecording14StreamingArrayE.field_8[eax]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96BA14 + 2), &CVehicleRecording__StreamingArray.gta_iv->pData);	// mov     edx, _ZN17CVehicleRecording14StreamingArrayE.pData[eax]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96BA68 + 2), &CVehicleRecording__StreamingArray.gta_iv->pData);	// mov     _ZN17CVehicleRecording14StreamingArrayE.pData[esi], eax
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96BA73 + 2), &CVehicleRecording__StreamingArray.gta_iv->pData);	// mov     ecx, _ZN17CVehicleRecording14StreamingArrayE.pData[esi]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96BA84 + 2), &CVehicleRecording__StreamingArray.gta_iv->field_8);	// mov     _ZN17CVehicleRecording14StreamingArrayE.field_8[esi], eax
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96BA91 + 2), &CVehicleRecording__StreamingArray.gta_iv->field_8);	// cmp     _ZN17CVehicleRecording14StreamingArrayE.field_8[esi], eax
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96BA99 + 2), &CVehicleRecording__StreamingArray.gta_iv->pData);	// mov     ecx, _ZN17CVehicleRecording14StreamingArrayE.pData[esi]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96BAB2 + 2), &CVehicleRecording__StreamingArray.gta_iv->field_8);	// mov     _ZN17CVehicleRecording14StreamingArrayE.field_8[esi], eax
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96BABB + 2), &CVehicleRecording__StreamingArray.gta_iv->field_8);	// cmp     eax, _ZN17CVehicleRecording14StreamingArrayE.field_8[esi]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96BB61 + 2), &CVehicleRecording__StreamingArray.gta_iv->usageCount);	// add     _ZN17CVehicleRecording14StreamingArrayE.usageCount[eax], 1
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96BB68 + 2), &CVehicleRecording__StreamingArray.gta_iv->field_8);	// mov     ecx, _ZN17CVehicleRecording14StreamingArrayE.field_8[eax]
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96BB6E + 2), &CVehicleRecording__StreamingArray.gta_iv->pData);	// mov     edx, _ZN17CVehicleRecording14StreamingArrayE.pData[eax]
		}
		#endif

		// Patch count
		#if TRUE
		{
			CPatch::PatchUINT32(g_mCalc.GetCurrentVAbyPreferedVA(0x96C5D3 + 1), GetFileIDcurrentLimit(FILE_TYPE_RRR));	// push    2072
		}
		#endif
	}
	#endif
}

// Patches ID limits if neccessary.
void FileIDlimit::PatchFileIDlimits_GTA_IV_1_0_8_0()
{
	// Patch limit of models
	#if TRUE
	{
		// Patch pointers to CModelInfo__ms_modelInfoPtrs.gta_iv
		#if TRUE
		{
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x491FA6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_491F70
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x492028 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_491FD0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x4928B6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_4BDB50
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x496154 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_496150
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x496638 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_4965E0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x496E40 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_496DF0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x496EB5 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_496DF0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x49740E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_497390
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x497483 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_497390
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x498B6C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; 
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x498BA2 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs	; 
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x49AC46 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_497080
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x49AC6E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_49AC60
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x49AD1C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_49ACE0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x49AD68 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_49ACE0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x49EC0C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_49EBF0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x4A6AEA + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_4A6980
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x4A7524 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4A6980
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x4A76D4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4A6980
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x4AE0AB + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_4AD200
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x4AE0B2 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_4AD200
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x4B0F12 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_4B0F00
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x4B18C6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_4B18C0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x4B9025 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_4B8F20
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x4CADFE + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_4CAAF0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x6DD3E2 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_6DD3D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x6DD504 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_6DD3D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x6EB475 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_6EB3E0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x6FA734 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_6FA550
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x6FB31E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_6FB2B0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x6FB35C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_6FB2B0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x6FC221 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_6FC210
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x6FC25E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_6FC210
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x6FC285 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_6FC210
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x70641D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_706100
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x7067EB + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_706100
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x70680B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_706100
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x7068E1 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_706100
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x7069A6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_706100
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x706AFA + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_706100
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x70702B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_706EE0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x70704B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_706EE0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x707102 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_706EE0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x708D7D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_708CC0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x70B2F9 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_70B100
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x70DCA5 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_70DC60
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x7130F0 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_713080
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x71339F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_713390
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x714E5A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_714E40
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x715018 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_714E40
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x716121 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_716110
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x71A2A6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_71A110
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x71A401 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_71A110
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x71A4EF + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_71A110
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x71A7AD + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_71A110
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x71AAB9 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_71A110
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x71AD25 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_71A110
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x71B094 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_71A110
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x72753E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_727440
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x7CC85F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_7CC820
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x7CFA3C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_7CF800
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x7DA179 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_7DA0F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x7E8502 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebp, _ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs	; sub_7E8100
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x7F7813 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_7F7450
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x80A82B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_80A800
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x80A951 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_80A920
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x80E8A6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_80E780
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x812E06 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; 
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x815A79 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_815A10
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x815B4A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_815B10
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8191B3 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_818FB0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x81BAB9 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_81B3E0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8242BA + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8242B0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x826764 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8263F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x82685B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8263F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x828A48 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8289C0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x82964A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8295D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x82A536 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_82A520
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x82AA22 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_82A980
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x837508 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8374C0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x83D3D4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_83D260
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x83D5DA + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_83D480
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x84AA96 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_84A8E0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x85E6C9 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_85E6C0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x85E7A7 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_85E790
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x85E8E9 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_85E8D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x85EB52 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_85EAE0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x85EC28 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_85EC20
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x85FBE8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_85FBE0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x85FD23 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_85FD20
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x85FD83 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_85FD80
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x860181 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_860170
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x861E62 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_861E50
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x862CD2 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_862CC0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x86347D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8633A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x863588 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8634F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x86366C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_863620
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x863793 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_863740
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x864028 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_864020
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x86411B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8640F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x86B47E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_86B400
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x86F9D3 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_86F6F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x86FB78 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_86F6F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x86FB9A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_86F6F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x86FE98 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_86F6F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8ADAB5 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_8ADAA0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8ADE08 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8ADDF0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8ADE6B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8ADDF0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8AEBDF + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8AEBB0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8B4258 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8B40F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8B703D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8B6EC0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8CC662 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; 
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8CC742 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; 
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8CEA6E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8CE9E0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8CEAA9 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8CE9E0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8CEB20 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8CE9E0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8CEB58 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8CE9E0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8D139F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8D1A90
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8DAC15 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; 
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8DAC45 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; 
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8DAC74 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; 
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8DAC84 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; 
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8DAC94 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; 
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8DACA7 + 1), CModelInfo__ms_modelInfoPtrs.gta_iv);	// push    offset _ZN10CModelInfo16ms_modelInfoPtrsE; void *	; sub_8DACA0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8DACD4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; 
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8DAD84 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; 
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8DAD9B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; 
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8DAED0 + 1), CModelInfo__ms_modelInfoPtrs.gta_iv + GetFileIDcurrentLimit(FILE_TYPE_MODEL));	// mov     eax, offset unk_124A5C8	; sub_8DAED0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8DB95B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; 
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8DBEE4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_8DBEE0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8DBF56 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; 
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8DBFC6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8DBFC0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8DC022 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8DC010
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8DC136 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; 
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8DC1B2 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs	; 
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8DD0D5 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; 
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8DD2C3 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8DD270
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8DD319 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8DD2D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8DD3C6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_8DD3C0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8DD48A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4], esi; CModelInfo::ms_modelInfoPtrs	; sub_8DD450
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8DD4F9 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4], esi; CModelInfo::ms_modelInfoPtrs	; sub_8DD4C0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8DD569 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4], esi; CModelInfo::ms_modelInfoPtrs	; sub_8DD530
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8DD5D9 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4], esi; CModelInfo::ms_modelInfoPtrs	; sub_8DD5A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8DD64C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4], esi; CModelInfo::ms_modelInfoPtrs	; sub_8DD610
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8DD6BC + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4], esi; CModelInfo::ms_modelInfoPtrs	; sub_8DD680
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8DD729 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4], esi; CModelInfo::ms_modelInfoPtrs	; sub_8DD6F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8DDA10 + 1), CModelInfo__ms_modelInfoPtrs.gta_iv + GetFileIDcurrentLimit(FILE_TYPE_MODEL));	// mov     ecx, offset unk_124A5C8	; sub_8DD7B0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8E5734 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8E5730
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8E5D9C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8E5D80
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8E5EF6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8E5E00
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8E82F3 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8E82E0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8E8945 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_8E8930
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8E94F8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8E94F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8E9B02 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8E99E0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8E9C44 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8E9C30
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8EAF4B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8EAEA0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8EB5B1 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8EB560
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8EC85B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8EC710
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8ECE83 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8ECDA0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8EDAD8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8EDAD0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8EDB21 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8EDB10
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8EDDA9 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8EDD80
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8EE03C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8EE000
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8EE51B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebp, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8EE510
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8EE8AC + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8EE510
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8EF5B2 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8EF5A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8EFDC5 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8EFD90
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8EFF44 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8EFF40
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F0300 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8F02F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F04AB + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8F0470
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F04D4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8F0470
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F0537 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_8F0520
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F056C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_8F0520
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F05E6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8F05E0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F0A6B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8F0A60
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F0A84 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8F0A80
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F0A94 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8F0A90
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F2304 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8F2300
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F238B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8F2360
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F240C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8F2360
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F24AA + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8F2360
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F2553 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8F2360
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F267D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8F2630
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F2900 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8F2630
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F29B6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8F2990
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F2A37 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8F2990
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F2E94 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8F2DC0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F2EF4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8F2DC0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F2F45 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8F2DC0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F4CB6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8F4C30
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F57C8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8F4E80
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F6636 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8F65F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F6C07 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8F6C00
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F6C77 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8F6C70
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F6D25 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8F6CE0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8FB26B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8FB0B0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8FCD7F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8FCD50
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8FD210 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_8FCD50
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8FD274 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_8FCD50
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8FD2D7 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_8FCD50
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8FD33B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_8FCD50
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8FD39E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_8FCD50
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8FD464 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8FCD50
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8FDF59 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8FDF40
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8FE009 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8FDFD0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8FE13A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8FE130
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8FE2B4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8FE2B0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8FE549 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8FE540
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8FE62B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs	; sub_8FE540
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8FE6B9 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs	; sub_8FE540
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8FE877 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8FE760
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8FE8A8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8FE760
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8FE977 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebp, _ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8FE960
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8FED59 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8FED40
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8FED94 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8FED90
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8FEEC4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8FEE40
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8FEF72 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8FEF50
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8FF334 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8FEF50
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8FF598 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8FEF50
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8FF5D8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8FEF50
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x900142 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_900120
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9002BC + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9002B0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x902A5D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_902A30
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x902C7F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_902C60
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9031AF + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_903170
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x903C36 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_903B70
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9042D8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9042D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x904421 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_904410
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x90493A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_904920
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x904B30 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_904B10
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x904B74 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_904B70
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x904B84 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_904B80
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x904B94 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_904B90
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x904BA5 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_904BA0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x904E70 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_904E60
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x904E84 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_904E80
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x904EC2 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_904EB0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x905055 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_905030
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9050DE + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_905080
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x905234 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_905230
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x90524B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_905240
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x905331 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_905320
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9053B1 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9053A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x90543F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_905420
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x905B8A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_905B40
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x905CBD + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs	; sub_905CB0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x905D9C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_905D80
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x905F58 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_905F40
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x90630D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_906300
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x90641D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_906410
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x906520 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_906510
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x90660F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_906510
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x906782 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_906770
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9067D1 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9067C0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x906875 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_906860
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x906921 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_906860
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9069DF + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9069C0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x906A86 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9069C0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x906E3F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_906E20
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x907698 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_907640
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x907C5B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebp, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_907C50
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x907E56 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_907C50
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9081C6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_907EC0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x90869F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_908650
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x908CA0 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_908C80
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x908E2F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_908E10
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x90B0D3 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_90AD20
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x90B4EE + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_90AD20
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x90B5A6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_90AD20
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x90B67A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_90AD20
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x90EB7E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_90EA80
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x90F09A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_90F010
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x914244 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_914240
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9142D4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9142D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x914304 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_914300
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x914334 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_914330
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x914364 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_914360
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x914394 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_914390
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9143C4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9143C0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9143E4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9143E0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x914414 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_914410
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x914444 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_914440
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x914464 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_914460
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9144B4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9144B0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9144F4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9144F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x914534 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_914530
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x914574 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_914570
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9145B4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9145B0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9145F4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9145F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x914634 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_914630
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x914654 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_914650
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x914D99 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_914D90
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x914FF3 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_914FC0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x91511D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_915110
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x915BD4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_915BD0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9168C1 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9168A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x917267 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_917030
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9174AF + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_917030
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9175A4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_917030
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x917681 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_917670
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9176EB + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_917670
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x917CF1 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_917CE0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9180B3 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_918070
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x918565 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9184A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x918A57 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_918A50
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x918E5C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_918E30
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x91CEE2 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_91CED0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x91D51E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_91D3D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x91D581 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_91D570
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x91D5F3 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_91D5E0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x91DA00 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_91D9E0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x91E389 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; 
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x923C79 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_923C40
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x923CBA + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_923C90
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x923D9D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_923D60
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x923E6D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_923D60
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x92F5E4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_92F590
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x92F93E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_92F880
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x92FF96 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebp, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_92FF00
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9301F7 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_9301A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9303D4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9303D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9303F4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9303F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x930424 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_930420
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x931B18 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_931AB0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x931E59 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_931DC0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9324CC + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_932240
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x932602 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_932520
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9331DA + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_933150
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9337AB + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_933790
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x93384F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_933790
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x93385C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_933790
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x934458 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_934240
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x934723 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_934520
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x934BC3 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_934B80
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x934F23 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_934EE0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9356FF + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_935690
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x937F6B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_937F10
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x93800D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_937F10
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9397EB + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_9397E0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9398B8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_939890
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x939B2B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_939A60
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x93AED3 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_93AE50
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x93AF3B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_93AF30
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x93B5E8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_93B5E0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x93B650 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_93B620
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x93B704 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_93B700
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x93B93E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_836550
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x93C2C3 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_93C260
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x93D04F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_93CF60
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x93D6F2 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_93D670
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x940321 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_940230
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9409B1 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_940980
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x940C8A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_940C60
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9411C8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_941150
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x94128F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_941150
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x94134D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_941150
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9413A2 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_941150
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9413EF + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_941150
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x941482 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_941150
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x94150C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_941150
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x94155D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_941150
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9415DC + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_941150
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x941648 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_941150
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9418D2 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_941860
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x941A34 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_941860
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9421E7 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_942190
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x942208 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_942190
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x942B9E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_942960
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x942BC6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_942960
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9438C4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9438B0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x944024 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_943F70
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x944089 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_943F70
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9440F1 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_943F70
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x944529 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_944300
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x944592 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_944300
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x944809 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_944610
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9455FA + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_945470
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9467F2 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9466C0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x947BB5 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_947BB0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x947C4D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_947BB0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9483F2 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_948330
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x949250 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_949040
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x94D069 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_94D050
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x94D0CA + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_94D0C0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9502E5 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9502A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x95071C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9504B0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x951789 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9516D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x951C5C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_951C30
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x95268D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_952680
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x95281E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_952810
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x952D1D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_952D00
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x952D3E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_952D00
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x952F23 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_952E20
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x954044 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_953DB0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x954884 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_954120
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x956870 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_956820
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x956EBC + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_956EB0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x956FBC + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_956EB0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9571EA + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_957140
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x957637 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_957450
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x957BBE + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_957B50
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x958932 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_958870
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x958B36 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_958A10
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x958BF5 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_958A10
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x95E764 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebp, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_95E6A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x95E941 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_95E930
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x95EB10 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_95EAB0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x95EEC0 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebp, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_95EE70
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x95EF4F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_95EE70
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x95F035 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_95EFE0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x95F19D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_95EFE0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x960548 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9604A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x960686 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9604A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9606B5 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9604A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x967F30 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_967810
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96AA5D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_96A9B0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96B6A9 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_96B6A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96BC1D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_96BC10
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96C167 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_96C150
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96C339 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_96C330
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96C3D8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_96C3D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96C440 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_96C430
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96C5AA + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_96C5A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96C67D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_96C610
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96CD07 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_96CCF0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96D0B9 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_96D0B0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96D411 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_96D3D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x978964 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_978860
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x97B540 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_97B520
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x97B5B0 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_97B590
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x97F3CC + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_97F130
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x981F58 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_981490
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x982004 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_981490
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x982D84 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9822B0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x987E48 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_987E20
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x98842F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; 
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x98845A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; 
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x988488 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; 
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9884B6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; 
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9884E4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; 
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x988512 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; 
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x988540 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; 
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x98856E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; 
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x98859C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; 
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9885CA + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; 
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x988972 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_988910
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x989AE5 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_989AE0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x989FF1 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_989FB0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x98B30E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_98B2F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x98B326 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_98B2F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x98B528 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_98B2F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x98B550 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_98B2F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x98F0C8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_98F020
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x98F768 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_98F2C0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9962CB + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_996260
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x996AA2 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9967F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x999426 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_999420
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9A2E2B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9A2DE0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9A2EDB + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9A2E90
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9A2F8B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9A2F40
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9A46A4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; 
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9A4866 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; 
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9A4945 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9A4910
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9A4A01 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9A49C0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9A753A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9A6FE0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9B2CD4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9B2BE0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9B4D15 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9B4C80
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9B4FA3 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9B4F60
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9B517E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9B5130
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9B5E14 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9B5C80
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9B61A5 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9B6140
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9B61EC + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9B6140
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9B6224 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9B6140
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9B66D7 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9B6680
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9B6919 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9B68A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9B6D2D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9B6D00
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9B7044 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9B6FE0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9B7177 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9B7140
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9B7383 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9B72D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9B73BE + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9B72D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9B73F4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9B72D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9B746E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9B72D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9B7582 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9B72D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9B770D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9B7640
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9B7BFE + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9B7B30
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9B7C21 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9B7B30
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9B7C53 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9B7B30
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9B7DFC + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9B7D90
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9B812A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9B8070
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9B8155 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9B8070
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9B8181 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9B8070
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9B81AE + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9B8070
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9B82D0 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9B8290
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9B8CE2 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9B8C50
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9B8E09 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9B8DD0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9B8E89 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9B8DD0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9B90CB + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9B8FB0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9B94BB + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9B92D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9B977E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9B9680
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9B97F0 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9B9680
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9B9B2A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9B98B0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9BA274 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9BA150
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9BC960 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9BC8D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9BC9E3 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_9BC8D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9BCA3C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_9BC8D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9BCB61 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9BCB50
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9BD0BD + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9BD0B0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9BDB9D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9BDB10
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9BDD96 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_9BDB10
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9BDFCD + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9BDB10
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9BE087 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9BDB10
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9BE8D2 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9BE8C0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9BE8F4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9BE8F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9BEC98 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9BEBE0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9BF258 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9BF250
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9BF629 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9BF560
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9C1434 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_9C1040
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9C14C7 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_9C1040
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9C1534 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_9C1040
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9C1606 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9C1040
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9C1AC9 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9C1760
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9C1B28 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9C1760
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9C1C6B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9C1760
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9C2116 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9C1D60
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9C2761 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9C26C0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9C2F66 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9C2EE0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9C362A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9C30C0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9C3B4B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_9C30C0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9C3C8A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_9C30C0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9C3DC0 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_9C30C0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9C3F8A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9C30C0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9C4FA5 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_9C4F80
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9C51CF + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_9C4F80
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9C5C5E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9C5C40
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9C5CA2 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9C5C40
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9C5CEC + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9C5C40
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9C5D33 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9C5C40
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9C6536 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9C64D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9C65AD + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9C65A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9C65F8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9C65A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9C6634 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9C65A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9C6679 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9C65A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9C66B5 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9C65A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9C7BA1 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9C7B90
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9C7E82 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9C7E70
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9C83F0 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9C8200
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9C887E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9C8860
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9CA508 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_9CA410
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9CA546 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9CA410
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9CA6F4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9CA410
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9CA92E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9CA410
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9CBBF5 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_9CBBA0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9CBEC4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9CBE90
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9CC120 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9CC100
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9CE4D0 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9CE440
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9CE7AE + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9CE790
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9D035A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9D0190
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9D0E26 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9D0E00
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9D11E3 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9D1150
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9D28A8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9D26C0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9D29A2 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9D26C0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9D2BD4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9D26C0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9D3787 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9D3670
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9E4FA4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9E4F60
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9EA0F4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9EA0E0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9EAB15 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9EAB00
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9EADBB + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9EACC0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9EADD8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9EACC0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9EAE18 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9EACC0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9EAE35 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9EACC0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9EAE6B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9EACC0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9EB1DE + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9EB0D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9EB337 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9EB280
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9EBACC + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9EBA80
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9EC0DA + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9EC070
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9EC20D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9EC070
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9ED3A8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9ED220
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9EDDDC + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9EDD60
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9EED5B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9EED30
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9EF990 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9EF5E0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9EFA8F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9EF5E0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9EFBDA + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9EFBA0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9EFC36 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9EFBA0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9EFC5F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9EFBA0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9EFC89 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9EFBA0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9EFCB3 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9EFBA0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9F2020 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9F1F50
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9F206F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9F1F50
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9F20F9 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9F1F50
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9F2141 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9F1F50
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9F2508 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9F24E0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9F262A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9F24E0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9F26CD + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9F24E0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9F2F2A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9F2E60
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9F3219 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9F2E60
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9F32AA + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9F2E60
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9F34B4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9F34A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9F3574 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9F34A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9F387C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9F3780
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9F3A5A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9F3780
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9F3C50 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9F3780
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9F3E64 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9F3D70
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9F610B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9F5FB0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9F629B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9F6140
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9F7275 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9F71F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9F7348 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9F7300
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9F761C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9F7300
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9F8C5A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9F8BF0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9F8D3D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9F8BF0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9F93B5 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9F9380
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9F950E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9F94E0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9FA4B3 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebp, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9FA320
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA01063 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A00F90
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA028FA + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A027F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA07A3D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A079F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA09B52 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A099F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA09E2E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A09B80
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA0B73C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A0B4C0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA120A7 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A12090
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA120E7 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A120D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA122A3 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A12270
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA123E5 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_A12380
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA12529 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A124F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA162A9 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A16290
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA1DED2 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A1DEC0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA1E578 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A1E560
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA1E5DC + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A1E5A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA1F64B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A1F600
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA1F941 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A1F920
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA1FE27 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A1FE10
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA20360 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A1FE10
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA2217C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A22150
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA237E4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A237E0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA23804 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A23800
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA23BD8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A23B20
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA25584 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A25540
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA255E5 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A25540
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA26495 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A26370
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA2653E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A26370
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA2662B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A265C0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA26A50 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A26990
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA3169A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; 
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA3C490 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_A3C400
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA3D7B5 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_A3D790
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA3DA9B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_A3DA10
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA3DB2A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_A3DA10
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA3DCFC + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_A3DCC0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA3DDB5 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_A3DCC0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA3DE81 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A3DCC0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA3DF02 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A3DCC0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA3E05B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A3E050
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA3E0EE + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_A3E090
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA3E1F3 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A3E1C0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA3E222 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs	; sub_A3E1C0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA3E2A3 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A3E270
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA3E2D2 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs	; sub_A3E270
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA3E458 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A3E430
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA3E526 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A3E510
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA3E5AA + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A3E580
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA3E61B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_A3E580
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA3E6C6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A3E6B0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA3E726 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A3E710
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA3E806 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A3E7F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA3E876 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A3E860
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA3E8FD + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_A3E8D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA3E98D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_A3E960
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA3EA06 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A3E9F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA3EA7F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A3EA60
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA3EB28 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A3EB10
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA3F1C6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A3F030
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA3F1F0 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A3F030
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA3F204 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A3F030
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA3F21C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A3F030
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA3F22F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A3F030
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA3F272 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A3F030
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA3F299 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A3F030
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA3F2C0 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A3F030
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA3F442 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_A3F2F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA3F4A3 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A3F490
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA3F650 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A3F5C0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA3F77A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_A3F6F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA3F890 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A3F810
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA3FBDB + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_A3F900
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA3FBFF + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A3F900
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA3FC24 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A3FC20
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA3FCA8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A3FC50
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA3FD10 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A3FCF0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA3FDCF + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_A3FD50
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA401C8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_A401C0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA40226 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_A401C0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA402BF + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_A40290
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA40395 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_A40360
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA40449 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A40400
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA40518 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A404E0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA40588 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A404E0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA40633 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A404E0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA406B3 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A404E0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA408DA + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_A40890
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA4093E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_A40890
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA41261 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_A411F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA413E5 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_A411F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA41502 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_A41460
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA4174D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs	; sub_A41650
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA41798 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_A41790
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA417B7 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_A41790
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA417E9 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_A417E0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA417FD + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_A417E0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA41B08 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_A41B00
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA41B27 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_A41B00
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA41BB3 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_A41B50
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA41C70 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_A41B50
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA42118 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A41F60
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA4211F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A41F60
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA42378 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A41F60
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA4237F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A41F60
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA4247F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs	; sub_A41F60
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA42508 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A41F60
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA425B3 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_A41F60
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA42630 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_A41F60
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA464A4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A464A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA4B65E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A4B640
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA4B69F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A4B640
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA4B6E6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A4B640
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA4B72D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A4B640
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA4BFBA + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A4BF60
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA4C081 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A4C070
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA4C1FE + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A4C1E0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA51001 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A50EE0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA51156 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebp, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A51130
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA51177 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A51130
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA5120D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A51130
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA512BD + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A51130
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA5135E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A51130
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA51A64 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A51A60
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA51D9E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A51D70
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA52243 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A52200
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA52297 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A52200
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA5230C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A52200
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA52356 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A52200
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA5239E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A52200
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA523EB + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A52200
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA546C4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A546B0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA54E38 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A54E00
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA5623F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A56230
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA5835A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A58350
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA58B41 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A58900
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA62A60 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A623F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA62ADF + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A623F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA634D5 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A623F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA63664 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A623F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA66D93 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A66D60
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA678D1 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A67820
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA69B56 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A69B40
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA69BC1 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A69B40
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA69BE9 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A69B40
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA69C6F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A69B40
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA69F14 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A69EC0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA6A53D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A6A4F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA6C5D6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A6C5C0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA6C5FF + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A6C5C0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA6C819 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A6C800
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA6C852 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A6C800
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA6C882 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A6C800
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA6C8D8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A6C800
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA6D735 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A6D4A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA6D73C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A6D4A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA6D906 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A6D4A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA6D90D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A6D4A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA6D9C7 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A6D4A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA6D9D2 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A6D4A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA6DC76 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A6D4A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA6DC7D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A6D4A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA6E3A6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A6E120
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA70515 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A70440
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA789F6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A78980
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA799DA + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A799D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA79A1E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A79A10
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA7ADEB + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A7ACE0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA7B04C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A7B040
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA7B061 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A7B040
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA7C654 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A7C440
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA7D3DC + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A7D380
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA7FEF3 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A7FEC0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA82F16 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A82DD0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA848E8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A848E0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA85999 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A857B0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA85AFA + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A857B0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA85B16 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A857B0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA8BA3C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A8B9B0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA8BD7B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A8BCE0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA8DC1D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A8DAA0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA9CE8F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A9C560
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA9FE75 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A9FDF0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA9FF26 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A9FF00
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA9FFA7 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A9FF00
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA9FFD7 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A9FF00
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAA0809 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A9FF00
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAA13D5 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_AA13A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAA147E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_AA13A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAA44A0 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_AA40D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAA630C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_AA62D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAA670E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_AA66E0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAA677C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_AA6740
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAA70B5 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_AA7000
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAA71C0 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_AA7000
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAA8226 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_AA81E0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAA863D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_AA8390
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAA904C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_AA9010
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAAB2D0 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_AAB210
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAB3C69 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_AB31F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAB4BC7 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_AB4A20
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAB6988 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_AB6980
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAB69A8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_AB69A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAB69C8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_AB69C0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAB69E8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_AB69E0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAB6A7F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_AB6A00
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAB6D9F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_AB6D20
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAB70BF + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_AB7040
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAB735F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_AB72E0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAB75A4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_AB7520
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAB865E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_AB85E0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAB8EEE + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_AB8E80
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xABA979 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_ABA8E0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xABAEF6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_ABAE80
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xABB9A6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_ABB960
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xABE5F4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// cmp     _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4], 0; CModelInfo::ms_modelInfoPtrs	; sub_ABE5E0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xABE96F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_ABE930
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAC0235 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_AC0220
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAC2A64 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_AC2A60
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAC2F91 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_AC2F80
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAC3468 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_AC3460
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAC3555 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_AC3550
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAC35FD + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_AC35C0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAC3883 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_AC3820
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAC4946 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_AC4810
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAC59E3 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_AC59A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAC8E27 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_AC8E00
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAC9255 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_AC91D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAC9C5A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_AC9920
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xACA786 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_ACA5D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xACC5FD + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_ACC5E0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xACC730 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_ACC5E0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xACC766 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_ACC5E0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xACD82A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_ACD790
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xACF1E0 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_ACF190
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xACF1F7 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_ACF190
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAD14AD + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_AD12A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAD191B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_AD1710
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAD3844 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_AD3830
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAD3A40 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_AD39D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAD3C99 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_AD39D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAD3F1D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_AD3EA0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAD4183 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_AD4120
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAD419A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_AD4120
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAD4293 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_AD4210
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAD46C5 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_AD4690
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAD6F91 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_AD6F80
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAD9099 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_AD9090
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAD90D9 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_AD90D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAD99DA + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_AD99D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAD9F82 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_AD9F60
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xADA264 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_ADA240
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xADCA79 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_ADCA50
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xADCAEF + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_ADCA50
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xADE0BD + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_ADE070
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAE2635 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_AE2130
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAE285C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_AE2830
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAE2A8B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_AE2A80
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAE2B0B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_AE2B00
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAE3F43 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_AE3EF0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAE3FF0 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_AE3EF0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAEEC03 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_AEEAE0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAF8796 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_AF8770
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAF8D67 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_AF8CB0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB19DD5 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_B19DD0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB1BD27 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_B1BCB0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB1BD55 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; 
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB1BD89 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; 
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB1BDA5 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; 
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB1BE0B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_B1BDF0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB1E442 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_B1E420
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB1EC1E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_B1EB60
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB20104 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_B200F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB2013A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_B20120
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB21474 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_B21460
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB216FF + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_B216B0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB30177 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_B300A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB35DB5 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_B35DB0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB38E35 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; 
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB38E6A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; 
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB3968B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_B394F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB39712 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_B394F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB3EF44 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_B3EF30
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB3F646 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_B3F610
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB40754 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_B40740
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB407B4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_B407A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB628B2 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_B628A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB6300F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_B62FE0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB654EB + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_B652D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB667AF + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_B66770
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB67A71 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_B678E0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB67E21 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_B67B20
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB67FD4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_B67B20
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB6A3F9 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_B6A2D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB6A5E7 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_B6A2D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB6EAE1 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_B6E3A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB6EF14 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_B6EF00
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB6EF9A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_B6EF00
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB6F074 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_B6F060
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB6F0A8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_B6F060
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB6F1A4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_B6F140
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB6F234 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_B6F200
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB6F319 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_B6F280
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB6F3BE + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_B6F390
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB717F6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_B71750
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB89CBA + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_B89BA0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB89CD4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebp, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_B89BA0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB8B4DB + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_B8B470
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB8FEAA + 1), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, offset _ZN10CModelInfo16ms_modelInfoPtrsE; CModelInfo::ms_modelInfoPtrs	; sub_B8FEA0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB8FF56 + 2), CModelInfo__ms_modelInfoPtrs.gta_iv + GetFileIDcurrentLimit(FILE_TYPE_MODEL));	// cmp     esi, offset unk_124A5C8	; sub_B8FEA0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB97873 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_B977B0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB979DB + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_B977B0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB98333 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_B981E0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB9869E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_B98690
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB9871B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_B98700
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB987FC + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_B987F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB9A2AF + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_B9A0E0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB9A2E1 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_B9A0E0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB9A3C3 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_B9A0E0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB9DABF + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_B9DAB0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB9EC96 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_B9EC20
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBA0087 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BA0000
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBA09AC + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BA0940
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBA2D16 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_BA2D10
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBA2E23 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BA2E10
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBA7ABE + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BA7940
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBA7EEA + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_BA7EC0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBA8FEC + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_BA8FD0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBA9054 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_BA8FD0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBA911C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_BA9110
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBA9150 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BA9110
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBA9A21 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_BA9A00
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBAE50D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BAE300
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBAE76A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_BAE300
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBAE959 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_BAE300
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBAEA28 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_BAE9C0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBAEA8D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_BAE9C0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBAEB62 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BAE9C0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBB34B4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_BB3410
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBB6103 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BB5F30
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBBE7C7 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BBE470
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBBE7EB + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BBE470
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBBE9FB + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BBE960
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBBF10C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_BBEF00
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBC55A0 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_BC5000
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBC7007 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_BCECE0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBC710E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_BC7100
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBC7128 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BC7100
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBC7147 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BC7100
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBC7162 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BC7100
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBC71A6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_BC7190
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBC71C8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BC7190
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBC71E9 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_BC7190
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBC72FD + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_BC7190
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBC7375 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_BC7360
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBC7397 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BC7360
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBC73B8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_BC7360
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBC73D9 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_BC7360
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBC7435 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_BC7360
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBC74F4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BC7360
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBC75A0 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BC7560
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBC75E4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_BC7560
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBC76FA + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BC76F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBC77A3 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BC7750
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBC79F5 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BC7960
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBC7A0C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BC7960
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBC7A32 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BC7960
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBC7A49 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BC7960
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBC7DCA + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_BC7960
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBC7DE1 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BC7960
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBC83D3 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BC83C0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBC83F4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BC83C0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBC8609 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_BC85F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBC8628 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BC85F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBC8EA7 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BC8E90
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBC8EBF + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BC8E90
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBC8EDA + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BC8E90
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBC8EF5 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BC8E90
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBC8F14 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BC8E90
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBCA388 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BCA1D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBCA565 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BCA1D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBCA5E3 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BCA1D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBCA610 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BCA1D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBCA6F8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BCA1D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBCA753 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BCA1D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBCA791 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BCA1D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBCA7BE + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BCA1D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBCA84A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BCA1D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBCA8A5 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BCA1D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBCA97A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_BCA1D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBCAA3B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_BCA1D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBCAADD + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BCA1D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBCB182 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BCB160
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBCB19D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BCB160
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBCB1B8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BCB160
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBCB1D1 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BCB160
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBCB1E8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BCB160
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBCB203 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BCB160
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBCB968 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_BCB8B0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBCB9E8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BCB8B0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBCBB6E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BCB8B0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBCBC93 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BCB8B0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBCBD30 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BCB8B0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBCBDCE + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BCBDC0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBCBE48 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BCBDC0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBCBEBD + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BCBDC0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBCC00A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BCBF60
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBCC06C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_BCBF60
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBCC16F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_BCBF60
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBCC998 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BCC980
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBCCC73 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BCC980
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBCD707 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BCD6F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBCD72C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BCD6F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBCD87B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BCD6F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBCD8A4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BCD6F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBCD8C8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BCD6F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBCD8EC + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BCD6F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBCDE6D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_BCD6F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBCDE9F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BCD6F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBCDED4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BCD6F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBCDF55 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_BCD6F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBCE0A1 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BCD6F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBCE0EF + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BCD6F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBCE108 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BCD6F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBCE128 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BCD6F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBCE141 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BCD6F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBCE1C1 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BCD6F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBCE20F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BCD6F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBCE226 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BCD6F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBCE248 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BCD6F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBCE25F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BCD6F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBCE489 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BCE420
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBCE555 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BCE530
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBCED6E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BCED20
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBCEDA2 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_BCED90
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBCF517 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BCF500
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBCF5BB + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BCF500
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBCF6E1 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BCF500
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBCF7D1 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BCF500
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBCF847 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BCF500
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBCF8A2 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BCF500
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBCF8FD + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BCF500
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBCF914 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BCF500
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBCF936 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BCF500
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBCF94D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BCF500
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBCF96F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BCF500
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBCF988 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BCF500
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBCF9A8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BCF500
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBCF9BF + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BCF500
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBCF9E1 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BCF500
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBCFA01 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BCF500
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBCFC24 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BCF500
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBCFC8D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_BCF500
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBCFEE9 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BCFD00
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBD05A9 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BCFD00
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBD05CB + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BCFD00
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBD05EB + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BCFD00
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBD0AB5 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_BCFD00
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBD0AD5 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BCFD00
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBD0AF9 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BCFD00
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBD1131 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BD10A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBD114F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BD10A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBD1184 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BD10A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBD119B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BD10A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBD1317 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BD10A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBD1854 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BD10A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBD186B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BD10A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBD1886 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BD10A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBD189D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BD10A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBD1901 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BD10A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBD1918 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BD10A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBD1933 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BD10A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBD194A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BD10A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBD1C06 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BD1A20
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBD23C4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BD2380
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBD42AC + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_BD4290
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBD42E4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BD4290
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBD5052 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BD4CB0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBD5BA0 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BD5B90
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBD615D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_BD6120
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBE1B4C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BE1860
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBE1BC0 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_BE1860
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBE5391 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_BE4FC0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBE5417 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BE4FC0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBE5436 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_BE4FC0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBE54A5 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BE4FC0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBE573F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BE5500
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBE5850 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_BE5500
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBE6087 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BE6020
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBE6B1F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_BE6AF0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBE6B6C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BE6AF0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBE7885 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BE6EC0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBE914B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_BE9110
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBE9221 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BE9210
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBE9866 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BE97D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBEAAD8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_BEAA60
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBEAE6D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BEAA60
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBEC948 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; 
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC08906 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_C088F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC094C6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_C094B0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC0B656 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_C0B640
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC0BBC0 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_C0BAF0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC0BBE7 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_C0BAF0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC0BC5F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_C0BAF0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC0C202 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_C0BF70
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC0C330 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_C0C310
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC0C4BC + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_C0C4B0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC0DBBC + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_C0DA70
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC0DC08 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_C0DA70
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC119DD + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_C117B0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC121A2 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_C12100
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC1A826 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_C1A350
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC1B90C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_C1B740
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC23C25 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_C23A20
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC23E0C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_C23A20
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC2F29C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_C2EE50
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC2F336 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_C2EE50
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC2F88D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_C2EE50
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC32321 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_C316E0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC3CECC + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_C3C4A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC506EA + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_C506E0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC54AAE + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_C549A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC54C4B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_C549A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC54D30 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_C549A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC54EC8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_C54EC0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC5537C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_C55310
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC55A34 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_C55880
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC56200 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_C55F40
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC5D7D7 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_C5CDE0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC5F22A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_C5F080
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC607B2 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_C60770
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC6205B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_C61D10
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC6208C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_C61D10
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC6DB02 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_C6D020
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC6DCD7 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_C6D020
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC6E2E6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_C6E260
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC6EC76 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_C6EC00
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC6F4FF + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_C6EFC0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC702B7 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_C6FF00
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC722D7 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_C710D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC722F7 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_C710D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC72A6E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_C729D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC72CF2 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_C729D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC72DC3 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_C729D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC7312C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_C730B0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC74107 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_C73DB0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC741FF + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_C73DB0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC750BE + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_C74420
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC75106 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_C74420
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC75198 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_C74420
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC75EF0 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_C75EE0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC761E0 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_C761D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC763B5 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_C763A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC768A2 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_C76890
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC77909 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; 
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC7869E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_C78300
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC78922 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_C78910
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC7D442 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; 
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC7D5D3 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_C7D520
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC7F552 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_C7F510
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC81308 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_C81300
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC8CD44 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_C8CD40
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC8F080 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_C8F000
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC8FAD0 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_C8FAA0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC90088 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_C90080
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC98453 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_C98270
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC98C82 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_C98C00
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC9B3FC + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_C9B390
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC9D0F4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_C9D080
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC9F0B6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_C9EE60
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC9F0E0 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_C9EE60
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xCA3FB6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_CA3F90
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xCB7C2E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_CB7B60
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xCBA78B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_CBA730
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xCBB1AD + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_CBB180
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xCBB6F4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_CBB690
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xCC2D1A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_CC2AD0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xCCB6DB + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_CCB630
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xCCC0FF + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_CCC0A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xCCC512 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_CCC4A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xCCCF7D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_CCCF30
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xCE12A1 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_CE1290
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xCE258F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_CE24E0
		}
		#endif

		// Patch values
		#if TRUE
		{
			// 31000 - count of models
			{
				CPatch::PatchUINT32(g_mCalc.GetCurrentVAbyPreferedVA(0x8DBF41 + 2), GetFileIDcurrentLimit(FILE_TYPE_MODEL));	// cmp     edi, 7918h	; sub_8DBEE0
				CPatch::PatchUINT32(g_mCalc.GetCurrentVAbyPreferedVA(0x8DD146 + 2), GetFileIDcurrentLimit(FILE_TYPE_MODEL));	// cmp     edi, 7918h	; sub_8DD110
				CPatch::PatchUINT32(g_mCalc.GetCurrentVAbyPreferedVA(0x8DD7FC + 1), GetFileIDcurrentLimit(FILE_TYPE_MODEL));	// push    7918h	; sub_8DD7B0
				CPatch::PatchUINT32(g_mCalc.GetCurrentVAbyPreferedVA(0xA3C4B2 + 2), GetFileIDcurrentLimit(FILE_TYPE_MODEL));	// cmp     edi, 7918h	; sub_A3C400
				CPatch::PatchUINT32(g_mCalc.GetCurrentVAbyPreferedVA(0xA3C4BA + 2), GetFileIDcurrentLimit(FILE_TYPE_MODEL));	// cmp     edi, 7918h	; sub_A3C400
			}
		}
		#endif
	}
	#endif

	// Patch WTD limit
	#if TRUE
	{
		CPatch::PatchUINT32(g_mCalc.GetCurrentVAbyPreferedVA(0x8CCCD6 + 1), GetFileIDcurrentLimit(FILE_TYPE_WTD));	// push    6000	; sub_8CCCD0
		CPatch::PatchUINT32(g_mCalc.GetCurrentVAbyPreferedVA(0x8CCD4E + 1), GetFileIDcurrentLimit(FILE_TYPE_WTD));	// push    1770h	; sub_8CCCD0
	}
	#endif

	// Patch IPL/WPL limit
	#if TRUE
	{
		CPatch::PatchUINT32(g_mCalc.GetCurrentVAbyPreferedVA(0xA0C0D1 + 1), GetFileIDcurrentLimit(FILE_TYPE_MAP_PLACEMENT));	// push    500
		CPatch::PatchUINT32(g_mCalc.GetCurrentVAbyPreferedVA(0xA0C122 + 1), GetFileIDcurrentLimit(FILE_TYPE_MAP_PLACEMENT));	// push    500
	}
	#endif

	// Patch WAD limit
	#if TRUE
	{
		CPatch::PatchUINT32(g_mCalc.GetCurrentVAbyPreferedVA(0xA324C3 + 1), GetFileIDcurrentLimit(FILE_TYPE_WAD));	// push    1500
		CPatch::PatchUINT32(g_mCalc.GetCurrentVAbyPreferedVA(0xA32514 + 1), GetFileIDcurrentLimit(FILE_TYPE_WAD));	// push    1500
	}
	#endif

	// Patch RRR limit
	#if TRUE
	{
		// Patch pointers to CVehicleRecording::StreamingArray
		#if TRUE
		{
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96E15B + 1), &CVehicleRecording__StreamingArray.gta_iv->usageCount);	// mov     eax, offset _ZN17CVehicleRecording14StreamingArrayE.usageCount	; sub_96E130
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96E168 + 1), &CVehicleRecording__StreamingArray.gta_iv[GetFileIDcurrentLimit(FILE_TYPE_RRR)].usageCount);	// cmp     eax, offset unk_12D7AFC	; sub_96E130
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96E184 + 1), &CVehicleRecording__StreamingArray.gta_iv->pData);	// mov     esi, offset _ZN17CVehicleRecording14StreamingArrayE.pData	; sub_96E180
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96E1AA + 2), &CVehicleRecording__StreamingArray.gta_iv[GetFileIDcurrentLimit(FILE_TYPE_RRR)].pData);	// cmp     esi, offset unk_12D7AF4	; sub_96E180
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96E2C8 + 2), &CVehicleRecording__StreamingArray.gta_iv->usageCount);	// lea     eax, _ZN17CVehicleRecording14StreamingArrayE.usageCount[ecx]	; sub_96E270
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96E39D + 2), &CVehicleRecording__StreamingArray.gta_iv->number);	// mov     ecx, dword ptr _ZN17CVehicleRecording14StreamingArrayE.number[eax]; CVehicleRecording::StreamingArray	; sub_96E2F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96EA48 + 1), CVehicleRecording__StreamingArray.gta_iv);	// mov     edx, offset _ZN17CVehicleRecording14StreamingArrayE; CVehicleRecording::StreamingArray	; _ZN17CVehicleRecording21RegisterRecordingFileEPKc
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96EA63 + 2), &CVehicleRecording__StreamingArray.gta_iv->number);	// mov     dword ptr _ZN17CVehicleRecording14StreamingArrayE.number[eax], esi; CVehicleRecording::StreamingArray	; _ZN17CVehicleRecording21RegisterRecordingFileEPKc
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96EA6C + 2), &CVehicleRecording__StreamingArray.gta_iv->pData);	// mov     _ZN17CVehicleRecording14StreamingArrayE.pData[eax], 0	; _ZN17CVehicleRecording21RegisterRecordingFileEPKc
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96EAA1 + 1), CVehicleRecording__StreamingArray.gta_iv);	// mov     ecx, offset _ZN17CVehicleRecording14StreamingArrayE; CVehicleRecording::StreamingArray	; sub_96EA90
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96EAF2 + 2), &CVehicleRecording__StreamingArray.gta_iv->pData);	// cmp     _ZN17CVehicleRecording14StreamingArrayE.pData[eax], ecx	; sub_96EAE0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96EB09 + 1), &CVehicleRecording__StreamingArray.gta_iv->pData);	// mov     esi, offset _ZN17CVehicleRecording14StreamingArrayE.pData	; sub_96EB00
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96EB35 + 2), &CVehicleRecording__StreamingArray.gta_iv[GetFileIDcurrentLimit(FILE_TYPE_RRR)].pData);	// cmp     esi, offset unk_12D7AF4	; sub_96EB00
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96EB58 + 2), &CVehicleRecording__StreamingArray.gta_iv->pData);	// mov     eax, _ZN17CVehicleRecording14StreamingArrayE.pData[esi]	; sub_96EB50
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96EB5E + 2), &CVehicleRecording__StreamingArray.gta_iv->pData);	// lea     esi, _ZN17CVehicleRecording14StreamingArrayE.pData[esi]	; sub_96EB50
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96EB87 + 2), &CVehicleRecording__StreamingArray.gta_iv->field_8);	// mov     ecx, _ZN17CVehicleRecording14StreamingArrayE.field_8[edx]	; sub_96EB80
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96EB8D + 2), &CVehicleRecording__StreamingArray.gta_iv->pData);	// mov     eax, _ZN17CVehicleRecording14StreamingArrayE.pData[edx]	; sub_96EB80
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96EBED + 2), &CVehicleRecording__StreamingArray.gta_iv->field_8);	// mov     edi, _ZN17CVehicleRecording14StreamingArrayE.field_8[edx]	; sub_96EB80
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96ECD4 + 1), &CVehicleRecording__StreamingArray.gta_iv->usageCount);	// mov     esi, offset _ZN17CVehicleRecording14StreamingArrayE.usageCount	; sub_96ECD0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96ED00 + 2), &CVehicleRecording__StreamingArray.gta_iv[GetFileIDcurrentLimit(FILE_TYPE_RRR)].usageCount);	// cmp     esi, offset unk_12D7AFC	; sub_96ECD0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96F924 + 2), &CVehicleRecording__StreamingArray.gta_iv->pData);	// mov     eax, _ZN17CVehicleRecording14StreamingArrayE.pData[esi]	; sub_96F910
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96F941 + 2), &CVehicleRecording__StreamingArray.gta_iv->field_8);	// cmp     _ZN17CVehicleRecording14StreamingArrayE.field_8[esi], 0	; sub_96F910
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96F984 + 2), &CVehicleRecording__StreamingArray.gta_iv->field_8);	// cmp     eax, _ZN17CVehicleRecording14StreamingArrayE.field_8[esi]	; sub_96F910
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96F9AE + 2), &CVehicleRecording__StreamingArray.gta_iv->field_8);	// mov     ecx, _ZN17CVehicleRecording14StreamingArrayE.field_8[eax]	; sub_96F9A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96F9B4 + 2), &CVehicleRecording__StreamingArray.gta_iv->pData);	// mov     edx, _ZN17CVehicleRecording14StreamingArrayE.pData[eax]	; sub_96F9A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96FA08 + 2), &CVehicleRecording__StreamingArray.gta_iv->pData);	// mov     _ZN17CVehicleRecording14StreamingArrayE.pData[esi], eax	; sub_96F9D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96FA13 + 2), &CVehicleRecording__StreamingArray.gta_iv->pData);	// mov     ecx, _ZN17CVehicleRecording14StreamingArrayE.pData[esi]	; sub_96F9D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96FA24 + 2), &CVehicleRecording__StreamingArray.gta_iv->field_8);	// mov     _ZN17CVehicleRecording14StreamingArrayE.field_8[esi], eax	; sub_96F9D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96FA31 + 2), &CVehicleRecording__StreamingArray.gta_iv->field_8);	// cmp     _ZN17CVehicleRecording14StreamingArrayE.field_8[esi], eax	; sub_96F9D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96FA39 + 2), &CVehicleRecording__StreamingArray.gta_iv->pData);	// mov     ecx, _ZN17CVehicleRecording14StreamingArrayE.pData[esi]	; sub_96F9D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96FA52 + 2), &CVehicleRecording__StreamingArray.gta_iv->field_8);	// mov     _ZN17CVehicleRecording14StreamingArrayE.field_8[esi], eax	; sub_96F9D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96FA5B + 2), &CVehicleRecording__StreamingArray.gta_iv->field_8);	// cmp     eax, _ZN17CVehicleRecording14StreamingArrayE.field_8[esi]	; sub_96F9D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96FB01 + 2), &CVehicleRecording__StreamingArray.gta_iv->usageCount);	// add     _ZN17CVehicleRecording14StreamingArrayE.usageCount[eax], 1	; sub_96FA80
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96FB08 + 2), &CVehicleRecording__StreamingArray.gta_iv->field_8);	// mov     ecx, _ZN17CVehicleRecording14StreamingArrayE.field_8[eax]	; sub_96FA80
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96FB0E + 2), &CVehicleRecording__StreamingArray.gta_iv->pData);	// mov     edx, _ZN17CVehicleRecording14StreamingArrayE.pData[eax]	; sub_96FA80

		}
		#endif

		// Patch count
		#if TRUE
		{
			CPatch::PatchUINT32(g_mCalc.GetCurrentVAbyPreferedVA(0x970583 + 1), GetFileIDcurrentLimit(FILE_TYPE_RRR));	// push    2072
		}
		#endif
	}
	#endif
}

// Patches ID limits if neccessary.
void FileIDlimit::PatchFileIDlimits_GTA_EFLC_1_1_2_0()
{
	// Patch limit of models
	#if TRUE
	{
		// Patch pointers to CModelInfo__ms_modelInfoPtrs.gta_iv
		#if TRUE
		{
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x495132 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_495120
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x4968C6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_496890
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x496948 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4968F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x497206 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_4B7380
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x4999A4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4999A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x499DD8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_499D80
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x49A5E0 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_49A590
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x49A655 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_49A590
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x49ABAE + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_49AB30
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x49AC23 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_49AB30
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x49C30C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; 
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x49C342 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs	; 
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x49E3E6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_49A820
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x49E40E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_49E400
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x49E4BC + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_49E480
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x49E508 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_49E480
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x4A1B0A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_4A19A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x4A2544 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4A19A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x4A26F4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4A19A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x4AA6BC + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_4AA6A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x4B293B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_4B1A90
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x4B2942 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_4B1A90
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x4B4836 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_4B4830
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x4B7045 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_4B6F40
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x4C3FAE + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_4C3CA0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x510CE5 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_510C50
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x5141B2 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_5141A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x5142D4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_5141A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x55DC10 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_55DBA0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x55DEBF + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_55DEB0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x55F96A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_55F950
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x55FB28 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_55F950
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x560C71 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_560C60
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x564E06 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_564C70
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x564F61 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_564C70
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x56504F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_564C70
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x56530D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_564C70
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x565619 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_564C70
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x565885 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_564C70
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x565BF4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_564C70
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x5678F4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_567710
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x5684EE + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_568480
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x56852C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_568480
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x5693E1 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_5693D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x56941E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_5693D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x569445 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_5693D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x56C7EE + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_56C6F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x56FB25 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_56FAE0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x57466D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_5745B0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x579A2D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_579710
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x579DFB + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_579710
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x579E1B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_579710
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x579EF1 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_579710
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x579FB6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_579710
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x57A10A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_579710
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x57A62B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_57A4E0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x57A64B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_57A4E0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x57A702 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_57A4E0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x57C7E9 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_57C5F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x7D8448 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_7D8440
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x7D85EF + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_7D8550
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x7D8D99 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_7D8D90
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x7D8E77 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_7D8E60
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x7D8FC9 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_7D8FB0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x7D9252 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_7D91E0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x7D9328 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_7D9320
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x7DA2E8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_7DA2E0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x7DA423 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_7DA420
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x7DA483 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_7DA480
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x7DA851 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_7DA840
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x7DC402 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_7DC3F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x7DD272 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_7DD260
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x7DDA1D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_7DD940
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x7DDB38 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs	; sub_7DDAA0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x7DDC1C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_7DDBD0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x7DDD43 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_7DDCF0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x7E4B59 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_7E4AD0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x7F30A2 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebp, _ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs	; sub_7F2CA0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8253A6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8251F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x82DDD8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_82DDC0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x82DE3B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_82DDC0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x82EBAF + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_82EB80
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x834268 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_834100
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x836B4E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_836AD0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x83B0A3 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_83ADC0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x83B248 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_83ADC0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x83B26A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_83ADC0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x83B568 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_83ADC0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x83F512 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_83F500
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x83F534 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_83F530
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x83F8D8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_83F820
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x83FE88 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_83FE80
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x840259 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_840190
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x842044 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_841C50
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8420D7 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_841C50
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x842144 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_841C50
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x842216 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_841C50
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8426D9 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_842370
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x842738 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_842370
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x84287B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_842370
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x842D26 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_842970
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x843371 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8432D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x843B76 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_843AF0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x84423A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_843CD0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x84475B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_843CD0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x84489A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_843CD0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8449D0 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_843CD0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x844B9A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_843CD0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x845BB5 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_845B90
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x845DDF + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_845B90
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x848B9D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_848A20
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x84CC0B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_84CBC0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x84CCBB + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_84CC70
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x84CD6B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_84CD20
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x86B872 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_86B760
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x86B952 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_86B760
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x86C2CF + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_86C290
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x86F2FC + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_86F0C0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x879403 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_879040
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x87A31F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_87AA10
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x88D396 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_88D320
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x89ADE5 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_89AD10
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x89D1D0 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_89D180
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x89D81C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_89D810
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x89D91C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_89D810
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x89DB4A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_89DAA0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x89DF97 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_89DDB0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x89E51E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_89E4B0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x89F2C2 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_89F200
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x89F4C6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_89F3A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x89F585 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_89F3A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8A50B4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebp, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8A4FF0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8A5291 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8A5280
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8A5460 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8A5400
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8A5810 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebp, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8A57C0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8A589F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8A57C0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8A5985 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8A5930
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8A5AED + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8A5930
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8B82DA + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8B8270
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8B83BD + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8B8270
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8B8A35 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8B8A00
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8B8B8E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8B8B60
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8B9B33 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebp, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8B99A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8BA0F6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8BA050
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8BC265 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_8BC240
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8BC54B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_8BC4C0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8BC5DA + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_8BC4C0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8BC7AC + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_8BC770
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8BC865 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_8BC770
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8BC931 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8BC770
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8BC9B2 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8BC770
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8BCB0B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8BCB00
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8BCB9E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_8BCB40
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8BCCA3 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8BCC70
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8BCCD2 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs	; sub_8BCC70
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8BCD53 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8BCD20
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8BCD82 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs	; sub_8BCD20
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8BCF08 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8BCEE0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8BCFD6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8BCFC0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8BD05A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8BD030
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8BD0CB + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_8BD030
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8BD176 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8BD160
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8BD1D6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8BD1C0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8BD2B6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8BD2A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8BD326 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8BD310
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8BD3AD + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_8BD380
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8BD43D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_8BD410
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8BD4B6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8BD4A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8BD52F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8BD510
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8BD5D8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8BD5C0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8BDC76 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8BDAE0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8BDCA0 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8BDAE0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8BDCB4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8BDAE0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8BDCCC + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8BDAE0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8BDCDF + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8BDAE0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8BDD22 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8BDAE0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8BDD49 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8BDAE0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8BDD70 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8BDAE0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8BDEF2 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_8BDDA0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8BDF53 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8BDF40
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8BE100 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8BE070
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8BE22A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_8BE1A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8BE340 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8BE2C0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8BE68B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_8BE3B0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8BE6AF + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8BE3B0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8BE6D4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8BE6D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8BE758 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8BE700
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8BE7C0 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8BE7A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8BE87F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_8BE800
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8BEC78 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_8BEC70
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8BECD6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_8BEC70
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8BED6F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_8BED40
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8BEE45 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_8BEE10
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8BEEF9 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8BEEB0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8BEFC8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8BEF90
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8BF038 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8BEF90
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8BF0E3 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8BEF90
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8BF163 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8BEF90
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8BF38A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_8BF340
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8BF3EE + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_8BF340
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8BFD11 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_8BFCA0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8BFE95 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_8BFCA0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8BFFB2 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_8BFF10
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8C01FD + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs	; sub_8C0100
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8C0248 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_8C0240
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8C0267 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_8C0240
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8C0299 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_8C0290
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8C02AD + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_8C0290
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8C05B8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_8C05B0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8C05D7 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_8C05B0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8C0663 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_8C0600
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8C0720 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_8C0600
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8C0BC8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8C0A10
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8C0BCF + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8C0A10
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8C0E28 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8C0A10
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8C0E2F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8C0A10
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8C0F2F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs	; sub_8C0A10
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8C0FB8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8C0A10
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8C1063 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_8C0A10
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8C10E0 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_8C0A10
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8C1475 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_8C1460
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8C2028 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8C2020
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8C2B12 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8C29F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8C2C54 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8C2C40
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8C3F5B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8C3EB0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8C45C1 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8C4570
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8C586B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8C5720
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8C5E93 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8C5DB0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8C6AE8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8C6AE0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8C6B31 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8C6B20
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8C6DB9 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8C6D90
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8C704C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8C7010
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8C752B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebp, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8C7520
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8C78BC + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8C7520
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8CB2C0 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8CB2A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8CB330 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8CB310
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8CF17C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8CEEE0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8D1D08 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8D1240
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8D1DB4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8D1240
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8D2B34 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8D2060
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8D45F7 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8D45E0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8D4637 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8D4620
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8D47F3 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8D47C0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8D4935 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_8D48D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8D4A79 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8D4A40
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8D9DFA + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8D9D70
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8DFCA0 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8DF580
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8E446A + 1), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, offset _ZN10CModelInfo16ms_modelInfoPtrsE; CModelInfo::ms_modelInfoPtrs	; sub_8E4460
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8E4516 + 2), CModelInfo__ms_modelInfoPtrs.gta_iv + GetFileIDcurrentLimit(FILE_TYPE_MODEL));	// cmp     esi, offset unk_159B770	; sub_8E4460
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8E5A46 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8E57C0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8E66BB + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_8E66B0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8E6788 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8E6760
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8E69FB + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8E6930
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8E7DA3 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8E7D20
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8E7E0B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8E7E00
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8E84B8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8E84B0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8E8520 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8E84F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8E85D4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8E85D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8E880E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A94420
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8E9193 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8E9130
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8E9F0F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8E9E20
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8EA5B2 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8EA530
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8ED1F1 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8ED100
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8ED881 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8ED850
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8EDB5A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8EDB30
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8EE098 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_8EE020
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8EE15F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8EE020
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8EE21D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8EE020
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8EE272 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8EE020
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8EE2BF + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8EE020
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8EE352 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8EE020
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8EE3DC + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8EE020
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8EE42D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8EE020
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8EE4AC + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8EE020
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8EE518 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8EE020
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8EE7A2 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8EE730
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8EE904 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8EE730
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8EF0B7 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8EF060
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8EF0D8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8EF060
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8EFA6E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8EF830
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8EFA96 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8EF830
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F0794 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8F0780
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F0EF4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8F0E40
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F0F59 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8F0E40
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F0FC1 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8F0E40
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F13F9 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8F11D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F1462 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8F11D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F16D9 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8F14E0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F24CA + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8F2340
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F36C2 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8F3590
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F4A85 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8F4A80
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F4B1D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8F4A80
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F52C2 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_8F5200
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8F6120 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8F5F10
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8FD296 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8FD270
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8FD867 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_8FD7B0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x903E3E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_903E20
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x903E82 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_903E20
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x903ECC + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_903E20
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x903F13 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_903E20
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x904716 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9046B0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x90478D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_904780
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9047D8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_904780
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x904814 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_904780
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x904859 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_904780
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x904895 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_904780
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x905D81 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_905D70
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x906062 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_906050
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9065D0 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9063E0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x906A5E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_906A40
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9086E8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_9085F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x908726 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9085F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9088D4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9085F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x908B0E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9085F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x90AD47 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_912A10
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x90AE4E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_90AE40
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x90AE68 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_90AE40
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x90AE87 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_90AE40
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x90AEA2 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_90AE40
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x90AEE6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_90AED0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x90AF08 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_90AED0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x90AF29 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_90AED0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x90B03D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_90AED0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x90B0B5 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_90B0A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x90B0D7 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_90B0A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x90B0F8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_90B0A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x90B119 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_90B0A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x90B175 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_90B0A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x90B234 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_90B0A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x90B2E0 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_90B2A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x90B324 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_90B2A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x90B43A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_90B430
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x90B4E3 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_90B490
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x90B735 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_90B6A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x90B74C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_90B6A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x90B772 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_90B6A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x90B789 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_90B6A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x90BB0A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_90B6A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x90BB21 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_90B6A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x90C113 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_90C100
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x90C134 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_90C100
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x90C349 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_90C330
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x90C368 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_90C330
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x90CBE7 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_90CBD0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x90CBFF + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_90CBD0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x90CC1A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_90CBD0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x90CC35 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_90CBD0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x90CC54 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_90CBD0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x90E0C8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_90DF10
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x90E2A5 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_90DF10
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x90E323 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_90DF10
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x90E350 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_90DF10
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x90E438 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_90DF10
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x90E493 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_90DF10
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x90E4D1 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_90DF10
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x90E4FE + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_90DF10
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x90E58A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_90DF10
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x90E5E5 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_90DF10
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x90E6BA + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_90DF10
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x90E77B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_90DF10
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x90E81D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_90DF10
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x90EEC2 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_90EEA0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x90EEDD + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_90EEA0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x90EEF8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_90EEA0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x90EF11 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_90EEA0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x90EF28 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_90EEA0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x90EF43 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_90EEA0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x90F6A8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_90F5F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x90F728 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_90F5F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x90F8AE + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_90F5F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x90F9D3 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_90F5F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x90FA70 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_90F5F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x90FB0E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_90FB00
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x90FB88 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_90FB00
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x90FBFD + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_90FB00
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x90FD3A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_90FC90
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x90FD9C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_90FC90
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x90FE9F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_90FC90
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9106C8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9106B0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9109A3 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9106B0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x911437 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_911420
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x91145C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_911420
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9115AB + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_911420
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9115D4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_911420
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9115F8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_911420
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x91161C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_911420
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x911B9D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_911420
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x911BCF + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_911420
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x911C04 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_911420
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x911C85 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_911420
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x911DD1 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_911420
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x911E1F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_911420
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x911E38 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_911420
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x911E58 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_911420
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x911E71 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_911420
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x911EF1 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_911420
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x911F3F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_911420
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x911F56 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_911420
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x911F78 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_911420
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x911F8F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_911420
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9121B9 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_912150
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x912285 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_912260
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x912A9E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_912A50
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x912AD2 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_912AC0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x913247 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_913230
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9132EB + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_913230
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x913411 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_913230
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x913501 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_913230
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x913577 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_913230
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9135D2 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_913230
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x91362D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_913230
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x913644 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_913230
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x913666 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_913230
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x91367D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_913230
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x91369F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_913230
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9136B8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_913230
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9136D8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_913230
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9136EF + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_913230
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x913711 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_913230
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x913731 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_913230
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x913954 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_913230
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9139BD + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_913230
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x913C19 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_913A30
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9142D9 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_913A30
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9142FB + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_913A30
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x91431B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_913A30
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9147E5 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_913A30
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x914805 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_913A30
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x914829 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_913A30
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x914E61 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_914DD0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x914E7F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_914DD0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x914EB4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_914DD0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x914ECB + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_914DD0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x915047 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_914DD0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x915584 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_914DD0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x91559B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_914DD0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9155B6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_914DD0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9155CD + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_914DD0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x915631 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_914DD0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x915648 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_914DD0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x915663 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_914DD0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x91567A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_914DD0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x915936 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_915750
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x917953 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_917910
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x91ADC7 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_91ADA0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x91B1E5 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_91B160
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x91BC1A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_91B8E0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x91C776 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_91C5C0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x91E5FD + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_91E5E0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x91E730 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_91E5E0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x91E766 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_91E5E0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x91F01B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_91EFF0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x91FC50 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_91F8A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x91FD4F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_91F8A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x91FE9A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_91FE60
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x91FEF6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_91FE60
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x91FF1F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_91FE60
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x91FF49 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_91FE60
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x91FF73 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_91FE60
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9222F0 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_922220
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x92233F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_922220
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9223C9 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_922220
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x922411 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_922220
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9227D8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9227B0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9228FA + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9227B0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x92299D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9227B0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9231FA + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_923130
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9234E9 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_923130
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x92357A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_923130
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x923784 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_923770
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x923844 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_923770
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x923B4A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_923A50
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x923D2A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_923A50
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x923F20 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_923A50
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x924134 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_924040
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x92BABE + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_92B9C0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x92BFCA + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_92BF40
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x931D61 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_931990
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x931DE7 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_931990
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x931E06 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_931990
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x931E75 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_931990
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x93210F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_931ED0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x932220 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_931ED0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x932A57 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9329F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9334EF + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9334C0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x93353C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9334C0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x934255 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_933890
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x935B1B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_935AE0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x935BF1 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_935BE0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x936236 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9361A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9374A8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_937430
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x93783D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_937430
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x93997A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9398E0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x93BB76 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_93BB30
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x93E554 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// cmp     _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4], 0; CModelInfo::ms_modelInfoPtrs	; sub_93E540
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x93E8CF + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_93E890
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x940195 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_940180
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x942FFD + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_942F50
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x943C29 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_943C20
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x94419D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_944190
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x944637 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_944620
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x944809 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_944800
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9448A8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9448A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x944910 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_944900
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x944A7A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_944A70
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x944B4D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_944AE0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9451D7 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9451C0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x945589 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_945580
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9458E1 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9458A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x94F0BD + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_94EEB0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x94F31A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_94EEB0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x94F509 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_94EEB0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x94F5D8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_94F570
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x94F63D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_94F570
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x94F712 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_94F570
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9542D4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_954230
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9574C3 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9572F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x95980A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9596F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x959824 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebp, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9596F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x95B05B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_95AFF0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x95E59A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_95E040
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x962573 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_962450
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x962CE0 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_962CC0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x962E6C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_962E60
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96457C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_964430
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9645C8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_964430
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9683CD + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9681A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x968BC2 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_968B20
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96E368 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_96E340
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96E94F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; 
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96E97A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; 
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96E9A8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; 
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96E9D6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; 
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96EA04 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; 
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96EA32 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; 
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96EA60 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; 
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96EA8E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; 
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96EABC + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; 
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96EAEA + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; 
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x96EE92 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_96EE30
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x970005 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_970000
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x970511 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9704D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x97182E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_971810
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x971846 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_971810
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x971A48 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_971810
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x971A70 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_971810
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9750C6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_974E70
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9750F0 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_974E70
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x978830 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_978770
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x97C1D3 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_97BE20
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x97C5EE + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_97BE20
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x97C6A6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_97BE20
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x97C77A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_97BE20
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x993574 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_993480
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x994DB4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_994DB0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x99523E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_995220
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x99527F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_995220
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9952C6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_995220
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x99530D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_995220
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x995BCA + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_995B70
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x995C91 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_995C80
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x995E0E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_995DF0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x99A8E4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_99A8E0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x99AE11 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_99AE00
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x99B2E8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_99B2E0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x99B3D5 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_99B3D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x99B47D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_99B440
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x99B703 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_99B6A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x99C7C6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_99C690
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x99D06E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_99D050
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x99EB1A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_99E950
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x99F5E6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_99F5C0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x99F9A3 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_99F910
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9A1068 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9A0E80
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9A1162 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9A0E80
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9A1394 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9A0E80
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9A1FA7 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9A1E90
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9B2DF6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9B2D80
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9BA4D5 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; 
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9BA505 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; 
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9BA534 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; 
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9BA544 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; 
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9BA554 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; 
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9BA567 + 1), CModelInfo__ms_modelInfoPtrs.gta_iv);	// push    offset _ZN10CModelInfo16ms_modelInfoPtrsE; CModelInfo::ms_modelInfoPtrs	; sub_9BA560
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9BA594 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; 
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9BA644 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; 
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9BA65B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; 
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9BB1FB + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; 
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9BB784 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_9BB780
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9BB7F6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; 
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9BB866 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9BB860
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9BB8C2 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9BB8B0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9BB9D6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; 
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9BBA52 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs	; 
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9BC975 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; 
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9BCB63 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; _ZN10CModelInfo12GetModelInfoEPKcPi
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9BCBB9 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9BCB70
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9BCC66 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_9BCC60
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9BCD2A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4], esi; CModelInfo::ms_modelInfoPtrs	; sub_9BCCF0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9BCD99 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4], esi; CModelInfo::ms_modelInfoPtrs	; sub_9BCD60
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9BCE09 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4], esi; CModelInfo::ms_modelInfoPtrs	; sub_9BCDD0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9BCE79 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4], esi; CModelInfo::ms_modelInfoPtrs	; sub_9BCE40
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9BCEEC + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4], esi; CModelInfo::ms_modelInfoPtrs	; sub_9BCEB0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9BCF5C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4], esi; CModelInfo::ms_modelInfoPtrs	; sub_9BCF20
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9BCFC9 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4], esi; CModelInfo::ms_modelInfoPtrs	; sub_9BCF90
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9C211E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9C2090
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9C2159 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9C2090
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9C21D0 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9C2090
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9C2208 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9C2090
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9C3420 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_9C3390
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9CE514 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9CE4C0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9CF799 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9CF790
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9CF7D9 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9CF7D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9D00DA + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9D00D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9D0682 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9D0660
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9D0964 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9D0940
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9D1C4D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9D1C20
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9D1E6F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9D1E50
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9D239F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9D2360
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9D2E26 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9D2D60
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9D34C8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9D34C0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9D3611 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9D3600
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9D8B0A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9D8AF0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9D8D00 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9D8CE0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9D8D44 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9D8D40
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9D8D54 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9D8D50
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9D8D64 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9D8D60
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9D8D75 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9D8D70
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9D9040 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9D9030
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9D9054 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9D9050
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9D9092 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9D9080
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9D9225 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9D9200
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9D92AE + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9D9250
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9D9404 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9D9400
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9D941B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9D9410
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9D9501 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9D94F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9D9581 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9D9570
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9D960F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9D95F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9D9D5A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9D9D10
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9D9E8D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9D9E80
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9D9F6C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9D9F50
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9DA138 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9DA120
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9DA4ED + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9DA4E0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9DA5FD + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9DA5F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9DA700 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9DA6F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9DA7EF + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9DA6F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9DA962 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9DA950
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9DA9B1 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9DA9A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9DAA55 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9DAA40
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9DAB01 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9DAA40
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9DABBF + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9DABA0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9DAC66 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9DABA0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9DB01F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9DB000
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9DB848 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9DB7F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9DBE0B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebp, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9DBE00
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9DC006 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9DBE00
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9DC376 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9DC070
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9DC7EF + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9DC7A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9DCDF0 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9DCDD0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9DCF7F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9DCF60
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9DD625 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_9DD5D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9DD8F4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9DD8C0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9DDBA0 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9DDB80
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9DFF50 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9DFEC0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9E6EBA + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9E6EB0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9E6EFE + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9E6EF0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9E82BB + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9E81B0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9E851C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9E8510
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9E8531 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9E8510
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9E9B24 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9E9910
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9EA8AC + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9EA850
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9ED3C3 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9ED390
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9EF2D2 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9EF2C0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9EFAE5 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9EFAB0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9EFC64 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9EFC60
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9F0020 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9F0010
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9F01CB + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9F0190
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9F01F4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9F0190
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9F0257 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_9F0240
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9F028C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_9F0240
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9F0306 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9F0300
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9F078B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9F0780
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9F07A4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9F07A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9F07B4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9F07B0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9F201B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9F1FF0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9F209C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9F1FF0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9F213A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9F1FF0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9F21E3 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9F1FF0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9F230D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9F22C0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9F2590 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9F22C0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9F2646 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9F2620
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9F26C7 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9F2620
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9F2A64 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9F2990
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9F2AC4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9F2990
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9F2B15 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9F2990
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9F4886 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9F4800
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9F5398 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9F4A50
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9F6206 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9F61C0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9F67D7 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9F67D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9F6847 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9F6840
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9F68F5 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9F68B0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9FAE4B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9FAC90
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9FC95F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9FC930
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9FCDF0 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_9FC930
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9FCE54 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_9FC930
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9FCEB7 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_9FC930
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9FCF1B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_9FC930
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9FCF7E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_9FC930
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9FD044 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9FC930
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9FDB39 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9FDB20
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9FDBE9 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9FDBB0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9FDD1A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9FDD10
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9FDE94 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9FDE90
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9FE129 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9FE120
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9FE20B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs	; sub_9FE120
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9FE299 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs	; sub_9FE120
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9FE427 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9FE310
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9FE458 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9FE310
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9FE527 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebp, _ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9FE510
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9FE909 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9FE8F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9FE944 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9FE940
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9FEA74 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9FE9F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9FEB22 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9FEB00
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9FEEE4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_9FEB00
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9FF148 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9FEB00
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9FF188 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9FEB00
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9FFCF2 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9FFCD0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x9FFE3C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_9FFE30
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA08F20 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A088B0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA08F9F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A088B0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA09995 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A088B0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA09B24 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A088B0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA0B975 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A0B8F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA0BA26 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A0BA00
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA0BAA7 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A0BA00
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA0BAD7 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A0BA00
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA0C309 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A0BA00
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA0CED5 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A0CEA0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA0CF7E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A0CEA0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA10020 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A0FC50
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA177C6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_A177C0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA201D4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A200D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA2334B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A232E0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA23AF2 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A23840
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA2C2F8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; 
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA3CA06 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A3C980
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA3D386 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A3D310
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA3DBEF + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A3D6B0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA3E9B7 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A3E600
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA409D7 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A3F7D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA409F7 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A3F7D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA4116E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A410D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA413F2 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A410D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA414C3 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A410D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA4182C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A417B0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA42807 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A424B0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA428FF + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A424B0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA437BE + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A42B20
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA43806 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A42B20
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA43898 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A42B20
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA45292 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A45280
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA45A0F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A459E0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA47EBB + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A47CA0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA4917F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A49140
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA4A441 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A4A2B0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA4A7F1 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A4A4F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA4A9A4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A4A4F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA4CDF9 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A4CCD0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA4CFE7 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A4CCD0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA6453E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A64480
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA64B96 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebp, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A64B00
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA64DF7 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_A64DA0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA64FD4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A64FD0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA64FF4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A64FF0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA65024 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A65020
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA666B8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A66650
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA669F9 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A66960
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA6706C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A66DE0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA671A2 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_A670C0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA67D7A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A67CF0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA6834B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A68330
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA683EF + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_A68330
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA683FC + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_A68330
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA68CD8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A68AC0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA68FA3 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A68DA0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA69443 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A69400
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA697A3 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A69760
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA69F7F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A69F10
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA6C36B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_A6C310
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA6C40D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_A6C310
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA6DE55 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A6DBC0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA6DE5C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A6DBC0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA6E026 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A6DBC0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA6E02D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A6DBC0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA6E0E7 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A6DBC0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA6E0F2 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A6DBC0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA6E396 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A6DBC0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA6E39D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A6DBC0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA738C4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A738C0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA73944 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A73940
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA73974 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A73970
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA739A4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A739A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA739D4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A739D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA73A04 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A73A00
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA73A34 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A73A30
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA73A54 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A73A50
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA73A84 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A73A80
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA73AB4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A73AB0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA73AD4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A73AD0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA73B24 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A73B20
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA73B64 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A73B60
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA73BA4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A73BA0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA73BE4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A73BE0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA73C24 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A73C20
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA73C64 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A73C60
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA73CA4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A73CA0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA73CC4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A73CC0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA74309 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A74300
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA74563 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A74530
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA7468D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A74680
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA75144 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A75140
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA75E31 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A75E10
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA767D7 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A765A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA76A1F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A765A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA76B14 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A765A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA76BF1 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A76BE0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA76C5B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A76BE0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA77261 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A77250
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA77623 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A775E0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA77AD5 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A77A10
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA77FC7 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A77FC0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA783CC + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A783A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA7CEE4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A7CEE0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA7D2CC + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A7D2B0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA7D426 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A7D330
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA7F823 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A7F810
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA82907 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A825B0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA8292B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A825B0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA82B3B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A82AA0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA8324C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A83040
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA8B019 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A8A5A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA8BF77 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A8BDD0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA90102 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A8F620
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA902D7 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A8F620
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA935B8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A93510
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA93C58 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A937B0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA953D8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A95390
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA9A246 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A9A230
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA9A2B1 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A9A230
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA9A2D9 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_A9A230
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA9A35F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A9A230
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA9A604 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A9A5B0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xA9AC2D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_A9ABE0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAA1A19 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_AA19E0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAA1A5A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_AA1A30
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAA1B3D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_AA1B00
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAA1C0D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_AA1B00
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAA737B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_AA7320
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAA7D9D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_AA7D70
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAA82E4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_AA8280
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAAC7B4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_AAC7A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAAC83A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_AAC7A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAAC914 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_AAC900
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAAC948 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_AAC900
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAACA44 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_AAC9E0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAACAD4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_AACAA0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAACBB9 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_AACB20
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAACC5E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_AACC30
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAAE72C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_AAE6F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAAEB2E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_AAEB00
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAAEB9C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_AAEB60
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAAF485 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_AAF3D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAAF590 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_AAF3D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAB05F6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_AB05B0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAB0A0D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_AB0760
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAB149C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_AB1460
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAB318E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_AB30C0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAB5F94 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_AB5E20
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAB619A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_AB6040
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAC0A35 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_AC0830
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAC0C1C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_AC0830
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xACC12C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_ACBCE0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xACC1C6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_ACBCE0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xACC71D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_ACBCE0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xACF1B1 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_ACE570
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAD9DAC + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_AD9380
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xADFD70 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_ADF7D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAE1A87 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_AE1A00
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAE23AC + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_AE2340
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAE5906 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_AE5430
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAE69EC + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_AE6820
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAEC285 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_AEC1F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAEC513 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_AEC4D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAEC6EE + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_AEC6A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAED384 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_AED1F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAED715 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_AED6B0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAED75C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_AED6B0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAED794 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_AED6B0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAEDC47 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_AEDBF0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAEDE89 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_AEDE10
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAEE29D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_AEE270
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAEE5B4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_AEE550
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAEE6E7 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_AEE6B0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAEE8F3 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_AEE840
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAEE92E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_AEE840
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAEE964 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_AEE840
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAEE9DE + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_AEE840
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAEEAF2 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_AEE840
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAEEC7D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_AEEBB0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAEF16E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_AEF0A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAEF191 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_AEF0A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAEF1C3 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_AEF0A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAEF36C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_AEF300
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAEF69A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_AEF5E0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAEF6C5 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_AEF5E0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAEF6F1 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_AEF5E0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAEF71E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_AEF5E0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAEF840 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_AEF800
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAF01D2 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_AF0140
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAF02F9 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_AF02C0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAF0379 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_AF02C0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAF05BB + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_AF04A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAF09AB + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_AF07C0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAF0C6E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_AF0B70
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAF0CE0 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_AF0B70
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAF101A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_AF0DA0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAF1764 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_AF1640
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAF3AF2 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs	; sub_AF3990
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAF3DCE + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_AF3B20
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAF4380 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_AF42F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAF4403 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_AF42F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAF445C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_AF42F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAF4581 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_AF4570
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAF4ADD + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_AF4AD0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAF55BD + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_AF5530
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAF57B6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_AF5530
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAF59ED + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_AF5530
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAF5AA7 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_AF5530
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAFAE22 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_AFAE10
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAFB46E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_AFB320
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAFB4D1 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_AFB4C0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAFB543 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_AFB530
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAFB950 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_AFB930
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAFC2D9 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; 
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAFFCEC + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_AFFCD0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xAFFD24 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_AFFCD0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB00B12 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_B00770
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB01630 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_B01620
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB01BED + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_B01BB0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB0D5AC + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_B0D2C0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB0D620 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_B0D2C0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB12EA0 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_B12E20
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB138F0 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_B138C0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB13EA8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_B13EA0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB1828A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; GetVehicleGXTlabel
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB1A694 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; GetAudioSettingsForVehicle
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB1A78B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; GetAudioSettingsForVehicle
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB1C978 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_B1C8F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB1D58A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_B1D510
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB1E476 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_B1E460
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB1E962 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_B1E8C0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB2A6E9 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_B2A6C0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB2A75F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_B2A6C0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB2BD2D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_B2BCE0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB30295 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_B2FD90
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB304BC + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_B30490
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB306EB + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_B306E0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB3076B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_B30760
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB31BA3 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_B31B50
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB31C50 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_B31B50
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB33CE8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_B33CE0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB33D08 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_B33D00
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB33D28 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_B33D20
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB33D48 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_B33D40
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB33DDF + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_B33D60
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB340FF + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_B34080
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB3441F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_B343A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB346BF + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_B34640
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB34904 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_B34880
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB359BE + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_B35940
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB3624E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_B361E0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB37CD9 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_B37C40
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB38256 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_B381E0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB3BCE9 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs	; sub_B3BC30
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB3BF35 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_B3BF20
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB3C1FC + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_B3C1D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB3CC2D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_B3CC20
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB3CDBE + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_B3CDB0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB3D2CD + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_B3D2B0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB3D2EE + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_B3D2B0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB3D4D3 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_B3D3D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB3E5F4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_B3E360
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB3EE34 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_B3E6D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB3F42A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_B3F420
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB3FC11 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_B3F9D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB4122E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_B410B0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB4165A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_B41630
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB4275C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_B42740
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB427C4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_B42740
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB4288C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_B42880
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB428C0 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_B42880
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB42CBE + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_B42CB0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB42D3B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_B42D20
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB42E1C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_B42E10
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB4485F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_B44690
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB44891 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_B44690
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB44973 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_B44690
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB4806F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_B48060
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB4C64C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_B4C3D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB4FD84 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_B4FD80
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB6557B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_B65550
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB656A1 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_B65670
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB695E6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_B694C0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB6BE06 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; 
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB6EA79 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_B6EA10
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB6EB4A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_B6EB10
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB721B3 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_B71FB0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB74AB9 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_B743E0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB810F2 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_B810E0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB81798 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_B81780
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB817FC + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_B817C0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB8286B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_B82820
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB82B61 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_B82B40
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB83047 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_B83030
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB83580 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_B83030
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB8539C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_B85370
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB87BEA + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_B87AE0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB8ABD1 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_B8AB20
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB8C534 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_B8C4E0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB8C6F6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_B8C6E0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB8C7D5 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_B8C7A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB8C891 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_B8C850
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB8E6A8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_B8E600
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB8E7E6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_B8E600
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB8E815 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_B8E600
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB97174 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_B97130
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB9C284 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_B9C270
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB9CCA5 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_B9CC90
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB9CF5B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_B9CE60
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB9CF78 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_B9CE60
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB9CFB8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_B9CE60
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB9CFD5 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_B9CE60
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB9D00B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_B9CE60
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB9D37E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_B9D270
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB9D4D7 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_B9D420
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB9DC6C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_B9DC20
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB9E27A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_B9E210
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB9E3AD + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_B9E210
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB9F4D8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_B9F350
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xB9FF0C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_B9FE90
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBA13CB + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BA1270
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBA155B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BA1400
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBA2535 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BA24B0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBA2608 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BA25C0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBA28DC + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BA25C0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBAA353 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_BAA280
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBAE17D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BAE130
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBB3BE9 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BB3BD0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBDBB65 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_BDBB60
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBDD2F0 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BDD2A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBDD307 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BDD2A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBDF5CD + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BDF3C0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBDFA3B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BDF830
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBE1964 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BE1950
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBE1B60 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BE1AF0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBE1DB9 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BE1AF0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBE203D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BE1FC0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBE22A3 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_BE2240
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBE22BA + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BE2240
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBE23B3 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BE2330
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBE27E5 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BE27B0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBE50B1 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_BE50A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBE7077 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_BE7000
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBE70A5 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; 
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBE70D9 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; 
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBE70F5 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; 
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBE715B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BE7140
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBE9792 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BE9770
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBE9F6E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BE9EB0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBEB454 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BEB440
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBEB48A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BEB470
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBEC7B4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BEC7A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBECA3F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BEC9F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xBFB437 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_BFB360
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC01045 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_C01040
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC040E5 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; 
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC0411A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; 
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC0493B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_C047A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC049C2 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_C047A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC0A1D4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_C0A1C0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC0A8D6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_C0A8A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC0B9E4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_C0B9D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC0BA44 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_C0BA30
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC309E3 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_C309D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC31A4E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_C31940
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC31BEB + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_C31940
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC31CD0 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_C31940
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC31E68 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_C31E60
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC3231C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_C322B0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC329D4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_C32820
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC331A0 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_C32EE0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC3A757 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_C39D60
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC3C1AA + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_C3C000
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC3D732 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_C3D6F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC3EFEB + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_C3ECA0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC3F01C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_C3ECA0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC42D41 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_C42C20
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC42E96 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebp, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_C42E70
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC42EB7 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_C42E70
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC42F4D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_C42E70
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC42FFD + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_C42E70
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC4309E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_C42E70
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC437A4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_C437A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC43AFE + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_C43AD0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC43FF3 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_C43FB0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC44047 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_C43FB0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC440BC + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_C43FB0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC44106 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_C43FB0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC4414E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_C43FB0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC4419B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_C43FB0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC46374 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_C46360
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC46AE8 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_C46AB0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC47EFF + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_C47EF0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC49A61 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_C49320
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC508DF + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_C4FFB0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC52138 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_C52130
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC531D9 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_C52FF0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC5333A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_C52FF0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC53356 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_C52FF0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC617D9 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_C617C0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC6183A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_C61830
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC647E5 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_C647A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC64C1C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_C649B0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC66414 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_C66410
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC66434 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_C66430
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC66808 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_C66750
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC68274 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_C68230
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC682D5 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_C68230
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC69185 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_C69060
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC6922E + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_C69060
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC6931B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_C692B0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC69740 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_C69680
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC83B8C + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_C83B00
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC83ECB + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_C83E30
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC85D6D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_C85BF0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC8EADA + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_C8EAD0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC90543 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_C90510
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xC9F646 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_C9F620
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xCB0DF3 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_CB0C10
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xCB1622 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_CB15A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xCB6CCC + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_CB6C60
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xCB86C4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_CB8650
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xCBAD16 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_CBAD00
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xCBB8D6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_CBB8C0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xCBDA66 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_CBDA50
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xCBDFD0 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_CBDF00
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xCBDFF7 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_CBDF00
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xCBE06F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_CBDF00
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xCBE612 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_CBE380
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xCC12F2 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; 
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xCC1483 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_CC13D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xCC17B1 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_CC1790
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xCC2C16 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_CC2C00
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xCC2C3F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_CC2C00
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xCC2E59 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_CC2E40
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xCC2E92 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_CC2E40
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xCC2EC2 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_CC2E40
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xCC2F18 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_CC2E40
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xCC6712 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_CC66D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xCC7C04 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_CC7C00
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xCC8A36 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_CC88F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xCCF51A + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_CCF2D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xCD9E28 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_CD9E20
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xCE1BF3 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_CE1B30
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xCE1D5B + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_CE1B30
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xCE26B3 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_CE2560
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xCE33F6 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_CE33F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xCE7240 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_CE7230
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xCE7530 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_CE7520
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xCE7705 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_CE76F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xCE7BF2 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_CE7BE0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xCE8C59 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; 
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xCE99EE + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_CE9650
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xCE9C72 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ebx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_CE9C60
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xCF01EB + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_CF0140
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xCF0C0F + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_CF0BB0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xCF1022 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_CF0FB0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xCF502D + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_CF4FE0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xCF8FE4 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_CF8FA0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xCFAAD1 + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_CFAAC0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0xCFBDCF + 3), CModelInfo__ms_modelInfoPtrs.gta_iv);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_CFBD20
		}
		#endif

		// Patch values
		#if TRUE
		{
			// 31000 - count of models
			{
				CPatch::PatchUINT32(g_mCalc.GetCurrentVAbyPreferedVA(0x9BB7E1 + 2), GetFileIDcurrentLimit(FILE_TYPE_MODEL));	// cmp     edi, 31000	; sub_9BB780
				CPatch::PatchUINT32(g_mCalc.GetCurrentVAbyPreferedVA(0x9BC9E6 + 2), GetFileIDcurrentLimit(FILE_TYPE_MODEL));	// cmp     edi, 31000	; sub_9BC9B0
				CPatch::PatchUINT32(g_mCalc.GetCurrentVAbyPreferedVA(0x9BD09C + 1), GetFileIDcurrentLimit(FILE_TYPE_MODEL));	// push    31000	; sub_9BD050
				CPatch::PatchUINT32(g_mCalc.GetCurrentVAbyPreferedVA(0x9C3442 + 2), GetFileIDcurrentLimit(FILE_TYPE_MODEL));	// cmp     edi, 31000	; sub_9C3390
				CPatch::PatchUINT32(g_mCalc.GetCurrentVAbyPreferedVA(0x9C344A + 2), GetFileIDcurrentLimit(FILE_TYPE_MODEL));	// cmp     edi, 31000	; sub_9C3390
			}
		}
		#endif
	}
	#endif

	// Patch WTD limit
	#if TRUE
	{
		CPatch::PatchUINT32(g_mCalc.GetCurrentVAbyPreferedVA(0x86BEE6 + 1), GetFileIDcurrentLimit(FILE_TYPE_WTD));	// push    6000
		CPatch::PatchUINT32(g_mCalc.GetCurrentVAbyPreferedVA(0x86BF5E + 1), GetFileIDcurrentLimit(FILE_TYPE_WTD));	// push    6000
	}
	#endif
	
	// Patch IPL/WPL limit
	#if TRUE
	{
		CPatch::PatchUINT32(g_mCalc.GetCurrentVAbyPreferedVA(0xB4CF21 + 1), GetFileIDcurrentLimit(FILE_TYPE_MAP_PLACEMENT));	// push    500
		CPatch::PatchUINT32(g_mCalc.GetCurrentVAbyPreferedVA(0xB4CF72 + 1), GetFileIDcurrentLimit(FILE_TYPE_MAP_PLACEMENT));	// push    500
	}
	#endif

	// Patch WAD limit
	#if TRUE
	{
		CPatch::PatchUINT32(g_mCalc.GetCurrentVAbyPreferedVA(0x8DABE3 + 1), GetFileIDcurrentLimit(FILE_TYPE_WAD));	// push    1500
		CPatch::PatchUINT32(g_mCalc.GetCurrentVAbyPreferedVA(0x8DAC34 + 1), GetFileIDcurrentLimit(FILE_TYPE_WAD));	// push    1500
	}
	#endif

	// Patch RRR limit
	#if TRUE
	{
		// Patch pointers to CVehicleRecording::StreamingArray
		#if TRUE
		{
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8E1FAB + 1), &CVehicleRecording__StreamingArray.gta_iv->usageCount);	// mov     eax, offset _ZN17CVehicleRecording14StreamingArrayE.usageCount	; sub_8E1F80
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8E1FB8 + 1), &CVehicleRecording__StreamingArray.gta_iv[GetFileIDcurrentLimit(FILE_TYPE_RRR)].usageCount);	// cmp     eax, offset unk_11C137C	; sub_8E1F80
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8E1FD4 + 1), &CVehicleRecording__StreamingArray.gta_iv->pData);	// mov     esi, offset _ZN17CVehicleRecording14StreamingArrayE.pData	; sub_8E1FD0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8E1FFA + 2), &CVehicleRecording__StreamingArray.gta_iv[GetFileIDcurrentLimit(FILE_TYPE_RRR)].pData);	// cmp     esi, offset unk_11C1374	; sub_8E1FD0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8E2118 + 2), &CVehicleRecording__StreamingArray.gta_iv->usageCount);	// lea     eax, _ZN17CVehicleRecording14StreamingArrayE.usageCount[ecx]	; sub_8E20C0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8E21ED + 2), &CVehicleRecording__StreamingArray.gta_iv->number);	// mov     ecx, dword ptr _ZN17CVehicleRecording14StreamingArrayE.number[eax]; CVehicleRecording::StreamingArray	; sub_8E2140
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8E2898 + 1), CVehicleRecording__StreamingArray.gta_iv);	// mov     edx, offset _ZN17CVehicleRecording14StreamingArrayE; CVehicleRecording::StreamingArray	; sub_8E2850
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8E28B3 + 2), &CVehicleRecording__StreamingArray.gta_iv->number);	// mov     dword ptr _ZN17CVehicleRecording14StreamingArrayE.number[eax], esi; CVehicleRecording::StreamingArray	; sub_8E2850
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8E28BC + 2), &CVehicleRecording__StreamingArray.gta_iv->pData);	// mov     _ZN17CVehicleRecording14StreamingArrayE.pData[eax], 0	; sub_8E2850
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8E28F1 + 1), CVehicleRecording__StreamingArray.gta_iv);	// mov     ecx, offset _ZN17CVehicleRecording14StreamingArrayE; CVehicleRecording::StreamingArray	; sub_8E28E0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8E2942 + 2), &CVehicleRecording__StreamingArray.gta_iv->pData);	// cmp     _ZN17CVehicleRecording14StreamingArrayE.pData[eax], ecx	; sub_8E2930
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8E2959 + 1), &CVehicleRecording__StreamingArray.gta_iv->pData);	// mov     esi, offset _ZN17CVehicleRecording14StreamingArrayE.pData	; sub_8E2950
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8E2985 + 2), &CVehicleRecording__StreamingArray.gta_iv[GetFileIDcurrentLimit(FILE_TYPE_RRR)].pData);	// cmp     esi, offset unk_11C1374	; sub_8E2950
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8E29A8 + 2), &CVehicleRecording__StreamingArray.gta_iv->pData);	// mov     eax, _ZN17CVehicleRecording14StreamingArrayE.pData[esi]	; sub_8E29A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8E29AE + 2), &CVehicleRecording__StreamingArray.gta_iv->pData);	// lea     esi, _ZN17CVehicleRecording14StreamingArrayE.pData[esi]	; sub_8E29A0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8E29D7 + 2), &CVehicleRecording__StreamingArray.gta_iv->field_8);	// mov     ecx, _ZN17CVehicleRecording14StreamingArrayE.field_8[edx]	; sub_8E29D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8E29DD + 2), &CVehicleRecording__StreamingArray.gta_iv->pData);	// mov     eax, _ZN17CVehicleRecording14StreamingArrayE.pData[edx]	; sub_8E29D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8E2A3D + 2), &CVehicleRecording__StreamingArray.gta_iv->field_8);	// mov     edi, _ZN17CVehicleRecording14StreamingArrayE.field_8[edx]	; sub_8E29D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8E2B24 + 1), &CVehicleRecording__StreamingArray.gta_iv->usageCount);	// mov     esi, offset _ZN17CVehicleRecording14StreamingArrayE.usageCount	; sub_8E2B20
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8E2B50 + 2), &CVehicleRecording__StreamingArray.gta_iv[GetFileIDcurrentLimit(FILE_TYPE_RRR)].usageCount);	// cmp     esi, offset unk_11C137C	; sub_8E2B20
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8E3774 + 2), &CVehicleRecording__StreamingArray.gta_iv->pData);	// mov     eax, _ZN17CVehicleRecording14StreamingArrayE.pData[esi]	; sub_8E3760
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8E3791 + 2), &CVehicleRecording__StreamingArray.gta_iv->field_8);	// cmp     _ZN17CVehicleRecording14StreamingArrayE.field_8[esi], 0	; sub_8E3760
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8E37D4 + 2), &CVehicleRecording__StreamingArray.gta_iv->field_8);	// cmp     eax, _ZN17CVehicleRecording14StreamingArrayE.field_8[esi]	; sub_8E3760
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8E37FE + 2), &CVehicleRecording__StreamingArray.gta_iv->field_8);	// mov     ecx, _ZN17CVehicleRecording14StreamingArrayE.field_8[eax]	; sub_8E37F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8E3804 + 2), &CVehicleRecording__StreamingArray.gta_iv->pData);	// mov     edx, _ZN17CVehicleRecording14StreamingArrayE.pData[eax]	; sub_8E37F0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8E3858 + 2), &CVehicleRecording__StreamingArray.gta_iv->pData);	// mov     _ZN17CVehicleRecording14StreamingArrayE.pData[esi], eax	; sub_8E3820
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8E3863 + 2), &CVehicleRecording__StreamingArray.gta_iv->pData);	// mov     ecx, _ZN17CVehicleRecording14StreamingArrayE.pData[esi]	; sub_8E3820
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8E3874 + 2), &CVehicleRecording__StreamingArray.gta_iv->field_8);	// mov     _ZN17CVehicleRecording14StreamingArrayE.field_8[esi], eax	; sub_8E3820
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8E3881 + 2), &CVehicleRecording__StreamingArray.gta_iv->field_8);	// cmp     _ZN17CVehicleRecording14StreamingArrayE.field_8[esi], eax	; sub_8E3820
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8E3889 + 2), &CVehicleRecording__StreamingArray.gta_iv->pData);	// mov     ecx, _ZN17CVehicleRecording14StreamingArrayE.pData[esi]	; sub_8E3820
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8E38A2 + 2), &CVehicleRecording__StreamingArray.gta_iv->field_8);	// mov     _ZN17CVehicleRecording14StreamingArrayE.field_8[esi], eax	; sub_8E3820
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8E38AB + 2), &CVehicleRecording__StreamingArray.gta_iv->field_8);	// cmp     eax, _ZN17CVehicleRecording14StreamingArrayE.field_8[esi]	; sub_8E3820
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8E3951 + 2), &CVehicleRecording__StreamingArray.gta_iv->usageCount);	// add     _ZN17CVehicleRecording14StreamingArrayE.usageCount[eax], 1	; sub_8E38D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8E3958 + 2), &CVehicleRecording__StreamingArray.gta_iv->field_8);	// mov     ecx, _ZN17CVehicleRecording14StreamingArrayE.field_8[eax]	; sub_8E38D0
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x8E395E + 2), &CVehicleRecording__StreamingArray.gta_iv->pData);	// mov     edx, _ZN17CVehicleRecording14StreamingArrayE.pData[eax]	; sub_8E38D0
		}
		#endif

		// Patch count
		#if TRUE
		{
			CPatch::PatchUINT32(g_mCalc.GetCurrentVAbyPreferedVA(0x8E43C3 + 1), GetFileIDcurrentLimit(FILE_TYPE_RRR));	// push    2072
		}
		#endif
	}
	#endif
}

static unsigned int numberOfModelInfoFlagDWORDs = 0;

static void InitializeModelInfoFlags()
{
	memset(
		g_fileIDlimits.CSimpleModelInfo__m_UpdatedUVATable.ptr,
		NULL,
		numberOfModelInfoFlagDWORDs * sizeof(uint32_t)
	);
}

// Patches ID limits if neccessary.
void FileIDlimit::PatchFileIDlimits_Bully_SE_1_20()
{
	using namespace Game_BullyScholarschipEdition;

	if (!WriteToLogIfLimitsAreChanged())
		return;
				
	if(!CPatch::IsDebugModeActive())
		NumberOfStreamingInfos = GetCountOfAllFileIDs() + 4 /* loaded, requested */;

	// ID limit checks
	#if TRUE
	{
		// See if count is unsigned
		if (GetCountOfAllFileIDs() >= 32768)
			throw f92_runtime_error("Count of file IDs should be less than 32768."
				"\nUnsigned ID for Bully Scholarship Edition is not supported yet."
				"\nCount of IDs: %d", GetCountOfAllFileIDs()
				);

		// See if COL limit is a multiply of 5
		if (GetFileIDcurrentLimit(FILE_TYPE_COL) % 5 != 0)
			throw f92_runtime_error("COL limit should be a multiply by 5.\n%d models isn't a multiply of 5!",
				GetFileIDcurrentLimit(FILE_TYPE_COL)
				);
	}
	#endif

	CGenericLogStorage::SaveFormattedTextLn(
		"\nNew ID limits:\n"
		"%d - %d (%d) - NIF models defined within IDE files.\n"
		"%d - %d (%d) - NFT texture archives.\n"
		"%d - %d (%d) - COL collision archives.\n"
		"%d - %d (%d) - AGR\n"
		"%d - %d (%d) - IPB binary placement files\n"
		"%d - %d (%d) - LIP\n"
		"%d - %d (%d) - CAT compressed action trees.\n"
		"%d - %d (%d) - LUR Lua compiled scripts\n"
		"%d - count of all file IDs\n\n"
		"%d - count of possible streaming infos"
		,

		GetBaseID(FILE_TYPE_NIF), GetLastIDforType(FILE_TYPE_NIF), GetFileIDcurrentLimit(FILE_TYPE_NIF),
		GetBaseID(FILE_TYPE_NFT), GetLastIDforType(FILE_TYPE_NFT), GetFileIDcurrentLimit(FILE_TYPE_NFT),
		GetBaseID(FILE_TYPE_COL), GetLastIDforType(FILE_TYPE_COL), GetFileIDcurrentLimit(FILE_TYPE_COL),
		GetBaseID(FILE_TYPE_AGR), GetLastIDforType(FILE_TYPE_AGR), GetFileIDcurrentLimit(FILE_TYPE_AGR),
		GetBaseID(FILE_TYPE_IPB), GetLastIDforType(FILE_TYPE_IPB), GetFileIDcurrentLimit(FILE_TYPE_IPB),
		GetBaseID(FILE_TYPE_LIP), GetLastIDforType(FILE_TYPE_LIP), GetFileIDcurrentLimit(FILE_TYPE_LIP),
		GetBaseID(FILE_TYPE_CAT), GetLastIDforType(FILE_TYPE_CAT), GetFileIDcurrentLimit(FILE_TYPE_CAT),
		GetBaseID(FILE_TYPE_LUR), GetLastIDforType(FILE_TYPE_LUR), GetFileIDcurrentLimit(FILE_TYPE_LUR),

		GetCountOfAllFileIDs(),

		NumberOfStreamingInfos
		);

	// Allocate variables
	#if TRUE
	if (!CPatch::IsDebugModeActive())
	{
		// CModelInfo::ms_modelInfoPtrs
		if (GetFileIDcurrentLimit(FILE_TYPE_NIF) > FileIDlimit::GetFileIDdefaultLimit(FILE_TYPE_NIF))
		{
			CModelInfo__ms_modelInfoPtrs.bully_se = new CBaseModelInfo*[GetFileIDcurrentLimit(FILE_TYPE_NIF)];
			memset(
				CModelInfo__ms_modelInfoPtrs.bully_se,
				NULL,
				GetFileIDcurrentLimit(FILE_TYPE_NIF) * sizeof(CBaseModelInfo*)
			);

			CModelInfo__ms_modelInfoPtrs.bIsAllocated.Set(true);
		}

		// CSimpleModelInfo::m_UpdatedUVATable		
		{
			tFileTypeLimit top = GetFileIDcurrentLimit(FILE_TYPE_NIF);
			tFileTypeLimit rest = top % 32;

			if (rest != 0)
				top += 32 - rest;

			numberOfModelInfoFlagDWORDs = top / 32;

			if (GetFileIDcurrentLimit(FILE_TYPE_NIF) > FileIDlimit::GetFileIDdefaultLimit(FILE_TYPE_NIF))
			{
				CSimpleModelInfo__m_UpdatedUVATable.ptr = new uint32_t[numberOfModelInfoFlagDWORDs];
				CSimpleModelInfo__m_UpdatedUVATable.bIsAllocated.Set(true);
			}
		}

		// CStreaming::ms_apAGRBuffers
		if (GetFileIDcurrentLimit(FILE_TYPE_AGR) > GetFileIDdefaultLimit(FILE_TYPE_AGR))
		{
			CStreaming__ms_apAGRBuffers.bully_se = new void*[GetFileIDcurrentLimit(FILE_TYPE_AGR)];
			memset(CStreaming__ms_apAGRBuffers.bully_se, NULL, GetFileIDcurrentLimit(FILE_TYPE_AGR) * sizeof(void*));
			CStreaming__ms_apAGRBuffers.bIsAllocated.Set(true);
		}
		
		// CStreaming::ms_aInfoForModel
		if (NumberOfStreamingInfos > 17204)
		{
			CStreaming__ms_aInfoForModel.bully_se = new CStreamingInfo[NumberOfStreamingInfos];

			memset(
				CStreaming__ms_aInfoForModel.bully_se,
				NULL,
				NumberOfStreamingInfos * sizeof(CStreamingInfo)
			);

			CStreaming__ms_aInfoForModel.bIsAllocated.Set(true);
		}
	}
	#endif

	// CStreaming::ms_aInfoForModel
	#if TRUE
	{
		// Patch references to CStreaming::ms_aInfoForModel
		#if TRUE
		{				
			CPatch::PatchPointer(0x52DD40 + 1, CStreaming__ms_aInfoForModel.bully_se);	// add     eax, offset _ZN10CStreaming16ms_aInfoForModelE; CStreaming::ms_aInfoForModel	; sub_52DD30
			CPatch::PatchPointer(0x52DD5F + 2, CStreaming__ms_aInfoForModel.bully_se);	// sub     esi, offset _ZN10CStreaming16ms_aInfoForModelE; CStreaming::ms_aInfoForModel	; sub_52DD50
			CPatch::PatchPointer(0x52DED1 + 1, &CStreaming__ms_aInfoForModel.bully_se->usPrev);	// add     eax, offset _ZN10CStreaming16ms_aInfoForModelE.usPrev	; sub_52DDA0
			CPatch::PatchPointer(0x52DF73 + 1, &CStreaming__ms_aInfoForModel.bully_se->usPrev);	// add     eax, offset _ZN10CStreaming16ms_aInfoForModelE.usPrev	; sub_52DDA0
			CPatch::PatchPointer(0x52E002 + 1, CStreaming__ms_aInfoForModel.bully_se);	// add     eax, offset _ZN10CStreaming16ms_aInfoForModelE; CStreaming::ms_aInfoForModel	; sub_52DFF0
			CPatch::PatchPointer(0x52E2F3 + 1, CStreaming__ms_aInfoForModel.bully_se);	// add     eax, offset _ZN10CStreaming16ms_aInfoForModelE; CStreaming::ms_aInfoForModel	; _ZN10CStreaming15LoadCdDirectoryEPKci
			CPatch::PatchPointer(0x52E499 + 3, &CStreaming__ms_aInfoForModel.bully_se->usNextOnCd);	// mov     _ZN10CStreaming16ms_aInfoForModelE.usNextOnCd[eax], dx	; _ZN10CStreaming15LoadCdDirectoryEPKci
			CPatch::PatchPointer(0x52E72D + 1, CStreaming__ms_aInfoForModel.bully_se);	// add     eax, offset _ZN10CStreaming16ms_aInfoForModelE; CStreaming::ms_aInfoForModel	; sub_52E720
			CPatch::PatchPointer(0x52E78D + 1, CStreaming__ms_aInfoForModel.bully_se);	// add     eax, offset _ZN10CStreaming16ms_aInfoForModelE; CStreaming::ms_aInfoForModel	; sub_52E780
			CPatch::PatchPointer(0x52E806 + 2, CStreaming__ms_aInfoForModel.bully_se);	// add     esi, offset _ZN10CStreaming16ms_aInfoForModelE; CStreaming::ms_aInfoForModel	; sub_52E7F0
			CPatch::PatchPointer(0x52E880 + 1, CStreaming__ms_aInfoForModel.bully_se);	// add     eax, offset _ZN10CStreaming16ms_aInfoForModelE; CStreaming::ms_aInfoForModel	; sub_52E870
			CPatch::PatchPointer(0x52E8F3 + 1, CStreaming__ms_aInfoForModel.bully_se);	// add     eax, offset _ZN10CStreaming16ms_aInfoForModelE; CStreaming::ms_aInfoForModel	; sub_52E8D0
			CPatch::PatchPointer(0x52E9BD + 2, &CStreaming__ms_aInfoForModel.bully_se->ucFlags);	// or      _ZN10CStreaming16ms_aInfoForModelE.ucFlags[eax], 0C0h	; sub_52E8D0
			CPatch::PatchPointer(0x52E9C4 + 2, &CStreaming__ms_aInfoForModel.bully_se->ucFlags);	// lea     eax, _ZN10CStreaming16ms_aInfoForModelE.ucFlags[eax]	; sub_52E8D0
			CPatch::PatchPointer(0x52E9D9 + 2, &CStreaming__ms_aInfoForModel.bully_se->ucFlags);	// lea     eax, _ZN10CStreaming16ms_aInfoForModelE.ucFlags[eax]	; sub_52E8D0
			CPatch::PatchPointer(0x52EA16 + 2, &CStreaming__ms_aInfoForModel.bully_se->ucFlags);	// or      _ZN10CStreaming16ms_aInfoForModelE.ucFlags[eax], 0C0h	; sub_52E8D0
			CPatch::PatchPointer(0x52EA1D + 2, &CStreaming__ms_aInfoForModel.bully_se->ucFlags);	// lea     eax, _ZN10CStreaming16ms_aInfoForModelE.ucFlags[eax]	; sub_52E8D0
			CPatch::PatchPointer(0x52EA32 + 2, &CStreaming__ms_aInfoForModel.bully_se->ucFlags);	// lea     eax, _ZN10CStreaming16ms_aInfoForModelE.ucFlags[eax]	; sub_52E8D0
			CPatch::PatchPointer(0x52EA94 + 2, &CStreaming__ms_aInfoForModel.bully_se->ucFlags);	// test    _ZN10CStreaming16ms_aInfoForModelE.ucFlags[eax], 0C0h	; sub_52E8D0
			CPatch::PatchPointer(0x52EAA6 + 2, &CStreaming__ms_aInfoForModel.bully_se->ucFlags);	// lea     eax, _ZN10CStreaming16ms_aInfoForModelE.ucFlags[eax]	; sub_52E8D0
			CPatch::PatchPointer(0x52F652 + 2, CStreaming__ms_aInfoForModel.bully_se);	// add     edi, offset _ZN10CStreaming16ms_aInfoForModelE; CStreaming::ms_aInfoForModel	; _ZN10CStreaming11RemoveModelEi
			CPatch::PatchPointer(0x52F940 + 3, CStreaming__ms_aInfoForModel.bully_se);	// cmp     word ptr _ZN10CStreaming16ms_aInfoForModelE.usNext[eax], 0FFFFh; CStreaming::ms_aInfoForModel	; sub_52F8B0
			CPatch::PatchPointer(0x52FA34 + 3, CStreaming__ms_aInfoForModel.bully_se);	// cmp     word ptr _ZN10CStreaming16ms_aInfoForModelE.usNext[eax], 0FFFFh; CStreaming::ms_aInfoForModel	; sub_52F9A0
			CPatch::PatchPointer(0x52FE4E + 1, CStreaming__ms_aInfoForModel.bully_se);	// add     eax, offset _ZN10CStreaming16ms_aInfoForModelE; CStreaming::ms_aInfoForModel	; sub_52FE40
			CPatch::PatchPointer(0x52FF4E + 1, CStreaming__ms_aInfoForModel.bully_se);	// add     eax, offset _ZN10CStreaming16ms_aInfoForModelE; CStreaming::ms_aInfoForModel	; sub_52FF40
			CPatch::PatchPointer(0x53009D + 2, &CStreaming__ms_aInfoForModel.bully_se->ucFlags);	// test    _ZN10CStreaming16ms_aInfoForModelE.ucFlags[eax], 3	; sub_530060
			CPatch::PatchPointer(0x5300C0 + 2, &CStreaming__ms_aInfoForModel.bully_se->uiLoadFlag);	// cmp     _ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[eax], 1	; sub_530060
			CPatch::PatchPointer(0x530154 + 2, CStreaming__ms_aInfoForModel.bully_se);	// add     esi, offset _ZN10CStreaming16ms_aInfoForModelE; CStreaming::ms_aInfoForModel	; sub_530100
			CPatch::PatchPointer(0x53037E + 1, &CStreaming__ms_aInfoForModel.bully_se->ucFlags);	// mov     ebp, offset _ZN10CStreaming16ms_aInfoForModelE.ucFlags	; sub_530100
			CPatch::PatchPointer(0x530AE7 + 1, CStreaming__ms_aInfoForModel.bully_se);	// add     eax, offset _ZN10CStreaming16ms_aInfoForModelE; CStreaming::ms_aInfoForModel	; sub_530A70
			CPatch::PatchPointer(0x531CCF + 2, CStreaming__ms_aInfoForModel.bully_se);	// add     edi, offset _ZN10CStreaming16ms_aInfoForModelE; CStreaming::ms_aInfoForModel	; _ZN10CStreaming21ConvertBufferToObjectEPcii
			CPatch::PatchPointer(0x532036 + 2, CStreaming__ms_aInfoForModel.bully_se);	// add     esi, offset _ZN10CStreaming16ms_aInfoForModelE; CStreaming::ms_aInfoForModel	; _ZN10CStreaming21ProcessLoadingChannelEi
			CPatch::PatchPointer(0x532524 + 2, CStreaming__ms_aInfoForModel.bully_se);	// add     ebp, offset _ZN10CStreaming16ms_aInfoForModelE; CStreaming::ms_aInfoForModel	; sub_532510
			CPatch::PatchPointer(0x5327A7 + 2, &CStreaming__ms_aInfoForModel.bully_se->uiLoadFlag);	// cmp     _ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[eax], 4	; sub_532750
			CPatch::PatchPointer(0x53286C + 2, CStreaming__ms_aInfoForModel.bully_se);	// add     edi, offset _ZN10CStreaming16ms_aInfoForModelE; CStreaming::ms_aInfoForModel	; sub_532750
			CPatch::PatchPointer(0x532AA8 + 3, &CStreaming__ms_aInfoForModel.bully_se->uiLoadFlag);	// movzx   eax, _ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[eax]	; sub_532A10
			CPatch::PatchPointer(0x532AE5 + 3, &CStreaming__ms_aInfoForModel.bully_se->ucFlags);	// movzx   eax, _ZN10CStreaming16ms_aInfoForModelE.ucFlags[eax]	; sub_532A10
			CPatch::PatchPointer(0x532B57 + 3, &CStreaming__ms_aInfoForModel.bully_se->uiLoadFlag);	// movzx   eax, _ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[eax]	; sub_532A10
			CPatch::PatchPointer(0x532B85 + 1, CStreaming__ms_aInfoForModel.bully_se);	// add     eax, offset _ZN10CStreaming16ms_aInfoForModelE; CStreaming::ms_aInfoForModel	; sub_532A10
			CPatch::PatchPointer(0x532CD8 + 2, &CStreaming__ms_aInfoForModel.bully_se->ucFlags);	// test    _ZN10CStreaming16ms_aInfoForModelE.ucFlags[eax], bl	; sub_532C80
			CPatch::PatchPointer(0x532D33 + 1, CStreaming__ms_aInfoForModel.bully_se);	// add     eax, offset _ZN10CStreaming16ms_aInfoForModelE; CStreaming::ms_aInfoForModel	; sub_532C80
			CPatch::PatchPointer(0x532D79 + 2, &CStreaming__ms_aInfoForModel.bully_se->ucFlags);	// test    _ZN10CStreaming16ms_aInfoForModelE.ucFlags[eax], bl	; sub_532C80
			CPatch::PatchPointer(0x532D97 + 1, CStreaming__ms_aInfoForModel.bully_se);	// add     eax, offset _ZN10CStreaming16ms_aInfoForModelE; CStreaming::ms_aInfoForModel	; sub_532C80
			CPatch::PatchPointer(0x532DD4 + 2, CStreaming__ms_aInfoForModel.bully_se);	// add     esi, offset _ZN10CStreaming16ms_aInfoForModelE; CStreaming::ms_aInfoForModel	; sub_532C80
			CPatch::PatchPointer(0x532E57 + 1, CStreaming__ms_aInfoForModel.bully_se);	// add     eax, offset _ZN10CStreaming16ms_aInfoForModelE; CStreaming::ms_aInfoForModel	; sub_532C80
			CPatch::PatchPointer(0x532EB6 + 1, CStreaming__ms_aInfoForModel.bully_se);	// add     eax, offset _ZN10CStreaming16ms_aInfoForModelE; CStreaming::ms_aInfoForModel	; sub_532C80
			CPatch::PatchPointer(0x533041 + 2, &CStreaming__ms_aInfoForModel.bully_se->uiLoadFlag);	// cmp     _ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[eax], 2	; sub_533030
			CPatch::PatchPointer(0x533056 + 2, &CStreaming__ms_aInfoForModel.bully_se->uiLoadFlag);	// cmp     _ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[eax], 1	; sub_533030
			CPatch::PatchPointer(0x5333B5 + 1, CStreaming__ms_aInfoForModel.bully_se);	// add     eax, offset _ZN10CStreaming16ms_aInfoForModelE; CStreaming::ms_aInfoForModel	; sub_533200
			CPatch::PatchPointer(0x533433 + 1, CStreaming__ms_aInfoForModel.bully_se);	// add     eax, offset _ZN10CStreaming16ms_aInfoForModelE; CStreaming::ms_aInfoForModel	; sub_5333E0
			CPatch::PatchPointer(0x5334DF + 6, CStreaming__ms_aInfoForModel.bully_se);	// mov     dword_CA397C, offset _ZN10CStreaming16ms_aInfoForModelE; CStreaming::ms_aInfoForModel	; CStreaming::Init
			CPatch::PatchPointer(0x5334E9 + 6, &CStreaming__ms_aInfoForModel.bully_se[NumberOfStreamingInfos - 4]);	// mov     dword_CA3884, (offset _ZN10CStreaming16ms_aInfoForModelE.usNext+43300h)	; CStreaming::Init
			CPatch::PatchPointer(0x5334F3 + 6, &CStreaming__ms_aInfoForModel.bully_se[NumberOfStreamingInfos - 3]);	// mov     dword_CA3880, (offset _ZN10CStreaming16ms_aInfoForModelE.usNext+43310h)	; CStreaming::Init
			CPatch::PatchPointer(0x5334FD + 6, &CStreaming__ms_aInfoForModel.bully_se[NumberOfStreamingInfos - 2]);	// mov     dword_CA387C, (offset _ZN10CStreaming16ms_aInfoForModelE.usNext+43320h)	; CStreaming::Init
			CPatch::PatchPointer(0x533507 + 6, &CStreaming__ms_aInfoForModel.bully_se[NumberOfStreamingInfos - 1]);	// mov     dword_CA3878, (offset _ZN10CStreaming16ms_aInfoForModelE.usNext+43330h)	; CStreaming::Init
			CPatch::PatchPointer(0x53366A + 2, &CStreaming__ms_aInfoForModel.bully_se->uiLoadFlag);	// mov     _ZN10CStreaming16ms_aInfoForModelE.uiLoadFlag[eax], 1	; CStreaming::Init
			CPatch::PatchPointer(0x53384E + 2, &CStreaming__ms_aInfoForModel.bully_se->iBlockCount);	// cmp     _ZN10CStreaming16ms_aInfoForModelE.iBlockCount[eax], ebx	; CStreaming::Init
			CPatch::PatchPointer(0x5338B3 + 1, CStreaming__ms_aInfoForModel.bully_se);	// add     eax, offset _ZN10CStreaming16ms_aInfoForModelE; CStreaming::ms_aInfoForModel	; sub_533880
			CPatch::PatchPointer(0x5339A6 + 2, &CStreaming__ms_aInfoForModel.bully_se->ucFlags);	// test    _ZN10CStreaming16ms_aInfoForModelE.ucFlags[eax], 2	; sub_533980
			CPatch::PatchPointer(0x533FC1 + 2, CStreaming__ms_aInfoForModel.bully_se);	// add     esi, offset _ZN10CStreaming16ms_aInfoForModelE; CStreaming::ms_aInfoForModel	; _ZN10CStreaming22LoadAllRequestedModelsEb
			CPatch::PatchPointer(0x534680 + 1, &CStreaming__ms_aInfoForModel.bully_se->ucFlags);	// mov     eax, offset _ZN10CStreaming16ms_aInfoForModelE.ucFlags	; sub_534510
			CPatch::PatchPointer(0x534916 + 1, CStreaming__ms_aInfoForModel.bully_se);	// add     eax, offset _ZN10CStreaming16ms_aInfoForModelE; CStreaming::ms_aInfoForModel	; sub_5348D0
			CPatch::PatchPointer(0x5349AC + 2, &CStreaming__ms_aInfoForModel.bully_se->ucFlags);	// test    _ZN10CStreaming16ms_aInfoForModelE.ucFlags[eax], 0C0h	; sub_5348D0
			CPatch::PatchPointer(0x8DDB80 + 1, CStreaming__ms_aInfoForModel.bully_se);	// mov     eax, offset _ZN10CStreaming16ms_aInfoForModelE; CStreaming::ms_aInfoForModel	; ms_aInfoForModel_constructor
		}
		#endif

		// Patch 17203
		#if TRUE
		{
			CPatch::PatchUINT32(0x8DDB85 + 1, NumberOfStreamingInfos - 1);	// mov     ecx, 17203	; ms_aInfoForModel_constructor
		}
		#endif
	}
	#endif

	// Patch base IDs and multiplies
	#if TRUE
	{
		// 23993 (LUR base index)
		#if TRUE
		{
			CPatch::PatchUINT32(0x52E463 + 1, GetBaseID(FILE_TYPE_LUR));	// add     eax, 5DB9h	; _ZN10CStreaming15LoadCdDirectoryEPKci
			CPatch::PatchUINT32(0x52EC42 + 1, GetBaseID(FILE_TYPE_LUR));	// cmp     eax, 5DB9h	; getNameByFileID
			CPatch::PatchUINT32(0x52F721 + 2, GetBaseID(FILE_TYPE_LUR));	// cmp     esi, 5DB9h	; sub_52F640
			CPatch::PatchUINT32(0x52F834 + 2, GetBaseID(FILE_TYPE_LUR));	// cmp     esi, 5DB9h	; sub_52F640
			CPatch::PatchUINT32(0x5303F9 + 2, GetBaseID(FILE_TYPE_LUR));	// cmp     esi, 5DB9h	; sub_530100
			CPatch::PatchUINT32(0x530D5F + 2, GetBaseID(FILE_TYPE_LUR));	// cmp     esi, 5DB9h	; _ZN10CStreaming20RemoveLeastUsedModelEj
			CPatch::PatchUINT32(0x5318CB + 2, GetBaseID(FILE_TYPE_LUR));	// cmp     esi, 5DB9h	; _ZN10CStreaming12MakeSpaceForEi
			CPatch::PatchUINT32(0x531F37 + 2, GetBaseID(FILE_TYPE_LUR));	// cmp     esi, 5DB9h	; _ZN10CStreaming21ConvertBufferToObjectEPcii
			CPatch::PatchUINT32(0x531FE3 + 2, GetBaseID(FILE_TYPE_LUR));	// cmp     esi, 5DB9h	; _ZN10CStreaming21ConvertBufferToObjectEPcii
			CPatch::PatchUINT32(0x532932 + 2, GetBaseID(FILE_TYPE_LUR));	// cmp     esi, 5DB9h	; sub_532750
			CPatch::PatchUINT32(0x5BF6BE + 1, GetBaseID(FILE_TYPE_LUR));	// add     eax, 5DB9h	; sub_5BF6A0
			CPatch::PatchUINT32(0x5BF6FC + 1, GetBaseID(FILE_TYPE_LUR));	// add     eax, 5DB9h	; sub_5BF6E0
			CPatch::PatchUINT32(0x5DB060 + 2, GetBaseID(FILE_TYPE_LUR));	// lea     eax, [esi+5DB9h]	; sub_5334C0
		}
		#endif

		// -23993 (-LUR base index)
		#if TRUE
		{
			CPatch::PatchUINT32(0x52EC77 + 1, -GetBaseID(FILE_TYPE_LUR));	// add     eax, 0FFFFA247h	; getNameByFileID
			CPatch::PatchUINT32(0x52F729 + 2, -GetBaseID(FILE_TYPE_LUR));	// lea     edx, [esi-5DB9h]	; sub_52F640
			CPatch::PatchUINT32(0x52F83C + 2, -GetBaseID(FILE_TYPE_LUR));	// add     esi, 0FFFFA247h	; sub_52F640
			CPatch::PatchUINT32(0x531F65 + 2, -GetBaseID(FILE_TYPE_LUR));	// lea     eax, [esi-5DB9h]	; _ZN10CStreaming21ConvertBufferToObjectEPcii
		}
		#endif

		// 23873 (CAT base index)
		#if TRUE
		{
			CPatch::PatchUINT32(0x415D04 + 2, GetBaseID(FILE_TYPE_CAT));	// lea     esi, [eax+5D41h]	; sub_415AD0
			CPatch::PatchUINT32(0x460305 + 2, GetBaseID(FILE_TYPE_CAT));	// add     edi, 5D41h	; sub_45FAB0
			CPatch::PatchUINT32(0x4EA2F4 + 1, GetBaseID(FILE_TYPE_CAT));	// add     eax, 5D41h	; sub_4EA2F0
			CPatch::PatchUINT32(0x4ED8F1 + 2, GetBaseID(FILE_TYPE_CAT));	// lea     ebx, [edi+5D41h]	; sub_4ED8E0
			CPatch::PatchUINT32(0x51C410 + 2, GetBaseID(FILE_TYPE_CAT));	// lea     edx, [ebx+5D41h]	; sub_51C3D0
			CPatch::PatchUINT32(0x51C421 + 2, GetBaseID(FILE_TYPE_CAT));	// lea     eax, [edi+5D41h]	; sub_51C3D0
			CPatch::PatchUINT32(0x52E41E + 1, GetBaseID(FILE_TYPE_CAT));	// add     eax, 5D41h	; _ZN10CStreaming15LoadCdDirectoryEPKci
			CPatch::PatchUINT32(0x52EC1D + 1, GetBaseID(FILE_TYPE_CAT));	// cmp     eax, 5D41h	; getNameByFileID
			CPatch::PatchUINT32(0x5303ED + 2, GetBaseID(FILE_TYPE_CAT));	// cmp     esi, 5D41h	; sub_530100
			CPatch::PatchUINT32(0x53054D + 2, GetBaseID(FILE_TYPE_CAT));	// cmp     esi, 5D41h	; sub_530100
			CPatch::PatchUINT32(0x5318C3 + 2, GetBaseID(FILE_TYPE_CAT));	// cmp     esi, 5D41h	; _ZN10CStreaming12MakeSpaceForEi
			CPatch::PatchUINT32(0x531F1B + 2, GetBaseID(FILE_TYPE_CAT));	// cmp     esi, 5D41h	; _ZN10CStreaming21ConvertBufferToObjectEPcii
			CPatch::PatchUINT32(0x532164 + 2, GetBaseID(FILE_TYPE_CAT));	// add     ebp, 5D41h	; sub_532020
			CPatch::PatchUINT32(0x53292A + 2, GetBaseID(FILE_TYPE_CAT));	// cmp     esi, 5D41h	; sub_532750
			CPatch::PatchUINT32(0x5BF5E3 + 2, GetBaseID(FILE_TYPE_CAT));	// lea     eax, [esi+5D41h]	; sub_5BF5C0
			CPatch::PatchUINT32(0x5BF627 + 1, GetBaseID(FILE_TYPE_CAT));	// add     eax, 5D41h	; sub_5BF610
			CPatch::PatchUINT32(0x5C565A + 2, GetBaseID(FILE_TYPE_CAT));	// lea     edx, [edi+5D41h]	; sub_5C5400
			CPatch::PatchUINT32(0x5D9AB3 + 2, GetBaseID(FILE_TYPE_CAT));	// lea     eax, [esi+5D41h]	; sub_5D9A80
			CPatch::PatchUINT32(0x5FB5C4 + 1, GetBaseID(FILE_TYPE_CAT));	// add     eax, 5D41h	; sub_5FB550
		}
		#endif

		// -23873 (-CAT base index)
		#if TRUE
		{
			CPatch::PatchUINT32(0x52EC4E + 1, -GetBaseID(FILE_TYPE_CAT));	// add     eax, 0FFFFA2BFh	; getNameByFileID
			CPatch::PatchUINT32(0x531F48 + 2, -GetBaseID(FILE_TYPE_CAT));	// lea     edx, [esi-5D41h]	; _ZN10CStreaming21ConvertBufferToObjectEPcii
		}
		#endif

		// 23380 (LIP base index)
		#if TRUE
		{
			CPatch::PatchUINT32(0x52E3E4 + 1, GetBaseID(FILE_TYPE_LIP));	// add     eax, 5B54h	; _ZN10CStreaming15LoadCdDirectoryEPKci
			CPatch::PatchUINT32(0x52EBCB + 1, GetBaseID(FILE_TYPE_LIP));	// cmp     eax, 5B54h	; getNameByFileID
			CPatch::PatchUINT32(0x52F73C + 2, GetBaseID(FILE_TYPE_LIP));	// cmp     esi, 5B54h	; sub_52F640
			CPatch::PatchUINT32(0x52F883 + 2, GetBaseID(FILE_TYPE_LIP));	// cmp     esi, 5B54h	; sub_52F640
			CPatch::PatchUINT32(0x5303C5 + 2, GetBaseID(FILE_TYPE_LIP));	// cmp     esi, 5B54h	; sub_530100
			CPatch::PatchUINT32(0x5303E5 + 2, GetBaseID(FILE_TYPE_LIP));	// cmp     esi, 5B54h	; sub_530100
			CPatch::PatchUINT32(0x53052F + 2, GetBaseID(FILE_TYPE_LIP));	// cmp     esi, 5B54h	; sub_530100
			CPatch::PatchUINT32(0x5318A3 + 2, GetBaseID(FILE_TYPE_LIP));	// cmp     esi, 5B54h	; _ZN10CStreaming12MakeSpaceForEi
			CPatch::PatchUINT32(0x5318BB + 2, GetBaseID(FILE_TYPE_LIP));	// cmp     esi, 5B54h	; _ZN10CStreaming12MakeSpaceForEi
			CPatch::PatchUINT32(0x531EF7 + 2, GetBaseID(FILE_TYPE_LIP));	// cmp     esi, 5B54h	; _ZN10CStreaming21ConvertBufferToObjectEPcii
			CPatch::PatchUINT32(0x533810 + 2, GetBaseID(FILE_TYPE_LIP));	// lea     ecx, [esi+23380]	; CStreaming::Init
			CPatch::PatchUINT32(0x59FE24 + 1, GetBaseID(FILE_TYPE_LIP));	// add     eax, 5B54h	; sub_59FE20
			CPatch::PatchUINT32(0x5A53EA + 2, GetBaseID(FILE_TYPE_LIP));	// add     edx, 5B54h	; sub_5A5060
			CPatch::PatchUINT32(0x5A581E + 1, GetBaseID(FILE_TYPE_LIP));	// add     eax, 5B54h	; sub_5A57A0
			CPatch::PatchUINT32(0x68E165 + 2, GetBaseID(FILE_TYPE_LIP));	// add     edi, 5B54h	; sub_68E0B0
		}
		#endif

		// -23380 (-LIP base index)
		#if TRUE
		{
			CPatch::PatchUINT32(0x52E3E4 + 1, GetBaseID(FILE_TYPE_LIP));	// add     eax, 5B54h	; _ZN10CStreaming15LoadCdDirectoryEPKci
			CPatch::PatchUINT32(0x52EBCB + 1, GetBaseID(FILE_TYPE_LIP));	// cmp     eax, 23380	; getNameByFileID
			CPatch::PatchUINT32(0x52F73C + 2, GetBaseID(FILE_TYPE_LIP));	// cmp     esi, 5B54h	; sub_52F640
			CPatch::PatchUINT32(0x52F883 + 2, GetBaseID(FILE_TYPE_LIP));	// cmp     esi, 5B54h	; sub_52F640
			CPatch::PatchUINT32(0x5303C5 + 2, GetBaseID(FILE_TYPE_LIP));	// cmp     esi, 5B54h	; sub_530100
			CPatch::PatchUINT32(0x5303E5 + 2, GetBaseID(FILE_TYPE_LIP));	// cmp     esi, 5B54h	; sub_530100
			CPatch::PatchUINT32(0x53052F + 2, GetBaseID(FILE_TYPE_LIP));	// cmp     esi, 5B54h	; sub_530100
			CPatch::PatchUINT32(0x5318A3 + 2, GetBaseID(FILE_TYPE_LIP));	// cmp     esi, 5B54h	; _ZN10CStreaming12MakeSpaceForEi
			CPatch::PatchUINT32(0x5318BB + 2, GetBaseID(FILE_TYPE_LIP));	// cmp     esi, 5B54h	; _ZN10CStreaming12MakeSpaceForEi
			CPatch::PatchUINT32(0x531EF7 + 2, GetBaseID(FILE_TYPE_LIP));	// cmp     esi, 5B54h	; _ZN10CStreaming21ConvertBufferToObjectEPcii
			CPatch::PatchUINT32(0x533810 + 2, GetBaseID(FILE_TYPE_LIP));	// lea     ecx, [esi+23380]	; CStreaming::Init
			CPatch::PatchUINT32(0x59FE24 + 1, GetBaseID(FILE_TYPE_LIP));	// add     eax, 5B54h	; sub_59FE20
			CPatch::PatchUINT32(0x5A53EA + 2, GetBaseID(FILE_TYPE_LIP));	// add     edx, 5B54h	; sub_5A5060
			CPatch::PatchUINT32(0x5A581E + 1, GetBaseID(FILE_TYPE_LIP));	// add     eax, 5B54h	; sub_5A57A0
			CPatch::PatchUINT32(0x68E165 + 2, GetBaseID(FILE_TYPE_LIP));	// add     edi, 5B54h	; sub_68E0B0
		}
		#endif

		// 23280 (IPB base index)
		#if TRUE
		{
			CPatch::PatchUINT32(0x438656 + 2, GetBaseID(FILE_TYPE_IPB));	// lea     ecx, [esi+5AF0h]	; sub_4385E0
			CPatch::PatchUINT32(0x43870E + 2, GetBaseID(FILE_TYPE_IPB));	// lea     ecx, [esi+5AF0h]	; sub_438690
			CPatch::PatchUINT32(0x439449 + 2, GetBaseID(FILE_TYPE_IPB));	// lea     ecx, [esi+5AF0h]	; sub_439370
			CPatch::PatchUINT32(0x43951C + 2, GetBaseID(FILE_TYPE_IPB));	// lea     edx, [edi+5AF0h]	; sub_439370
			CPatch::PatchUINT32(0x52E3B7 + 1, GetBaseID(FILE_TYPE_IPB));	// add     eax, 5AF0h	; _ZN10CStreaming15LoadCdDirectoryEPKci
			CPatch::PatchUINT32(0x52EB64 + 1, GetBaseID(FILE_TYPE_IPB));	// cmp     eax, 23280	; getNameByFileID
			CPatch::PatchUINT32(0x52F6F9 + 2, GetBaseID(FILE_TYPE_IPB));	// cmp     esi, 5AF0h	; sub_52F640
			CPatch::PatchUINT32(0x52F855 + 2, GetBaseID(FILE_TYPE_IPB));	// cmp     esi, 5AF0h	; sub_52F640
			CPatch::PatchUINT32(0x52FE75 + 2, GetBaseID(FILE_TYPE_IPB));	// cmp     esi, 5AF0h	; sub_52FE40
			CPatch::PatchUINT32(0x52FF75 + 2, GetBaseID(FILE_TYPE_IPB));	// cmp     esi, 5AF0h	; sub_52FF40
			CPatch::PatchUINT32(0x5303AF + 2, GetBaseID(FILE_TYPE_IPB));	// cmp     esi, 5AF0h	; sub_530100
			CPatch::PatchUINT32(0x5303BD + 2, GetBaseID(FILE_TYPE_IPB));	// cmp     esi, 5AF0h	; sub_530100
			CPatch::PatchUINT32(0x5304D7 + 2, GetBaseID(FILE_TYPE_IPB));	// cmp     esi, 5AF0h	; sub_530100
			CPatch::PatchUINT32(0x53188D + 2, GetBaseID(FILE_TYPE_IPB));	// cmp     esi, 5AF0h	; _ZN10CStreaming12MakeSpaceForEi
			CPatch::PatchUINT32(0x53189B + 2, GetBaseID(FILE_TYPE_IPB));	// cmp     esi, 5AF0h	; _ZN10CStreaming12MakeSpaceForEi
			CPatch::PatchUINT32(0x531E3D + 2, GetBaseID(FILE_TYPE_IPB));	// cmp     esi, 23280	; _ZN10CStreaming21ConvertBufferToObjectEPcii
		}
		#endif

		// -23280 (-IPB base index)
		#if TRUE
		{
			CPatch::PatchUINT32(0x52EBD2 + 1, -GetBaseID(FILE_TYPE_IPB));	// add     eax, 0FFFFA510h	; getNameByFileID
			CPatch::PatchUINT32(0x52F744 + 2, -GetBaseID(FILE_TYPE_IPB));	// lea     eax, [esi-5AF0h]	; sub_52F640
			CPatch::PatchUINT32(0x52F88B + 2, -GetBaseID(FILE_TYPE_IPB));	// add     esi, 0FFFFA510h	; sub_52F640
			CPatch::PatchUINT32(0x531F09 + 2, -GetBaseID(FILE_TYPE_IPB));	// lea     eax, [esi-5AF0h]	; _ZN10CStreaming21ConvertBufferToObjectEPcii
			CPatch::PatchUINT32(0x5326D9 + 2, -GetBaseID(FILE_TYPE_IPB));	// lea     eax, [edi-5AF0h]	; sub_532510
			CPatch::PatchUINT32(0x5326F1 + 2, -GetBaseID(FILE_TYPE_IPB));	// lea     edx, [edi-5AF0h]	; sub_532510
		}
		#endif

		// 22700 (AGR base index)
		#if TRUE
		{
			CPatch::PatchUINT32(0x52DAB6 + 2, GetBaseID(FILE_TYPE_AGR));	// mov     dword ptr [ecx], 58ACh	; sub_52DA60
			CPatch::PatchUINT32(0x52EB39 + 1, GetBaseID(FILE_TYPE_AGR));	// cmp     eax, 58ACh	; getNameByFileID
			CPatch::PatchUINT32(0x52F6E3 + 2, GetBaseID(FILE_TYPE_AGR));	// cmp     esi, 58ACh	; sub_52F640
			CPatch::PatchUINT32(0x52F815 + 2, GetBaseID(FILE_TYPE_AGR));	// cmp     esi, 58ACh	; sub_52F640
			CPatch::PatchUINT32(0x52FE6D + 2, GetBaseID(FILE_TYPE_AGR));	// cmp     esi, 58ACh	; sub_52FE40
			CPatch::PatchUINT32(0x52FF6D + 2, GetBaseID(FILE_TYPE_AGR));	// cmp     esi, 58ACh	; sub_52FF40
			CPatch::PatchUINT32(0x5303A7 + 2, GetBaseID(FILE_TYPE_AGR));	// cmp     esi, 58ACh	; sub_530100
			CPatch::PatchUINT32(0x5303D9 + 2, GetBaseID(FILE_TYPE_AGR));	// cmp     esi, 58ACh	; sub_530100
			CPatch::PatchUINT32(0x53049E + 2, GetBaseID(FILE_TYPE_AGR));	// cmp     esi, 58ACh	; sub_530100
			CPatch::PatchUINT32(0x531885 + 2, GetBaseID(FILE_TYPE_AGR));	// cmp     esi, 58ACh	; _ZN10CStreaming12MakeSpaceForEi
			CPatch::PatchUINT32(0x5318B3 + 2, GetBaseID(FILE_TYPE_AGR));	// cmp     esi, 58ACh	; _ZN10CStreaming12MakeSpaceForEi
			CPatch::PatchUINT32(0x531DE1 + 2, GetBaseID(FILE_TYPE_AGR));	// cmp     esi, 58ACh	; _ZN10CStreaming21ConvertBufferToObjectEPcii
			CPatch::PatchUINT32(0x53382B + 1, GetBaseID(FILE_TYPE_AGR));	// mov     esi, 58ACh	; CStreaming::Init
		}
		#endif

		// 22700 (-AGR base index)
		#if TRUE
		{
			CPatch::PatchUINT32(0x530D0F + 2, -GetBaseID(FILE_TYPE_AGR));	// lea     edx, [esi-58ACh]	; _ZN10CStreaming20RemoveLeastUsedModelEj
			CPatch::PatchUINT32(0x531FD6 + 2, -GetBaseID(FILE_TYPE_AGR));	// lea     eax, [esi-58ACh]	; _ZN10CStreaming21ConvertBufferToObjectEPcii
			CPatch::PatchUINT32(0x53220B + 2, -GetBaseID(FILE_TYPE_AGR));	// lea     ecx, [edi-58ACh]	; sub_532020
			CPatch::PatchUINT32(0x532D0A + 2, -GetBaseID(FILE_TYPE_AGR));	// lea     eax, [ebp-58ACh]	; sub_532C80
			CPatch::PatchUINT32(0x53401E + 2, -GetBaseID(FILE_TYPE_AGR));	// lea     eax, [edi-58ACh]	; _ZN10CStreaming22LoadAllRequestedModelsEb
		}
		#endif

		// Pointers to pAGRfiles
		#if TRUE
		{
			CPatch::PatchPointer(0x531E49 + 3, CStreaming__ms_apAGRBuffers.bully_se - GetBaseID(FILE_TYPE_AGR));	// cmp     (pAGRfiles-162B0h)[esi*4], 0	; _ZN10CStreaming21ConvertBufferToObjectEPcii
			CPatch::PatchPointer(0x531E63 + 3, CStreaming__ms_apAGRBuffers.bully_se - GetBaseID(FILE_TYPE_AGR));	// mov     (pAGRfiles-162B0h)[esi*4], eax	; _ZN10CStreaming21ConvertBufferToObjectEPcii
			CPatch::PatchPointer(0x532219 + 3, CStreaming__ms_apAGRBuffers.bully_se - GetBaseID(FILE_TYPE_AGR));	// mov     eax, (pAGRfiles-162B0h)[edi*4]	; sub_532020

			CPatch::PatchPointer(0x52D86F + 1, CStreaming__ms_apAGRBuffers.bully_se);	// mov     esi, offset pAGRfiles	; sub_52D820
			CPatch::PatchPointer(0x533830 + 1, CStreaming__ms_apAGRBuffers.bully_se);	// mov     edi, offset pAGRfiles	; CStreaming::Init
			CPatch::PatchPointer(0x53402B + 3, CStreaming__ms_apAGRBuffers.bully_se);	// mov     eax, pAGRfiles[eax*4]	; _ZN10CStreaming22LoadAllRequestedModelsEb
		}
		#endif

		// 22200 (COL base index)
		#if TRUE
		{
			CPatch::PatchUINT32(0x52DFF4 + 1, GetBaseID(FILE_TYPE_COL));	// add     eax, 56B8h	; sub_52DFF0
			CPatch::PatchUINT32(0x52E344 + 1, GetBaseID(FILE_TYPE_COL));	// add     eax, 56B8h	; _ZN10CStreaming15LoadCdDirectoryEPKci
			CPatch::PatchUINT32(0x52EB0E + 1, GetBaseID(FILE_TYPE_COL));	// cmp     eax, 56B8h	; getNameByFileID
			CPatch::PatchUINT32(0x52F6CD + 2, GetBaseID(FILE_TYPE_COL));	// cmp     esi, 56B8h	; sub_52F640
			CPatch::PatchUINT32(0x52F7F6 + 2, GetBaseID(FILE_TYPE_COL));	// cmp     esi, 56B8h	; sub_52F640
			CPatch::PatchUINT32(0x52FE65 + 2, GetBaseID(FILE_TYPE_COL));	// cmp     esi, 56B8h	; sub_52FE40
			CPatch::PatchUINT32(0x52FF65 + 2, GetBaseID(FILE_TYPE_COL));	// cmp     esi, 56B8h	; sub_52FF40
			CPatch::PatchUINT32(0x5303D1 + 2, GetBaseID(FILE_TYPE_COL));	// cmp     esi, 56B8h	; sub_530100
			CPatch::PatchUINT32(0x530445 + 2, GetBaseID(FILE_TYPE_COL));	// cmp     esi, 56B8h	; sub_530100
			CPatch::PatchUINT32(0x530C0E + 2, GetBaseID(FILE_TYPE_COL));	// cmp     edi, 56B8h	; sub_530B50
			CPatch::PatchUINT32(0x530CD0 + 2, GetBaseID(FILE_TYPE_COL));	// cmp     esi, 56B8h	; _ZN10CStreaming20RemoveLeastUsedModelEj
			CPatch::PatchUINT32(0x5318AB + 2, GetBaseID(FILE_TYPE_COL));	// cmp     esi, 56B8h	; _ZN10CStreaming12MakeSpaceForEi
			CPatch::PatchUINT32(0x531D27 + 2, GetBaseID(FILE_TYPE_COL));	// cmp     esi, 56B8h	; _ZN10CStreaming21ConvertBufferToObjectEPcii
			CPatch::PatchUINT32(0x531FCE + 2, GetBaseID(FILE_TYPE_COL));	// cmp     esi, 56B8h	; _ZN10CStreaming21ConvertBufferToObjectEPcii
			CPatch::PatchUINT32(0x53269F + 2, GetBaseID(FILE_TYPE_COL));	// cmp     edi, 56B8h	; sub_532510
			CPatch::PatchUINT32(0x56F2AE + 1, GetBaseID(FILE_TYPE_COL));	// add     eax, 56B8h	; sub_56F260
			CPatch::PatchUINT32(0x56F2F3 + 2, GetBaseID(FILE_TYPE_COL));	// lea     edi, [esi+56B8h]	; sub_56F260
			CPatch::PatchUINT32(0x56F36A + 2, GetBaseID(FILE_TYPE_COL));	// lea     eax, [esi+56B8h]	; sub_56F330
			CPatch::PatchUINT32(0x56FF31 + 2, GetBaseID(FILE_TYPE_COL));	// lea     edx, [edi+56B8h]	; sub_56FE00
			CPatch::PatchUINT32(0x570135 + 2, GetBaseID(FILE_TYPE_COL));	// lea     edx, [ebx+56B8h]	; sub_570080
			CPatch::PatchUINT32(0x5704B1 + 2, GetBaseID(FILE_TYPE_COL));	// lea     ecx, [esi+56B8h]	; sub_570470
			CPatch::PatchUINT32(0x5C4779 + 1, GetBaseID(FILE_TYPE_COL));	// add     eax, 56B8h	; sub_5C46F0
			CPatch::PatchUINT32(0x5C4796 + 2, GetBaseID(FILE_TYPE_COL));	// add     edx, 56B8h	; sub_5C46F0
			CPatch::PatchUINT32(0x5C5995 + 1, GetBaseID(FILE_TYPE_COL));	// add     eax, 56B8h	; sub_5C5870
		}
		#endif

		// -22200 (-COL base index)
		#if TRUE
		{
			CPatch::PatchUINT32(0x52EB40 + 1, -GetBaseID(FILE_TYPE_COL));	// add     eax, 0FFFFA948h	; getNameByFileID
			CPatch::PatchUINT32(0x52F6EB + 2, -GetBaseID(FILE_TYPE_COL));	// lea     ecx, [esi-56B8h]	; sub_52F640
			CPatch::PatchUINT32(0x52F81D + 2, -GetBaseID(FILE_TYPE_COL));	// add     esi, 0FFFFA948h	; sub_52F640
			CPatch::PatchUINT32(0x5304AF + 2, -GetBaseID(FILE_TYPE_COL));	// lea     ecx, [esi-56B8h]	; sub_530100
			CPatch::PatchUINT32(0x531DF1 + 2, -GetBaseID(FILE_TYPE_COL));	// lea     ecx, [esi-56B8h]	; _ZN10CStreaming21ConvertBufferToObjectEPcii
			CPatch::PatchUINT32(0x532234 + 2, -GetBaseID(FILE_TYPE_COL));	// lea     edx, [edi-56B8h]	; sub_532020
			CPatch::PatchUINT32(0x532242 + 2, -GetBaseID(FILE_TYPE_COL));	// lea     eax, [edi-56B8h]	; sub_532020
		}
		#endif

		// 12200 (NFT base index)
		#if TRUE
		{
			CPatch::PatchUINT32(0x437B85 + 2, GetBaseID(FILE_TYPE_NFT));	// cmp     edi, 2FA8h	; sub_4377F0
			CPatch::PatchUINT32(0x464829 + 2, GetBaseID(FILE_TYPE_NFT));	// cmp     esi, 2FA8h	; sub_464800
			CPatch::PatchUINT32(0x4648A9 + 2, GetBaseID(FILE_TYPE_NFT));	// cmp     edi, 2FA8h	; sub_464800
			CPatch::PatchUINT32(0x51C1E3 + 2, GetBaseID(FILE_TYPE_NFT));	// cmp     ecx, 2FA8h	; sub_51C190
			CPatch::PatchUINT32(0x51C379 + 1, GetBaseID(FILE_TYPE_NFT));	// cmp     eax, 2FA8h	; sub_51C330				
			CPatch::PatchUINT32(0x52D794 + 2, GetBaseID(FILE_TYPE_NFT));	// add     ecx, 2FA8h	; sub_52D790
			CPatch::PatchUINT32(0x52E2E1 + 1, GetBaseID(FILE_TYPE_NFT));	// add     eax, 2FA8h	; _ZN10CStreaming15LoadCdDirectoryEPKci
			CPatch::PatchUINT32(0x52E69A + 2, GetBaseID(FILE_TYPE_NFT));	// cmp     esi, 2FA8h	; IsMemoryAvailableForNewFile
			CPatch::PatchUINT32(0x52E8DD + 1, GetBaseID(FILE_TYPE_NFT));	// mov     esi, 2FA8h	; sub_52E8D0
			CPatch::PatchUINT32(0x52E96C + 2, GetBaseID(FILE_TYPE_NFT));	// lea     ebx, [ebp+2FA8h]	; sub_52E8D0
			CPatch::PatchUINT32(0x52E972 + 2, GetBaseID(FILE_TYPE_NFT));	// lea     eax, [esi+2FA8h]	; sub_52E8D0
			CPatch::PatchUINT32(0x52E9CA + 2, GetBaseID(FILE_TYPE_NFT));	// add     esi, 2FA8h	; sub_52E8D0
			CPatch::PatchUINT32(0x52EA23 + 2, GetBaseID(FILE_TYPE_NFT));	// add     esi, 2FA8h	; sub_52E8D0
			CPatch::PatchUINT32(0x52EA4A + 2, GetBaseID(FILE_TYPE_NFT));	// lea     ecx, [esi+2FA8h]	; sub_52E8D0
			CPatch::PatchUINT32(0x52EA83 + 1, GetBaseID(FILE_TYPE_NFT));	// add     eax, 2FA8h	; sub_52E8D0
			CPatch::PatchUINT32(0x52EAB5 + 2, GetBaseID(FILE_TYPE_NFT));	// cmp     esi, 2FA8h	; sub_52E8D0
			CPatch::PatchUINT32(0x52EAE4 + 1, GetBaseID(FILE_TYPE_NFT));	// cmp     eax, 2FA8h	; getNameByFileID
			CPatch::PatchUINT32(0x52F66F + 2, GetBaseID(FILE_TYPE_NFT));	// cmp     esi, 2FA8h	; sub_52F640
			CPatch::PatchUINT32(0x52F7EA + 2, GetBaseID(FILE_TYPE_NFT));	// cmp     esi, 2FA8h	; sub_52F640
			CPatch::PatchUINT32(0x52FAC1 + 1, GetBaseID(FILE_TYPE_NFT));	// cmp     eax, 2FA8h	; sub_52FA90
			CPatch::PatchUINT32(0x52FAF0 + 1, GetBaseID(FILE_TYPE_NFT));	// cmp     eax, 2FA8h	; sub_52FA90
			CPatch::PatchUINT32(0x52FB0D + 1, GetBaseID(FILE_TYPE_NFT));	// cmp     eax, 2FA8h	; sub_52FA90
			CPatch::PatchUINT32(0x52FBAC + 1, GetBaseID(FILE_TYPE_NFT));	// cmp     eax, 2FA8h	; sub_52FB40
			CPatch::PatchUINT32(0x52FC1F + 1, GetBaseID(FILE_TYPE_NFT));	// cmp     eax, 2FA8h	; sub_52FB40
			CPatch::PatchUINT32(0x52FC6C + 1, GetBaseID(FILE_TYPE_NFT));	// cmp     eax, 2FA8h	; sub_52FB40
			CPatch::PatchUINT32(0x52FE5D + 2, GetBaseID(FILE_TYPE_NFT));	// cmp     esi, 2FA8h	; sub_52FE40
			CPatch::PatchUINT32(0x52FEB5 + 2, GetBaseID(FILE_TYPE_NFT));	// cmp     esi, 2FA8h	; sub_52FE40
			CPatch::PatchUINT32(0x52FEC8 + 2, GetBaseID(FILE_TYPE_NFT));	// add     edx, 2FA8h	; sub_52FE40
			CPatch::PatchUINT32(0x52FF5D + 2, GetBaseID(FILE_TYPE_NFT));	// cmp     esi, 2FA8h	; sub_52FF40
			CPatch::PatchUINT32(0x52FFB5 + 2, GetBaseID(FILE_TYPE_NFT));	// cmp     esi, 2FA8h	; sub_52FF40
			CPatch::PatchUINT32(0x52FFC8 + 2, GetBaseID(FILE_TYPE_NFT));	// add     edx, 2FA8h	; sub_52FF40
			CPatch::PatchUINT32(0x530405 + 2, GetBaseID(FILE_TYPE_NFT));	// cmp     esi, 2FA8h	; sub_530100
			CPatch::PatchUINT32(0x530864 + 1, GetBaseID(FILE_TYPE_NFT));	// add     eax, 2FA8h	; sub_530860
			CPatch::PatchUINT32(0x530B36 + 2, GetBaseID(FILE_TYPE_NFT));	// cmp     edi, 2FA8h	; sub_530A70
			CPatch::PatchUINT32(0x530BDA + 2, GetBaseID(FILE_TYPE_NFT));	// cmp     edi, 2FA8h	; sub_530B50
			CPatch::PatchUINT32(0x530C16 + 2, GetBaseID(FILE_TYPE_NFT));	// sub     edi, 2FA8h	; sub_530B50
			CPatch::PatchUINT32(0x530C36 + 2, GetBaseID(FILE_TYPE_NFT));	// add     edi, 2FA8h	; sub_530B50
			CPatch::PatchUINT32(0x530CB5 + 2, GetBaseID(FILE_TYPE_NFT));	// cmp     esi, 2FA8h	; _ZN10CStreaming20RemoveLeastUsedModelEj
			CPatch::PatchUINT32(0x530CD8 + 2, GetBaseID(FILE_TYPE_NFT));	// sub     esi, 2FA8h	; _ZN10CStreaming20RemoveLeastUsedModelEj
			CPatch::PatchUINT32(0x530CF8 + 2, GetBaseID(FILE_TYPE_NFT));	// add     esi, 2FA8h	; _ZN10CStreaming20RemoveLeastUsedModelEj
			CPatch::PatchUINT32(0x531C6D + 2, GetBaseID(FILE_TYPE_NFT));	// add     ecx, 2FA8h	; sub_531950
			CPatch::PatchUINT32(0x531CEF + 2, GetBaseID(FILE_TYPE_NFT));	// cmp     esi, 2FA8h	; _ZN10CStreaming21ConvertBufferToObjectEPcii
			CPatch::PatchUINT32(0x531F8D + 2, GetBaseID(FILE_TYPE_NFT));	// cmp     esi, 2FA8h	; _ZN10CStreaming21ConvertBufferToObjectEPcii
			CPatch::PatchUINT32(0x532077 + 2, GetBaseID(FILE_TYPE_NFT));	// cmp     edi, 2FA8h	; sub_532020
			CPatch::PatchUINT32(0x532114 + 2, GetBaseID(FILE_TYPE_NFT));	// cmp     edi, 2FA8h	; sub_532020
			CPatch::PatchUINT32(0x532538 + 2, GetBaseID(FILE_TYPE_NFT));	// cmp     edi, 2FA8h	; sub_532510
			CPatch::PatchUINT32(0x532579 + 2, GetBaseID(FILE_TYPE_NFT));	// cmp     edi, 2FA8h	; sub_532510
			CPatch::PatchUINT32(0x532878 + 2, GetBaseID(FILE_TYPE_NFT));	// cmp     esi, 2FA8h	; sub_532750
			CPatch::PatchUINT32(0x532A83 + 2, GetBaseID(FILE_TYPE_NFT));	// cmp     ebx, 2FA8h	; sub_532A10
			CPatch::PatchUINT32(0x532A9A + 1, GetBaseID(FILE_TYPE_NFT));	// add     eax, 2FA8h	; sub_532A10
			CPatch::PatchUINT32(0x532ADF + 2, GetBaseID(FILE_TYPE_NFT));	// add     ecx, 2FA8h	; sub_532A10
			CPatch::PatchUINT32(0x532E21 + 2, GetBaseID(FILE_TYPE_NFT));	// cmp     ebp, 2FA8h	; sub_532C80
			CPatch::PatchUINT32(0x532E49 + 1, GetBaseID(FILE_TYPE_NFT));	// add     eax, 2FA8h	; sub_532C80
			CPatch::PatchUINT32(0x532F0F + 2, GetBaseID(FILE_TYPE_NFT));	// cmp     ebp, 2FA8h	; sub_532C80
			CPatch::PatchUINT32(0x5334DA + 1, GetBaseID(FILE_TYPE_NFT));	// push    2FA8h	; CStreaming::Init
			CPatch::PatchUINT32(0x533652 + 2, GetBaseID(FILE_TYPE_NFT));	// lea     eax, [esi+2FA8h]	; CStreaming::Init
			CPatch::PatchUINT32(0x5338D4 + 2, GetBaseID(FILE_TYPE_NFT));	// cmp     esi, 2FA8h	; sub_533880
			CPatch::PatchUINT32(0x5339BB + 2, GetBaseID(FILE_TYPE_NFT));	// cmp     esi, 2FA8h	; sub_533980
			CPatch::PatchUINT32(0x5340F1 + 2, GetBaseID(FILE_TYPE_NFT));	// cmp     edi, 2FA8h	; _ZN10CStreaming22LoadAllRequestedModelsEb
			CPatch::PatchUINT32(0x534949 + 2, GetBaseID(FILE_TYPE_NFT));	// cmp     edi, 2FA8h	; sub_5348D0
			CPatch::PatchUINT32(0x534996 + 2, GetBaseID(FILE_TYPE_NFT));	// lea     eax, [edi+2FA8h]	; sub_5348D0
			CPatch::PatchUINT32(0x5349FA + 2, GetBaseID(FILE_TYPE_NFT));	// cmp     ebp, 2FA8h	; sub_5348D0
			CPatch::PatchUINT32(0x5391DD + 1, GetBaseID(FILE_TYPE_NFT));	// add     eax, 2FA8h	; sub_5391A0
			CPatch::PatchUINT32(0x5391F4 + 1, GetBaseID(FILE_TYPE_NFT));	// add     eax, 2FA8h	; sub_5391A0
			CPatch::PatchUINT32(0x53936A + 1, GetBaseID(FILE_TYPE_NFT));	// add     eax, 2FA8h	; sub_539240
			CPatch::PatchUINT32(0x539754 + 1, GetBaseID(FILE_TYPE_NFT));	// add     eax, 2FA8h	; sub_5396D0
			CPatch::PatchUINT32(0x539A6B + 1, GetBaseID(FILE_TYPE_NFT));	// add     eax, 2FA8h	; sub_539A00
			CPatch::PatchUINT32(0x539E3E + 1, GetBaseID(FILE_TYPE_NFT));	// add     eax, 2FA8h	; sub_539E20
			CPatch::PatchUINT32(0x549308 + 2, GetBaseID(FILE_TYPE_NFT));	// add     ecx, 2FA8h	; sub_5492E0
			CPatch::PatchUINT32(0x549336 + 2, GetBaseID(FILE_TYPE_NFT));	// add     esi, 2FA8h	; sub_5492E0
			CPatch::PatchUINT32(0x5497CD + 1, GetBaseID(FILE_TYPE_NFT));	// add     eax, 2FA8h	; sub_549740
			CPatch::PatchUINT32(0x54D265 + 2, GetBaseID(FILE_TYPE_NFT));	// lea     edx, [esi+2FA8h]	; sub_54D1F0
			CPatch::PatchUINT32(0x551223 + 2, GetBaseID(FILE_TYPE_NFT));	// add     edi, 2FA8h	; sub_550EC0
			CPatch::PatchUINT32(0x5F1194 + 2, GetBaseID(FILE_TYPE_NFT));	// add     edx, 2FA8h	; sub_5F1160
			CPatch::PatchUINT32(0x6A07FB + 2, GetBaseID(FILE_TYPE_NFT));	// lea     eax, [edi+2FA8h]	; sub_6A07D0
			CPatch::PatchUINT32(0x6CCE00 + 1, GetBaseID(FILE_TYPE_NFT));	// add     eax, 2FA8h	; sub_6CCDB0
			CPatch::PatchUINT32(0x6CD01C + 1, GetBaseID(FILE_TYPE_NFT));	// add     eax, 2FA8h	; sub_6CCFD0
			CPatch::PatchUINT32(0x6CEF38 + 1, GetBaseID(FILE_TYPE_NFT));	// add     eax, 2FA8h	; sub_6CEF10
			CPatch::PatchUINT32(0x6CF05A + 1, GetBaseID(FILE_TYPE_NFT));	// add     eax, 2FA8h	; sub_6CF020
			CPatch::PatchUINT32(0x6CFFAF + 1, GetBaseID(FILE_TYPE_NFT));	// add     eax, 2FA8h	; sub_6CFF80
			CPatch::PatchUINT32(0x6D00EC + 1, GetBaseID(FILE_TYPE_NFT));	// add     eax, 2FA8h	; sub_6D00A0
			CPatch::PatchUINT32(0x6D10C0 + 2, GetBaseID(FILE_TYPE_NFT));	// lea     edi, [esi+2FA8h]	; sub_6D10A0
			CPatch::PatchUINT32(0x6E1046 + 2, GetBaseID(FILE_TYPE_NFT));	// add     edx, 2FA8h	; sub_6E1010
			CPatch::PatchUINT32(0x6E1685 + 1, GetBaseID(FILE_TYPE_NFT));	// add     eax, 2FA8h	; sub_6E15F0
			CPatch::PatchUINT32(0x6E38F5 + 1, GetBaseID(FILE_TYPE_NFT));	// add     eax, 2FA8h	; sub_6E38B0
			CPatch::PatchUINT32(0x6E396B + 1, GetBaseID(FILE_TYPE_NFT));	// add     eax, 2FA8h	; sub_6E38B0
			CPatch::PatchUINT32(0x6E39F4 + 1, GetBaseID(FILE_TYPE_NFT));	// add     eax, 2FA8h	; sub_6E38B0
			CPatch::PatchUINT32(0x6E3AC2 + 1, GetBaseID(FILE_TYPE_NFT));	// add     eax, 2FA8h	; sub_6E3A70
			CPatch::PatchUINT32(0x6E3CB7 + 1, GetBaseID(FILE_TYPE_NFT));	// add     eax, 2FA8h	; sub_6E3AF0
			CPatch::PatchUINT32(0x6E3DAF + 1, GetBaseID(FILE_TYPE_NFT));	// add     eax, 2FA8h	; sub_6E3D60
			CPatch::PatchUINT32(0x6EEF49 + 1, GetBaseID(FILE_TYPE_NFT));	// add     eax, 2FA8h	; sub_6EE6A0
			CPatch::PatchUINT32(0x6EF7EB + 1, GetBaseID(FILE_TYPE_NFT));	// add     eax, 2FA8h	; sub_6EEFD0
			CPatch::PatchUINT32(0x6F008D + 1, GetBaseID(FILE_TYPE_NFT));	// add     eax, 2FA8h	; sub_6EF870
			CPatch::PatchUINT32(0x6F09B2 + 1, GetBaseID(FILE_TYPE_NFT));	// add     eax, 2FA8h	; sub_6F0110
			CPatch::PatchUINT32(0x6F12DE + 1, GetBaseID(FILE_TYPE_NFT));	// add     eax, 2FA8h	; sub_6F0A40
			CPatch::PatchUINT32(0x6F1A73 + 1, GetBaseID(FILE_TYPE_NFT));	// add     eax, 2FA8h	; sub_6F1360
			CPatch::PatchUINT32(0x6F232D + 1, GetBaseID(FILE_TYPE_NFT));	// add     eax, 2FA8h	; sub_6F1B00
			CPatch::PatchUINT32(0x6F2837 + 1, GetBaseID(FILE_TYPE_NFT));	// add     eax, 2FA8h	; sub_6F23B0
			CPatch::PatchUINT32(0x6F2D9F + 1, GetBaseID(FILE_TYPE_NFT));	// add     eax, 2FA8h	; sub_6F28C0
			CPatch::PatchUINT32(0x704D46 + 1, GetBaseID(FILE_TYPE_NFT));	// add     eax, 2FA8h	; sub_704D30
			CPatch::PatchUINT32(0x704E14 + 1, GetBaseID(FILE_TYPE_NFT));	// add     eax, 2FA8h	; sub_704DE0
			CPatch::PatchUINT32(0x704E72 + 2, GetBaseID(FILE_TYPE_NFT));	// add     ecx, 2FA8h	; sub_704E40
			CPatch::PatchUINT32(0x70535A + 1, GetBaseID(FILE_TYPE_NFT));	// add     eax, 2FA8h	; sub_705340
			CPatch::PatchUINT32(0x7057EB + 2, GetBaseID(FILE_TYPE_NFT));	// lea     eax, [esi+2FA8h]	; sub_7057B0
			CPatch::PatchUINT32(0x705883 + 2, GetBaseID(FILE_TYPE_NFT));	// add     ebx, 2FA8h	; sub_705850
			CPatch::PatchUINT32(0x70753D + 1, GetBaseID(FILE_TYPE_NFT));	// add     eax, 2FA8h	; sub_707510
			CPatch::PatchUINT32(0x72C1AB + 1, GetBaseID(FILE_TYPE_NFT));	// add     eax, 2FA8h	; sub_72C180
			CPatch::PatchUINT32(0x72C1CE + 1, GetBaseID(FILE_TYPE_NFT));	// add     eax, 2FA8h	; sub_72C180
			CPatch::PatchUINT32(0x72DF75 + 2, GetBaseID(FILE_TYPE_NFT));	// add     ebx, 2FA8h	; sub_72DDF0
			CPatch::PatchUINT32(0x72DFE9 + 1, GetBaseID(FILE_TYPE_NFT));	// add     eax, 2FA8h	; sub_72DDF0
			CPatch::PatchUINT32(0x72E108 + 2, GetBaseID(FILE_TYPE_NFT));	// lea     esi, [ecx+2FA8h]	; sub_72E090
			CPatch::PatchUINT32(0x72E133 + 2, GetBaseID(FILE_TYPE_NFT));	// add     edx, 2FA8h	; sub_72E090
			CPatch::PatchUINT32(0x72E2DF + 2, GetBaseID(FILE_TYPE_NFT));	// lea     edi, [ecx+2FA8h]	; sub_72E090
			CPatch::PatchUINT32(0x72E306 + 2, GetBaseID(FILE_TYPE_NFT));	// add     edx, 2FA8h	; sub_72E090
		}
		#endif

		// -12200 (-NFT base index)
		#if TRUE
		{
			CPatch::PatchUINT32(0x52E65A + 2, -GetBaseID(FILE_TYPE_NFT));	// lea     ecx, [esi-12200]	; IsMemoryAvailableForNewFile
			CPatch::PatchUINT32(0x52E6B8 + 2, -GetBaseID(FILE_TYPE_NFT));	// add     esi, 0FFFFD058h	; IsMemoryAvailableForNewFile
			CPatch::PatchUINT32(0x52EB15 + 1, -GetBaseID(FILE_TYPE_NFT));	// add     eax, 0FFFFD058h	; getNameByFileID
			CPatch::PatchUINT32(0x52F6D5 + 2, -GetBaseID(FILE_TYPE_NFT));	// lea     eax, [esi-2FA8h]	; sub_52F640
			CPatch::PatchUINT32(0x52F7FE + 2, -GetBaseID(FILE_TYPE_NFT));	// add     esi, 0FFFFD058h	; sub_52F640
			CPatch::PatchUINT32(0x530452 + 2, -GetBaseID(FILE_TYPE_NFT));	// lea     ebx, [esi-2FA8h]	; sub_530100
			CPatch::PatchUINT32(0x53074B + 2, -GetBaseID(FILE_TYPE_NFT));	// add     esi, 0FFFFD058h	; sub_530100
			CPatch::PatchUINT32(0x531D39 + 2, -GetBaseID(FILE_TYPE_NFT));	// lea     edx, [esi-2FA8h]	; _ZN10CStreaming21ConvertBufferToObjectEPcii
			CPatch::PatchUINT32(0x531D7A + 2, -GetBaseID(FILE_TYPE_NFT));	// lea     ebp, [esi-2FA8h]	; _ZN10CStreaming21ConvertBufferToObjectEPcii
			CPatch::PatchUINT32(0x5326A7 + 2, -GetBaseID(FILE_TYPE_NFT));	// lea     esi, [edi-2FA8h]	; sub_532510
			CPatch::PatchUINT32(0x532CF0 + 2, -GetBaseID(FILE_TYPE_NFT));	// lea     eax, [ebp-2FA8h]	; sub_532C80
		}
		#endif

		// others
		#if TRUE
		{
			CPatch::PatchUINT32(0x52E901 + 2, GetBaseID(FILE_TYPE_NFT) + GetFileIDdefaultLimit(FILE_TYPE_NFT));	// cmp     esi, 56B8h	; sub_52E8D0
			CPatch::PatchUINT32(0x52EA50 + 2, GetBaseID(FILE_TYPE_NFT) + GetFileIDdefaultLimit(FILE_TYPE_NFT));	// cmp     ecx, 56B8h	; sub_52E8D0
		}
		#endif

	}
	#endif

	// Patch limits itself
	#if TRUE
	{
		// NIF limit
		#if TRUE
		{
			// Patch references to CModelInfo::ms_modelInfoPtrs
			{
				CPatch::PatchPointer(0x4C43CC + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ebp, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4] ; CModelInfo::ms_modelInfoPtrs

				CPatch::PatchPointer(0x40B181 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs	; sub_40AD80
				CPatch::PatchPointer(0x41672C + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_416690
				CPatch::PatchPointer(0x41698A + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4168A0
				CPatch::PatchPointer(0x41C55B + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_41C540
				CPatch::PatchPointer(0x41C616 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_41C5F0
				CPatch::PatchPointer(0x41C8C2 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs	; sub_41C5F0
				CPatch::PatchPointer(0x41CCB4 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_41CC70
				CPatch::PatchPointer(0x41CDB6 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_41CC70
				CPatch::PatchPointer(0x41D379 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_41D260
				CPatch::PatchPointer(0x41D384 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_41D260
				CPatch::PatchPointer(0x41D756 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_41D6D0
				CPatch::PatchPointer(0x41DA27 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_41D930
				CPatch::PatchPointer(0x41F4A8 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_41F400
				CPatch::PatchPointer(0x41F639 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_41F5A0
				CPatch::PatchPointer(0x41F7D9 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_41F740
				CPatch::PatchPointer(0x41FFB9 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_41FF70
				CPatch::PatchPointer(0x420583 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_41FF70
				CPatch::PatchPointer(0x4205C0 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_41FF70
				CPatch::PatchPointer(0x4207E1 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_420740
				CPatch::PatchPointer(0x421685 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs	; sub_421370
				CPatch::PatchPointer(0x421745 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs	; sub_421370
				CPatch::PatchPointer(0x421858 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_421370
				CPatch::PatchPointer(0x423544 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_423460
				CPatch::PatchPointer(0x42356D + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_423460
				CPatch::PatchPointer(0x423C90 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_423460
				CPatch::PatchPointer(0x423CB9 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_423460
				CPatch::PatchPointer(0x42B8DC + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs	; sub_42B760
				CPatch::PatchPointer(0x42BA1F + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ebp, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_42B9D0
				CPatch::PatchPointer(0x42C4F8 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_42C120
				CPatch::PatchPointer(0x42C7A1 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ebp, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; _ZN11CFileLoader17LoadCollisionFileEPhjh
				CPatch::PatchPointer(0x42C9BF + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ebx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; _ZN11CFileLoader26LoadCollisionFileFirstTimeEPhjh
				CPatch::PatchPointer(0x4302CD + 1, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     esi, offset _ZN10CModelInfo16ms_modelInfoPtrsE; CModelInfo::ms_modelInfoPtrs	; sub_4302A0
				CPatch::PatchPointer(0x430300 + 2, CModelInfo__ms_modelInfoPtrs.bully_se + GetFileIDcurrentLimit(FILE_TYPE_NIF));	// cmp     esi, offset dword_C735D8	; sub_4302A0
				CPatch::PatchPointer(0x4363B1 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_436310
				CPatch::PatchPointer(0x436C6B + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_436C40
				CPatch::PatchPointer(0x436F7F + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_436F50
				CPatch::PatchPointer(0x43722D + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_437220
				CPatch::PatchPointer(0x43770E + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// cmp     _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4], 0; CModelInfo::ms_modelInfoPtrs	; sub_4373F0
				CPatch::PatchPointer(0x43772D + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4373F0
				CPatch::PatchPointer(0x437B44 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_4377F0
				CPatch::PatchPointer(0x437C97 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_437C90
				CPatch::PatchPointer(0x4388F3 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ebp, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4388E0
				CPatch::PatchPointer(0x438A3B + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4388E0
				CPatch::PatchPointer(0x438A5D + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs	; sub_4388E0
				CPatch::PatchPointer(0x438AB1 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_4388E0
				CPatch::PatchPointer(0x439020 + 2, CModelInfo__ms_modelInfoPtrs.bully_se + 10082);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE+9D88h	; sub_438F50
				CPatch::PatchPointer(0x44490A + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4448F0
				CPatch::PatchPointer(0x445608 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_445540
				CPatch::PatchPointer(0x4456C9 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_445540
				CPatch::PatchPointer(0x44573C + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_445540
				CPatch::PatchPointer(0x44CD3A + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_44CD30
				CPatch::PatchPointer(0x44CF3F + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_44CE70
				CPatch::PatchPointer(0x44CFA1 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_44CE70
				CPatch::PatchPointer(0x44ECA2 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_44EAE0
				CPatch::PatchPointer(0x44ED56 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_44EAE0
				CPatch::PatchPointer(0x44EDAE + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_44EAE0
				CPatch::PatchPointer(0x44EE2E + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_44EAE0
				CPatch::PatchPointer(0x44EE98 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_44EAE0
				CPatch::PatchPointer(0x451671 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_451660
				CPatch::PatchPointer(0x4517C2 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_451660
				CPatch::PatchPointer(0x4517FD + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_451660
				CPatch::PatchPointer(0x45182C + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ebx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_451660
				CPatch::PatchPointer(0x451EB0 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_451E70
				CPatch::PatchPointer(0x452010 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_452000
				CPatch::PatchPointer(0x4520EA + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_452000
				CPatch::PatchPointer(0x452115 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_452000
				CPatch::PatchPointer(0x4521F0 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs	; sub_452000
				CPatch::PatchPointer(0x452CAA + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_452C70
				CPatch::PatchPointer(0x454570 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_454540
				CPatch::PatchPointer(0x456ABA + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_456A70
				CPatch::PatchPointer(0x456AD3 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_456A70
				CPatch::PatchPointer(0x456B33 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_456A70
				CPatch::PatchPointer(0x456BD7 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_456A70
				CPatch::PatchPointer(0x456C5E + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_456A70
				CPatch::PatchPointer(0x45EB8B + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_45EAE0
				CPatch::PatchPointer(0x45EE9B + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_45EE90
				CPatch::PatchPointer(0x45EF7F + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_45EEC0
				CPatch::PatchPointer(0x45EFE6 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_45EEC0
				CPatch::PatchPointer(0x45F15C + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_45F0E0
				CPatch::PatchPointer(0x45F242 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_45F230
				CPatch::PatchPointer(0x45F7BD + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_45F670
				CPatch::PatchPointer(0x45F81B + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_45F670
				CPatch::PatchPointer(0x45F9CC + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_45F940
				CPatch::PatchPointer(0x4602D3 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ebx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_45FAB0
				CPatch::PatchPointer(0x46040D + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_45FAB0
				CPatch::PatchPointer(0x4606EC + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_460670
				CPatch::PatchPointer(0x460ACF + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4608F0
				CPatch::PatchPointer(0x460B1D + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4608F0
				CPatch::PatchPointer(0x462BA2 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_462B20
				CPatch::PatchPointer(0x462F3F + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_462B20
				CPatch::PatchPointer(0x4637E7 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4637D0
				CPatch::PatchPointer(0x464908 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4648C0
				CPatch::PatchPointer(0x4649E2 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_4648C0
				CPatch::PatchPointer(0x464B4A + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_464B30
				CPatch::PatchPointer(0x464BB3 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_464B70
				CPatch::PatchPointer(0x464C32 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_464C20
				CPatch::PatchPointer(0x464C86 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_464C50
				CPatch::PatchPointer(0x464D0E + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ebx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_464C50
				CPatch::PatchPointer(0x464D9B + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_464C50
				CPatch::PatchPointer(0x464FB3 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_464FA0
				CPatch::PatchPointer(0x46626E + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_466260
				CPatch::PatchPointer(0x4662FE + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ebx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4662F0
				CPatch::PatchPointer(0x4663B7 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_4662F0
				CPatch::PatchPointer(0x46670A + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4666F0
				CPatch::PatchPointer(0x466750 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_466740
				CPatch::PatchPointer(0x4667D5 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4667A0
				CPatch::PatchPointer(0x466801 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_4667A0
				CPatch::PatchPointer(0x466868 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_4667A0
				CPatch::PatchPointer(0x4668B6 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_4667A0
				CPatch::PatchPointer(0x466CD3 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_466CC0
				CPatch::PatchPointer(0x466D78 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_466D40
				CPatch::PatchPointer(0x466F01 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_466D40
				CPatch::PatchPointer(0x467BA1 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_467B80
				CPatch::PatchPointer(0x467BC7 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_467BC0
				CPatch::PatchPointer(0x467C0E + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_467C00
				CPatch::PatchPointer(0x467E02 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_467D80
				CPatch::PatchPointer(0x467E43 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_467E30
				CPatch::PatchPointer(0x468335 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_468190
				CPatch::PatchPointer(0x46AB05 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_46AAF0
				CPatch::PatchPointer(0x46B5F6 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_46AEE0
				CPatch::PatchPointer(0x46B725 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_46AEE0
				CPatch::PatchPointer(0x46C799 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_46C600
				CPatch::PatchPointer(0x471856 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// cmp     _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4], 0; CModelInfo::ms_modelInfoPtrs	; sub_471810
				CPatch::PatchPointer(0x477DC7 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_477DC0
				CPatch::PatchPointer(0x478A65 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ebp, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_478A20
				CPatch::PatchPointer(0x478E50 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_478A20
				CPatch::PatchPointer(0x478F9D + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_478F80
				CPatch::PatchPointer(0x47AC1C + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_47A9E0
				CPatch::PatchPointer(0x47AD23 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_47ACC0
				CPatch::PatchPointer(0x47AF1F + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_47AE30
				CPatch::PatchPointer(0x47B20D + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_47B0B0
				CPatch::PatchPointer(0x47C9AE + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_47C820
				CPatch::PatchPointer(0x47CF17 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_47CF10
				CPatch::PatchPointer(0x47CF53 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_47CF30
				CPatch::PatchPointer(0x47D063 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_47D030
				CPatch::PatchPointer(0x47F505 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_47F0A0
				CPatch::PatchPointer(0x480995 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs	; sub_480130
				CPatch::PatchPointer(0x480DEA + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ebp, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_480CC0
				CPatch::PatchPointer(0x481799 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_481420
				CPatch::PatchPointer(0x481C1A + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_481420
				CPatch::PatchPointer(0x48392A + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_483690
				CPatch::PatchPointer(0x48A867 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_48A700
				CPatch::PatchPointer(0x48B62F + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_48B510
				CPatch::PatchPointer(0x48BE42 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_48BAC0
				CPatch::PatchPointer(0x4976E6 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_497570
				CPatch::PatchPointer(0x497B41 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_497B00
				CPatch::PatchPointer(0x497E17 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_497E10
				CPatch::PatchPointer(0x497E38 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_497E30
				CPatch::PatchPointer(0x498001 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_497FE0
				CPatch::PatchPointer(0x4980BF + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_497FE0
				CPatch::PatchPointer(0x4988E2 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_497FE0
				CPatch::PatchPointer(0x498938 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_497FE0
				CPatch::PatchPointer(0x498B54 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_498B40
				CPatch::PatchPointer(0x49BBD4 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_49BBD0
				CPatch::PatchPointer(0x49BE79 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_49BC30
				CPatch::PatchPointer(0x49E4F0 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_49E4E0
				CPatch::PatchPointer(0x49E728 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_49E4E0
				CPatch::PatchPointer(0x49EE54 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_49EE40
				CPatch::PatchPointer(0x49EE88 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_49EE70
				CPatch::PatchPointer(0x49F108 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs	; sub_49F0C0
				CPatch::PatchPointer(0x49F70E + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_49F6C0
				CPatch::PatchPointer(0x49FE05 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_49FDF0
				CPatch::PatchPointer(0x49FE55 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_49FE40
				CPatch::PatchPointer(0x4A021C + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_4A00B0
				CPatch::PatchPointer(0x4A0866 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4A0790
				CPatch::PatchPointer(0x4A092F + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_4A0790
				CPatch::PatchPointer(0x4A6FF5 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_4A6FA0
				CPatch::PatchPointer(0x4A704E + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4A7040
				CPatch::PatchPointer(0x4AAB15 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ebp, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4AAB00
				CPatch::PatchPointer(0x4AB2C4 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4AB100
				CPatch::PatchPointer(0x4AB368 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4AB100
				CPatch::PatchPointer(0x4AC1C8 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4AC110
				CPatch::PatchPointer(0x4AC83F + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ebp, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4AC830
				CPatch::PatchPointer(0x4ACCE7 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4ACCD0
				CPatch::PatchPointer(0x4AD8F9 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4AD740
				CPatch::PatchPointer(0x4AE514 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_4AE430
				CPatch::PatchPointer(0x4B2934 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4B2540
				CPatch::PatchPointer(0x4B2DBA + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_4B2540
				CPatch::PatchPointer(0x4B357B + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_4B2540
				CPatch::PatchPointer(0x4B3E3D + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_4B2540
				CPatch::PatchPointer(0x4B4900 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4B2540
				CPatch::PatchPointer(0x4B5232 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4B2540
				CPatch::PatchPointer(0x4B6432 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_4B63C0
				CPatch::PatchPointer(0x4B6CD0 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4B6C50
				CPatch::PatchPointer(0x4B6E58 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_4B6C50
				CPatch::PatchPointer(0x4B6FE4 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4B6F80
				CPatch::PatchPointer(0x4B7D55 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4B7B80
				CPatch::PatchPointer(0x4B82A4 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_4B8110
				CPatch::PatchPointer(0x4B8453 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4B8110
				CPatch::PatchPointer(0x4B86DB + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_4B85F0
				CPatch::PatchPointer(0x4BD86F + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4BD850
				CPatch::PatchPointer(0x4BE09A + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_4BDF80
				CPatch::PatchPointer(0x4BE0B7 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_4BDF80
				CPatch::PatchPointer(0x4BE0ED + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4BE0E0
				CPatch::PatchPointer(0x4BF6B7 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4BF6B0
				CPatch::PatchPointer(0x4BFD16 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_4BFCC0
				CPatch::PatchPointer(0x4BFE49 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4BFD90
				CPatch::PatchPointer(0x4C0288 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_4C0200
				CPatch::PatchPointer(0x4C036A + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4C0350
				CPatch::PatchPointer(0x4C03F5 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4C0350
				CPatch::PatchPointer(0x4C0BF9 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4C09E0
				CPatch::PatchPointer(0x4C0EC7 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4C0EC0
				CPatch::PatchPointer(0x4C0FF4 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4C0EF0
				CPatch::PatchPointer(0x4C179C + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4C1730
				CPatch::PatchPointer(0x4C2424 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4C22D0
				CPatch::PatchPointer(0x4C2D14 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_4C22D0
				CPatch::PatchPointer(0x4C31F4 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_4C22D0
				CPatch::PatchPointer(0x4C3944 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_4C22D0
				CPatch::PatchPointer(0x4C3CAA + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_4C3C30
				CPatch::PatchPointer(0x4C5112 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_4C4E90
				CPatch::PatchPointer(0x4C512F + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_4C4E90
				CPatch::PatchPointer(0x4C51FB + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_4C4E90
				CPatch::PatchPointer(0x4C54E7 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_4C54A0
				CPatch::PatchPointer(0x4C6716 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_4C6660
				CPatch::PatchPointer(0x4C759D + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4C7590
				CPatch::PatchPointer(0x4C761F + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4C7590
				CPatch::PatchPointer(0x4C77C2 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_4C7700
				CPatch::PatchPointer(0x4C83F5 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4C83E0
				CPatch::PatchPointer(0x4C96CF + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_4C9680
				CPatch::PatchPointer(0x4CC831 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_4CC4A0
				CPatch::PatchPointer(0x4CCA1D + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4CCA10
				CPatch::PatchPointer(0x4CCBBA + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4CCBB0
				CPatch::PatchPointer(0x4CD625 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4CD2A0
				CPatch::PatchPointer(0x4CEB09 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_4CEAA0
				CPatch::PatchPointer(0x4CFD69 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_4CFCE0
				CPatch::PatchPointer(0x4D00CD + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_4D0040
				CPatch::PatchPointer(0x4D0CA4 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4D0CA0
				CPatch::PatchPointer(0x4D0CD4 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4D0CD0
				CPatch::PatchPointer(0x4D0D1E + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_4D0D10
				CPatch::PatchPointer(0x4D0D49 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_4D0D10
				CPatch::PatchPointer(0x4D0D64 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_4D0D10
				CPatch::PatchPointer(0x4D0D80 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_4D0D10
				CPatch::PatchPointer(0x4D0D9C + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_4D0D10
				CPatch::PatchPointer(0x4D0DB8 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_4D0D10
				CPatch::PatchPointer(0x4D234D + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_4D22B0
				CPatch::PatchPointer(0x4D4B41 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs	; sub_4D4AF0
				CPatch::PatchPointer(0x4D4DE9 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs	; sub_4D4D90
				CPatch::PatchPointer(0x4D8900 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_4D84C0
				CPatch::PatchPointer(0x4DA430 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_4DA420
				CPatch::PatchPointer(0x4DA460 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_4DA420
				CPatch::PatchPointer(0x4DA691 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_4DA680
				CPatch::PatchPointer(0x4DA6D7 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4DA680
				CPatch::PatchPointer(0x4DA71C + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_4DA680
				CPatch::PatchPointer(0x4E9E5B + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_4E9DF0
				CPatch::PatchPointer(0x4E9F48 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_4E9EF0
				CPatch::PatchPointer(0x4E9FDF + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4E9FD0
				CPatch::PatchPointer(0x4EB2E5 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_4EB200
				CPatch::PatchPointer(0x4EB67C + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4EB5C0
				CPatch::PatchPointer(0x4EB7EE + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_4EB7E0
				CPatch::PatchPointer(0x4ED04A + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4ED040
				CPatch::PatchPointer(0x4ED078 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4ED070
				CPatch::PatchPointer(0x4ED0BC + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ebp, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4ED0B0
				CPatch::PatchPointer(0x4ED42A + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4ED410
				CPatch::PatchPointer(0x4ED51E + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4ED4E0
				CPatch::PatchPointer(0x4ED90D + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4ED8E0
				CPatch::PatchPointer(0x4EDBF0 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_4EDB20
				CPatch::PatchPointer(0x4EDD90 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4EDD80
				CPatch::PatchPointer(0x4EDDD8 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_4EDD80
				CPatch::PatchPointer(0x4EDEC9 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4EDEB0
				CPatch::PatchPointer(0x4EE137 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_4EE0D0
				CPatch::PatchPointer(0x4EE6A9 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4EE690
				CPatch::PatchPointer(0x4EE6C4 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_4EE690
				CPatch::PatchPointer(0x4EE9AE + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_4EE950
				CPatch::PatchPointer(0x4EEA27 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_4EE950
				CPatch::PatchPointer(0x4EEAEC + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_4EE950
				CPatch::PatchPointer(0x4EEBAB + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4EEBA0
				CPatch::PatchPointer(0x4EEE56 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_4EED10
				CPatch::PatchPointer(0x4EEF3C + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_4EEEC0
				CPatch::PatchPointer(0x4EEFDC + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ebp, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4EEFA0
				CPatch::PatchPointer(0x4EF2AB + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4EF2A0
				CPatch::PatchPointer(0x4EF35A + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_4EF340
				CPatch::PatchPointer(0x4EF408 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4EF400
				CPatch::PatchPointer(0x4EF4BE + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4EF400
				CPatch::PatchPointer(0x4EF5E5 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4EF570
				CPatch::PatchPointer(0x4EF610 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_4EF570
				CPatch::PatchPointer(0x4EF64C + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ebp, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4EF630
				CPatch::PatchPointer(0x4EF67E + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_4EF630
				CPatch::PatchPointer(0x4EFF7F + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_4EF630
				CPatch::PatchPointer(0x4EFFEB + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_4EFFB0
				CPatch::PatchPointer(0x4F01D8 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_4EFFB0
				CPatch::PatchPointer(0x512767 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_512760
				CPatch::PatchPointer(0x5127E1 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_512790
				CPatch::PatchPointer(0x5137B4 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_513780
				CPatch::PatchPointer(0x513888 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_513780
				CPatch::PatchPointer(0x51A964 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_51A960
				CPatch::PatchPointer(0x51AE31 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_51AE20
				CPatch::PatchPointer(0x51AF73 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_51AF50
				CPatch::PatchPointer(0x51AF8B + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_51AF50
				CPatch::PatchPointer(0x51AFA4 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_51AFA0
				CPatch::PatchPointer(0x51B365 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_51B360
				CPatch::PatchPointer(0x51B3C4 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_51B3B0
				CPatch::PatchPointer(0x51B404 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_51B3F0
				CPatch::PatchPointer(0x51B434 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_51B420
				CPatch::PatchPointer(0x51B459 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_51B450
				CPatch::PatchPointer(0x51BAA4 + 2, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx]; CModelInfo::ms_modelInfoPtrs	; sub_51BA60
				CPatch::PatchPointer(0x51BAEC + 2, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[edi]; CModelInfo::ms_modelInfoPtrs	; sub_51BA60
				CPatch::PatchPointer(0x51BD15 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[esi+esi]; CModelInfo::ms_modelInfoPtrs	; sub_51BCE0
				CPatch::PatchPointer(0x51BDB9 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx+edx]; CModelInfo::ms_modelInfoPtrs	; sub_51BD90
				CPatch::PatchPointer(0x51C1BF + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_51C190
				CPatch::PatchPointer(0x51C1D0 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_51C190
				CPatch::PatchPointer(0x51C22A + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_51C200
				CPatch::PatchPointer(0x51C240 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_51C200
				CPatch::PatchPointer(0x51C290 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_51C260
				CPatch::PatchPointer(0x51C2A5 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_51C260
				CPatch::PatchPointer(0x51C313 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_51C2E0
				CPatch::PatchPointer(0x51C360 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_51C330
				CPatch::PatchPointer(0x51C612 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_51C600
				CPatch::PatchPointer(0x51C628 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_51C600
				CPatch::PatchPointer(0x51C63E + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4], 0; CModelInfo::ms_modelInfoPtrs	; sub_51C600
				CPatch::PatchPointer(0x51C655 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_51C650
				CPatch::PatchPointer(0x51C678 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_51C650
				CPatch::PatchPointer(0x51C688 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4], 0; CModelInfo::ms_modelInfoPtrs	; sub_51C650
				CPatch::PatchPointer(0x51C6A9 + 1, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edi, offset _ZN10CModelInfo16ms_modelInfoPtrsE; CModelInfo::ms_modelInfoPtrs	; _ZN10CModelInfo10InitialiseEv
				CPatch::PatchPointer(0x51C785 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4], esi; CModelInfo::ms_modelInfoPtrs	; sub_51C740
				CPatch::PatchPointer(0x51C7EC + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4], esi; CModelInfo::ms_modelInfoPtrs	; sub_51C7A0
				CPatch::PatchPointer(0x51C845 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4], esi; CModelInfo::ms_modelInfoPtrs	; sub_51C800
				CPatch::PatchPointer(0x51C8A3 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4], esi; CModelInfo::ms_modelInfoPtrs	; sub_51C860
				CPatch::PatchPointer(0x51C93E + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4], esi; CModelInfo::ms_modelInfoPtrs	; sub_51C8C0
				CPatch::PatchPointer(0x51C9B6 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4], esi; CModelInfo::ms_modelInfoPtrs	; sub_51C960
				CPatch::PatchPointer(0x51CA3D + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4], esi; CModelInfo::ms_modelInfoPtrs	; sub_51C9D0
				CPatch::PatchPointer(0x51CAA3 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4], esi; CModelInfo::ms_modelInfoPtrs	; sub_51CA60
				CPatch::PatchPointer(0x51CAC4 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_51CAC0
				CPatch::PatchPointer(0x51CB37 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_51CB30
				CPatch::PatchPointer(0x51CB74 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4], eax; CModelInfo::ms_modelInfoPtrs	; sub_51CB30
				CPatch::PatchPointer(0x51CB90 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_51CB30
				CPatch::PatchPointer(0x51CBAF + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4], esi; CModelInfo::ms_modelInfoPtrs	; sub_51CB30
				CPatch::PatchPointer(0x51CBE3 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_51CBD0
				CPatch::PatchPointer(0x51CBF5 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_51CBD0
				CPatch::PatchPointer(0x51CC1F + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4], 0; CModelInfo::ms_modelInfoPtrs	; sub_51CBD0
				CPatch::PatchPointer(0x51CC33 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4], eax; CModelInfo::ms_modelInfoPtrs	; sub_51CBD0
				CPatch::PatchPointer(0x520CA1 + 1, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     esi, offset _ZN10CModelInfo16ms_modelInfoPtrsE; CModelInfo::ms_modelInfoPtrs	; sub_520CA0
				CPatch::PatchPointer(0x520CBA + 2, CModelInfo__ms_modelInfoPtrs.bully_se + GetFileIDcurrentLimit(FILE_TYPE_NIF));	// cmp     esi, offset dword_C735D8	; sub_520CA0
				CPatch::PatchPointer(0x522C60 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_522C30
				CPatch::PatchPointer(0x522C77 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_522C30
				CPatch::PatchPointer(0x522D1B + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_522CB0
				CPatch::PatchPointer(0x522D41 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_522CB0
				CPatch::PatchPointer(0x5268CB + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_526830
				CPatch::PatchPointer(0x52EA74 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_52E8D0
				CPatch::PatchPointer(0x52EAEB + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; getNameByFileID
				CPatch::PatchPointer(0x52F677 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ebx, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_52F640
				CPatch::PatchPointer(0x52F952 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_52F8B0
				CPatch::PatchPointer(0x52FA46 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_52F9A0
				CPatch::PatchPointer(0x52FAC8 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_52FA90
				CPatch::PatchPointer(0x52FAF7 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_52FA90
				CPatch::PatchPointer(0x52FB14 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_52FA90
				CPatch::PatchPointer(0x52FBB3 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_52FB40
				CPatch::PatchPointer(0x52FC26 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_52FB40
				CPatch::PatchPointer(0x52FC73 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_52FB40
				CPatch::PatchPointer(0x52FDBB + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs	; sub_52FCD0
				CPatch::PatchPointer(0x52FE0C + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs	; sub_52FCD0
				CPatch::PatchPointer(0x52FE7D + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_52FE40
				CPatch::PatchPointer(0x52FEBD + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_52FE40
				CPatch::PatchPointer(0x52FF7D + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_52FF40
				CPatch::PatchPointer(0x52FFBD + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_52FF40
				CPatch::PatchPointer(0x5300A6 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_530060
				CPatch::PatchPointer(0x53040D + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_530100
				CPatch::PatchPointer(0x53090E + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_5308D0
				CPatch::PatchPointer(0x530964 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_5308D0
				CPatch::PatchPointer(0x5309F9 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_5308D0
				CPatch::PatchPointer(0x530A4F + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_5308D0
				CPatch::PatchPointer(0x530AFA + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_530A70
				CPatch::PatchPointer(0x530BE2 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_530B50
				CPatch::PatchPointer(0x530CBD + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; _ZN10CStreaming20RemoveLeastUsedModelEj
				CPatch::PatchPointer(0x531CFB + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; _ZN10CStreaming21ConvertBufferToObjectEPcii
				CPatch::PatchPointer(0x53207F + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_532020
				CPatch::PatchPointer(0x532120 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_532020
				CPatch::PatchPointer(0x532540 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_532510
				CPatch::PatchPointer(0x532589 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_532510
				CPatch::PatchPointer(0x532888 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_532750
				CPatch::PatchPointer(0x5328C4 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_532750
				CPatch::PatchPointer(0x532A8F + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs	; sub_532A10
				CPatch::PatchPointer(0x532E32 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs	; sub_532C80
				CPatch::PatchPointer(0x532F17 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs	; sub_532C80
				CPatch::PatchPointer(0x53320E + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ebp, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_533200
				CPatch::PatchPointer(0x53340F + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_5333E0
				CPatch::PatchPointer(0x5337F0 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; CStreaming::Init
				CPatch::PatchPointer(0x5338A0 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// cmp     _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4], 0; CModelInfo::ms_modelInfoPtrs	; sub_533880
				CPatch::PatchPointer(0x533990 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// cmp     _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4], 0; CModelInfo::ms_modelInfoPtrs	; sub_533980
				CPatch::PatchPointer(0x533BF4 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_533B30
				CPatch::PatchPointer(0x533E57 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_533DA0
				CPatch::PatchPointer(0x5340F9 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; _ZN10CStreaming22LoadAllRequestedModelsEb
				CPatch::PatchPointer(0x53492B + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_5348D0
				CPatch::PatchPointer(0x534980 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs	; sub_5348D0
				CPatch::PatchPointer(0x5359D4 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_5359D0
				CPatch::PatchPointer(0x535CF5 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_5359F0
				CPatch::PatchPointer(0x535E08 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// cmp     _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4], esi; CModelInfo::ms_modelInfoPtrs	; sub_535D90
				CPatch::PatchPointer(0x54F087 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_54EE80
				CPatch::PatchPointer(0x563A0E + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_563970
				CPatch::PatchPointer(0x563A92 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_563970
				CPatch::PatchPointer(0x563C86 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_563BF0
				CPatch::PatchPointer(0x56EF9C + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_56EF60
				CPatch::PatchPointer(0x56F275 + 1, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edi, offset _ZN10CModelInfo16ms_modelInfoPtrsE; CModelInfo::ms_modelInfoPtrs	; sub_56F260
				CPatch::PatchPointer(0x56F2D1 + 2, CModelInfo__ms_modelInfoPtrs.bully_se + GetFileIDcurrentLimit(FILE_TYPE_NIF));	// cmp     edi, offset dword_C735D8	; sub_56F260
				CPatch::PatchPointer(0x56FD10 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs	; sub_56FC40
				CPatch::PatchPointer(0x56FDAA + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_56FC40
				CPatch::PatchPointer(0x57097E + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_5708E0
				CPatch::PatchPointer(0x5B345D + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_5B33D0
				CPatch::PatchPointer(0x5BD117 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_5BCD70
				CPatch::PatchPointer(0x5BD2B5 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ebp, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_5BD1F0
				CPatch::PatchPointer(0x5C1B3D + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_5C1B30
				CPatch::PatchPointer(0x5C1B7D + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_5C1B60
				CPatch::PatchPointer(0x5C1C68 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs	; sub_5C1BA0
				CPatch::PatchPointer(0x5C3072 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_5C3030
				CPatch::PatchPointer(0x5C362E + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_5C35E0
				CPatch::PatchPointer(0x5C4743 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_5C46F0
				CPatch::PatchPointer(0x5C4985 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ebp, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_5C47E0
				CPatch::PatchPointer(0x5C50E9 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_5C4FB0
				CPatch::PatchPointer(0x5C592A + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ebp, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_5C5870
				CPatch::PatchPointer(0x5C5C24 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_5C5C10
				CPatch::PatchPointer(0x5C6078 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_5C5DB0
				CPatch::PatchPointer(0x5CE026 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_5CE010
				CPatch::PatchPointer(0x5CE06A + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_5CE050
				CPatch::PatchPointer(0x5CE2F8 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_5CE240
				CPatch::PatchPointer(0x5CE59C + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_5CE590
				CPatch::PatchPointer(0x5CF08F + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_5CF060
				CPatch::PatchPointer(0x5CF0EE + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_5CF0C0
				CPatch::PatchPointer(0x5D2C6F + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_5D2C60
				CPatch::PatchPointer(0x5D758F + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_5D73B0
				CPatch::PatchPointer(0x5D78C2 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_5D7870
				CPatch::PatchPointer(0x5D792A + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_5D7870
				CPatch::PatchPointer(0x5D9241 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ebx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_5D9230
				CPatch::PatchPointer(0x5DA21E + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_5DA170
				CPatch::PatchPointer(0x5DA2A2 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_5DA170
				CPatch::PatchPointer(0x5DA813 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_5DA7F0
				CPatch::PatchPointer(0x5DC1B3 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_5DC1A0
				CPatch::PatchPointer(0x5DE3EF + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_5DE2A0
				CPatch::PatchPointer(0x5DF45D + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_5DF1E0
				CPatch::PatchPointer(0x5E0724 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_5E0340
				CPatch::PatchPointer(0x5E6533 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_5E6510
				CPatch::PatchPointer(0x5E68FF + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_4EA430
				CPatch::PatchPointer(0x5E694B + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_5E6920
				CPatch::PatchPointer(0x5E90D9 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_5E90C0
				CPatch::PatchPointer(0x5E90E0 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_5E90C0
				CPatch::PatchPointer(0x5EC332 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_5EC200
				CPatch::PatchPointer(0x5F3028 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ebp, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_5F2FE0
				CPatch::PatchPointer(0x5F30CF + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ebx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_5F2FE0
				CPatch::PatchPointer(0x5F3494 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_5F3470
				CPatch::PatchPointer(0x5F3527 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_5F3500
				CPatch::PatchPointer(0x5F8CE3 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_5F8B60
				CPatch::PatchPointer(0x5F9027 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_5F8F80
				CPatch::PatchPointer(0x5F947B + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_5F92F0
				CPatch::PatchPointer(0x5F96B7 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_5F9670
				CPatch::PatchPointer(0x60332B + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_6032D0
				CPatch::PatchPointer(0x60334D + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_6032D0
				CPatch::PatchPointer(0x60733F + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_607300
				CPatch::PatchPointer(0x6123F4 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_6123D0
				CPatch::PatchPointer(0x619F03 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_619E10
				CPatch::PatchPointer(0x624276 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_6236D0
				CPatch::PatchPointer(0x6242A8 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_6236D0
				CPatch::PatchPointer(0x6555FB + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_6555E0
				CPatch::PatchPointer(0x6574DD + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_657420
				CPatch::PatchPointer(0x65776C + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_657740
				CPatch::PatchPointer(0x657B52 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_657AE0
				CPatch::PatchPointer(0x657EF2 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_657DF0
				CPatch::PatchPointer(0x658244 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_658210
				CPatch::PatchPointer(0x6582C1 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_658270
				CPatch::PatchPointer(0x65832A + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_6582F0
				CPatch::PatchPointer(0x65BE79 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs	; sub_65BDF0
				CPatch::PatchPointer(0x65BF04 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_65BEE0
				CPatch::PatchPointer(0x65BF25 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_65BEE0
				CPatch::PatchPointer(0x65BFF2 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[esi*4]; CModelInfo::ms_modelInfoPtrs	; sub_65BFB0
				CPatch::PatchPointer(0x65C14B + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_65C0E0
				CPatch::PatchPointer(0x65C1BB + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_65C190
				CPatch::PatchPointer(0x65C53B + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_65C3B0
				CPatch::PatchPointer(0x65C64A + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_65C3B0
				CPatch::PatchPointer(0x6622A1 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_662200
				CPatch::PatchPointer(0x66275B + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_662750
				CPatch::PatchPointer(0x66316D + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs	; sub_663120
				CPatch::PatchPointer(0x663291 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ebx*4]; CModelInfo::ms_modelInfoPtrs	; sub_6631D0
				CPatch::PatchPointer(0x66442D + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_664350
				CPatch::PatchPointer(0x6646E6 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_664640
				CPatch::PatchPointer(0x6652B3 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_6651E0
				CPatch::PatchPointer(0x665460 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_6653D0
				CPatch::PatchPointer(0x681264 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_6810B0
				CPatch::PatchPointer(0x6A1354 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_6A1350
				CPatch::PatchPointer(0x6A21D4 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_6A21D0
				CPatch::PatchPointer(0x6C410A + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_6C4030
				CPatch::PatchPointer(0x6C41F0 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[ebp*4]; CModelInfo::ms_modelInfoPtrs	; sub_6C4030
				CPatch::PatchPointer(0x6C4396 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[edi*4]; CModelInfo::ms_modelInfoPtrs	; sub_6C4370
				CPatch::PatchPointer(0x6CBF30 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     edi, _ZN10CModelInfo16ms_modelInfoPtrsE[ecx*4]; CModelInfo::ms_modelInfoPtrs	; sub_6CBF10
				CPatch::PatchPointer(0x6CC9A6 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_6CC7C0
				CPatch::PatchPointer(0x6CCCE2 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_6CCB20
				CPatch::PatchPointer(0x6D1984 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     esi, _ZN10CModelInfo16ms_modelInfoPtrsE[edx*4]; CModelInfo::ms_modelInfoPtrs	; sub_6D1940
				CPatch::PatchPointer(0x6D834F + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_6D82E0
				CPatch::PatchPointer(0x6D83EA + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     eax, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_6D83D0
				CPatch::PatchPointer(0x72C9D2 + 3, CModelInfo__ms_modelInfoPtrs.bully_se);	// mov     ecx, _ZN10CModelInfo16ms_modelInfoPtrsE[eax*4]; CModelInfo::ms_modelInfoPtrs	; sub_72C9C0
			}
			
			// Patch references to CModelInfo::modelInfoFlags		
			{
				CPatch::PatchPointer(0x527480 + 3, CSimpleModelInfo__m_UpdatedUVATable.ptr);	// mov     _ZN10CModelInfo17ms_modelInfoFlagsE[eax*4], edx; CModelInfo::ms_modelInfoFlags	; InitializeModelInfoFlags
				CPatch::PatchPointer(0x527487 + 3, CSimpleModelInfo__m_UpdatedUVATable.ptr + 1);	// mov     _ZN10CModelInfo17ms_modelInfoFlagsE+4[eax*4], edx	; InitializeModelInfoFlags
				CPatch::PatchPointer(0x52748E + 3, CSimpleModelInfo__m_UpdatedUVATable.ptr + 2);	// mov     _ZN10CModelInfo17ms_modelInfoFlagsE+8[eax*4], edx	; InitializeModelInfoFlags
				CPatch::PatchPointer(0x527495 + 3, CSimpleModelInfo__m_UpdatedUVATable.ptr + 3);	// mov     _ZN10CModelInfo17ms_modelInfoFlagsE+0Ch[eax*4], edx	; InitializeModelInfoFlags
				CPatch::PatchPointer(0x5274BC + 3, CSimpleModelInfo__m_UpdatedUVATable.ptr + 3);	// mov     _ZN10CModelInfo17ms_modelInfoFlagsE+0Ch[eax*4], edx	; InitializeModelInfoFlags
				CPatch::PatchPointer(0x5274C3 + 3, CSimpleModelInfo__m_UpdatedUVATable.ptr + 2);	// mov     _ZN10CModelInfo17ms_modelInfoFlagsE+8[eax*4], edx	; InitializeModelInfoFlags
				CPatch::PatchPointer(0x5274CA + 3, CSimpleModelInfo__m_UpdatedUVATable.ptr + 1);	// mov     _ZN10CModelInfo17ms_modelInfoFlagsE+4[eax*4], edx	; InitializeModelInfoFlags
				CPatch::PatchPointer(0x5274F3 + 3, CSimpleModelInfo__m_UpdatedUVATable.ptr);	// and     eax, _ZN10CModelInfo17ms_modelInfoFlagsE[edx*4]; CModelInfo::ms_modelInfoFlags	; sub_5274E0
				CPatch::PatchPointer(0x527527 + 3, CSimpleModelInfo__m_UpdatedUVATable.ptr);	// mov     _ZN10CModelInfo17ms_modelInfoFlagsE[eax*4], edx; CModelInfo::ms_modelInfoFlags	; CModelInfo::MarkModelExists
			}

			// Patch CModelInfo::modelInfoFlags initialization
			{
				CPatch::RedirectFunction(0x527470, &InitializeModelInfoFlags);
			}

			// Patch 12200
			{
				CPatch::PatchUINT32(0x51C6A4 + 1, GetBaseID(FILE_TYPE_NFT));	// mov     ecx, 12200	; _ZN10CModelInfo10InitialiseEv
			}

			// Patch 12199 (count of IDs - 1)
			{
				CPatch::PatchUINT32(0x4A79E8 + 1, GetFileIDcurrentLimit(FILE_TYPE_NIF) - 1);	// push    2FA7h	; sub_4A79C0
				CPatch::PatchUINT32(0x4A7E9C + 1, GetFileIDcurrentLimit(FILE_TYPE_NIF) - 1);	// push    2FA7h	; sub_4A7E50
				CPatch::PatchUINT32(0x4A7F4B + 1, GetFileIDcurrentLimit(FILE_TYPE_NIF) - 1);	// push    2FA7h	; sub_4A7E50
				CPatch::PatchUINT32(0x5209FD + 1, GetFileIDcurrentLimit(FILE_TYPE_NIF) - 1);	// push    2FA7h	; sub_520960
				CPatch::PatchUINT32(0x5FCFA7 + 1, GetFileIDcurrentLimit(FILE_TYPE_NIF) - 1);	// push    2FA7h	; sub_5FCFA0
				CPatch::PatchUINT32(0x5FCFB7 + 1, GetFileIDcurrentLimit(FILE_TYPE_NIF) - 1);	// push    2FA7h	; sub_5FCFA0
				CPatch::PatchUINT32(0x5FD027 + 1, GetFileIDcurrentLimit(FILE_TYPE_NIF) - 1);	// push    2FA7h	; sub_5FD020
				CPatch::PatchUINT32(0x60428C + 1, GetFileIDcurrentLimit(FILE_TYPE_NIF) - 1);	// push    2FA7h	; sub_604280
				CPatch::PatchUINT32(0x60AA37 + 1, GetFileIDcurrentLimit(FILE_TYPE_NIF) - 1);	// push    2FA7h	; sub_60AA30
				CPatch::PatchUINT32(0x656356 + 1, GetFileIDcurrentLimit(FILE_TYPE_NIF) - 1);	// push    2FA7h	; sub_656330
				CPatch::PatchUINT32(0x656417 + 1, GetFileIDcurrentLimit(FILE_TYPE_NIF) - 1);	// push    2FA7h	; sub_656330
				CPatch::PatchUINT32(0x65C3EE + 1, GetFileIDcurrentLimit(FILE_TYPE_NIF) - 1);	// push    2FA7h	; sub_65C3B0
				CPatch::PatchUINT32(0x65C422 + 1, GetFileIDcurrentLimit(FILE_TYPE_NIF) - 1);	// push    2FA7h	; sub_65C3B0
			}

			// Patch some container
			{
				CPatch::PatchUINT32(0x51C70D + 1, GetFileIDcurrentLimit(FILE_TYPE_NIF) * 4 + 2);	// push    48802           ; unsigned int
				
				CPatch::PatchUINT32(0x51BA80 + 3, GetFileIDcurrentLimit(FILE_TYPE_NIF) * 4);	// movzx   eax, word ptr [esi+0BEA0h]	; sub_51BA60
				CPatch::PatchUINT32(0x51BA91 + 3, GetFileIDcurrentLimit(FILE_TYPE_NIF) * 4);	// mov     [esi+0BEA0h], cx	; sub_51BA60
				CPatch::PatchUINT32(0x51BB86 + 3, GetFileIDcurrentLimit(FILE_TYPE_NIF) * 4);	// movzx   edi, word ptr [esi+0BEA0h]	; sub_51BA60
				CPatch::PatchUINT32(0x51BC7D + 3, GetFileIDcurrentLimit(FILE_TYPE_NIF) * 4);	// mov     [esi+0BEA0h], dx	; sub_51BA60
				CPatch::PatchUINT32(0x51BCEB + 3, GetFileIDcurrentLimit(FILE_TYPE_NIF) * 4);	// movzx   edi, word ptr [ecx+0BEA0h]	; sub_51BCE0
				CPatch::PatchUINT32(0x51BD99 + 3, GetFileIDcurrentLimit(FILE_TYPE_NIF) * 4);	// movzx   edi, word ptr [esi+0BEA0h]	; sub_51BD90
				CPatch::PatchUINT32(0x51BF1F + 3, GetFileIDcurrentLimit(FILE_TYPE_NIF) * 4);	// mov     [esi+0BEA0h], ax	; sub_51BD90
				CPatch::PatchUINT32(0x51BFCE + 3, GetFileIDcurrentLimit(FILE_TYPE_NIF) * 4);	// mov     [esi+0BEA0h], di	; sub_51BD90
				CPatch::PatchUINT32(0x51C023 + 3, GetFileIDcurrentLimit(FILE_TYPE_NIF) * 4);	// movzx   edi, word ptr [esi+0BEA0h]	; sub_51BD90
				CPatch::PatchUINT32(0x51C16B + 3, GetFileIDcurrentLimit(FILE_TYPE_NIF) * 4);	// mov     [esi+0BEA0h], di	; sub_51BD90
				CPatch::PatchUINT32(0x51C71F + 3, GetFileIDcurrentLimit(FILE_TYPE_NIF) * 4);	// mov     word ptr [eax+48800], 0FFFFh	; _ZN10CModelInfo10InitialiseEv
			}

			// Patch something
			{
				CPatch::PatchUINT32(0x5334D5 + 1, FileIDlimit::GetCountOfAllFileIDs() - GetBaseID(FILE_TYPE_NFT));
			}

		}
		#endif

		// NFT limit
		#if TRUE
		{
			// base ID of next type
			{
				CPatch::PatchUINT32(0x52E901 + 2, GetBaseID(FILE_TYPE_COL));	// cmp     esi, 56B8h	; sub_52E8D0
				CPatch::PatchUINT32(0x52EA50 + 2, GetBaseID(FILE_TYPE_COL));	// cmp     ecx, 56B8h	; sub_52E8D0
			}

			// 10000, (count of NFT files)
			{
				CPatch::PatchUINT32(0x533674 + 2, GetFileIDcurrentLimit(FILE_TYPE_NFT));	// cmp     esi, 2710h	; CStreaming::Init
				CPatch::PatchUINT32(0x534965 + 1, GetFileIDcurrentLimit(FILE_TYPE_NFT));	// mov     ecx, 2710h	; sub_5348D0
				CPatch::PatchUINT32(0x5F1901 + 2, GetFileIDcurrentLimit(FILE_TYPE_NFT));	// cmp     esi, 2710h	; sub_5F18C0
				CPatch::PatchUINT32(0x5F262D + 3, GetFileIDcurrentLimit(FILE_TYPE_NFT));	// mov     dword ptr [esi+8], 10000	; _ZN9CTxdStore10InitialiseEv
				CPatch::PatchUINT32(0x5F26CE + 2, GetFileIDcurrentLimit(FILE_TYPE_NFT));	// cmp     esi, 2710h	; sub_5F2690
				CPatch::PatchUINT32(0x6D1AF4 + 1, GetFileIDcurrentLimit(FILE_TYPE_NFT));	// cmp     eax, 2710h	; sub_6D1A60
			}

			// 9999, (count of NFT files - 1)
			{
				CPatch::PatchUINT32(0x52E660 + 2, GetFileIDcurrentLimit(FILE_TYPE_NFT) - 1);	// cmp     ecx, 9999	; IsMemoryAvailableForNewFile
				CPatch::PatchUINT32(0x52E6BE + 2, GetFileIDcurrentLimit(FILE_TYPE_NFT) - 1);	// cmp     esi, 270Fh	; IsMemoryAvailableForNewFile
				CPatch::PatchUINT32(0x530751 + 2, GetFileIDcurrentLimit(FILE_TYPE_NFT) - 1);	// cmp     esi, 270Fh	; sub_530100
				CPatch::PatchUINT32(0x532CF6 + 1, GetFileIDcurrentLimit(FILE_TYPE_NFT) - 1);	// cmp     eax, 270Fh	; sub_532C80
			}

			// patch hash container
			{
				CPatch::PatchUINT32(0x5F2650 + 1, GetFileIDcurrentLimit(FILE_TYPE_NFT) * 4 + 2);	// push    40002

				CPatch::PatchUINT32(0x534956 + 1, GetFileIDcurrentLimit(FILE_TYPE_NFT) * 4);	// push    40000                           ; unsigned int	; sub_5348D0
				CPatch::PatchUINT32(0x5F192B + 3, GetFileIDcurrentLimit(FILE_TYPE_NFT) * 4);	// movzx   eax, word ptr [ebx+9C40h]	; sub_5F1910
				CPatch::PatchUINT32(0x5F193C + 3, GetFileIDcurrentLimit(FILE_TYPE_NFT) * 4);	// mov     [ebx+9C40h], cx	; sub_5F1910
				CPatch::PatchUINT32(0x5F1AE7 + 3, GetFileIDcurrentLimit(FILE_TYPE_NFT) * 4);	// movzx   esi, word ptr [ebx+9C40h]	; sub_5F1910
				CPatch::PatchUINT32(0x5F1BF4 + 3, GetFileIDcurrentLimit(FILE_TYPE_NFT) * 4);	// mov     [ebx+9C40h], dx	; sub_5F1910
				CPatch::PatchUINT32(0x5F1C55 + 3, GetFileIDcurrentLimit(FILE_TYPE_NFT) * 4);	// movzx   ebp, word ptr [ecx+9C40h]	; sub_5F1C50
				CPatch::PatchUINT32(0x5F1D78 + 3, GetFileIDcurrentLimit(FILE_TYPE_NFT) * 4);	// movzx   ebp, word ptr [esi+9C40h]	; sub_5F1D70
				CPatch::PatchUINT32(0x5F1F4F + 3, GetFileIDcurrentLimit(FILE_TYPE_NFT) * 4);	// mov     [esi+9C40h], ax	; sub_5F1D70
				CPatch::PatchUINT32(0x5F1FE5 + 3, GetFileIDcurrentLimit(FILE_TYPE_NFT) * 4);	// mov     [esi+9C40h], bp	; sub_5F1D70
				CPatch::PatchUINT32(0x5F202E + 3, GetFileIDcurrentLimit(FILE_TYPE_NFT) * 4);	// movzx   edi, word ptr [esi+9C40h]	; sub_5F1D70
				CPatch::PatchUINT32(0x5F2179 + 3, GetFileIDcurrentLimit(FILE_TYPE_NFT) * 4);	// mov     [esi+9C40h], di	; sub_5F1D70
				CPatch::PatchUINT32(0x5F2661 + 3, GetFileIDcurrentLimit(FILE_TYPE_NFT) * 4);	// mov     word ptr [eax+9C40h], 0FFFFh	; _ZN9CTxdStore10InitialiseEv
			}
		}
		#endif

		// COL limit
		#if TRUE
		{
			// 500
			{
				CPatch::PatchUINT32(0x56F319 + 2, GetFileIDcurrentLimit(FILE_TYPE_COL));	// cmp     esi, 1F4h	; sub_56F260
				CPatch::PatchUINT32(0x56F37E + 2, GetFileIDcurrentLimit(FILE_TYPE_COL));	// cmp     esi, 1F4h	; sub_56F330
				CPatch::PatchUINT32(0x56F633 + 2, GetFileIDcurrentLimit(FILE_TYPE_COL));	// cmp     ecx, 1F4h	; _ZN9CColStore10InitialiseEv
				CPatch::PatchUINT32(0x56F817 + 2, GetFileIDcurrentLimit(FILE_TYPE_COL));	// cmp     ecx, 1F4h	; _ZN9CColStore10InitialiseEv
				CPatch::PatchUINT32(0x56FF7D + 2, GetFileIDcurrentLimit(FILE_TYPE_COL));	// cmp     edi, 1F4h	; sub_56FE00
				CPatch::PatchUINT32(0x57019D + 2, GetFileIDcurrentLimit(FILE_TYPE_COL));	// cmp     ebx, 1F4h	; sub_570080
				CPatch::PatchUINT32(0x570244 + 2, GetFileIDcurrentLimit(FILE_TYPE_COL));	// cmp     esi, 1F4h	; sub_5701C0
				CPatch::PatchUINT32(0x570313 + 2, GetFileIDcurrentLimit(FILE_TYPE_COL));	// cmp     edi, 1F4h	; sub_570260
				CPatch::PatchUINT32(0x570439 + 1, GetFileIDcurrentLimit(FILE_TYPE_COL));	// cmp     eax, 1F4h	; sub_570330
				CPatch::PatchUINT32(0x5704CB + 2, GetFileIDcurrentLimit(FILE_TYPE_COL));	// cmp     esi, 1F4h	; sub_570470
				CPatch::PatchUINT32(0x5705CD + 3, GetFileIDcurrentLimit(FILE_TYPE_COL));	// mov     dword ptr [esi+8], 500	; sub_5705A0
				CPatch::PatchUINT32(0x570680 + 2, GetFileIDcurrentLimit(FILE_TYPE_COL));	// cmp     ebp, 1F4h	; sub_570620
			}

			// 499
			{
				CPatch::PatchUINT32(0x53223A + 2, GetFileIDcurrentLimit(FILE_TYPE_COL) - 1);	// cmp     edx, 1F3h	; _ZN10CStreaming21ProcessLoadingChannelEi
			}
		}
		#endif

		// AGR limit
		#if TRUE
		{
			// Patch pointers to pAGRfiles
			#if TRUE
			{
				CPatch::PatchPointer(0x52D86F + 1, CStreaming__ms_apAGRBuffers.bully_se);	// mov     esi, offset pAGRfiles	; sub_52D820
				CPatch::PatchPointer(0x52D888 + 2, CStreaming__ms_apAGRBuffers.bully_se + GetFileIDcurrentLimit(FILE_TYPE_AGR));	// cmp     esi, offset dword_CA3670	; sub_52D820
				CPatch::PatchPointer(0x533830 + 1, CStreaming__ms_apAGRBuffers.bully_se);	// mov     edi, offset pAGRfiles	; _ZN10CStreaming4InitEv
				CPatch::PatchPointer(0x53402B + 3, CStreaming__ms_apAGRBuffers.bully_se);	// mov     eax, pAGRfiles[eax*4]	; _ZN10CStreaming22LoadAllRequestedModelsEb
			}
			#endif

			// Patch 580 (count of AGR files)
			#if TRUE
			{
				CPatch::PatchUINT32(0x533835 + 1, GetFileIDcurrentLimit(FILE_TYPE_AGR));	// mov     ebp, 244h	; _ZN10CStreaming4InitEv
			}
			#endif

			// Patch 579 (count of AGR files - 1)
			#if TRUE
			{
				CPatch::PatchUINT32(0x530D15 + 2, GetFileIDcurrentLimit(FILE_TYPE_AGR) - 1);	// cmp     edx, 243h	; _ZN10CStreaming20RemoveLeastUsedModelEj
				CPatch::PatchUINT32(0x531FDC + 1, GetFileIDcurrentLimit(FILE_TYPE_AGR) - 1);	// cmp     eax, 243h	; _ZN10CStreaming21ConvertBufferToObjectEPcii
				CPatch::PatchUINT32(0x532211 + 2, GetFileIDcurrentLimit(FILE_TYPE_AGR) - 1);	// cmp     ecx, 243h	; _ZN10CStreaming21ProcessLoadingChannelEi
				CPatch::PatchUINT32(0x532D10 + 1, GetFileIDcurrentLimit(FILE_TYPE_AGR) - 1);	// cmp     eax, 579	; sub_532C80
				CPatch::PatchUINT32(0x534024 + 1, GetFileIDcurrentLimit(FILE_TYPE_AGR) - 1);	// cmp     eax, 243h	; _ZN10CStreaming22LoadAllRequestedModelsEb
			}
			#endif
		}
		#endif

		// IBP limit
		#if TRUE
		{
			// Patch 100 (count of IDB files)
			#if TRUE
			{
				CPatch::PatchUINT32(0x436AFE + 1, GetFileIDcurrentLimit(FILE_TYPE_IPB));	// mov     eax, 64h	; sub_436AF0
				CPatch::PatchUINT32(0x43997D + 3, GetFileIDcurrentLimit(FILE_TYPE_IPB));	// mov     dword ptr [esi+8], 64h	; _ZN9CIplStore10InitialiseEv
				CPatch::PatchUINT32(0x439A34 + 4, GetFileIDcurrentLimit(FILE_TYPE_IPB));	// mov     [esp+2Ch+var_18], 64h	; sub_4399F0	
			}
			#endif

		}
		#endif

		// CAT limit
		#if TRUE
		{
			// Patch 120
			#if TRUE
			{
				CCodeMover::SetVariable("MAX_NUMBER_OF_CAT_FILES", GetFileIDcurrentLimit(FILE_TYPE_CAT));

				#if TRUE
				{
					// Count of active addresses: 1
					CCodeMover::FixOnAddressRel(0x5334C9, 7, "\x01\x01\x68\x05\x04""MAX_NUMBER_OF_CAT_FILES\x00\x02\x05\xCB\x34\x53\x00\x00", 0x5334D0);	// line 1, 0x5334C0 - _ZN10CStreaming4InitEv : push    120
				}
				#endif

				CCodeMover::DeleteVariable("MAX_NUMBER_OF_CAT_FILES");
			}
			#endif
		}
		#endif
	}
	#endif

	// finish
	CGenericLogStorage::WriteLineSeparator();
}
#endif

// Patches ID limit if neccessary.
void FileIDlimit::PatchFileIDlimits()
{
	MAKE_VAR_GAME_VERSION();

	this->UpdateBaseIDs();

	CPatch::EnterNewLevel();
	// CPatch::EnableDebugMode();
	
	MAKE_DEAD_IF();
	#ifdef IS_PLATFORM_WIN_X86
	else if (gameVersion == GAME_VERSION_GTA_III_1_0_WIN_X86)
		this->PatchFileIDlimits_GTA_III_1_0();
	else if (gameVersion == GAME_VERSION_GTA_VC_1_0_WIN_X86)
		this->PatchFileIDlimits_GTA_VC_1_0();
	else if (CGameVersion::IsAny_GTA_IV_or_EFLC(gameVersion))
		this->PatchFileIDlimits_GTA_IV_or_EFLC();
	else if (gameVersion == GAME_VERSION_BULLY_SE_1_2_WIN_X86)
		this->PatchFileIDlimits_Bully_SE_1_20();
	#endif
	else if (gameVersion == GAME_VERSION_GTA_SA_1_0_US_HOODLUM_WIN_X86 || gameVersion == GAME_VERSION_GTA_SA_2_00_ANDROID_ARM32)
		this->PatchFileIDlimits_GTA_SA();	

	CPatch::LeaveThisLevel();
}

// Patches save loading functions.
void FileIDlimit::PatchSaveBlockLoadingFunctions()
{
	// Patch functions that read from save.
	if (bIsPatchForLoadingSaveBlocksApplied)
		return;

	MAKE_VAR_GAME_VERSION();

	MAKE_DEAD_IF();
	#ifdef IS_PLATFORM_WIN_X86
	else if (CGameVersion::Is_GTA_SA_1_0_US_WIN_X86(gameVersion))
	{
		using namespace Game_GTASA;

		// CIplStore::Load
		{
			CPatch::RedirectCode(0x5D5514, &patch_CIplStore__Load_5D5514);
		}

		// 0x5D29E0 - CStreaming::Load
		CPatch::RedirectFunction(0x5D29E0, &CStreaming__Load);

		// Count of killeable IDs warnings
		CPatch::RedirectCode(0x5B6FCC, &patch_CFileLoader__LoadVehicleObject_5B6FCC);
		CPatch::RedirectCode(0x5B74A2, &patch_CFileLoader__LoadPedObject_5B74A2);

		// Pickups - model ID warning
		CPatch::RedirectCode(0x457290, &patch_CPickups__GenerateNewOne_457290);

		bIsPatchForLoadingSaveBlocksApplied = true;
	}
	#elif defined(IS_PLATFORM_ANDROID_ARM32)
	else if(gameVersion == GAME_VERSION_GTA_SA_2_00_ANDROID_ARM32)
	{
		using namespace Game_GTASA;		

		// CIplStore::Load		
		{
			Address_CIplStore__Load_48E426_thumb = g_mCalc.GetCurrentVAbyPreferedVA(ASM_GET_THUMB_ADDRESS_FOR_JUMP(0x48E426));
			Address_CIplStore__Load_48E476_thumb = g_mCalc.GetCurrentVAbyPreferedVA(ASM_GET_THUMB_ADDRESS_FOR_JUMP(0x48E476));
			CPatch::RedirectCodeEx(INSTRUCTION_SET_THUMB, g_mCalc.GetCurrentVAbyPreferedVA(0x48E46C),
				(void*)&patch_CIplStore__Load_48E46C
			);
		}

		CPatch::RedirectFunction((uintptr_t)Library::GetSymbolAddress(&g_LimitAdjuster.hModule_of_game,
			"_ZN10CStreaming4LoadEv"), (void*)&CStreaming__Load);
		
		// Count of killeable IDs warnings
		Address_CFileLoader__LoadVehicleObject_193F14_arm = g_mCalc.GetCurrentVAbyPreferedVA(ASM_GET_ARM_ADDRESS_FOR_JUMP(0x193F14));
		Address_CFileLoader__LoadVehicleObject_469A88_thumb = g_mCalc.GetCurrentVAbyPreferedVA(ASM_GET_THUMB_ADDRESS_FOR_JUMP(0x469A88));
		CPatch::RedirectCodeEx(INSTRUCTION_SET_THUMB, g_mCalc.GetCurrentVAbyPreferedVA(0x469A7E),
			(void*)&patch_CFileLoader__LoadVehicleObject_469A7E
		);

		Address_CFileLoader__LoadPedObject_1A1440_arm = g_mCalc.GetCurrentVAbyPreferedVA(ASM_GET_ARM_ADDRESS_FOR_JUMP(0x1A1440));
		Address_CFileLoader__LoadPedObject_469EBC_thumb = g_mCalc.GetCurrentVAbyPreferedVA(ASM_GET_THUMB_ADDRESS_FOR_JUMP(0x469EBC));
		CPatch::RedirectCodeEx(INSTRUCTION_SET_THUMB, g_mCalc.GetCurrentVAbyPreferedVA(0x469EB4),
			(void*)&patch_CFileLoader__LoadPedObject_469EB4
		);

		bIsPatchForLoadingSaveBlocksApplied = true;
	}
	#endif
}

// Allocates model info extension item
CBaseModelInfo_extension* FileIDlimit::AllocateModelInfoExtensionItemIfNeccessary(int modelID)
{
	if (FileIDlimit::CModelInfo__ms_modelInfoExtensionPtrs[modelID])
		return FileIDlimit::CModelInfo__ms_modelInfoExtensionPtrs[modelID];
	else
	{
		CBaseModelInfo_extension* ptr = FileIDlimit::CModelInfo__ms_modelInfoExtensionStore.AllocateItem();
		ptr->Init();

		CModelInfo__ms_modelInfoExtensionPtrs[modelID] = ptr;
		return ptr;
	}
}

// Enables model info model extension array
void FileIDlimit::EnableModelInfoExtensionArray()
{		
	this->bIsModelInfoExtensionArrayEnabled = true;
}

// Allocates model extension if enabled
void FileIDlimit::AllocateModelInfoExtensionArrayIfEnabled()
{
	MAKE_VAR_GAME_VERSION();	

	if (this->bIsModelInfoExtensionArrayEnabled && !CModelInfo__ms_modelInfoExtensionPtrs)
	{
		int modelLimit = GetFileIDcurrentLimit(GENERAL_FILE_TYPE_MODEL_ARCHIVE);
			
		if (modelLimit > 0)
		{
			CModelInfo__ms_modelInfoExtensionPtrs
				= new CBaseModelInfo_extension*[modelLimit];

			memset(
				CModelInfo__ms_modelInfoExtensionPtrs,
				0,
				modelLimit * sizeof(CBaseModelInfo_extension*)
			);
		}
	}
}

// Enables file info model extension array
void FileIDlimit::EnableFileInfoExtensionArray()
{
	this->bIsFileInfoExtensionArrayEnabled = true;
}

/*
NAKED void patch_CColStore__Initialise_finish()
{
	__asm
	{
		mov     ecx, [esp + 20h - 0xC];
		mov     fs:[0], ecx;
		add     esp, 20h;
		retn;
	}
}

void DoAllocateCollisionFileExtensionArrayIfEnabled()
{

}

NAKED void patch_CColStore__Initialise_4114AB()
{
	__asm
	{
		call DoAllocateCollisionFileExtensionArrayIfEnabled;
		jmp patch_CColStore__Initialise_finish;
	}
}
*/

// Enables collision file extension array
void FileIDlimit::EnableCollisionFileExtensionArray()
{
	FileIDlimit::bIsCollisionFileExtensionArrayEnabled = true;
}

// Allocates file extension if enabled
void FileIDlimit::AllocateFileInfoExtensionArrayIfEnabled()
{
	MAKE_VAR_GAME_VERSION();

	if (this->bIsFileInfoExtensionArrayEnabled || this->bIsIDbaseInt32)
	{
		MAKE_DEAD_IF();
#ifdef IS_PLATFORM_WIN_X86
	else if (CGameVersion::Is_GTA_SA_1_0_US_WIN_X86(gameVersion))
	{
		using namespace Game_GTASA;

		if (this->bIsFileInfoExtensionArrayEnabled && !CStreaming__ms_aInfoForModelExtension)
		{
			CStreaming__ms_aInfoForModelExtension
				= new CStreamingInfoExtension[this->GetCountOfAllFileIDs()];

			memset(
				CStreaming__ms_aInfoForModelExtension,
				NULL,
				this->GetCountOfAllFileIDs() * sizeof(CStreamingInfoExtension)
			);
		}

		CPatch::RedirectCode(0x5B8AE7, &patch_CStreaming__Init2_5B8AE7);
	}
#endif
	}
}

// Allocates collision file extension if enabled
void FileIDlimit::AllocateCollisionFileExtensionArrayIfEnabled()
{
	MAKE_VAR_GAME_VERSION();

	if (FileIDlimit::bIsCollisionFileExtensionArrayEnabled && !CColStore__ms_aColPoolExtension)
	{
		CColStore__ms_aColPoolExtension
			= new CCollisionFile_extension[this->GetFileIDcurrentLimit(FILE_TYPE_COL)];

		memset(CColStore__ms_aColPoolExtension,
			0,
			this->GetFileIDcurrentLimit(FILE_TYPE_COL) * sizeof(CStreamingInfoExtension));		
	}
}

// Allocates neccessary arrays
void FileIDlimit::AllocateNeccessaryArrays()
{
	FileIDlimit::AllocateFileInfoExtensionArrayIfEnabled();
	FileIDlimit::AllocateModelInfoExtensionArrayIfEnabled();
	FileIDlimit::AllocateCollisionFileExtensionArrayIfEnabled();
}

// Is special model ID?
bool FileIDlimit::IsSpecialModelID(int32_t ID)
{
	return
		(ID >= 384 && ID <= 397)
		|| (ID >= 300 && ID <= 319);
}

namespace Game_GTASA
{
	extern "C"
	{
		unsigned int CDarkel__CountOfKillableModelIDs = 0;
	}

	#ifdef IS_PLATFORM_ANDROID_ARM32
	// patch for 0x3048F4
	extern "C"
	{
		uintptr_t Address_CDarkel__ResetModelsKilledByPlayer_3048EE_thumb = 0;
		uintptr_t Address_CDarkel__ResetModelsKilledByPlayer_3048FC_thumb = 0;
	}

	static TARGET_THUMB NAKED void patch_CDarkel__ResetModelsKilledByPlayer_3048F4()
	{
		__asm(
		".thumb\n"
			ASM_CMP_4BYTE_VALUE_STORED_ON_SYMBOL(R2, CDarkel__CountOfKillableModelIDs)
			"BNE 1f\n"
			"BX LR\n"
			ASM_JUMP_TO_ADDRESS_STORED_ON_SYMBOL(Address_CDarkel__ResetModelsKilledByPlayer_3048FC_thumb)

			// Branches
			"1:\n"	// loc_3048EE
			ASM_JUMP_TO_ADDRESS_STORED_ON_SYMBOL(Address_CDarkel__ResetModelsKilledByPlayer_3048EE_thumb)
			);
	}
	#endif
}

// Patches count of killable model IDs
void FileIDlimit::PatchCountOfKillableModelIDs(int iCountOfKillableModelIDs)
{
	MAKE_VAR_GAME_VERSION();
	CPatch::EnterNewLevel();
	// CPatch::EnableDebugMode();

	if (CPatch::IsDebugModeActive())
		iCountOfKillableModelIDs = this->CDarkel__CountOfKillableModelIDs;

	MAKE_DEAD_IF();
	#if defined(IS_PLATFORM_WIN_X86) || defined(IS_PLATFORM_ANDROID_ARM32)
	else if (CGameVersion::Is_GTA_SA_1_0_US_WIN_X86(gameVersion)
		|| gameVersion == GAME_VERSION_GTA_SA_2_00_ANDROID_ARM32)
	{
		using namespace Game_GTASA;

		if (iCountOfKillableModelIDs > this->CDarkel__CountOfKillableModelIDs)
		{
			CDarkel__RegisteredKills.ptr = new int16_t[iCountOfKillableModelIDs * PlayerLimits::ms_NumberOfPlayers];
			CDarkel__RegisteredKills.bIsAllocated.Set(true);
		}

		// Set variable
		Game_GTASA::CDarkel__CountOfKillableModelIDs = iCountOfKillableModelIDs;

		MAKE_DEAD_IF();
		#ifdef IS_PLATFORM_WIN_X86
		else if(CGameVersion::Is_GTA_SA_1_0_US_WIN_X86(gameVersion))

		// CDarkel::FindTotalPedsKilledByPlayer
		CPatch::RedirectFunction(0x43D6E0, &CDarkel::FindTotalPedsKilledByPlayer);

		// .text:0043D6C0     ; CDarkel::QueryModelsKilledByPlayer(int, int)
		// .text:0043D6C0     _ZN7CDarkel25QueryModelsKilledByPlayerEii proc near
		if (gameVersion == GAME_VERSION_GTA_SA_1_0_US_HOODLUM_WIN_X86)
			CPatch::PatchPointer(0x1564C2B + 4, CDarkel__RegisteredKills.ptr);
		else if (gameVersion == GAME_VERSION_GTA_SA_1_0_US_COMPACT_WIN_X86)
			CPatch::PatchPointer(0x43D6CB + 4, CDarkel__RegisteredKills.ptr);


		// .text:0043DF20     ; void __cdecl CDarkel::RegisterCarBlownUpByPlayer(CEntity *a1, float a2)
		// .text:0043DF20     _ZN7CDarkel26RegisterCarBlownUpByPlayerEP8CVehiclei proc near	
		CPatch::PatchPointer(0x43DF5B + 4, CDarkel__RegisteredKills.ptr);

		// .text:0043DCD0     ; char __cdecl CDarkel::RegisterKillByPlayer(CPed *killedPed, int a2, char a3, int a4)
		// .text:0043DCD0     _ZN7CDarkel20RegisterKillByPlayerEPK4CPed11eWeaponTypebi proc near
		CPatch::PatchPointer(0x43DE6C + 4, CDarkel__RegisteredKills.ptr);

		// .text:0043D6A0     ; CDarkel::ResetModelsKilledByPlayer(int)
		// .text:0043D6A0     _ZN7CDarkel25ResetModelsKilledByPlayerEi proc near
		if (gameVersion == GAME_VERSION_GTA_SA_1_0_US_HOODLUM_WIN_X86)
		{
			CPatch::PatchPointer(0x1561634 + 3, CDarkel__RegisteredKills.ptr);
			CPatch::PatchUINT32(0x156163B + 1, CDarkel__CountOfKillableModelIDs);
		}
		else if (gameVersion == GAME_VERSION_GTA_SA_1_0_US_COMPACT_WIN_X86)
		{
			CPatch::PatchPointer(0x43D6A4 + 3, CDarkel__RegisteredKills.ptr);
			CPatch::PatchUINT32(0x43D6AB + 1, CDarkel__CountOfKillableModelIDs);
		}
		#elif defined(IS_PLATFORM_ANDROID_ARM32)
		else if(gameVersion == GAME_VERSION_GTA_SA_2_00_ANDROID_ARM32)
		{
			// Patch pointer
			CPatch::PatchPointer(g_mCalc.GetCurrentVAbyPreferedVA(0x678AAC), CDarkel__RegisteredKills.ptr);

			// Changes
			Address_CDarkel__ResetModelsKilledByPlayer_3048EE_thumb = g_mCalc.GetCurrentVAbyPreferedVA(ASM_GET_THUMB_ADDRESS_FOR_JUMP(0x3048EE));
			Address_CDarkel__ResetModelsKilledByPlayer_3048FC_thumb = g_mCalc.GetCurrentVAbyPreferedVA(ASM_GET_THUMB_ADDRESS_FOR_JUMP(0x3048FC));
			CPatch::RedirectCodeEx(INSTRUCTION_SET_THUMB, g_mCalc.GetCurrentVAbyPreferedVA(0x3048F4),
				(void*)&patch_CDarkel__ResetModelsKilledByPlayer_3048F4, 8
			);
		}
		#endif
	}
	#endif
	else
	{
		CPatch::LeaveThisLevel();
		return;
	}

	this->CDarkel__CountOfKillableModelIDs = iCountOfKillableModelIDs;

	CPatch::LeaveThisLevel();

	CGenericLogStorage::SaveFormattedTextLn("Modified count of killable model IDs: %d", iCountOfKillableModelIDs);
	CGenericLogStorage::WriteLineSeparator();
}

// Patch: file binary search by name
void FileIDlimit::PatchUseFileBinarySearchByName()
{
	MAKE_VAR_GAME_VERSION();
	CPatch::EnterNewLevel();
	// CPatch::EnableDebugMode();

	MAKE_DEAD_IF();
	#ifdef IS_PLATFORM_WIN_X86
	else if (CGameVersion::Is_GTA_SA_1_0_US_WIN_X86(gameVersion))
	{
		using namespace Game_GTASA;

		// models
		#if TRUE
		{
			CPatch::RedirectCode(0x5B6FE8, &patch_CFileLoader__LoadVehicleObject_5B6FE8);
			CPatch::RedirectCode(0x5B3DBD, &patch_CFileLoader__LoadObject_5B3DBD);
			CPatch::RedirectCode(0x5B74C6, &patch_CFileLoader__LoadPedObject_5B74C6);
			CPatch::RedirectCode(0x5B3F56, &patch_CFileLoader__LoadTimeObject_5B3F56);
			CPatch::RedirectCode(0x5B414C, &patch_CFileLoader__LoadAnimatedClumpObject_5B414C);
			CPatch::RedirectCode(0x5B408F, &patch_CFileLoader__LoadClumpObject_5B408F);
			CPatch::RedirectCode(0x5B3FFE, &patch_CFileLoader__LoadWeaponObject_5B3FFE);
			CPatch::RedirectCode(0x409E71, &patch_CStreaming__RequestSpecialModel_409E71);

			CPatch::RedirectFunction(0x4C5940, &CModelInfo__GetModelInfo);
			CPatch::RedirectFunction(0x4C59B0, &CModelInfo__GetModelInfoFromHashKey);
			CPatch::RedirectMethod(0x4C47E0, &CTimeInfo_patch_4C47E0::FindOtherTimeModel);
			CPatch::PutCallFunction(0x5B4F00, &CModelInfo__GetModelInfoLinearSearch);
			CPatch::PutCallFunction(0x5B10F0, &CModelInfo__GetModelInfoLinearSearch);
		}
		#endif

		// TXD files
		#if TRUE
		{
			CPatch::RedirectCode(0x731CC8, &patch_CTxdStore__AddTxdSlot_731CC8);
			CPatch::RedirectCode(0x731CF1, &patch_CTxdStore__RemoveTxdSlot_731CF1);

			CPatch::RedirectFunction(0x731850, &CTxdStore__FindTxdSlot);
			g_InlinedCodeRemoving.SwitchHookEnabled(eInlineRemove_CTxdStore__AddTxdSlot, true);
		}
		#endif
	}
	#endif
	else
	{
		CPatch::LeaveThisLevel();
		return;
	}

	CPatch::LeaveThisLevel();

	CGenericLogStorage::SaveTextLn("File binary search by name enabled.");
	CGenericLogStorage::WriteLineSeparator();
}

// Returns usPrev 
int32_t FileIDlimit::GetFileInfoPrev(int32_t fileID)
{
	MAKE_VAR_GAME_VERSION();

	if (this->bIsIDbaseInt32)
		return FileIDlimit::CStreaming__ms_aInfoForModelExtension[fileID].exUsPrev;
	else
	{
		int32_t result = -1;

		MAKE_DEAD_IF();
#ifdef IS_PLATFORM_WIN_X86
		else if (gameVersion == GAME_VERSION_GTA_III_1_0_WIN_X86)
			result = this->CStreaming__ms_aInfoForModel.gta_iii[fileID].usPrev - this->CStreaming__ms_aInfoForModel.gta_iii;
		else if(gameVersion == GAME_VERSION_GTA_VC_1_0_WIN_X86)
			result = this->CStreaming__ms_aInfoForModel.gta_vc[fileID].usPrev - this->CStreaming__ms_aInfoForModel.gta_vc;
		else if (CGameVersion::Is_GTA_SA_1_0_US_WIN_X86(gameVersion))
			result = this->CStreaming__ms_aInfoForModel.gta_sa[fileID].usPrev.raw;
		else if (gameVersion == GAME_VERSION_BULLY_SE_1_2_WIN_X86)
			result = this->CStreaming__ms_aInfoForModel.bully_se[fileID].usPrev;
#endif

		return result == 0xFFFF ? -1 : result;
	}
}

// Returns Next value
int32_t FileIDlimit::GetFileInfoNext(int32_t fileID)
{
	MAKE_VAR_GAME_VERSION();

	if (this->bIsIDbaseInt32)
		return FileIDlimit::CStreaming__ms_aInfoForModelExtension[fileID].exUsNext;
	else
	{
		int32_t result = -1;
		
		MAKE_DEAD_IF();
		#ifdef IS_PLATFORM_WIN_X86
		else if (gameVersion == GAME_VERSION_GTA_III_1_0_WIN_X86)
			result = this->CStreaming__ms_aInfoForModel.gta_iii[fileID].usNext - this->CStreaming__ms_aInfoForModel.gta_iii;
		else if (gameVersion == GAME_VERSION_GTA_VC_1_0_WIN_X86)
			result = this->CStreaming__ms_aInfoForModel.gta_vc[fileID].usNext - this->CStreaming__ms_aInfoForModel.gta_vc;
		else if(CGameVersion::Is_GTA_SA_1_0_US_WIN_X86(gameVersion))
			result = this->CStreaming__ms_aInfoForModel.gta_sa[fileID].usNext.raw;
		else if (gameVersion == GAME_VERSION_BULLY_SE_1_2_WIN_X86)
			result = this->CStreaming__ms_aInfoForModel.bully_se[fileID].usNext;
		#endif

		return result == 0xFFFF ? -1 : result;
	}
}

// Returns NextOnCD value
int32_t FileIDlimit::GetFileInfoNextOnCD(int32_t fileID)
{
	MAKE_VAR_GAME_VERSION();

	if (this->bIsIDbaseInt32)
		return this->CStreaming__ms_aInfoForModelExtension[fileID].exNextOnCd;
	else
	{
		int32_t result = -1;
		
		MAKE_DEAD_IF();
		#ifdef IS_PLATFORM_WIN_X86
		else if (gameVersion == GAME_VERSION_GTA_III_1_0_WIN_X86)
			result = this->CStreaming__ms_aInfoForModel.gta_iii[fileID].usNextOnCd;
		else if (gameVersion == GAME_VERSION_GTA_VC_1_0_WIN_X86)
			result = this->CStreaming__ms_aInfoForModel.gta_vc[fileID].usNextOnCd;
		else if (CGameVersion::Is_GTA_SA_1_0_US_WIN_X86(gameVersion))
			result = this->CStreaming__ms_aInfoForModel.gta_sa[fileID].usNextOnCd.raw;
		else if (gameVersion == GAME_VERSION_BULLY_SE_1_2_WIN_X86)
			result = this->CStreaming__ms_aInfoForModel.bully_se[fileID].usNextOnCd;
		#endif

		return result == 0xFFFF ? -1 : result;
	}
}

// Sets Prev value 
void FileIDlimit::SetFileInfoPrev(int32_t fileID, int32_t newValue)
{
	MAKE_VAR_GAME_VERSION();

	if (this->bIsIDbaseInt32)
		CStreaming__ms_aInfoForModelExtension[fileID].exUsPrev = newValue;
	else
	{
		MAKE_DEAD_IF();
		#ifdef IS_PLATFORM_WIN_X86
		else if(gameVersion == GAME_VERSION_GTA_III_1_0_WIN_X86)
			CStreaming__ms_aInfoForModel.gta_iii[fileID].usPrev = CStreaming__ms_aInfoForModel.gta_iii + newValue;
		else if (gameVersion == GAME_VERSION_GTA_VC_1_0_WIN_X86)
			CStreaming__ms_aInfoForModel.gta_vc[fileID].usPrev = CStreaming__ms_aInfoForModel.gta_vc + newValue;
		else if (CGameVersion::Is_GTA_SA_1_0_US_WIN_X86(gameVersion))
			this->CStreaming__ms_aInfoForModel.gta_sa[fileID].usPrev.raw = newValue;
		else if (gameVersion == GAME_VERSION_BULLY_SE_1_2_WIN_X86)
			this->CStreaming__ms_aInfoForModel.bully_se[fileID].usPrev = newValue;
		#endif
	}
}

// Sets Next value
void FileIDlimit::SetFileInfoNext(int32_t fileID, int32_t newValue)
{
	MAKE_VAR_GAME_VERSION();

	if (this->bIsIDbaseInt32)
		CStreaming__ms_aInfoForModelExtension[fileID].exUsNext = newValue;
	else
	{
		MAKE_DEAD_IF();
		#ifdef IS_PLATFORM_WIN_X86
		else if (gameVersion == GAME_VERSION_GTA_III_1_0_WIN_X86)
			CStreaming__ms_aInfoForModel.gta_iii[fileID].usNext = CStreaming__ms_aInfoForModel.gta_iii + newValue;
		else if (gameVersion == GAME_VERSION_GTA_VC_1_0_WIN_X86)
			CStreaming__ms_aInfoForModel.gta_vc[fileID].usNext = CStreaming__ms_aInfoForModel.gta_vc + newValue;
		else if (CGameVersion::Is_GTA_SA_1_0_US_WIN_X86(gameVersion))
			this->CStreaming__ms_aInfoForModel.gta_sa[fileID].usNext.raw = newValue;
		else if (gameVersion == GAME_VERSION_BULLY_SE_1_2_WIN_X86)
			this->CStreaming__ms_aInfoForModel.bully_se[fileID].usNext = newValue;
		#endif
	}
}

// Sets NextOnCD value
void FileIDlimit::SetFileInfoNextOnCD(int32_t fileID, int32_t newValue)
{
	MAKE_VAR_GAME_VERSION();

	if (this->bIsIDbaseInt32)
		CStreaming__ms_aInfoForModelExtension[fileID].exNextOnCd = newValue;
	else
	{
		MAKE_DEAD_IF();
		#ifdef IS_PLATFORM_WIN_X86
		else if(gameVersion == GAME_VERSION_GTA_III_1_0_WIN_X86)
			CStreaming__ms_aInfoForModel.gta_iii[fileID].usNextOnCd = newValue;
		else if (gameVersion == GAME_VERSION_GTA_VC_1_0_WIN_X86)
			CStreaming__ms_aInfoForModel.gta_vc[fileID].usNextOnCd = newValue;
		else if (CGameVersion::Is_GTA_SA_1_0_US_WIN_X86(gameVersion))
			this->CStreaming__ms_aInfoForModel.gta_sa[fileID].usNextOnCd.raw = newValue;
		else if (gameVersion == GAME_VERSION_BULLY_SE_1_2_WIN_X86)
			this->CStreaming__ms_aInfoForModel.bully_se[fileID].usNextOnCd = newValue;
		#endif
	}
}

////////////////		

// Finds model ID by model info
int32_t FileIDlimit::FindModelIDbyModelInfo(CBaseModelInfo_generic* pModelInfo)
{
	if (!CModelInfo__ms_modelInfoPtrs.generic)
		return UNDEFINED_FILE_ID_INT32;

	auto numberOfModels = GetFileIDcurrentLimit(GENERAL_FILE_TYPE_MODEL_ARCHIVE);

	for (int i = 0; i < numberOfModels; i++)
	{
		if (CModelInfo__ms_modelInfoPtrs.generic[i] == pModelInfo)
			return i;
	}

	return UNDEFINED_FILE_ID_INT32;
}

// Initialize must be called before any other function
void FileIDlimit::Initialise()
{
	MAKE_VAR_GAME_VERSION();
		
	this->countOfTypes = 0;

	this->NumberOfStreamingInfos = 0;

	this->maxCountOfIDs = 0;

	tFileTypeBaseID maxLimitValue;

	if (gameVersion == GAME_VERSION_GTA_III_1_0_WIN_X86)
	{
		using namespace Game_GTAIII;

		this->RegisterFileType(FILE_TYPE_DFF, 5500, SHRT_MAX);	// base ID		0
		this->RegisterFileType(FILE_TYPE_TXD, 850, SHRT_MAX);		// base ID		5500

		if (g_LimitAdjuster.IsGameVersionSetUpWithMemory())
		{
			#ifdef IS_PLATFORM_WIN_X86		
			this->CModelInfo__ms_modelInfoPtrs.gta_iii = (CBaseModelInfo**)0x83D408;
			this->CStreaming__ms_aInfoForModel.gta_iii = (CStreamingInfo*)0x6C7088;
			#endif
		}

		// Max count of file IDs
		this->maxCountOfIDs = SHRT_MAX;
	}	
	else if (gameVersion == GAME_VERSION_GTA_VC_1_0_WIN_X86)
	{
		using namespace Game_GTAVC;

		this->RegisterFileType(FILE_TYPE_DFF, 6500, SHRT_MAX);	// base ID		0
		this->RegisterFileType(FILE_TYPE_TXD, 1385, SHRT_MAX);	// TXD, base ID		6500
		this->RegisterFileType(FILE_TYPE_COL, 31, UCHAR_MAX + 1);		// COL, base ID		7885
		this->RegisterFileType(FILE_TYPE_IFP, 35, SHRT_MAX + 1);		// IFP, base ID		7916

		if (g_LimitAdjuster.IsGameVersionSetUpWithMemory())
		{
			#ifdef IS_PLATFORM_WIN_X86
			this->CModelInfo__ms_modelInfoPtrs.gta_vc = (CBaseModelInfo**)0x92D4C8;
			this->CStreaming__ms_aInfoForModel.gta_vc = (CStreamingInfo*)0x94DDD0;
			this->CAnimManager__ms_aAnimBlocks.gta_vc = (CAnimBlock*)0x7DB428;
			#endif
		}

		// Max count of file IDs
		this->maxCountOfIDs = SHRT_MAX;
	}
	else if (CGameVersion::Is_GTA_SA_1_0_US_WIN_X86(gameVersion)
		|| CGameVersion::Is_GTA_SA_1_0_1_EU_WIN_X86(gameVersion)
		|| CGameVersion::Is_GTA_SA_Android_Renderware(gameVersion)
		|| CGameVersion::IsAny_GTA_SA_UnrealEngine(gameVersion))
	{
		using namespace Game_GTASA;

		// DFF
		if (gameVersion == GAME_VERSION_GTA_SA_1_0_US_HOODLUM_WIN_X86)
			maxLimitValue = INT_MAX;
		else if (gameVersion == GAME_VERSION_GTA_SA_2_00_ANDROID_ARM32)
			maxLimitValue = SHRT_MAX;
		else
			maxLimitValue = false;

		this->RegisterFileType(FILE_TYPE_DFF, 20000, maxLimitValue);		// base ID		0

		// TXD
		if (gameVersion == GAME_VERSION_GTA_SA_1_0_US_HOODLUM_WIN_X86)
			maxLimitValue = INT_MAX;
		else if (gameVersion == GAME_VERSION_GTA_SA_2_00_ANDROID_ARM32)
			maxLimitValue = SHRT_MAX;
		else
			maxLimitValue = false;

		this->RegisterFileType(FILE_TYPE_TXD, 5000, maxLimitValue);		// base ID		20000

		// COL
		// for Android COL file ID in CColModel needs to be extended
		if (gameVersion == GAME_VERSION_GTA_SA_1_0_US_HOODLUM_WIN_X86 || gameVersion == GAME_VERSION_GTA_SA_2_00_ANDROID_ARM32)
			maxLimitValue = INT_MAX;
		else
			maxLimitValue = false;

		this->RegisterFileType(FILE_TYPE_COL, 255, maxLimitValue);			// base ID		25000

		// IPL
		// for Android IPL ID needs to be extended in CBuilding
		if (gameVersion == GAME_VERSION_GTA_SA_1_0_US_HOODLUM_WIN_X86)
			maxLimitValue = INT_MAX;
		else
			maxLimitValue = false;

		this->RegisterFileType(FILE_TYPE_IPL, 256, maxLimitValue);			// base ID		25255

		// DAT
		if (gameVersion == GAME_VERSION_GTA_SA_1_0_US_HOODLUM_WIN_X86 || gameVersion == GAME_VERSION_GTA_SA_2_00_ANDROID_ARM32)
			maxLimitValue = USHRT_MAX;
		else
			maxLimitValue = false;

		this->RegisterFileType(FILE_TYPE_DAT, 64, maxLimitValue);			// base ID		25511

		// IFP
		if (gameVersion == GAME_VERSION_GTA_SA_1_0_US_HOODLUM_WIN_X86 || gameVersion == GAME_VERSION_GTA_SA_2_00_ANDROID_ARM32)
			maxLimitValue = INT_MAX;
		else
			maxLimitValue = false;

		this->RegisterFileType(FILE_TYPE_IFP, 180, maxLimitValue);			// base ID		25575

		// RRR
		if (gameVersion == GAME_VERSION_GTA_SA_1_0_US_HOODLUM_WIN_X86 || gameVersion == GAME_VERSION_GTA_SA_2_00_ANDROID_ARM32)
			maxLimitValue = INT_MAX;
		else
			maxLimitValue = false;

		this->RegisterFileType(FILE_TYPE_RRR, 475, maxLimitValue);			// base ID		25755

		// SCM
		if (gameVersion == GAME_VERSION_GTA_SA_1_0_US_HOODLUM_WIN_X86)
			maxLimitValue = SCHAR_MAX;
		else
			maxLimitValue = false;

		this->RegisterFileType(FILE_TYPE_SCM, 82, maxLimitValue);			// base ID		26230

		// Loaded
		this->RegisterFileType(FILE_TYPE_LOADED_START, 2, false);	// base ID		26312

		// Requested
		this->RegisterFileType(FILE_TYPE_REQUESTED_START, 2, false);// base ID		26314

		// Max count of file IDs		
		if (gameVersion == GAME_VERSION_GTA_SA_1_0_US_HOODLUM_WIN_X86)
			this->maxCountOfIDs = INT_MAX;
		else
			this->maxCountOfIDs = SHRT_MAX;

		// Count of killable model IDs
		this->CDarkel__CountOfKillableModelIDs = 800;

		// Memory addresses
		if(g_LimitAdjuster.IsGameVersionSetUpWithMemory())
		{
			#ifdef IS_PLATFORM_WIN_X86
			if (CGameVersion::Is_GTA_SA_1_0_US_WIN_X86(gameVersion))
			{
				// Variables
				this->CModelInfo__ms_modelInfoPtrs.gta_sa = (CBaseModelInfo**)0xA9B0C8;

				this->CTxdStore__pTxdPool.gta_sa = (CPool_GTASA<TxdDef>**)0xC8800C;
				this->CStreaming__ms_aInfoForModel.gta_sa = (CStreamingInfo*)0x8E4CC0;
				this->CAnimManager__ms_aAnimBlocks.gta_sa = (CAnimBlock*)0xB5D4A0;
				this->CVehicleRecording__StreamingArray.gta_sa = (CVehicleRecordingFileInfo*)0x97D880;
				this->CTheScripts__StreamedScripts.ptr = (char*)0xA47B60;
				this->CIplStore__ms_pPool.gta_sa = (CPool_GTASA<CIplFile>**)0x8E3FB0;
				this->CColStore__ms_pColPool.gta_sa = (CPool_GTASA<Game_GTASA::CCollisionFile>**)0x965560;

				// Registered kills
				this->CDarkel__RegisteredKills.ptr = (int16_t*)0x969A50;
				
				// Lists
				this->CStreaming__ms_pStartRequestedList = (void**)0x8E4C58;
				this->CStreaming__ms_pEndRequestedList = (void**)0x8E4C54;
				this->CStreaming__ms_pStartLoadedList = (void**)0x8E4C60;
				this->CStreaming__ms_pEndLoadedList = (void**)0x8E4C5C;
			}
			#elif defined(IS_PLATFORM_WIN_X64)
			if (gameVersion == GAME_VERSION_GTA_TRILOGY_SA_1_0_0_14718_WIN_X64)
			{
				// Variables
				this->CModelInfo__ms_modelInfoPtrs.gta_sa = (CBaseModelInfo**)g_mCalc.GetCurrentVAbyPreferedVA(0x14508E730);
			}
			#elif defined(IS_PLATFORM_ANDROID_ARM32)
			if (gameVersion == GAME_VERSION_GTA_SA_2_00_ANDROID_ARM32)
			{
				// Variables
				this->CModelInfo__ms_modelInfoPtrs.gta_sa = (CBaseModelInfo**)Library::GetSymbolAddress(
					&g_LimitAdjuster.hModule_of_game, "_ZN10CModelInfo16ms_modelInfoPtrsE"
				);

				this->CTxdStore__pTxdPool.gta_sa = (CPool_GTASA<TxdDef>**)Library::GetSymbolAddress(
					&g_LimitAdjuster.hModule_of_game, "_ZN9CTxdStore11ms_pTxdPoolE"
				);
				this->CStreaming__ms_aInfoForModel.gta_sa = (CStreamingInfo*)Library::GetSymbolAddress(
					&g_LimitAdjuster.hModule_of_game, "_ZN10CStreaming16ms_aInfoForModelE"
				);
				this->CAnimManager__ms_aAnimBlocks.gta_sa = (CAnimBlock*)Library::GetSymbolAddress(
					&g_LimitAdjuster.hModule_of_game, "_ZN12CAnimManager14ms_aAnimBlocksE"
				);
				this->CVehicleRecording__StreamingArray.gta_sa = (CVehicleRecordingFileInfo*)Library::GetSymbolAddress(
					&g_LimitAdjuster.hModule_of_game, "_ZN17CVehicleRecording14StreamingArrayE"
				);
				this->CTheScripts__StreamedScripts.ptr = (char*)Library::GetSymbolAddress(
					&g_LimitAdjuster.hModule_of_game, "_ZN11CTheScripts15StreamedScriptsE"
				);
				this->CIplStore__ms_pPool.gta_sa = (CPool_GTASA<CIplFile>**)Library::GetSymbolAddress(
					&g_LimitAdjuster.hModule_of_game, "_ZN9CIplStore8ms_pPoolE"
				);

				this->CColStore__ms_pColPool.gta_sa = (CPool_GTASA<Game_GTASA::CCollisionFile>**)Library::GetSymbolAddress(
					&g_LimitAdjuster.hModule_of_game, "_ZN9CColStore11ms_pColPoolE"
				);

				// Registered kills
				this->CDarkel__RegisteredKills.ptr = (int16_t*)Library::GetSymbolAddress(
					&g_LimitAdjuster.hModule_of_game, "_ZN7CDarkel15RegisteredKillsE"
				);

				// Lists
				this->CStreaming__ms_pStartRequestedList = (void**)Library::GetSymbolAddress(
					&g_LimitAdjuster.hModule_of_game, "_ZN10CStreaming22ms_pStartRequestedListE"
				);
				this->CStreaming__ms_pEndRequestedList = (void**)(void**)Library::GetSymbolAddress(
					&g_LimitAdjuster.hModule_of_game, "_ZN10CStreaming20ms_pEndRequestedListE"
				);
				this->CStreaming__ms_pStartLoadedList = (void**)(void**)Library::GetSymbolAddress(
					&g_LimitAdjuster.hModule_of_game, "_ZN10CStreaming19ms_pStartLoadedListE"
				);
				this->CStreaming__ms_pEndLoadedList = (void**)(void**)Library::GetSymbolAddress(
					&g_LimitAdjuster.hModule_of_game, "_ZN10CStreaming17ms_pEndLoadedListE"
				);
			}
			#endif
		}
	}
	 
	else if (gameVersion == GAME_VERSION_GTA_SA_1_0_PS2)
	{
		this->RegisterFileType(FILE_TYPE_DFF, 15080, false);		// base ID		0
		this->RegisterFileType(FILE_TYPE_TXD, 3610, false);		// base ID		15080
		this->RegisterFileType(FILE_TYPE_COL, 255, false);			// base ID		18690
		this->RegisterFileType(FILE_TYPE_IPL, 256, false);			// base ID		18945
		this->RegisterFileType(FILE_TYPE_DAT, 64, false);			// base ID		19201
		this->RegisterFileType(FILE_TYPE_IFP, 180, false);			// base ID		19265
		this->RegisterFileType(FILE_TYPE_RRR, 475, false);			// base ID		19445
		this->RegisterFileType(FILE_TYPE_SCM, 82, false);			// base ID		19920

		// Loaded
		this->RegisterFileType(FILE_TYPE_LOADED_START, 2, false);	// base ID		20002

		// Requested
		this->RegisterFileType(FILE_TYPE_REQUESTED_START, 2, false);// base ID		20004
	}
	else if (gameVersion == GAME_VERSION_GTA_LCS_2_4_ANDROID_ARM32)
	{
		// DFF
		this->RegisterFileType(FILE_TYPE_DFF, 4900, false);	// base ID		0

		// TXD / MDL / TEX
		this->RegisterFileType(FILE_TYPE_TXD, 1200, false);		// base ID		4900

		// COL / COL2
		this->RegisterFileType(FILE_TYPE_COL, 15, false);		// base ID		6100

		// IFP / ANIM
		this->RegisterFileType(FILE_TYPE_IFP, 60, false);		// base ID		6115
	}
	else if (gameVersion == GAME_VERSION_GTA_VCS_1_0_PS2)
	{

	}
	else if (CGameVersion::IsAny_GTA_IV_or_EFLC(gameVersion))
	{
		using namespace Game_GTAIV;

		// MODEL
		if (gameVersion == GAME_VERSION_GTA_IV_1_0_7_0_WIN_X86
			|| gameVersion == GAME_VERSION_GTA_IV_1_0_8_0_WIN_X86
			|| gameVersion == GAME_VERSION_GTA_EFLC_1_1_2_0_WIN_X86)
			maxLimitValue = SHRT_MAX;
		else
			maxLimitValue = false;

		this->RegisterFileType(FILE_TYPE_MODEL, 31000, maxLimitValue);		// MODEL (WDR / WFT / LOD / NTH)

		// WTD
		if (gameVersion == GAME_VERSION_GTA_IV_1_0_7_0_WIN_X86
			|| gameVersion == GAME_VERSION_GTA_IV_1_0_8_0_WIN_X86
			|| gameVersion == GAME_VERSION_GTA_EFLC_1_1_2_0_WIN_X86)
			maxLimitValue = SHRT_MAX + 1;
		else
			maxLimitValue = false;

		this->RegisterFileType(FILE_TYPE_WTD, 6000, maxLimitValue);		// WTD

		// MAP_PLACEMENT
		if (gameVersion == GAME_VERSION_GTA_IV_1_0_7_0_WIN_X86
			|| gameVersion == GAME_VERSION_GTA_IV_1_0_8_0_WIN_X86
			|| gameVersion == GAME_VERSION_GTA_EFLC_1_1_2_0_WIN_X86)
			maxLimitValue = SHRT_MAX + 1;
		else
			maxLimitValue = false;

		this->RegisterFileType(FILE_TYPE_MAP_PLACEMENT, 500, maxLimitValue);		// MAP_PLACEMENT (IPL / WPL)

		// WAD
		if (gameVersion == GAME_VERSION_GTA_IV_1_0_7_0_WIN_X86
			|| gameVersion == GAME_VERSION_GTA_IV_1_0_8_0_WIN_X86
			|| gameVersion == GAME_VERSION_GTA_EFLC_1_1_2_0_WIN_X86)
			maxLimitValue = SHRT_MAX + 1;
		else
			maxLimitValue = false;

		this->RegisterFileType(FILE_TYPE_WAD, 1500, maxLimitValue);		// WAD

		// RRR
		if (gameVersion == GAME_VERSION_GTA_IV_1_0_7_0_WIN_X86
			|| gameVersion == GAME_VERSION_GTA_IV_1_0_8_0_WIN_X86
			|| gameVersion == GAME_VERSION_GTA_EFLC_1_1_2_0_WIN_X86)
			maxLimitValue = SHRT_MAX + 1;
		else
			maxLimitValue = false;

		this->RegisterFileType(FILE_TYPE_RRR, 2072, maxLimitValue);		// RRR

		// Memory addresses
		if (g_LimitAdjuster.IsGameVersionSetUpWithMemory())
		{
			#ifdef IS_PLATFORM_WIN_X86
			if (gameVersion == GAME_VERSION_GTA_IV_1_0_4_0_WIN_X86)
			{
				this->CModelInfo__ms_modelInfoPtrs.gta_iv = (CBaseModelInfo**)g_mCalc.GetCurrentVAbyPreferedVA(0x10C90F8);
			}
			else if (gameVersion == GAME_VERSION_GTA_IV_1_0_7_0_WIN_X86)
			{
				this->CModelInfo__ms_modelInfoPtrs.gta_iv = (CBaseModelInfo**)g_mCalc.GetCurrentVAbyPreferedVA(0x15F73B0);
				this->CTxdStore__pTxdPool.gta_iv = (CPool_GTAIV<Game_GTAIV::TxdDef>**)g_mCalc.GetCurrentVAbyPreferedVA(0x11D52D4);
				this->CVehicleRecording__StreamingArray.gta_iv = (CVehicleRecordingFileInfo*)g_mCalc.GetCurrentVAbyPreferedVA(0x15DB170);				
			}
			else if (gameVersion == GAME_VERSION_GTA_IV_1_0_8_0_WIN_X86)
			{
				this->CModelInfo__ms_modelInfoPtrs.gta_iv = (CBaseModelInfo**)g_mCalc.GetCurrentVAbyPreferedVA(0x122C168);
				this->CTxdStore__pTxdPool.gta_iv = (CPool_GTAIV<Game_GTAIV::TxdDef>**)g_mCalc.GetCurrentVAbyPreferedVA(0x12291E0);
				this->CVehicleRecording__StreamingArray.gta_iv = (CVehicleRecordingFileInfo*)g_mCalc.GetCurrentVAbyPreferedVA(0x12CF970);				
			}
			else if (gameVersion == GAME_VERSION_GTA_EFLC_1_1_2_0_WIN_X86)
			{
				this->CModelInfo__ms_modelInfoPtrs.gta_iv = (CBaseModelInfo**)g_mCalc.GetCurrentVAbyPreferedVA(0x157D310);
				this->CTxdStore__pTxdPool.gta_iv = (CPool_GTAIV<Game_GTAIV::TxdDef>**)g_mCalc.GetCurrentVAbyPreferedVA(0x10D5BE8);
				this->CVehicleRecording__StreamingArray.gta_iv = (CVehicleRecordingFileInfo*)g_mCalc.GetCurrentVAbyPreferedVA(0x11B91F0);				
			}
			else if (gameVersion == GAME_VERSION_GTA_EFLC_1_1_3_0_WIN_X86)
			{
				this->CModelInfo__ms_modelInfoPtrs.gta_iv = (CBaseModelInfo**)g_mCalc.GetCurrentVAbyPreferedVA(0x1199ED8);
			}
			#endif
		}
	}

	else if (gameVersion == GAME_VERSION_BULLY_SE_1_2_WIN_X86)
	{
		using namespace Game_BullyScholarschipEdition;

		this->RegisterFileType(FILE_TYPE_NIF, 12200, SHRT_MAX);	// base ID 0
		this->RegisterFileType(FILE_TYPE_NFT, 10000, SHRT_MAX + 1);	// base ID 12200
		this->RegisterFileType(FILE_TYPE_COL, 500, SHRT_MAX + 1);		// base ID 22200
		this->RegisterFileType(FILE_TYPE_AGR, 580, SHRT_MAX + 1);		// base ID 22700
		this->RegisterFileType(FILE_TYPE_IPB, 100, SHRT_MAX + 1);		// base ID 23280
		this->RegisterFileType(FILE_TYPE_LIP, 493, false);		// base ID 23380
		this->RegisterFileType(FILE_TYPE_CAT, 120, SHRT_MAX + 1); 	// base ID 23873
		this->RegisterFileType(FILE_TYPE_LUR, 75, false);		// base ID 23993

		this->CStreaming__ms_aInfoForModel.bully_se = (CStreamingInfo*)0xCA39B8;
		this->NumberOfStreamingInfos = 17204;

		#ifdef IS_PLATFORM_WIN_X86
		this->CModelInfo__ms_modelInfoPtrs.bully_se = (CBaseModelInfo**)0xC67738;
		this->CSimpleModelInfo__m_UpdatedUVATable.ptr = (uint32_t*)0xC9DD58;
		this->CStreaming__ms_apAGRBuffers.bully_se = (void**)0xCA2D60;
		#endif
	}

	this->UpdateBaseIDs();

	/////////
	this->CModelInfo__ms_modelInfoExtensionPtrs = NULL;
	this->CStreaming__ms_aInfoForModelExtension = NULL;
	this->CColStore__ms_aColPoolExtension = NULL;

	this->bIsModelInfoExtensionArrayEnabled = false;
	this->bIsFileInfoExtensionArrayEnabled = false;
	this->bIsCollisionFileExtensionArrayEnabled = false;
	this->bIsIDunsigned = false;
	this->bIsIDbaseInt32 = false;
	this->bAreDifficultFileTypesInt32 = false;
	this->bIsCOLIDsizeIncreased = false;
	this->bIsIPLIDsizeIncreased = false;
	this->bIsPatchForLoadingSaveBlocksApplied = false;
}

// Shutdown
void FileIDlimit::Shutdown()
{
	MAKE_VAR_GAME_VERSION();
		
	delete[] CStreaming__ms_aInfoForModelExtension;
	delete[] CModelInfo__ms_modelInfoExtensionPtrs;
	delete[] CColStore__ms_aColPoolExtension;

	if (CModelInfo__ms_modelInfoPtrs.bIsAllocated.Get())
		delete CModelInfo__ms_modelInfoPtrs.generic;

	if (CSimpleModelInfo__m_UpdatedUVATable.bIsAllocated.Get())
		delete[] CSimpleModelInfo__m_UpdatedUVATable.ptr;
	
	if (CStreaming__ms_aInfoForModel.bIsAllocated.Get())
		delete CStreaming__ms_aInfoForModel.generic;

	if (CAnimManager__ms_aAnimBlocks.bIsAllocated.Get())
		delete CAnimManager__ms_aAnimBlocks.generic;

	if (CVehicleRecording__StreamingArray.bIsAllocated.Get())
		delete CVehicleRecording__StreamingArray.generic;
	
	if (CStreaming__ms_apAGRBuffers.bIsAllocated.Get())
		delete CStreaming__ms_apAGRBuffers.generic;

	if (CTheScripts__StreamedScripts.bIsAllocated.Get())
		delete CTheScripts__StreamedScripts.ptr;

	if (CDarkel__RegisteredKills.bIsAllocated.Get())
		delete[] CDarkel__RegisteredKills.ptr;
}

// Initialize
void CStreamingInfoExtension::Init()
{
	this->dwSize = sizeof(CStreamingInfoExtension);
	this->exUsNext = -1;
	this->exUsPrev = -1;
	this->exNextOnCd = -1;

	this->m_NumberOfPaddedBytesInAlignedOriginalSize = 0;
	this->m_PackedSizeInSectors = 0;
	this->m_exFlags = 0;

	#ifdef STORE_FILE_NAME_FIELD_IN_CSTREAMINGINFO_EXTENSION
	this->m_Filename[0] = 0;
	#endif
}